







open Ctypes
open Foreign

let so_ext = if Sys.win32 || Sys.cygwin then "dll" else "so"
let c_lib_name = Format.sprintf "libadalang.%s" so_ext
let c_lib = Dl.dlopen ~filename:c_lib_name ~flags:[Dl.RTLD_NOW]

exception BadTypeError of string

exception OutOfBoundsError of string

exception InvalidInput of string

exception InvalidSymbolError of string

exception InvalidUnitNameError of string

exception NativeException of string

exception PreconditionFailure of string

exception PropertyError of string

exception TemplateArgsError of string

exception TemplateFormatError of string

exception TemplateInstantiationError of string

exception StaleReferenceError of string

exception UnknownCharset of string



      (**
 * Raised when an error occurs while loading a project file.
 *)
exception InvalidProjectError of string



exception SyntaxError

module Exception = struct

  type t = {
    kind : int;
    information : string;
  }

  let c_struct : t structure typ = structure "exception"
  let kind = field c_struct "kind" int
  let information = field c_struct "information" string
  let () = seal c_struct

  let wrap c_value_ptr =
    if is_null c_value_ptr then
      None
    else
      let c_value = !@ c_value_ptr in
      Some {
        kind = getf c_value kind;
        information = getf c_value information;
      }

  let unwrap value =
    match value with
    | None ->
        from_voidp c_struct null
    | Some value ->
        let c_value = make c_struct in
        setf c_value kind value.kind;
        setf c_value information value.information;
        allocate c_struct c_value

  let c_type = view (ptr c_struct) ~read:wrap ~write:unwrap

end

let get_last_exception = foreign ~from:c_lib
  "ada_get_last_exception"
  (void @-> returning Exception.c_type)

(* When declaring an imported function with foreign, use raisable instead of
 returning, to check get_last_exception before returning *)
let raisable typ =
  let read value =
    match get_last_exception () with
    | None -> value
    | Some exc ->
        (match exc.kind with
         | 0 ->
             raise (BadTypeError exc.information)
         | 1 ->
             raise (OutOfBoundsError exc.information)
         | 2 ->
             raise (InvalidInput exc.information)
         | 3 ->
             raise (InvalidSymbolError exc.information)
         | 4 ->
             raise (InvalidUnitNameError exc.information)
         | 5 ->
             raise (NativeException exc.information)
         | 6 ->
             raise (PreconditionFailure exc.information)
         | 7 ->
             raise (PropertyError exc.information)
         | 8 ->
             raise (TemplateArgsError exc.information)
         | 9 ->
             raise (TemplateFormatError exc.information)
         | 10 ->
             raise (TemplateInstantiationError exc.information)
         | 11 ->
             raise (StaleReferenceError exc.information)
         | 12 ->
             raise (UnknownCharset exc.information)
         | _ -> assert false)
  in
  let write value = value in
  let new_typ = view typ ~read ~write in
  returning new_typ

(* Module used to encode/decode UTF32 strings *)

(* Camomile needs to know the location of its standard library to work,
   so we use the following heuristic:
   - if the directory chosen at build time exists, we assume the installation
     is ok
   - otherwise we look for a directory 'share/camomile' next to the binary
   - otherwise we fail
*)

module type CamomileConfig = module type of CamomileLibrary.DefaultConfig

module CamomileDefaultConfig : CamomileConfig = CamomileLibrary.DefaultConfig

let ( ^/ ) = Filename.concat

let build_camomile_config root_path = (module struct
  let share_dir = root_path ^/ "share" ^/ "camomile"

  let datadir = share_dir ^/ "database"

  let localedir = share_dir ^/ "locales"

  let charmapdir = share_dir ^/ "charmaps"

  let unimapdir = share_dir ^/ "mappings"

  end : CamomileConfig)

module CamomileShareConfig =
  (val build_camomile_config
    (Filename.dirname Sys.executable_name ^/ Filename.parent_dir_name)
    : CamomileConfig)

(* In case we are building through an opam-installed env, find
   Camomile's stdlib through the appropriate opam env variable *)
module CamomileOpamConfig =
  (val
    let opam_dir = try Sys.getenv "OPAM_SWITCH_PREFIX" with _ -> "DUMMY" in
    build_camomile_config opam_dir : CamomileConfig)

let camomile_config =
  if Sys.file_exists CamomileDefaultConfig.datadir then
    (module CamomileDefaultConfig : CamomileConfig )
  else if Sys.file_exists CamomileShareConfig.datadir then
    (module CamomileShareConfig : CamomileConfig )
  else if Sys.file_exists CamomileOpamConfig.datadir then
    (module CamomileOpamConfig : CamomileConfig)
  else failwith "no camomile library found"

module CamomileConfig = (val camomile_config)

module Camomile = CamomileLibrary.Make (CamomileConfig)

module Text = struct
  type t = string

  let c_struct : t structure typ = structure "text"

  let chars = field c_struct "chars" (ptr uint32_t)

  let length = field c_struct "length" size_t

  let is_allocated = field c_struct "is_allocated" bool

  let () = seal c_struct

  let destroy_text = foreign ~from:c_lib "ada_destroy_text"
    (ptr c_struct @-> raisable void)

  module UCS4Encoding = Camomile.CharEncoding.Make (Camomile.UCS4)

  let wrap (c_value : t structure) : t =
    let open Unsigned.Size_t in
    let open Camomile in
    let length = to_int (getf c_value length) in
    let chars = getf c_value chars in
    let f i =
      UChar.chr_of_uint (Unsigned.UInt32.to_int !@ (chars +@ i))
    in
    let result = UCS4.init length f in
    (* Now that the value is fully transformed to an ocaml value, we can
      free it by calling destroy_text *)
    destroy_text (addr c_value) ;
    UCS4Encoding.encode CharEncoding.utf8 result

  let unwrap (value : t) : t structure =
    let open Unsigned in
    let open Camomile in
    let text = UCS4Encoding.decode CharEncoding.utf8 value in
    let struct_length = Size_t.of_int (UCS4.length text) in
    let struct_chars = allocate_n uint32_t ~count:(UCS4.length text) in
    let i = ref 0 in
    let f c =
      struct_chars +@ !i <-@ (UInt32.of_int (UChar.code c));
      i := !i + 1
    in
    UCS4.iter f text ;
    let c_value = make c_struct in
    setf c_value length struct_length ;
    setf c_value chars struct_chars ;
    setf c_value is_allocated false ;
    (* We don't need to care about calling destroy_text here since we
     manually allocated the pointer, ctypes will take care of freeing the
     memory *)
    c_value

  let c_type = view c_struct ~read:wrap ~write:unwrap
end

module Character = struct
  (* Characters are encoded as strings because ocaml char are not unicode
   characters *)
  type t = string

  module UCharEncoding = Camomile.CharEncoding.Make (Camomile.UText)

  let of_int i =
    let open Camomile in
    let uchar = UChar.chr i in
    UCharEncoding.encode CharEncoding.utf8 (UText.init 1 (fun _ -> uchar))

  let of_int32 i =
    of_int (Unsigned.UInt32.to_int i)

  let wrap (c_value : Unsigned.UInt32.t) : t =
    of_int32 c_value

  let unwrap (value : string) : Unsigned.UInt32.t =
    let open Camomile in
    let text = UCharEncoding.decode CharEncoding.utf8 value in
    let uchar = UText.get text 0 in
    Unsigned.UInt32.of_int (UChar.code uchar)

  let c_type = view uint32_t ~read:wrap ~write:unwrap
end

module BigInteger = struct

  type t = Z.t

  let c_type = ptr void

  let create = foreign ~from:c_lib "ada_create_big_integer"
    (ptr Text.c_type @-> raisable c_type)

  let text = foreign ~from:c_lib "ada_big_integer_text"
    (c_type @-> ptr Text.c_type @-> raisable void)

  let decref = foreign ~from:c_lib "ada_big_integer_decref"
    (c_type @-> raisable void)

  let wrap (c_value : unit ptr) : t =
    let c_text_ptr = allocate_n Text.c_type ~count:1 in
    text c_value c_text_ptr;
    decref c_value;
    Z.of_string (!@ c_text_ptr)

  let unwrap (value : t) : unit ptr =
    create (allocate Text.c_type (Z.to_string value))
end

module AnalysisUnitKind = struct
  type t =
  | UnitSpecification
  | UnitBody

  let name () = "AnalysisUnitKind"

  let wrap (c_value : int) : t =
    match c_value with
    | 0 -> UnitSpecification
    | 1 -> UnitBody
    | _ -> assert false

  let unwrap (value : t) : int =
    match value with
    | UnitSpecification -> 0
    | UnitBody -> 1

   let c_type = view int ~read:wrap ~write:unwrap
end

module LookupKind = struct
  type t =
  | Recursive
  | Flat
  | Minimal

  let name () = "LookupKind"

  let wrap (c_value : int) : t =
    match c_value with
    | 0 -> Recursive
    | 1 -> Flat
    | 2 -> Minimal
    | _ -> assert false

  let unwrap (value : t) : int =
    match value with
    | Recursive -> 0
    | Flat -> 1
    | Minimal -> 2

   let c_type = view int ~read:wrap ~write:unwrap
end

module FindAllMode = struct
  type t =
  | References
  | Derivedtypes

  let name () = "FindAllMode"

  let wrap (c_value : int) : t =
    match c_value with
    | 0 -> References
    | 1 -> Derivedtypes
    | _ -> assert false

  let unwrap (value : t) : int =
    match value with
    | References -> 0
    | Derivedtypes -> 1

   let c_type = view int ~read:wrap ~write:unwrap
end

module RefResultKind = struct
  type t =
  | Noref
  | Precise
  | Imprecise
  | Error

  let name () = "RefResultKind"

  let wrap (c_value : int) : t =
    match c_value with
    | 0 -> Noref
    | 1 -> Precise
    | 2 -> Imprecise
    | 3 -> Error
    | _ -> assert false

  let unwrap (value : t) : int =
    match value with
    | Noref -> 0
    | Precise -> 1
    | Imprecise -> 2
    | Error -> 3

   let c_type = view int ~read:wrap ~write:unwrap
end

module GrammarRule = struct
  type t =
  | ParentListRule
  | ProtectedTypeDeclRule
  | ProtectedOpRule
  | ProtectedElRule
  | ProtectedDefRule
  | ProtectedDeclRule
  | TaskItemRule
  | TaskDefRule
  | TaskTypeDeclRule
  | SubtypeDeclRule
  | InterfaceTypeDefRule
  | UnconstrainedIndexRule
  | ArrayTypeDefRule
  | DiscreteSubtypeDefinitionRule
  | ConstraintListRule
  | SignedIntTypeDefRule
  | ModIntTypeDefRule
  | DerivedTypeDefRule
  | DiscriminantAssocRule
  | DiscriminantConstraintRule
  | IndexConstraintRule
  | DigitsConstraintRule
  | DeltaConstraintRule
  | RangeConstraintRule
  | ConstraintRule
  | DiscriminantSpecRule
  | DiscrSpecListRule
  | DiscriminantPartRule
  | EnumLiteralDeclRule
  | FormalDiscreteTypeDefRule
  | RecordDefRule
  | RangeSpecRule
  | RealTypeDefRule
  | SexprOrBoxRule
  | OrdinaryFixedPointDefRule
  | DecimalFixedPointDefRule
  | FloatingPointDefRule
  | RecordTypeDefRule
  | AccessDefRule
  | EnumTypeDefRule
  | TypeDefRule
  | VariantRule
  | AnonymousTypeDeclRule
  | TypeDeclRule
  | VariantPartRule
  | ComponentDefRule
  | ComponentItemRule
  | ComponentDeclRule
  | ComponentListRule
  | GenericDeclRule
  | GenericFormalPartRule
  | GenericFormalDeclRule
  | FormalSubpDeclRule
  | RenamingClauseRule
  | GenericRenamingDeclRule
  | GenericInstantiationRule
  | ExceptionDeclRule
  | BasicDeclsRule
  | PackageRenamingDeclRule
  | PackageDeclRule
  | BasicDeclRule
  | ObjectDeclRule
  | SubObjectDeclRule
  | ExtRetStmtObjectDeclRule
  | DefiningIdListRule
  | NumberDeclRule
  | ContractCaseAssocRule
  | ContractCasesExprRule
  | AbstractStateDeclRule
  | MultiAbstractStateDeclRule
  | AspectAssocRule
  | AspectSpecRule
  | SingleTaskDeclRule
  | OverridingIndicatorRule
  | EntryDeclRule
  | ComponentClauseRule
  | AspectClauseRule
  | ParamSpecRule
  | ParamSpecsRule
  | SubpSpecRule
  | ExprFnRule
  | NullSubpDeclRule
  | AbstractSubpDeclRule
  | SubpRenamingDeclRule
  | SimpleSubpDeclRule
  | SubpDeclRule
  | WithClauseRule
  | ContextItemRule
  | UseClauseRule
  | UsePackageClauseRule
  | UseTypeClauseRule
  | SubtypeIndicationRule
  | DiscreteSubtypeIndicationRule
  | ConstrainedSubtypeIndicationRule
  | TypeExprRule
  | AnonymousTypeRule
  | ModeRule
  | PragmaArgumentRule
  | PragmaRule
  | SubunitRule
  | LibraryUnitBodyRule
  | LibraryUnitRenamingDeclRule
  | LibraryItemRule
  | CompilationUnitRule
  | CompilationRule
  | DeclPartRule
  | EntryBodyRule
  | ProtectedBodyRule
  | ProtectedBodyStubRule
  | TaskBodyRule
  | TaskBodyStubRule
  | PackageBodyStubRule
  | PackageBodyRule
  | TerminateAlternativeRule
  | SelectStmtRule
  | AcceptStmtRule
  | CaseAltRule
  | CaseStmtRule
  | ExtReturnStmtRule
  | IblockStmtRule
  | BlockStmtRule
  | IloopStmtRule
  | LoopStmtRule
  | CompoundStmtRule
  | IfStmtRule
  | RaiseStmtRule
  | DelayStmtRule
  | AbortStmtRule
  | BodyRule
  | BodyStubRule
  | SubpBodyStubRule
  | RecovDeclPartRule
  | SubpBodyRule
  | HandledStmtsRule
  | ExceptionHandlerRule
  | StmtsRule
  | LabelRule
  | StmtRule
  | CallStmtRule
  | SimpleStmtRule
  | NullStmtRule
  | AssignmentStmtRule
  | GotoStmtRule
  | ExitStmtRule
  | ReturnStmtRule
  | RequeueStmtRule
  | IdentifierRule
  | CharLiteralRule
  | StringLiteralRule
  | DefiningIdRule
  | DecLiteralRule
  | IntLiteralRule
  | NumLiteralRule
  | NullLiteralRule
  | AllocatorRule
  | ForLoopParamSpecRule
  | QuantifiedExprRule
  | CaseExprRule
  | CaseExprAltRule
  | RaiseExprRule
  | IfExprRule
  | ConditionalExprRule
  | BoxExprRule
  | OthersDesignatorRule
  | AggregateAssocRule
  | RegularAggregateRule
  | BracketAggregateRule
  | AggregateRule
  | DirectNameRule
  | ParamAssocRule
  | CallSuffixRule
  | QualifiedNameRule
  | QualNameInternalRule
  | NameRule
  | DefiningNameRule
  | DirectNameOrTargetNameRule
  | TargetNameRule
  | UpdateAttrAggregateRule
  | UpdateAttrContentRule
  | MultidimArrayAssocRule
  | SubtypeNameRule
  | StaticNameRule
  | PrimaryRule
  | ParenExprRule
  | DeclareExprRule
  | FactorRule
  | TermRule
  | UnopTermRule
  | SimpleExprRule
  | BooleanOpRule
  | DiscreteRangeRule
  | ChoiceRule
  | ChoiceListRule
  | RelOpRule
  | MembershipChoiceRule
  | MembershipChoiceListRule
  | RelationRule
  | ExprRule

  let name () = "GrammarRule"

  let wrap (c_value : int) : t =
    match c_value with
    | 0 -> ParentListRule
    | 1 -> ProtectedTypeDeclRule
    | 2 -> ProtectedOpRule
    | 3 -> ProtectedElRule
    | 4 -> ProtectedDefRule
    | 5 -> ProtectedDeclRule
    | 6 -> TaskItemRule
    | 7 -> TaskDefRule
    | 8 -> TaskTypeDeclRule
    | 9 -> SubtypeDeclRule
    | 10 -> InterfaceTypeDefRule
    | 11 -> UnconstrainedIndexRule
    | 12 -> ArrayTypeDefRule
    | 13 -> DiscreteSubtypeDefinitionRule
    | 14 -> ConstraintListRule
    | 15 -> SignedIntTypeDefRule
    | 16 -> ModIntTypeDefRule
    | 17 -> DerivedTypeDefRule
    | 18 -> DiscriminantAssocRule
    | 19 -> DiscriminantConstraintRule
    | 20 -> IndexConstraintRule
    | 21 -> DigitsConstraintRule
    | 22 -> DeltaConstraintRule
    | 23 -> RangeConstraintRule
    | 24 -> ConstraintRule
    | 25 -> DiscriminantSpecRule
    | 26 -> DiscrSpecListRule
    | 27 -> DiscriminantPartRule
    | 28 -> EnumLiteralDeclRule
    | 29 -> FormalDiscreteTypeDefRule
    | 30 -> RecordDefRule
    | 31 -> RangeSpecRule
    | 32 -> RealTypeDefRule
    | 33 -> SexprOrBoxRule
    | 34 -> OrdinaryFixedPointDefRule
    | 35 -> DecimalFixedPointDefRule
    | 36 -> FloatingPointDefRule
    | 37 -> RecordTypeDefRule
    | 38 -> AccessDefRule
    | 39 -> EnumTypeDefRule
    | 40 -> TypeDefRule
    | 41 -> VariantRule
    | 42 -> AnonymousTypeDeclRule
    | 43 -> TypeDeclRule
    | 44 -> VariantPartRule
    | 45 -> ComponentDefRule
    | 46 -> ComponentItemRule
    | 47 -> ComponentDeclRule
    | 48 -> ComponentListRule
    | 49 -> GenericDeclRule
    | 50 -> GenericFormalPartRule
    | 51 -> GenericFormalDeclRule
    | 52 -> FormalSubpDeclRule
    | 53 -> RenamingClauseRule
    | 54 -> GenericRenamingDeclRule
    | 55 -> GenericInstantiationRule
    | 56 -> ExceptionDeclRule
    | 57 -> BasicDeclsRule
    | 58 -> PackageRenamingDeclRule
    | 59 -> PackageDeclRule
    | 60 -> BasicDeclRule
    | 61 -> ObjectDeclRule
    | 62 -> SubObjectDeclRule
    | 63 -> ExtRetStmtObjectDeclRule
    | 64 -> DefiningIdListRule
    | 65 -> NumberDeclRule
    | 66 -> ContractCaseAssocRule
    | 67 -> ContractCasesExprRule
    | 68 -> AbstractStateDeclRule
    | 69 -> MultiAbstractStateDeclRule
    | 70 -> AspectAssocRule
    | 71 -> AspectSpecRule
    | 72 -> SingleTaskDeclRule
    | 73 -> OverridingIndicatorRule
    | 74 -> EntryDeclRule
    | 75 -> ComponentClauseRule
    | 76 -> AspectClauseRule
    | 77 -> ParamSpecRule
    | 78 -> ParamSpecsRule
    | 79 -> SubpSpecRule
    | 80 -> ExprFnRule
    | 81 -> NullSubpDeclRule
    | 82 -> AbstractSubpDeclRule
    | 83 -> SubpRenamingDeclRule
    | 84 -> SimpleSubpDeclRule
    | 85 -> SubpDeclRule
    | 86 -> WithClauseRule
    | 87 -> ContextItemRule
    | 88 -> UseClauseRule
    | 89 -> UsePackageClauseRule
    | 90 -> UseTypeClauseRule
    | 91 -> SubtypeIndicationRule
    | 92 -> DiscreteSubtypeIndicationRule
    | 93 -> ConstrainedSubtypeIndicationRule
    | 94 -> TypeExprRule
    | 95 -> AnonymousTypeRule
    | 96 -> ModeRule
    | 97 -> PragmaArgumentRule
    | 98 -> PragmaRule
    | 99 -> SubunitRule
    | 100 -> LibraryUnitBodyRule
    | 101 -> LibraryUnitRenamingDeclRule
    | 102 -> LibraryItemRule
    | 103 -> CompilationUnitRule
    | 104 -> CompilationRule
    | 105 -> DeclPartRule
    | 106 -> EntryBodyRule
    | 107 -> ProtectedBodyRule
    | 108 -> ProtectedBodyStubRule
    | 109 -> TaskBodyRule
    | 110 -> TaskBodyStubRule
    | 111 -> PackageBodyStubRule
    | 112 -> PackageBodyRule
    | 113 -> TerminateAlternativeRule
    | 114 -> SelectStmtRule
    | 115 -> AcceptStmtRule
    | 116 -> CaseAltRule
    | 117 -> CaseStmtRule
    | 118 -> ExtReturnStmtRule
    | 119 -> IblockStmtRule
    | 120 -> BlockStmtRule
    | 121 -> IloopStmtRule
    | 122 -> LoopStmtRule
    | 123 -> CompoundStmtRule
    | 124 -> IfStmtRule
    | 125 -> RaiseStmtRule
    | 126 -> DelayStmtRule
    | 127 -> AbortStmtRule
    | 128 -> BodyRule
    | 129 -> BodyStubRule
    | 130 -> SubpBodyStubRule
    | 131 -> RecovDeclPartRule
    | 132 -> SubpBodyRule
    | 133 -> HandledStmtsRule
    | 134 -> ExceptionHandlerRule
    | 135 -> StmtsRule
    | 136 -> LabelRule
    | 137 -> StmtRule
    | 138 -> CallStmtRule
    | 139 -> SimpleStmtRule
    | 140 -> NullStmtRule
    | 141 -> AssignmentStmtRule
    | 142 -> GotoStmtRule
    | 143 -> ExitStmtRule
    | 144 -> ReturnStmtRule
    | 145 -> RequeueStmtRule
    | 146 -> IdentifierRule
    | 147 -> CharLiteralRule
    | 148 -> StringLiteralRule
    | 149 -> DefiningIdRule
    | 150 -> DecLiteralRule
    | 151 -> IntLiteralRule
    | 152 -> NumLiteralRule
    | 153 -> NullLiteralRule
    | 154 -> AllocatorRule
    | 155 -> ForLoopParamSpecRule
    | 156 -> QuantifiedExprRule
    | 157 -> CaseExprRule
    | 158 -> CaseExprAltRule
    | 159 -> RaiseExprRule
    | 160 -> IfExprRule
    | 161 -> ConditionalExprRule
    | 162 -> BoxExprRule
    | 163 -> OthersDesignatorRule
    | 164 -> AggregateAssocRule
    | 165 -> RegularAggregateRule
    | 166 -> BracketAggregateRule
    | 167 -> AggregateRule
    | 168 -> DirectNameRule
    | 169 -> ParamAssocRule
    | 170 -> CallSuffixRule
    | 171 -> QualifiedNameRule
    | 172 -> QualNameInternalRule
    | 173 -> NameRule
    | 174 -> DefiningNameRule
    | 175 -> DirectNameOrTargetNameRule
    | 176 -> TargetNameRule
    | 177 -> UpdateAttrAggregateRule
    | 178 -> UpdateAttrContentRule
    | 179 -> MultidimArrayAssocRule
    | 180 -> SubtypeNameRule
    | 181 -> StaticNameRule
    | 182 -> PrimaryRule
    | 183 -> ParenExprRule
    | 184 -> DeclareExprRule
    | 185 -> FactorRule
    | 186 -> TermRule
    | 187 -> UnopTermRule
    | 188 -> SimpleExprRule
    | 189 -> BooleanOpRule
    | 190 -> DiscreteRangeRule
    | 191 -> ChoiceRule
    | 192 -> ChoiceListRule
    | 193 -> RelOpRule
    | 194 -> MembershipChoiceRule
    | 195 -> MembershipChoiceListRule
    | 196 -> RelationRule
    | 197 -> ExprRule
    | _ -> assert false

  let unwrap (value : t) : int =
    match value with
    | ParentListRule -> 0
    | ProtectedTypeDeclRule -> 1
    | ProtectedOpRule -> 2
    | ProtectedElRule -> 3
    | ProtectedDefRule -> 4
    | ProtectedDeclRule -> 5
    | TaskItemRule -> 6
    | TaskDefRule -> 7
    | TaskTypeDeclRule -> 8
    | SubtypeDeclRule -> 9
    | InterfaceTypeDefRule -> 10
    | UnconstrainedIndexRule -> 11
    | ArrayTypeDefRule -> 12
    | DiscreteSubtypeDefinitionRule -> 13
    | ConstraintListRule -> 14
    | SignedIntTypeDefRule -> 15
    | ModIntTypeDefRule -> 16
    | DerivedTypeDefRule -> 17
    | DiscriminantAssocRule -> 18
    | DiscriminantConstraintRule -> 19
    | IndexConstraintRule -> 20
    | DigitsConstraintRule -> 21
    | DeltaConstraintRule -> 22
    | RangeConstraintRule -> 23
    | ConstraintRule -> 24
    | DiscriminantSpecRule -> 25
    | DiscrSpecListRule -> 26
    | DiscriminantPartRule -> 27
    | EnumLiteralDeclRule -> 28
    | FormalDiscreteTypeDefRule -> 29
    | RecordDefRule -> 30
    | RangeSpecRule -> 31
    | RealTypeDefRule -> 32
    | SexprOrBoxRule -> 33
    | OrdinaryFixedPointDefRule -> 34
    | DecimalFixedPointDefRule -> 35
    | FloatingPointDefRule -> 36
    | RecordTypeDefRule -> 37
    | AccessDefRule -> 38
    | EnumTypeDefRule -> 39
    | TypeDefRule -> 40
    | VariantRule -> 41
    | AnonymousTypeDeclRule -> 42
    | TypeDeclRule -> 43
    | VariantPartRule -> 44
    | ComponentDefRule -> 45
    | ComponentItemRule -> 46
    | ComponentDeclRule -> 47
    | ComponentListRule -> 48
    | GenericDeclRule -> 49
    | GenericFormalPartRule -> 50
    | GenericFormalDeclRule -> 51
    | FormalSubpDeclRule -> 52
    | RenamingClauseRule -> 53
    | GenericRenamingDeclRule -> 54
    | GenericInstantiationRule -> 55
    | ExceptionDeclRule -> 56
    | BasicDeclsRule -> 57
    | PackageRenamingDeclRule -> 58
    | PackageDeclRule -> 59
    | BasicDeclRule -> 60
    | ObjectDeclRule -> 61
    | SubObjectDeclRule -> 62
    | ExtRetStmtObjectDeclRule -> 63
    | DefiningIdListRule -> 64
    | NumberDeclRule -> 65
    | ContractCaseAssocRule -> 66
    | ContractCasesExprRule -> 67
    | AbstractStateDeclRule -> 68
    | MultiAbstractStateDeclRule -> 69
    | AspectAssocRule -> 70
    | AspectSpecRule -> 71
    | SingleTaskDeclRule -> 72
    | OverridingIndicatorRule -> 73
    | EntryDeclRule -> 74
    | ComponentClauseRule -> 75
    | AspectClauseRule -> 76
    | ParamSpecRule -> 77
    | ParamSpecsRule -> 78
    | SubpSpecRule -> 79
    | ExprFnRule -> 80
    | NullSubpDeclRule -> 81
    | AbstractSubpDeclRule -> 82
    | SubpRenamingDeclRule -> 83
    | SimpleSubpDeclRule -> 84
    | SubpDeclRule -> 85
    | WithClauseRule -> 86
    | ContextItemRule -> 87
    | UseClauseRule -> 88
    | UsePackageClauseRule -> 89
    | UseTypeClauseRule -> 90
    | SubtypeIndicationRule -> 91
    | DiscreteSubtypeIndicationRule -> 92
    | ConstrainedSubtypeIndicationRule -> 93
    | TypeExprRule -> 94
    | AnonymousTypeRule -> 95
    | ModeRule -> 96
    | PragmaArgumentRule -> 97
    | PragmaRule -> 98
    | SubunitRule -> 99
    | LibraryUnitBodyRule -> 100
    | LibraryUnitRenamingDeclRule -> 101
    | LibraryItemRule -> 102
    | CompilationUnitRule -> 103
    | CompilationRule -> 104
    | DeclPartRule -> 105
    | EntryBodyRule -> 106
    | ProtectedBodyRule -> 107
    | ProtectedBodyStubRule -> 108
    | TaskBodyRule -> 109
    | TaskBodyStubRule -> 110
    | PackageBodyStubRule -> 111
    | PackageBodyRule -> 112
    | TerminateAlternativeRule -> 113
    | SelectStmtRule -> 114
    | AcceptStmtRule -> 115
    | CaseAltRule -> 116
    | CaseStmtRule -> 117
    | ExtReturnStmtRule -> 118
    | IblockStmtRule -> 119
    | BlockStmtRule -> 120
    | IloopStmtRule -> 121
    | LoopStmtRule -> 122
    | CompoundStmtRule -> 123
    | IfStmtRule -> 124
    | RaiseStmtRule -> 125
    | DelayStmtRule -> 126
    | AbortStmtRule -> 127
    | BodyRule -> 128
    | BodyStubRule -> 129
    | SubpBodyStubRule -> 130
    | RecovDeclPartRule -> 131
    | SubpBodyRule -> 132
    | HandledStmtsRule -> 133
    | ExceptionHandlerRule -> 134
    | StmtsRule -> 135
    | LabelRule -> 136
    | StmtRule -> 137
    | CallStmtRule -> 138
    | SimpleStmtRule -> 139
    | NullStmtRule -> 140
    | AssignmentStmtRule -> 141
    | GotoStmtRule -> 142
    | ExitStmtRule -> 143
    | ReturnStmtRule -> 144
    | RequeueStmtRule -> 145
    | IdentifierRule -> 146
    | CharLiteralRule -> 147
    | StringLiteralRule -> 148
    | DefiningIdRule -> 149
    | DecLiteralRule -> 150
    | IntLiteralRule -> 151
    | NumLiteralRule -> 152
    | NullLiteralRule -> 153
    | AllocatorRule -> 154
    | ForLoopParamSpecRule -> 155
    | QuantifiedExprRule -> 156
    | CaseExprRule -> 157
    | CaseExprAltRule -> 158
    | RaiseExprRule -> 159
    | IfExprRule -> 160
    | ConditionalExprRule -> 161
    | BoxExprRule -> 162
    | OthersDesignatorRule -> 163
    | AggregateAssocRule -> 164
    | RegularAggregateRule -> 165
    | BracketAggregateRule -> 166
    | AggregateRule -> 167
    | DirectNameRule -> 168
    | ParamAssocRule -> 169
    | CallSuffixRule -> 170
    | QualifiedNameRule -> 171
    | QualNameInternalRule -> 172
    | NameRule -> 173
    | DefiningNameRule -> 174
    | DirectNameOrTargetNameRule -> 175
    | TargetNameRule -> 176
    | UpdateAttrAggregateRule -> 177
    | UpdateAttrContentRule -> 178
    | MultidimArrayAssocRule -> 179
    | SubtypeNameRule -> 180
    | StaticNameRule -> 181
    | PrimaryRule -> 182
    | ParenExprRule -> 183
    | DeclareExprRule -> 184
    | FactorRule -> 185
    | TermRule -> 186
    | UnopTermRule -> 187
    | SimpleExprRule -> 188
    | BooleanOpRule -> 189
    | DiscreteRangeRule -> 190
    | ChoiceRule -> 191
    | ChoiceListRule -> 192
    | RelOpRule -> 193
    | MembershipChoiceRule -> 194
    | MembershipChoiceListRule -> 195
    | RelationRule -> 196
    | ExprRule -> 197

   let c_type = view int ~read:wrap ~write:unwrap
end


let free = foreign ~from:c_lib
  "ada_free"
  (ptr void @-> returning void)

(** Assuming char_ptr is a valid char*, convert it to a native Ocaml
  * string and free the C pointer.
  *)
let unwrap_str char_ptr =
  let str = Ctypes.coerce (ptr char) string char_ptr in
  free (Ctypes.coerce (ptr char) (ptr void) char_ptr);
  str


let default_grammar_rule = GrammarRule.CompilationRule

module Sloc = struct
  type t = {
    line : int;
    column : int;
  }

  let c_struct : t structure typ = structure "sloc"
  let line = field c_struct "line" uint32_t
  let column = field c_struct "column" uint16_t
  let () = seal c_struct

  let wrap (c_value : t structure) : t = {
    line = Unsigned.UInt32.to_int (getf c_value line);
    column = Unsigned.UInt16.to_int (getf c_value column);
  }

  let unwrap (value : t) : t structure =
    let c_value = make c_struct in
    setf c_value line (Unsigned.UInt32.of_int (value.line));
    setf c_value column (Unsigned.UInt16.of_int (value.column));
    c_value

  let c_type = view c_struct ~read:wrap ~write:unwrap
end

module SlocRange = struct
  type t = {
    loc_start : Sloc.t;
    loc_end : Sloc.t;
  }

  let c_struct : t structure typ = structure "sloc_range"
  let loc_start = field c_struct "loc_start" Sloc.c_type
  let loc_end = field c_struct "loc_end" Sloc.c_type
  let () = seal c_struct

  let wrap (c_value : t structure) : t = {
    loc_start = getf c_value loc_start;
    loc_end = getf c_value loc_end;
  }

  let unwrap (value : t) : t structure =
    let c_value = make c_struct in
    setf c_value loc_start value.loc_start;
    setf c_value loc_end value.loc_end;
    c_value

  let c_type = view c_struct ~read:wrap ~write:unwrap

  let pp fmt sloc_range =
    Format.fprintf fmt "<SlocRange %d:%d-%d:%d>"
      sloc_range.loc_start.line
      sloc_range.loc_start.column
      sloc_range.loc_end.line
      sloc_range.loc_end.column
end

module Diagnostic = struct
  type t = {
    sloc_range : SlocRange.t;
    message : string
  }

  let c_struct : t structure typ = structure "diagnostic"
  let sloc_range = field c_struct "sloc_range" SlocRange.c_type
  let message = field c_struct "message" Text.c_type
  let () = seal c_struct

  let wrap (c_value : t structure) : t = {
    sloc_range = getf c_value sloc_range;
    message = getf c_value message;
  }

  let unwrap (value : t) : t structure =
    let c_value = make c_struct in
    setf c_value sloc_range value.sloc_range;
    setf c_value message value.message;
    c_value

  let c_type = view c_struct ~read:wrap ~write:unwrap
end

module TokenData = struct
  type t = unit ptr
end

module Token = struct
  type t = {
    token_data : TokenData.t;
    token_index : int;
    trivia_index : int;
    kind : int;
    text : string;
    sloc_range : SlocRange.t;
  }

  let c_struct : t structure typ = structure "token"
  let token_data = field c_struct "token_data" (ptr void)
  let token_index = field c_struct "token_index" int
  let trivia_index = field c_struct "trivia_index" int
  let kind = field c_struct "kind" int
  let text = field c_struct "text" Text.c_type
  let sloc_range = field c_struct "sloc_range" SlocRange.c_type
  let () = seal c_struct

  let wrap (c_value : t structure) : t = {
    token_data = getf c_value token_data;
    token_index = getf c_value token_index;
    trivia_index = getf c_value trivia_index;
    kind = getf c_value kind;
    text = getf c_value text;
    sloc_range = getf c_value sloc_range;
  }

  let unwrap (value : t) : t structure =
    let c_value = make c_struct in
    setf c_value token_data value.token_data;
    setf c_value token_index value.token_index;
    setf c_value trivia_index value.trivia_index;
    setf c_value kind value.kind;
    setf c_value text value.text;
    setf c_value sloc_range value.sloc_range;
    c_value

  let c_type = view c_struct ~read:wrap ~write:unwrap

  let _token_kind_name = foreign ~from:c_lib
    "ada_token_kind_name"
    (int @-> raisable (ptr char))

  let token_kind_name kind =
    unwrap_str (_token_kind_name kind)

  let kind_name token = token_kind_name token.kind

  let token_range_text = foreign ~from:c_lib
    "ada_token_range_text"
    (ptr c_type @-> ptr c_type @-> ptr Text.c_type @-> raisable int)

  let token_next = foreign ~from:c_lib
    "ada_token_next"
    (ptr c_type @-> ptr c_type @-> raisable void)

  let token_previous = foreign ~from:c_lib
    "ada_token_previous"
    (ptr c_type @-> ptr c_type @-> raisable void)

  let is_equivalent = foreign ~from:c_lib
    "ada_token_is_equivalent"
    (ptr c_type @-> ptr c_type @-> raisable bool)

  let pp fmt token =
    let pp_text fmt = function
      | "" -> Format.pp_print_string fmt ""
      | _ as text -> Format.fprintf fmt " %S" text
    in
    Format.fprintf fmt "<Token %s%a at %a>"
      (kind_name token)
      pp_text token.text
      SlocRange.pp token.sloc_range

  let text_range token_first token_last =
    let c_result_ptr = allocate_n Text.c_type ~count:1 in
    let res =
      token_range_text
        (allocate c_type token_first)
        (allocate c_type token_last)
        c_result_ptr
    in
    if res = 0 then
      raise (Invalid_argument
        (Format.asprintf "%a and %a come from different units"
          pp token_first
          pp token_last));
    !@ c_result_ptr

  let next token =
    let c_next_token_ptr = allocate_n c_type ~count:1 in
    token_next (allocate c_type token) c_next_token_ptr ;
    !@ c_next_token_ptr

  let previous token =
    let c_next_token_ptr = allocate_n c_type ~count:1 in
    token_previous (allocate c_type token) c_next_token_ptr ;
    !@ c_next_token_ptr

  let is_trivia token =
    token.trivia_index != 0

  let index token =
    match token.trivia_index with
    | 0 ->
        token.token_index - 1
    | _ ->
        token.trivia_index - 1

  let compare one other =
    let open Stdlib in
    let compare_token_data = compare one.token_data other.token_data in
    if compare_token_data = 0 then
      let compare_token_index = compare one.token_index other.token_index in
      if compare_token_index = 0 then
        compare one.trivia_index other.trivia_index
      else
        compare_token_index
    else
      compare_token_data

  let equal one other =
    compare one other = 0

  let hash token =
    Hashtbl.hash
      (token.token_data
       , token.token_index
       , token.trivia_index)

  let is_equivalent one other =
    is_equivalent (allocate c_type one) (allocate c_type other)

end

module UnitProvider = struct
  (* The real C type of a context is a void*. But we use a pointer to this
     type, to be able to allocate a value of t and attach a finalizer to it. *)
  type t = unit ptr ptr

  let c_type = ptr void

  let null = allocate c_type null

  
        type project_scenario_variable

  let c_project_scenario_variable : project_scenario_variable structure typ =
    structure "project_scenario_variable"

  let scenario_var_name = field c_project_scenario_variable "name" string

  let scenario_var_value = field c_project_scenario_variable "value" string

  let () = seal c_project_scenario_variable

  let create_project_unit_provider =
    foreign ~from:c_lib "ada_create_project_unit_provider"
    (string @-> string @-> ptr c_project_scenario_variable @-> string
     @-> string @-> raisable c_type)

  let dec_ref =
    let f =
      foreign ~from:c_lib "ada_dec_ref_unit_provider"
        (c_type @-> raisable void)
    in
    (fun v ->
      f (!@ v))

  let wrap v =
    allocate ~finalise:dec_ref (ptr void) v

  let for_project
      ?(project = "")
      ?(scenario_vars = [])
      ?(target = "")
      ?(runtime = "")
      project_file =
    (* One more to store (null, null)*)
    let scenario_vars_len = List.length scenario_vars + 1 in
    let c_scenario_vars =
      (* Allocates a fresh array with given size, fill with zeros. Thus,
       the last cell is already (null, null) *)
      allocate_n c_project_scenario_variable ~count:scenario_vars_len
    in
    let fill_scenario_vars i (name, value) =
      let c_struct = make c_project_scenario_variable in
      setf c_struct scenario_var_name name ;
      setf c_struct scenario_var_value value ;
      c_scenario_vars +@ i <-@ c_struct
    in
    List.iteri fill_scenario_vars scenario_vars ;
    let result = create_project_unit_provider
                 project_file project
                 c_scenario_vars
                 target runtime
    in
    if is_null result then
      raise (InvalidProjectError ("Cannot open or interpret project " ^ project_file))
    else
      wrap result

  let create_auto_provider =
    foreign ~from:c_lib "ada_create_auto_provider"
      (ptr (ptr char) @-> string @-> raisable c_type)

  let auto input_files =
    (* Convert the names of the input files into pointers to C strings. We used
       to use the high-level type [Ctypes.string] type, but this was causing
       memory corruption problems. We switched to [ptr char] instead. *)
    let cstrings =
      List.map (fun f -> CArray.(start (of_string f))) input_files
    in
    (* Add a null pointer at the end. This is part of the LAL calling
       convention. *)
    let null_ptr = from_voidp char Ctypes.null in
    let cstrings_null = List.rev_append cstrings [null_ptr] in
    (* Create an array with all these pointers *)
    let array = CArray.of_list (ptr char) cstrings_null in
    let ptr = CArray.start array in
    let result = create_auto_provider ptr "" in
    wrap result


end

module BareNode = struct
  type t = unit ptr
end

module Rebindings = struct
  type t = unit ptr
end



      
module MetadataStruct = struct
  type t

   
  let c_type : t structure typ = structure "metadata"
  let dottable_subp =
    field c_type "dottable_subp" bool
  let access_entity =
    field c_type "access_entity" bool
  let primitive =
    field c_type "primitive" (ptr void)
  let primitive_real_type =
    field c_type "primitive_real_type" (ptr void)
  let () = seal c_type

end

      
module EntityInfoStruct = struct
  type t

   
  let c_type : t structure typ = structure "entity_info"
  let md =
    field c_type "md" MetadataStruct.c_type
  let rebindings =
    field c_type "rebindings" (ptr void)
  let from_rebound =
    field c_type "from_rebound" bool
  let () = seal c_type

end

         
module EntityStruct = struct
  type t

   
  let c_type : t structure typ = structure "ada_node"
  let node =
    field c_type "node" (ptr void)
  let info =
    field c_type "info" EntityInfoStruct.c_type
  let () = seal c_type

end

      
module AnalysisUnitStruct : sig
  type t = unit ptr

  val c_type : t typ

  val unit_root : t -> EntityStruct.t structure ptr -> unit

  val unit_diagnostic_count : t -> int

  val unit_diagnostic : t -> int -> Diagnostic.t ptr -> int

  val unit_filename : t -> char ptr

  val unit_reparse_from_file : t -> string -> int

  val unit_reparse_from_buffer :
    t -> string -> string -> Unsigned.size_t -> int

  val unit_first_token : t -> Token.t ptr -> unit

  val unit_last_token : t -> Token.t ptr -> unit

  val unit_token_count : t -> int

  val unit_trivia_count : t -> int
end = struct
  (* Module defining the c structure of an analysis unit *)

  type t = unit ptr
  let c_type = ptr void

  let unit_root = foreign ~from:c_lib "ada_unit_root"
    (c_type @-> ptr EntityStruct.c_type @-> raisable void)

  let unit_diagnostic_count = foreign ~from:c_lib
    "ada_unit_diagnostic_count"
    (c_type @-> raisable int)

  let unit_diagnostic = foreign ~from:c_lib
    "ada_unit_diagnostic"
    (c_type @-> int @-> ptr Diagnostic.c_type @-> raisable int)

  let unit_filename = foreign ~from:c_lib
    "ada_unit_filename"
    (c_type @-> raisable (ptr char))

  let unit_reparse_from_file = foreign ~from:c_lib
    "ada_unit_reparse_from_file"
    (c_type
     @-> string
     @-> raisable int)

  let unit_reparse_from_buffer = foreign ~from:c_lib
    "ada_unit_reparse_from_buffer"
    (c_type
     @-> string
     @-> string
     @-> size_t
     @-> raisable int)

  let unit_first_token = foreign ~from:c_lib
    "ada_unit_first_token"
    (c_type
     @-> ptr Token.c_type
     @-> raisable void)

  let unit_last_token = foreign ~from:c_lib
    "ada_unit_last_token"
    (c_type
     @-> ptr Token.c_type
     @-> raisable void)

  let unit_token_count = foreign ~from:c_lib
    "ada_unit_token_count"
    (c_type @-> raisable int)

  let unit_trivia_count = foreign ~from:c_lib
    "ada_unit_trivia_count"
    (c_type @-> raisable int)
end


      
module AnalysisContextStruct : sig
  type t

  val c_type : t typ

  val create_analysis_context :
    string -> unit ptr -> unit ptr -> bool -> int -> t

  val get_analysis_unit_from_file :
    t -> string -> string -> bool -> GrammarRule.t -> AnalysisUnitStruct.t

  val get_analysis_unit_from_buffer :
    t
    -> string (* Filename *)
    -> string (* Charset *)
    -> string (* Buffer *)
    -> Unsigned.size_t (* Buffer size *)
    -> GrammarRule.t
    -> AnalysisUnitStruct.t
end = struct
  (* Module defining the c structure of an analysis context *)

  (* The real C type of a context is a void*. But we use a pointer to this
     type, to be able to allocate a value of t and attach a finalizer to it.
     See wrap function *)
  type t = unit ptr ptr

  let unwrap (value : t) : unit ptr = !@value

  let context_decref =
    let f =
      foreign ~from:c_lib "ada_context_decref"
        (ptr void @-> raisable void)
    in
    fun ctx -> f (unwrap ctx)

  let wrap (c_value : unit ptr) : t =
    (* To deallocate cleanly the context, we need to call context_decref.
       Allocate a value and attach a finalizer to it *)
    allocate ~finalise:context_decref (ptr void) c_value

  let c_type = view (ptr void) ~read:wrap ~write:unwrap

  let create_analysis_context =
    foreign ~from:c_lib "ada_create_analysis_context"
      (string @-> ptr void @-> UnitProvider.c_type @-> bool @-> int
      @-> raisable c_type)

  let get_analysis_unit_from_file =
    foreign ~from:c_lib "ada_get_analysis_unit_from_file"
      ( c_type @-> string @-> string @-> bool @-> GrammarRule.c_type
      @-> raisable AnalysisUnitStruct.c_type )

  let get_analysis_unit_from_buffer =
    foreign ~from:c_lib "ada_get_analysis_unit_from_buffer"
      ( c_type @-> string (* Filename *) @-> string (* Charset *)
      @-> string (* Buffer *) @-> size_t (* Buffer size *)
      @-> GrammarRule.c_type
      @-> raisable AnalysisUnitStruct.c_type )

end

      
module Symbol : sig
  type t = string

  val c_type : t structure typ

  val wrap : (t structure) -> t

  val unwrap : AnalysisContextStruct.t -> t -> (t structure)

  val symbol_text : t structure ptr -> string ptr -> unit

  val context_symbol :
    AnalysisContextStruct.t -> string ptr -> t structure ptr -> int
end = struct
  type t = string

  let c_type : t structure typ = structure "symbol"
  let data = field c_type "data" (ptr void)
  let bounds = field c_type "bounds" (ptr void)
  let () = seal c_type

  let symbol_text = foreign ~from:c_lib "ada_symbol_text"
    (ptr c_type @-> ptr Text.c_type @-> raisable void)

  let wrap (c_value : t structure) : t =
    let c_result_ptr = allocate_n Text.c_type ~count:1 in
    symbol_text (addr c_value) c_result_ptr;
    !@ c_result_ptr

  let context_symbol = foreign ~from:c_lib "ada_context_symbol"
    (AnalysisContextStruct.c_type
     @-> ptr Text.c_type
     @-> ptr c_type
     @-> raisable int)

  let unwrap (ctx : AnalysisContextStruct.t) (value : t) : t structure =
    let result = make c_type in
    let code =
      context_symbol ctx (allocate Text.c_type value) (addr result)
    in
    if code = 0 then
      raise (InvalidSymbolError value) ;
    result
end

      
module AspectStruct = struct
  type t

   
  let c_type : t structure typ = structure "aspect"
  let exists =
    field c_type "exists" bool
  let node =
    field c_type "node" EntityStruct.c_type
  let value =
    field c_type "value" EntityStruct.c_type
  let () = seal c_type

end

      
module CompletionItemStruct = struct
  type t

   
  let c_type : t structure typ = structure "completion_item"
  let decl =
    field c_type "decl" EntityStruct.c_type
  let is_dot_call =
    field c_type "is_dot_call" bool
  let is_visible =
    field c_type "is_visible" bool
  let () = seal c_type

end

      
module DiscreteRangeStruct = struct
  type t

   
  let c_type : t structure typ = structure "discrete_range"
  let low_bound =
    field c_type "low_bound" EntityStruct.c_type
  let high_bound =
    field c_type "high_bound" EntityStruct.c_type
  let () = seal c_type

end

      
module DiscriminantValuesStruct = struct
  type t

   
  let c_type : t structure typ = structure "discriminant_values"
  let discriminant =
    field c_type "discriminant" EntityStruct.c_type
  let values =
    field c_type "values" EntityStruct.c_type
  let () = seal c_type

end

         
module TextTypeStruct = struct
  

  type t

  let c_struct : t structure typ = structure "text_type"
  let n = field c_struct "n" int
  let _ = field c_struct "ref_count" int
  let items = field c_struct "items" uint32_t
  let () = seal c_struct

  let c_type = ptr c_struct

  let create = foreign ~from:c_lib "ada_text_type_create"
    (int @-> raisable c_type)
  let dec_ref = foreign ~from:c_lib "ada_text_type_dec_ref"
    (c_type @-> raisable void)

end

      
module DocAnnotationStruct = struct
  type t

   
  let c_type : t structure typ = structure "doc_annotation"
  let key =
    field c_type "key" TextTypeStruct.c_type
  let value =
    field c_type "value" TextTypeStruct.c_type
  let () = seal c_type

end

      
module ParamActualStruct = struct
  type t

   
  let c_type : t structure typ = structure "param_actual"
  let param =
    field c_type "param" EntityStruct.c_type
  let actual =
    field c_type "actual" EntityStruct.c_type
  let () = seal c_type

end

      
module RefResultStruct = struct
  type t

   
  let c_type : t structure typ = structure "ref_result"
  let ref =
    field c_type "ref" EntityStruct.c_type
  let kind =
    field c_type "kind" RefResultKind.c_type
  let () = seal c_type

end

      
module RefdDeclStruct = struct
  type t

   
  let c_type : t structure typ = structure "refd_decl"
  let decl =
    field c_type "decl" EntityStruct.c_type
  let kind =
    field c_type "kind" RefResultKind.c_type
  let () = seal c_type

end

      
module RefdDefStruct = struct
  type t

   
  let c_type : t structure typ = structure "refd_def"
  let def_name =
    field c_type "def_name" EntityStruct.c_type
  let kind =
    field c_type "kind" RefResultKind.c_type
  let () = seal c_type

end

         
module AdaNodeArrayStruct = struct
  

  type t

  let c_struct : t structure typ = structure "ada_node_array"
  let n = field c_struct "n" int
  let _ = field c_struct "ref_count" int
  let items = field c_struct "items"
    EntityStruct.c_type
  let () = seal c_struct

  let c_type = ptr c_struct

  let create = foreign ~from:c_lib "ada_ada_node_array_create"
    (int @-> raisable c_type)
  let dec_ref = foreign ~from:c_lib "ada_ada_node_array_dec_ref"
    (c_type @-> raisable void)

end

         
module DiscriminantValuesArrayStruct = struct
  

  type t

  let c_struct : t structure typ = structure "discriminant_values_array"
  let n = field c_struct "n" int
  let _ = field c_struct "ref_count" int
  let items = field c_struct "items"
    DiscriminantValuesStruct.c_type
  let () = seal c_struct

  let c_type = ptr c_struct

  let create = foreign ~from:c_lib "ada_discriminant_values_array_create"
    (int @-> raisable c_type)
  let dec_ref = foreign ~from:c_lib "ada_discriminant_values_array_dec_ref"
    (c_type @-> raisable void)

end

      
module ShapeStruct = struct
  type t

   
  let c_type : t structure typ = structure "shape"
  let components =
    field c_type "components" AdaNodeArrayStruct.c_type
  let discriminants_values =
    field c_type "discriminants_values" DiscriminantValuesArrayStruct.c_type
  let () = seal c_type

end

      
module SubstitutionStruct = struct
  type t

   
  let c_type : t structure typ = structure "substitution"
  let from_decl =
    field c_type "from_decl" EntityStruct.c_type
  let to_value =
    field c_type "to_value" BigInteger.c_type
  let value_type =
    field c_type "value_type" EntityStruct.c_type
  let () = seal c_type

end

         
module DocAnnotationArrayStruct = struct
  

  type t

  let c_struct : t structure typ = structure "doc_annotation_array"
  let n = field c_struct "n" int
  let _ = field c_struct "ref_count" int
  let items = field c_struct "items"
    DocAnnotationStruct.c_type
  let () = seal c_struct

  let c_type = ptr c_struct

  let create = foreign ~from:c_lib "ada_doc_annotation_array_create"
    (int @-> raisable c_type)
  let dec_ref = foreign ~from:c_lib "ada_doc_annotation_array_dec_ref"
    (c_type @-> raisable void)

end

         
module ParamActualArrayStruct = struct
  

  type t

  let c_struct : t structure typ = structure "param_actual_array"
  let n = field c_struct "n" int
  let _ = field c_struct "ref_count" int
  let items = field c_struct "items"
    ParamActualStruct.c_type
  let () = seal c_struct

  let c_type = ptr c_struct

  let create = foreign ~from:c_lib "ada_param_actual_array_create"
    (int @-> raisable c_type)
  let dec_ref = foreign ~from:c_lib "ada_param_actual_array_dec_ref"
    (c_type @-> raisable void)

end

         
module RefResultArrayStruct = struct
  

  type t

  let c_struct : t structure typ = structure "ref_result_array"
  let n = field c_struct "n" int
  let _ = field c_struct "ref_count" int
  let items = field c_struct "items"
    RefResultStruct.c_type
  let () = seal c_struct

  let c_type = ptr c_struct

  let create = foreign ~from:c_lib "ada_ref_result_array_create"
    (int @-> raisable c_type)
  let dec_ref = foreign ~from:c_lib "ada_ref_result_array_dec_ref"
    (c_type @-> raisable void)

end

         
module ShapeArrayStruct = struct
  

  type t

  let c_struct : t structure typ = structure "shape_array"
  let n = field c_struct "n" int
  let _ = field c_struct "ref_count" int
  let items = field c_struct "items"
    ShapeStruct.c_type
  let () = seal c_struct

  let c_type = ptr c_struct

  let create = foreign ~from:c_lib "ada_shape_array_create"
    (int @-> raisable c_type)
  let dec_ref = foreign ~from:c_lib "ada_shape_array_dec_ref"
    (c_type @-> raisable void)

end

         
module SubstitutionArrayStruct = struct
  

  type t

  let c_struct : t structure typ = structure "substitution_array"
  let n = field c_struct "n" int
  let _ = field c_struct "ref_count" int
  let items = field c_struct "items"
    SubstitutionStruct.c_type
  let () = seal c_struct

  let c_type = ptr c_struct

  let create = foreign ~from:c_lib "ada_substitution_array_create"
    (int @-> raisable c_type)
  let dec_ref = foreign ~from:c_lib "ada_substitution_array_dec_ref"
    (c_type @-> raisable void)

end

         
module AnalysisUnitArrayStruct = struct
  

  type t

  let c_struct : t structure typ = structure "analysis_unit_array"
  let n = field c_struct "n" int
  let _ = field c_struct "ref_count" int
  let items = field c_struct "items"
    AnalysisUnitStruct.c_type
  let () = seal c_struct

  let c_type = ptr c_struct

  let create = foreign ~from:c_lib "ada_analysis_unit_array_create"
    (int @-> raisable c_type)
  let dec_ref = foreign ~from:c_lib "ada_analysis_unit_array_dec_ref"
    (c_type @-> raisable void)

end

         
module UnboundedTextTypeArrayStruct = struct
  

  type t

  let c_struct : t structure typ = structure "unbounded_text_type_array"
  let n = field c_struct "n" int
  let _ = field c_struct "ref_count" int
  let items = field c_struct "items"
    Symbol.c_type
  let () = seal c_struct

  let c_type = ptr c_struct

  let create = foreign ~from:c_lib "ada_unbounded_text_type_array_create"
    (int @-> raisable c_type)
  let dec_ref = foreign ~from:c_lib "ada_unbounded_text_type_array_dec_ref"
    (c_type @-> raisable void)

end


module CFunctions = struct
  let node_kind = foreign ~from:c_lib "ada_node_kind"
    (ptr EntityStruct.c_type @-> raisable int)

  let image = foreign ~from:c_lib "ada_node_image"
    (ptr EntityStruct.c_type
     @-> ptr Text.c_type
     @-> raisable void)

  let node_sloc_range = foreign ~from:c_lib
    "ada_node_sloc_range"
    (ptr EntityStruct.c_type
     @-> ptr SlocRange.c_type
     @-> raisable void)

  let lookup_in_node = foreign ~from:c_lib
    "ada_lookup_in_node"
    (ptr EntityStruct.c_type
     @-> ptr Sloc.c_type
     @-> ptr EntityStruct.c_type
     @-> raisable void)

  let entity_image = foreign ~from:c_lib
    "ada_entity_image"
    (ptr EntityStruct.c_type
     @-> ptr Text.c_type
     @-> raisable void)

  let node_is_token_node = foreign ~from:c_lib
    "ada_node_is_token_node"
    (ptr EntityStruct.c_type
     @-> raisable bool)

  let ada_node_p_declarative_scope = foreign ~from:c_lib
    "ada_ada_node_p_declarative_scope"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)


  let ada_node_p_valid_keywords = foreign ~from:c_lib
    "ada_ada_node_p_valid_keywords"
    (ptr EntityStruct.c_type
    @-> ptr UnboundedTextTypeArrayStruct.c_type
    @-> raisable int)

  let ada_node_p_generic_instantiations = foreign ~from:c_lib
    "ada_ada_node_p_generic_instantiations"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let ada_node_p_semantic_parent = foreign ~from:c_lib
    "ada_ada_node_p_semantic_parent"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_p_parent_basic_decl = foreign ~from:c_lib
    "ada_ada_node_p_parent_basic_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_p_filter_is_imported_by = foreign ~from:c_lib
    "ada_ada_node_p_filter_is_imported_by"
    (ptr EntityStruct.c_type
        
    @-> AnalysisUnitArrayStruct.c_type
        
    @-> bool
    @-> ptr AnalysisUnitArrayStruct.c_type
    @-> raisable int)

  let ada_node_p_xref_entry_point = foreign ~from:c_lib
    "ada_ada_node_p_xref_entry_point"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let ada_node_p_resolve_names = foreign ~from:c_lib
    "ada_ada_node_p_resolve_names"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let ada_node_p_standard_unit = foreign ~from:c_lib
    "ada_ada_node_p_standard_unit"
    (ptr EntityStruct.c_type
    @-> ptr AnalysisUnitStruct.c_type
    @-> raisable int)

  let ada_node_p_std_entity = foreign ~from:c_lib
    "ada_ada_node_p_std_entity"
    (ptr EntityStruct.c_type
        
    @-> ptr Symbol.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_p_bool_type = foreign ~from:c_lib
    "ada_ada_node_p_bool_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_p_int_type = foreign ~from:c_lib
    "ada_ada_node_p_int_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_p_universal_int_type = foreign ~from:c_lib
    "ada_ada_node_p_universal_int_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_p_universal_real_type = foreign ~from:c_lib
    "ada_ada_node_p_universal_real_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_p_top_level_decl = foreign ~from:c_lib
    "ada_ada_node_p_top_level_decl"
    (ptr EntityStruct.c_type
        
    @-> AnalysisUnitStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_p_choice_match = foreign ~from:c_lib
    "ada_ada_node_p_choice_match"
    (ptr EntityStruct.c_type
        
    @-> ptr BigInteger.c_type
    @-> ptr bool
    @-> raisable int)

  let ada_node_p_gnat_xref = foreign ~from:c_lib
    "ada_ada_node_p_gnat_xref"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_parent = foreign ~from:c_lib
    "ada_ada_node_parent"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_parents = foreign ~from:c_lib
    "ada_ada_node_parents"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let ada_node_children = foreign ~from:c_lib
    "ada_ada_node_children"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let ada_node_token_start = foreign ~from:c_lib
    "ada_ada_node_token_start"
    (ptr EntityStruct.c_type
    @-> ptr Token.c_type
    @-> raisable int)

  let ada_node_token_end = foreign ~from:c_lib
    "ada_ada_node_token_end"
    (ptr EntityStruct.c_type
    @-> ptr Token.c_type
    @-> raisable int)

  let ada_node_child_index = foreign ~from:c_lib
    "ada_ada_node_child_index"
    (ptr EntityStruct.c_type
    @-> ptr int
    @-> raisable int)

  let ada_node_previous_sibling = foreign ~from:c_lib
    "ada_ada_node_previous_sibling"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_next_sibling = foreign ~from:c_lib
    "ada_ada_node_next_sibling"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ada_node_unit = foreign ~from:c_lib
    "ada_ada_node_unit"
    (ptr EntityStruct.c_type
    @-> ptr AnalysisUnitStruct.c_type
    @-> raisable int)

  let ada_node_is_ghost = foreign ~from:c_lib
    "ada_ada_node_is_ghost"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let ada_node_full_sloc_image = foreign ~from:c_lib
    "ada_ada_node_full_sloc_image"
    (ptr EntityStruct.c_type
    @-> ptr TextTypeStruct.c_type
    @-> raisable int)

  let abort_node_p_as_bool = foreign ~from:c_lib
    "ada_abort_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let abstract_node_p_as_bool = foreign ~from:c_lib
    "ada_abstract_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let assoc_list_p_zip_with_params = foreign ~from:c_lib
    "ada_assoc_list_p_zip_with_params"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr ParamActualArrayStruct.c_type
    @-> raisable int)

  let aliased_node_p_as_bool = foreign ~from:c_lib
    "ada_aliased_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let all_node_p_as_bool = foreign ~from:c_lib
    "ada_all_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let constrained_array_indices_f_list = foreign ~from:c_lib
    "ada_constrained_array_indices_f_list"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let unconstrained_array_indices_f_types = foreign ~from:c_lib
    "ada_unconstrained_array_indices_f_types"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let aspect_assoc_f_id = foreign ~from:c_lib
    "ada_aspect_assoc_f_id"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let aspect_assoc_f_expr = foreign ~from:c_lib
    "ada_aspect_assoc_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let at_clause_f_name = foreign ~from:c_lib
    "ada_at_clause_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let at_clause_f_expr = foreign ~from:c_lib
    "ada_at_clause_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let attribute_def_clause_f_attribute_expr = foreign ~from:c_lib
    "ada_attribute_def_clause_f_attribute_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let attribute_def_clause_f_expr = foreign ~from:c_lib
    "ada_attribute_def_clause_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let enum_rep_clause_f_type_name = foreign ~from:c_lib
    "ada_enum_rep_clause_f_type_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let enum_rep_clause_f_aggregate = foreign ~from:c_lib
    "ada_enum_rep_clause_f_aggregate"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let record_rep_clause_f_name = foreign ~from:c_lib
    "ada_record_rep_clause_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let record_rep_clause_f_at_expr = foreign ~from:c_lib
    "ada_record_rep_clause_f_at_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let record_rep_clause_f_components = foreign ~from:c_lib
    "ada_record_rep_clause_f_components"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let aspect_spec_f_aspect_assocs = foreign ~from:c_lib
    "ada_aspect_spec_f_aspect_assocs"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_assoc_p_assoc_expr = foreign ~from:c_lib
    "ada_base_assoc_p_assoc_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let contract_case_assoc_f_guard = foreign ~from:c_lib
    "ada_contract_case_assoc_f_guard"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let contract_case_assoc_f_consequence = foreign ~from:c_lib
    "ada_contract_case_assoc_f_consequence"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let pragma_argument_assoc_f_id = foreign ~from:c_lib
    "ada_pragma_argument_assoc_f_id"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let pragma_argument_assoc_f_expr = foreign ~from:c_lib
    "ada_pragma_argument_assoc_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_formal_param_holder_p_abstract_formal_params = foreign ~from:c_lib
    "ada_base_formal_param_holder_p_abstract_formal_params"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let base_formal_param_holder_p_nb_min_params = foreign ~from:c_lib
    "ada_base_formal_param_holder_p_nb_min_params"
    (ptr EntityStruct.c_type
    @-> ptr int
    @-> raisable int)

  let base_formal_param_holder_p_nb_max_params = foreign ~from:c_lib
    "ada_base_formal_param_holder_p_nb_max_params"
    (ptr EntityStruct.c_type
    @-> ptr int
    @-> raisable int)

  let base_formal_param_holder_p_param_types = foreign ~from:c_lib
    "ada_base_formal_param_holder_p_param_types"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let base_subp_spec_p_returns = foreign ~from:c_lib
    "ada_base_subp_spec_p_returns"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_subp_spec_p_params = foreign ~from:c_lib
    "ada_base_subp_spec_p_params"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let base_subp_spec_p_primitive_subp_types = foreign ~from:c_lib
    "ada_base_subp_spec_p_primitive_subp_types"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let base_subp_spec_p_primitive_subp_first_type = foreign ~from:c_lib
    "ada_base_subp_spec_p_primitive_subp_first_type"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_subp_spec_p_primitive_subp_tagged_type = foreign ~from:c_lib
    "ada_base_subp_spec_p_primitive_subp_tagged_type"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_subp_spec_p_return_type = foreign ~from:c_lib
    "ada_base_subp_spec_p_return_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_spec_f_entry_name = foreign ~from:c_lib
    "ada_entry_spec_f_entry_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_spec_f_family_type = foreign ~from:c_lib
    "ada_entry_spec_f_family_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_spec_f_entry_params = foreign ~from:c_lib
    "ada_entry_spec_f_entry_params"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subp_spec_f_subp_kind = foreign ~from:c_lib
    "ada_subp_spec_f_subp_kind"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subp_spec_f_subp_name = foreign ~from:c_lib
    "ada_subp_spec_f_subp_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subp_spec_f_subp_params = foreign ~from:c_lib
    "ada_subp_spec_f_subp_params"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subp_spec_f_subp_returns = foreign ~from:c_lib
    "ada_subp_spec_f_subp_returns"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let component_list_f_components = foreign ~from:c_lib
    "ada_component_list_f_components"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let component_list_f_variant_part = foreign ~from:c_lib
    "ada_component_list_f_variant_part"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let known_discriminant_part_f_discr_specs = foreign ~from:c_lib
    "ada_known_discriminant_part_f_discr_specs"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_completion_formal_params_f_params = foreign ~from:c_lib
    "ada_entry_completion_formal_params_f_params"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_formal_part_f_decls = foreign ~from:c_lib
    "ada_generic_formal_part_f_decls"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_record_def_f_components = foreign ~from:c_lib
    "ada_base_record_def_f_components"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_assoc_p_get_params = foreign ~from:c_lib
    "ada_basic_assoc_p_get_params"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let aggregate_assoc_f_designators = foreign ~from:c_lib
    "ada_aggregate_assoc_f_designators"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let aggregate_assoc_f_r_expr = foreign ~from:c_lib
    "ada_aggregate_assoc_f_r_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let discriminant_assoc_f_ids = foreign ~from:c_lib
    "ada_discriminant_assoc_f_ids"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let discriminant_assoc_f_discr_expr = foreign ~from:c_lib
    "ada_discriminant_assoc_f_discr_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let iterated_assoc_f_spec = foreign ~from:c_lib
    "ada_iterated_assoc_f_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let iterated_assoc_f_r_expr = foreign ~from:c_lib
    "ada_iterated_assoc_f_r_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let param_assoc_f_designator = foreign ~from:c_lib
    "ada_param_assoc_f_designator"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let param_assoc_f_r_expr = foreign ~from:c_lib
    "ada_param_assoc_f_r_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_is_formal = foreign ~from:c_lib
    "ada_basic_decl_p_is_formal"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let basic_decl_p_doc_annotations = foreign ~from:c_lib
    "ada_basic_decl_p_doc_annotations"
    (ptr EntityStruct.c_type
    @-> ptr DocAnnotationArrayStruct.c_type
    @-> raisable int)

  let basic_decl_p_doc = foreign ~from:c_lib
    "ada_basic_decl_p_doc"
    (ptr EntityStruct.c_type
    @-> ptr TextTypeStruct.c_type
    @-> raisable int)

  let basic_decl_p_previous_part_for_decl = foreign ~from:c_lib
    "ada_basic_decl_p_previous_part_for_decl"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_canonical_part = foreign ~from:c_lib
    "ada_basic_decl_p_canonical_part"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_all_parts = foreign ~from:c_lib
    "ada_basic_decl_p_all_parts"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let basic_decl_p_is_static_decl = foreign ~from:c_lib
    "ada_basic_decl_p_is_static_decl"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr bool
    @-> raisable int)

  let basic_decl_p_is_imported = foreign ~from:c_lib
    "ada_basic_decl_p_is_imported"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let basic_decl_f_aspects = foreign ~from:c_lib
    "ada_basic_decl_f_aspects"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_get_aspect_assoc = foreign ~from:c_lib
    "ada_basic_decl_p_get_aspect_assoc"
    (ptr EntityStruct.c_type
        
    @-> ptr Symbol.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_get_aspect_spec_expr = foreign ~from:c_lib
    "ada_basic_decl_p_get_aspect_spec_expr"
    (ptr EntityStruct.c_type
        
    @-> ptr Symbol.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_get_aspect = foreign ~from:c_lib
    "ada_basic_decl_p_get_aspect"
    (ptr EntityStruct.c_type
        
    @-> ptr Symbol.c_type
        
    @-> bool
    @-> ptr AspectStruct.c_type
    @-> raisable int)

  let basic_decl_p_has_aspect = foreign ~from:c_lib
    "ada_basic_decl_p_has_aspect"
    (ptr EntityStruct.c_type
        
    @-> ptr Symbol.c_type
        
    @-> bool
    @-> ptr bool
    @-> raisable int)

  let basic_decl_p_get_pragma = foreign ~from:c_lib
    "ada_basic_decl_p_get_pragma"
    (ptr EntityStruct.c_type
        
    @-> ptr Symbol.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_get_representation_clause = foreign ~from:c_lib
    "ada_basic_decl_p_get_representation_clause"
    (ptr EntityStruct.c_type
        
    @-> ptr Symbol.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_is_compilation_unit_root = foreign ~from:c_lib
    "ada_basic_decl_p_is_compilation_unit_root"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let basic_decl_p_is_visible = foreign ~from:c_lib
    "ada_basic_decl_p_is_visible"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let basic_decl_p_base_subp_declarations = foreign ~from:c_lib
    "ada_basic_decl_p_base_subp_declarations"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let basic_decl_p_root_subp_declarations = foreign ~from:c_lib
    "ada_basic_decl_p_root_subp_declarations"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let basic_decl_p_find_all_overrides = foreign ~from:c_lib
    "ada_basic_decl_p_find_all_overrides"
    (ptr EntityStruct.c_type
        
    @-> AnalysisUnitArrayStruct.c_type
        
    @-> bool
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let basic_decl_p_defining_names = foreign ~from:c_lib
    "ada_basic_decl_p_defining_names"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let basic_decl_p_defining_name = foreign ~from:c_lib
    "ada_basic_decl_p_defining_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_type_expression = foreign ~from:c_lib
    "ada_basic_decl_p_type_expression"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_subp_spec_or_null = foreign ~from:c_lib
    "ada_basic_decl_p_subp_spec_or_null"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_is_subprogram = foreign ~from:c_lib
    "ada_basic_decl_p_is_subprogram"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let basic_decl_p_relative_name = foreign ~from:c_lib
    "ada_basic_decl_p_relative_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_relative_name_text = foreign ~from:c_lib
    "ada_basic_decl_p_relative_name_text"
    (ptr EntityStruct.c_type
    @-> ptr Symbol.c_type
    @-> raisable int)

  let basic_decl_p_next_part_for_decl = foreign ~from:c_lib
    "ada_basic_decl_p_next_part_for_decl"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_body_part_for_decl = foreign ~from:c_lib
    "ada_basic_decl_p_body_part_for_decl"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let basic_decl_p_fully_qualified_name_array = foreign ~from:c_lib
    "ada_basic_decl_p_fully_qualified_name_array"
    (ptr EntityStruct.c_type
    @-> ptr UnboundedTextTypeArrayStruct.c_type
    @-> raisable int)

  let basic_decl_p_fully_qualified_name = foreign ~from:c_lib
    "ada_basic_decl_p_fully_qualified_name"
    (ptr EntityStruct.c_type
    @-> ptr TextTypeStruct.c_type
    @-> raisable int)

  let basic_decl_p_canonical_fully_qualified_name = foreign ~from:c_lib
    "ada_basic_decl_p_canonical_fully_qualified_name"
    (ptr EntityStruct.c_type
    @-> ptr TextTypeStruct.c_type
    @-> raisable int)

  let basic_decl_p_unique_identifying_name = foreign ~from:c_lib
    "ada_basic_decl_p_unique_identifying_name"
    (ptr EntityStruct.c_type
    @-> ptr TextTypeStruct.c_type
    @-> raisable int)

  let abstract_state_decl_f_name = foreign ~from:c_lib
    "ada_abstract_state_decl_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let anonymous_expr_decl_f_expr = foreign ~from:c_lib
    "ada_anonymous_expr_decl_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_formal_param_decl_p_formal_type = foreign ~from:c_lib
    "ada_base_formal_param_decl_p_formal_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let component_decl_f_ids = foreign ~from:c_lib
    "ada_component_decl_f_ids"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let component_decl_f_component_def = foreign ~from:c_lib
    "ada_component_decl_f_component_def"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let component_decl_f_default_expr = foreign ~from:c_lib
    "ada_component_decl_f_default_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let discriminant_spec_f_ids = foreign ~from:c_lib
    "ada_discriminant_spec_f_ids"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let discriminant_spec_f_type_expr = foreign ~from:c_lib
    "ada_discriminant_spec_f_type_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let discriminant_spec_f_default_expr = foreign ~from:c_lib
    "ada_discriminant_spec_f_default_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_formal_f_decl = foreign ~from:c_lib
    "ada_generic_formal_f_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let param_spec_f_ids = foreign ~from:c_lib
    "ada_param_spec_f_ids"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let param_spec_f_has_aliased = foreign ~from:c_lib
    "ada_param_spec_f_has_aliased"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let param_spec_f_mode = foreign ~from:c_lib
    "ada_param_spec_f_mode"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let param_spec_f_type_expr = foreign ~from:c_lib
    "ada_param_spec_f_type_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let param_spec_f_default_expr = foreign ~from:c_lib
    "ada_param_spec_f_default_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_package_decl_f_package_name = foreign ~from:c_lib
    "ada_base_package_decl_f_package_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_package_decl_f_public_part = foreign ~from:c_lib
    "ada_base_package_decl_f_public_part"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_package_decl_f_private_part = foreign ~from:c_lib
    "ada_base_package_decl_f_private_part"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_package_decl_f_end_name = foreign ~from:c_lib
    "ada_base_package_decl_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_package_decl_p_body_part = foreign ~from:c_lib
    "ada_base_package_decl_p_body_part"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_f_name = foreign ~from:c_lib
    "ada_base_type_decl_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_base_subtype = foreign ~from:c_lib
    "ada_base_type_decl_p_base_subtype"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_private_completion = foreign ~from:c_lib
    "ada_base_type_decl_p_private_completion"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_is_inherited_primitive = foreign ~from:c_lib
    "ada_base_type_decl_p_is_inherited_primitive"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_get_record_representation_clause = foreign ~from:c_lib
    "ada_base_type_decl_p_get_record_representation_clause"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_get_enum_representation_clause = foreign ~from:c_lib
    "ada_base_type_decl_p_get_enum_representation_clause"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_is_record_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_record_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_is_array_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_array_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_find_derived_types = foreign ~from:c_lib
    "ada_base_type_decl_p_find_derived_types"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let base_type_decl_p_is_real_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_real_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_is_float_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_float_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_is_fixed_point = foreign ~from:c_lib
    "ada_base_type_decl_p_is_fixed_point"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_is_enum_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_enum_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_is_access_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_access_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_is_char_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_char_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_discrete_range = foreign ~from:c_lib
    "ada_base_type_decl_p_discrete_range"
    (ptr EntityStruct.c_type
    @-> ptr DiscreteRangeStruct.c_type
    @-> raisable int)

  let base_type_decl_p_is_discrete_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_discrete_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_is_int_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_int_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_accessed_type = foreign ~from:c_lib
    "ada_base_type_decl_p_accessed_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_is_tagged_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_tagged_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_base_type = foreign ~from:c_lib
    "ada_base_type_decl_p_base_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_base_types = foreign ~from:c_lib
    "ada_base_type_decl_p_base_types"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let base_type_decl_p_find_all_derived_types = foreign ~from:c_lib
    "ada_base_type_decl_p_find_all_derived_types"
    (ptr EntityStruct.c_type
        
    @-> AnalysisUnitArrayStruct.c_type
        
    @-> bool
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let base_type_decl_p_comp_type = foreign ~from:c_lib
    "ada_base_type_decl_p_comp_type"
    (ptr EntityStruct.c_type
        
    @-> bool
        
    @-> ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_index_type = foreign ~from:c_lib
    "ada_base_type_decl_p_index_type"
    (ptr EntityStruct.c_type
        
    @-> int
        
    @-> ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_is_derived_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_derived_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_is_interface_type = foreign ~from:c_lib
    "ada_base_type_decl_p_is_interface_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_matching_type = foreign ~from:c_lib
    "ada_base_type_decl_p_matching_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_canonical_type = foreign ~from:c_lib
    "ada_base_type_decl_p_canonical_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_previous_part = foreign ~from:c_lib
    "ada_base_type_decl_p_previous_part"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_next_part = foreign ~from:c_lib
    "ada_base_type_decl_p_next_part"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_full_view = foreign ~from:c_lib
    "ada_base_type_decl_p_full_view"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_is_definite_subtype = foreign ~from:c_lib
    "ada_base_type_decl_p_is_definite_subtype"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_is_private = foreign ~from:c_lib
    "ada_base_type_decl_p_is_private"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let base_type_decl_p_discriminants_list = foreign ~from:c_lib
    "ada_base_type_decl_p_discriminants_list"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let base_type_decl_p_root_type = foreign ~from:c_lib
    "ada_base_type_decl_p_root_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_type_decl_p_shapes = foreign ~from:c_lib
    "ada_base_type_decl_p_shapes"
    (ptr EntityStruct.c_type
        
    @-> bool
        
    @-> ptr EntityStruct.c_type
    @-> ptr ShapeArrayStruct.c_type
    @-> raisable int)

  let base_subtype_decl_p_get_type = foreign ~from:c_lib
    "ada_base_subtype_decl_p_get_type"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subtype_decl_f_subtype = foreign ~from:c_lib
    "ada_subtype_decl_f_subtype"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let incomplete_type_decl_f_discriminants = foreign ~from:c_lib
    "ada_incomplete_type_decl_f_discriminants"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let incomplete_tagged_type_decl_f_has_abstract = foreign ~from:c_lib
    "ada_incomplete_tagged_type_decl_f_has_abstract"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let protected_type_decl_f_discriminants = foreign ~from:c_lib
    "ada_protected_type_decl_f_discriminants"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let protected_type_decl_f_interfaces = foreign ~from:c_lib
    "ada_protected_type_decl_f_interfaces"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let protected_type_decl_f_definition = foreign ~from:c_lib
    "ada_protected_type_decl_f_definition"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let task_type_decl_f_discriminants = foreign ~from:c_lib
    "ada_task_type_decl_f_discriminants"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let task_type_decl_f_definition = foreign ~from:c_lib
    "ada_task_type_decl_f_definition"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let type_decl_f_discriminants = foreign ~from:c_lib
    "ada_type_decl_f_discriminants"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let type_decl_f_type_def = foreign ~from:c_lib
    "ada_type_decl_f_type_def"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let type_decl_p_get_primitives = foreign ~from:c_lib
    "ada_type_decl_p_get_primitives"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let basic_subp_decl_p_subp_decl_spec = foreign ~from:c_lib
    "ada_basic_subp_decl_p_subp_decl_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let classic_subp_decl_f_overriding = foreign ~from:c_lib
    "ada_classic_subp_decl_f_overriding"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let classic_subp_decl_f_subp_spec = foreign ~from:c_lib
    "ada_classic_subp_decl_f_subp_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let classic_subp_decl_p_body_part = foreign ~from:c_lib
    "ada_classic_subp_decl_p_body_part"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let formal_subp_decl_f_default_expr = foreign ~from:c_lib
    "ada_formal_subp_decl_f_default_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_decl_f_overriding = foreign ~from:c_lib
    "ada_entry_decl_f_overriding"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_decl_f_spec = foreign ~from:c_lib
    "ada_entry_decl_f_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_decl_p_body_part = foreign ~from:c_lib
    "ada_entry_decl_p_body_part"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let enum_literal_decl_f_name = foreign ~from:c_lib
    "ada_enum_literal_decl_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let enum_literal_decl_p_enum_type = foreign ~from:c_lib
    "ada_enum_literal_decl_p_enum_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_internal_f_subp_spec = foreign ~from:c_lib
    "ada_generic_subp_internal_f_subp_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let body_node_p_previous_part = foreign ~from:c_lib
    "ada_body_node_p_previous_part"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let body_node_p_decl_part = foreign ~from:c_lib
    "ada_body_node_p_decl_part"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let body_node_p_subunit_root = foreign ~from:c_lib
    "ada_body_node_p_subunit_root"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_subp_body_f_overriding = foreign ~from:c_lib
    "ada_base_subp_body_f_overriding"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_subp_body_f_subp_spec = foreign ~from:c_lib
    "ada_base_subp_body_f_subp_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let expr_function_f_expr = foreign ~from:c_lib
    "ada_expr_function_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subp_body_f_decls = foreign ~from:c_lib
    "ada_subp_body_f_decls"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subp_body_f_stmts = foreign ~from:c_lib
    "ada_subp_body_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subp_body_f_end_name = foreign ~from:c_lib
    "ada_subp_body_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subp_renaming_decl_f_renames = foreign ~from:c_lib
    "ada_subp_renaming_decl_f_renames"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let body_stub_p_syntactic_fully_qualified_name = foreign ~from:c_lib
    "ada_body_stub_p_syntactic_fully_qualified_name"
    (ptr EntityStruct.c_type
    @-> ptr UnboundedTextTypeArrayStruct.c_type
    @-> raisable int)

  let package_body_stub_f_name = foreign ~from:c_lib
    "ada_package_body_stub_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let protected_body_stub_f_name = foreign ~from:c_lib
    "ada_protected_body_stub_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subp_body_stub_f_overriding = foreign ~from:c_lib
    "ada_subp_body_stub_f_overriding"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subp_body_stub_f_subp_spec = foreign ~from:c_lib
    "ada_subp_body_stub_f_subp_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let task_body_stub_f_name = foreign ~from:c_lib
    "ada_task_body_stub_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_body_f_entry_name = foreign ~from:c_lib
    "ada_entry_body_f_entry_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_body_f_index_spec = foreign ~from:c_lib
    "ada_entry_body_f_index_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_body_f_params = foreign ~from:c_lib
    "ada_entry_body_f_params"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_body_f_barrier = foreign ~from:c_lib
    "ada_entry_body_f_barrier"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_body_f_decls = foreign ~from:c_lib
    "ada_entry_body_f_decls"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_body_f_stmts = foreign ~from:c_lib
    "ada_entry_body_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_body_f_end_name = foreign ~from:c_lib
    "ada_entry_body_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let package_body_f_package_name = foreign ~from:c_lib
    "ada_package_body_f_package_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let package_body_f_decls = foreign ~from:c_lib
    "ada_package_body_f_decls"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let package_body_f_stmts = foreign ~from:c_lib
    "ada_package_body_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let package_body_f_end_name = foreign ~from:c_lib
    "ada_package_body_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let protected_body_f_name = foreign ~from:c_lib
    "ada_protected_body_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let protected_body_f_decls = foreign ~from:c_lib
    "ada_protected_body_f_decls"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let protected_body_f_end_name = foreign ~from:c_lib
    "ada_protected_body_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let task_body_f_name = foreign ~from:c_lib
    "ada_task_body_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let task_body_f_decls = foreign ~from:c_lib
    "ada_task_body_f_decls"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let task_body_f_stmts = foreign ~from:c_lib
    "ada_task_body_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let task_body_f_end_name = foreign ~from:c_lib
    "ada_task_body_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_index_spec_f_id = foreign ~from:c_lib
    "ada_entry_index_spec_f_id"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let entry_index_spec_f_subtype = foreign ~from:c_lib
    "ada_entry_index_spec_f_subtype"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let exception_decl_f_ids = foreign ~from:c_lib
    "ada_exception_decl_f_ids"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let exception_decl_f_renames = foreign ~from:c_lib
    "ada_exception_decl_f_renames"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let exception_handler_f_exception_name = foreign ~from:c_lib
    "ada_exception_handler_f_exception_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let exception_handler_f_handled_exceptions = foreign ~from:c_lib
    "ada_exception_handler_f_handled_exceptions"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let exception_handler_f_stmts = foreign ~from:c_lib
    "ada_exception_handler_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let for_loop_var_decl_f_id = foreign ~from:c_lib
    "ada_for_loop_var_decl_f_id"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let for_loop_var_decl_f_id_type = foreign ~from:c_lib
    "ada_for_loop_var_decl_f_id_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_decl_f_formal_part = foreign ~from:c_lib
    "ada_generic_decl_f_formal_part"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_package_decl_f_package_decl = foreign ~from:c_lib
    "ada_generic_package_decl_f_package_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_package_decl_p_body_part = foreign ~from:c_lib
    "ada_generic_package_decl_p_body_part"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_decl_f_subp_decl = foreign ~from:c_lib
    "ada_generic_subp_decl_f_subp_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_decl_p_body_part = foreign ~from:c_lib
    "ada_generic_subp_decl_p_body_part"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_instantiation_p_designated_generic_decl = foreign ~from:c_lib
    "ada_generic_instantiation_p_designated_generic_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_package_instantiation_f_name = foreign ~from:c_lib
    "ada_generic_package_instantiation_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_package_instantiation_f_generic_pkg_name = foreign ~from:c_lib
    "ada_generic_package_instantiation_f_generic_pkg_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_package_instantiation_f_params = foreign ~from:c_lib
    "ada_generic_package_instantiation_f_params"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_instantiation_f_overriding = foreign ~from:c_lib
    "ada_generic_subp_instantiation_f_overriding"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_instantiation_f_kind = foreign ~from:c_lib
    "ada_generic_subp_instantiation_f_kind"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_instantiation_f_subp_name = foreign ~from:c_lib
    "ada_generic_subp_instantiation_f_subp_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_instantiation_f_generic_subp_name = foreign ~from:c_lib
    "ada_generic_subp_instantiation_f_generic_subp_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_instantiation_f_params = foreign ~from:c_lib
    "ada_generic_subp_instantiation_f_params"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_instantiation_p_designated_subp = foreign ~from:c_lib
    "ada_generic_subp_instantiation_p_designated_subp"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_package_renaming_decl_f_name = foreign ~from:c_lib
    "ada_generic_package_renaming_decl_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_package_renaming_decl_f_renames = foreign ~from:c_lib
    "ada_generic_package_renaming_decl_f_renames"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_renaming_decl_f_kind = foreign ~from:c_lib
    "ada_generic_subp_renaming_decl_f_kind"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_renaming_decl_f_name = foreign ~from:c_lib
    "ada_generic_subp_renaming_decl_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let generic_subp_renaming_decl_f_renames = foreign ~from:c_lib
    "ada_generic_subp_renaming_decl_f_renames"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let label_decl_f_name = foreign ~from:c_lib
    "ada_label_decl_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let named_stmt_decl_f_name = foreign ~from:c_lib
    "ada_named_stmt_decl_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let number_decl_f_ids = foreign ~from:c_lib
    "ada_number_decl_f_ids"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let number_decl_f_expr = foreign ~from:c_lib
    "ada_number_decl_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let object_decl_f_ids = foreign ~from:c_lib
    "ada_object_decl_f_ids"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let object_decl_f_has_aliased = foreign ~from:c_lib
    "ada_object_decl_f_has_aliased"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let object_decl_f_has_constant = foreign ~from:c_lib
    "ada_object_decl_f_has_constant"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let object_decl_f_mode = foreign ~from:c_lib
    "ada_object_decl_f_mode"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let object_decl_f_type_expr = foreign ~from:c_lib
    "ada_object_decl_f_type_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let object_decl_f_default_expr = foreign ~from:c_lib
    "ada_object_decl_f_default_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let object_decl_f_renaming_clause = foreign ~from:c_lib
    "ada_object_decl_f_renaming_clause"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let object_decl_p_public_part_decl = foreign ~from:c_lib
    "ada_object_decl_p_public_part_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let object_decl_p_private_part_decl = foreign ~from:c_lib
    "ada_object_decl_p_private_part_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let package_renaming_decl_f_name = foreign ~from:c_lib
    "ada_package_renaming_decl_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let package_renaming_decl_f_renames = foreign ~from:c_lib
    "ada_package_renaming_decl_f_renames"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let package_renaming_decl_p_renamed_package = foreign ~from:c_lib
    "ada_package_renaming_decl_p_renamed_package"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let package_renaming_decl_p_final_renamed_package = foreign ~from:c_lib
    "ada_package_renaming_decl_p_final_renamed_package"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let single_protected_decl_f_name = foreign ~from:c_lib
    "ada_single_protected_decl_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let single_protected_decl_f_interfaces = foreign ~from:c_lib
    "ada_single_protected_decl_f_interfaces"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let single_protected_decl_f_definition = foreign ~from:c_lib
    "ada_single_protected_decl_f_definition"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let single_task_decl_f_task_type = foreign ~from:c_lib
    "ada_single_task_decl_f_task_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let case_stmt_alternative_f_choices = foreign ~from:c_lib
    "ada_case_stmt_alternative_f_choices"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let case_stmt_alternative_f_stmts = foreign ~from:c_lib
    "ada_case_stmt_alternative_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let compilation_unit_f_prelude = foreign ~from:c_lib
    "ada_compilation_unit_f_prelude"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let compilation_unit_f_body = foreign ~from:c_lib
    "ada_compilation_unit_f_body"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let compilation_unit_f_pragmas = foreign ~from:c_lib
    "ada_compilation_unit_f_pragmas"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let compilation_unit_p_syntactic_fully_qualified_name = foreign ~from:c_lib
    "ada_compilation_unit_p_syntactic_fully_qualified_name"
    (ptr EntityStruct.c_type
    @-> ptr UnboundedTextTypeArrayStruct.c_type
    @-> raisable int)

  let compilation_unit_p_unit_kind = foreign ~from:c_lib
    "ada_compilation_unit_p_unit_kind"
    (ptr EntityStruct.c_type
    @-> ptr AnalysisUnitKind.c_type
    @-> raisable int)

  let compilation_unit_p_withed_units = foreign ~from:c_lib
    "ada_compilation_unit_p_withed_units"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let compilation_unit_p_imported_units = foreign ~from:c_lib
    "ada_compilation_unit_p_imported_units"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let compilation_unit_p_unit_dependencies = foreign ~from:c_lib
    "ada_compilation_unit_p_unit_dependencies"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let compilation_unit_p_decl = foreign ~from:c_lib
    "ada_compilation_unit_p_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let compilation_unit_p_is_preelaborable = foreign ~from:c_lib
    "ada_compilation_unit_p_is_preelaborable"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr bool
    @-> raisable int)

  let component_clause_f_id = foreign ~from:c_lib
    "ada_component_clause_f_id"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let component_clause_f_position = foreign ~from:c_lib
    "ada_component_clause_f_position"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let component_clause_f_range = foreign ~from:c_lib
    "ada_component_clause_f_range"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let component_def_f_has_aliased = foreign ~from:c_lib
    "ada_component_def_f_has_aliased"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let component_def_f_has_constant = foreign ~from:c_lib
    "ada_component_def_f_has_constant"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let component_def_f_type_expr = foreign ~from:c_lib
    "ada_component_def_f_type_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let constant_node_p_as_bool = foreign ~from:c_lib
    "ada_constant_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let delta_constraint_f_digits = foreign ~from:c_lib
    "ada_delta_constraint_f_digits"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let delta_constraint_f_range = foreign ~from:c_lib
    "ada_delta_constraint_f_range"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let digits_constraint_f_digits = foreign ~from:c_lib
    "ada_digits_constraint_f_digits"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let digits_constraint_f_range = foreign ~from:c_lib
    "ada_digits_constraint_f_range"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let discriminant_constraint_f_constraints = foreign ~from:c_lib
    "ada_discriminant_constraint_f_constraints"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let index_constraint_f_constraints = foreign ~from:c_lib
    "ada_index_constraint_f_constraints"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let range_constraint_f_range = foreign ~from:c_lib
    "ada_range_constraint_f_range"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let declarative_part_f_decls = foreign ~from:c_lib
    "ada_declarative_part_f_decls"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let elsif_expr_part_f_cond_expr = foreign ~from:c_lib
    "ada_elsif_expr_part_f_cond_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let elsif_expr_part_f_then_expr = foreign ~from:c_lib
    "ada_elsif_expr_part_f_then_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let elsif_stmt_part_f_cond_expr = foreign ~from:c_lib
    "ada_elsif_stmt_part_f_cond_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let elsif_stmt_part_f_stmts = foreign ~from:c_lib
    "ada_elsif_stmt_part_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let expr_p_expression_type = foreign ~from:c_lib
    "ada_expr_p_expression_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let expr_p_is_dynamically_tagged = foreign ~from:c_lib
    "ada_expr_p_is_dynamically_tagged"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr bool
    @-> raisable int)

  let expr_p_is_dispatching_call = foreign ~from:c_lib
    "ada_expr_p_is_dispatching_call"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr bool
    @-> raisable int)

  let expr_p_is_static_expr = foreign ~from:c_lib
    "ada_expr_p_is_static_expr"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr bool
    @-> raisable int)

  let expr_p_first_corresponding_decl = foreign ~from:c_lib
    "ada_expr_p_first_corresponding_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let expr_p_eval_as_int = foreign ~from:c_lib
    "ada_expr_p_eval_as_int"
    (ptr EntityStruct.c_type
    @-> ptr BigInteger.c_type
    @-> raisable int)

  let expr_p_eval_as_int_in_env = foreign ~from:c_lib
    "ada_expr_p_eval_as_int_in_env"
    (ptr EntityStruct.c_type
        
    @-> SubstitutionArrayStruct.c_type
    @-> ptr BigInteger.c_type
    @-> raisable int)

  let expr_p_eval_as_string = foreign ~from:c_lib
    "ada_expr_p_eval_as_string"
    (ptr EntityStruct.c_type
    @-> ptr TextTypeStruct.c_type
    @-> raisable int)

  let expr_p_eval_as_string_in_env = foreign ~from:c_lib
    "ada_expr_p_eval_as_string_in_env"
    (ptr EntityStruct.c_type
        
    @-> SubstitutionArrayStruct.c_type
    @-> ptr TextTypeStruct.c_type
    @-> raisable int)

  let expr_p_matching_nodes = foreign ~from:c_lib
    "ada_expr_p_matching_nodes"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let abstract_state_decl_expr_f_state_decl = foreign ~from:c_lib
    "ada_abstract_state_decl_expr_f_state_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let allocator_f_subpool = foreign ~from:c_lib
    "ada_allocator_f_subpool"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let allocator_f_type_or_expr = foreign ~from:c_lib
    "ada_allocator_f_type_or_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let allocator_p_get_allocated_type = foreign ~from:c_lib
    "ada_allocator_p_get_allocated_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_aggregate_f_ancestor_expr = foreign ~from:c_lib
    "ada_base_aggregate_f_ancestor_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_aggregate_f_assocs = foreign ~from:c_lib
    "ada_base_aggregate_f_assocs"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let bin_op_f_left = foreign ~from:c_lib
    "ada_bin_op_f_left"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let bin_op_f_op = foreign ~from:c_lib
    "ada_bin_op_f_op"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let bin_op_f_right = foreign ~from:c_lib
    "ada_bin_op_f_right"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let case_expr_alternative_f_choices = foreign ~from:c_lib
    "ada_case_expr_alternative_f_choices"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let case_expr_alternative_f_expr = foreign ~from:c_lib
    "ada_case_expr_alternative_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let cond_expr_p_dependent_exprs = foreign ~from:c_lib
    "ada_cond_expr_p_dependent_exprs"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let case_expr_f_expr = foreign ~from:c_lib
    "ada_case_expr_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let case_expr_f_cases = foreign ~from:c_lib
    "ada_case_expr_f_cases"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let if_expr_f_cond_expr = foreign ~from:c_lib
    "ada_if_expr_f_cond_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let if_expr_f_then_expr = foreign ~from:c_lib
    "ada_if_expr_f_then_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let if_expr_f_alternatives = foreign ~from:c_lib
    "ada_if_expr_f_alternatives"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let if_expr_f_else_expr = foreign ~from:c_lib
    "ada_if_expr_f_else_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let contract_cases_f_contract_cases = foreign ~from:c_lib
    "ada_contract_cases_f_contract_cases"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let decl_expr_f_decls = foreign ~from:c_lib
    "ada_decl_expr_f_decls"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let decl_expr_f_expr = foreign ~from:c_lib
    "ada_decl_expr_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let membership_expr_f_expr = foreign ~from:c_lib
    "ada_membership_expr_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let membership_expr_f_op = foreign ~from:c_lib
    "ada_membership_expr_f_op"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let membership_expr_f_membership_exprs = foreign ~from:c_lib
    "ada_membership_expr_f_membership_exprs"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let name_p_enclosing_defining_name = foreign ~from:c_lib
    "ada_name_p_enclosing_defining_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let name_p_is_defining = foreign ~from:c_lib
    "ada_name_p_is_defining"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let name_p_name_is = foreign ~from:c_lib
    "ada_name_p_name_is"
    (ptr EntityStruct.c_type
        
    @-> ptr Symbol.c_type
    @-> ptr bool
    @-> raisable int)

  let name_p_is_direct_call = foreign ~from:c_lib
    "ada_name_p_is_direct_call"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let name_p_is_access_call = foreign ~from:c_lib
    "ada_name_p_is_access_call"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let name_p_is_call = foreign ~from:c_lib
    "ada_name_p_is_call"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let name_p_is_dot_call = foreign ~from:c_lib
    "ada_name_p_is_dot_call"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr bool
    @-> raisable int)

  let name_p_failsafe_referenced_def_name = foreign ~from:c_lib
    "ada_name_p_failsafe_referenced_def_name"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr RefdDefStruct.c_type
    @-> raisable int)

  let name_p_referenced_defining_name = foreign ~from:c_lib
    "ada_name_p_referenced_defining_name"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let name_p_all_env_elements = foreign ~from:c_lib
    "ada_name_p_all_env_elements"
    (ptr EntityStruct.c_type
        
    @-> bool
        
    @-> ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let name_p_called_subp_spec = foreign ~from:c_lib
    "ada_name_p_called_subp_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let name_p_referenced_decl = foreign ~from:c_lib
    "ada_name_p_referenced_decl"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let name_p_failsafe_referenced_decl = foreign ~from:c_lib
    "ada_name_p_failsafe_referenced_decl"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr RefdDeclStruct.c_type
    @-> raisable int)

  let name_p_referenced_decl_internal = foreign ~from:c_lib
    "ada_name_p_referenced_decl_internal"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr RefdDeclStruct.c_type
    @-> raisable int)

  let name_p_name_designated_type = foreign ~from:c_lib
    "ada_name_p_name_designated_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let name_p_is_static_subtype = foreign ~from:c_lib
    "ada_name_p_is_static_subtype"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr bool
    @-> raisable int)

  let name_p_name_matches = foreign ~from:c_lib
    "ada_name_p_name_matches"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let name_p_relative_name = foreign ~from:c_lib
    "ada_name_p_relative_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let name_p_is_operator_name = foreign ~from:c_lib
    "ada_name_p_is_operator_name"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let name_p_is_write_reference = foreign ~from:c_lib
    "ada_name_p_is_write_reference"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr bool
    @-> raisable int)

  let name_p_is_static_call = foreign ~from:c_lib
    "ada_name_p_is_static_call"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr bool
    @-> raisable int)

  let name_p_as_symbol_array = foreign ~from:c_lib
    "ada_name_p_as_symbol_array"
    (ptr EntityStruct.c_type
    @-> ptr UnboundedTextTypeArrayStruct.c_type
    @-> raisable int)

  let attribute_ref_f_prefix = foreign ~from:c_lib
    "ada_attribute_ref_f_prefix"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let attribute_ref_f_attribute = foreign ~from:c_lib
    "ada_attribute_ref_f_attribute"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let attribute_ref_f_args = foreign ~from:c_lib
    "ada_attribute_ref_f_args"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let call_expr_f_name = foreign ~from:c_lib
    "ada_call_expr_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let call_expr_f_suffix = foreign ~from:c_lib
    "ada_call_expr_f_suffix"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let call_expr_p_is_array_slice = foreign ~from:c_lib
    "ada_call_expr_p_is_array_slice"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let defining_name_f_name = foreign ~from:c_lib
    "ada_defining_name_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let defining_name_p_basic_decl = foreign ~from:c_lib
    "ada_defining_name_p_basic_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let defining_name_p_find_refs = foreign ~from:c_lib
    "ada_defining_name_p_find_refs"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr RefResultArrayStruct.c_type
    @-> raisable int)

  let defining_name_p_find_all_references = foreign ~from:c_lib
    "ada_defining_name_p_find_all_references"
    (ptr EntityStruct.c_type
        
    @-> AnalysisUnitArrayStruct.c_type
        
    @-> bool
        
    @-> bool
    @-> ptr RefResultArrayStruct.c_type
    @-> raisable int)

  let defining_name_p_find_all_calls = foreign ~from:c_lib
    "ada_defining_name_p_find_all_calls"
    (ptr EntityStruct.c_type
        
    @-> AnalysisUnitArrayStruct.c_type
        
    @-> bool
        
    @-> bool
    @-> ptr RefResultArrayStruct.c_type
    @-> raisable int)

  let defining_name_p_next_part = foreign ~from:c_lib
    "ada_defining_name_p_next_part"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let defining_name_p_previous_part = foreign ~from:c_lib
    "ada_defining_name_p_previous_part"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let defining_name_p_canonical_part = foreign ~from:c_lib
    "ada_defining_name_p_canonical_part"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let discrete_subtype_name_f_subtype = foreign ~from:c_lib
    "ada_discrete_subtype_name_f_subtype"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let dotted_name_f_prefix = foreign ~from:c_lib
    "ada_dotted_name_f_prefix"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let dotted_name_f_suffix = foreign ~from:c_lib
    "ada_dotted_name_f_suffix"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let end_name_f_name = foreign ~from:c_lib
    "ada_end_name_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let end_name_p_basic_decl = foreign ~from:c_lib
    "ada_end_name_p_basic_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let explicit_deref_f_prefix = foreign ~from:c_lib
    "ada_explicit_deref_f_prefix"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let qual_expr_f_prefix = foreign ~from:c_lib
    "ada_qual_expr_f_prefix"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let qual_expr_f_suffix = foreign ~from:c_lib
    "ada_qual_expr_f_suffix"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let single_tok_node_p_canonical_text = foreign ~from:c_lib
    "ada_single_tok_node_p_canonical_text"
    (ptr EntityStruct.c_type
    @-> ptr Symbol.c_type
    @-> raisable int)

  let char_literal_p_denoted_value = foreign ~from:c_lib
    "ada_char_literal_p_denoted_value"
    (ptr EntityStruct.c_type
    @-> ptr Character.c_type
    @-> raisable int)

  let string_literal_p_denoted_value = foreign ~from:c_lib
    "ada_string_literal_p_denoted_value"
    (ptr EntityStruct.c_type
    @-> ptr TextTypeStruct.c_type
    @-> raisable int)

  let int_literal_p_denoted_value = foreign ~from:c_lib
    "ada_int_literal_p_denoted_value"
    (ptr EntityStruct.c_type
    @-> ptr BigInteger.c_type
    @-> raisable int)

  let paren_expr_f_expr = foreign ~from:c_lib
    "ada_paren_expr_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let quantified_expr_f_quantifier = foreign ~from:c_lib
    "ada_quantified_expr_f_quantifier"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let quantified_expr_f_loop_spec = foreign ~from:c_lib
    "ada_quantified_expr_f_loop_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let quantified_expr_f_expr = foreign ~from:c_lib
    "ada_quantified_expr_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let raise_expr_f_exception_name = foreign ~from:c_lib
    "ada_raise_expr_f_exception_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let raise_expr_f_error_message = foreign ~from:c_lib
    "ada_raise_expr_f_error_message"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let un_op_f_op = foreign ~from:c_lib
    "ada_un_op_f_op"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let un_op_f_expr = foreign ~from:c_lib
    "ada_un_op_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let handled_stmts_f_stmts = foreign ~from:c_lib
    "ada_handled_stmts_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let handled_stmts_f_exceptions = foreign ~from:c_lib
    "ada_handled_stmts_f_exceptions"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let library_item_f_has_private = foreign ~from:c_lib
    "ada_library_item_f_has_private"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let library_item_f_item = foreign ~from:c_lib
    "ada_library_item_f_item"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let limited_node_p_as_bool = foreign ~from:c_lib
    "ada_limited_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let for_loop_spec_f_var_decl = foreign ~from:c_lib
    "ada_for_loop_spec_f_var_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let for_loop_spec_f_loop_type = foreign ~from:c_lib
    "ada_for_loop_spec_f_loop_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let for_loop_spec_f_has_reverse = foreign ~from:c_lib
    "ada_for_loop_spec_f_has_reverse"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let for_loop_spec_f_iter_expr = foreign ~from:c_lib
    "ada_for_loop_spec_f_iter_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let while_loop_spec_f_expr = foreign ~from:c_lib
    "ada_while_loop_spec_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let multi_abstract_state_decl_f_decls = foreign ~from:c_lib
    "ada_multi_abstract_state_decl_f_decls"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let not_null_p_as_bool = foreign ~from:c_lib
    "ada_not_null_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let params_f_params = foreign ~from:c_lib
    "ada_params_f_params"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let paren_abstract_state_decl_f_decl = foreign ~from:c_lib
    "ada_paren_abstract_state_decl_f_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let pragma_node_f_id = foreign ~from:c_lib
    "ada_pragma_node_f_id"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let pragma_node_f_args = foreign ~from:c_lib
    "ada_pragma_node_f_args"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let pragma_node_p_associated_decls = foreign ~from:c_lib
    "ada_pragma_node_p_associated_decls"
    (ptr EntityStruct.c_type
    @-> ptr AdaNodeArrayStruct.c_type
    @-> raisable int)

  let private_node_p_as_bool = foreign ~from:c_lib
    "ada_private_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let protected_def_f_public_part = foreign ~from:c_lib
    "ada_protected_def_f_public_part"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let protected_def_f_private_part = foreign ~from:c_lib
    "ada_protected_def_f_private_part"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let protected_def_f_end_name = foreign ~from:c_lib
    "ada_protected_def_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let protected_node_p_as_bool = foreign ~from:c_lib
    "ada_protected_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let range_spec_f_range = foreign ~from:c_lib
    "ada_range_spec_f_range"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let renaming_clause_f_renamed_object = foreign ~from:c_lib
    "ada_renaming_clause_f_renamed_object"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let reverse_node_p_as_bool = foreign ~from:c_lib
    "ada_reverse_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let select_when_part_f_cond_expr = foreign ~from:c_lib
    "ada_select_when_part_f_cond_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let select_when_part_f_stmts = foreign ~from:c_lib
    "ada_select_when_part_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let accept_stmt_f_name = foreign ~from:c_lib
    "ada_accept_stmt_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let accept_stmt_f_entry_index_expr = foreign ~from:c_lib
    "ada_accept_stmt_f_entry_index_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let accept_stmt_f_params = foreign ~from:c_lib
    "ada_accept_stmt_f_params"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let accept_stmt_with_stmts_f_stmts = foreign ~from:c_lib
    "ada_accept_stmt_with_stmts_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let accept_stmt_with_stmts_f_end_name = foreign ~from:c_lib
    "ada_accept_stmt_with_stmts_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_loop_stmt_f_spec = foreign ~from:c_lib
    "ada_base_loop_stmt_f_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_loop_stmt_f_stmts = foreign ~from:c_lib
    "ada_base_loop_stmt_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let base_loop_stmt_f_end_name = foreign ~from:c_lib
    "ada_base_loop_stmt_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let begin_block_f_stmts = foreign ~from:c_lib
    "ada_begin_block_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let begin_block_f_end_name = foreign ~from:c_lib
    "ada_begin_block_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let decl_block_f_decls = foreign ~from:c_lib
    "ada_decl_block_f_decls"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let decl_block_f_stmts = foreign ~from:c_lib
    "ada_decl_block_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let decl_block_f_end_name = foreign ~from:c_lib
    "ada_decl_block_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let case_stmt_f_expr = foreign ~from:c_lib
    "ada_case_stmt_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let case_stmt_f_alternatives = foreign ~from:c_lib
    "ada_case_stmt_f_alternatives"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let extended_return_stmt_f_decl = foreign ~from:c_lib
    "ada_extended_return_stmt_f_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let extended_return_stmt_f_stmts = foreign ~from:c_lib
    "ada_extended_return_stmt_f_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let if_stmt_f_cond_expr = foreign ~from:c_lib
    "ada_if_stmt_f_cond_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let if_stmt_f_then_stmts = foreign ~from:c_lib
    "ada_if_stmt_f_then_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let if_stmt_f_alternatives = foreign ~from:c_lib
    "ada_if_stmt_f_alternatives"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let if_stmt_f_else_stmts = foreign ~from:c_lib
    "ada_if_stmt_f_else_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let named_stmt_f_decl = foreign ~from:c_lib
    "ada_named_stmt_f_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let named_stmt_f_stmt = foreign ~from:c_lib
    "ada_named_stmt_f_stmt"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let select_stmt_f_guards = foreign ~from:c_lib
    "ada_select_stmt_f_guards"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let select_stmt_f_else_stmts = foreign ~from:c_lib
    "ada_select_stmt_f_else_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let select_stmt_f_abort_stmts = foreign ~from:c_lib
    "ada_select_stmt_f_abort_stmts"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let abort_stmt_f_names = foreign ~from:c_lib
    "ada_abort_stmt_f_names"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let assign_stmt_f_dest = foreign ~from:c_lib
    "ada_assign_stmt_f_dest"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let assign_stmt_f_expr = foreign ~from:c_lib
    "ada_assign_stmt_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let call_stmt_f_call = foreign ~from:c_lib
    "ada_call_stmt_f_call"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let delay_stmt_f_has_until = foreign ~from:c_lib
    "ada_delay_stmt_f_has_until"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let delay_stmt_f_expr = foreign ~from:c_lib
    "ada_delay_stmt_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let exit_stmt_f_loop_name = foreign ~from:c_lib
    "ada_exit_stmt_f_loop_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let exit_stmt_f_cond_expr = foreign ~from:c_lib
    "ada_exit_stmt_f_cond_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let goto_stmt_f_label_name = foreign ~from:c_lib
    "ada_goto_stmt_f_label_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let label_f_decl = foreign ~from:c_lib
    "ada_label_f_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let raise_stmt_f_exception_name = foreign ~from:c_lib
    "ada_raise_stmt_f_exception_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let raise_stmt_f_error_message = foreign ~from:c_lib
    "ada_raise_stmt_f_error_message"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let requeue_stmt_f_call_name = foreign ~from:c_lib
    "ada_requeue_stmt_f_call_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let requeue_stmt_f_has_abort = foreign ~from:c_lib
    "ada_requeue_stmt_f_has_abort"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let return_stmt_f_return_expr = foreign ~from:c_lib
    "ada_return_stmt_f_return_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subunit_f_name = foreign ~from:c_lib
    "ada_subunit_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subunit_f_body = foreign ~from:c_lib
    "ada_subunit_f_body"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subunit_p_body_root = foreign ~from:c_lib
    "ada_subunit_p_body_root"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let synchronized_node_p_as_bool = foreign ~from:c_lib
    "ada_synchronized_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let tagged_node_p_as_bool = foreign ~from:c_lib
    "ada_tagged_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let task_def_f_interfaces = foreign ~from:c_lib
    "ada_task_def_f_interfaces"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let task_def_f_public_part = foreign ~from:c_lib
    "ada_task_def_f_public_part"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let task_def_f_private_part = foreign ~from:c_lib
    "ada_task_def_f_private_part"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let task_def_f_end_name = foreign ~from:c_lib
    "ada_task_def_f_end_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let access_def_f_has_not_null = foreign ~from:c_lib
    "ada_access_def_f_has_not_null"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let access_to_subp_def_f_has_protected = foreign ~from:c_lib
    "ada_access_to_subp_def_f_has_protected"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let access_to_subp_def_f_subp_spec = foreign ~from:c_lib
    "ada_access_to_subp_def_f_subp_spec"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let anonymous_type_access_def_f_type_decl = foreign ~from:c_lib
    "ada_anonymous_type_access_def_f_type_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let type_access_def_f_has_all = foreign ~from:c_lib
    "ada_type_access_def_f_has_all"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let type_access_def_f_has_constant = foreign ~from:c_lib
    "ada_type_access_def_f_has_constant"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let type_access_def_f_subtype_indication = foreign ~from:c_lib
    "ada_type_access_def_f_subtype_indication"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let array_type_def_f_indices = foreign ~from:c_lib
    "ada_array_type_def_f_indices"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let array_type_def_f_component_type = foreign ~from:c_lib
    "ada_array_type_def_f_component_type"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let derived_type_def_f_has_abstract = foreign ~from:c_lib
    "ada_derived_type_def_f_has_abstract"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let derived_type_def_f_has_limited = foreign ~from:c_lib
    "ada_derived_type_def_f_has_limited"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let derived_type_def_f_has_synchronized = foreign ~from:c_lib
    "ada_derived_type_def_f_has_synchronized"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let derived_type_def_f_subtype_indication = foreign ~from:c_lib
    "ada_derived_type_def_f_subtype_indication"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let derived_type_def_f_interfaces = foreign ~from:c_lib
    "ada_derived_type_def_f_interfaces"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let derived_type_def_f_record_extension = foreign ~from:c_lib
    "ada_derived_type_def_f_record_extension"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let derived_type_def_f_has_with_private = foreign ~from:c_lib
    "ada_derived_type_def_f_has_with_private"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let enum_type_def_f_enum_literals = foreign ~from:c_lib
    "ada_enum_type_def_f_enum_literals"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let interface_type_def_f_interface_kind = foreign ~from:c_lib
    "ada_interface_type_def_f_interface_kind"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let interface_type_def_f_interfaces = foreign ~from:c_lib
    "ada_interface_type_def_f_interfaces"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let mod_int_type_def_f_expr = foreign ~from:c_lib
    "ada_mod_int_type_def_f_expr"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let private_type_def_f_has_abstract = foreign ~from:c_lib
    "ada_private_type_def_f_has_abstract"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let private_type_def_f_has_tagged = foreign ~from:c_lib
    "ada_private_type_def_f_has_tagged"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let private_type_def_f_has_limited = foreign ~from:c_lib
    "ada_private_type_def_f_has_limited"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let decimal_fixed_point_def_f_delta = foreign ~from:c_lib
    "ada_decimal_fixed_point_def_f_delta"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let decimal_fixed_point_def_f_digits = foreign ~from:c_lib
    "ada_decimal_fixed_point_def_f_digits"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let decimal_fixed_point_def_f_range = foreign ~from:c_lib
    "ada_decimal_fixed_point_def_f_range"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let floating_point_def_f_num_digits = foreign ~from:c_lib
    "ada_floating_point_def_f_num_digits"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let floating_point_def_f_range = foreign ~from:c_lib
    "ada_floating_point_def_f_range"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ordinary_fixed_point_def_f_delta = foreign ~from:c_lib
    "ada_ordinary_fixed_point_def_f_delta"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let ordinary_fixed_point_def_f_range = foreign ~from:c_lib
    "ada_ordinary_fixed_point_def_f_range"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let record_type_def_f_has_abstract = foreign ~from:c_lib
    "ada_record_type_def_f_has_abstract"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let record_type_def_f_has_tagged = foreign ~from:c_lib
    "ada_record_type_def_f_has_tagged"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let record_type_def_f_has_limited = foreign ~from:c_lib
    "ada_record_type_def_f_has_limited"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let record_type_def_f_record_def = foreign ~from:c_lib
    "ada_record_type_def_f_record_def"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let signed_int_type_def_f_range = foreign ~from:c_lib
    "ada_signed_int_type_def_f_range"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let type_expr_p_type_name = foreign ~from:c_lib
    "ada_type_expr_p_type_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let type_expr_p_designated_type_decl = foreign ~from:c_lib
    "ada_type_expr_p_designated_type_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let type_expr_p_designated_type_decl_from = foreign ~from:c_lib
    "ada_type_expr_p_designated_type_decl_from"
    (ptr EntityStruct.c_type
        
    @-> ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let anonymous_type_f_type_decl = foreign ~from:c_lib
    "ada_anonymous_type_f_type_decl"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subtype_indication_f_has_not_null = foreign ~from:c_lib
    "ada_subtype_indication_f_has_not_null"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subtype_indication_f_name = foreign ~from:c_lib
    "ada_subtype_indication_f_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subtype_indication_f_constraint = foreign ~from:c_lib
    "ada_subtype_indication_f_constraint"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let subtype_indication_p_is_static_subtype = foreign ~from:c_lib
    "ada_subtype_indication_p_is_static_subtype"
    (ptr EntityStruct.c_type
        
    @-> bool
    @-> ptr bool
    @-> raisable int)

  let unconstrained_array_index_f_subtype_indication = foreign ~from:c_lib
    "ada_unconstrained_array_index_f_subtype_indication"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let until_node_p_as_bool = foreign ~from:c_lib
    "ada_until_node_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)

  let use_package_clause_f_packages = foreign ~from:c_lib
    "ada_use_package_clause_f_packages"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let use_type_clause_f_has_all = foreign ~from:c_lib
    "ada_use_type_clause_f_has_all"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let use_type_clause_f_types = foreign ~from:c_lib
    "ada_use_type_clause_f_types"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let variant_f_choices = foreign ~from:c_lib
    "ada_variant_f_choices"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let variant_f_components = foreign ~from:c_lib
    "ada_variant_f_components"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let variant_part_f_discr_name = foreign ~from:c_lib
    "ada_variant_part_f_discr_name"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let variant_part_f_variant = foreign ~from:c_lib
    "ada_variant_part_f_variant"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let with_clause_f_has_limited = foreign ~from:c_lib
    "ada_with_clause_f_has_limited"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let with_clause_f_has_private = foreign ~from:c_lib
    "ada_with_clause_f_has_private"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let with_clause_f_packages = foreign ~from:c_lib
    "ada_with_clause_f_packages"
    (ptr EntityStruct.c_type
    @-> ptr EntityStruct.c_type
    @-> raisable int)

  let with_private_p_as_bool = foreign ~from:c_lib
    "ada_with_private_p_as_bool"
    (ptr EntityStruct.c_type
    @-> ptr bool
    @-> raisable int)


end

module rec AdaNodeType : sig
  
   
  (**
    * AbortNodeType.t
    * AbstractNodeType.t
    * AdaListType.t
    * AliasedNodeType.t
    * AllNodeType.t
    * ArrayIndicesType.t
    * AspectAssocType.t
    * AspectClauseType.t
    * AspectSpecType.t
    * BaseAssocType.t
    * BaseFormalParamHolderType.t
    * BaseRecordDefType.t
    * BasicAssocType.t
    * BasicDeclType.t
    * CaseStmtAlternativeType.t
    * CompilationUnitType.t
    * ComponentClauseType.t
    * ComponentDefType.t
    * ConstantNodeType.t
    * ConstraintType.t
    * DeclarativePartType.t
    * ElsifExprPartType.t
    * ElsifStmtPartType.t
    * ExprType.t
    * HandledStmtsType.t
    * InterfaceKindType.t
    * IterTypeType.t
    * LibraryItemType.t
    * LimitedNodeType.t
    * LoopSpecType.t
    * ModeType.t
    * MultiAbstractStateDeclType.t
    * NotNullType.t
    * NullComponentDeclType.t
    * OthersDesignatorType.t
    * OverridingNodeType.t
    * ParamsType.t
    * ParenAbstractStateDeclType.t
    * PragmaNodeType.t
    * PrivateNodeType.t
    * ProtectedDefType.t
    * ProtectedNodeType.t
    * QuantifierType.t
    * RangeSpecType.t
    * RenamingClauseType.t
    * ReverseNodeType.t
    * SelectWhenPartType.t
    * StmtType.t
    * SubpKindType.t
    * SubunitType.t
    * SynchronizedNodeType.t
    * TaggedNodeType.t
    * TaskDefType.t
    * TypeDefType.t
    * TypeExprType.t
    * UnconstrainedArrayIndexType.t
    * UntilNodeType.t
    * UseClauseType.t
    * VariantType.t
    * VariantPartType.t
    * WithClauseType.t
    * WithPrivateType.t
    *)
  type t = [
    | `AbortAbsent
        of AbortAbsentType.fields
    | `AbortPresent
        of AbortPresentType.fields
    | `AbstractAbsent
        of AbstractAbsentType.fields
    | `AbstractPresent
        of AbstractPresentType.fields
    | `AdaNodeList
        of AdaNodeListType.fields
    | `AbstractStateDeclList
        of AbstractStateDeclListType.fields
    | `AlternativesList
        of AlternativesListType.fields
    | `ConstraintList
        of ConstraintListType.fields
    | `DeclList
        of DeclListType.fields
    | `StmtList
        of StmtListType.fields
    | `AspectAssocList
        of AspectAssocListType.fields
    | `BaseAssocList
        of BaseAssocListType.fields
    | `AssocList
        of AssocListType.fields
    | `BasicDeclList
        of BasicDeclListType.fields
    | `CaseExprAlternativeList
        of CaseExprAlternativeListType.fields
    | `CaseStmtAlternativeList
        of CaseStmtAlternativeListType.fields
    | `CompilationUnitList
        of CompilationUnitListType.fields
    | `ContractCaseAssocList
        of ContractCaseAssocListType.fields
    | `DefiningNameList
        of DefiningNameListType.fields
    | `DiscriminantSpecList
        of DiscriminantSpecListType.fields
    | `ElsifExprPartList
        of ElsifExprPartListType.fields
    | `ElsifStmtPartList
        of ElsifStmtPartListType.fields
    | `EnumLiteralDeclList
        of EnumLiteralDeclListType.fields
    | `ExprAlternativesList
        of ExprAlternativesListType.fields
    | `DiscriminantChoiceList
        of DiscriminantChoiceListType.fields
    | `NameList
        of NameListType.fields
    | `ParentList
        of ParentListType.fields
    | `ParamSpecList
        of ParamSpecListType.fields
    | `PragmaNodeList
        of PragmaNodeListType.fields
    | `SelectWhenPartList
        of SelectWhenPartListType.fields
    | `UnconstrainedArrayIndexList
        of UnconstrainedArrayIndexListType.fields
    | `VariantList
        of VariantListType.fields
    | `AliasedAbsent
        of AliasedAbsentType.fields
    | `AliasedPresent
        of AliasedPresentType.fields
    | `AllAbsent
        of AllAbsentType.fields
    | `AllPresent
        of AllPresentType.fields
    | `ConstrainedArrayIndices
        of ConstrainedArrayIndicesType.fields
    | `UnconstrainedArrayIndices
        of UnconstrainedArrayIndicesType.fields
    | `AspectAssoc
        of AspectAssocType.fields
    | `AtClause
        of AtClauseType.fields
    | `AttributeDefClause
        of AttributeDefClauseType.fields
    | `EnumRepClause
        of EnumRepClauseType.fields
    | `RecordRepClause
        of RecordRepClauseType.fields
    | `AspectSpec
        of AspectSpecType.fields
    | `ContractCaseAssoc
        of ContractCaseAssocType.fields
    | `PragmaArgumentAssoc
        of PragmaArgumentAssocType.fields
    | `EntrySpec
        of EntrySpecType.fields
    | `EnumSubpSpec
        of EnumSubpSpecType.fields
    | `SubpSpec
        of SubpSpecType.fields
    | `ComponentList
        of ComponentListType.fields
    | `KnownDiscriminantPart
        of KnownDiscriminantPartType.fields
    | `UnknownDiscriminantPart
        of UnknownDiscriminantPartType.fields
    | `EntryCompletionFormalParams
        of EntryCompletionFormalParamsType.fields
    | `GenericFormalPart
        of GenericFormalPartType.fields
    | `NullRecordDef
        of NullRecordDefType.fields
    | `RecordDef
        of RecordDefType.fields
    | `AggregateAssoc
        of AggregateAssocType.fields
    | `MultiDimArrayAssoc
        of MultiDimArrayAssocType.fields
    | `DiscriminantAssoc
        of DiscriminantAssocType.fields
    | `IteratedAssoc
        of IteratedAssocType.fields
    | `ParamAssoc
        of ParamAssocType.fields
    | `AbstractStateDecl
        of AbstractStateDeclType.fields
    | `AnonymousExprDecl
        of AnonymousExprDeclType.fields
    | `ComponentDecl
        of ComponentDeclType.fields
    | `DiscriminantSpec
        of DiscriminantSpecType.fields
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
    | `ParamSpec
        of ParamSpecType.fields
    | `GenericPackageInternal
        of GenericPackageInternalType.fields
    | `PackageDecl
        of PackageDeclType.fields
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
    | `ClasswideTypeDecl
        of ClasswideTypeDeclType.fields
    | `IncompleteTypeDecl
        of IncompleteTypeDeclType.fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
    | `ProtectedTypeDecl
        of ProtectedTypeDeclType.fields
    | `TaskTypeDecl
        of TaskTypeDeclType.fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
    | `TypeDecl
        of TypeDeclType.fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
    | `EntryDecl
        of EntryDeclType.fields
    | `EnumLiteralDecl
        of EnumLiteralDeclType.fields
    | `GenericSubpInternal
        of GenericSubpInternalType.fields
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
    | `EntryBody
        of EntryBodyType.fields
    | `PackageBody
        of PackageBodyType.fields
    | `ProtectedBody
        of ProtectedBodyType.fields
    | `TaskBody
        of TaskBodyType.fields
    | `EntryIndexSpec
        of EntryIndexSpecType.fields
    | `ErrorDecl
        of ErrorDeclType.fields
    | `ExceptionDecl
        of ExceptionDeclType.fields
    | `ExceptionHandler
        of ExceptionHandlerType.fields
    | `ForLoopVarDecl
        of ForLoopVarDeclType.fields
    | `GenericPackageDecl
        of GenericPackageDeclType.fields
    | `GenericSubpDecl
        of GenericSubpDeclType.fields
    | `GenericPackageInstantiation
        of GenericPackageInstantiationType.fields
    | `GenericSubpInstantiation
        of GenericSubpInstantiationType.fields
    | `GenericPackageRenamingDecl
        of GenericPackageRenamingDeclType.fields
    | `GenericSubpRenamingDecl
        of GenericSubpRenamingDeclType.fields
    | `LabelDecl
        of LabelDeclType.fields
    | `NamedStmtDecl
        of NamedStmtDeclType.fields
    | `NumberDecl
        of NumberDeclType.fields
    | `ObjectDecl
        of ObjectDeclType.fields
    | `ExtendedReturnStmtObjectDecl
        of ExtendedReturnStmtObjectDeclType.fields
    | `PackageRenamingDecl
        of PackageRenamingDeclType.fields
    | `SingleProtectedDecl
        of SingleProtectedDeclType.fields
    | `SingleTaskDecl
        of SingleTaskDeclType.fields
    | `CaseStmtAlternative
        of CaseStmtAlternativeType.fields
    | `CompilationUnit
        of CompilationUnitType.fields
    | `ComponentClause
        of ComponentClauseType.fields
    | `ComponentDef
        of ComponentDefType.fields
    | `ConstantAbsent
        of ConstantAbsentType.fields
    | `ConstantPresent
        of ConstantPresentType.fields
    | `DeltaConstraint
        of DeltaConstraintType.fields
    | `DigitsConstraint
        of DigitsConstraintType.fields
    | `DiscriminantConstraint
        of DiscriminantConstraintType.fields
    | `IndexConstraint
        of IndexConstraintType.fields
    | `RangeConstraint
        of RangeConstraintType.fields
    | `DeclarativePart
        of DeclarativePartType.fields
    | `PrivatePart
        of PrivatePartType.fields
    | `PublicPart
        of PublicPartType.fields
    | `ElsifExprPart
        of ElsifExprPartType.fields
    | `ElsifStmtPart
        of ElsifStmtPartType.fields
    | `AbstractStateDeclExpr
        of AbstractStateDeclExprType.fields
    | `Allocator
        of AllocatorType.fields
    | `Aggregate
        of AggregateType.fields
    | `BracketAggregate
        of BracketAggregateType.fields
    | `DeltaAggregate
        of DeltaAggregateType.fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
    | `NullRecordAggregate
        of NullRecordAggregateType.fields
    | `BinOp
        of BinOpType.fields
    | `RelationOp
        of RelationOpType.fields
    | `BoxExpr
        of BoxExprType.fields
    | `CaseExprAlternative
        of CaseExprAlternativeType.fields
    | `CaseExpr
        of CaseExprType.fields
    | `IfExpr
        of IfExprType.fields
    | `ContractCases
        of ContractCasesType.fields
    | `DeclExpr
        of DeclExprType.fields
    | `MembershipExpr
        of MembershipExprType.fields
    | `AttributeRef
        of AttributeRefType.fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
    | `CallExpr
        of CallExprType.fields
    | `DefiningName
        of DefiningNameType.fields
    | `DiscreteSubtypeName
        of DiscreteSubtypeNameType.fields
    | `DottedName
        of DottedNameType.fields
    | `EndName
        of EndNameType.fields
    | `ExplicitDeref
        of ExplicitDerefType.fields
    | `QualExpr
        of QualExprType.fields
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
    | `TargetName
        of TargetNameType.fields
    | `ParenExpr
        of ParenExprType.fields
    | `QuantifiedExpr
        of QuantifiedExprType.fields
    | `RaiseExpr
        of RaiseExprType.fields
    | `UnOp
        of UnOpType.fields
    | `HandledStmts
        of HandledStmtsType.fields
    | `InterfaceKindLimited
        of InterfaceKindLimitedType.fields
    | `InterfaceKindProtected
        of InterfaceKindProtectedType.fields
    | `InterfaceKindSynchronized
        of InterfaceKindSynchronizedType.fields
    | `InterfaceKindTask
        of InterfaceKindTaskType.fields
    | `IterTypeIn
        of IterTypeInType.fields
    | `IterTypeOf
        of IterTypeOfType.fields
    | `LibraryItem
        of LibraryItemType.fields
    | `LimitedAbsent
        of LimitedAbsentType.fields
    | `LimitedPresent
        of LimitedPresentType.fields
    | `ForLoopSpec
        of ForLoopSpecType.fields
    | `WhileLoopSpec
        of WhileLoopSpecType.fields
    | `ModeDefault
        of ModeDefaultType.fields
    | `ModeIn
        of ModeInType.fields
    | `ModeInOut
        of ModeInOutType.fields
    | `ModeOut
        of ModeOutType.fields
    | `MultiAbstractStateDecl
        of MultiAbstractStateDeclType.fields
    | `NotNullAbsent
        of NotNullAbsentType.fields
    | `NotNullPresent
        of NotNullPresentType.fields
    | `NullComponentDecl
        of NullComponentDeclType.fields
    | `OthersDesignator
        of OthersDesignatorType.fields
    | `OverridingNotOverriding
        of OverridingNotOverridingType.fields
    | `OverridingOverriding
        of OverridingOverridingType.fields
    | `OverridingUnspecified
        of OverridingUnspecifiedType.fields
    | `Params
        of ParamsType.fields
    | `ParenAbstractStateDecl
        of ParenAbstractStateDeclType.fields
    | `PragmaNode
        of PragmaNodeType.fields
    | `PrivateAbsent
        of PrivateAbsentType.fields
    | `PrivatePresent
        of PrivatePresentType.fields
    | `ProtectedDef
        of ProtectedDefType.fields
    | `ProtectedAbsent
        of ProtectedAbsentType.fields
    | `ProtectedPresent
        of ProtectedPresentType.fields
    | `QuantifierAll
        of QuantifierAllType.fields
    | `QuantifierSome
        of QuantifierSomeType.fields
    | `RangeSpec
        of RangeSpecType.fields
    | `RenamingClause
        of RenamingClauseType.fields
    | `SyntheticRenamingClause
        of SyntheticRenamingClauseType.fields
    | `ReverseAbsent
        of ReverseAbsentType.fields
    | `ReversePresent
        of ReversePresentType.fields
    | `SelectWhenPart
        of SelectWhenPartType.fields
    | `AcceptStmt
        of AcceptStmtType.fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
    | `CaseStmt
        of CaseStmtType.fields
    | `ExtendedReturnStmt
        of ExtendedReturnStmtType.fields
    | `IfStmt
        of IfStmtType.fields
    | `NamedStmt
        of NamedStmtType.fields
    | `SelectStmt
        of SelectStmtType.fields
    | `ErrorStmt
        of ErrorStmtType.fields
    | `AbortStmt
        of AbortStmtType.fields
    | `AssignStmt
        of AssignStmtType.fields
    | `CallStmt
        of CallStmtType.fields
    | `DelayStmt
        of DelayStmtType.fields
    | `ExitStmt
        of ExitStmtType.fields
    | `GotoStmt
        of GotoStmtType.fields
    | `Label
        of LabelType.fields
    | `NullStmt
        of NullStmtType.fields
    | `RaiseStmt
        of RaiseStmtType.fields
    | `RequeueStmt
        of RequeueStmtType.fields
    | `ReturnStmt
        of ReturnStmtType.fields
    | `TerminateAlternative
        of TerminateAlternativeType.fields
    | `SubpKindFunction
        of SubpKindFunctionType.fields
    | `SubpKindProcedure
        of SubpKindProcedureType.fields
    | `Subunit
        of SubunitType.fields
    | `SynchronizedAbsent
        of SynchronizedAbsentType.fields
    | `SynchronizedPresent
        of SynchronizedPresentType.fields
    | `TaggedAbsent
        of TaggedAbsentType.fields
    | `TaggedPresent
        of TaggedPresentType.fields
    | `TaskDef
        of TaskDefType.fields
    | `AccessToSubpDef
        of AccessToSubpDefType.fields
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
    | `ArrayTypeDef
        of ArrayTypeDefType.fields
    | `DerivedTypeDef
        of DerivedTypeDefType.fields
    | `EnumTypeDef
        of EnumTypeDefType.fields
    | `FormalDiscreteTypeDef
        of FormalDiscreteTypeDefType.fields
    | `InterfaceTypeDef
        of InterfaceTypeDefType.fields
    | `ModIntTypeDef
        of ModIntTypeDefType.fields
    | `PrivateTypeDef
        of PrivateTypeDefType.fields
    | `DecimalFixedPointDef
        of DecimalFixedPointDefType.fields
    | `FloatingPointDef
        of FloatingPointDefType.fields
    | `OrdinaryFixedPointDef
        of OrdinaryFixedPointDefType.fields
    | `RecordTypeDef
        of RecordTypeDefType.fields
    | `SignedIntTypeDef
        of SignedIntTypeDefType.fields
    | `AnonymousType
        of AnonymousTypeType.fields
    | `EnumLitSynthTypeExpr
        of EnumLitSynthTypeExprType.fields
    | `SubtypeIndication
        of SubtypeIndicationType.fields
    | `ConstrainedSubtypeIndication
        of ConstrainedSubtypeIndicationType.fields
    | `DiscreteSubtypeIndication
        of DiscreteSubtypeIndicationType.fields
    | `UnconstrainedArrayIndex
        of UnconstrainedArrayIndexType.fields
    | `UntilAbsent
        of UntilAbsentType.fields
    | `UntilPresent
        of UntilPresentType.fields
    | `UsePackageClause
        of UsePackageClauseType.fields
    | `UseTypeClause
        of UseTypeClauseType.fields
    | `Variant
        of VariantType.fields
    | `VariantPart
        of VariantPartType.fields
    | `WithClause
        of WithClauseType.fields
    | `WithPrivateAbsent
        of WithPrivateAbsentType.fields
    | `WithPrivatePresent
        of WithPrivatePresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

  val unwrap : [< t ] -> EntityStruct.t structure

end = struct
  
   
  (**
    * AbortNodeType.t
    * AbstractNodeType.t
    * AdaListType.t
    * AliasedNodeType.t
    * AllNodeType.t
    * ArrayIndicesType.t
    * AspectAssocType.t
    * AspectClauseType.t
    * AspectSpecType.t
    * BaseAssocType.t
    * BaseFormalParamHolderType.t
    * BaseRecordDefType.t
    * BasicAssocType.t
    * BasicDeclType.t
    * CaseStmtAlternativeType.t
    * CompilationUnitType.t
    * ComponentClauseType.t
    * ComponentDefType.t
    * ConstantNodeType.t
    * ConstraintType.t
    * DeclarativePartType.t
    * ElsifExprPartType.t
    * ElsifStmtPartType.t
    * ExprType.t
    * HandledStmtsType.t
    * InterfaceKindType.t
    * IterTypeType.t
    * LibraryItemType.t
    * LimitedNodeType.t
    * LoopSpecType.t
    * ModeType.t
    * MultiAbstractStateDeclType.t
    * NotNullType.t
    * NullComponentDeclType.t
    * OthersDesignatorType.t
    * OverridingNodeType.t
    * ParamsType.t
    * ParenAbstractStateDeclType.t
    * PragmaNodeType.t
    * PrivateNodeType.t
    * ProtectedDefType.t
    * ProtectedNodeType.t
    * QuantifierType.t
    * RangeSpecType.t
    * RenamingClauseType.t
    * ReverseNodeType.t
    * SelectWhenPartType.t
    * StmtType.t
    * SubpKindType.t
    * SubunitType.t
    * SynchronizedNodeType.t
    * TaggedNodeType.t
    * TaskDefType.t
    * TypeDefType.t
    * TypeExprType.t
    * UnconstrainedArrayIndexType.t
    * UntilNodeType.t
    * UseClauseType.t
    * VariantType.t
    * VariantPartType.t
    * WithClauseType.t
    * WithPrivateType.t
    *)
  type t = [
    | `AbortAbsent
        of AbortAbsentType.fields
    | `AbortPresent
        of AbortPresentType.fields
    | `AbstractAbsent
        of AbstractAbsentType.fields
    | `AbstractPresent
        of AbstractPresentType.fields
    | `AdaNodeList
        of AdaNodeListType.fields
    | `AbstractStateDeclList
        of AbstractStateDeclListType.fields
    | `AlternativesList
        of AlternativesListType.fields
    | `ConstraintList
        of ConstraintListType.fields
    | `DeclList
        of DeclListType.fields
    | `StmtList
        of StmtListType.fields
    | `AspectAssocList
        of AspectAssocListType.fields
    | `BaseAssocList
        of BaseAssocListType.fields
    | `AssocList
        of AssocListType.fields
    | `BasicDeclList
        of BasicDeclListType.fields
    | `CaseExprAlternativeList
        of CaseExprAlternativeListType.fields
    | `CaseStmtAlternativeList
        of CaseStmtAlternativeListType.fields
    | `CompilationUnitList
        of CompilationUnitListType.fields
    | `ContractCaseAssocList
        of ContractCaseAssocListType.fields
    | `DefiningNameList
        of DefiningNameListType.fields
    | `DiscriminantSpecList
        of DiscriminantSpecListType.fields
    | `ElsifExprPartList
        of ElsifExprPartListType.fields
    | `ElsifStmtPartList
        of ElsifStmtPartListType.fields
    | `EnumLiteralDeclList
        of EnumLiteralDeclListType.fields
    | `ExprAlternativesList
        of ExprAlternativesListType.fields
    | `DiscriminantChoiceList
        of DiscriminantChoiceListType.fields
    | `NameList
        of NameListType.fields
    | `ParentList
        of ParentListType.fields
    | `ParamSpecList
        of ParamSpecListType.fields
    | `PragmaNodeList
        of PragmaNodeListType.fields
    | `SelectWhenPartList
        of SelectWhenPartListType.fields
    | `UnconstrainedArrayIndexList
        of UnconstrainedArrayIndexListType.fields
    | `VariantList
        of VariantListType.fields
    | `AliasedAbsent
        of AliasedAbsentType.fields
    | `AliasedPresent
        of AliasedPresentType.fields
    | `AllAbsent
        of AllAbsentType.fields
    | `AllPresent
        of AllPresentType.fields
    | `ConstrainedArrayIndices
        of ConstrainedArrayIndicesType.fields
    | `UnconstrainedArrayIndices
        of UnconstrainedArrayIndicesType.fields
    | `AspectAssoc
        of AspectAssocType.fields
    | `AtClause
        of AtClauseType.fields
    | `AttributeDefClause
        of AttributeDefClauseType.fields
    | `EnumRepClause
        of EnumRepClauseType.fields
    | `RecordRepClause
        of RecordRepClauseType.fields
    | `AspectSpec
        of AspectSpecType.fields
    | `ContractCaseAssoc
        of ContractCaseAssocType.fields
    | `PragmaArgumentAssoc
        of PragmaArgumentAssocType.fields
    | `EntrySpec
        of EntrySpecType.fields
    | `EnumSubpSpec
        of EnumSubpSpecType.fields
    | `SubpSpec
        of SubpSpecType.fields
    | `ComponentList
        of ComponentListType.fields
    | `KnownDiscriminantPart
        of KnownDiscriminantPartType.fields
    | `UnknownDiscriminantPart
        of UnknownDiscriminantPartType.fields
    | `EntryCompletionFormalParams
        of EntryCompletionFormalParamsType.fields
    | `GenericFormalPart
        of GenericFormalPartType.fields
    | `NullRecordDef
        of NullRecordDefType.fields
    | `RecordDef
        of RecordDefType.fields
    | `AggregateAssoc
        of AggregateAssocType.fields
    | `MultiDimArrayAssoc
        of MultiDimArrayAssocType.fields
    | `DiscriminantAssoc
        of DiscriminantAssocType.fields
    | `IteratedAssoc
        of IteratedAssocType.fields
    | `ParamAssoc
        of ParamAssocType.fields
    | `AbstractStateDecl
        of AbstractStateDeclType.fields
    | `AnonymousExprDecl
        of AnonymousExprDeclType.fields
    | `ComponentDecl
        of ComponentDeclType.fields
    | `DiscriminantSpec
        of DiscriminantSpecType.fields
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
    | `ParamSpec
        of ParamSpecType.fields
    | `GenericPackageInternal
        of GenericPackageInternalType.fields
    | `PackageDecl
        of PackageDeclType.fields
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
    | `ClasswideTypeDecl
        of ClasswideTypeDeclType.fields
    | `IncompleteTypeDecl
        of IncompleteTypeDeclType.fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
    | `ProtectedTypeDecl
        of ProtectedTypeDeclType.fields
    | `TaskTypeDecl
        of TaskTypeDeclType.fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
    | `TypeDecl
        of TypeDeclType.fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
    | `EntryDecl
        of EntryDeclType.fields
    | `EnumLiteralDecl
        of EnumLiteralDeclType.fields
    | `GenericSubpInternal
        of GenericSubpInternalType.fields
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
    | `EntryBody
        of EntryBodyType.fields
    | `PackageBody
        of PackageBodyType.fields
    | `ProtectedBody
        of ProtectedBodyType.fields
    | `TaskBody
        of TaskBodyType.fields
    | `EntryIndexSpec
        of EntryIndexSpecType.fields
    | `ErrorDecl
        of ErrorDeclType.fields
    | `ExceptionDecl
        of ExceptionDeclType.fields
    | `ExceptionHandler
        of ExceptionHandlerType.fields
    | `ForLoopVarDecl
        of ForLoopVarDeclType.fields
    | `GenericPackageDecl
        of GenericPackageDeclType.fields
    | `GenericSubpDecl
        of GenericSubpDeclType.fields
    | `GenericPackageInstantiation
        of GenericPackageInstantiationType.fields
    | `GenericSubpInstantiation
        of GenericSubpInstantiationType.fields
    | `GenericPackageRenamingDecl
        of GenericPackageRenamingDeclType.fields
    | `GenericSubpRenamingDecl
        of GenericSubpRenamingDeclType.fields
    | `LabelDecl
        of LabelDeclType.fields
    | `NamedStmtDecl
        of NamedStmtDeclType.fields
    | `NumberDecl
        of NumberDeclType.fields
    | `ObjectDecl
        of ObjectDeclType.fields
    | `ExtendedReturnStmtObjectDecl
        of ExtendedReturnStmtObjectDeclType.fields
    | `PackageRenamingDecl
        of PackageRenamingDeclType.fields
    | `SingleProtectedDecl
        of SingleProtectedDeclType.fields
    | `SingleTaskDecl
        of SingleTaskDeclType.fields
    | `CaseStmtAlternative
        of CaseStmtAlternativeType.fields
    | `CompilationUnit
        of CompilationUnitType.fields
    | `ComponentClause
        of ComponentClauseType.fields
    | `ComponentDef
        of ComponentDefType.fields
    | `ConstantAbsent
        of ConstantAbsentType.fields
    | `ConstantPresent
        of ConstantPresentType.fields
    | `DeltaConstraint
        of DeltaConstraintType.fields
    | `DigitsConstraint
        of DigitsConstraintType.fields
    | `DiscriminantConstraint
        of DiscriminantConstraintType.fields
    | `IndexConstraint
        of IndexConstraintType.fields
    | `RangeConstraint
        of RangeConstraintType.fields
    | `DeclarativePart
        of DeclarativePartType.fields
    | `PrivatePart
        of PrivatePartType.fields
    | `PublicPart
        of PublicPartType.fields
    | `ElsifExprPart
        of ElsifExprPartType.fields
    | `ElsifStmtPart
        of ElsifStmtPartType.fields
    | `AbstractStateDeclExpr
        of AbstractStateDeclExprType.fields
    | `Allocator
        of AllocatorType.fields
    | `Aggregate
        of AggregateType.fields
    | `BracketAggregate
        of BracketAggregateType.fields
    | `DeltaAggregate
        of DeltaAggregateType.fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
    | `NullRecordAggregate
        of NullRecordAggregateType.fields
    | `BinOp
        of BinOpType.fields
    | `RelationOp
        of RelationOpType.fields
    | `BoxExpr
        of BoxExprType.fields
    | `CaseExprAlternative
        of CaseExprAlternativeType.fields
    | `CaseExpr
        of CaseExprType.fields
    | `IfExpr
        of IfExprType.fields
    | `ContractCases
        of ContractCasesType.fields
    | `DeclExpr
        of DeclExprType.fields
    | `MembershipExpr
        of MembershipExprType.fields
    | `AttributeRef
        of AttributeRefType.fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
    | `CallExpr
        of CallExprType.fields
    | `DefiningName
        of DefiningNameType.fields
    | `DiscreteSubtypeName
        of DiscreteSubtypeNameType.fields
    | `DottedName
        of DottedNameType.fields
    | `EndName
        of EndNameType.fields
    | `ExplicitDeref
        of ExplicitDerefType.fields
    | `QualExpr
        of QualExprType.fields
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
    | `TargetName
        of TargetNameType.fields
    | `ParenExpr
        of ParenExprType.fields
    | `QuantifiedExpr
        of QuantifiedExprType.fields
    | `RaiseExpr
        of RaiseExprType.fields
    | `UnOp
        of UnOpType.fields
    | `HandledStmts
        of HandledStmtsType.fields
    | `InterfaceKindLimited
        of InterfaceKindLimitedType.fields
    | `InterfaceKindProtected
        of InterfaceKindProtectedType.fields
    | `InterfaceKindSynchronized
        of InterfaceKindSynchronizedType.fields
    | `InterfaceKindTask
        of InterfaceKindTaskType.fields
    | `IterTypeIn
        of IterTypeInType.fields
    | `IterTypeOf
        of IterTypeOfType.fields
    | `LibraryItem
        of LibraryItemType.fields
    | `LimitedAbsent
        of LimitedAbsentType.fields
    | `LimitedPresent
        of LimitedPresentType.fields
    | `ForLoopSpec
        of ForLoopSpecType.fields
    | `WhileLoopSpec
        of WhileLoopSpecType.fields
    | `ModeDefault
        of ModeDefaultType.fields
    | `ModeIn
        of ModeInType.fields
    | `ModeInOut
        of ModeInOutType.fields
    | `ModeOut
        of ModeOutType.fields
    | `MultiAbstractStateDecl
        of MultiAbstractStateDeclType.fields
    | `NotNullAbsent
        of NotNullAbsentType.fields
    | `NotNullPresent
        of NotNullPresentType.fields
    | `NullComponentDecl
        of NullComponentDeclType.fields
    | `OthersDesignator
        of OthersDesignatorType.fields
    | `OverridingNotOverriding
        of OverridingNotOverridingType.fields
    | `OverridingOverriding
        of OverridingOverridingType.fields
    | `OverridingUnspecified
        of OverridingUnspecifiedType.fields
    | `Params
        of ParamsType.fields
    | `ParenAbstractStateDecl
        of ParenAbstractStateDeclType.fields
    | `PragmaNode
        of PragmaNodeType.fields
    | `PrivateAbsent
        of PrivateAbsentType.fields
    | `PrivatePresent
        of PrivatePresentType.fields
    | `ProtectedDef
        of ProtectedDefType.fields
    | `ProtectedAbsent
        of ProtectedAbsentType.fields
    | `ProtectedPresent
        of ProtectedPresentType.fields
    | `QuantifierAll
        of QuantifierAllType.fields
    | `QuantifierSome
        of QuantifierSomeType.fields
    | `RangeSpec
        of RangeSpecType.fields
    | `RenamingClause
        of RenamingClauseType.fields
    | `SyntheticRenamingClause
        of SyntheticRenamingClauseType.fields
    | `ReverseAbsent
        of ReverseAbsentType.fields
    | `ReversePresent
        of ReversePresentType.fields
    | `SelectWhenPart
        of SelectWhenPartType.fields
    | `AcceptStmt
        of AcceptStmtType.fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
    | `CaseStmt
        of CaseStmtType.fields
    | `ExtendedReturnStmt
        of ExtendedReturnStmtType.fields
    | `IfStmt
        of IfStmtType.fields
    | `NamedStmt
        of NamedStmtType.fields
    | `SelectStmt
        of SelectStmtType.fields
    | `ErrorStmt
        of ErrorStmtType.fields
    | `AbortStmt
        of AbortStmtType.fields
    | `AssignStmt
        of AssignStmtType.fields
    | `CallStmt
        of CallStmtType.fields
    | `DelayStmt
        of DelayStmtType.fields
    | `ExitStmt
        of ExitStmtType.fields
    | `GotoStmt
        of GotoStmtType.fields
    | `Label
        of LabelType.fields
    | `NullStmt
        of NullStmtType.fields
    | `RaiseStmt
        of RaiseStmtType.fields
    | `RequeueStmt
        of RequeueStmtType.fields
    | `ReturnStmt
        of ReturnStmtType.fields
    | `TerminateAlternative
        of TerminateAlternativeType.fields
    | `SubpKindFunction
        of SubpKindFunctionType.fields
    | `SubpKindProcedure
        of SubpKindProcedureType.fields
    | `Subunit
        of SubunitType.fields
    | `SynchronizedAbsent
        of SynchronizedAbsentType.fields
    | `SynchronizedPresent
        of SynchronizedPresentType.fields
    | `TaggedAbsent
        of TaggedAbsentType.fields
    | `TaggedPresent
        of TaggedPresentType.fields
    | `TaskDef
        of TaskDefType.fields
    | `AccessToSubpDef
        of AccessToSubpDefType.fields
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
    | `ArrayTypeDef
        of ArrayTypeDefType.fields
    | `DerivedTypeDef
        of DerivedTypeDefType.fields
    | `EnumTypeDef
        of EnumTypeDefType.fields
    | `FormalDiscreteTypeDef
        of FormalDiscreteTypeDefType.fields
    | `InterfaceTypeDef
        of InterfaceTypeDefType.fields
    | `ModIntTypeDef
        of ModIntTypeDefType.fields
    | `PrivateTypeDef
        of PrivateTypeDefType.fields
    | `DecimalFixedPointDef
        of DecimalFixedPointDefType.fields
    | `FloatingPointDef
        of FloatingPointDefType.fields
    | `OrdinaryFixedPointDef
        of OrdinaryFixedPointDefType.fields
    | `RecordTypeDef
        of RecordTypeDefType.fields
    | `SignedIntTypeDef
        of SignedIntTypeDefType.fields
    | `AnonymousType
        of AnonymousTypeType.fields
    | `EnumLitSynthTypeExpr
        of EnumLitSynthTypeExprType.fields
    | `SubtypeIndication
        of SubtypeIndicationType.fields
    | `ConstrainedSubtypeIndication
        of ConstrainedSubtypeIndicationType.fields
    | `DiscreteSubtypeIndication
        of DiscreteSubtypeIndicationType.fields
    | `UnconstrainedArrayIndex
        of UnconstrainedArrayIndexType.fields
    | `UntilAbsent
        of UntilAbsentType.fields
    | `UntilPresent
        of UntilPresentType.fields
    | `UsePackageClause
        of UsePackageClauseType.fields
    | `UseTypeClause
        of UseTypeClauseType.fields
    | `Variant
        of VariantType.fields
    | `VariantPart
        of VariantPartType.fields
    | `WithClause
        of WithClauseType.fields
    | `WithPrivateAbsent
        of WithPrivateAbsentType.fields
    | `WithPrivatePresent
        of WithPrivatePresentType.fields
  ]


  let wrap context c_value =
    (* Top level wrap function that dispatch to wrap function of concrete types
      depending on the node kind *)
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      let kind = CFunctions.node_kind (addr c_value) in
      match kind with
      | 1 ->
          (AbortAbsentType.wrap context (c_value)
           :> t)
      | 2 ->
          (AbortPresentType.wrap context (c_value)
           :> t)
      | 3 ->
          (AbstractAbsentType.wrap context (c_value)
           :> t)
      | 4 ->
          (AbstractPresentType.wrap context (c_value)
           :> t)
      | 5 ->
          (AdaNodeListType.wrap context (c_value)
           :> t)
      | 6 ->
          (AbstractStateDeclListType.wrap context (c_value)
           :> t)
      | 7 ->
          (AlternativesListType.wrap context (c_value)
           :> t)
      | 8 ->
          (ConstraintListType.wrap context (c_value)
           :> t)
      | 9 ->
          (DeclListType.wrap context (c_value)
           :> t)
      | 10 ->
          (StmtListType.wrap context (c_value)
           :> t)
      | 11 ->
          (AspectAssocListType.wrap context (c_value)
           :> t)
      | 12 ->
          (BaseAssocListType.wrap context (c_value)
           :> t)
      | 13 ->
          (AssocListType.wrap context (c_value)
           :> t)
      | 14 ->
          (BasicDeclListType.wrap context (c_value)
           :> t)
      | 15 ->
          (CaseExprAlternativeListType.wrap context (c_value)
           :> t)
      | 16 ->
          (CaseStmtAlternativeListType.wrap context (c_value)
           :> t)
      | 17 ->
          (CompilationUnitListType.wrap context (c_value)
           :> t)
      | 18 ->
          (ContractCaseAssocListType.wrap context (c_value)
           :> t)
      | 19 ->
          (DefiningNameListType.wrap context (c_value)
           :> t)
      | 20 ->
          (DiscriminantSpecListType.wrap context (c_value)
           :> t)
      | 21 ->
          (ElsifExprPartListType.wrap context (c_value)
           :> t)
      | 22 ->
          (ElsifStmtPartListType.wrap context (c_value)
           :> t)
      | 23 ->
          (EnumLiteralDeclListType.wrap context (c_value)
           :> t)
      | 24 ->
          (ExprAlternativesListType.wrap context (c_value)
           :> t)
      | 25 ->
          (DiscriminantChoiceListType.wrap context (c_value)
           :> t)
      | 26 ->
          (NameListType.wrap context (c_value)
           :> t)
      | 27 ->
          (ParentListType.wrap context (c_value)
           :> t)
      | 28 ->
          (ParamSpecListType.wrap context (c_value)
           :> t)
      | 29 ->
          (PragmaNodeListType.wrap context (c_value)
           :> t)
      | 30 ->
          (SelectWhenPartListType.wrap context (c_value)
           :> t)
      | 31 ->
          (UnconstrainedArrayIndexListType.wrap context (c_value)
           :> t)
      | 32 ->
          (VariantListType.wrap context (c_value)
           :> t)
      | 33 ->
          (AliasedAbsentType.wrap context (c_value)
           :> t)
      | 34 ->
          (AliasedPresentType.wrap context (c_value)
           :> t)
      | 35 ->
          (AllAbsentType.wrap context (c_value)
           :> t)
      | 36 ->
          (AllPresentType.wrap context (c_value)
           :> t)
      | 37 ->
          (ConstrainedArrayIndicesType.wrap context (c_value)
           :> t)
      | 38 ->
          (UnconstrainedArrayIndicesType.wrap context (c_value)
           :> t)
      | 39 ->
          (AspectAssocType.wrap context (c_value)
           :> t)
      | 40 ->
          (AtClauseType.wrap context (c_value)
           :> t)
      | 41 ->
          (AttributeDefClauseType.wrap context (c_value)
           :> t)
      | 42 ->
          (EnumRepClauseType.wrap context (c_value)
           :> t)
      | 43 ->
          (RecordRepClauseType.wrap context (c_value)
           :> t)
      | 44 ->
          (AspectSpecType.wrap context (c_value)
           :> t)
      | 45 ->
          (ContractCaseAssocType.wrap context (c_value)
           :> t)
      | 46 ->
          (PragmaArgumentAssocType.wrap context (c_value)
           :> t)
      | 47 ->
          (EntrySpecType.wrap context (c_value)
           :> t)
      | 48 ->
          (EnumSubpSpecType.wrap context (c_value)
           :> t)
      | 49 ->
          (SubpSpecType.wrap context (c_value)
           :> t)
      | 50 ->
          (ComponentListType.wrap context (c_value)
           :> t)
      | 51 ->
          (KnownDiscriminantPartType.wrap context (c_value)
           :> t)
      | 52 ->
          (UnknownDiscriminantPartType.wrap context (c_value)
           :> t)
      | 53 ->
          (EntryCompletionFormalParamsType.wrap context (c_value)
           :> t)
      | 54 ->
          (GenericFormalPartType.wrap context (c_value)
           :> t)
      | 55 ->
          (NullRecordDefType.wrap context (c_value)
           :> t)
      | 56 ->
          (RecordDefType.wrap context (c_value)
           :> t)
      | 57 ->
          (AggregateAssocType.wrap context (c_value)
           :> t)
      | 58 ->
          (MultiDimArrayAssocType.wrap context (c_value)
           :> t)
      | 59 ->
          (DiscriminantAssocType.wrap context (c_value)
           :> t)
      | 60 ->
          (IteratedAssocType.wrap context (c_value)
           :> t)
      | 61 ->
          (ParamAssocType.wrap context (c_value)
           :> t)
      | 62 ->
          (AbstractStateDeclType.wrap context (c_value)
           :> t)
      | 63 ->
          (AnonymousExprDeclType.wrap context (c_value)
           :> t)
      | 64 ->
          (ComponentDeclType.wrap context (c_value)
           :> t)
      | 65 ->
          (DiscriminantSpecType.wrap context (c_value)
           :> t)
      | 66 ->
          (GenericFormalObjDeclType.wrap context (c_value)
           :> t)
      | 67 ->
          (GenericFormalPackageType.wrap context (c_value)
           :> t)
      | 68 ->
          (GenericFormalSubpDeclType.wrap context (c_value)
           :> t)
      | 69 ->
          (GenericFormalTypeDeclType.wrap context (c_value)
           :> t)
      | 70 ->
          (ParamSpecType.wrap context (c_value)
           :> t)
      | 71 ->
          (GenericPackageInternalType.wrap context (c_value)
           :> t)
      | 72 ->
          (PackageDeclType.wrap context (c_value)
           :> t)
      | 73 ->
          (DiscreteBaseSubtypeDeclType.wrap context (c_value)
           :> t)
      | 74 ->
          (SubtypeDeclType.wrap context (c_value)
           :> t)
      | 75 ->
          (ClasswideTypeDeclType.wrap context (c_value)
           :> t)
      | 76 ->
          (IncompleteTypeDeclType.wrap context (c_value)
           :> t)
      | 77 ->
          (IncompleteTaggedTypeDeclType.wrap context (c_value)
           :> t)
      | 78 ->
          (ProtectedTypeDeclType.wrap context (c_value)
           :> t)
      | 79 ->
          (TaskTypeDeclType.wrap context (c_value)
           :> t)
      | 80 ->
          (SingleTaskTypeDeclType.wrap context (c_value)
           :> t)
      | 81 ->
          (TypeDeclType.wrap context (c_value)
           :> t)
      | 82 ->
          (AnonymousTypeDeclType.wrap context (c_value)
           :> t)
      | 83 ->
          (SynthAnonymousTypeDeclType.wrap context (c_value)
           :> t)
      | 84 ->
          (AbstractSubpDeclType.wrap context (c_value)
           :> t)
      | 85 ->
          (AbstractFormalSubpDeclType.wrap context (c_value)
           :> t)
      | 86 ->
          (ConcreteFormalSubpDeclType.wrap context (c_value)
           :> t)
      | 87 ->
          (SubpDeclType.wrap context (c_value)
           :> t)
      | 88 ->
          (EntryDeclType.wrap context (c_value)
           :> t)
      | 89 ->
          (EnumLiteralDeclType.wrap context (c_value)
           :> t)
      | 90 ->
          (GenericSubpInternalType.wrap context (c_value)
           :> t)
      | 91 ->
          (ExprFunctionType.wrap context (c_value)
           :> t)
      | 92 ->
          (NullSubpDeclType.wrap context (c_value)
           :> t)
      | 93 ->
          (SubpBodyType.wrap context (c_value)
           :> t)
      | 94 ->
          (SubpRenamingDeclType.wrap context (c_value)
           :> t)
      | 95 ->
          (PackageBodyStubType.wrap context (c_value)
           :> t)
      | 96 ->
          (ProtectedBodyStubType.wrap context (c_value)
           :> t)
      | 97 ->
          (SubpBodyStubType.wrap context (c_value)
           :> t)
      | 98 ->
          (TaskBodyStubType.wrap context (c_value)
           :> t)
      | 99 ->
          (EntryBodyType.wrap context (c_value)
           :> t)
      | 100 ->
          (PackageBodyType.wrap context (c_value)
           :> t)
      | 101 ->
          (ProtectedBodyType.wrap context (c_value)
           :> t)
      | 102 ->
          (TaskBodyType.wrap context (c_value)
           :> t)
      | 103 ->
          (EntryIndexSpecType.wrap context (c_value)
           :> t)
      | 104 ->
          (ErrorDeclType.wrap context (c_value)
           :> t)
      | 105 ->
          (ExceptionDeclType.wrap context (c_value)
           :> t)
      | 106 ->
          (ExceptionHandlerType.wrap context (c_value)
           :> t)
      | 107 ->
          (ForLoopVarDeclType.wrap context (c_value)
           :> t)
      | 108 ->
          (GenericPackageDeclType.wrap context (c_value)
           :> t)
      | 109 ->
          (GenericSubpDeclType.wrap context (c_value)
           :> t)
      | 110 ->
          (GenericPackageInstantiationType.wrap context (c_value)
           :> t)
      | 111 ->
          (GenericSubpInstantiationType.wrap context (c_value)
           :> t)
      | 112 ->
          (GenericPackageRenamingDeclType.wrap context (c_value)
           :> t)
      | 113 ->
          (GenericSubpRenamingDeclType.wrap context (c_value)
           :> t)
      | 114 ->
          (LabelDeclType.wrap context (c_value)
           :> t)
      | 115 ->
          (NamedStmtDeclType.wrap context (c_value)
           :> t)
      | 116 ->
          (NumberDeclType.wrap context (c_value)
           :> t)
      | 117 ->
          (ObjectDeclType.wrap context (c_value)
           :> t)
      | 118 ->
          (ExtendedReturnStmtObjectDeclType.wrap context (c_value)
           :> t)
      | 119 ->
          (PackageRenamingDeclType.wrap context (c_value)
           :> t)
      | 120 ->
          (SingleProtectedDeclType.wrap context (c_value)
           :> t)
      | 121 ->
          (SingleTaskDeclType.wrap context (c_value)
           :> t)
      | 122 ->
          (CaseStmtAlternativeType.wrap context (c_value)
           :> t)
      | 123 ->
          (CompilationUnitType.wrap context (c_value)
           :> t)
      | 124 ->
          (ComponentClauseType.wrap context (c_value)
           :> t)
      | 125 ->
          (ComponentDefType.wrap context (c_value)
           :> t)
      | 126 ->
          (ConstantAbsentType.wrap context (c_value)
           :> t)
      | 127 ->
          (ConstantPresentType.wrap context (c_value)
           :> t)
      | 128 ->
          (DeltaConstraintType.wrap context (c_value)
           :> t)
      | 129 ->
          (DigitsConstraintType.wrap context (c_value)
           :> t)
      | 130 ->
          (DiscriminantConstraintType.wrap context (c_value)
           :> t)
      | 131 ->
          (IndexConstraintType.wrap context (c_value)
           :> t)
      | 132 ->
          (RangeConstraintType.wrap context (c_value)
           :> t)
      | 133 ->
          (DeclarativePartType.wrap context (c_value)
           :> t)
      | 134 ->
          (PrivatePartType.wrap context (c_value)
           :> t)
      | 135 ->
          (PublicPartType.wrap context (c_value)
           :> t)
      | 136 ->
          (ElsifExprPartType.wrap context (c_value)
           :> t)
      | 137 ->
          (ElsifStmtPartType.wrap context (c_value)
           :> t)
      | 138 ->
          (AbstractStateDeclExprType.wrap context (c_value)
           :> t)
      | 139 ->
          (AllocatorType.wrap context (c_value)
           :> t)
      | 140 ->
          (AggregateType.wrap context (c_value)
           :> t)
      | 141 ->
          (BracketAggregateType.wrap context (c_value)
           :> t)
      | 142 ->
          (DeltaAggregateType.wrap context (c_value)
           :> t)
      | 143 ->
          (BracketDeltaAggregateType.wrap context (c_value)
           :> t)
      | 144 ->
          (NullRecordAggregateType.wrap context (c_value)
           :> t)
      | 145 ->
          (BinOpType.wrap context (c_value)
           :> t)
      | 146 ->
          (RelationOpType.wrap context (c_value)
           :> t)
      | 147 ->
          (BoxExprType.wrap context (c_value)
           :> t)
      | 148 ->
          (CaseExprAlternativeType.wrap context (c_value)
           :> t)
      | 149 ->
          (CaseExprType.wrap context (c_value)
           :> t)
      | 150 ->
          (IfExprType.wrap context (c_value)
           :> t)
      | 151 ->
          (ContractCasesType.wrap context (c_value)
           :> t)
      | 152 ->
          (DeclExprType.wrap context (c_value)
           :> t)
      | 153 ->
          (MembershipExprType.wrap context (c_value)
           :> t)
      | 154 ->
          (AttributeRefType.wrap context (c_value)
           :> t)
      | 155 ->
          (UpdateAttributeRefType.wrap context (c_value)
           :> t)
      | 156 ->
          (CallExprType.wrap context (c_value)
           :> t)
      | 157 ->
          (DefiningNameType.wrap context (c_value)
           :> t)
      | 158 ->
          (DiscreteSubtypeNameType.wrap context (c_value)
           :> t)
      | 159 ->
          (DottedNameType.wrap context (c_value)
           :> t)
      | 160 ->
          (EndNameType.wrap context (c_value)
           :> t)
      | 161 ->
          (ExplicitDerefType.wrap context (c_value)
           :> t)
      | 162 ->
          (QualExprType.wrap context (c_value)
           :> t)
      | 163 ->
          (CharLiteralType.wrap context (c_value)
           :> t)
      | 164 ->
          (IdentifierType.wrap context (c_value)
           :> t)
      | 165 ->
          (OpAbsType.wrap context (c_value)
           :> t)
      | 166 ->
          (OpAndType.wrap context (c_value)
           :> t)
      | 167 ->
          (OpAndThenType.wrap context (c_value)
           :> t)
      | 168 ->
          (OpConcatType.wrap context (c_value)
           :> t)
      | 169 ->
          (OpDivType.wrap context (c_value)
           :> t)
      | 170 ->
          (OpDoubleDotType.wrap context (c_value)
           :> t)
      | 171 ->
          (OpEqType.wrap context (c_value)
           :> t)
      | 172 ->
          (OpGtType.wrap context (c_value)
           :> t)
      | 173 ->
          (OpGteType.wrap context (c_value)
           :> t)
      | 174 ->
          (OpInType.wrap context (c_value)
           :> t)
      | 175 ->
          (OpLtType.wrap context (c_value)
           :> t)
      | 176 ->
          (OpLteType.wrap context (c_value)
           :> t)
      | 177 ->
          (OpMinusType.wrap context (c_value)
           :> t)
      | 178 ->
          (OpModType.wrap context (c_value)
           :> t)
      | 179 ->
          (OpMultType.wrap context (c_value)
           :> t)
      | 180 ->
          (OpNeqType.wrap context (c_value)
           :> t)
      | 181 ->
          (OpNotType.wrap context (c_value)
           :> t)
      | 182 ->
          (OpNotInType.wrap context (c_value)
           :> t)
      | 183 ->
          (OpOrType.wrap context (c_value)
           :> t)
      | 184 ->
          (OpOrElseType.wrap context (c_value)
           :> t)
      | 185 ->
          (OpPlusType.wrap context (c_value)
           :> t)
      | 186 ->
          (OpPowType.wrap context (c_value)
           :> t)
      | 187 ->
          (OpRemType.wrap context (c_value)
           :> t)
      | 188 ->
          (OpXorType.wrap context (c_value)
           :> t)
      | 189 ->
          (StringLiteralType.wrap context (c_value)
           :> t)
      | 190 ->
          (NullLiteralType.wrap context (c_value)
           :> t)
      | 191 ->
          (IntLiteralType.wrap context (c_value)
           :> t)
      | 192 ->
          (RealLiteralType.wrap context (c_value)
           :> t)
      | 193 ->
          (TargetNameType.wrap context (c_value)
           :> t)
      | 194 ->
          (ParenExprType.wrap context (c_value)
           :> t)
      | 195 ->
          (QuantifiedExprType.wrap context (c_value)
           :> t)
      | 196 ->
          (RaiseExprType.wrap context (c_value)
           :> t)
      | 197 ->
          (UnOpType.wrap context (c_value)
           :> t)
      | 198 ->
          (HandledStmtsType.wrap context (c_value)
           :> t)
      | 199 ->
          (InterfaceKindLimitedType.wrap context (c_value)
           :> t)
      | 200 ->
          (InterfaceKindProtectedType.wrap context (c_value)
           :> t)
      | 201 ->
          (InterfaceKindSynchronizedType.wrap context (c_value)
           :> t)
      | 202 ->
          (InterfaceKindTaskType.wrap context (c_value)
           :> t)
      | 203 ->
          (IterTypeInType.wrap context (c_value)
           :> t)
      | 204 ->
          (IterTypeOfType.wrap context (c_value)
           :> t)
      | 205 ->
          (LibraryItemType.wrap context (c_value)
           :> t)
      | 206 ->
          (LimitedAbsentType.wrap context (c_value)
           :> t)
      | 207 ->
          (LimitedPresentType.wrap context (c_value)
           :> t)
      | 208 ->
          (ForLoopSpecType.wrap context (c_value)
           :> t)
      | 209 ->
          (WhileLoopSpecType.wrap context (c_value)
           :> t)
      | 210 ->
          (ModeDefaultType.wrap context (c_value)
           :> t)
      | 211 ->
          (ModeInType.wrap context (c_value)
           :> t)
      | 212 ->
          (ModeInOutType.wrap context (c_value)
           :> t)
      | 213 ->
          (ModeOutType.wrap context (c_value)
           :> t)
      | 214 ->
          (MultiAbstractStateDeclType.wrap context (c_value)
           :> t)
      | 215 ->
          (NotNullAbsentType.wrap context (c_value)
           :> t)
      | 216 ->
          (NotNullPresentType.wrap context (c_value)
           :> t)
      | 217 ->
          (NullComponentDeclType.wrap context (c_value)
           :> t)
      | 218 ->
          (OthersDesignatorType.wrap context (c_value)
           :> t)
      | 219 ->
          (OverridingNotOverridingType.wrap context (c_value)
           :> t)
      | 220 ->
          (OverridingOverridingType.wrap context (c_value)
           :> t)
      | 221 ->
          (OverridingUnspecifiedType.wrap context (c_value)
           :> t)
      | 222 ->
          (ParamsType.wrap context (c_value)
           :> t)
      | 223 ->
          (ParenAbstractStateDeclType.wrap context (c_value)
           :> t)
      | 224 ->
          (PragmaNodeType.wrap context (c_value)
           :> t)
      | 225 ->
          (PrivateAbsentType.wrap context (c_value)
           :> t)
      | 226 ->
          (PrivatePresentType.wrap context (c_value)
           :> t)
      | 227 ->
          (ProtectedDefType.wrap context (c_value)
           :> t)
      | 228 ->
          (ProtectedAbsentType.wrap context (c_value)
           :> t)
      | 229 ->
          (ProtectedPresentType.wrap context (c_value)
           :> t)
      | 230 ->
          (QuantifierAllType.wrap context (c_value)
           :> t)
      | 231 ->
          (QuantifierSomeType.wrap context (c_value)
           :> t)
      | 232 ->
          (RangeSpecType.wrap context (c_value)
           :> t)
      | 233 ->
          (RenamingClauseType.wrap context (c_value)
           :> t)
      | 234 ->
          (SyntheticRenamingClauseType.wrap context (c_value)
           :> t)
      | 235 ->
          (ReverseAbsentType.wrap context (c_value)
           :> t)
      | 236 ->
          (ReversePresentType.wrap context (c_value)
           :> t)
      | 237 ->
          (SelectWhenPartType.wrap context (c_value)
           :> t)
      | 238 ->
          (AcceptStmtType.wrap context (c_value)
           :> t)
      | 239 ->
          (AcceptStmtWithStmtsType.wrap context (c_value)
           :> t)
      | 240 ->
          (ForLoopStmtType.wrap context (c_value)
           :> t)
      | 241 ->
          (LoopStmtType.wrap context (c_value)
           :> t)
      | 242 ->
          (WhileLoopStmtType.wrap context (c_value)
           :> t)
      | 243 ->
          (BeginBlockType.wrap context (c_value)
           :> t)
      | 244 ->
          (DeclBlockType.wrap context (c_value)
           :> t)
      | 245 ->
          (CaseStmtType.wrap context (c_value)
           :> t)
      | 246 ->
          (ExtendedReturnStmtType.wrap context (c_value)
           :> t)
      | 247 ->
          (IfStmtType.wrap context (c_value)
           :> t)
      | 248 ->
          (NamedStmtType.wrap context (c_value)
           :> t)
      | 249 ->
          (SelectStmtType.wrap context (c_value)
           :> t)
      | 250 ->
          (ErrorStmtType.wrap context (c_value)
           :> t)
      | 251 ->
          (AbortStmtType.wrap context (c_value)
           :> t)
      | 252 ->
          (AssignStmtType.wrap context (c_value)
           :> t)
      | 253 ->
          (CallStmtType.wrap context (c_value)
           :> t)
      | 254 ->
          (DelayStmtType.wrap context (c_value)
           :> t)
      | 255 ->
          (ExitStmtType.wrap context (c_value)
           :> t)
      | 256 ->
          (GotoStmtType.wrap context (c_value)
           :> t)
      | 257 ->
          (LabelType.wrap context (c_value)
           :> t)
      | 258 ->
          (NullStmtType.wrap context (c_value)
           :> t)
      | 259 ->
          (RaiseStmtType.wrap context (c_value)
           :> t)
      | 260 ->
          (RequeueStmtType.wrap context (c_value)
           :> t)
      | 261 ->
          (ReturnStmtType.wrap context (c_value)
           :> t)
      | 262 ->
          (TerminateAlternativeType.wrap context (c_value)
           :> t)
      | 263 ->
          (SubpKindFunctionType.wrap context (c_value)
           :> t)
      | 264 ->
          (SubpKindProcedureType.wrap context (c_value)
           :> t)
      | 265 ->
          (SubunitType.wrap context (c_value)
           :> t)
      | 266 ->
          (SynchronizedAbsentType.wrap context (c_value)
           :> t)
      | 267 ->
          (SynchronizedPresentType.wrap context (c_value)
           :> t)
      | 268 ->
          (TaggedAbsentType.wrap context (c_value)
           :> t)
      | 269 ->
          (TaggedPresentType.wrap context (c_value)
           :> t)
      | 270 ->
          (TaskDefType.wrap context (c_value)
           :> t)
      | 271 ->
          (AccessToSubpDefType.wrap context (c_value)
           :> t)
      | 272 ->
          (AnonymousTypeAccessDefType.wrap context (c_value)
           :> t)
      | 273 ->
          (TypeAccessDefType.wrap context (c_value)
           :> t)
      | 274 ->
          (ArrayTypeDefType.wrap context (c_value)
           :> t)
      | 275 ->
          (DerivedTypeDefType.wrap context (c_value)
           :> t)
      | 276 ->
          (EnumTypeDefType.wrap context (c_value)
           :> t)
      | 277 ->
          (FormalDiscreteTypeDefType.wrap context (c_value)
           :> t)
      | 278 ->
          (InterfaceTypeDefType.wrap context (c_value)
           :> t)
      | 279 ->
          (ModIntTypeDefType.wrap context (c_value)
           :> t)
      | 280 ->
          (PrivateTypeDefType.wrap context (c_value)
           :> t)
      | 281 ->
          (DecimalFixedPointDefType.wrap context (c_value)
           :> t)
      | 282 ->
          (FloatingPointDefType.wrap context (c_value)
           :> t)
      | 283 ->
          (OrdinaryFixedPointDefType.wrap context (c_value)
           :> t)
      | 284 ->
          (RecordTypeDefType.wrap context (c_value)
           :> t)
      | 285 ->
          (SignedIntTypeDefType.wrap context (c_value)
           :> t)
      | 286 ->
          (AnonymousTypeType.wrap context (c_value)
           :> t)
      | 287 ->
          (EnumLitSynthTypeExprType.wrap context (c_value)
           :> t)
      | 288 ->
          (SubtypeIndicationType.wrap context (c_value)
           :> t)
      | 289 ->
          (ConstrainedSubtypeIndicationType.wrap context (c_value)
           :> t)
      | 290 ->
          (DiscreteSubtypeIndicationType.wrap context (c_value)
           :> t)
      | 291 ->
          (UnconstrainedArrayIndexType.wrap context (c_value)
           :> t)
      | 292 ->
          (UntilAbsentType.wrap context (c_value)
           :> t)
      | 293 ->
          (UntilPresentType.wrap context (c_value)
           :> t)
      | 294 ->
          (UsePackageClauseType.wrap context (c_value)
           :> t)
      | 295 ->
          (UseTypeClauseType.wrap context (c_value)
           :> t)
      | 296 ->
          (VariantType.wrap context (c_value)
           :> t)
      | 297 ->
          (VariantPartType.wrap context (c_value)
           :> t)
      | 298 ->
          (WithClauseType.wrap context (c_value)
           :> t)
      | 299 ->
          (WithPrivateAbsentType.wrap context (c_value)
           :> t)
      | 300 ->
          (WithPrivatePresentType.wrap context (c_value)
           :> t)
      | _ -> assert false

  let unwrap value =
    (* This is the unique unwrap function that can be called for any node. *)
    match (value :> AdaNodeType.t) with
    | `AbortAbsent fields -> fields.c_value
    | `AbortPresent fields -> fields.c_value
    | `AbstractAbsent fields -> fields.c_value
    | `AbstractPresent fields -> fields.c_value
    | `AdaNodeList fields -> fields.c_value
    | `AbstractStateDeclList fields -> fields.c_value
    | `AlternativesList fields -> fields.c_value
    | `ConstraintList fields -> fields.c_value
    | `DeclList fields -> fields.c_value
    | `StmtList fields -> fields.c_value
    | `AspectAssocList fields -> fields.c_value
    | `BaseAssocList fields -> fields.c_value
    | `AssocList fields -> fields.c_value
    | `BasicDeclList fields -> fields.c_value
    | `CaseExprAlternativeList fields -> fields.c_value
    | `CaseStmtAlternativeList fields -> fields.c_value
    | `CompilationUnitList fields -> fields.c_value
    | `ContractCaseAssocList fields -> fields.c_value
    | `DefiningNameList fields -> fields.c_value
    | `DiscriminantSpecList fields -> fields.c_value
    | `ElsifExprPartList fields -> fields.c_value
    | `ElsifStmtPartList fields -> fields.c_value
    | `EnumLiteralDeclList fields -> fields.c_value
    | `ExprAlternativesList fields -> fields.c_value
    | `DiscriminantChoiceList fields -> fields.c_value
    | `NameList fields -> fields.c_value
    | `ParentList fields -> fields.c_value
    | `ParamSpecList fields -> fields.c_value
    | `PragmaNodeList fields -> fields.c_value
    | `SelectWhenPartList fields -> fields.c_value
    | `UnconstrainedArrayIndexList fields -> fields.c_value
    | `VariantList fields -> fields.c_value
    | `AliasedAbsent fields -> fields.c_value
    | `AliasedPresent fields -> fields.c_value
    | `AllAbsent fields -> fields.c_value
    | `AllPresent fields -> fields.c_value
    | `ConstrainedArrayIndices fields -> fields.c_value
    | `UnconstrainedArrayIndices fields -> fields.c_value
    | `AspectAssoc fields -> fields.c_value
    | `AtClause fields -> fields.c_value
    | `AttributeDefClause fields -> fields.c_value
    | `EnumRepClause fields -> fields.c_value
    | `RecordRepClause fields -> fields.c_value
    | `AspectSpec fields -> fields.c_value
    | `ContractCaseAssoc fields -> fields.c_value
    | `PragmaArgumentAssoc fields -> fields.c_value
    | `EntrySpec fields -> fields.c_value
    | `EnumSubpSpec fields -> fields.c_value
    | `SubpSpec fields -> fields.c_value
    | `ComponentList fields -> fields.c_value
    | `KnownDiscriminantPart fields -> fields.c_value
    | `UnknownDiscriminantPart fields -> fields.c_value
    | `EntryCompletionFormalParams fields -> fields.c_value
    | `GenericFormalPart fields -> fields.c_value
    | `NullRecordDef fields -> fields.c_value
    | `RecordDef fields -> fields.c_value
    | `AggregateAssoc fields -> fields.c_value
    | `MultiDimArrayAssoc fields -> fields.c_value
    | `DiscriminantAssoc fields -> fields.c_value
    | `IteratedAssoc fields -> fields.c_value
    | `ParamAssoc fields -> fields.c_value
    | `AbstractStateDecl fields -> fields.c_value
    | `AnonymousExprDecl fields -> fields.c_value
    | `ComponentDecl fields -> fields.c_value
    | `DiscriminantSpec fields -> fields.c_value
    | `GenericFormalObjDecl fields -> fields.c_value
    | `GenericFormalPackage fields -> fields.c_value
    | `GenericFormalSubpDecl fields -> fields.c_value
    | `GenericFormalTypeDecl fields -> fields.c_value
    | `ParamSpec fields -> fields.c_value
    | `GenericPackageInternal fields -> fields.c_value
    | `PackageDecl fields -> fields.c_value
    | `DiscreteBaseSubtypeDecl fields -> fields.c_value
    | `SubtypeDecl fields -> fields.c_value
    | `ClasswideTypeDecl fields -> fields.c_value
    | `IncompleteTypeDecl fields -> fields.c_value
    | `IncompleteTaggedTypeDecl fields -> fields.c_value
    | `ProtectedTypeDecl fields -> fields.c_value
    | `TaskTypeDecl fields -> fields.c_value
    | `SingleTaskTypeDecl fields -> fields.c_value
    | `TypeDecl fields -> fields.c_value
    | `AnonymousTypeDecl fields -> fields.c_value
    | `SynthAnonymousTypeDecl fields -> fields.c_value
    | `AbstractSubpDecl fields -> fields.c_value
    | `AbstractFormalSubpDecl fields -> fields.c_value
    | `ConcreteFormalSubpDecl fields -> fields.c_value
    | `SubpDecl fields -> fields.c_value
    | `EntryDecl fields -> fields.c_value
    | `EnumLiteralDecl fields -> fields.c_value
    | `GenericSubpInternal fields -> fields.c_value
    | `ExprFunction fields -> fields.c_value
    | `NullSubpDecl fields -> fields.c_value
    | `SubpBody fields -> fields.c_value
    | `SubpRenamingDecl fields -> fields.c_value
    | `PackageBodyStub fields -> fields.c_value
    | `ProtectedBodyStub fields -> fields.c_value
    | `SubpBodyStub fields -> fields.c_value
    | `TaskBodyStub fields -> fields.c_value
    | `EntryBody fields -> fields.c_value
    | `PackageBody fields -> fields.c_value
    | `ProtectedBody fields -> fields.c_value
    | `TaskBody fields -> fields.c_value
    | `EntryIndexSpec fields -> fields.c_value
    | `ErrorDecl fields -> fields.c_value
    | `ExceptionDecl fields -> fields.c_value
    | `ExceptionHandler fields -> fields.c_value
    | `ForLoopVarDecl fields -> fields.c_value
    | `GenericPackageDecl fields -> fields.c_value
    | `GenericSubpDecl fields -> fields.c_value
    | `GenericPackageInstantiation fields -> fields.c_value
    | `GenericSubpInstantiation fields -> fields.c_value
    | `GenericPackageRenamingDecl fields -> fields.c_value
    | `GenericSubpRenamingDecl fields -> fields.c_value
    | `LabelDecl fields -> fields.c_value
    | `NamedStmtDecl fields -> fields.c_value
    | `NumberDecl fields -> fields.c_value
    | `ObjectDecl fields -> fields.c_value
    | `ExtendedReturnStmtObjectDecl fields -> fields.c_value
    | `PackageRenamingDecl fields -> fields.c_value
    | `SingleProtectedDecl fields -> fields.c_value
    | `SingleTaskDecl fields -> fields.c_value
    | `CaseStmtAlternative fields -> fields.c_value
    | `CompilationUnit fields -> fields.c_value
    | `ComponentClause fields -> fields.c_value
    | `ComponentDef fields -> fields.c_value
    | `ConstantAbsent fields -> fields.c_value
    | `ConstantPresent fields -> fields.c_value
    | `DeltaConstraint fields -> fields.c_value
    | `DigitsConstraint fields -> fields.c_value
    | `DiscriminantConstraint fields -> fields.c_value
    | `IndexConstraint fields -> fields.c_value
    | `RangeConstraint fields -> fields.c_value
    | `DeclarativePart fields -> fields.c_value
    | `PrivatePart fields -> fields.c_value
    | `PublicPart fields -> fields.c_value
    | `ElsifExprPart fields -> fields.c_value
    | `ElsifStmtPart fields -> fields.c_value
    | `AbstractStateDeclExpr fields -> fields.c_value
    | `Allocator fields -> fields.c_value
    | `Aggregate fields -> fields.c_value
    | `BracketAggregate fields -> fields.c_value
    | `DeltaAggregate fields -> fields.c_value
    | `BracketDeltaAggregate fields -> fields.c_value
    | `NullRecordAggregate fields -> fields.c_value
    | `BinOp fields -> fields.c_value
    | `RelationOp fields -> fields.c_value
    | `BoxExpr fields -> fields.c_value
    | `CaseExprAlternative fields -> fields.c_value
    | `CaseExpr fields -> fields.c_value
    | `IfExpr fields -> fields.c_value
    | `ContractCases fields -> fields.c_value
    | `DeclExpr fields -> fields.c_value
    | `MembershipExpr fields -> fields.c_value
    | `AttributeRef fields -> fields.c_value
    | `UpdateAttributeRef fields -> fields.c_value
    | `CallExpr fields -> fields.c_value
    | `DefiningName fields -> fields.c_value
    | `DiscreteSubtypeName fields -> fields.c_value
    | `DottedName fields -> fields.c_value
    | `EndName fields -> fields.c_value
    | `ExplicitDeref fields -> fields.c_value
    | `QualExpr fields -> fields.c_value
    | `CharLiteral fields -> fields.c_value
    | `Identifier fields -> fields.c_value
    | `OpAbs fields -> fields.c_value
    | `OpAnd fields -> fields.c_value
    | `OpAndThen fields -> fields.c_value
    | `OpConcat fields -> fields.c_value
    | `OpDiv fields -> fields.c_value
    | `OpDoubleDot fields -> fields.c_value
    | `OpEq fields -> fields.c_value
    | `OpGt fields -> fields.c_value
    | `OpGte fields -> fields.c_value
    | `OpIn fields -> fields.c_value
    | `OpLt fields -> fields.c_value
    | `OpLte fields -> fields.c_value
    | `OpMinus fields -> fields.c_value
    | `OpMod fields -> fields.c_value
    | `OpMult fields -> fields.c_value
    | `OpNeq fields -> fields.c_value
    | `OpNot fields -> fields.c_value
    | `OpNotIn fields -> fields.c_value
    | `OpOr fields -> fields.c_value
    | `OpOrElse fields -> fields.c_value
    | `OpPlus fields -> fields.c_value
    | `OpPow fields -> fields.c_value
    | `OpRem fields -> fields.c_value
    | `OpXor fields -> fields.c_value
    | `StringLiteral fields -> fields.c_value
    | `NullLiteral fields -> fields.c_value
    | `IntLiteral fields -> fields.c_value
    | `RealLiteral fields -> fields.c_value
    | `TargetName fields -> fields.c_value
    | `ParenExpr fields -> fields.c_value
    | `QuantifiedExpr fields -> fields.c_value
    | `RaiseExpr fields -> fields.c_value
    | `UnOp fields -> fields.c_value
    | `HandledStmts fields -> fields.c_value
    | `InterfaceKindLimited fields -> fields.c_value
    | `InterfaceKindProtected fields -> fields.c_value
    | `InterfaceKindSynchronized fields -> fields.c_value
    | `InterfaceKindTask fields -> fields.c_value
    | `IterTypeIn fields -> fields.c_value
    | `IterTypeOf fields -> fields.c_value
    | `LibraryItem fields -> fields.c_value
    | `LimitedAbsent fields -> fields.c_value
    | `LimitedPresent fields -> fields.c_value
    | `ForLoopSpec fields -> fields.c_value
    | `WhileLoopSpec fields -> fields.c_value
    | `ModeDefault fields -> fields.c_value
    | `ModeIn fields -> fields.c_value
    | `ModeInOut fields -> fields.c_value
    | `ModeOut fields -> fields.c_value
    | `MultiAbstractStateDecl fields -> fields.c_value
    | `NotNullAbsent fields -> fields.c_value
    | `NotNullPresent fields -> fields.c_value
    | `NullComponentDecl fields -> fields.c_value
    | `OthersDesignator fields -> fields.c_value
    | `OverridingNotOverriding fields -> fields.c_value
    | `OverridingOverriding fields -> fields.c_value
    | `OverridingUnspecified fields -> fields.c_value
    | `Params fields -> fields.c_value
    | `ParenAbstractStateDecl fields -> fields.c_value
    | `PragmaNode fields -> fields.c_value
    | `PrivateAbsent fields -> fields.c_value
    | `PrivatePresent fields -> fields.c_value
    | `ProtectedDef fields -> fields.c_value
    | `ProtectedAbsent fields -> fields.c_value
    | `ProtectedPresent fields -> fields.c_value
    | `QuantifierAll fields -> fields.c_value
    | `QuantifierSome fields -> fields.c_value
    | `RangeSpec fields -> fields.c_value
    | `RenamingClause fields -> fields.c_value
    | `SyntheticRenamingClause fields -> fields.c_value
    | `ReverseAbsent fields -> fields.c_value
    | `ReversePresent fields -> fields.c_value
    | `SelectWhenPart fields -> fields.c_value
    | `AcceptStmt fields -> fields.c_value
    | `AcceptStmtWithStmts fields -> fields.c_value
    | `ForLoopStmt fields -> fields.c_value
    | `LoopStmt fields -> fields.c_value
    | `WhileLoopStmt fields -> fields.c_value
    | `BeginBlock fields -> fields.c_value
    | `DeclBlock fields -> fields.c_value
    | `CaseStmt fields -> fields.c_value
    | `ExtendedReturnStmt fields -> fields.c_value
    | `IfStmt fields -> fields.c_value
    | `NamedStmt fields -> fields.c_value
    | `SelectStmt fields -> fields.c_value
    | `ErrorStmt fields -> fields.c_value
    | `AbortStmt fields -> fields.c_value
    | `AssignStmt fields -> fields.c_value
    | `CallStmt fields -> fields.c_value
    | `DelayStmt fields -> fields.c_value
    | `ExitStmt fields -> fields.c_value
    | `GotoStmt fields -> fields.c_value
    | `Label fields -> fields.c_value
    | `NullStmt fields -> fields.c_value
    | `RaiseStmt fields -> fields.c_value
    | `RequeueStmt fields -> fields.c_value
    | `ReturnStmt fields -> fields.c_value
    | `TerminateAlternative fields -> fields.c_value
    | `SubpKindFunction fields -> fields.c_value
    | `SubpKindProcedure fields -> fields.c_value
    | `Subunit fields -> fields.c_value
    | `SynchronizedAbsent fields -> fields.c_value
    | `SynchronizedPresent fields -> fields.c_value
    | `TaggedAbsent fields -> fields.c_value
    | `TaggedPresent fields -> fields.c_value
    | `TaskDef fields -> fields.c_value
    | `AccessToSubpDef fields -> fields.c_value
    | `AnonymousTypeAccessDef fields -> fields.c_value
    | `TypeAccessDef fields -> fields.c_value
    | `ArrayTypeDef fields -> fields.c_value
    | `DerivedTypeDef fields -> fields.c_value
    | `EnumTypeDef fields -> fields.c_value
    | `FormalDiscreteTypeDef fields -> fields.c_value
    | `InterfaceTypeDef fields -> fields.c_value
    | `ModIntTypeDef fields -> fields.c_value
    | `PrivateTypeDef fields -> fields.c_value
    | `DecimalFixedPointDef fields -> fields.c_value
    | `FloatingPointDef fields -> fields.c_value
    | `OrdinaryFixedPointDef fields -> fields.c_value
    | `RecordTypeDef fields -> fields.c_value
    | `SignedIntTypeDef fields -> fields.c_value
    | `AnonymousType fields -> fields.c_value
    | `EnumLitSynthTypeExpr fields -> fields.c_value
    | `SubtypeIndication fields -> fields.c_value
    | `ConstrainedSubtypeIndication fields -> fields.c_value
    | `DiscreteSubtypeIndication fields -> fields.c_value
    | `UnconstrainedArrayIndex fields -> fields.c_value
    | `UntilAbsent fields -> fields.c_value
    | `UntilPresent fields -> fields.c_value
    | `UsePackageClause fields -> fields.c_value
    | `UseTypeClause fields -> fields.c_value
    | `Variant fields -> fields.c_value
    | `VariantPart fields -> fields.c_value
    | `WithClause fields -> fields.c_value
    | `WithPrivateAbsent fields -> fields.c_value
    | `WithPrivatePresent fields -> fields.c_value

end

      
and AbortNodeType : sig
   
  
   
  type t = [
    | `AbortAbsent
        of AbortAbsentType.fields
    | `AbortPresent
        of AbortPresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbortAbsent
        of AbortAbsentType.fields
    | `AbortPresent
        of AbortPresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AbortAbsent _
      | `AbortPresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AbortAbsentType : sig
   
  
   
  type t = [
    | `AbortAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbortAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AbortAbsent {
        c_value = c_value;
        context = context
      }

end

      
and AbortPresentType : sig
   
  
   
  type t = [
    | `AbortPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbortPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AbortPresent {
        c_value = c_value;
        context = context
      }

end

      
and AbstractNodeType : sig
   
  
   
  type t = [
    | `AbstractAbsent
        of AbstractAbsentType.fields
    | `AbstractPresent
        of AbstractPresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbstractAbsent
        of AbstractAbsentType.fields
    | `AbstractPresent
        of AbstractPresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AbstractAbsent _
      | `AbstractPresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AbstractAbsentType : sig
   
  
   
  type t = [
    | `AbstractAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbstractAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AbstractAbsent {
        c_value = c_value;
        context = context
      }

end

      
and AbstractPresentType : sig
   
  
   
  type t = [
    | `AbstractPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbstractPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AbstractPresent {
        c_value = c_value;
        context = context
      }

end

      
and AdaListType : sig
   
  
   
  (**
    * AdaNodeListType.t
    * AspectAssocListType.t
    * BaseAssocListType.t
    * BasicAssocListType.t
    * BasicDeclListType.t
    * CaseExprAlternativeListType.t
    * CaseStmtAlternativeListType.t
    * CompilationUnitListType.t
    * ContractCaseAssocListType.t
    * DefiningNameListType.t
    * DiscriminantSpecListType.t
    * ElsifExprPartListType.t
    * ElsifStmtPartListType.t
    * EnumLiteralDeclListType.t
    * ExprListType.t
    * IdentifierListType.t
    * NameListType.t
    * ParamSpecListType.t
    * PragmaNodeListType.t
    * SelectWhenPartListType.t
    * UnconstrainedArrayIndexListType.t
    * VariantListType.t
    *)
  type t = [
    | `AdaNodeList
        of AdaNodeListType.fields
    | `AbstractStateDeclList
        of AbstractStateDeclListType.fields
    | `AlternativesList
        of AlternativesListType.fields
    | `ConstraintList
        of ConstraintListType.fields
    | `DeclList
        of DeclListType.fields
    | `StmtList
        of StmtListType.fields
    | `AspectAssocList
        of AspectAssocListType.fields
    | `BaseAssocList
        of BaseAssocListType.fields
    | `AssocList
        of AssocListType.fields
    | `BasicDeclList
        of BasicDeclListType.fields
    | `CaseExprAlternativeList
        of CaseExprAlternativeListType.fields
    | `CaseStmtAlternativeList
        of CaseStmtAlternativeListType.fields
    | `CompilationUnitList
        of CompilationUnitListType.fields
    | `ContractCaseAssocList
        of ContractCaseAssocListType.fields
    | `DefiningNameList
        of DefiningNameListType.fields
    | `DiscriminantSpecList
        of DiscriminantSpecListType.fields
    | `ElsifExprPartList
        of ElsifExprPartListType.fields
    | `ElsifStmtPartList
        of ElsifStmtPartListType.fields
    | `EnumLiteralDeclList
        of EnumLiteralDeclListType.fields
    | `ExprAlternativesList
        of ExprAlternativesListType.fields
    | `DiscriminantChoiceList
        of DiscriminantChoiceListType.fields
    | `NameList
        of NameListType.fields
    | `ParentList
        of ParentListType.fields
    | `ParamSpecList
        of ParamSpecListType.fields
    | `PragmaNodeList
        of PragmaNodeListType.fields
    | `SelectWhenPartList
        of SelectWhenPartListType.fields
    | `UnconstrainedArrayIndexList
        of UnconstrainedArrayIndexListType.fields
    | `VariantList
        of VariantListType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * AdaNodeListType.t
    * AspectAssocListType.t
    * BaseAssocListType.t
    * BasicAssocListType.t
    * BasicDeclListType.t
    * CaseExprAlternativeListType.t
    * CaseStmtAlternativeListType.t
    * CompilationUnitListType.t
    * ContractCaseAssocListType.t
    * DefiningNameListType.t
    * DiscriminantSpecListType.t
    * ElsifExprPartListType.t
    * ElsifStmtPartListType.t
    * EnumLiteralDeclListType.t
    * ExprListType.t
    * IdentifierListType.t
    * NameListType.t
    * ParamSpecListType.t
    * PragmaNodeListType.t
    * SelectWhenPartListType.t
    * UnconstrainedArrayIndexListType.t
    * VariantListType.t
    *)
  type t = [
    | `AdaNodeList
        of AdaNodeListType.fields
    | `AbstractStateDeclList
        of AbstractStateDeclListType.fields
    | `AlternativesList
        of AlternativesListType.fields
    | `ConstraintList
        of ConstraintListType.fields
    | `DeclList
        of DeclListType.fields
    | `StmtList
        of StmtListType.fields
    | `AspectAssocList
        of AspectAssocListType.fields
    | `BaseAssocList
        of BaseAssocListType.fields
    | `AssocList
        of AssocListType.fields
    | `BasicDeclList
        of BasicDeclListType.fields
    | `CaseExprAlternativeList
        of CaseExprAlternativeListType.fields
    | `CaseStmtAlternativeList
        of CaseStmtAlternativeListType.fields
    | `CompilationUnitList
        of CompilationUnitListType.fields
    | `ContractCaseAssocList
        of ContractCaseAssocListType.fields
    | `DefiningNameList
        of DefiningNameListType.fields
    | `DiscriminantSpecList
        of DiscriminantSpecListType.fields
    | `ElsifExprPartList
        of ElsifExprPartListType.fields
    | `ElsifStmtPartList
        of ElsifStmtPartListType.fields
    | `EnumLiteralDeclList
        of EnumLiteralDeclListType.fields
    | `ExprAlternativesList
        of ExprAlternativesListType.fields
    | `DiscriminantChoiceList
        of DiscriminantChoiceListType.fields
    | `NameList
        of NameListType.fields
    | `ParentList
        of ParentListType.fields
    | `ParamSpecList
        of ParamSpecListType.fields
    | `PragmaNodeList
        of PragmaNodeListType.fields
    | `SelectWhenPartList
        of SelectWhenPartListType.fields
    | `UnconstrainedArrayIndexList
        of UnconstrainedArrayIndexListType.fields
    | `VariantList
        of VariantListType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AdaNodeList _
      | `AbstractStateDeclList _
      | `AlternativesList _
      | `ConstraintList _
      | `DeclList _
      | `StmtList _
      | `AspectAssocList _
      | `BaseAssocList _
      | `AssocList _
      | `BasicDeclList _
      | `CaseExprAlternativeList _
      | `CaseStmtAlternativeList _
      | `CompilationUnitList _
      | `ContractCaseAssocList _
      | `DefiningNameList _
      | `DiscriminantSpecList _
      | `ElsifExprPartList _
      | `ElsifStmtPartList _
      | `EnumLiteralDeclList _
      | `ExprAlternativesList _
      | `DiscriminantChoiceList _
      | `NameList _
      | `ParentList _
      | `ParamSpecList _
      | `PragmaNodeList _
      | `SelectWhenPartList _
      | `UnconstrainedArrayIndexList _
      | `VariantList _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AdaNodeListType : sig
   
  
   
  type t = [
    | `AdaNodeList
        of fields
    | `AbstractStateDeclList
        of AbstractStateDeclListType.fields
    | `AlternativesList
        of AlternativesListType.fields
    | `ConstraintList
        of ConstraintListType.fields
    | `DeclList
        of DeclListType.fields
    | `StmtList
        of StmtListType.fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AdaNodeList
        of fields
    | `AbstractStateDeclList
        of AbstractStateDeclListType.fields
    | `AlternativesList
        of AlternativesListType.fields
    | `ConstraintList
        of ConstraintListType.fields
    | `DeclList
        of DeclListType.fields
    | `StmtList
        of StmtListType.fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        AdaNodeType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AdaNodeList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and AbstractStateDeclListType : sig
   
  
   
  type t = [
    | `AbstractStateDeclList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbstractStateDeclList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        AdaNodeType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AbstractStateDeclList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and AlternativesListType : sig
   
  
   
  type t = [
    | `AlternativesList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AlternativesList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        AdaNodeType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AlternativesList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and ConstraintListType : sig
   
  
   
  type t = [
    | `ConstraintList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ConstraintList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        AdaNodeType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ConstraintList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and DeclListType : sig
   
  
   
  type t = [
    | `DeclList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DeclList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        AdaNodeType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DeclList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and StmtListType : sig
   
  
   
  type t = [
    | `StmtList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `StmtList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        AdaNodeType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `StmtList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and AspectAssocListType : sig
   
  
   
  type t = [
    | `AspectAssocList
        of fields
  ]
  and fields = 
  {
    list : AspectAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AspectAssocList
        of fields
  ]
  and fields = 
  {
    list : AspectAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        AspectAssocType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AspectAssocList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and BaseAssocListType : sig
   
  
   
  type t = [
    | `BaseAssocList
        of fields
  ]
  and fields = 
  {
    list : BaseAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `BaseAssocList
        of fields
  ]
  and fields = 
  {
    list : BaseAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        BaseAssocType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `BaseAssocList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and BasicAssocListType : sig
   
  
   
  type t = [
    | `AssocList
        of AssocListType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AssocList
        of AssocListType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AssocList _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AssocListType : sig
   
  
   
  type t = [
    | `AssocList
        of fields
  ]
  and fields = 
  {
    list : BasicAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AssocList
        of fields
  ]
  and fields = 
  {
    list : BasicAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        BasicAssocType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AssocList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and BasicDeclListType : sig
   
  
   
  type t = [
    | `BasicDeclList
        of fields
  ]
  and fields = 
  {
    list : BasicDeclType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `BasicDeclList
        of fields
  ]
  and fields = 
  {
    list : BasicDeclType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        BasicDeclType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `BasicDeclList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and CaseExprAlternativeListType : sig
   
  
   
  type t = [
    | `CaseExprAlternativeList
        of fields
  ]
  and fields = 
  {
    list : CaseExprAlternativeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CaseExprAlternativeList
        of fields
  ]
  and fields = 
  {
    list : CaseExprAlternativeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        CaseExprAlternativeType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `CaseExprAlternativeList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and CaseStmtAlternativeListType : sig
   
  
   
  type t = [
    | `CaseStmtAlternativeList
        of fields
  ]
  and fields = 
  {
    list : CaseStmtAlternativeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CaseStmtAlternativeList
        of fields
  ]
  and fields = 
  {
    list : CaseStmtAlternativeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        CaseStmtAlternativeType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `CaseStmtAlternativeList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and CompilationUnitListType : sig
   
  
   
  type t = [
    | `CompilationUnitList
        of fields
  ]
  and fields = 
  {
    list : CompilationUnitType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CompilationUnitList
        of fields
  ]
  and fields = 
  {
    list : CompilationUnitType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        CompilationUnitType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `CompilationUnitList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and ContractCaseAssocListType : sig
   
  
   
  type t = [
    | `ContractCaseAssocList
        of fields
  ]
  and fields = 
  {
    list : ContractCaseAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ContractCaseAssocList
        of fields
  ]
  and fields = 
  {
    list : ContractCaseAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        ContractCaseAssocType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ContractCaseAssocList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and DefiningNameListType : sig
   
  
   
  type t = [
    | `DefiningNameList
        of fields
  ]
  and fields = 
  {
    list : DefiningNameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DefiningNameList
        of fields
  ]
  and fields = 
  {
    list : DefiningNameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        DefiningNameType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DefiningNameList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and DiscriminantSpecListType : sig
   
  
   
  type t = [
    | `DiscriminantSpecList
        of fields
  ]
  and fields = 
  {
    list : DiscriminantSpecType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DiscriminantSpecList
        of fields
  ]
  and fields = 
  {
    list : DiscriminantSpecType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        DiscriminantSpecType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DiscriminantSpecList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and ElsifExprPartListType : sig
   
  
   
  type t = [
    | `ElsifExprPartList
        of fields
  ]
  and fields = 
  {
    list : ElsifExprPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ElsifExprPartList
        of fields
  ]
  and fields = 
  {
    list : ElsifExprPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        ElsifExprPartType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ElsifExprPartList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and ElsifStmtPartListType : sig
   
  
   
  type t = [
    | `ElsifStmtPartList
        of fields
  ]
  and fields = 
  {
    list : ElsifStmtPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ElsifStmtPartList
        of fields
  ]
  and fields = 
  {
    list : ElsifStmtPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        ElsifStmtPartType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ElsifStmtPartList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and EnumLiteralDeclListType : sig
   
  
   
  type t = [
    | `EnumLiteralDeclList
        of fields
  ]
  and fields = 
  {
    list : EnumLiteralDeclType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EnumLiteralDeclList
        of fields
  ]
  and fields = 
  {
    list : EnumLiteralDeclType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        EnumLiteralDeclType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EnumLiteralDeclList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and ExprListType : sig
   
  
   
  type t = [
    | `ExprAlternativesList
        of ExprAlternativesListType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ExprAlternativesList
        of ExprAlternativesListType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ExprAlternativesList _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ExprAlternativesListType : sig
   
  
   
  type t = [
    | `ExprAlternativesList
        of fields
  ]
  and fields = 
  {
    list : ExprType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ExprAlternativesList
        of fields
  ]
  and fields = 
  {
    list : ExprType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        ExprType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ExprAlternativesList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and IdentifierListType : sig
   
  
   
  type t = [
    | `DiscriminantChoiceList
        of DiscriminantChoiceListType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DiscriminantChoiceList
        of DiscriminantChoiceListType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `DiscriminantChoiceList _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and DiscriminantChoiceListType : sig
   
  
   
  type t = [
    | `DiscriminantChoiceList
        of fields
  ]
  and fields = 
  {
    list : IdentifierType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DiscriminantChoiceList
        of fields
  ]
  and fields = 
  {
    list : IdentifierType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        IdentifierType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DiscriminantChoiceList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and NameListType : sig
   
  
   
  type t = [
    | `NameList
        of fields
    | `ParentList
        of ParentListType.fields
  ]
  and fields = 
  {
    list : NameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NameList
        of fields
    | `ParentList
        of ParentListType.fields
  ]
  and fields = 
  {
    list : NameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        NameType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NameList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and ParentListType : sig
   
  
   
  type t = [
    | `ParentList
        of fields
  ]
  and fields = 
  {
    list : NameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ParentList
        of fields
  ]
  and fields = 
  {
    list : NameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        NameType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ParentList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and ParamSpecListType : sig
   
  
   
  type t = [
    | `ParamSpecList
        of fields
  ]
  and fields = 
  {
    list : ParamSpecType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ParamSpecList
        of fields
  ]
  and fields = 
  {
    list : ParamSpecType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        ParamSpecType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ParamSpecList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and PragmaNodeListType : sig
   
  
   
  type t = [
    | `PragmaNodeList
        of fields
  ]
  and fields = 
  {
    list : PragmaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PragmaNodeList
        of fields
  ]
  and fields = 
  {
    list : PragmaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        PragmaNodeType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PragmaNodeList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and SelectWhenPartListType : sig
   
  
   
  type t = [
    | `SelectWhenPartList
        of fields
  ]
  and fields = 
  {
    list : SelectWhenPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SelectWhenPartList
        of fields
  ]
  and fields = 
  {
    list : SelectWhenPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        SelectWhenPartType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SelectWhenPartList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and UnconstrainedArrayIndexListType : sig
   
  
   
  type t = [
    | `UnconstrainedArrayIndexList
        of fields
  ]
  and fields = 
  {
    list : UnconstrainedArrayIndexType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UnconstrainedArrayIndexList
        of fields
  ]
  and fields = 
  {
    list : UnconstrainedArrayIndexType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        UnconstrainedArrayIndexType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `UnconstrainedArrayIndexList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and VariantListType : sig
   
  
   
  type t = [
    | `VariantList
        of fields
  ]
  and fields = 
  {
    list : VariantType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `VariantList
        of fields
  ]
  and fields = 
  {
    list : VariantType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let list () =
      let c_value_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.ada_node_children
          (addr c_value)
          (c_value_ptr)
      in
      let c_value = !@(!@(c_value_ptr)) in
      let length = getf c_value AdaNodeArrayStruct.n in
      let items = c_value @. AdaNodeArrayStruct.items in
      let f i =
        let fresh = allocate EntityStruct.c_type !@(items +@ i) in
        (* This can raise a SyntaxError, which is expected *)
        VariantType.wrap context ((!@ fresh))
      in
      let result = List.init length f in
      AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
      result
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `VariantList {
        list = Lazy.from_fun list;
        c_value = c_value;
        context = context
      }

end

      
and AliasedNodeType : sig
   
  
   
  type t = [
    | `AliasedAbsent
        of AliasedAbsentType.fields
    | `AliasedPresent
        of AliasedPresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AliasedAbsent
        of AliasedAbsentType.fields
    | `AliasedPresent
        of AliasedPresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AliasedAbsent _
      | `AliasedPresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AliasedAbsentType : sig
   
  
   
  type t = [
    | `AliasedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AliasedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AliasedAbsent {
        c_value = c_value;
        context = context
      }

end

      
and AliasedPresentType : sig
   
  
   
  type t = [
    | `AliasedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AliasedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AliasedPresent {
        c_value = c_value;
        context = context
      }

end

      
and AllNodeType : sig
   
  
   
  type t = [
    | `AllAbsent
        of AllAbsentType.fields
    | `AllPresent
        of AllPresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AllAbsent
        of AllAbsentType.fields
    | `AllPresent
        of AllPresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AllAbsent _
      | `AllPresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AllAbsentType : sig
   
  
   
  type t = [
    | `AllAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AllAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AllAbsent {
        c_value = c_value;
        context = context
      }

end

      
and AllPresentType : sig
   
  
   
  type t = [
    | `AllPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AllPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AllPresent {
        c_value = c_value;
        context = context
      }

end

      
and ArrayIndicesType : sig
   
  
   
  type t = [
    | `ConstrainedArrayIndices
        of ConstrainedArrayIndicesType.fields
    | `UnconstrainedArrayIndices
        of UnconstrainedArrayIndicesType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ConstrainedArrayIndices
        of ConstrainedArrayIndicesType.fields
    | `UnconstrainedArrayIndices
        of UnconstrainedArrayIndicesType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ConstrainedArrayIndices _
      | `UnconstrainedArrayIndices _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ConstrainedArrayIndicesType : sig
   
  
   
  type t = [
    | `ConstrainedArrayIndices
        of fields
  ]
  and fields = 
  {
         
    f_list: ConstraintListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ConstrainedArrayIndices
        of fields
  ]
  and fields = 
  {
         
    f_list: ConstraintListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_list () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.constrained_array_indices_f_list
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ConstraintListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ConstrainedArrayIndices {
        f_list
          = Lazy.from_fun f_list ;
        c_value = c_value;
        context = context
      }

end

      
and UnconstrainedArrayIndicesType : sig
   
  
   
  type t = [
    | `UnconstrainedArrayIndices
        of fields
  ]
  and fields = 
  {
         
    f_types: UnconstrainedArrayIndexListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UnconstrainedArrayIndices
        of fields
  ]
  and fields = 
  {
         
    f_types: UnconstrainedArrayIndexListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_types () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.unconstrained_array_indices_f_types
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         UnconstrainedArrayIndexListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `UnconstrainedArrayIndices {
        f_types
          = Lazy.from_fun f_types ;
        c_value = c_value;
        context = context
      }

end

      
and AspectAssocType : sig
   
  
   
  type t = [
    | `AspectAssoc
        of fields
  ]
  and fields = 
  {
         
    f_id: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `AbstractStateDeclExpr
          of AbstractStateDeclExprType.fields
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ContractCases
          of ContractCasesType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AspectAssoc
        of fields
  ]
  and fields = 
  {
         
    f_id: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `AbstractStateDeclExpr
          of AbstractStateDeclExprType.fields
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ContractCases
          of ContractCasesType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_id () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.aspect_assoc_f_id
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.aspect_assoc_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `AbstractStateDeclExpr _
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `ContractCases _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AspectAssoc {
        f_id
          = Lazy.from_fun f_id ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and AspectClauseType : sig
   
  
   
  type t = [
    | `AtClause
        of AtClauseType.fields
    | `AttributeDefClause
        of AttributeDefClauseType.fields
    | `EnumRepClause
        of EnumRepClauseType.fields
    | `RecordRepClause
        of RecordRepClauseType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AtClause
        of AtClauseType.fields
    | `AttributeDefClause
        of AttributeDefClauseType.fields
    | `EnumRepClause
        of EnumRepClauseType.fields
    | `RecordRepClause
        of RecordRepClauseType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AtClause _
      | `AttributeDefClause _
      | `EnumRepClause _
      | `RecordRepClause _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AtClauseType : sig
   
  
   
  type t = [
    | `AtClause
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AtClause
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.at_clause_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BaseIdType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.at_clause_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AtClause {
        f_name
          = Lazy.from_fun f_name ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and AttributeDefClauseType : sig
   
  
   
  type t = [
    | `AttributeDefClause
        of fields
  ]
  and fields = 
  {
         
    f_attribute_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AttributeDefClause
        of fields
  ]
  and fields = 
  {
         
    f_attribute_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_attribute_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.attribute_def_clause_f_attribute_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.attribute_def_clause_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AttributeDefClause {
        f_attribute_expr
          = Lazy.from_fun f_attribute_expr ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and EnumRepClauseType : sig
   
  
   
  type t = [
    | `EnumRepClause
        of fields
  ]
  and fields = 
  {
         
    f_type_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_aggregate: BaseAggregateType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EnumRepClause
        of fields
  ]
  and fields = 
  {
         
    f_type_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_aggregate: BaseAggregateType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_type_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.enum_rep_clause_f_type_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    let f_aggregate () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.enum_rep_clause_f_aggregate
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BaseAggregateType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EnumRepClause {
        f_type_name
          = Lazy.from_fun f_type_name ;
        f_aggregate
          = Lazy.from_fun f_aggregate ;
        c_value = c_value;
        context = context
      }

end

      
and RecordRepClauseType : sig
   
  
   
  type t = [
    | `RecordRepClause
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_at_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_components: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `RecordRepClause
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_at_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_components: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.record_rep_clause_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    let f_at_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.record_rep_clause_f_at_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    let f_components () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.record_rep_clause_f_components
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `RecordRepClause {
        f_name
          = Lazy.from_fun f_name ;
        f_at_expr
          = Lazy.from_fun f_at_expr ;
        f_components
          = Lazy.from_fun f_components ;
        c_value = c_value;
        context = context
      }

end

      
and AspectSpecType : sig
   
  
   
  type t = [
    | `AspectSpec
        of fields
  ]
  and fields = 
  {
         
    f_aspect_assocs: AspectAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AspectSpec
        of fields
  ]
  and fields = 
  {
         
    f_aspect_assocs: AspectAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_aspect_assocs () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.aspect_spec_f_aspect_assocs
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AspectAssocListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AspectSpec {
        f_aspect_assocs
          = Lazy.from_fun f_aspect_assocs ;
        c_value = c_value;
        context = context
      }

end

      
and BaseAssocType : sig
   
  
   
  type t = [
    | `ContractCaseAssoc
        of ContractCaseAssocType.fields
    | `PragmaArgumentAssoc
        of PragmaArgumentAssocType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ContractCaseAssoc
        of ContractCaseAssocType.fields
    | `PragmaArgumentAssoc
        of PragmaArgumentAssocType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ContractCaseAssoc _
      | `PragmaArgumentAssoc _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ContractCaseAssocType : sig
   
  
   
  type t = [
    | `ContractCaseAssoc
        of fields
  ]
  and fields = 
  {
         
    f_guard: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `OthersDesignator
          of OthersDesignatorType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_consequence: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ContractCaseAssoc
        of fields
  ]
  and fields = 
  {
         
    f_guard: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `OthersDesignator
          of OthersDesignatorType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_consequence: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_guard () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.contract_case_assoc_f_guard
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `OthersDesignator _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_consequence () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.contract_case_assoc_f_consequence
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ContractCaseAssoc {
        f_guard
          = Lazy.from_fun f_guard ;
        f_consequence
          = Lazy.from_fun f_consequence ;
        c_value = c_value;
        context = context
      }

end

      
and PragmaArgumentAssocType : sig
   
  
   
  type t = [
    | `PragmaArgumentAssoc
        of fields
  ]
  and fields = 
  {
         
    f_id: IdentifierType.t
    option
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PragmaArgumentAssoc
        of fields
  ]
  and fields = 
  {
         
    f_id: IdentifierType.t
    option
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_id () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.pragma_argument_assoc_f_id
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (IdentifierType.wrap context (field_c_value))
      in
         

      node
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.pragma_argument_assoc_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PragmaArgumentAssoc {
        f_id
          = Lazy.from_fun f_id ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and BaseFormalParamHolderType : sig
   
  
   
  (**
    * BaseSubpSpecType.t
    * ComponentListType.t
    * DiscriminantPartType.t
    * EntryCompletionFormalParamsType.t
    * GenericFormalPartType.t
    *)
  type t = [
    | `EntrySpec
        of EntrySpecType.fields
    | `EnumSubpSpec
        of EnumSubpSpecType.fields
    | `SubpSpec
        of SubpSpecType.fields
    | `ComponentList
        of ComponentListType.fields
    | `KnownDiscriminantPart
        of KnownDiscriminantPartType.fields
    | `UnknownDiscriminantPart
        of UnknownDiscriminantPartType.fields
    | `EntryCompletionFormalParams
        of EntryCompletionFormalParamsType.fields
    | `GenericFormalPart
        of GenericFormalPartType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * BaseSubpSpecType.t
    * ComponentListType.t
    * DiscriminantPartType.t
    * EntryCompletionFormalParamsType.t
    * GenericFormalPartType.t
    *)
  type t = [
    | `EntrySpec
        of EntrySpecType.fields
    | `EnumSubpSpec
        of EnumSubpSpecType.fields
    | `SubpSpec
        of SubpSpecType.fields
    | `ComponentList
        of ComponentListType.fields
    | `KnownDiscriminantPart
        of KnownDiscriminantPartType.fields
    | `UnknownDiscriminantPart
        of UnknownDiscriminantPartType.fields
    | `EntryCompletionFormalParams
        of EntryCompletionFormalParamsType.fields
    | `GenericFormalPart
        of GenericFormalPartType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `EntrySpec _
      | `EnumSubpSpec _
      | `SubpSpec _
      | `ComponentList _
      | `KnownDiscriminantPart _
      | `UnknownDiscriminantPart _
      | `EntryCompletionFormalParams _
      | `GenericFormalPart _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and BaseSubpSpecType : sig
   
  
   
  type t = [
    | `EntrySpec
        of EntrySpecType.fields
    | `EnumSubpSpec
        of EnumSubpSpecType.fields
    | `SubpSpec
        of SubpSpecType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EntrySpec
        of EntrySpecType.fields
    | `EnumSubpSpec
        of EnumSubpSpecType.fields
    | `SubpSpec
        of SubpSpecType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `EntrySpec _
      | `EnumSubpSpec _
      | `SubpSpec _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and EntrySpecType : sig
   
  
   
  type t = [
    | `EntrySpec
        of fields
  ]
  and fields = 
  {
         
    f_entry_name: DefiningNameType.t
    Lazy.t;
         
    f_family_type: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_entry_params: ParamsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EntrySpec
        of fields
  ]
  and fields = 
  {
         
    f_entry_name: DefiningNameType.t
    Lazy.t;
         
    f_family_type: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_entry_params: ParamsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_entry_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_spec_f_entry_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_family_type () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_spec_f_family_type
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AdaNodeType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `CallExpr _
      | Some `CharLiteral _
      | Some `ConstrainedSubtypeIndication _
      | Some `DiscreteSubtypeIndication _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `QualExpr _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `SubtypeIndication _
      | Some `TargetName _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    let f_entry_params () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_spec_f_entry_params
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ParamsType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EntrySpec {
        f_entry_name
          = Lazy.from_fun f_entry_name ;
        f_family_type
          = Lazy.from_fun f_family_type ;
        f_entry_params
          = Lazy.from_fun f_entry_params ;
        c_value = c_value;
        context = context
      }

end

      
and EnumSubpSpecType : sig
   
  
   
  type t = [
    | `EnumSubpSpec
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EnumSubpSpec
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EnumSubpSpec {
        c_value = c_value;
        context = context
      }

end

      
and SubpSpecType : sig
   
  
   
  type t = [
    | `SubpSpec
        of fields
  ]
  and fields = 
  {
         
    f_subp_kind: SubpKindType.t
    Lazy.t;
         
    f_subp_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_subp_params: ParamsType.t
    option
    Lazy.t;
         
    f_subp_returns: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SubpSpec
        of fields
  ]
  and fields = 
  {
         
    f_subp_kind: SubpKindType.t
    Lazy.t;
         
    f_subp_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_subp_params: ParamsType.t
    option
    Lazy.t;
         
    f_subp_returns: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_subp_kind () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subp_spec_f_subp_kind
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpKindType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subp_spec_f_subp_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_subp_params () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subp_spec_f_subp_params
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ParamsType.wrap context (field_c_value))
      in
         

      node
    in
    let f_subp_returns () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subp_spec_f_subp_returns
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (TypeExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `AnonymousType _
      | Some `ConstrainedSubtypeIndication _
      | Some `DiscreteSubtypeIndication _
      | Some `SubtypeIndication _
      | None as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SubpSpec {
        f_subp_kind
          = Lazy.from_fun f_subp_kind ;
        f_subp_name
          = Lazy.from_fun f_subp_name ;
        f_subp_params
          = Lazy.from_fun f_subp_params ;
        f_subp_returns
          = Lazy.from_fun f_subp_returns ;
        c_value = c_value;
        context = context
      }

end

      
and ComponentListType : sig
   
  
   
  type t = [
    | `ComponentList
        of fields
  ]
  and fields = 
  {
         
    f_components: AdaNodeListType.t
    option
    Lazy.t;
         
    f_variant_part: VariantPartType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ComponentList
        of fields
  ]
  and fields = 
  {
         
    f_components: AdaNodeListType.t
    option
    Lazy.t;
         
    f_variant_part: VariantPartType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_components () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.component_list_f_components
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AdaNodeListType.wrap context (field_c_value))
      in
         

      node
    in
    let f_variant_part () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.component_list_f_variant_part
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (VariantPartType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ComponentList {
        f_components
          = Lazy.from_fun f_components ;
        f_variant_part
          = Lazy.from_fun f_variant_part ;
        c_value = c_value;
        context = context
      }

end

      
and DiscriminantPartType : sig
   
  
   
  type t = [
    | `KnownDiscriminantPart
        of KnownDiscriminantPartType.fields
    | `UnknownDiscriminantPart
        of UnknownDiscriminantPartType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `KnownDiscriminantPart
        of KnownDiscriminantPartType.fields
    | `UnknownDiscriminantPart
        of UnknownDiscriminantPartType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `KnownDiscriminantPart _
      | `UnknownDiscriminantPart _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and KnownDiscriminantPartType : sig
   
  
   
  type t = [
    | `KnownDiscriminantPart
        of fields
  ]
  and fields = 
  {
         
    f_discr_specs: DiscriminantSpecListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `KnownDiscriminantPart
        of fields
  ]
  and fields = 
  {
         
    f_discr_specs: DiscriminantSpecListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_discr_specs () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.known_discriminant_part_f_discr_specs
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DiscriminantSpecListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `KnownDiscriminantPart {
        f_discr_specs
          = Lazy.from_fun f_discr_specs ;
        c_value = c_value;
        context = context
      }

end

      
and UnknownDiscriminantPartType : sig
   
  
   
  type t = [
    | `UnknownDiscriminantPart
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UnknownDiscriminantPart
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `UnknownDiscriminantPart {
        c_value = c_value;
        context = context
      }

end

      
and EntryCompletionFormalParamsType : sig
   
  
   
  type t = [
    | `EntryCompletionFormalParams
        of fields
  ]
  and fields = 
  {
         
    f_params: ParamsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EntryCompletionFormalParams
        of fields
  ]
  and fields = 
  {
         
    f_params: ParamsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_params () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_completion_formal_params_f_params
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ParamsType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EntryCompletionFormalParams {
        f_params
          = Lazy.from_fun f_params ;
        c_value = c_value;
        context = context
      }

end

      
and GenericFormalPartType : sig
   
  
   
  type t = [
    | `GenericFormalPart
        of fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericFormalPart
        of fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_formal_part_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericFormalPart {
        f_decls
          = Lazy.from_fun f_decls ;
        c_value = c_value;
        context = context
      }

end

      
and BaseRecordDefType : sig
   
  
   
  type t = [
    | `NullRecordDef
        of NullRecordDefType.fields
    | `RecordDef
        of RecordDefType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NullRecordDef
        of NullRecordDefType.fields
    | `RecordDef
        of RecordDefType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `NullRecordDef _
      | `RecordDef _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and NullRecordDefType : sig
   
  
   
  type t = [
    | `NullRecordDef
        of fields
  ]
  and fields = 
  {
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NullRecordDef
        of fields
  ]
  and fields = 
  {
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_components () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_record_def_f_components
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ComponentListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NullRecordDef {
        f_components
          = Lazy.from_fun f_components ;
        c_value = c_value;
        context = context
      }

end

      
and RecordDefType : sig
   
  
   
  type t = [
    | `RecordDef
        of fields
  ]
  and fields = 
  {
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `RecordDef
        of fields
  ]
  and fields = 
  {
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_components () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_record_def_f_components
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ComponentListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `RecordDef {
        f_components
          = Lazy.from_fun f_components ;
        c_value = c_value;
        context = context
      }

end

      
and BasicAssocType : sig
   
  
   
  (**
    * AggregateAssocType.t
    * DiscriminantAssocType.t
    * IteratedAssocType.t
    * ParamAssocType.t
    *)
  type t = [
    | `AggregateAssoc
        of AggregateAssocType.fields
    | `MultiDimArrayAssoc
        of MultiDimArrayAssocType.fields
    | `DiscriminantAssoc
        of DiscriminantAssocType.fields
    | `IteratedAssoc
        of IteratedAssocType.fields
    | `ParamAssoc
        of ParamAssocType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * AggregateAssocType.t
    * DiscriminantAssocType.t
    * IteratedAssocType.t
    * ParamAssocType.t
    *)
  type t = [
    | `AggregateAssoc
        of AggregateAssocType.fields
    | `MultiDimArrayAssoc
        of MultiDimArrayAssocType.fields
    | `DiscriminantAssoc
        of DiscriminantAssocType.fields
    | `IteratedAssoc
        of IteratedAssocType.fields
    | `ParamAssoc
        of ParamAssocType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AggregateAssoc _
      | `MultiDimArrayAssoc _
      | `DiscriminantAssoc _
      | `IteratedAssoc _
      | `ParamAssoc _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AggregateAssocType : sig
   
  
   
  type t = [
    | `AggregateAssoc
        of fields
    | `MultiDimArrayAssoc
        of MultiDimArrayAssocType.fields
  ]
  and fields = 
  {
         
    f_designators: AlternativesListType.t
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AggregateAssoc
        of fields
    | `MultiDimArrayAssoc
        of MultiDimArrayAssocType.fields
  ]
  and fields = 
  {
         
    f_designators: AlternativesListType.t
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_designators () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.aggregate_assoc_f_designators
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AlternativesListType.wrap context (field_c_value))
      in
         

      node
    in
    let f_r_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.aggregate_assoc_f_r_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BoxExpr _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AggregateAssoc {
        f_designators
          = Lazy.from_fun f_designators ;
        f_r_expr
          = Lazy.from_fun f_r_expr ;
        c_value = c_value;
        context = context
      }

end

      
and MultiDimArrayAssocType : sig
   
  
   
  type t = [
    | `MultiDimArrayAssoc
        of fields
  ]
  and fields = 
  {
         
    f_designators: AlternativesListType.t
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `MultiDimArrayAssoc
        of fields
  ]
  and fields = 
  {
         
    f_designators: AlternativesListType.t
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_designators () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.aggregate_assoc_f_designators
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AlternativesListType.wrap context (field_c_value))
      in
         

      node
    in
    let f_r_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.aggregate_assoc_f_r_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BoxExpr _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `MultiDimArrayAssoc {
        f_designators
          = Lazy.from_fun f_designators ;
        f_r_expr
          = Lazy.from_fun f_r_expr ;
        c_value = c_value;
        context = context
      }

end

      
and DiscriminantAssocType : sig
   
  
   
  type t = [
    | `DiscriminantAssoc
        of fields
  ]
  and fields = 
  {
         
    f_ids: IdentifierListType.t
    option
    Lazy.t;
         
    f_discr_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DiscriminantAssoc
        of fields
  ]
  and fields = 
  {
         
    f_ids: IdentifierListType.t
    option
    Lazy.t;
         
    f_discr_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ids () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.discriminant_assoc_f_ids
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DiscriminantChoiceListType.wrap context (field_c_value))
      in
         

      node
    in
    let f_discr_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.discriminant_assoc_f_discr_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DiscriminantAssoc {
        f_ids
          = Lazy.from_fun f_ids ;
        f_discr_expr
          = Lazy.from_fun f_discr_expr ;
        c_value = c_value;
        context = context
      }

end

      
and IteratedAssocType : sig
   
  
   
  type t = [
    | `IteratedAssoc
        of fields
  ]
  and fields = 
  {
         
    f_spec: ForLoopSpecType.t
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `IteratedAssoc
        of fields
  ]
  and fields = 
  {
         
    f_spec: ForLoopSpecType.t
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.iterated_assoc_f_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ForLoopSpecType.wrap context (field_c_value)
      in
         

      node
    in
    let f_r_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.iterated_assoc_f_r_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `IteratedAssoc {
        f_spec
          = Lazy.from_fun f_spec ;
        f_r_expr
          = Lazy.from_fun f_r_expr ;
        c_value = c_value;
        context = context
      }

end

      
and ParamAssocType : sig
   
  
   
  type t = [
    | `ParamAssoc
        of fields
  ]
  and fields = 
  {
         
    f_designator: [
      | `Identifier
          of IdentifierType.fields
      | `OthersDesignator
          of OthersDesignatorType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ParamAssoc
        of fields
  ]
  and fields = 
  {
         
    f_designator: [
      | `Identifier
          of IdentifierType.fields
      | `OthersDesignator
          of OthersDesignatorType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_designator () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.param_assoc_f_designator
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AdaNodeType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Identifier _
      | Some `OthersDesignator _
      | Some `StringLiteral _
      | None as e -> e
      | _ -> assert false
    in
    let f_r_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.param_assoc_f_r_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BoxExpr _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ParamAssoc {
        f_designator
          = Lazy.from_fun f_designator ;
        f_r_expr
          = Lazy.from_fun f_r_expr ;
        c_value = c_value;
        context = context
      }

end

      
and BasicDeclType : sig
   
  
   
  (**
    * AbstractStateDeclType.t
    * AnonymousExprDeclType.t
    * BaseFormalParamDeclType.t
    * BasePackageDeclType.t
    * BaseTypeDeclType.t
    * BasicSubpDeclType.t
    * BodyNodeType.t
    * EntryIndexSpecType.t
    * ErrorDeclType.t
    * ExceptionDeclType.t
    * ExceptionHandlerType.t
    * ForLoopVarDeclType.t
    * GenericDeclType.t
    * GenericInstantiationType.t
    * GenericRenamingDeclType.t
    * LabelDeclType.t
    * NamedStmtDeclType.t
    * NumberDeclType.t
    * ObjectDeclType.t
    * PackageRenamingDeclType.t
    * SingleProtectedDeclType.t
    * SingleTaskDeclType.t
    *)
  type t = [
    | `AbstractStateDecl
        of AbstractStateDeclType.fields
    | `AnonymousExprDecl
        of AnonymousExprDeclType.fields
    | `ComponentDecl
        of ComponentDeclType.fields
    | `DiscriminantSpec
        of DiscriminantSpecType.fields
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
    | `ParamSpec
        of ParamSpecType.fields
    | `GenericPackageInternal
        of GenericPackageInternalType.fields
    | `PackageDecl
        of PackageDeclType.fields
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
    | `ClasswideTypeDecl
        of ClasswideTypeDeclType.fields
    | `IncompleteTypeDecl
        of IncompleteTypeDeclType.fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
    | `ProtectedTypeDecl
        of ProtectedTypeDeclType.fields
    | `TaskTypeDecl
        of TaskTypeDeclType.fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
    | `TypeDecl
        of TypeDeclType.fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
    | `EntryDecl
        of EntryDeclType.fields
    | `EnumLiteralDecl
        of EnumLiteralDeclType.fields
    | `GenericSubpInternal
        of GenericSubpInternalType.fields
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
    | `EntryBody
        of EntryBodyType.fields
    | `PackageBody
        of PackageBodyType.fields
    | `ProtectedBody
        of ProtectedBodyType.fields
    | `TaskBody
        of TaskBodyType.fields
    | `EntryIndexSpec
        of EntryIndexSpecType.fields
    | `ErrorDecl
        of ErrorDeclType.fields
    | `ExceptionDecl
        of ExceptionDeclType.fields
    | `ExceptionHandler
        of ExceptionHandlerType.fields
    | `ForLoopVarDecl
        of ForLoopVarDeclType.fields
    | `GenericPackageDecl
        of GenericPackageDeclType.fields
    | `GenericSubpDecl
        of GenericSubpDeclType.fields
    | `GenericPackageInstantiation
        of GenericPackageInstantiationType.fields
    | `GenericSubpInstantiation
        of GenericSubpInstantiationType.fields
    | `GenericPackageRenamingDecl
        of GenericPackageRenamingDeclType.fields
    | `GenericSubpRenamingDecl
        of GenericSubpRenamingDeclType.fields
    | `LabelDecl
        of LabelDeclType.fields
    | `NamedStmtDecl
        of NamedStmtDeclType.fields
    | `NumberDecl
        of NumberDeclType.fields
    | `ObjectDecl
        of ObjectDeclType.fields
    | `ExtendedReturnStmtObjectDecl
        of ExtendedReturnStmtObjectDeclType.fields
    | `PackageRenamingDecl
        of PackageRenamingDeclType.fields
    | `SingleProtectedDecl
        of SingleProtectedDeclType.fields
    | `SingleTaskDecl
        of SingleTaskDeclType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * AbstractStateDeclType.t
    * AnonymousExprDeclType.t
    * BaseFormalParamDeclType.t
    * BasePackageDeclType.t
    * BaseTypeDeclType.t
    * BasicSubpDeclType.t
    * BodyNodeType.t
    * EntryIndexSpecType.t
    * ErrorDeclType.t
    * ExceptionDeclType.t
    * ExceptionHandlerType.t
    * ForLoopVarDeclType.t
    * GenericDeclType.t
    * GenericInstantiationType.t
    * GenericRenamingDeclType.t
    * LabelDeclType.t
    * NamedStmtDeclType.t
    * NumberDeclType.t
    * ObjectDeclType.t
    * PackageRenamingDeclType.t
    * SingleProtectedDeclType.t
    * SingleTaskDeclType.t
    *)
  type t = [
    | `AbstractStateDecl
        of AbstractStateDeclType.fields
    | `AnonymousExprDecl
        of AnonymousExprDeclType.fields
    | `ComponentDecl
        of ComponentDeclType.fields
    | `DiscriminantSpec
        of DiscriminantSpecType.fields
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
    | `ParamSpec
        of ParamSpecType.fields
    | `GenericPackageInternal
        of GenericPackageInternalType.fields
    | `PackageDecl
        of PackageDeclType.fields
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
    | `ClasswideTypeDecl
        of ClasswideTypeDeclType.fields
    | `IncompleteTypeDecl
        of IncompleteTypeDeclType.fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
    | `ProtectedTypeDecl
        of ProtectedTypeDeclType.fields
    | `TaskTypeDecl
        of TaskTypeDeclType.fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
    | `TypeDecl
        of TypeDeclType.fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
    | `EntryDecl
        of EntryDeclType.fields
    | `EnumLiteralDecl
        of EnumLiteralDeclType.fields
    | `GenericSubpInternal
        of GenericSubpInternalType.fields
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
    | `EntryBody
        of EntryBodyType.fields
    | `PackageBody
        of PackageBodyType.fields
    | `ProtectedBody
        of ProtectedBodyType.fields
    | `TaskBody
        of TaskBodyType.fields
    | `EntryIndexSpec
        of EntryIndexSpecType.fields
    | `ErrorDecl
        of ErrorDeclType.fields
    | `ExceptionDecl
        of ExceptionDeclType.fields
    | `ExceptionHandler
        of ExceptionHandlerType.fields
    | `ForLoopVarDecl
        of ForLoopVarDeclType.fields
    | `GenericPackageDecl
        of GenericPackageDeclType.fields
    | `GenericSubpDecl
        of GenericSubpDeclType.fields
    | `GenericPackageInstantiation
        of GenericPackageInstantiationType.fields
    | `GenericSubpInstantiation
        of GenericSubpInstantiationType.fields
    | `GenericPackageRenamingDecl
        of GenericPackageRenamingDeclType.fields
    | `GenericSubpRenamingDecl
        of GenericSubpRenamingDeclType.fields
    | `LabelDecl
        of LabelDeclType.fields
    | `NamedStmtDecl
        of NamedStmtDeclType.fields
    | `NumberDecl
        of NumberDeclType.fields
    | `ObjectDecl
        of ObjectDeclType.fields
    | `ExtendedReturnStmtObjectDecl
        of ExtendedReturnStmtObjectDeclType.fields
    | `PackageRenamingDecl
        of PackageRenamingDeclType.fields
    | `SingleProtectedDecl
        of SingleProtectedDeclType.fields
    | `SingleTaskDecl
        of SingleTaskDeclType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AbstractStateDecl _
      | `AnonymousExprDecl _
      | `ComponentDecl _
      | `DiscriminantSpec _
      | `GenericFormalObjDecl _
      | `GenericFormalPackage _
      | `GenericFormalSubpDecl _
      | `GenericFormalTypeDecl _
      | `ParamSpec _
      | `GenericPackageInternal _
      | `PackageDecl _
      | `DiscreteBaseSubtypeDecl _
      | `SubtypeDecl _
      | `ClasswideTypeDecl _
      | `IncompleteTypeDecl _
      | `IncompleteTaggedTypeDecl _
      | `ProtectedTypeDecl _
      | `TaskTypeDecl _
      | `SingleTaskTypeDecl _
      | `TypeDecl _
      | `AnonymousTypeDecl _
      | `SynthAnonymousTypeDecl _
      | `AbstractSubpDecl _
      | `AbstractFormalSubpDecl _
      | `ConcreteFormalSubpDecl _
      | `SubpDecl _
      | `EntryDecl _
      | `EnumLiteralDecl _
      | `GenericSubpInternal _
      | `ExprFunction _
      | `NullSubpDecl _
      | `SubpBody _
      | `SubpRenamingDecl _
      | `PackageBodyStub _
      | `ProtectedBodyStub _
      | `SubpBodyStub _
      | `TaskBodyStub _
      | `EntryBody _
      | `PackageBody _
      | `ProtectedBody _
      | `TaskBody _
      | `EntryIndexSpec _
      | `ErrorDecl _
      | `ExceptionDecl _
      | `ExceptionHandler _
      | `ForLoopVarDecl _
      | `GenericPackageDecl _
      | `GenericSubpDecl _
      | `GenericPackageInstantiation _
      | `GenericSubpInstantiation _
      | `GenericPackageRenamingDecl _
      | `GenericSubpRenamingDecl _
      | `LabelDecl _
      | `NamedStmtDecl _
      | `NumberDecl _
      | `ObjectDecl _
      | `ExtendedReturnStmtObjectDecl _
      | `PackageRenamingDecl _
      | `SingleProtectedDecl _
      | `SingleTaskDecl _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AbstractStateDeclType : sig
   
  
   
  type t = [
    | `AbstractStateDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbstractStateDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.abstract_state_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AbstractStateDecl {
        f_name
          = Lazy.from_fun f_name ;
        c_value = c_value;
        context = context
      }

end

      
and AnonymousExprDeclType : sig
   
  
   
  type t = [
    | `AnonymousExprDecl
        of fields
  ]
  and fields = 
  {
         
    f_expr: ExprType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AnonymousExprDecl
        of fields
  ]
  and fields = 
  {
         
    f_expr: ExprType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.anonymous_expr_decl_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AnonymousExprDecl {
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and BaseFormalParamDeclType : sig
   
  
   
  (**
    * ComponentDeclType.t
    * DiscriminantSpecType.t
    * GenericFormalType.t
    * ParamSpecType.t
    *)
  type t = [
    | `ComponentDecl
        of ComponentDeclType.fields
    | `DiscriminantSpec
        of DiscriminantSpecType.fields
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
    | `ParamSpec
        of ParamSpecType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * ComponentDeclType.t
    * DiscriminantSpecType.t
    * GenericFormalType.t
    * ParamSpecType.t
    *)
  type t = [
    | `ComponentDecl
        of ComponentDeclType.fields
    | `DiscriminantSpec
        of DiscriminantSpecType.fields
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
    | `ParamSpec
        of ParamSpecType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ComponentDecl _
      | `DiscriminantSpec _
      | `GenericFormalObjDecl _
      | `GenericFormalPackage _
      | `GenericFormalSubpDecl _
      | `GenericFormalTypeDecl _
      | `ParamSpec _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ComponentDeclType : sig
   
  
   
  type t = [
    | `ComponentDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_component_def: ComponentDefType.t
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ComponentDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_component_def: ComponentDefType.t
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ids () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.component_decl_f_ids
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_component_def () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.component_decl_f_component_def
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ComponentDefType.wrap context (field_c_value)
      in
         

      node
    in
    let f_default_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.component_decl_f_default_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ComponentDecl {
        f_ids
          = Lazy.from_fun f_ids ;
        f_component_def
          = Lazy.from_fun f_component_def ;
        f_default_expr
          = Lazy.from_fun f_default_expr ;
        c_value = c_value;
        context = context
      }

end

      
and DiscriminantSpecType : sig
   
  
   
  type t = [
    | `DiscriminantSpec
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DiscriminantSpec
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ids () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.discriminant_spec_f_ids
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_type_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.discriminant_spec_f_type_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         TypeExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AnonymousType _
      | `ConstrainedSubtypeIndication _
      | `DiscreteSubtypeIndication _
      | `SubtypeIndication _
      as e -> e
      | _ -> assert false
    in
    let f_default_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.discriminant_spec_f_default_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DiscriminantSpec {
        f_ids
          = Lazy.from_fun f_ids ;
        f_type_expr
          = Lazy.from_fun f_type_expr ;
        f_default_expr
          = Lazy.from_fun f_default_expr ;
        c_value = c_value;
        context = context
      }

end

      
and GenericFormalType : sig
   
  
   
  type t = [
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `GenericFormalObjDecl _
      | `GenericFormalPackage _
      | `GenericFormalSubpDecl _
      | `GenericFormalTypeDecl _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and GenericFormalObjDeclType : sig
   
  
   
  type t = [
    | `GenericFormalObjDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericFormalObjDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_formal_f_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BasicDeclType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AbstractFormalSubpDecl _
      | `AnonymousTypeDecl _
      | `ConcreteFormalSubpDecl _
      | `ExtendedReturnStmtObjectDecl _
      | `GenericPackageInstantiation _
      | `GenericSubpInstantiation _
      | `IncompleteTaggedTypeDecl _
      | `IncompleteTypeDecl _
      | `NumberDecl _
      | `ObjectDecl _
      | `SingleProtectedDecl _
      | `SingleTaskDecl _
      | `SynthAnonymousTypeDecl _
      | `TypeDecl _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericFormalObjDecl {
        f_decl
          = Lazy.from_fun f_decl ;
        c_value = c_value;
        context = context
      }

end

      
and GenericFormalPackageType : sig
   
  
   
  type t = [
    | `GenericFormalPackage
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericFormalPackage
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_formal_f_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BasicDeclType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AbstractFormalSubpDecl _
      | `AnonymousTypeDecl _
      | `ConcreteFormalSubpDecl _
      | `ExtendedReturnStmtObjectDecl _
      | `GenericPackageInstantiation _
      | `GenericSubpInstantiation _
      | `IncompleteTaggedTypeDecl _
      | `IncompleteTypeDecl _
      | `NumberDecl _
      | `ObjectDecl _
      | `SingleProtectedDecl _
      | `SingleTaskDecl _
      | `SynthAnonymousTypeDecl _
      | `TypeDecl _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericFormalPackage {
        f_decl
          = Lazy.from_fun f_decl ;
        c_value = c_value;
        context = context
      }

end

      
and GenericFormalSubpDeclType : sig
   
  
   
  type t = [
    | `GenericFormalSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericFormalSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_formal_f_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BasicDeclType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AbstractFormalSubpDecl _
      | `AnonymousTypeDecl _
      | `ConcreteFormalSubpDecl _
      | `ExtendedReturnStmtObjectDecl _
      | `GenericPackageInstantiation _
      | `GenericSubpInstantiation _
      | `IncompleteTaggedTypeDecl _
      | `IncompleteTypeDecl _
      | `NumberDecl _
      | `ObjectDecl _
      | `SingleProtectedDecl _
      | `SingleTaskDecl _
      | `SynthAnonymousTypeDecl _
      | `TypeDecl _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericFormalSubpDecl {
        f_decl
          = Lazy.from_fun f_decl ;
        c_value = c_value;
        context = context
      }

end

      
and GenericFormalTypeDeclType : sig
   
  
   
  type t = [
    | `GenericFormalTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericFormalTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_formal_f_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BasicDeclType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AbstractFormalSubpDecl _
      | `AnonymousTypeDecl _
      | `ConcreteFormalSubpDecl _
      | `ExtendedReturnStmtObjectDecl _
      | `GenericPackageInstantiation _
      | `GenericSubpInstantiation _
      | `IncompleteTaggedTypeDecl _
      | `IncompleteTypeDecl _
      | `NumberDecl _
      | `ObjectDecl _
      | `SingleProtectedDecl _
      | `SingleTaskDecl _
      | `SynthAnonymousTypeDecl _
      | `TypeDecl _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericFormalTypeDecl {
        f_decl
          = Lazy.from_fun f_decl ;
        c_value = c_value;
        context = context
      }

end

      
and ParamSpecType : sig
   
  
   
  type t = [
    | `ParamSpec
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ParamSpec
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ids () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.param_spec_f_ids
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_aliased () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.param_spec_f_has_aliased
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AliasedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_mode () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.param_spec_f_mode
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ModeType.wrap context (field_c_value))
      in
         

      node
    in
    let f_type_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.param_spec_f_type_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         TypeExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AnonymousType _
      | `ConstrainedSubtypeIndication _
      | `DiscreteSubtypeIndication _
      | `SubtypeIndication _
      as e -> e
      | _ -> assert false
    in
    let f_default_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.param_spec_f_default_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ParamSpec {
        f_ids
          = Lazy.from_fun f_ids ;
        f_has_aliased
          = Lazy.from_fun f_has_aliased ;
        f_mode
          = Lazy.from_fun f_mode ;
        f_type_expr
          = Lazy.from_fun f_type_expr ;
        f_default_expr
          = Lazy.from_fun f_default_expr ;
        c_value = c_value;
        context = context
      }

end

      
and BasePackageDeclType : sig
   
  
   
  type t = [
    | `GenericPackageInternal
        of GenericPackageInternalType.fields
    | `PackageDecl
        of PackageDeclType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericPackageInternal
        of GenericPackageInternalType.fields
    | `PackageDecl
        of PackageDeclType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `GenericPackageInternal _
      | `PackageDecl _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and GenericPackageInternalType : sig
   
  
   
  type t = [
    | `GenericPackageInternal
        of fields
  ]
  and fields = 
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericPackageInternal
        of fields
  ]
  and fields = 
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_package_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_package_decl_f_package_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_public_part () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_package_decl_f_public_part
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         PublicPartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_private_part () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_package_decl_f_private_part
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (PrivatePartType.wrap context (field_c_value))
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_package_decl_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericPackageInternal {
        f_package_name
          = Lazy.from_fun f_package_name ;
        f_public_part
          = Lazy.from_fun f_public_part ;
        f_private_part
          = Lazy.from_fun f_private_part ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and PackageDeclType : sig
   
  
   
  type t = [
    | `PackageDecl
        of fields
  ]
  and fields = 
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PackageDecl
        of fields
  ]
  and fields = 
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_package_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_package_decl_f_package_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_public_part () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_package_decl_f_public_part
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         PublicPartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_private_part () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_package_decl_f_private_part
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (PrivatePartType.wrap context (field_c_value))
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_package_decl_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PackageDecl {
        f_package_name
          = Lazy.from_fun f_package_name ;
        f_public_part
          = Lazy.from_fun f_public_part ;
        f_private_part
          = Lazy.from_fun f_private_part ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and BaseTypeDeclType : sig
   
  
   
  (**
    * BaseSubtypeDeclType.t
    * ClasswideTypeDeclType.t
    * IncompleteTypeDeclType.t
    * ProtectedTypeDeclType.t
    * TaskTypeDeclType.t
    * TypeDeclType.t
    *)
  type t = [
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
    | `ClasswideTypeDecl
        of ClasswideTypeDeclType.fields
    | `IncompleteTypeDecl
        of IncompleteTypeDeclType.fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
    | `ProtectedTypeDecl
        of ProtectedTypeDeclType.fields
    | `TaskTypeDecl
        of TaskTypeDeclType.fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
    | `TypeDecl
        of TypeDeclType.fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * BaseSubtypeDeclType.t
    * ClasswideTypeDeclType.t
    * IncompleteTypeDeclType.t
    * ProtectedTypeDeclType.t
    * TaskTypeDeclType.t
    * TypeDeclType.t
    *)
  type t = [
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
    | `ClasswideTypeDecl
        of ClasswideTypeDeclType.fields
    | `IncompleteTypeDecl
        of IncompleteTypeDeclType.fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
    | `ProtectedTypeDecl
        of ProtectedTypeDeclType.fields
    | `TaskTypeDecl
        of TaskTypeDeclType.fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
    | `TypeDecl
        of TypeDeclType.fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `DiscreteBaseSubtypeDecl _
      | `SubtypeDecl _
      | `ClasswideTypeDecl _
      | `IncompleteTypeDecl _
      | `IncompleteTaggedTypeDecl _
      | `ProtectedTypeDecl _
      | `TaskTypeDecl _
      | `SingleTaskTypeDecl _
      | `TypeDecl _
      | `AnonymousTypeDecl _
      | `SynthAnonymousTypeDecl _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and BaseSubtypeDeclType : sig
   
  
   
  type t = [
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `DiscreteBaseSubtypeDecl _
      | `SubtypeDecl _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and DiscreteBaseSubtypeDeclType : sig
   
  
   
  type t = [
    | `DiscreteBaseSubtypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DiscreteBaseSubtypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_type_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DiscreteBaseSubtypeDecl {
        f_name
          = Lazy.from_fun f_name ;
        c_value = c_value;
        context = context
      }

end

      
and SubtypeDeclType : sig
   
  
   
  type t = [
    | `SubtypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_subtype: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SubtypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_subtype: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_type_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_subtype () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subtype_decl_f_subtype
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubtypeIndicationType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SubtypeDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_subtype
          = Lazy.from_fun f_subtype ;
        c_value = c_value;
        context = context
      }

end

      
and ClasswideTypeDeclType : sig
   
  
   
  type t = [
    | `ClasswideTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ClasswideTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_type_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ClasswideTypeDecl {
        f_name
          = Lazy.from_fun f_name ;
        c_value = c_value;
        context = context
      }

end

      
and IncompleteTypeDeclType : sig
   
  
   
  type t = [
    | `IncompleteTypeDecl
        of fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `IncompleteTypeDecl
        of fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_type_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_discriminants () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.incomplete_type_decl_f_discriminants
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DiscriminantPartType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `IncompleteTypeDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_discriminants
          = Lazy.from_fun f_discriminants ;
        c_value = c_value;
        context = context
      }

end

      
and IncompleteTaggedTypeDeclType : sig
   
  
   
  type t = [
    | `IncompleteTaggedTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `IncompleteTaggedTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_type_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_discriminants () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.incomplete_type_decl_f_discriminants
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DiscriminantPartType.wrap context (field_c_value))
      in
         

      node
    in
    let f_has_abstract () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.incomplete_tagged_type_decl_f_has_abstract
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AbstractNodeType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `IncompleteTaggedTypeDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_discriminants
          = Lazy.from_fun f_discriminants ;
        f_has_abstract
          = Lazy.from_fun f_has_abstract ;
        c_value = c_value;
        context = context
      }

end

      
and ProtectedTypeDeclType : sig
   
  
   
  type t = [
    | `ProtectedTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_definition: ProtectedDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ProtectedTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_definition: ProtectedDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_type_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_discriminants () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.protected_type_decl_f_discriminants
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DiscriminantPartType.wrap context (field_c_value))
      in
         

      node
    in
    let f_interfaces () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.protected_type_decl_f_interfaces
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ParentListType.wrap context (field_c_value))
      in
         

      node
    in
    let f_definition () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.protected_type_decl_f_definition
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ProtectedDefType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ProtectedTypeDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_discriminants
          = Lazy.from_fun f_discriminants ;
        f_interfaces
          = Lazy.from_fun f_interfaces ;
        f_definition
          = Lazy.from_fun f_definition ;
        c_value = c_value;
        context = context
      }

end

      
and TaskTypeDeclType : sig
   
  
   
  type t = [
    | `TaskTypeDecl
        of fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_definition: TaskDefType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `TaskTypeDecl
        of fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_definition: TaskDefType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_type_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_discriminants () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_type_decl_f_discriminants
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DiscriminantPartType.wrap context (field_c_value))
      in
         

      node
    in
    let f_definition () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_type_decl_f_definition
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (TaskDefType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `TaskTypeDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_discriminants
          = Lazy.from_fun f_discriminants ;
        f_definition
          = Lazy.from_fun f_definition ;
        c_value = c_value;
        context = context
      }

end

      
and SingleTaskTypeDeclType : sig
   
  
   
  type t = [
    | `SingleTaskTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_definition: TaskDefType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SingleTaskTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_definition: TaskDefType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_type_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_discriminants () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_type_decl_f_discriminants
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DiscriminantPartType.wrap context (field_c_value))
      in
         

      node
    in
    let f_definition () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_type_decl_f_definition
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (TaskDefType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SingleTaskTypeDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_discriminants
          = Lazy.from_fun f_discriminants ;
        f_definition
          = Lazy.from_fun f_definition ;
        c_value = c_value;
        context = context
      }

end

      
and TypeDeclType : sig
   
  
   
  type t = [
    | `TypeDecl
        of fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `TypeDecl
        of fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_type_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_discriminants () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.type_decl_f_discriminants
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DiscriminantPartType.wrap context (field_c_value))
      in
         

      node
    in
    let f_type_def () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.type_decl_f_type_def
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         TypeDefType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `TypeDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_discriminants
          = Lazy.from_fun f_discriminants ;
        f_type_def
          = Lazy.from_fun f_type_def ;
        c_value = c_value;
        context = context
      }

end

      
and AnonymousTypeDeclType : sig
   
  
   
  type t = [
    | `AnonymousTypeDecl
        of fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AnonymousTypeDecl
        of fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_type_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_discriminants () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.type_decl_f_discriminants
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DiscriminantPartType.wrap context (field_c_value))
      in
         

      node
    in
    let f_type_def () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.type_decl_f_type_def
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         TypeDefType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AnonymousTypeDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_discriminants
          = Lazy.from_fun f_discriminants ;
        f_type_def
          = Lazy.from_fun f_type_def ;
        c_value = c_value;
        context = context
      }

end

      
and SynthAnonymousTypeDeclType : sig
   
  
   
  type t = [
    | `SynthAnonymousTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SynthAnonymousTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_type_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_discriminants () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.type_decl_f_discriminants
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DiscriminantPartType.wrap context (field_c_value))
      in
         

      node
    in
    let f_type_def () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.type_decl_f_type_def
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         TypeDefType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SynthAnonymousTypeDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_discriminants
          = Lazy.from_fun f_discriminants ;
        f_type_def
          = Lazy.from_fun f_type_def ;
        c_value = c_value;
        context = context
      }

end

      
and BasicSubpDeclType : sig
   
  
   
  (**
    * ClassicSubpDeclType.t
    * EntryDeclType.t
    * EnumLiteralDeclType.t
    * GenericSubpInternalType.t
    *)
  type t = [
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
    | `EntryDecl
        of EntryDeclType.fields
    | `EnumLiteralDecl
        of EnumLiteralDeclType.fields
    | `GenericSubpInternal
        of GenericSubpInternalType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * ClassicSubpDeclType.t
    * EntryDeclType.t
    * EnumLiteralDeclType.t
    * GenericSubpInternalType.t
    *)
  type t = [
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
    | `EntryDecl
        of EntryDeclType.fields
    | `EnumLiteralDecl
        of EnumLiteralDeclType.fields
    | `GenericSubpInternal
        of GenericSubpInternalType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AbstractSubpDecl _
      | `AbstractFormalSubpDecl _
      | `ConcreteFormalSubpDecl _
      | `SubpDecl _
      | `EntryDecl _
      | `EnumLiteralDecl _
      | `GenericSubpInternal _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ClassicSubpDeclType : sig
   
  
   
  (**
    * AbstractSubpDeclType.t
    * FormalSubpDeclType.t
    * SubpDeclType.t
    *)
  type t = [
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * AbstractSubpDeclType.t
    * FormalSubpDeclType.t
    * SubpDeclType.t
    *)
  type t = [
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AbstractSubpDecl _
      | `AbstractFormalSubpDecl _
      | `ConcreteFormalSubpDecl _
      | `SubpDecl _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AbstractSubpDeclType : sig
   
  
   
  type t = [
    | `AbstractSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbstractSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_overriding () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.classic_subp_decl_f_overriding
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OverridingNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.classic_subp_decl_f_subp_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpSpecType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AbstractSubpDecl {
        f_overriding
          = Lazy.from_fun f_overriding ;
        f_subp_spec
          = Lazy.from_fun f_subp_spec ;
        c_value = c_value;
        context = context
      }

end

      
and FormalSubpDeclType : sig
   
  
   
  type t = [
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AbstractFormalSubpDecl _
      | `ConcreteFormalSubpDecl _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AbstractFormalSubpDeclType : sig
   
  
   
  type t = [
    | `AbstractFormalSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_default_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbstractFormalSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_default_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_overriding () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.classic_subp_decl_f_overriding
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OverridingNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.classic_subp_decl_f_subp_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpSpecType.wrap context (field_c_value)
      in
         

      node
    in
    let f_default_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.formal_subp_decl_f_default_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `AttributeRef _
      | Some `BoxExpr _
      | Some `CallExpr _
      | Some `CharLiteral _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `NullLiteral _
      | Some `QualExpr _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AbstractFormalSubpDecl {
        f_overriding
          = Lazy.from_fun f_overriding ;
        f_subp_spec
          = Lazy.from_fun f_subp_spec ;
        f_default_expr
          = Lazy.from_fun f_default_expr ;
        c_value = c_value;
        context = context
      }

end

      
and ConcreteFormalSubpDeclType : sig
   
  
   
  type t = [
    | `ConcreteFormalSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_default_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ConcreteFormalSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_default_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_overriding () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.classic_subp_decl_f_overriding
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OverridingNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.classic_subp_decl_f_subp_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpSpecType.wrap context (field_c_value)
      in
         

      node
    in
    let f_default_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.formal_subp_decl_f_default_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `AttributeRef _
      | Some `BoxExpr _
      | Some `CallExpr _
      | Some `CharLiteral _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `NullLiteral _
      | Some `QualExpr _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ConcreteFormalSubpDecl {
        f_overriding
          = Lazy.from_fun f_overriding ;
        f_subp_spec
          = Lazy.from_fun f_subp_spec ;
        f_default_expr
          = Lazy.from_fun f_default_expr ;
        c_value = c_value;
        context = context
      }

end

      
and SubpDeclType : sig
   
  
   
  type t = [
    | `SubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_overriding () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.classic_subp_decl_f_overriding
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OverridingNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.classic_subp_decl_f_subp_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpSpecType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SubpDecl {
        f_overriding
          = Lazy.from_fun f_overriding ;
        f_subp_spec
          = Lazy.from_fun f_subp_spec ;
        c_value = c_value;
        context = context
      }

end

      
and EntryDeclType : sig
   
  
   
  type t = [
    | `EntryDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_spec: EntrySpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EntryDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_spec: EntrySpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_overriding () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_decl_f_overriding
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OverridingNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_decl_f_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         EntrySpecType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EntryDecl {
        f_overriding
          = Lazy.from_fun f_overriding ;
        f_spec
          = Lazy.from_fun f_spec ;
        c_value = c_value;
        context = context
      }

end

      
and EnumLiteralDeclType : sig
   
  
   
  type t = [
    | `EnumLiteralDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EnumLiteralDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.enum_literal_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EnumLiteralDecl {
        f_name
          = Lazy.from_fun f_name ;
        c_value = c_value;
        context = context
      }

end

      
and GenericSubpInternalType : sig
   
  
   
  type t = [
    | `GenericSubpInternal
        of fields
  ]
  and fields = 
  {
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericSubpInternal
        of fields
  ]
  and fields = 
  {
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_subp_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_subp_internal_f_subp_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpSpecType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericSubpInternal {
        f_subp_spec
          = Lazy.from_fun f_subp_spec ;
        c_value = c_value;
        context = context
      }

end

      
and BodyNodeType : sig
   
  
   
  (**
    * BaseSubpBodyType.t
    * BodyStubType.t
    * EntryBodyType.t
    * PackageBodyType.t
    * ProtectedBodyType.t
    * TaskBodyType.t
    *)
  type t = [
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
    | `EntryBody
        of EntryBodyType.fields
    | `PackageBody
        of PackageBodyType.fields
    | `ProtectedBody
        of ProtectedBodyType.fields
    | `TaskBody
        of TaskBodyType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * BaseSubpBodyType.t
    * BodyStubType.t
    * EntryBodyType.t
    * PackageBodyType.t
    * ProtectedBodyType.t
    * TaskBodyType.t
    *)
  type t = [
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
    | `EntryBody
        of EntryBodyType.fields
    | `PackageBody
        of PackageBodyType.fields
    | `ProtectedBody
        of ProtectedBodyType.fields
    | `TaskBody
        of TaskBodyType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ExprFunction _
      | `NullSubpDecl _
      | `SubpBody _
      | `SubpRenamingDecl _
      | `PackageBodyStub _
      | `ProtectedBodyStub _
      | `SubpBodyStub _
      | `TaskBodyStub _
      | `EntryBody _
      | `PackageBody _
      | `ProtectedBody _
      | `TaskBody _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and BaseSubpBodyType : sig
   
  
   
  type t = [
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ExprFunction _
      | `NullSubpDecl _
      | `SubpBody _
      | `SubpRenamingDecl _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ExprFunctionType : sig
   
  
   
  type t = [
    | `ExprFunction
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ExprFunction
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_overriding () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_subp_body_f_overriding
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OverridingNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_subp_body_f_subp_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpSpecType.wrap context (field_c_value)
      in
         

      node
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.expr_function_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `DeltaAggregate _
      | `NullRecordAggregate _
      | `ParenExpr _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ExprFunction {
        f_overriding
          = Lazy.from_fun f_overriding ;
        f_subp_spec
          = Lazy.from_fun f_subp_spec ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and NullSubpDeclType : sig
   
  
   
  type t = [
    | `NullSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NullSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_overriding () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_subp_body_f_overriding
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OverridingNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_subp_body_f_subp_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpSpecType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NullSubpDecl {
        f_overriding
          = Lazy.from_fun f_overriding ;
        f_subp_spec
          = Lazy.from_fun f_subp_spec ;
        c_value = c_value;
        context = context
      }

end

      
and SubpBodyType : sig
   
  
   
  type t = [
    | `SubpBody
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SubpBody
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_overriding () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_subp_body_f_overriding
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OverridingNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_subp_body_f_subp_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpSpecType.wrap context (field_c_value)
      in
         

      node
    in
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subp_body_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DeclarativePartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subp_body_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         HandledStmtsType.wrap context (field_c_value)
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subp_body_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SubpBody {
        f_overriding
          = Lazy.from_fun f_overriding ;
        f_subp_spec
          = Lazy.from_fun f_subp_spec ;
        f_decls
          = Lazy.from_fun f_decls ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and SubpRenamingDeclType : sig
   
  
   
  type t = [
    | `SubpRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SubpRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_overriding () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_subp_body_f_overriding
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OverridingNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_subp_body_f_subp_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpSpecType.wrap context (field_c_value)
      in
         

      node
    in
    let f_renames () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subp_renaming_decl_f_renames
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         RenamingClauseType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SubpRenamingDecl {
        f_overriding
          = Lazy.from_fun f_overriding ;
        f_subp_spec
          = Lazy.from_fun f_subp_spec ;
        f_renames
          = Lazy.from_fun f_renames ;
        c_value = c_value;
        context = context
      }

end

      
and BodyStubType : sig
   
  
   
  type t = [
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `PackageBodyStub _
      | `ProtectedBodyStub _
      | `SubpBodyStub _
      | `TaskBodyStub _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and PackageBodyStubType : sig
   
  
   
  type t = [
    | `PackageBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PackageBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.package_body_stub_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PackageBodyStub {
        f_name
          = Lazy.from_fun f_name ;
        c_value = c_value;
        context = context
      }

end

      
and ProtectedBodyStubType : sig
   
  
   
  type t = [
    | `ProtectedBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ProtectedBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.protected_body_stub_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ProtectedBodyStub {
        f_name
          = Lazy.from_fun f_name ;
        c_value = c_value;
        context = context
      }

end

      
and SubpBodyStubType : sig
   
  
   
  type t = [
    | `SubpBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SubpBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_overriding () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subp_body_stub_f_overriding
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OverridingNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subp_body_stub_f_subp_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpSpecType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SubpBodyStub {
        f_overriding
          = Lazy.from_fun f_overriding ;
        f_subp_spec
          = Lazy.from_fun f_subp_spec ;
        c_value = c_value;
        context = context
      }

end

      
and TaskBodyStubType : sig
   
  
   
  type t = [
    | `TaskBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `TaskBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_body_stub_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `TaskBodyStub {
        f_name
          = Lazy.from_fun f_name ;
        c_value = c_value;
        context = context
      }

end

      
and EntryBodyType : sig
   
  
   
  type t = [
    | `EntryBody
        of fields
  ]
  and fields = 
  {
         
    f_entry_name: DefiningNameType.t
    Lazy.t;
         
    f_index_spec: EntryIndexSpecType.t
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
         
    f_barrier: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EntryBody
        of fields
  ]
  and fields = 
  {
         
    f_entry_name: DefiningNameType.t
    Lazy.t;
         
    f_index_spec: EntryIndexSpecType.t
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
         
    f_barrier: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_entry_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_body_f_entry_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_index_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_body_f_index_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EntryIndexSpecType.wrap context (field_c_value))
      in
         

      node
    in
    let f_params () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_body_f_params
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         EntryCompletionFormalParamsType.wrap context (field_c_value)
      in
         

      node
    in
    let f_barrier () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_body_f_barrier
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_body_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DeclarativePartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_body_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         HandledStmtsType.wrap context (field_c_value)
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_body_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EntryBody {
        f_entry_name
          = Lazy.from_fun f_entry_name ;
        f_index_spec
          = Lazy.from_fun f_index_spec ;
        f_params
          = Lazy.from_fun f_params ;
        f_barrier
          = Lazy.from_fun f_barrier ;
        f_decls
          = Lazy.from_fun f_decls ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and PackageBodyType : sig
   
  
   
  type t = [
    | `PackageBody
        of fields
  ]
  and fields = 
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PackageBody
        of fields
  ]
  and fields = 
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_package_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.package_body_f_package_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.package_body_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DeclarativePartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.package_body_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (HandledStmtsType.wrap context (field_c_value))
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.package_body_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PackageBody {
        f_package_name
          = Lazy.from_fun f_package_name ;
        f_decls
          = Lazy.from_fun f_decls ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and ProtectedBodyType : sig
   
  
   
  type t = [
    | `ProtectedBody
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ProtectedBody
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.protected_body_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.protected_body_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DeclarativePartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.protected_body_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ProtectedBody {
        f_name
          = Lazy.from_fun f_name ;
        f_decls
          = Lazy.from_fun f_decls ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and TaskBodyType : sig
   
  
   
  type t = [
    | `TaskBody
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `TaskBody
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_body_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_body_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DeclarativePartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_body_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         HandledStmtsType.wrap context (field_c_value)
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_body_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `TaskBody {
        f_name
          = Lazy.from_fun f_name ;
        f_decls
          = Lazy.from_fun f_decls ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and EntryIndexSpecType : sig
   
  
   
  type t = [
    | `EntryIndexSpec
        of fields
  ]
  and fields = 
  {
         
    f_id: DefiningNameType.t
    Lazy.t;
         
    f_subtype: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EntryIndexSpec
        of fields
  ]
  and fields = 
  {
         
    f_id: DefiningNameType.t
    Lazy.t;
         
    f_subtype: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_id () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_index_spec_f_id
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subtype () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.entry_index_spec_f_subtype
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `BinOp _
      | `CallExpr _
      | `CharLiteral _
      | `ConstrainedSubtypeIndication _
      | `DiscreteSubtypeIndication _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `RelationOp _
      | `StringLiteral _
      | `SubtypeIndication _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EntryIndexSpec {
        f_id
          = Lazy.from_fun f_id ;
        f_subtype
          = Lazy.from_fun f_subtype ;
        c_value = c_value;
        context = context
      }

end

      
and ErrorDeclType : sig
   
  
   
  type t = [
    | `ErrorDecl
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ErrorDecl
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ErrorDecl {
        c_value = c_value;
        context = context
      }

end

      
and ExceptionDeclType : sig
   
  
   
  type t = [
    | `ExceptionDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ExceptionDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ids () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.exception_decl_f_ids
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_renames () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.exception_decl_f_renames
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (RenamingClauseType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ExceptionDecl {
        f_ids
          = Lazy.from_fun f_ids ;
        f_renames
          = Lazy.from_fun f_renames ;
        c_value = c_value;
        context = context
      }

end

      
and ExceptionHandlerType : sig
   
  
   
  type t = [
    | `ExceptionHandler
        of fields
  ]
  and fields = 
  {
         
    f_exception_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_handled_exceptions: AlternativesListType.t
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ExceptionHandler
        of fields
  ]
  and fields = 
  {
         
    f_exception_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_handled_exceptions: AlternativesListType.t
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_exception_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.exception_handler_f_exception_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (DefiningNameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_handled_exceptions () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.exception_handler_f_handled_exceptions
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AlternativesListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.exception_handler_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         StmtListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ExceptionHandler {
        f_exception_name
          = Lazy.from_fun f_exception_name ;
        f_handled_exceptions
          = Lazy.from_fun f_handled_exceptions ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        c_value = c_value;
        context = context
      }

end

      
and ForLoopVarDeclType : sig
   
  
   
  type t = [
    | `ForLoopVarDecl
        of fields
  ]
  and fields = 
  {
         
    f_id: DefiningNameType.t
    Lazy.t;
         
    f_id_type: SubtypeIndicationType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ForLoopVarDecl
        of fields
  ]
  and fields = 
  {
         
    f_id: DefiningNameType.t
    Lazy.t;
         
    f_id_type: SubtypeIndicationType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_id () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.for_loop_var_decl_f_id
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_id_type () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.for_loop_var_decl_f_id_type
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (SubtypeIndicationType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ForLoopVarDecl {
        f_id
          = Lazy.from_fun f_id ;
        f_id_type
          = Lazy.from_fun f_id_type ;
        c_value = c_value;
        context = context
      }

end

      
and GenericDeclType : sig
   
  
   
  type t = [
    | `GenericPackageDecl
        of GenericPackageDeclType.fields
    | `GenericSubpDecl
        of GenericSubpDeclType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericPackageDecl
        of GenericPackageDeclType.fields
    | `GenericSubpDecl
        of GenericSubpDeclType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `GenericPackageDecl _
      | `GenericSubpDecl _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and GenericPackageDeclType : sig
   
  
   
  type t = [
    | `GenericPackageDecl
        of fields
  ]
  and fields = 
  {
         
    f_formal_part: GenericFormalPartType.t
    Lazy.t;
         
    f_package_decl: GenericPackageInternalType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericPackageDecl
        of fields
  ]
  and fields = 
  {
         
    f_formal_part: GenericFormalPartType.t
    Lazy.t;
         
    f_package_decl: GenericPackageInternalType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_formal_part () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_decl_f_formal_part
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         GenericFormalPartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_package_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_package_decl_f_package_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         GenericPackageInternalType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericPackageDecl {
        f_formal_part
          = Lazy.from_fun f_formal_part ;
        f_package_decl
          = Lazy.from_fun f_package_decl ;
        c_value = c_value;
        context = context
      }

end

      
and GenericSubpDeclType : sig
   
  
   
  type t = [
    | `GenericSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_formal_part: GenericFormalPartType.t
    Lazy.t;
         
    f_subp_decl: GenericSubpInternalType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_formal_part: GenericFormalPartType.t
    Lazy.t;
         
    f_subp_decl: GenericSubpInternalType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_formal_part () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_decl_f_formal_part
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         GenericFormalPartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_subp_decl_f_subp_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         GenericSubpInternalType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericSubpDecl {
        f_formal_part
          = Lazy.from_fun f_formal_part ;
        f_subp_decl
          = Lazy.from_fun f_subp_decl ;
        c_value = c_value;
        context = context
      }

end

      
and GenericInstantiationType : sig
   
  
   
  type t = [
    | `GenericPackageInstantiation
        of GenericPackageInstantiationType.fields
    | `GenericSubpInstantiation
        of GenericSubpInstantiationType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericPackageInstantiation
        of GenericPackageInstantiationType.fields
    | `GenericSubpInstantiation
        of GenericSubpInstantiationType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `GenericPackageInstantiation _
      | `GenericSubpInstantiation _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and GenericPackageInstantiationType : sig
   
  
   
  type t = [
    | `GenericPackageInstantiation
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_generic_pkg_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_params: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericPackageInstantiation
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_generic_pkg_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_params: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_package_instantiation_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_generic_pkg_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_package_instantiation_f_generic_pkg_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    let f_params () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_package_instantiation_f_params
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AssocListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericPackageInstantiation {
        f_name
          = Lazy.from_fun f_name ;
        f_generic_pkg_name
          = Lazy.from_fun f_generic_pkg_name ;
        f_params
          = Lazy.from_fun f_params ;
        c_value = c_value;
        context = context
      }

end

      
and GenericSubpInstantiationType : sig
   
  
   
  type t = [
    | `GenericSubpInstantiation
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_kind: SubpKindType.t
    Lazy.t;
         
    f_subp_name: DefiningNameType.t
    Lazy.t;
         
    f_generic_subp_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_params: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericSubpInstantiation
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_kind: SubpKindType.t
    Lazy.t;
         
    f_subp_name: DefiningNameType.t
    Lazy.t;
         
    f_generic_subp_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_params: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_overriding () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_subp_instantiation_f_overriding
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OverridingNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_kind () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_subp_instantiation_f_kind
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpKindType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_subp_instantiation_f_subp_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_generic_subp_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_subp_instantiation_f_generic_subp_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    let f_params () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_subp_instantiation_f_params
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AssocListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericSubpInstantiation {
        f_overriding
          = Lazy.from_fun f_overriding ;
        f_kind
          = Lazy.from_fun f_kind ;
        f_subp_name
          = Lazy.from_fun f_subp_name ;
        f_generic_subp_name
          = Lazy.from_fun f_generic_subp_name ;
        f_params
          = Lazy.from_fun f_params ;
        c_value = c_value;
        context = context
      }

end

      
and GenericRenamingDeclType : sig
   
  
   
  type t = [
    | `GenericPackageRenamingDecl
        of GenericPackageRenamingDeclType.fields
    | `GenericSubpRenamingDecl
        of GenericSubpRenamingDeclType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericPackageRenamingDecl
        of GenericPackageRenamingDeclType.fields
    | `GenericSubpRenamingDecl
        of GenericSubpRenamingDeclType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `GenericPackageRenamingDecl _
      | `GenericSubpRenamingDecl _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and GenericPackageRenamingDeclType : sig
   
  
   
  type t = [
    | `GenericPackageRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericPackageRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_package_renaming_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_renames () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_package_renaming_decl_f_renames
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericPackageRenamingDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_renames
          = Lazy.from_fun f_renames ;
        c_value = c_value;
        context = context
      }

end

      
and GenericSubpRenamingDeclType : sig
   
  
   
  type t = [
    | `GenericSubpRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_kind: SubpKindType.t
    Lazy.t;
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GenericSubpRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_kind: SubpKindType.t
    Lazy.t;
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_kind () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_subp_renaming_decl_f_kind
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpKindType.wrap context (field_c_value)
      in
         

      node
    in
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_subp_renaming_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_renames () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.generic_subp_renaming_decl_f_renames
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GenericSubpRenamingDecl {
        f_kind
          = Lazy.from_fun f_kind ;
        f_name
          = Lazy.from_fun f_name ;
        f_renames
          = Lazy.from_fun f_renames ;
        c_value = c_value;
        context = context
      }

end

      
and LabelDeclType : sig
   
  
   
  type t = [
    | `LabelDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `LabelDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.label_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `LabelDecl {
        f_name
          = Lazy.from_fun f_name ;
        c_value = c_value;
        context = context
      }

end

      
and NamedStmtDeclType : sig
   
  
   
  type t = [
    | `NamedStmtDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NamedStmtDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.named_stmt_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NamedStmtDecl {
        f_name
          = Lazy.from_fun f_name ;
        c_value = c_value;
        context = context
      }

end

      
and NumberDeclType : sig
   
  
   
  type t = [
    | `NumberDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NumberDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ids () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.number_decl_f_ids
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.number_decl_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NumberDecl {
        f_ids
          = Lazy.from_fun f_ids ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and ObjectDeclType : sig
   
  
   
  type t = [
    | `ObjectDecl
        of fields
    | `ExtendedReturnStmtObjectDecl
        of ExtendedReturnStmtObjectDeclType.fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_renaming_clause: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ObjectDecl
        of fields
    | `ExtendedReturnStmtObjectDecl
        of ExtendedReturnStmtObjectDeclType.fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_renaming_clause: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ids () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_ids
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_aliased () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_has_aliased
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AliasedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_constant () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_has_constant
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ConstantNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_mode () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_mode
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ModeType.wrap context (field_c_value))
      in
         

      node
    in
    let f_type_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_type_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         TypeExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AnonymousType _
      | `ConstrainedSubtypeIndication _
      | `DiscreteSubtypeIndication _
      | `SubtypeIndication _
      as e -> e
      | _ -> assert false
    in
    let f_default_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_default_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    let f_renaming_clause () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_renaming_clause
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (RenamingClauseType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ObjectDecl {
        f_ids
          = Lazy.from_fun f_ids ;
        f_has_aliased
          = Lazy.from_fun f_has_aliased ;
        f_has_constant
          = Lazy.from_fun f_has_constant ;
        f_mode
          = Lazy.from_fun f_mode ;
        f_type_expr
          = Lazy.from_fun f_type_expr ;
        f_default_expr
          = Lazy.from_fun f_default_expr ;
        f_renaming_clause
          = Lazy.from_fun f_renaming_clause ;
        c_value = c_value;
        context = context
      }

end

      
and ExtendedReturnStmtObjectDeclType : sig
   
  
   
  type t = [
    | `ExtendedReturnStmtObjectDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_renaming_clause: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ExtendedReturnStmtObjectDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_renaming_clause: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ids () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_ids
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_aliased () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_has_aliased
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AliasedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_constant () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_has_constant
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ConstantNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_mode () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_mode
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ModeType.wrap context (field_c_value))
      in
         

      node
    in
    let f_type_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_type_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         TypeExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AnonymousType _
      | `ConstrainedSubtypeIndication _
      | `DiscreteSubtypeIndication _
      | `SubtypeIndication _
      as e -> e
      | _ -> assert false
    in
    let f_default_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_default_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    let f_renaming_clause () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.object_decl_f_renaming_clause
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (RenamingClauseType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ExtendedReturnStmtObjectDecl {
        f_ids
          = Lazy.from_fun f_ids ;
        f_has_aliased
          = Lazy.from_fun f_has_aliased ;
        f_has_constant
          = Lazy.from_fun f_has_constant ;
        f_mode
          = Lazy.from_fun f_mode ;
        f_type_expr
          = Lazy.from_fun f_type_expr ;
        f_default_expr
          = Lazy.from_fun f_default_expr ;
        f_renaming_clause
          = Lazy.from_fun f_renaming_clause ;
        c_value = c_value;
        context = context
      }

end

      
and PackageRenamingDeclType : sig
   
  
   
  type t = [
    | `PackageRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PackageRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.package_renaming_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_renames () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.package_renaming_decl_f_renames
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         RenamingClauseType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PackageRenamingDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_renames
          = Lazy.from_fun f_renames ;
        c_value = c_value;
        context = context
      }

end

      
and SingleProtectedDeclType : sig
   
  
   
  type t = [
    | `SingleProtectedDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_definition: ProtectedDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SingleProtectedDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_definition: ProtectedDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.single_protected_decl_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DefiningNameType.wrap context (field_c_value)
      in
         

      node
    in
    let f_interfaces () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.single_protected_decl_f_interfaces
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ParentListType.wrap context (field_c_value))
      in
         

      node
    in
    let f_definition () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.single_protected_decl_f_definition
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ProtectedDefType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SingleProtectedDecl {
        f_name
          = Lazy.from_fun f_name ;
        f_interfaces
          = Lazy.from_fun f_interfaces ;
        f_definition
          = Lazy.from_fun f_definition ;
        c_value = c_value;
        context = context
      }

end

      
and SingleTaskDeclType : sig
   
  
   
  type t = [
    | `SingleTaskDecl
        of fields
  ]
  and fields = 
  {
         
    f_task_type: SingleTaskTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SingleTaskDecl
        of fields
  ]
  and fields = 
  {
         
    f_task_type: SingleTaskTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_task_type () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.single_task_decl_f_task_type
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SingleTaskTypeDeclType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SingleTaskDecl {
        f_task_type
          = Lazy.from_fun f_task_type ;
        c_value = c_value;
        context = context
      }

end

      
and CaseStmtAlternativeType : sig
   
  
   
  type t = [
    | `CaseStmtAlternative
        of fields
  ]
  and fields = 
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CaseStmtAlternative
        of fields
  ]
  and fields = 
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_choices () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.case_stmt_alternative_f_choices
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AlternativesListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.case_stmt_alternative_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         StmtListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `CaseStmtAlternative {
        f_choices
          = Lazy.from_fun f_choices ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        c_value = c_value;
        context = context
      }

end

      
and CompilationUnitType : sig
   
  
   
  type t = [
    | `CompilationUnit
        of fields
  ]
  and fields = 
  {
         
    f_prelude: AdaNodeListType.t
    Lazy.t;
         
    f_body: [
      | `LibraryItem
          of LibraryItemType.fields
      | `Subunit
          of SubunitType.fields
    ]
    Lazy.t;
         
    f_pragmas: PragmaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CompilationUnit
        of fields
  ]
  and fields = 
  {
         
    f_prelude: AdaNodeListType.t
    Lazy.t;
         
    f_body: [
      | `LibraryItem
          of LibraryItemType.fields
      | `Subunit
          of SubunitType.fields
    ]
    Lazy.t;
         
    f_pragmas: PragmaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_prelude () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.compilation_unit_f_prelude
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_body () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.compilation_unit_f_body
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeType.wrap context (field_c_value)
      in
         

      match node with
            
      | `LibraryItem _
      | `Subunit _
      as e -> e
      | _ -> assert false
    in
    let f_pragmas () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.compilation_unit_f_pragmas
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         PragmaNodeListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `CompilationUnit {
        f_prelude
          = Lazy.from_fun f_prelude ;
        f_body
          = Lazy.from_fun f_body ;
        f_pragmas
          = Lazy.from_fun f_pragmas ;
        c_value = c_value;
        context = context
      }

end

      
and ComponentClauseType : sig
   
  
   
  type t = [
    | `ComponentClause
        of fields
  ]
  and fields = 
  {
         
    f_id: IdentifierType.t
    Lazy.t;
         
    f_position: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ComponentClause
        of fields
  ]
  and fields = 
  {
         
    f_id: IdentifierType.t
    Lazy.t;
         
    f_position: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_id () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.component_clause_f_id
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         IdentifierType.wrap context (field_c_value)
      in
         

      node
    in
    let f_position () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.component_clause_f_position
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_range () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.component_clause_f_range
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         RangeSpecType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ComponentClause {
        f_id
          = Lazy.from_fun f_id ;
        f_position
          = Lazy.from_fun f_position ;
        f_range
          = Lazy.from_fun f_range ;
        c_value = c_value;
        context = context
      }

end

      
and ComponentDefType : sig
   
  
   
  type t = [
    | `ComponentDef
        of fields
  ]
  and fields = 
  {
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ComponentDef
        of fields
  ]
  and fields = 
  {
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_aliased () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.component_def_f_has_aliased
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AliasedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_constant () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.component_def_f_has_constant
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ConstantNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_type_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.component_def_f_type_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         TypeExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AnonymousType _
      | `ConstrainedSubtypeIndication _
      | `DiscreteSubtypeIndication _
      | `SubtypeIndication _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ComponentDef {
        f_has_aliased
          = Lazy.from_fun f_has_aliased ;
        f_has_constant
          = Lazy.from_fun f_has_constant ;
        f_type_expr
          = Lazy.from_fun f_type_expr ;
        c_value = c_value;
        context = context
      }

end

      
and ConstantNodeType : sig
   
  
   
  type t = [
    | `ConstantAbsent
        of ConstantAbsentType.fields
    | `ConstantPresent
        of ConstantPresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ConstantAbsent
        of ConstantAbsentType.fields
    | `ConstantPresent
        of ConstantPresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ConstantAbsent _
      | `ConstantPresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ConstantAbsentType : sig
   
  
   
  type t = [
    | `ConstantAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ConstantAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ConstantAbsent {
        c_value = c_value;
        context = context
      }

end

      
and ConstantPresentType : sig
   
  
   
  type t = [
    | `ConstantPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ConstantPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ConstantPresent {
        c_value = c_value;
        context = context
      }

end

      
and ConstraintType : sig
   
  
   
  type t = [
    | `DeltaConstraint
        of DeltaConstraintType.fields
    | `DigitsConstraint
        of DigitsConstraintType.fields
    | `DiscriminantConstraint
        of DiscriminantConstraintType.fields
    | `IndexConstraint
        of IndexConstraintType.fields
    | `RangeConstraint
        of RangeConstraintType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DeltaConstraint
        of DeltaConstraintType.fields
    | `DigitsConstraint
        of DigitsConstraintType.fields
    | `DiscriminantConstraint
        of DiscriminantConstraintType.fields
    | `IndexConstraint
        of IndexConstraintType.fields
    | `RangeConstraint
        of RangeConstraintType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `DeltaConstraint _
      | `DigitsConstraint _
      | `DiscriminantConstraint _
      | `IndexConstraint _
      | `RangeConstraint _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and DeltaConstraintType : sig
   
  
   
  type t = [
    | `DeltaConstraint
        of fields
  ]
  and fields = 
  {
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DeltaConstraint
        of fields
  ]
  and fields = 
  {
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_digits () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.delta_constraint_f_digits
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_range () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.delta_constraint_f_range
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (RangeSpecType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DeltaConstraint {
        f_digits
          = Lazy.from_fun f_digits ;
        f_range
          = Lazy.from_fun f_range ;
        c_value = c_value;
        context = context
      }

end

      
and DigitsConstraintType : sig
   
  
   
  type t = [
    | `DigitsConstraint
        of fields
  ]
  and fields = 
  {
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DigitsConstraint
        of fields
  ]
  and fields = 
  {
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_digits () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.digits_constraint_f_digits
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_range () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.digits_constraint_f_range
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (RangeSpecType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DigitsConstraint {
        f_digits
          = Lazy.from_fun f_digits ;
        f_range
          = Lazy.from_fun f_range ;
        c_value = c_value;
        context = context
      }

end

      
and DiscriminantConstraintType : sig
   
  
   
  type t = [
    | `DiscriminantConstraint
        of fields
  ]
  and fields = 
  {
         
    f_constraints: BasicAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DiscriminantConstraint
        of fields
  ]
  and fields = 
  {
         
    f_constraints: BasicAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_constraints () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.discriminant_constraint_f_constraints
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AssocListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DiscriminantConstraint {
        f_constraints
          = Lazy.from_fun f_constraints ;
        c_value = c_value;
        context = context
      }

end

      
and IndexConstraintType : sig
   
  
   
  type t = [
    | `IndexConstraint
        of fields
  ]
  and fields = 
  {
         
    f_constraints: ConstraintListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `IndexConstraint
        of fields
  ]
  and fields = 
  {
         
    f_constraints: ConstraintListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_constraints () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.index_constraint_f_constraints
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ConstraintListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `IndexConstraint {
        f_constraints
          = Lazy.from_fun f_constraints ;
        c_value = c_value;
        context = context
      }

end

      
and RangeConstraintType : sig
   
  
   
  type t = [
    | `RangeConstraint
        of fields
  ]
  and fields = 
  {
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `RangeConstraint
        of fields
  ]
  and fields = 
  {
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_range () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.range_constraint_f_range
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         RangeSpecType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `RangeConstraint {
        f_range
          = Lazy.from_fun f_range ;
        c_value = c_value;
        context = context
      }

end

      
and DeclarativePartType : sig
   
  
   
  type t = [
    | `DeclarativePart
        of fields
    | `PrivatePart
        of PrivatePartType.fields
    | `PublicPart
        of PublicPartType.fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DeclarativePart
        of fields
    | `PrivatePart
        of PrivatePartType.fields
    | `PublicPart
        of PublicPartType.fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.declarative_part_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DeclarativePart {
        f_decls
          = Lazy.from_fun f_decls ;
        c_value = c_value;
        context = context
      }

end

      
and PrivatePartType : sig
   
  
   
  type t = [
    | `PrivatePart
        of fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PrivatePart
        of fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.declarative_part_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PrivatePart {
        f_decls
          = Lazy.from_fun f_decls ;
        c_value = c_value;
        context = context
      }

end

      
and PublicPartType : sig
   
  
   
  type t = [
    | `PublicPart
        of fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PublicPart
        of fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.declarative_part_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PublicPart {
        f_decls
          = Lazy.from_fun f_decls ;
        c_value = c_value;
        context = context
      }

end

      
and ElsifExprPartType : sig
   
  
   
  type t = [
    | `ElsifExprPart
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ElsifExprPart
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_cond_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.elsif_expr_part_f_cond_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_then_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.elsif_expr_part_f_then_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ElsifExprPart {
        f_cond_expr
          = Lazy.from_fun f_cond_expr ;
        f_then_expr
          = Lazy.from_fun f_then_expr ;
        c_value = c_value;
        context = context
      }

end

      
and ElsifStmtPartType : sig
   
  
   
  type t = [
    | `ElsifStmtPart
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ElsifStmtPart
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_cond_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.elsif_stmt_part_f_cond_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.elsif_stmt_part_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         StmtListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ElsifStmtPart {
        f_cond_expr
          = Lazy.from_fun f_cond_expr ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        c_value = c_value;
        context = context
      }

end

      
and ExprType : sig
   
  
   
  (**
    * AbstractStateDeclExprType.t
    * AllocatorType.t
    * BaseAggregateType.t
    * BinOpType.t
    * BoxExprType.t
    * CaseExprAlternativeType.t
    * CondExprType.t
    * ContractCasesType.t
    * DeclExprType.t
    * MembershipExprType.t
    * NameType.t
    * ParenExprType.t
    * QuantifiedExprType.t
    * RaiseExprType.t
    * UnOpType.t
    *)
  type t = [
    | `AbstractStateDeclExpr
        of AbstractStateDeclExprType.fields
    | `Allocator
        of AllocatorType.fields
    | `Aggregate
        of AggregateType.fields
    | `BracketAggregate
        of BracketAggregateType.fields
    | `DeltaAggregate
        of DeltaAggregateType.fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
    | `NullRecordAggregate
        of NullRecordAggregateType.fields
    | `BinOp
        of BinOpType.fields
    | `RelationOp
        of RelationOpType.fields
    | `BoxExpr
        of BoxExprType.fields
    | `CaseExprAlternative
        of CaseExprAlternativeType.fields
    | `CaseExpr
        of CaseExprType.fields
    | `IfExpr
        of IfExprType.fields
    | `ContractCases
        of ContractCasesType.fields
    | `DeclExpr
        of DeclExprType.fields
    | `MembershipExpr
        of MembershipExprType.fields
    | `AttributeRef
        of AttributeRefType.fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
    | `CallExpr
        of CallExprType.fields
    | `DefiningName
        of DefiningNameType.fields
    | `DiscreteSubtypeName
        of DiscreteSubtypeNameType.fields
    | `DottedName
        of DottedNameType.fields
    | `EndName
        of EndNameType.fields
    | `ExplicitDeref
        of ExplicitDerefType.fields
    | `QualExpr
        of QualExprType.fields
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
    | `TargetName
        of TargetNameType.fields
    | `ParenExpr
        of ParenExprType.fields
    | `QuantifiedExpr
        of QuantifiedExprType.fields
    | `RaiseExpr
        of RaiseExprType.fields
    | `UnOp
        of UnOpType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * AbstractStateDeclExprType.t
    * AllocatorType.t
    * BaseAggregateType.t
    * BinOpType.t
    * BoxExprType.t
    * CaseExprAlternativeType.t
    * CondExprType.t
    * ContractCasesType.t
    * DeclExprType.t
    * MembershipExprType.t
    * NameType.t
    * ParenExprType.t
    * QuantifiedExprType.t
    * RaiseExprType.t
    * UnOpType.t
    *)
  type t = [
    | `AbstractStateDeclExpr
        of AbstractStateDeclExprType.fields
    | `Allocator
        of AllocatorType.fields
    | `Aggregate
        of AggregateType.fields
    | `BracketAggregate
        of BracketAggregateType.fields
    | `DeltaAggregate
        of DeltaAggregateType.fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
    | `NullRecordAggregate
        of NullRecordAggregateType.fields
    | `BinOp
        of BinOpType.fields
    | `RelationOp
        of RelationOpType.fields
    | `BoxExpr
        of BoxExprType.fields
    | `CaseExprAlternative
        of CaseExprAlternativeType.fields
    | `CaseExpr
        of CaseExprType.fields
    | `IfExpr
        of IfExprType.fields
    | `ContractCases
        of ContractCasesType.fields
    | `DeclExpr
        of DeclExprType.fields
    | `MembershipExpr
        of MembershipExprType.fields
    | `AttributeRef
        of AttributeRefType.fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
    | `CallExpr
        of CallExprType.fields
    | `DefiningName
        of DefiningNameType.fields
    | `DiscreteSubtypeName
        of DiscreteSubtypeNameType.fields
    | `DottedName
        of DottedNameType.fields
    | `EndName
        of EndNameType.fields
    | `ExplicitDeref
        of ExplicitDerefType.fields
    | `QualExpr
        of QualExprType.fields
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
    | `TargetName
        of TargetNameType.fields
    | `ParenExpr
        of ParenExprType.fields
    | `QuantifiedExpr
        of QuantifiedExprType.fields
    | `RaiseExpr
        of RaiseExprType.fields
    | `UnOp
        of UnOpType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AbstractStateDeclExpr _
      | `Allocator _
      | `Aggregate _
      | `BracketAggregate _
      | `DeltaAggregate _
      | `BracketDeltaAggregate _
      | `NullRecordAggregate _
      | `BinOp _
      | `RelationOp _
      | `BoxExpr _
      | `CaseExprAlternative _
      | `CaseExpr _
      | `IfExpr _
      | `ContractCases _
      | `DeclExpr _
      | `MembershipExpr _
      | `AttributeRef _
      | `UpdateAttributeRef _
      | `CallExpr _
      | `DefiningName _
      | `DiscreteSubtypeName _
      | `DottedName _
      | `EndName _
      | `ExplicitDeref _
      | `QualExpr _
      | `CharLiteral _
      | `Identifier _
      | `OpAbs _
      | `OpAnd _
      | `OpAndThen _
      | `OpConcat _
      | `OpDiv _
      | `OpDoubleDot _
      | `OpEq _
      | `OpGt _
      | `OpGte _
      | `OpIn _
      | `OpLt _
      | `OpLte _
      | `OpMinus _
      | `OpMod _
      | `OpMult _
      | `OpNeq _
      | `OpNot _
      | `OpNotIn _
      | `OpOr _
      | `OpOrElse _
      | `OpPlus _
      | `OpPow _
      | `OpRem _
      | `OpXor _
      | `StringLiteral _
      | `NullLiteral _
      | `IntLiteral _
      | `RealLiteral _
      | `TargetName _
      | `ParenExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `UnOp _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AbstractStateDeclExprType : sig
   
  
   
  type t = [
    | `AbstractStateDeclExpr
        of fields
  ]
  and fields = 
  {
         
    f_state_decl: [
      | `AbstractStateDecl
          of AbstractStateDeclType.fields
      | `MultiAbstractStateDecl
          of MultiAbstractStateDeclType.fields
      | `ParenAbstractStateDecl
          of ParenAbstractStateDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbstractStateDeclExpr
        of fields
  ]
  and fields = 
  {
         
    f_state_decl: [
      | `AbstractStateDecl
          of AbstractStateDeclType.fields
      | `MultiAbstractStateDecl
          of MultiAbstractStateDeclType.fields
      | `ParenAbstractStateDecl
          of ParenAbstractStateDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_state_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.abstract_state_decl_expr_f_state_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AbstractStateDecl _
      | `MultiAbstractStateDecl _
      | `ParenAbstractStateDecl _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AbstractStateDeclExpr {
        f_state_decl
          = Lazy.from_fun f_state_decl ;
        c_value = c_value;
        context = context
      }

end

      
and AllocatorType : sig
   
  
   
  type t = [
    | `Allocator
        of fields
  ]
  and fields = 
  {
         
    f_subpool: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_type_or_expr: [
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `QualExpr
          of QualExprType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `Allocator
        of fields
  ]
  and fields = 
  {
         
    f_subpool: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_type_or_expr: [
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `QualExpr
          of QualExprType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_subpool () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.allocator_f_subpool
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (NameType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `AttributeRef _
      | Some `CallExpr _
      | Some `CharLiteral _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `QualExpr _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    let f_type_or_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.allocator_f_type_or_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeType.wrap context (field_c_value)
      in
         

      match node with
            
      | `ConstrainedSubtypeIndication _
      | `DiscreteSubtypeIndication _
      | `QualExpr _
      | `SubtypeIndication _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `Allocator {
        f_subpool
          = Lazy.from_fun f_subpool ;
        f_type_or_expr
          = Lazy.from_fun f_type_or_expr ;
        c_value = c_value;
        context = context
      }

end

      
and BaseAggregateType : sig
   
  
   
  (**
    * AggregateType.t
    * DeltaAggregateType.t
    * NullRecordAggregateType.t
    *)
  type t = [
    | `Aggregate
        of AggregateType.fields
    | `BracketAggregate
        of BracketAggregateType.fields
    | `DeltaAggregate
        of DeltaAggregateType.fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
    | `NullRecordAggregate
        of NullRecordAggregateType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * AggregateType.t
    * DeltaAggregateType.t
    * NullRecordAggregateType.t
    *)
  type t = [
    | `Aggregate
        of AggregateType.fields
    | `BracketAggregate
        of BracketAggregateType.fields
    | `DeltaAggregate
        of DeltaAggregateType.fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
    | `NullRecordAggregate
        of NullRecordAggregateType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `Aggregate _
      | `BracketAggregate _
      | `DeltaAggregate _
      | `BracketDeltaAggregate _
      | `NullRecordAggregate _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AggregateType : sig
   
  
   
  type t = [
    | `Aggregate
        of fields
    | `BracketAggregate
        of BracketAggregateType.fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `Aggregate
        of fields
    | `BracketAggregate
        of BracketAggregateType.fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ancestor_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_aggregate_f_ancestor_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      node
    in
    let f_assocs () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_aggregate_f_assocs
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AssocListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `Aggregate {
        f_ancestor_expr
          = Lazy.from_fun f_ancestor_expr ;
        f_assocs
          = Lazy.from_fun f_assocs ;
        c_value = c_value;
        context = context
      }

end

      
and BracketAggregateType : sig
   
  
   
  type t = [
    | `BracketAggregate
        of fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `BracketAggregate
        of fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ancestor_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_aggregate_f_ancestor_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      node
    in
    let f_assocs () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_aggregate_f_assocs
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AssocListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `BracketAggregate {
        f_ancestor_expr
          = Lazy.from_fun f_ancestor_expr ;
        f_assocs
          = Lazy.from_fun f_assocs ;
        c_value = c_value;
        context = context
      }

end

      
and DeltaAggregateType : sig
   
  
   
  type t = [
    | `DeltaAggregate
        of fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DeltaAggregate
        of fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ancestor_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_aggregate_f_ancestor_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      node
    in
    let f_assocs () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_aggregate_f_assocs
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AssocListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DeltaAggregate {
        f_ancestor_expr
          = Lazy.from_fun f_ancestor_expr ;
        f_assocs
          = Lazy.from_fun f_assocs ;
        c_value = c_value;
        context = context
      }

end

      
and BracketDeltaAggregateType : sig
   
  
   
  type t = [
    | `BracketDeltaAggregate
        of fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `BracketDeltaAggregate
        of fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ancestor_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_aggregate_f_ancestor_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      node
    in
    let f_assocs () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_aggregate_f_assocs
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AssocListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `BracketDeltaAggregate {
        f_ancestor_expr
          = Lazy.from_fun f_ancestor_expr ;
        f_assocs
          = Lazy.from_fun f_assocs ;
        c_value = c_value;
        context = context
      }

end

      
and NullRecordAggregateType : sig
   
  
   
  type t = [
    | `NullRecordAggregate
        of fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NullRecordAggregate
        of fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_ancestor_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_aggregate_f_ancestor_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      node
    in
    let f_assocs () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_aggregate_f_assocs
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AssocListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NullRecordAggregate {
        f_ancestor_expr
          = Lazy.from_fun f_ancestor_expr ;
        f_assocs
          = Lazy.from_fun f_assocs ;
        c_value = c_value;
        context = context
      }

end

      
and BinOpType : sig
   
  
   
  type t = [
    | `BinOp
        of fields
    | `RelationOp
        of RelationOpType.fields
  ]
  and fields = 
  {
         
    f_left: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpAnd
          of OpAndType.fields
      | `OpAndThen
          of OpAndThenType.fields
      | `OpConcat
          of OpConcatType.fields
      | `OpDiv
          of OpDivType.fields
      | `OpDoubleDot
          of OpDoubleDotType.fields
      | `OpEq
          of OpEqType.fields
      | `OpGt
          of OpGtType.fields
      | `OpGte
          of OpGteType.fields
      | `OpLt
          of OpLtType.fields
      | `OpLte
          of OpLteType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpMod
          of OpModType.fields
      | `OpMult
          of OpMultType.fields
      | `OpNeq
          of OpNeqType.fields
      | `OpOr
          of OpOrType.fields
      | `OpOrElse
          of OpOrElseType.fields
      | `OpPlus
          of OpPlusType.fields
      | `OpPow
          of OpPowType.fields
      | `OpRem
          of OpRemType.fields
      | `OpXor
          of OpXorType.fields
    ]
    Lazy.t;
         
    f_right: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `BinOp
        of fields
    | `RelationOp
        of RelationOpType.fields
  ]
  and fields = 
  {
         
    f_left: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpAnd
          of OpAndType.fields
      | `OpAndThen
          of OpAndThenType.fields
      | `OpConcat
          of OpConcatType.fields
      | `OpDiv
          of OpDivType.fields
      | `OpDoubleDot
          of OpDoubleDotType.fields
      | `OpEq
          of OpEqType.fields
      | `OpGt
          of OpGtType.fields
      | `OpGte
          of OpGteType.fields
      | `OpLt
          of OpLtType.fields
      | `OpLte
          of OpLteType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpMod
          of OpModType.fields
      | `OpMult
          of OpMultType.fields
      | `OpNeq
          of OpNeqType.fields
      | `OpOr
          of OpOrType.fields
      | `OpOrElse
          of OpOrElseType.fields
      | `OpPlus
          of OpPlusType.fields
      | `OpPow
          of OpPowType.fields
      | `OpRem
          of OpRemType.fields
      | `OpXor
          of OpXorType.fields
    ]
    Lazy.t;
         
    f_right: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_left () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.bin_op_f_left
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_op () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.bin_op_f_op
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OpType.wrap context (field_c_value)
      in
         

      match node with
            
      | `OpAnd _
      | `OpAndThen _
      | `OpConcat _
      | `OpDiv _
      | `OpDoubleDot _
      | `OpEq _
      | `OpGt _
      | `OpGte _
      | `OpLt _
      | `OpLte _
      | `OpMinus _
      | `OpMod _
      | `OpMult _
      | `OpNeq _
      | `OpOr _
      | `OpOrElse _
      | `OpPlus _
      | `OpPow _
      | `OpRem _
      | `OpXor _
      as e -> e
      | _ -> assert false
    in
    let f_right () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.bin_op_f_right
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `BinOp {
        f_left
          = Lazy.from_fun f_left ;
        f_op
          = Lazy.from_fun f_op ;
        f_right
          = Lazy.from_fun f_right ;
        c_value = c_value;
        context = context
      }

end

      
and RelationOpType : sig
   
  
   
  type t = [
    | `RelationOp
        of fields
  ]
  and fields = 
  {
         
    f_left: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpAnd
          of OpAndType.fields
      | `OpAndThen
          of OpAndThenType.fields
      | `OpConcat
          of OpConcatType.fields
      | `OpDiv
          of OpDivType.fields
      | `OpDoubleDot
          of OpDoubleDotType.fields
      | `OpEq
          of OpEqType.fields
      | `OpGt
          of OpGtType.fields
      | `OpGte
          of OpGteType.fields
      | `OpLt
          of OpLtType.fields
      | `OpLte
          of OpLteType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpMod
          of OpModType.fields
      | `OpMult
          of OpMultType.fields
      | `OpNeq
          of OpNeqType.fields
      | `OpOr
          of OpOrType.fields
      | `OpOrElse
          of OpOrElseType.fields
      | `OpPlus
          of OpPlusType.fields
      | `OpPow
          of OpPowType.fields
      | `OpRem
          of OpRemType.fields
      | `OpXor
          of OpXorType.fields
    ]
    Lazy.t;
         
    f_right: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `RelationOp
        of fields
  ]
  and fields = 
  {
         
    f_left: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpAnd
          of OpAndType.fields
      | `OpAndThen
          of OpAndThenType.fields
      | `OpConcat
          of OpConcatType.fields
      | `OpDiv
          of OpDivType.fields
      | `OpDoubleDot
          of OpDoubleDotType.fields
      | `OpEq
          of OpEqType.fields
      | `OpGt
          of OpGtType.fields
      | `OpGte
          of OpGteType.fields
      | `OpLt
          of OpLtType.fields
      | `OpLte
          of OpLteType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpMod
          of OpModType.fields
      | `OpMult
          of OpMultType.fields
      | `OpNeq
          of OpNeqType.fields
      | `OpOr
          of OpOrType.fields
      | `OpOrElse
          of OpOrElseType.fields
      | `OpPlus
          of OpPlusType.fields
      | `OpPow
          of OpPowType.fields
      | `OpRem
          of OpRemType.fields
      | `OpXor
          of OpXorType.fields
    ]
    Lazy.t;
         
    f_right: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_left () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.bin_op_f_left
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_op () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.bin_op_f_op
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OpType.wrap context (field_c_value)
      in
         

      match node with
            
      | `OpAnd _
      | `OpAndThen _
      | `OpConcat _
      | `OpDiv _
      | `OpDoubleDot _
      | `OpEq _
      | `OpGt _
      | `OpGte _
      | `OpLt _
      | `OpLte _
      | `OpMinus _
      | `OpMod _
      | `OpMult _
      | `OpNeq _
      | `OpOr _
      | `OpOrElse _
      | `OpPlus _
      | `OpPow _
      | `OpRem _
      | `OpXor _
      as e -> e
      | _ -> assert false
    in
    let f_right () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.bin_op_f_right
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `RelationOp {
        f_left
          = Lazy.from_fun f_left ;
        f_op
          = Lazy.from_fun f_op ;
        f_right
          = Lazy.from_fun f_right ;
        c_value = c_value;
        context = context
      }

end

      
and BoxExprType : sig
   
  
   
  type t = [
    | `BoxExpr
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `BoxExpr
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `BoxExpr {
        c_value = c_value;
        context = context
      }

end

      
and CaseExprAlternativeType : sig
   
  
   
  type t = [
    | `CaseExprAlternative
        of fields
  ]
  and fields = 
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CaseExprAlternative
        of fields
  ]
  and fields = 
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_choices () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.case_expr_alternative_f_choices
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AlternativesListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.case_expr_alternative_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `CaseExprAlternative {
        f_choices
          = Lazy.from_fun f_choices ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and CondExprType : sig
   
  
   
  type t = [
    | `CaseExpr
        of CaseExprType.fields
    | `IfExpr
        of IfExprType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CaseExpr
        of CaseExprType.fields
    | `IfExpr
        of IfExprType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `CaseExpr _
      | `IfExpr _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and CaseExprType : sig
   
  
   
  type t = [
    | `CaseExpr
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_cases: CaseExprAlternativeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CaseExpr
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_cases: CaseExprAlternativeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.case_expr_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_cases () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.case_expr_f_cases
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         CaseExprAlternativeListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `CaseExpr {
        f_expr
          = Lazy.from_fun f_expr ;
        f_cases
          = Lazy.from_fun f_cases ;
        c_value = c_value;
        context = context
      }

end

      
and IfExprType : sig
   
  
   
  type t = [
    | `IfExpr
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_alternatives: ElsifExprPartListType.t
    Lazy.t;
         
    f_else_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `IfExpr
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_alternatives: ElsifExprPartListType.t
    Lazy.t;
         
    f_else_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_cond_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.if_expr_f_cond_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_then_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.if_expr_f_then_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_alternatives () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.if_expr_f_alternatives
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ElsifExprPartListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_else_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.if_expr_f_else_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `IfExpr {
        f_cond_expr
          = Lazy.from_fun f_cond_expr ;
        f_then_expr
          = Lazy.from_fun f_then_expr ;
        f_alternatives
          = Lazy.from_fun f_alternatives ;
        f_else_expr
          = Lazy.from_fun f_else_expr ;
        c_value = c_value;
        context = context
      }

end

      
and ContractCasesType : sig
   
  
   
  type t = [
    | `ContractCases
        of fields
  ]
  and fields = 
  {
         
    f_contract_cases: ContractCaseAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ContractCases
        of fields
  ]
  and fields = 
  {
         
    f_contract_cases: ContractCaseAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_contract_cases () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.contract_cases_f_contract_cases
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ContractCaseAssocListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ContractCases {
        f_contract_cases
          = Lazy.from_fun f_contract_cases ;
        c_value = c_value;
        context = context
      }

end

      
and DeclExprType : sig
   
  
   
  type t = [
    | `DeclExpr
        of fields
  ]
  and fields = 
  {
         
    f_decls: BasicDeclListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DeclExpr
        of fields
  ]
  and fields = 
  {
         
    f_decls: BasicDeclListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.decl_expr_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BasicDeclListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.decl_expr_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DeclExpr {
        f_decls
          = Lazy.from_fun f_decls ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and MembershipExprType : sig
   
  
   
  type t = [
    | `MembershipExpr
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpIn
          of OpInType.fields
      | `OpNotIn
          of OpNotInType.fields
    ]
    Lazy.t;
         
    f_membership_exprs: ExprListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `MembershipExpr
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpIn
          of OpInType.fields
      | `OpNotIn
          of OpNotInType.fields
    ]
    Lazy.t;
         
    f_membership_exprs: ExprListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.membership_expr_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_op () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.membership_expr_f_op
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OpType.wrap context (field_c_value)
      in
         

      match node with
            
      | `OpIn _
      | `OpNotIn _
      as e -> e
      | _ -> assert false
    in
    let f_membership_exprs () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.membership_expr_f_membership_exprs
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprAlternativesListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `MembershipExpr {
        f_expr
          = Lazy.from_fun f_expr ;
        f_op
          = Lazy.from_fun f_op ;
        f_membership_exprs
          = Lazy.from_fun f_membership_exprs ;
        c_value = c_value;
        context = context
      }

end

      
and NameType : sig
   
  
   
  (**
    * AttributeRefType.t
    * CallExprType.t
    * DefiningNameType.t
    * DiscreteSubtypeNameType.t
    * DottedNameType.t
    * EndNameType.t
    * ExplicitDerefType.t
    * QualExprType.t
    * SingleTokNodeType.t
    * TargetNameType.t
    *)
  type t = [
    | `AttributeRef
        of AttributeRefType.fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
    | `CallExpr
        of CallExprType.fields
    | `DefiningName
        of DefiningNameType.fields
    | `DiscreteSubtypeName
        of DiscreteSubtypeNameType.fields
    | `DottedName
        of DottedNameType.fields
    | `EndName
        of EndNameType.fields
    | `ExplicitDeref
        of ExplicitDerefType.fields
    | `QualExpr
        of QualExprType.fields
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
    | `TargetName
        of TargetNameType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * AttributeRefType.t
    * CallExprType.t
    * DefiningNameType.t
    * DiscreteSubtypeNameType.t
    * DottedNameType.t
    * EndNameType.t
    * ExplicitDerefType.t
    * QualExprType.t
    * SingleTokNodeType.t
    * TargetNameType.t
    *)
  type t = [
    | `AttributeRef
        of AttributeRefType.fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
    | `CallExpr
        of CallExprType.fields
    | `DefiningName
        of DefiningNameType.fields
    | `DiscreteSubtypeName
        of DiscreteSubtypeNameType.fields
    | `DottedName
        of DottedNameType.fields
    | `EndName
        of EndNameType.fields
    | `ExplicitDeref
        of ExplicitDerefType.fields
    | `QualExpr
        of QualExprType.fields
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
    | `TargetName
        of TargetNameType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AttributeRef _
      | `UpdateAttributeRef _
      | `CallExpr _
      | `DefiningName _
      | `DiscreteSubtypeName _
      | `DottedName _
      | `EndName _
      | `ExplicitDeref _
      | `QualExpr _
      | `CharLiteral _
      | `Identifier _
      | `OpAbs _
      | `OpAnd _
      | `OpAndThen _
      | `OpConcat _
      | `OpDiv _
      | `OpDoubleDot _
      | `OpEq _
      | `OpGt _
      | `OpGte _
      | `OpIn _
      | `OpLt _
      | `OpLte _
      | `OpMinus _
      | `OpMod _
      | `OpMult _
      | `OpNeq _
      | `OpNot _
      | `OpNotIn _
      | `OpOr _
      | `OpOrElse _
      | `OpPlus _
      | `OpPow _
      | `OpRem _
      | `OpXor _
      | `StringLiteral _
      | `NullLiteral _
      | `IntLiteral _
      | `RealLiteral _
      | `TargetName _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AttributeRefType : sig
   
  
   
  type t = [
    | `AttributeRef
        of fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_attribute: IdentifierType.t
    Lazy.t;
         
    f_args: AdaNodeType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AttributeRef
        of fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_attribute: IdentifierType.t
    Lazy.t;
         
    f_args: AdaNodeType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_prefix () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.attribute_ref_f_prefix
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_attribute () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.attribute_ref_f_attribute
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         IdentifierType.wrap context (field_c_value)
      in
         

      node
    in
    let f_args () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.attribute_ref_f_args
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AdaNodeType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AttributeRef {
        f_prefix
          = Lazy.from_fun f_prefix ;
        f_attribute
          = Lazy.from_fun f_attribute ;
        f_args
          = Lazy.from_fun f_args ;
        c_value = c_value;
        context = context
      }

end

      
and UpdateAttributeRefType : sig
   
  
   
  type t = [
    | `UpdateAttributeRef
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of fields
    ]
    Lazy.t;
         
    f_attribute: IdentifierType.t
    Lazy.t;
         
    f_args: AdaNodeType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UpdateAttributeRef
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of fields
    ]
    Lazy.t;
         
    f_attribute: IdentifierType.t
    Lazy.t;
         
    f_args: AdaNodeType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_prefix () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.attribute_ref_f_prefix
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_attribute () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.attribute_ref_f_attribute
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         IdentifierType.wrap context (field_c_value)
      in
         

      node
    in
    let f_args () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.attribute_ref_f_args
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AdaNodeType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `UpdateAttributeRef {
        f_prefix
          = Lazy.from_fun f_prefix ;
        f_attribute
          = Lazy.from_fun f_attribute ;
        f_args
          = Lazy.from_fun f_args ;
        c_value = c_value;
        context = context
      }

end

      
and CallExprType : sig
   
  
   
  type t = [
    | `CallExpr
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `AssocList
          of AssocListType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CallExpr
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `AssocList
          of AssocListType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.call_expr_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_suffix () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.call_expr_f_suffix
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AssocList _
      | `AttributeRef _
      | `BinOp _
      | `CallExpr _
      | `CharLiteral _
      | `DiscreteSubtypeIndication _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `CallExpr {
        f_name
          = Lazy.from_fun f_name ;
        f_suffix
          = Lazy.from_fun f_suffix ;
        c_value = c_value;
        context = context
      }

end

      
and DefiningNameType : sig
   
  
   
  type t = [
    | `DefiningName
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DefiningName
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.defining_name_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DefiningName {
        f_name
          = Lazy.from_fun f_name ;
        c_value = c_value;
        context = context
      }

end

      
and DiscreteSubtypeNameType : sig
   
  
   
  type t = [
    | `DiscreteSubtypeName
        of fields
  ]
  and fields = 
  {
         
    f_subtype: DiscreteSubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DiscreteSubtypeName
        of fields
  ]
  and fields = 
  {
         
    f_subtype: DiscreteSubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_subtype () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.discrete_subtype_name_f_subtype
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DiscreteSubtypeIndicationType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DiscreteSubtypeName {
        f_subtype
          = Lazy.from_fun f_subtype ;
        c_value = c_value;
        context = context
      }

end

      
and DottedNameType : sig
   
  
   
  type t = [
    | `DottedName
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `CharLiteral
          of CharLiteralType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DottedName
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `CharLiteral
          of CharLiteralType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_prefix () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.dotted_name_f_prefix
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_suffix () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.dotted_name_f_suffix
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BaseIdType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DottedName {
        f_prefix
          = Lazy.from_fun f_prefix ;
        f_suffix
          = Lazy.from_fun f_suffix ;
        c_value = c_value;
        context = context
      }

end

      
and EndNameType : sig
   
  
   
  type t = [
    | `EndName
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EndName
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.end_name_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EndName {
        f_name
          = Lazy.from_fun f_name ;
        c_value = c_value;
        context = context
      }

end

      
and ExplicitDerefType : sig
   
  
   
  type t = [
    | `ExplicitDeref
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ExplicitDeref
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_prefix () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.explicit_deref_f_prefix
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ExplicitDeref {
        f_prefix
          = Lazy.from_fun f_prefix ;
        c_value = c_value;
        context = context
      }

end

      
and QualExprType : sig
   
  
   
  type t = [
    | `QualExpr
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `Aggregate
          of AggregateType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `QualExpr
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `Aggregate
          of AggregateType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_prefix () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.qual_expr_f_prefix
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_suffix () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.qual_expr_f_suffix
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `DeltaAggregate _
      | `NullRecordAggregate _
      | `ParenExpr _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `QualExpr {
        f_prefix
          = Lazy.from_fun f_prefix ;
        f_suffix
          = Lazy.from_fun f_suffix ;
        c_value = c_value;
        context = context
      }

end

      
and SingleTokNodeType : sig
   
  
   
  (**
    * BaseIdType.t
    * NullLiteralType.t
    * NumLiteralType.t
    *)
  type t = [
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * BaseIdType.t
    * NullLiteralType.t
    * NumLiteralType.t
    *)
  type t = [
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `CharLiteral _
      | `Identifier _
      | `OpAbs _
      | `OpAnd _
      | `OpAndThen _
      | `OpConcat _
      | `OpDiv _
      | `OpDoubleDot _
      | `OpEq _
      | `OpGt _
      | `OpGte _
      | `OpIn _
      | `OpLt _
      | `OpLte _
      | `OpMinus _
      | `OpMod _
      | `OpMult _
      | `OpNeq _
      | `OpNot _
      | `OpNotIn _
      | `OpOr _
      | `OpOrElse _
      | `OpPlus _
      | `OpPow _
      | `OpRem _
      | `OpXor _
      | `StringLiteral _
      | `NullLiteral _
      | `IntLiteral _
      | `RealLiteral _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and BaseIdType : sig
   
  
   
  (**
    * CharLiteralType.t
    * IdentifierType.t
    * OpType.t
    * StringLiteralType.t
    *)
  type t = [
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * CharLiteralType.t
    * IdentifierType.t
    * OpType.t
    * StringLiteralType.t
    *)
  type t = [
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `CharLiteral _
      | `Identifier _
      | `OpAbs _
      | `OpAnd _
      | `OpAndThen _
      | `OpConcat _
      | `OpDiv _
      | `OpDoubleDot _
      | `OpEq _
      | `OpGt _
      | `OpGte _
      | `OpIn _
      | `OpLt _
      | `OpLte _
      | `OpMinus _
      | `OpMod _
      | `OpMult _
      | `OpNeq _
      | `OpNot _
      | `OpNotIn _
      | `OpOr _
      | `OpOrElse _
      | `OpPlus _
      | `OpPow _
      | `OpRem _
      | `OpXor _
      | `StringLiteral _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and CharLiteralType : sig
   
  
   
  type t = [
    | `CharLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CharLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `CharLiteral {
        c_value = c_value;
        context = context
      }

end

      
and IdentifierType : sig
   
  
   
  type t = [
    | `Identifier
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `Identifier
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `Identifier {
        c_value = c_value;
        context = context
      }

end

      
and OpType : sig
   
  
   
  type t = [
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `OpAbs _
      | `OpAnd _
      | `OpAndThen _
      | `OpConcat _
      | `OpDiv _
      | `OpDoubleDot _
      | `OpEq _
      | `OpGt _
      | `OpGte _
      | `OpIn _
      | `OpLt _
      | `OpLte _
      | `OpMinus _
      | `OpMod _
      | `OpMult _
      | `OpNeq _
      | `OpNot _
      | `OpNotIn _
      | `OpOr _
      | `OpOrElse _
      | `OpPlus _
      | `OpPow _
      | `OpRem _
      | `OpXor _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and OpAbsType : sig
   
  
   
  type t = [
    | `OpAbs
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpAbs
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpAbs {
        c_value = c_value;
        context = context
      }

end

      
and OpAndType : sig
   
  
   
  type t = [
    | `OpAnd
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpAnd
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpAnd {
        c_value = c_value;
        context = context
      }

end

      
and OpAndThenType : sig
   
  
   
  type t = [
    | `OpAndThen
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpAndThen
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpAndThen {
        c_value = c_value;
        context = context
      }

end

      
and OpConcatType : sig
   
  
   
  type t = [
    | `OpConcat
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpConcat
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpConcat {
        c_value = c_value;
        context = context
      }

end

      
and OpDivType : sig
   
  
   
  type t = [
    | `OpDiv
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpDiv
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpDiv {
        c_value = c_value;
        context = context
      }

end

      
and OpDoubleDotType : sig
   
  
   
  type t = [
    | `OpDoubleDot
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpDoubleDot
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpDoubleDot {
        c_value = c_value;
        context = context
      }

end

      
and OpEqType : sig
   
  
   
  type t = [
    | `OpEq
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpEq
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpEq {
        c_value = c_value;
        context = context
      }

end

      
and OpGtType : sig
   
  
   
  type t = [
    | `OpGt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpGt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpGt {
        c_value = c_value;
        context = context
      }

end

      
and OpGteType : sig
   
  
   
  type t = [
    | `OpGte
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpGte
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpGte {
        c_value = c_value;
        context = context
      }

end

      
and OpInType : sig
   
  
   
  type t = [
    | `OpIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpIn {
        c_value = c_value;
        context = context
      }

end

      
and OpLtType : sig
   
  
   
  type t = [
    | `OpLt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpLt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpLt {
        c_value = c_value;
        context = context
      }

end

      
and OpLteType : sig
   
  
   
  type t = [
    | `OpLte
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpLte
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpLte {
        c_value = c_value;
        context = context
      }

end

      
and OpMinusType : sig
   
  
   
  type t = [
    | `OpMinus
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpMinus
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpMinus {
        c_value = c_value;
        context = context
      }

end

      
and OpModType : sig
   
  
   
  type t = [
    | `OpMod
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpMod
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpMod {
        c_value = c_value;
        context = context
      }

end

      
and OpMultType : sig
   
  
   
  type t = [
    | `OpMult
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpMult
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpMult {
        c_value = c_value;
        context = context
      }

end

      
and OpNeqType : sig
   
  
   
  type t = [
    | `OpNeq
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpNeq
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpNeq {
        c_value = c_value;
        context = context
      }

end

      
and OpNotType : sig
   
  
   
  type t = [
    | `OpNot
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpNot
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpNot {
        c_value = c_value;
        context = context
      }

end

      
and OpNotInType : sig
   
  
   
  type t = [
    | `OpNotIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpNotIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpNotIn {
        c_value = c_value;
        context = context
      }

end

      
and OpOrType : sig
   
  
   
  type t = [
    | `OpOr
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpOr
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpOr {
        c_value = c_value;
        context = context
      }

end

      
and OpOrElseType : sig
   
  
   
  type t = [
    | `OpOrElse
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpOrElse
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpOrElse {
        c_value = c_value;
        context = context
      }

end

      
and OpPlusType : sig
   
  
   
  type t = [
    | `OpPlus
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpPlus
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpPlus {
        c_value = c_value;
        context = context
      }

end

      
and OpPowType : sig
   
  
   
  type t = [
    | `OpPow
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpPow
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpPow {
        c_value = c_value;
        context = context
      }

end

      
and OpRemType : sig
   
  
   
  type t = [
    | `OpRem
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpRem
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpRem {
        c_value = c_value;
        context = context
      }

end

      
and OpXorType : sig
   
  
   
  type t = [
    | `OpXor
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OpXor
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OpXor {
        c_value = c_value;
        context = context
      }

end

      
and StringLiteralType : sig
   
  
   
  type t = [
    | `StringLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `StringLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `StringLiteral {
        c_value = c_value;
        context = context
      }

end

      
and NullLiteralType : sig
   
  
   
  type t = [
    | `NullLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NullLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NullLiteral {
        c_value = c_value;
        context = context
      }

end

      
and NumLiteralType : sig
   
  
   
  type t = [
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `IntLiteral _
      | `RealLiteral _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and IntLiteralType : sig
   
  
   
  type t = [
    | `IntLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `IntLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `IntLiteral {
        c_value = c_value;
        context = context
      }

end

      
and RealLiteralType : sig
   
  
   
  type t = [
    | `RealLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `RealLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `RealLiteral {
        c_value = c_value;
        context = context
      }

end

      
and TargetNameType : sig
   
  
   
  type t = [
    | `TargetName
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `TargetName
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `TargetName {
        c_value = c_value;
        context = context
      }

end

      
and ParenExprType : sig
   
  
   
  type t = [
    | `ParenExpr
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ParenExpr
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.paren_expr_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ParenExpr {
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and QuantifiedExprType : sig
   
  
   
  type t = [
    | `QuantifiedExpr
        of fields
  ]
  and fields = 
  {
         
    f_quantifier: QuantifierType.t
    Lazy.t;
         
    f_loop_spec: ForLoopSpecType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `QuantifiedExpr
        of fields
  ]
  and fields = 
  {
         
    f_quantifier: QuantifierType.t
    Lazy.t;
         
    f_loop_spec: ForLoopSpecType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_quantifier () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.quantified_expr_f_quantifier
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         QuantifierType.wrap context (field_c_value)
      in
         

      node
    in
    let f_loop_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.quantified_expr_f_loop_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ForLoopSpecType.wrap context (field_c_value)
      in
         

      node
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.quantified_expr_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `QuantifiedExpr {
        f_quantifier
          = Lazy.from_fun f_quantifier ;
        f_loop_spec
          = Lazy.from_fun f_loop_spec ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and RaiseExprType : sig
   
  
   
  type t = [
    | `RaiseExpr
        of fields
  ]
  and fields = 
  {
         
    f_exception_name: NameType.t
    option
    Lazy.t;
         
    f_error_message: ExprType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `RaiseExpr
        of fields
  ]
  and fields = 
  {
         
    f_exception_name: NameType.t
    option
    Lazy.t;
         
    f_error_message: ExprType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_exception_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.raise_expr_f_exception_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (NameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_error_message () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.raise_expr_f_error_message
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `RaiseExpr {
        f_exception_name
          = Lazy.from_fun f_exception_name ;
        f_error_message
          = Lazy.from_fun f_error_message ;
        c_value = c_value;
        context = context
      }

end

      
and UnOpType : sig
   
  
   
  type t = [
    | `UnOp
        of fields
  ]
  and fields = 
  {
         
    f_op: [
      | `OpAbs
          of OpAbsType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpNot
          of OpNotType.fields
      | `OpPlus
          of OpPlusType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UnOp
        of fields
  ]
  and fields = 
  {
         
    f_op: [
      | `OpAbs
          of OpAbsType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpNot
          of OpNotType.fields
      | `OpPlus
          of OpPlusType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_op () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.un_op_f_op
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         OpType.wrap context (field_c_value)
      in
         

      match node with
            
      | `OpAbs _
      | `OpMinus _
      | `OpNot _
      | `OpPlus _
      as e -> e
      | _ -> assert false
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.un_op_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `UnOp {
        f_op
          = Lazy.from_fun f_op ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and HandledStmtsType : sig
   
  
   
  type t = [
    | `HandledStmts
        of fields
  ]
  and fields = 
  {
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_exceptions: AdaNodeListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `HandledStmts
        of fields
  ]
  and fields = 
  {
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_exceptions: AdaNodeListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.handled_stmts_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         StmtListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_exceptions () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.handled_stmts_f_exceptions
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (AdaNodeListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `HandledStmts {
        f_stmts
          = Lazy.from_fun f_stmts ;
        f_exceptions
          = Lazy.from_fun f_exceptions ;
        c_value = c_value;
        context = context
      }

end

      
and InterfaceKindType : sig
   
  
   
  type t = [
    | `InterfaceKindLimited
        of InterfaceKindLimitedType.fields
    | `InterfaceKindProtected
        of InterfaceKindProtectedType.fields
    | `InterfaceKindSynchronized
        of InterfaceKindSynchronizedType.fields
    | `InterfaceKindTask
        of InterfaceKindTaskType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `InterfaceKindLimited
        of InterfaceKindLimitedType.fields
    | `InterfaceKindProtected
        of InterfaceKindProtectedType.fields
    | `InterfaceKindSynchronized
        of InterfaceKindSynchronizedType.fields
    | `InterfaceKindTask
        of InterfaceKindTaskType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `InterfaceKindLimited _
      | `InterfaceKindProtected _
      | `InterfaceKindSynchronized _
      | `InterfaceKindTask _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and InterfaceKindLimitedType : sig
   
  
   
  type t = [
    | `InterfaceKindLimited
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `InterfaceKindLimited
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `InterfaceKindLimited {
        c_value = c_value;
        context = context
      }

end

      
and InterfaceKindProtectedType : sig
   
  
   
  type t = [
    | `InterfaceKindProtected
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `InterfaceKindProtected
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `InterfaceKindProtected {
        c_value = c_value;
        context = context
      }

end

      
and InterfaceKindSynchronizedType : sig
   
  
   
  type t = [
    | `InterfaceKindSynchronized
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `InterfaceKindSynchronized
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `InterfaceKindSynchronized {
        c_value = c_value;
        context = context
      }

end

      
and InterfaceKindTaskType : sig
   
  
   
  type t = [
    | `InterfaceKindTask
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `InterfaceKindTask
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `InterfaceKindTask {
        c_value = c_value;
        context = context
      }

end

      
and IterTypeType : sig
   
  
   
  type t = [
    | `IterTypeIn
        of IterTypeInType.fields
    | `IterTypeOf
        of IterTypeOfType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `IterTypeIn
        of IterTypeInType.fields
    | `IterTypeOf
        of IterTypeOfType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `IterTypeIn _
      | `IterTypeOf _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and IterTypeInType : sig
   
  
   
  type t = [
    | `IterTypeIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `IterTypeIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `IterTypeIn {
        c_value = c_value;
        context = context
      }

end

      
and IterTypeOfType : sig
   
  
   
  type t = [
    | `IterTypeOf
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `IterTypeOf
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `IterTypeOf {
        c_value = c_value;
        context = context
      }

end

      
and LibraryItemType : sig
   
  
   
  type t = [
    | `LibraryItem
        of fields
  ]
  and fields = 
  {
         
    f_has_private: PrivateNodeType.t
    Lazy.t;
         
    f_item: [
      | `AbstractSubpDecl
          of AbstractSubpDeclType.fields
      | `ErrorDecl
          of ErrorDeclType.fields
      | `ExprFunction
          of ExprFunctionType.fields
      | `GenericPackageDecl
          of GenericPackageDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericPackageRenamingDecl
          of GenericPackageRenamingDeclType.fields
      | `GenericSubpDecl
          of GenericSubpDeclType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `GenericSubpRenamingDecl
          of GenericSubpRenamingDeclType.fields
      | `NullSubpDecl
          of NullSubpDeclType.fields
      | `PackageBody
          of PackageBodyType.fields
      | `PackageDecl
          of PackageDeclType.fields
      | `PackageRenamingDecl
          of PackageRenamingDeclType.fields
      | `SubpBody
          of SubpBodyType.fields
      | `SubpDecl
          of SubpDeclType.fields
      | `SubpRenamingDecl
          of SubpRenamingDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `LibraryItem
        of fields
  ]
  and fields = 
  {
         
    f_has_private: PrivateNodeType.t
    Lazy.t;
         
    f_item: [
      | `AbstractSubpDecl
          of AbstractSubpDeclType.fields
      | `ErrorDecl
          of ErrorDeclType.fields
      | `ExprFunction
          of ExprFunctionType.fields
      | `GenericPackageDecl
          of GenericPackageDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericPackageRenamingDecl
          of GenericPackageRenamingDeclType.fields
      | `GenericSubpDecl
          of GenericSubpDeclType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `GenericSubpRenamingDecl
          of GenericSubpRenamingDeclType.fields
      | `NullSubpDecl
          of NullSubpDeclType.fields
      | `PackageBody
          of PackageBodyType.fields
      | `PackageDecl
          of PackageDeclType.fields
      | `PackageRenamingDecl
          of PackageRenamingDeclType.fields
      | `SubpBody
          of SubpBodyType.fields
      | `SubpDecl
          of SubpDeclType.fields
      | `SubpRenamingDecl
          of SubpRenamingDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_private () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.library_item_f_has_private
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         PrivateNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_item () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.library_item_f_item
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BasicDeclType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AbstractSubpDecl _
      | `ErrorDecl _
      | `ExprFunction _
      | `GenericPackageDecl _
      | `GenericPackageInstantiation _
      | `GenericPackageRenamingDecl _
      | `GenericSubpDecl _
      | `GenericSubpInstantiation _
      | `GenericSubpRenamingDecl _
      | `NullSubpDecl _
      | `PackageBody _
      | `PackageDecl _
      | `PackageRenamingDecl _
      | `SubpBody _
      | `SubpDecl _
      | `SubpRenamingDecl _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `LibraryItem {
        f_has_private
          = Lazy.from_fun f_has_private ;
        f_item
          = Lazy.from_fun f_item ;
        c_value = c_value;
        context = context
      }

end

      
and LimitedNodeType : sig
   
  
   
  type t = [
    | `LimitedAbsent
        of LimitedAbsentType.fields
    | `LimitedPresent
        of LimitedPresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `LimitedAbsent
        of LimitedAbsentType.fields
    | `LimitedPresent
        of LimitedPresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `LimitedAbsent _
      | `LimitedPresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and LimitedAbsentType : sig
   
  
   
  type t = [
    | `LimitedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `LimitedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `LimitedAbsent {
        c_value = c_value;
        context = context
      }

end

      
and LimitedPresentType : sig
   
  
   
  type t = [
    | `LimitedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `LimitedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `LimitedPresent {
        c_value = c_value;
        context = context
      }

end

      
and LoopSpecType : sig
   
  
   
  type t = [
    | `ForLoopSpec
        of ForLoopSpecType.fields
    | `WhileLoopSpec
        of WhileLoopSpecType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ForLoopSpec
        of ForLoopSpecType.fields
    | `WhileLoopSpec
        of WhileLoopSpecType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ForLoopSpec _
      | `WhileLoopSpec _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ForLoopSpecType : sig
   
  
   
  type t = [
    | `ForLoopSpec
        of fields
  ]
  and fields = 
  {
         
    f_var_decl: ForLoopVarDeclType.t
    Lazy.t;
         
    f_loop_type: IterTypeType.t
    Lazy.t;
         
    f_has_reverse: ReverseNodeType.t
    Lazy.t;
         
    f_iter_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ForLoopSpec
        of fields
  ]
  and fields = 
  {
         
    f_var_decl: ForLoopVarDeclType.t
    Lazy.t;
         
    f_loop_type: IterTypeType.t
    Lazy.t;
         
    f_has_reverse: ReverseNodeType.t
    Lazy.t;
         
    f_iter_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_var_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.for_loop_spec_f_var_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ForLoopVarDeclType.wrap context (field_c_value)
      in
         

      node
    in
    let f_loop_type () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.for_loop_spec_f_loop_type
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         IterTypeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_reverse () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.for_loop_spec_f_has_reverse
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ReverseNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_iter_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.for_loop_spec_f_iter_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `BinOp _
      | `CallExpr _
      | `CharLiteral _
      | `DiscreteSubtypeIndication _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ForLoopSpec {
        f_var_decl
          = Lazy.from_fun f_var_decl ;
        f_loop_type
          = Lazy.from_fun f_loop_type ;
        f_has_reverse
          = Lazy.from_fun f_has_reverse ;
        f_iter_expr
          = Lazy.from_fun f_iter_expr ;
        c_value = c_value;
        context = context
      }

end

      
and WhileLoopSpecType : sig
   
  
   
  type t = [
    | `WhileLoopSpec
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `WhileLoopSpec
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.while_loop_spec_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `WhileLoopSpec {
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and ModeType : sig
   
  
   
  type t = [
    | `ModeDefault
        of ModeDefaultType.fields
    | `ModeIn
        of ModeInType.fields
    | `ModeInOut
        of ModeInOutType.fields
    | `ModeOut
        of ModeOutType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ModeDefault
        of ModeDefaultType.fields
    | `ModeIn
        of ModeInType.fields
    | `ModeInOut
        of ModeInOutType.fields
    | `ModeOut
        of ModeOutType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ModeDefault _
      | `ModeIn _
      | `ModeInOut _
      | `ModeOut _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ModeDefaultType : sig
   
  
   
  type t = [
    | `ModeDefault
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ModeDefault
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ModeDefault {
        c_value = c_value;
        context = context
      }

end

      
and ModeInType : sig
   
  
   
  type t = [
    | `ModeIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ModeIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ModeIn {
        c_value = c_value;
        context = context
      }

end

      
and ModeInOutType : sig
   
  
   
  type t = [
    | `ModeInOut
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ModeInOut
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ModeInOut {
        c_value = c_value;
        context = context
      }

end

      
and ModeOutType : sig
   
  
   
  type t = [
    | `ModeOut
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ModeOut
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ModeOut {
        c_value = c_value;
        context = context
      }

end

      
and MultiAbstractStateDeclType : sig
   
  
   
  type t = [
    | `MultiAbstractStateDecl
        of fields
  ]
  and fields = 
  {
         
    f_decls: AbstractStateDeclListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `MultiAbstractStateDecl
        of fields
  ]
  and fields = 
  {
         
    f_decls: AbstractStateDeclListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.multi_abstract_state_decl_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AbstractStateDeclListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `MultiAbstractStateDecl {
        f_decls
          = Lazy.from_fun f_decls ;
        c_value = c_value;
        context = context
      }

end

      
and NotNullType : sig
   
  
   
  type t = [
    | `NotNullAbsent
        of NotNullAbsentType.fields
    | `NotNullPresent
        of NotNullPresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NotNullAbsent
        of NotNullAbsentType.fields
    | `NotNullPresent
        of NotNullPresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `NotNullAbsent _
      | `NotNullPresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and NotNullAbsentType : sig
   
  
   
  type t = [
    | `NotNullAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NotNullAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NotNullAbsent {
        c_value = c_value;
        context = context
      }

end

      
and NotNullPresentType : sig
   
  
   
  type t = [
    | `NotNullPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NotNullPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NotNullPresent {
        c_value = c_value;
        context = context
      }

end

      
and NullComponentDeclType : sig
   
  
   
  type t = [
    | `NullComponentDecl
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NullComponentDecl
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NullComponentDecl {
        c_value = c_value;
        context = context
      }

end

      
and OthersDesignatorType : sig
   
  
   
  type t = [
    | `OthersDesignator
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OthersDesignator
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OthersDesignator {
        c_value = c_value;
        context = context
      }

end

      
and OverridingNodeType : sig
   
  
   
  type t = [
    | `OverridingNotOverriding
        of OverridingNotOverridingType.fields
    | `OverridingOverriding
        of OverridingOverridingType.fields
    | `OverridingUnspecified
        of OverridingUnspecifiedType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OverridingNotOverriding
        of OverridingNotOverridingType.fields
    | `OverridingOverriding
        of OverridingOverridingType.fields
    | `OverridingUnspecified
        of OverridingUnspecifiedType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `OverridingNotOverriding _
      | `OverridingOverriding _
      | `OverridingUnspecified _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and OverridingNotOverridingType : sig
   
  
   
  type t = [
    | `OverridingNotOverriding
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OverridingNotOverriding
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OverridingNotOverriding {
        c_value = c_value;
        context = context
      }

end

      
and OverridingOverridingType : sig
   
  
   
  type t = [
    | `OverridingOverriding
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OverridingOverriding
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OverridingOverriding {
        c_value = c_value;
        context = context
      }

end

      
and OverridingUnspecifiedType : sig
   
  
   
  type t = [
    | `OverridingUnspecified
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OverridingUnspecified
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OverridingUnspecified {
        c_value = c_value;
        context = context
      }

end

      
and ParamsType : sig
   
  
   
  type t = [
    | `Params
        of fields
  ]
  and fields = 
  {
         
    f_params: ParamSpecListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `Params
        of fields
  ]
  and fields = 
  {
         
    f_params: ParamSpecListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_params () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.params_f_params
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ParamSpecListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `Params {
        f_params
          = Lazy.from_fun f_params ;
        c_value = c_value;
        context = context
      }

end

      
and ParenAbstractStateDeclType : sig
   
  
   
  type t = [
    | `ParenAbstractStateDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractStateDecl
          of AbstractStateDeclType.fields
      | `ParenAbstractStateDecl
          of fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ParenAbstractStateDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractStateDecl
          of AbstractStateDeclType.fields
      | `ParenAbstractStateDecl
          of fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.paren_abstract_state_decl_f_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AdaNodeType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AbstractStateDecl _
      | `ParenAbstractStateDecl _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ParenAbstractStateDecl {
        f_decl
          = Lazy.from_fun f_decl ;
        c_value = c_value;
        context = context
      }

end

      
and PragmaNodeType : sig
   
  
   
  type t = [
    | `PragmaNode
        of fields
  ]
  and fields = 
  {
         
    f_id: IdentifierType.t
    Lazy.t;
         
    f_args: BaseAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PragmaNode
        of fields
  ]
  and fields = 
  {
         
    f_id: IdentifierType.t
    Lazy.t;
         
    f_args: BaseAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_id () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.pragma_node_f_id
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         IdentifierType.wrap context (field_c_value)
      in
         

      node
    in
    let f_args () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.pragma_node_f_args
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (BaseAssocListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PragmaNode {
        f_id
          = Lazy.from_fun f_id ;
        f_args
          = Lazy.from_fun f_args ;
        c_value = c_value;
        context = context
      }

end

      
and PrivateNodeType : sig
   
  
   
  type t = [
    | `PrivateAbsent
        of PrivateAbsentType.fields
    | `PrivatePresent
        of PrivatePresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PrivateAbsent
        of PrivateAbsentType.fields
    | `PrivatePresent
        of PrivatePresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `PrivateAbsent _
      | `PrivatePresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and PrivateAbsentType : sig
   
  
   
  type t = [
    | `PrivateAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PrivateAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PrivateAbsent {
        c_value = c_value;
        context = context
      }

end

      
and PrivatePresentType : sig
   
  
   
  type t = [
    | `PrivatePresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PrivatePresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PrivatePresent {
        c_value = c_value;
        context = context
      }

end

      
and ProtectedDefType : sig
   
  
   
  type t = [
    | `ProtectedDef
        of fields
  ]
  and fields = 
  {
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ProtectedDef
        of fields
  ]
  and fields = 
  {
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_public_part () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.protected_def_f_public_part
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         PublicPartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_private_part () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.protected_def_f_private_part
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (PrivatePartType.wrap context (field_c_value))
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.protected_def_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ProtectedDef {
        f_public_part
          = Lazy.from_fun f_public_part ;
        f_private_part
          = Lazy.from_fun f_private_part ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and ProtectedNodeType : sig
   
  
   
  type t = [
    | `ProtectedAbsent
        of ProtectedAbsentType.fields
    | `ProtectedPresent
        of ProtectedPresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ProtectedAbsent
        of ProtectedAbsentType.fields
    | `ProtectedPresent
        of ProtectedPresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ProtectedAbsent _
      | `ProtectedPresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ProtectedAbsentType : sig
   
  
   
  type t = [
    | `ProtectedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ProtectedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ProtectedAbsent {
        c_value = c_value;
        context = context
      }

end

      
and ProtectedPresentType : sig
   
  
   
  type t = [
    | `ProtectedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ProtectedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ProtectedPresent {
        c_value = c_value;
        context = context
      }

end

      
and QuantifierType : sig
   
  
   
  type t = [
    | `QuantifierAll
        of QuantifierAllType.fields
    | `QuantifierSome
        of QuantifierSomeType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `QuantifierAll
        of QuantifierAllType.fields
    | `QuantifierSome
        of QuantifierSomeType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `QuantifierAll _
      | `QuantifierSome _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and QuantifierAllType : sig
   
  
   
  type t = [
    | `QuantifierAll
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `QuantifierAll
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `QuantifierAll {
        c_value = c_value;
        context = context
      }

end

      
and QuantifierSomeType : sig
   
  
   
  type t = [
    | `QuantifierSome
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `QuantifierSome
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `QuantifierSome {
        c_value = c_value;
        context = context
      }

end

      
and RangeSpecType : sig
   
  
   
  type t = [
    | `RangeSpec
        of fields
  ]
  and fields = 
  {
         
    f_range: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `RangeSpec
        of fields
  ]
  and fields = 
  {
         
    f_range: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_range () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.range_spec_f_range
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `BinOp _
      | `BoxExpr _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `RangeSpec {
        f_range
          = Lazy.from_fun f_range ;
        c_value = c_value;
        context = context
      }

end

      
and RenamingClauseType : sig
   
  
   
  type t = [
    | `RenamingClause
        of fields
    | `SyntheticRenamingClause
        of SyntheticRenamingClauseType.fields
  ]
  and fields = 
  {
         
    f_renamed_object: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `RenamingClause
        of fields
    | `SyntheticRenamingClause
        of SyntheticRenamingClauseType.fields
  ]
  and fields = 
  {
         
    f_renamed_object: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_renamed_object () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.renaming_clause_f_renamed_object
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `RenamingClause {
        f_renamed_object
          = Lazy.from_fun f_renamed_object ;
        c_value = c_value;
        context = context
      }

end

      
and SyntheticRenamingClauseType : sig
   
  
   
  type t = [
    | `SyntheticRenamingClause
        of fields
  ]
  and fields = 
  {
         
    f_renamed_object: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SyntheticRenamingClause
        of fields
  ]
  and fields = 
  {
         
    f_renamed_object: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_renamed_object () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.renaming_clause_f_renamed_object
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SyntheticRenamingClause {
        f_renamed_object
          = Lazy.from_fun f_renamed_object ;
        c_value = c_value;
        context = context
      }

end

      
and ReverseNodeType : sig
   
  
   
  type t = [
    | `ReverseAbsent
        of ReverseAbsentType.fields
    | `ReversePresent
        of ReversePresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ReverseAbsent
        of ReverseAbsentType.fields
    | `ReversePresent
        of ReversePresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ReverseAbsent _
      | `ReversePresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ReverseAbsentType : sig
   
  
   
  type t = [
    | `ReverseAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ReverseAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ReverseAbsent {
        c_value = c_value;
        context = context
      }

end

      
and ReversePresentType : sig
   
  
   
  type t = [
    | `ReversePresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ReversePresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ReversePresent {
        c_value = c_value;
        context = context
      }

end

      
and SelectWhenPartType : sig
   
  
   
  type t = [
    | `SelectWhenPart
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SelectWhenPart
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_cond_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.select_when_part_f_cond_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.select_when_part_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         StmtListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SelectWhenPart {
        f_cond_expr
          = Lazy.from_fun f_cond_expr ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        c_value = c_value;
        context = context
      }

end

      
and StmtType : sig
   
  
   
  (**
    * CompositeStmtType.t
    * ErrorStmtType.t
    * SimpleStmtType.t
    *)
  type t = [
    | `AcceptStmt
        of AcceptStmtType.fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
    | `CaseStmt
        of CaseStmtType.fields
    | `ExtendedReturnStmt
        of ExtendedReturnStmtType.fields
    | `IfStmt
        of IfStmtType.fields
    | `NamedStmt
        of NamedStmtType.fields
    | `SelectStmt
        of SelectStmtType.fields
    | `ErrorStmt
        of ErrorStmtType.fields
    | `AbortStmt
        of AbortStmtType.fields
    | `AssignStmt
        of AssignStmtType.fields
    | `CallStmt
        of CallStmtType.fields
    | `DelayStmt
        of DelayStmtType.fields
    | `ExitStmt
        of ExitStmtType.fields
    | `GotoStmt
        of GotoStmtType.fields
    | `Label
        of LabelType.fields
    | `NullStmt
        of NullStmtType.fields
    | `RaiseStmt
        of RaiseStmtType.fields
    | `RequeueStmt
        of RequeueStmtType.fields
    | `ReturnStmt
        of ReturnStmtType.fields
    | `TerminateAlternative
        of TerminateAlternativeType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * CompositeStmtType.t
    * ErrorStmtType.t
    * SimpleStmtType.t
    *)
  type t = [
    | `AcceptStmt
        of AcceptStmtType.fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
    | `CaseStmt
        of CaseStmtType.fields
    | `ExtendedReturnStmt
        of ExtendedReturnStmtType.fields
    | `IfStmt
        of IfStmtType.fields
    | `NamedStmt
        of NamedStmtType.fields
    | `SelectStmt
        of SelectStmtType.fields
    | `ErrorStmt
        of ErrorStmtType.fields
    | `AbortStmt
        of AbortStmtType.fields
    | `AssignStmt
        of AssignStmtType.fields
    | `CallStmt
        of CallStmtType.fields
    | `DelayStmt
        of DelayStmtType.fields
    | `ExitStmt
        of ExitStmtType.fields
    | `GotoStmt
        of GotoStmtType.fields
    | `Label
        of LabelType.fields
    | `NullStmt
        of NullStmtType.fields
    | `RaiseStmt
        of RaiseStmtType.fields
    | `RequeueStmt
        of RequeueStmtType.fields
    | `ReturnStmt
        of ReturnStmtType.fields
    | `TerminateAlternative
        of TerminateAlternativeType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AcceptStmt _
      | `AcceptStmtWithStmts _
      | `ForLoopStmt _
      | `LoopStmt _
      | `WhileLoopStmt _
      | `BeginBlock _
      | `DeclBlock _
      | `CaseStmt _
      | `ExtendedReturnStmt _
      | `IfStmt _
      | `NamedStmt _
      | `SelectStmt _
      | `ErrorStmt _
      | `AbortStmt _
      | `AssignStmt _
      | `CallStmt _
      | `DelayStmt _
      | `ExitStmt _
      | `GotoStmt _
      | `Label _
      | `NullStmt _
      | `RaiseStmt _
      | `RequeueStmt _
      | `ReturnStmt _
      | `TerminateAlternative _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and CompositeStmtType : sig
   
  
   
  (**
    * AcceptStmtType.t
    * BaseLoopStmtType.t
    * BlockStmtType.t
    * CaseStmtType.t
    * ExtendedReturnStmtType.t
    * IfStmtType.t
    * NamedStmtType.t
    * SelectStmtType.t
    *)
  type t = [
    | `AcceptStmt
        of AcceptStmtType.fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
    | `CaseStmt
        of CaseStmtType.fields
    | `ExtendedReturnStmt
        of ExtendedReturnStmtType.fields
    | `IfStmt
        of IfStmtType.fields
    | `NamedStmt
        of NamedStmtType.fields
    | `SelectStmt
        of SelectStmtType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * AcceptStmtType.t
    * BaseLoopStmtType.t
    * BlockStmtType.t
    * CaseStmtType.t
    * ExtendedReturnStmtType.t
    * IfStmtType.t
    * NamedStmtType.t
    * SelectStmtType.t
    *)
  type t = [
    | `AcceptStmt
        of AcceptStmtType.fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
    | `CaseStmt
        of CaseStmtType.fields
    | `ExtendedReturnStmt
        of ExtendedReturnStmtType.fields
    | `IfStmt
        of IfStmtType.fields
    | `NamedStmt
        of NamedStmtType.fields
    | `SelectStmt
        of SelectStmtType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AcceptStmt _
      | `AcceptStmtWithStmts _
      | `ForLoopStmt _
      | `LoopStmt _
      | `WhileLoopStmt _
      | `BeginBlock _
      | `DeclBlock _
      | `CaseStmt _
      | `ExtendedReturnStmt _
      | `IfStmt _
      | `NamedStmt _
      | `SelectStmt _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AcceptStmtType : sig
   
  
   
  type t = [
    | `AcceptStmt
        of fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
  ]
  and fields = 
  {
         
    f_name: IdentifierType.t
    Lazy.t;
         
    f_entry_index_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AcceptStmt
        of fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
  ]
  and fields = 
  {
         
    f_name: IdentifierType.t
    Lazy.t;
         
    f_entry_index_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.accept_stmt_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         IdentifierType.wrap context (field_c_value)
      in
         

      node
    in
    let f_entry_index_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.accept_stmt_f_entry_index_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    let f_params () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.accept_stmt_f_params
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         EntryCompletionFormalParamsType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AcceptStmt {
        f_name
          = Lazy.from_fun f_name ;
        f_entry_index_expr
          = Lazy.from_fun f_entry_index_expr ;
        f_params
          = Lazy.from_fun f_params ;
        c_value = c_value;
        context = context
      }

end

      
and AcceptStmtWithStmtsType : sig
   
  
   
  type t = [
    | `AcceptStmtWithStmts
        of fields
  ]
  and fields = 
  {
         
    f_name: IdentifierType.t
    Lazy.t;
         
    f_entry_index_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AcceptStmtWithStmts
        of fields
  ]
  and fields = 
  {
         
    f_name: IdentifierType.t
    Lazy.t;
         
    f_entry_index_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.accept_stmt_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         IdentifierType.wrap context (field_c_value)
      in
         

      node
    in
    let f_entry_index_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.accept_stmt_f_entry_index_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    let f_params () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.accept_stmt_f_params
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         EntryCompletionFormalParamsType.wrap context (field_c_value)
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.accept_stmt_with_stmts_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         HandledStmtsType.wrap context (field_c_value)
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.accept_stmt_with_stmts_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AcceptStmtWithStmts {
        f_name
          = Lazy.from_fun f_name ;
        f_entry_index_expr
          = Lazy.from_fun f_entry_index_expr ;
        f_params
          = Lazy.from_fun f_params ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and BaseLoopStmtType : sig
   
  
   
  type t = [
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `ForLoopStmt _
      | `LoopStmt _
      | `WhileLoopStmt _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and ForLoopStmtType : sig
   
  
   
  type t = [
    | `ForLoopStmt
        of fields
  ]
  and fields = 
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ForLoopStmt
        of fields
  ]
  and fields = 
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_loop_stmt_f_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (LoopSpecType.wrap context (field_c_value))
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_loop_stmt_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         StmtListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_loop_stmt_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ForLoopStmt {
        f_spec
          = Lazy.from_fun f_spec ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and LoopStmtType : sig
   
  
   
  type t = [
    | `LoopStmt
        of fields
  ]
  and fields = 
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `LoopStmt
        of fields
  ]
  and fields = 
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_loop_stmt_f_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (LoopSpecType.wrap context (field_c_value))
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_loop_stmt_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         StmtListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_loop_stmt_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `LoopStmt {
        f_spec
          = Lazy.from_fun f_spec ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and WhileLoopStmtType : sig
   
  
   
  type t = [
    | `WhileLoopStmt
        of fields
  ]
  and fields = 
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `WhileLoopStmt
        of fields
  ]
  and fields = 
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_loop_stmt_f_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (LoopSpecType.wrap context (field_c_value))
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_loop_stmt_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         StmtListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.base_loop_stmt_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `WhileLoopStmt {
        f_spec
          = Lazy.from_fun f_spec ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and BlockStmtType : sig
   
  
   
  type t = [
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `BeginBlock _
      | `DeclBlock _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and BeginBlockType : sig
   
  
   
  type t = [
    | `BeginBlock
        of fields
  ]
  and fields = 
  {
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `BeginBlock
        of fields
  ]
  and fields = 
  {
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.begin_block_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         HandledStmtsType.wrap context (field_c_value)
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.begin_block_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `BeginBlock {
        f_stmts
          = Lazy.from_fun f_stmts ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and DeclBlockType : sig
   
  
   
  type t = [
    | `DeclBlock
        of fields
  ]
  and fields = 
  {
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DeclBlock
        of fields
  ]
  and fields = 
  {
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decls () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.decl_block_f_decls
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         DeclarativePartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.decl_block_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         HandledStmtsType.wrap context (field_c_value)
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.decl_block_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DeclBlock {
        f_decls
          = Lazy.from_fun f_decls ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and CaseStmtType : sig
   
  
   
  type t = [
    | `CaseStmt
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_alternatives: CaseStmtAlternativeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CaseStmt
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_alternatives: CaseStmtAlternativeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.case_stmt_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_alternatives () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.case_stmt_f_alternatives
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         CaseStmtAlternativeListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `CaseStmt {
        f_expr
          = Lazy.from_fun f_expr ;
        f_alternatives
          = Lazy.from_fun f_alternatives ;
        c_value = c_value;
        context = context
      }

end

      
and ExtendedReturnStmtType : sig
   
  
   
  type t = [
    | `ExtendedReturnStmt
        of fields
  ]
  and fields = 
  {
         
    f_decl: ExtendedReturnStmtObjectDeclType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ExtendedReturnStmt
        of fields
  ]
  and fields = 
  {
         
    f_decl: ExtendedReturnStmtObjectDeclType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.extended_return_stmt_f_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExtendedReturnStmtObjectDeclType.wrap context (field_c_value)
      in
         

      node
    in
    let f_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.extended_return_stmt_f_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (HandledStmtsType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ExtendedReturnStmt {
        f_decl
          = Lazy.from_fun f_decl ;
        f_stmts
          = Lazy.from_fun f_stmts ;
        c_value = c_value;
        context = context
      }

end

      
and IfStmtType : sig
   
  
   
  type t = [
    | `IfStmt
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_stmts: StmtListType.t
    Lazy.t;
         
    f_alternatives: ElsifStmtPartListType.t
    Lazy.t;
         
    f_else_stmts: StmtListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `IfStmt
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_stmts: StmtListType.t
    Lazy.t;
         
    f_alternatives: ElsifStmtPartListType.t
    Lazy.t;
         
    f_else_stmts: StmtListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_cond_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.if_stmt_f_cond_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_then_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.if_stmt_f_then_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         StmtListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_alternatives () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.if_stmt_f_alternatives
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ElsifStmtPartListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_else_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.if_stmt_f_else_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (StmtListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `IfStmt {
        f_cond_expr
          = Lazy.from_fun f_cond_expr ;
        f_then_stmts
          = Lazy.from_fun f_then_stmts ;
        f_alternatives
          = Lazy.from_fun f_alternatives ;
        f_else_stmts
          = Lazy.from_fun f_else_stmts ;
        c_value = c_value;
        context = context
      }

end

      
and NamedStmtType : sig
   
  
   
  type t = [
    | `NamedStmt
        of fields
  ]
  and fields = 
  {
         
    f_decl: NamedStmtDeclType.t
    Lazy.t;
         
    f_stmt: [
      | `BeginBlock
          of BeginBlockType.fields
      | `DeclBlock
          of DeclBlockType.fields
      | `ForLoopStmt
          of ForLoopStmtType.fields
      | `LoopStmt
          of LoopStmtType.fields
      | `WhileLoopStmt
          of WhileLoopStmtType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NamedStmt
        of fields
  ]
  and fields = 
  {
         
    f_decl: NamedStmtDeclType.t
    Lazy.t;
         
    f_stmt: [
      | `BeginBlock
          of BeginBlockType.fields
      | `DeclBlock
          of DeclBlockType.fields
      | `ForLoopStmt
          of ForLoopStmtType.fields
      | `LoopStmt
          of LoopStmtType.fields
      | `WhileLoopStmt
          of WhileLoopStmtType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.named_stmt_f_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NamedStmtDeclType.wrap context (field_c_value)
      in
         

      node
    in
    let f_stmt () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.named_stmt_f_stmt
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         CompositeStmtType.wrap context (field_c_value)
      in
         

      match node with
            
      | `BeginBlock _
      | `DeclBlock _
      | `ForLoopStmt _
      | `LoopStmt _
      | `WhileLoopStmt _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NamedStmt {
        f_decl
          = Lazy.from_fun f_decl ;
        f_stmt
          = Lazy.from_fun f_stmt ;
        c_value = c_value;
        context = context
      }

end

      
and SelectStmtType : sig
   
  
   
  type t = [
    | `SelectStmt
        of fields
  ]
  and fields = 
  {
         
    f_guards: SelectWhenPartListType.t
    Lazy.t;
         
    f_else_stmts: StmtListType.t
    option
    Lazy.t;
         
    f_abort_stmts: StmtListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SelectStmt
        of fields
  ]
  and fields = 
  {
         
    f_guards: SelectWhenPartListType.t
    Lazy.t;
         
    f_else_stmts: StmtListType.t
    option
    Lazy.t;
         
    f_abort_stmts: StmtListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_guards () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.select_stmt_f_guards
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SelectWhenPartListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_else_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.select_stmt_f_else_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (StmtListType.wrap context (field_c_value))
      in
         

      node
    in
    let f_abort_stmts () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.select_stmt_f_abort_stmts
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (StmtListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SelectStmt {
        f_guards
          = Lazy.from_fun f_guards ;
        f_else_stmts
          = Lazy.from_fun f_else_stmts ;
        f_abort_stmts
          = Lazy.from_fun f_abort_stmts ;
        c_value = c_value;
        context = context
      }

end

      
and ErrorStmtType : sig
   
  
   
  type t = [
    | `ErrorStmt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ErrorStmt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ErrorStmt {
        c_value = c_value;
        context = context
      }

end

      
and SimpleStmtType : sig
   
  
   
  type t = [
    | `AbortStmt
        of AbortStmtType.fields
    | `AssignStmt
        of AssignStmtType.fields
    | `CallStmt
        of CallStmtType.fields
    | `DelayStmt
        of DelayStmtType.fields
    | `ExitStmt
        of ExitStmtType.fields
    | `GotoStmt
        of GotoStmtType.fields
    | `Label
        of LabelType.fields
    | `NullStmt
        of NullStmtType.fields
    | `RaiseStmt
        of RaiseStmtType.fields
    | `RequeueStmt
        of RequeueStmtType.fields
    | `ReturnStmt
        of ReturnStmtType.fields
    | `TerminateAlternative
        of TerminateAlternativeType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbortStmt
        of AbortStmtType.fields
    | `AssignStmt
        of AssignStmtType.fields
    | `CallStmt
        of CallStmtType.fields
    | `DelayStmt
        of DelayStmtType.fields
    | `ExitStmt
        of ExitStmtType.fields
    | `GotoStmt
        of GotoStmtType.fields
    | `Label
        of LabelType.fields
    | `NullStmt
        of NullStmtType.fields
    | `RaiseStmt
        of RaiseStmtType.fields
    | `RequeueStmt
        of RequeueStmtType.fields
    | `ReturnStmt
        of ReturnStmtType.fields
    | `TerminateAlternative
        of TerminateAlternativeType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AbortStmt _
      | `AssignStmt _
      | `CallStmt _
      | `DelayStmt _
      | `ExitStmt _
      | `GotoStmt _
      | `Label _
      | `NullStmt _
      | `RaiseStmt _
      | `RequeueStmt _
      | `ReturnStmt _
      | `TerminateAlternative _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AbortStmtType : sig
   
  
   
  type t = [
    | `AbortStmt
        of fields
  ]
  and fields = 
  {
         
    f_names: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AbortStmt
        of fields
  ]
  and fields = 
  {
         
    f_names: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_names () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.abort_stmt_f_names
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AbortStmt {
        f_names
          = Lazy.from_fun f_names ;
        c_value = c_value;
        context = context
      }

end

      
and AssignStmtType : sig
   
  
   
  type t = [
    | `AssignStmt
        of fields
  ]
  and fields = 
  {
         
    f_dest: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AssignStmt
        of fields
  ]
  and fields = 
  {
         
    f_dest: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_dest () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.assign_stmt_f_dest
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.assign_stmt_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AssignStmt {
        f_dest
          = Lazy.from_fun f_dest ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and CallStmtType : sig
   
  
   
  type t = [
    | `CallStmt
        of fields
  ]
  and fields = 
  {
         
    f_call: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `CallStmt
        of fields
  ]
  and fields = 
  {
         
    f_call: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_call () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.call_stmt_f_call
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `CallStmt {
        f_call
          = Lazy.from_fun f_call ;
        c_value = c_value;
        context = context
      }

end

      
and DelayStmtType : sig
   
  
   
  type t = [
    | `DelayStmt
        of fields
  ]
  and fields = 
  {
         
    f_has_until: UntilNodeType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DelayStmt
        of fields
  ]
  and fields = 
  {
         
    f_has_until: UntilNodeType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_until () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.delay_stmt_f_has_until
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         UntilNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.delay_stmt_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `MembershipExpr _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DelayStmt {
        f_has_until
          = Lazy.from_fun f_has_until ;
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and ExitStmtType : sig
   
  
   
  type t = [
    | `ExitStmt
        of fields
  ]
  and fields = 
  {
         
    f_loop_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    option
    Lazy.t;
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ExitStmt
        of fields
  ]
  and fields = 
  {
         
    f_loop_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    option
    Lazy.t;
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_loop_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.exit_stmt_f_loop_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (NameType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `CharLiteral _
      | Some `DottedName _
      | Some `Identifier _
      | Some `StringLiteral _
      | None as e -> e
      | _ -> assert false
    in
    let f_cond_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.exit_stmt_f_cond_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ExitStmt {
        f_loop_name
          = Lazy.from_fun f_loop_name ;
        f_cond_expr
          = Lazy.from_fun f_cond_expr ;
        c_value = c_value;
        context = context
      }

end

      
and GotoStmtType : sig
   
  
   
  type t = [
    | `GotoStmt
        of fields
  ]
  and fields = 
  {
         
    f_label_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `GotoStmt
        of fields
  ]
  and fields = 
  {
         
    f_label_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_label_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.goto_stmt_f_label_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `GotoStmt {
        f_label_name
          = Lazy.from_fun f_label_name ;
        c_value = c_value;
        context = context
      }

end

      
and LabelType : sig
   
  
   
  type t = [
    | `Label
        of fields
  ]
  and fields = 
  {
         
    f_decl: LabelDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `Label
        of fields
  ]
  and fields = 
  {
         
    f_decl: LabelDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.label_f_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         LabelDeclType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `Label {
        f_decl
          = Lazy.from_fun f_decl ;
        c_value = c_value;
        context = context
      }

end

      
and NullStmtType : sig
   
  
   
  type t = [
    | `NullStmt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `NullStmt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `NullStmt {
        c_value = c_value;
        context = context
      }

end

      
and RaiseStmtType : sig
   
  
   
  type t = [
    | `RaiseStmt
        of fields
  ]
  and fields = 
  {
         
    f_exception_name: NameType.t
    option
    Lazy.t;
         
    f_error_message: ExprType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `RaiseStmt
        of fields
  ]
  and fields = 
  {
         
    f_exception_name: NameType.t
    option
    Lazy.t;
         
    f_error_message: ExprType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_exception_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.raise_stmt_f_exception_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (NameType.wrap context (field_c_value))
      in
         

      node
    in
    let f_error_message () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.raise_stmt_f_error_message
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `RaiseStmt {
        f_exception_name
          = Lazy.from_fun f_exception_name ;
        f_error_message
          = Lazy.from_fun f_error_message ;
        c_value = c_value;
        context = context
      }

end

      
and RequeueStmtType : sig
   
  
   
  type t = [
    | `RequeueStmt
        of fields
  ]
  and fields = 
  {
         
    f_call_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_has_abort: AbortNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `RequeueStmt
        of fields
  ]
  and fields = 
  {
         
    f_call_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_has_abort: AbortNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_call_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.requeue_stmt_f_call_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CallExpr _
      | `CharLiteral _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `QualExpr _
      | `StringLiteral _
      | `TargetName _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_has_abort () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.requeue_stmt_f_has_abort
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AbortNodeType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `RequeueStmt {
        f_call_name
          = Lazy.from_fun f_call_name ;
        f_has_abort
          = Lazy.from_fun f_has_abort ;
        c_value = c_value;
        context = context
      }

end

      
and ReturnStmtType : sig
   
  
   
  type t = [
    | `ReturnStmt
        of fields
  ]
  and fields = 
  {
         
    f_return_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ReturnStmt
        of fields
  ]
  and fields = 
  {
         
    f_return_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_return_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.return_stmt_f_return_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ExprType.wrap context (field_c_value))
      in
         

      match node with
            
      | Some `Aggregate _
      | Some `Allocator _
      | Some `AttributeRef _
      | Some `BinOp _
      | Some `BracketAggregate _
      | Some `BracketDeltaAggregate _
      | Some `CallExpr _
      | Some `CaseExpr _
      | Some `CharLiteral _
      | Some `DeclExpr _
      | Some `DeltaAggregate _
      | Some `DottedName _
      | Some `ExplicitDeref _
      | Some `Identifier _
      | Some `IfExpr _
      | Some `IntLiteral _
      | Some `MembershipExpr _
      | Some `NullLiteral _
      | Some `NullRecordAggregate _
      | Some `ParenExpr _
      | Some `QualExpr _
      | Some `QuantifiedExpr _
      | Some `RaiseExpr _
      | Some `RealLiteral _
      | Some `RelationOp _
      | Some `StringLiteral _
      | Some `TargetName _
      | Some `UnOp _
      | Some `UpdateAttributeRef _
      | None as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ReturnStmt {
        f_return_expr
          = Lazy.from_fun f_return_expr ;
        c_value = c_value;
        context = context
      }

end

      
and TerminateAlternativeType : sig
   
  
   
  type t = [
    | `TerminateAlternative
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `TerminateAlternative
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `TerminateAlternative {
        c_value = c_value;
        context = context
      }

end

      
and SubpKindType : sig
   
  
   
  type t = [
    | `SubpKindFunction
        of SubpKindFunctionType.fields
    | `SubpKindProcedure
        of SubpKindProcedureType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SubpKindFunction
        of SubpKindFunctionType.fields
    | `SubpKindProcedure
        of SubpKindProcedureType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `SubpKindFunction _
      | `SubpKindProcedure _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and SubpKindFunctionType : sig
   
  
   
  type t = [
    | `SubpKindFunction
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SubpKindFunction
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SubpKindFunction {
        c_value = c_value;
        context = context
      }

end

      
and SubpKindProcedureType : sig
   
  
   
  type t = [
    | `SubpKindProcedure
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SubpKindProcedure
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SubpKindProcedure {
        c_value = c_value;
        context = context
      }

end

      
and SubunitType : sig
   
  
   
  type t = [
    | `Subunit
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_body: [
      | `PackageBody
          of PackageBodyType.fields
      | `ProtectedBody
          of ProtectedBodyType.fields
      | `SubpBody
          of SubpBodyType.fields
      | `TaskBody
          of TaskBodyType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `Subunit
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_body: [
      | `PackageBody
          of PackageBodyType.fields
      | `ProtectedBody
          of ProtectedBodyType.fields
      | `SubpBody
          of SubpBodyType.fields
      | `TaskBody
          of TaskBodyType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subunit_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      as e -> e
      | _ -> assert false
    in
    let f_body () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subunit_f_body
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BodyNodeType.wrap context (field_c_value)
      in
         

      match node with
            
      | `PackageBody _
      | `ProtectedBody _
      | `SubpBody _
      | `TaskBody _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `Subunit {
        f_name
          = Lazy.from_fun f_name ;
        f_body
          = Lazy.from_fun f_body ;
        c_value = c_value;
        context = context
      }

end

      
and SynchronizedNodeType : sig
   
  
   
  type t = [
    | `SynchronizedAbsent
        of SynchronizedAbsentType.fields
    | `SynchronizedPresent
        of SynchronizedPresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SynchronizedAbsent
        of SynchronizedAbsentType.fields
    | `SynchronizedPresent
        of SynchronizedPresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `SynchronizedAbsent _
      | `SynchronizedPresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and SynchronizedAbsentType : sig
   
  
   
  type t = [
    | `SynchronizedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SynchronizedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SynchronizedAbsent {
        c_value = c_value;
        context = context
      }

end

      
and SynchronizedPresentType : sig
   
  
   
  type t = [
    | `SynchronizedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SynchronizedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SynchronizedPresent {
        c_value = c_value;
        context = context
      }

end

      
and TaggedNodeType : sig
   
  
   
  type t = [
    | `TaggedAbsent
        of TaggedAbsentType.fields
    | `TaggedPresent
        of TaggedPresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `TaggedAbsent
        of TaggedAbsentType.fields
    | `TaggedPresent
        of TaggedPresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `TaggedAbsent _
      | `TaggedPresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and TaggedAbsentType : sig
   
  
   
  type t = [
    | `TaggedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `TaggedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `TaggedAbsent {
        c_value = c_value;
        context = context
      }

end

      
and TaggedPresentType : sig
   
  
   
  type t = [
    | `TaggedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `TaggedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `TaggedPresent {
        c_value = c_value;
        context = context
      }

end

      
and TaskDefType : sig
   
  
   
  type t = [
    | `TaskDef
        of fields
  ]
  and fields = 
  {
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `TaskDef
        of fields
  ]
  and fields = 
  {
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_interfaces () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_def_f_interfaces
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ParentListType.wrap context (field_c_value))
      in
         

      node
    in
    let f_public_part () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_def_f_public_part
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         PublicPartType.wrap context (field_c_value)
      in
         

      node
    in
    let f_private_part () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_def_f_private_part
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (PrivatePartType.wrap context (field_c_value))
      in
         

      node
    in
    let f_end_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.task_def_f_end_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (EndNameType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `TaskDef {
        f_interfaces
          = Lazy.from_fun f_interfaces ;
        f_public_part
          = Lazy.from_fun f_public_part ;
        f_private_part
          = Lazy.from_fun f_private_part ;
        f_end_name
          = Lazy.from_fun f_end_name ;
        c_value = c_value;
        context = context
      }

end

      
and TypeDefType : sig
   
  
   
  (**
    * AccessDefType.t
    * ArrayTypeDefType.t
    * DerivedTypeDefType.t
    * EnumTypeDefType.t
    * FormalDiscreteTypeDefType.t
    * InterfaceTypeDefType.t
    * ModIntTypeDefType.t
    * PrivateTypeDefType.t
    * RealTypeDefType.t
    * RecordTypeDefType.t
    * SignedIntTypeDefType.t
    *)
  type t = [
    | `AccessToSubpDef
        of AccessToSubpDefType.fields
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
    | `ArrayTypeDef
        of ArrayTypeDefType.fields
    | `DerivedTypeDef
        of DerivedTypeDefType.fields
    | `EnumTypeDef
        of EnumTypeDefType.fields
    | `FormalDiscreteTypeDef
        of FormalDiscreteTypeDefType.fields
    | `InterfaceTypeDef
        of InterfaceTypeDefType.fields
    | `ModIntTypeDef
        of ModIntTypeDefType.fields
    | `PrivateTypeDef
        of PrivateTypeDefType.fields
    | `DecimalFixedPointDef
        of DecimalFixedPointDefType.fields
    | `FloatingPointDef
        of FloatingPointDefType.fields
    | `OrdinaryFixedPointDef
        of OrdinaryFixedPointDefType.fields
    | `RecordTypeDef
        of RecordTypeDefType.fields
    | `SignedIntTypeDef
        of SignedIntTypeDefType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * AccessDefType.t
    * ArrayTypeDefType.t
    * DerivedTypeDefType.t
    * EnumTypeDefType.t
    * FormalDiscreteTypeDefType.t
    * InterfaceTypeDefType.t
    * ModIntTypeDefType.t
    * PrivateTypeDefType.t
    * RealTypeDefType.t
    * RecordTypeDefType.t
    * SignedIntTypeDefType.t
    *)
  type t = [
    | `AccessToSubpDef
        of AccessToSubpDefType.fields
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
    | `ArrayTypeDef
        of ArrayTypeDefType.fields
    | `DerivedTypeDef
        of DerivedTypeDefType.fields
    | `EnumTypeDef
        of EnumTypeDefType.fields
    | `FormalDiscreteTypeDef
        of FormalDiscreteTypeDefType.fields
    | `InterfaceTypeDef
        of InterfaceTypeDefType.fields
    | `ModIntTypeDef
        of ModIntTypeDefType.fields
    | `PrivateTypeDef
        of PrivateTypeDefType.fields
    | `DecimalFixedPointDef
        of DecimalFixedPointDefType.fields
    | `FloatingPointDef
        of FloatingPointDefType.fields
    | `OrdinaryFixedPointDef
        of OrdinaryFixedPointDefType.fields
    | `RecordTypeDef
        of RecordTypeDefType.fields
    | `SignedIntTypeDef
        of SignedIntTypeDefType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AccessToSubpDef _
      | `AnonymousTypeAccessDef _
      | `TypeAccessDef _
      | `ArrayTypeDef _
      | `DerivedTypeDef _
      | `EnumTypeDef _
      | `FormalDiscreteTypeDef _
      | `InterfaceTypeDef _
      | `ModIntTypeDef _
      | `PrivateTypeDef _
      | `DecimalFixedPointDef _
      | `FloatingPointDef _
      | `OrdinaryFixedPointDef _
      | `RecordTypeDef _
      | `SignedIntTypeDef _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AccessDefType : sig
   
  
   
  (**
    * AccessToSubpDefType.t
    * BaseTypeAccessDefType.t
    *)
  type t = [
    | `AccessToSubpDef
        of AccessToSubpDefType.fields
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * AccessToSubpDefType.t
    * BaseTypeAccessDefType.t
    *)
  type t = [
    | `AccessToSubpDef
        of AccessToSubpDefType.fields
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AccessToSubpDef _
      | `AnonymousTypeAccessDef _
      | `TypeAccessDef _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AccessToSubpDefType : sig
   
  
   
  type t = [
    | `AccessToSubpDef
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_has_protected: ProtectedNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AccessToSubpDef
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_has_protected: ProtectedNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_not_null () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.access_def_f_has_not_null
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NotNullType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_protected () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.access_to_subp_def_f_has_protected
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ProtectedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subp_spec () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.access_to_subp_def_f_subp_spec
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubpSpecType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AccessToSubpDef {
        f_has_not_null
          = Lazy.from_fun f_has_not_null ;
        f_has_protected
          = Lazy.from_fun f_has_protected ;
        f_subp_spec
          = Lazy.from_fun f_subp_spec ;
        c_value = c_value;
        context = context
      }

end

      
and BaseTypeAccessDefType : sig
   
  
   
  type t = [
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AnonymousTypeAccessDef _
      | `TypeAccessDef _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AnonymousTypeAccessDefType : sig
   
  
   
  type t = [
    | `AnonymousTypeAccessDef
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_type_decl: BaseTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AnonymousTypeAccessDef
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_type_decl: BaseTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_not_null () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.access_def_f_has_not_null
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NotNullType.wrap context (field_c_value)
      in
         

      node
    in
    let f_type_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.anonymous_type_access_def_f_type_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BaseTypeDeclType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AnonymousTypeAccessDef {
        f_has_not_null
          = Lazy.from_fun f_has_not_null ;
        f_type_decl
          = Lazy.from_fun f_type_decl ;
        c_value = c_value;
        context = context
      }

end

      
and TypeAccessDefType : sig
   
  
   
  type t = [
    | `TypeAccessDef
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_has_all: AllNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `TypeAccessDef
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_has_all: AllNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_not_null () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.access_def_f_has_not_null
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NotNullType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_all () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.type_access_def_f_has_all
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AllNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_constant () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.type_access_def_f_has_constant
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ConstantNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subtype_indication () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.type_access_def_f_subtype_indication
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubtypeIndicationType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `TypeAccessDef {
        f_has_not_null
          = Lazy.from_fun f_has_not_null ;
        f_has_all
          = Lazy.from_fun f_has_all ;
        f_has_constant
          = Lazy.from_fun f_has_constant ;
        f_subtype_indication
          = Lazy.from_fun f_subtype_indication ;
        c_value = c_value;
        context = context
      }

end

      
and ArrayTypeDefType : sig
   
  
   
  type t = [
    | `ArrayTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_indices: ArrayIndicesType.t
    Lazy.t;
         
    f_component_type: ComponentDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ArrayTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_indices: ArrayIndicesType.t
    Lazy.t;
         
    f_component_type: ComponentDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_indices () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.array_type_def_f_indices
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ArrayIndicesType.wrap context (field_c_value)
      in
         

      node
    in
    let f_component_type () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.array_type_def_f_component_type
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ComponentDefType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ArrayTypeDef {
        f_indices
          = Lazy.from_fun f_indices ;
        f_component_type
          = Lazy.from_fun f_component_type ;
        c_value = c_value;
        context = context
      }

end

      
and DerivedTypeDefType : sig
   
  
   
  type t = [
    | `DerivedTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_has_synchronized: SynchronizedNodeType.t
    Lazy.t;
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_record_extension: BaseRecordDefType.t
    option
    Lazy.t;
         
    f_has_with_private: WithPrivateType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DerivedTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_has_synchronized: SynchronizedNodeType.t
    Lazy.t;
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_record_extension: BaseRecordDefType.t
    option
    Lazy.t;
         
    f_has_with_private: WithPrivateType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_abstract () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.derived_type_def_f_has_abstract
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AbstractNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_limited () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.derived_type_def_f_has_limited
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         LimitedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_synchronized () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.derived_type_def_f_has_synchronized
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SynchronizedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_subtype_indication () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.derived_type_def_f_subtype_indication
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubtypeIndicationType.wrap context (field_c_value)
      in
         

      node
    in
    let f_interfaces () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.derived_type_def_f_interfaces
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ParentListType.wrap context (field_c_value))
      in
         

      node
    in
    let f_record_extension () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.derived_type_def_f_record_extension
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (BaseRecordDefType.wrap context (field_c_value))
      in
         

      node
    in
    let f_has_with_private () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.derived_type_def_f_has_with_private
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         WithPrivateType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DerivedTypeDef {
        f_has_abstract
          = Lazy.from_fun f_has_abstract ;
        f_has_limited
          = Lazy.from_fun f_has_limited ;
        f_has_synchronized
          = Lazy.from_fun f_has_synchronized ;
        f_subtype_indication
          = Lazy.from_fun f_subtype_indication ;
        f_interfaces
          = Lazy.from_fun f_interfaces ;
        f_record_extension
          = Lazy.from_fun f_record_extension ;
        f_has_with_private
          = Lazy.from_fun f_has_with_private ;
        c_value = c_value;
        context = context
      }

end

      
and EnumTypeDefType : sig
   
  
   
  type t = [
    | `EnumTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_enum_literals: EnumLiteralDeclListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EnumTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_enum_literals: EnumLiteralDeclListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_enum_literals () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.enum_type_def_f_enum_literals
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         EnumLiteralDeclListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EnumTypeDef {
        f_enum_literals
          = Lazy.from_fun f_enum_literals ;
        c_value = c_value;
        context = context
      }

end

      
and FormalDiscreteTypeDefType : sig
   
  
   
  type t = [
    | `FormalDiscreteTypeDef
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `FormalDiscreteTypeDef
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `FormalDiscreteTypeDef {
        c_value = c_value;
        context = context
      }

end

      
and InterfaceTypeDefType : sig
   
  
   
  type t = [
    | `InterfaceTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_interface_kind: InterfaceKindType.t
    option
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `InterfaceTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_interface_kind: InterfaceKindType.t
    option
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_interface_kind () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.interface_type_def_f_interface_kind
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (InterfaceKindType.wrap context (field_c_value))
      in
         

      node
    in
    let f_interfaces () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.interface_type_def_f_interfaces
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ParentListType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `InterfaceTypeDef {
        f_interface_kind
          = Lazy.from_fun f_interface_kind ;
        f_interfaces
          = Lazy.from_fun f_interfaces ;
        c_value = c_value;
        context = context
      }

end

      
and ModIntTypeDefType : sig
   
  
   
  type t = [
    | `ModIntTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ModIntTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_expr () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.mod_int_type_def_f_expr
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BoxExpr _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ModIntTypeDef {
        f_expr
          = Lazy.from_fun f_expr ;
        c_value = c_value;
        context = context
      }

end

      
and PrivateTypeDefType : sig
   
  
   
  type t = [
    | `PrivateTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_tagged: TaggedNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `PrivateTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_tagged: TaggedNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_abstract () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.private_type_def_f_has_abstract
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AbstractNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_tagged () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.private_type_def_f_has_tagged
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         TaggedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_limited () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.private_type_def_f_has_limited
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         LimitedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `PrivateTypeDef {
        f_has_abstract
          = Lazy.from_fun f_has_abstract ;
        f_has_tagged
          = Lazy.from_fun f_has_tagged ;
        f_has_limited
          = Lazy.from_fun f_has_limited ;
        c_value = c_value;
        context = context
      }

end

      
and RealTypeDefType : sig
   
  
   
  type t = [
    | `DecimalFixedPointDef
        of DecimalFixedPointDefType.fields
    | `FloatingPointDef
        of FloatingPointDefType.fields
    | `OrdinaryFixedPointDef
        of OrdinaryFixedPointDefType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DecimalFixedPointDef
        of DecimalFixedPointDefType.fields
    | `FloatingPointDef
        of FloatingPointDefType.fields
    | `OrdinaryFixedPointDef
        of OrdinaryFixedPointDefType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `DecimalFixedPointDef _
      | `FloatingPointDef _
      | `OrdinaryFixedPointDef _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and DecimalFixedPointDefType : sig
   
  
   
  type t = [
    | `DecimalFixedPointDef
        of fields
  ]
  and fields = 
  {
         
    f_delta: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DecimalFixedPointDef
        of fields
  ]
  and fields = 
  {
         
    f_delta: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_delta () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.decimal_fixed_point_def_f_delta
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BoxExpr _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_digits () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.decimal_fixed_point_def_f_digits
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BoxExpr _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_range () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.decimal_fixed_point_def_f_range
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (RangeSpecType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DecimalFixedPointDef {
        f_delta
          = Lazy.from_fun f_delta ;
        f_digits
          = Lazy.from_fun f_digits ;
        f_range
          = Lazy.from_fun f_range ;
        c_value = c_value;
        context = context
      }

end

      
and FloatingPointDefType : sig
   
  
   
  type t = [
    | `FloatingPointDef
        of fields
  ]
  and fields = 
  {
         
    f_num_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `FloatingPointDef
        of fields
  ]
  and fields = 
  {
         
    f_num_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_num_digits () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.floating_point_def_f_num_digits
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BoxExpr _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_range () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.floating_point_def_f_range
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (RangeSpecType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `FloatingPointDef {
        f_num_digits
          = Lazy.from_fun f_num_digits ;
        f_range
          = Lazy.from_fun f_range ;
        c_value = c_value;
        context = context
      }

end

      
and OrdinaryFixedPointDefType : sig
   
  
   
  type t = [
    | `OrdinaryFixedPointDef
        of fields
  ]
  and fields = 
  {
         
    f_delta: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `OrdinaryFixedPointDef
        of fields
  ]
  and fields = 
  {
         
    f_delta: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_delta () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.ordinary_fixed_point_def_f_delta
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ExprType.wrap context (field_c_value)
      in
         

      match node with
            
      | `Aggregate _
      | `Allocator _
      | `AttributeRef _
      | `BinOp _
      | `BoxExpr _
      | `BracketAggregate _
      | `BracketDeltaAggregate _
      | `CallExpr _
      | `CaseExpr _
      | `CharLiteral _
      | `DeclExpr _
      | `DeltaAggregate _
      | `DottedName _
      | `ExplicitDeref _
      | `Identifier _
      | `IfExpr _
      | `IntLiteral _
      | `NullLiteral _
      | `NullRecordAggregate _
      | `ParenExpr _
      | `QualExpr _
      | `QuantifiedExpr _
      | `RaiseExpr _
      | `RealLiteral _
      | `RelationOp _
      | `StringLiteral _
      | `TargetName _
      | `UnOp _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_range () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.ordinary_fixed_point_def_f_range
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (RangeSpecType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `OrdinaryFixedPointDef {
        f_delta
          = Lazy.from_fun f_delta ;
        f_range
          = Lazy.from_fun f_range ;
        c_value = c_value;
        context = context
      }

end

      
and RecordTypeDefType : sig
   
  
   
  type t = [
    | `RecordTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_tagged: TaggedNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_record_def: BaseRecordDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `RecordTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_tagged: TaggedNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_record_def: BaseRecordDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_abstract () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.record_type_def_f_has_abstract
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AbstractNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_tagged () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.record_type_def_f_has_tagged
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         TaggedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_limited () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.record_type_def_f_has_limited
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         LimitedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_record_def () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.record_type_def_f_record_def
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         BaseRecordDefType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `RecordTypeDef {
        f_has_abstract
          = Lazy.from_fun f_has_abstract ;
        f_has_tagged
          = Lazy.from_fun f_has_tagged ;
        f_has_limited
          = Lazy.from_fun f_has_limited ;
        f_record_def
          = Lazy.from_fun f_record_def ;
        c_value = c_value;
        context = context
      }

end

      
and SignedIntTypeDefType : sig
   
  
   
  type t = [
    | `SignedIntTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SignedIntTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_range () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.signed_int_type_def_f_range
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         RangeSpecType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SignedIntTypeDef {
        f_range
          = Lazy.from_fun f_range ;
        c_value = c_value;
        context = context
      }

end

      
and TypeExprType : sig
   
  
   
  (**
    * AnonymousTypeType.t
    * EnumLitSynthTypeExprType.t
    * SubtypeIndicationType.t
    *)
  type t = [
    | `AnonymousType
        of AnonymousTypeType.fields
    | `EnumLitSynthTypeExpr
        of EnumLitSynthTypeExprType.fields
    | `SubtypeIndication
        of SubtypeIndicationType.fields
    | `ConstrainedSubtypeIndication
        of ConstrainedSubtypeIndicationType.fields
    | `DiscreteSubtypeIndication
        of DiscreteSubtypeIndicationType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  (**
    * AnonymousTypeType.t
    * EnumLitSynthTypeExprType.t
    * SubtypeIndicationType.t
    *)
  type t = [
    | `AnonymousType
        of AnonymousTypeType.fields
    | `EnumLitSynthTypeExpr
        of EnumLitSynthTypeExprType.fields
    | `SubtypeIndication
        of SubtypeIndicationType.fields
    | `ConstrainedSubtypeIndication
        of ConstrainedSubtypeIndicationType.fields
    | `DiscreteSubtypeIndication
        of DiscreteSubtypeIndicationType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `AnonymousType _
      | `EnumLitSynthTypeExpr _
      | `SubtypeIndication _
      | `ConstrainedSubtypeIndication _
      | `DiscreteSubtypeIndication _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and AnonymousTypeType : sig
   
  
   
  type t = [
    | `AnonymousType
        of fields
  ]
  and fields = 
  {
         
    f_type_decl: AnonymousTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `AnonymousType
        of fields
  ]
  and fields = 
  {
         
    f_type_decl: AnonymousTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_type_decl () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.anonymous_type_f_type_decl
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AnonymousTypeDeclType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `AnonymousType {
        f_type_decl
          = Lazy.from_fun f_type_decl ;
        c_value = c_value;
        context = context
      }

end

      
and EnumLitSynthTypeExprType : sig
   
  
   
  type t = [
    | `EnumLitSynthTypeExpr
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `EnumLitSynthTypeExpr
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `EnumLitSynthTypeExpr {
        c_value = c_value;
        context = context
      }

end

      
and SubtypeIndicationType : sig
   
  
   
  type t = [
    | `SubtypeIndication
        of fields
    | `ConstrainedSubtypeIndication
        of ConstrainedSubtypeIndicationType.fields
    | `DiscreteSubtypeIndication
        of DiscreteSubtypeIndicationType.fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `SubtypeIndication
        of fields
    | `ConstrainedSubtypeIndication
        of ConstrainedSubtypeIndicationType.fields
    | `DiscreteSubtypeIndication
        of DiscreteSubtypeIndicationType.fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_not_null () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subtype_indication_f_has_not_null
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NotNullType.wrap context (field_c_value)
      in
         

      node
    in
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subtype_indication_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_constraint () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subtype_indication_f_constraint
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ConstraintType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `SubtypeIndication {
        f_has_not_null
          = Lazy.from_fun f_has_not_null ;
        f_name
          = Lazy.from_fun f_name ;
        f_constraint
          = Lazy.from_fun f_constraint ;
        c_value = c_value;
        context = context
      }

end

      
and ConstrainedSubtypeIndicationType : sig
   
  
   
  type t = [
    | `ConstrainedSubtypeIndication
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `ConstrainedSubtypeIndication
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_not_null () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subtype_indication_f_has_not_null
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NotNullType.wrap context (field_c_value)
      in
         

      node
    in
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subtype_indication_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_constraint () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subtype_indication_f_constraint
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ConstraintType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `ConstrainedSubtypeIndication {
        f_has_not_null
          = Lazy.from_fun f_has_not_null ;
        f_name
          = Lazy.from_fun f_name ;
        f_constraint
          = Lazy.from_fun f_constraint ;
        c_value = c_value;
        context = context
      }

end

      
and DiscreteSubtypeIndicationType : sig
   
  
   
  type t = [
    | `DiscreteSubtypeIndication
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `DiscreteSubtypeIndication
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_not_null () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subtype_indication_f_has_not_null
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NotNullType.wrap context (field_c_value)
      in
         

      node
    in
    let f_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subtype_indication_f_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameType.wrap context (field_c_value)
      in
         

      match node with
            
      | `AttributeRef _
      | `CharLiteral _
      | `DottedName _
      | `Identifier _
      | `StringLiteral _
      | `UpdateAttributeRef _
      as e -> e
      | _ -> assert false
    in
    let f_constraint () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.subtype_indication_f_constraint
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         if is_null (getf field_c_value EntityStruct.node) then None else Some (ConstraintType.wrap context (field_c_value))
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `DiscreteSubtypeIndication {
        f_has_not_null
          = Lazy.from_fun f_has_not_null ;
        f_name
          = Lazy.from_fun f_name ;
        f_constraint
          = Lazy.from_fun f_constraint ;
        c_value = c_value;
        context = context
      }

end

      
and UnconstrainedArrayIndexType : sig
   
  
   
  type t = [
    | `UnconstrainedArrayIndex
        of fields
  ]
  and fields = 
  {
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UnconstrainedArrayIndex
        of fields
  ]
  and fields = 
  {
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_subtype_indication () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.unconstrained_array_index_f_subtype_indication
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         SubtypeIndicationType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `UnconstrainedArrayIndex {
        f_subtype_indication
          = Lazy.from_fun f_subtype_indication ;
        c_value = c_value;
        context = context
      }

end

      
and UntilNodeType : sig
   
  
   
  type t = [
    | `UntilAbsent
        of UntilAbsentType.fields
    | `UntilPresent
        of UntilPresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UntilAbsent
        of UntilAbsentType.fields
    | `UntilPresent
        of UntilPresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `UntilAbsent _
      | `UntilPresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and UntilAbsentType : sig
   
  
   
  type t = [
    | `UntilAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UntilAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `UntilAbsent {
        c_value = c_value;
        context = context
      }

end

      
and UntilPresentType : sig
   
  
   
  type t = [
    | `UntilPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UntilPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `UntilPresent {
        c_value = c_value;
        context = context
      }

end

      
and UseClauseType : sig
   
  
   
  type t = [
    | `UsePackageClause
        of UsePackageClauseType.fields
    | `UseTypeClause
        of UseTypeClauseType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UsePackageClause
        of UsePackageClauseType.fields
    | `UseTypeClause
        of UseTypeClauseType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `UsePackageClause _
      | `UseTypeClause _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and UsePackageClauseType : sig
   
  
   
  type t = [
    | `UsePackageClause
        of fields
  ]
  and fields = 
  {
         
    f_packages: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UsePackageClause
        of fields
  ]
  and fields = 
  {
         
    f_packages: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_packages () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.use_package_clause_f_packages
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `UsePackageClause {
        f_packages
          = Lazy.from_fun f_packages ;
        c_value = c_value;
        context = context
      }

end

      
and UseTypeClauseType : sig
   
  
   
  type t = [
    | `UseTypeClause
        of fields
  ]
  and fields = 
  {
         
    f_has_all: AllNodeType.t
    Lazy.t;
         
    f_types: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `UseTypeClause
        of fields
  ]
  and fields = 
  {
         
    f_has_all: AllNodeType.t
    Lazy.t;
         
    f_types: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_all () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.use_type_clause_f_has_all
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AllNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_types () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.use_type_clause_f_types
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `UseTypeClause {
        f_has_all
          = Lazy.from_fun f_has_all ;
        f_types
          = Lazy.from_fun f_types ;
        c_value = c_value;
        context = context
      }

end

      
and VariantType : sig
   
  
   
  type t = [
    | `Variant
        of fields
  ]
  and fields = 
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `Variant
        of fields
  ]
  and fields = 
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_choices () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.variant_f_choices
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         AlternativesListType.wrap context (field_c_value)
      in
         

      node
    in
    let f_components () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.variant_f_components
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         ComponentListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `Variant {
        f_choices
          = Lazy.from_fun f_choices ;
        f_components
          = Lazy.from_fun f_components ;
        c_value = c_value;
        context = context
      }

end

      
and VariantPartType : sig
   
  
   
  type t = [
    | `VariantPart
        of fields
  ]
  and fields = 
  {
         
    f_discr_name: IdentifierType.t
    Lazy.t;
         
    f_variant: VariantListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `VariantPart
        of fields
  ]
  and fields = 
  {
         
    f_discr_name: IdentifierType.t
    Lazy.t;
         
    f_variant: VariantListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_discr_name () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.variant_part_f_discr_name
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         IdentifierType.wrap context (field_c_value)
      in
         

      node
    in
    let f_variant () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.variant_part_f_variant
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         VariantListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `VariantPart {
        f_discr_name
          = Lazy.from_fun f_discr_name ;
        f_variant
          = Lazy.from_fun f_variant ;
        c_value = c_value;
        context = context
      }

end

      
and WithClauseType : sig
   
  
   
  type t = [
    | `WithClause
        of fields
  ]
  and fields = 
  {
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_has_private: PrivateNodeType.t
    Lazy.t;
         
    f_packages: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `WithClause
        of fields
  ]
  and fields = 
  {
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_has_private: PrivateNodeType.t
    Lazy.t;
         
    f_packages: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    let f_has_limited () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.with_clause_f_has_limited
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         LimitedNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_has_private () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.with_clause_f_has_private
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         PrivateNodeType.wrap context (field_c_value)
      in
         

      node
    in
    let f_packages () =
      let field_c_value = make EntityStruct.c_type in
      let _ : int = CFunctions.with_clause_f_packages
        (addr c_value)
        (addr field_c_value)
      in
      let node =
         NameListType.wrap context (field_c_value)
      in
         

      node
    in
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `WithClause {
        f_has_limited
          = Lazy.from_fun f_has_limited ;
        f_has_private
          = Lazy.from_fun f_has_private ;
        f_packages
          = Lazy.from_fun f_packages ;
        c_value = c_value;
        context = context
      }

end

      
and WithPrivateType : sig
   
  
   
  type t = [
    | `WithPrivateAbsent
        of WithPrivateAbsentType.fields
    | `WithPrivatePresent
        of WithPrivatePresentType.fields
  ]


  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `WithPrivateAbsent
        of WithPrivateAbsentType.fields
    | `WithPrivatePresent
        of WithPrivatePresentType.fields
  ]


  let wrap context c_value =
    (* This is an abstract node, call the root wrap function and filter to get
     the desired type *)
    match AdaNodeType.wrap context (c_value) with
      | `WithPrivateAbsent _
      | `WithPrivatePresent _
      as e -> e
      | _ ->
          (* This should not happen if the types are correct *)
          assert false

end

      
and WithPrivateAbsentType : sig
   
  
   
  type t = [
    | `WithPrivateAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `WithPrivateAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `WithPrivateAbsent {
        c_value = c_value;
        context = context
      }

end

      
and WithPrivatePresentType : sig
   
  
   
  type t = [
    | `WithPrivatePresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



  val wrap :
    AnalysisContext.t
    -> EntityStruct.t structure
    -> t

end = struct
   
  
   
  type t = [
    | `WithPrivatePresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }




  let wrap context c_value =
    if is_null (getf c_value EntityStruct.node) then
      raise SyntaxError
    else
      `WithPrivatePresent {
        c_value = c_value;
        context = context
      }

end


and Entity : sig
  type t = EntityStruct.t structure

  val info : t -> EntityInfo.t

  val equal : t -> t -> bool

  val compare : t -> t -> int

  val hash : t -> int
end = struct
  type t = EntityStruct.t structure

  let info value =
    EntityInfo.wrap (getf value EntityStruct.info)

  let compare e1 e2 =
    let open Stdlib in
    let compare_node =
      compare (getf e1 EntityStruct.node) (getf e2 EntityStruct.node)
    in
    if compare_node = 0 then
      compare
        (getf (getf e1 EntityStruct.info) EntityInfoStruct.rebindings)
        (getf (getf e2 EntityStruct.info) EntityInfoStruct.rebindings)
    else
      compare_node

  let equal e1 e2 =
    compare e1 e2 = 0

  let hash e =
    Hashtbl.hash
      ( getf e EntityStruct.node
      , getf (getf e EntityStruct.info) EntityInfoStruct.rebindings )
end

and AnalysisUnit : sig
  type t = {
    c_value : AnalysisUnitStruct.t;
    context : AnalysisContext.t
  }

  val root : t -> AdaNodeType.t option
  val diagnostics : t -> Diagnostic.t list
  val filename : t -> string
  val reparse : ?charset:string -> ?buffer:string -> t -> unit
  val first_token : t -> Token.t
  val last_token : t -> Token.t
  val token_count : t -> int
  val trivia_count : t -> int

  
  val fold_tokens : ('a -> Token.t -> 'a) -> 'a -> t -> 'a
  (**
   * Fold all the token this node contains by calling f on each token.
   *)

  val iter_tokens : (Token.t -> unit) -> t-> unit
  (**
   * Iterate over all token this node contains by calling f on each token.
   *)

  val map_tokens : (Token.t -> 'a) -> t -> 'a list
  (**
   * Map each token calling the given function
   *)

  val tokens : t -> Token.t list
  (**
   * Return a list of tokens for the given node
   *)


  val wrap : AnalysisContext.t -> AnalysisUnitStruct.t -> t
  val unwrap : t -> AnalysisUnitStruct.t
end = struct
  type t = {
    c_value : AnalysisUnitStruct.t;
    context : AnalysisContext.t
  }

  let root analysis_unit =
    let c_value = make EntityStruct.c_type in
    AnalysisUnitStruct.unit_root
      (AnalysisUnit.unwrap (analysis_unit))
      (addr c_value);
    if is_null (getf c_value EntityStruct.node) then None else Some (AdaNodeType.wrap analysis_unit.context (c_value))

  let diagnostics unit =
    let c_unit = AnalysisUnit.unwrap (unit) in
    let length = AnalysisUnitStruct.unit_diagnostic_count c_unit in
    let f i =
      let diag = allocate_n Diagnostic.c_type ~count:1 in
      let _ : int = AnalysisUnitStruct.unit_diagnostic c_unit i diag in
      !@ diag
    in
    List.init length f

  let filename unit =
    unwrap_str( AnalysisUnitStruct.unit_filename
      (AnalysisUnit.unwrap (unit)))

  let reparse ?charset:(charset="") ?buffer ctx =
    match buffer with
    | None ->
        ignore
          (AnalysisUnitStruct.unit_reparse_from_file ctx.c_value charset)
    | Some buffer ->
        ignore (AnalysisUnitStruct.unit_reparse_from_buffer ctx.c_value
          charset buffer (Unsigned.Size_t.of_int (String.length buffer)))

  let first_token unit =
    let c_unit = AnalysisUnit.unwrap (unit) in
    let result_ptr = allocate_n Token.c_type ~count:1 in
    AnalysisUnitStruct.unit_first_token c_unit result_ptr ;
    !@ result_ptr

  let last_token unit =
    let c_unit = AnalysisUnit.unwrap (unit) in
    let result_ptr = allocate_n Token.c_type ~count:1 in
    AnalysisUnitStruct.unit_last_token c_unit result_ptr ;
    !@ result_ptr

  let token_count unit =
    AnalysisUnitStruct.unit_token_count
      (AnalysisUnit.unwrap (unit))

  let trivia_count unit =
    AnalysisUnitStruct.unit_trivia_count
      (AnalysisUnit.unwrap (unit))

  
  let fold_tokens f init node =
    let tok_start = first_token node in
    let tok_end = last_token node in
    let rec aux acc tok_curr =
      let new_acc = f acc tok_curr in
      if Token.equal tok_curr tok_end then
        new_acc
      else
        aux new_acc (Token.next tok_curr)
    in
    aux init tok_start

  let iter_tokens f node =
    let tok_start = first_token node in
    let tok_end = last_token node in
    let rec aux tok_curr =
      f tok_curr;
      if not (Token.equal tok_curr tok_end) then
        aux (Token.next tok_curr)
    in
    aux tok_start

  let map_tokens f node =
    let tok_start = first_token node in
    let tok_end = last_token node in
    let rec aux tok_curr =
      let value = f tok_curr in
      if Token.equal tok_curr tok_end then
        [value]
      else
        value :: aux (Token.next tok_curr)
    in
    aux tok_start

  let tokens node =
    map_tokens (fun x -> x) node


  let wrap context c_value = {
    c_value=c_value;
    context=context;
  }

  let unwrap {c_value} = c_value
end

and AnalysisContext : sig
  type t = {
    c_value : AnalysisContextStruct.t;
    unit_provider : UnitProvider.t }

  val create :
    ?charset:string
    -> ?with_trivia:bool
    -> ?tab_stop:int
    -> ?unit_provider:UnitProvider.t
    -> unit
    -> t

  val get_from_file :
    ?charset:string
    -> ?reparse:bool
    -> ?grammar_rule:GrammarRule.t
    -> t
    -> string
    -> AnalysisUnit.t

  val get_from_buffer :
    ?charset:string
    -> ?grammar_rule:GrammarRule.t
    -> t
    -> string
    -> string
    -> AnalysisUnit.t
end = struct
  type t = {
    c_value : AnalysisContextStruct.t;
    unit_provider : UnitProvider.t }

  let create
    ?charset:(charset="")
    ?with_trivia:(with_trivia=true)
    ?tab_stop:(tab_stop=8)
    ?unit_provider:(unit_provider=UnitProvider.null) () : t =
    if tab_stop < 1 then
      raise (Invalid_argument "Invalid tab_stop (positive integer expected)") ;
    let c_context =
       AnalysisContextStruct.create_analysis_context
         charset
         Ctypes.null (* TODO: bind the file readers API to OCaml*)
         (!@unit_provider)
         with_trivia
         tab_stop
    in
    { c_value= c_context
      ; unit_provider= unit_provider }

  let get_from_file
    ?charset:(charset="")
    ?reparse:(reparse=false)
    ?grammar_rule:(grammar_rule=default_grammar_rule)
    ctx
    filename : AnalysisUnit.t =

    AnalysisUnit.wrap ctx (AnalysisContextStruct.get_analysis_unit_from_file ctx.c_value filename charset reparse grammar_rule)

  let get_from_buffer
    ?charset:(charset="")
    ?grammar_rule:(grammar_rule=default_grammar_rule)
    ctx
    filename
    buffer : AnalysisUnit.t =

    AnalysisUnit.wrap ctx (AnalysisContextStruct.get_analysis_unit_from_buffer ctx.c_value filename charset buffer (Unsigned.Size_t.of_int (String.length buffer)) grammar_rule)
end


and EntityInfo : sig
   
  
  type t = {
      md :
         Metadata.t;
      rebindings :
         Rebindings.t;
      from_rebound :
         bool;
  }


  val wrap : EntityInfoStruct.t structure -> t

  val unwrap : t -> EntityInfoStruct.t structure


end = struct
   
  
  type t = {
      md :
         Metadata.t;
      rebindings :
         Rebindings.t;
      from_rebound :
         bool;
  }


  let wrap c_value = {
    md = Metadata.wrap ((getf c_value EntityInfoStruct.md));
    rebindings = (getf c_value EntityInfoStruct.rebindings);
    from_rebound = (getf c_value EntityInfoStruct.from_rebound);
  }

  let unwrap value =
    let c_value = make EntityInfoStruct.c_type in
    setf c_value
      EntityInfoStruct.md
      (Metadata.unwrap (value.md));
    setf c_value
      EntityInfoStruct.rebindings
      (value.rebindings);
    setf c_value
      EntityInfoStruct.from_rebound
      (value.from_rebound);
    c_value


end



and Metadata : sig
   
  
  type t = {
      dottable_subp :
         bool;
      access_entity :
         bool;
      primitive :
         BareNode.t;
      primitive_real_type :
         BareNode.t;
  }


  val wrap : MetadataStruct.t structure -> t

  val unwrap : t -> MetadataStruct.t structure


end = struct
   
  
  type t = {
      dottable_subp :
         bool;
      access_entity :
         bool;
      primitive :
         BareNode.t;
      primitive_real_type :
         BareNode.t;
  }


  let wrap c_value = {
    dottable_subp = (getf c_value MetadataStruct.dottable_subp);
    access_entity = (getf c_value MetadataStruct.access_entity);
    primitive = (getf c_value MetadataStruct.primitive);
    primitive_real_type = (getf c_value MetadataStruct.primitive_real_type);
  }

  let unwrap value =
    let c_value = make MetadataStruct.c_type in
    setf c_value
      MetadataStruct.dottable_subp
      (value.dottable_subp);
    setf c_value
      MetadataStruct.access_entity
      (value.access_entity);
    setf c_value
      MetadataStruct.primitive
      (value.primitive);
    setf c_value
      MetadataStruct.primitive_real_type
      (value.primitive_real_type);
    c_value


end


   
module Aspect : sig
   
  
  type t = {
      exists :
         bool;
      node :
         AdaNodeType.t option;
      value :
         ExprType.t option;
  }


  val wrap : AnalysisContext.t -> AspectStruct.t structure -> t

  val unwrap : t -> AspectStruct.t structure


end = struct
   
  
  type t = {
      exists :
         bool;
      node :
         AdaNodeType.t option;
      value :
         ExprType.t option;
  }


  let wrap context c_value = {
    exists = (getf c_value AspectStruct.exists);
    node = if is_null (getf (getf c_value AspectStruct.node) EntityStruct.node) then None else Some (AdaNodeType.wrap context ((getf c_value AspectStruct.node)));
    value = if is_null (getf (getf c_value AspectStruct.value) EntityStruct.node) then None else Some (ExprType.wrap context ((getf c_value AspectStruct.value)));
  }

  let unwrap value =
    let c_value = make AspectStruct.c_type in
    setf c_value
      AspectStruct.exists
      (value.exists);
    setf c_value
      AspectStruct.node
      (match value.node with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    setf c_value
      AspectStruct.value
      (match value.value with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    c_value


end

   
module CompletionItem : sig
   
  
  type t = {
      decl :
         BasicDeclType.t option;
      is_dot_call :
         bool;
      is_visible :
         bool;
  }


  val wrap : AnalysisContext.t -> CompletionItemStruct.t structure -> t

  val unwrap : t -> CompletionItemStruct.t structure


end = struct
   
  
  type t = {
      decl :
         BasicDeclType.t option;
      is_dot_call :
         bool;
      is_visible :
         bool;
  }


  let wrap context c_value = {
    decl = if is_null (getf (getf c_value CompletionItemStruct.decl) EntityStruct.node) then None else Some (BasicDeclType.wrap context ((getf c_value CompletionItemStruct.decl)));
    is_dot_call = (getf c_value CompletionItemStruct.is_dot_call);
    is_visible = (getf c_value CompletionItemStruct.is_visible);
  }

  let unwrap value =
    let c_value = make CompletionItemStruct.c_type in
    setf c_value
      CompletionItemStruct.decl
      (match value.decl with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    setf c_value
      CompletionItemStruct.is_dot_call
      (value.is_dot_call);
    setf c_value
      CompletionItemStruct.is_visible
      (value.is_visible);
    c_value


end

   
module DiscreteRange : sig
   
  
  type t = {
      low_bound :
         ExprType.t option;
      high_bound :
         ExprType.t option;
  }


  val wrap : AnalysisContext.t -> DiscreteRangeStruct.t structure -> t

  val unwrap : t -> DiscreteRangeStruct.t structure


end = struct
   
  
  type t = {
      low_bound :
         ExprType.t option;
      high_bound :
         ExprType.t option;
  }


  let wrap context c_value = {
    low_bound = if is_null (getf (getf c_value DiscreteRangeStruct.low_bound) EntityStruct.node) then None else Some (ExprType.wrap context ((getf c_value DiscreteRangeStruct.low_bound)));
    high_bound = if is_null (getf (getf c_value DiscreteRangeStruct.high_bound) EntityStruct.node) then None else Some (ExprType.wrap context ((getf c_value DiscreteRangeStruct.high_bound)));
  }

  let unwrap value =
    let c_value = make DiscreteRangeStruct.c_type in
    setf c_value
      DiscreteRangeStruct.low_bound
      (match value.low_bound with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    setf c_value
      DiscreteRangeStruct.high_bound
      (match value.high_bound with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    c_value


end

   
module DiscriminantValues : sig
   
  
  type t = {
      discriminant :
         IdentifierType.t option;
      values :
         AlternativesListType.t option;
  }


  val wrap : AnalysisContext.t -> DiscriminantValuesStruct.t structure -> t

  val unwrap : t -> DiscriminantValuesStruct.t structure


end = struct
   
  
  type t = {
      discriminant :
         IdentifierType.t option;
      values :
         AlternativesListType.t option;
  }


  let wrap context c_value = {
    discriminant = if is_null (getf (getf c_value DiscriminantValuesStruct.discriminant) EntityStruct.node) then None else Some (IdentifierType.wrap context ((getf c_value DiscriminantValuesStruct.discriminant)));
    values = if is_null (getf (getf c_value DiscriminantValuesStruct.values) EntityStruct.node) then None else Some (AlternativesListType.wrap context ((getf c_value DiscriminantValuesStruct.values)));
  }

  let unwrap value =
    let c_value = make DiscriminantValuesStruct.c_type in
    setf c_value
      DiscriminantValuesStruct.discriminant
      (match value.discriminant with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    setf c_value
      DiscriminantValuesStruct.values
      (match value.values with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    c_value


end

   
module TextType : sig
   
  type t = string

  val wrap : TextTypeStruct.t structure ptr -> t

  val unwrap : t -> TextTypeStruct.t structure ptr

end = struct
   
  type t = string

  let wrap c_value_ptr =
    let open Text in
    let open Camomile in
    let c_value = !@ c_value_ptr in
    let length = getf c_value TextTypeStruct.n in
    let items = c_value @. TextTypeStruct.items in
    (* We use Camomile to encode utf32 strings to an ocaml string *)
    let f i = UChar.chr_of_uint (Unsigned.UInt32.to_int !@(items +@ i)) in
    let result =
      UCS4Encoding.encode CharEncoding.utf8 (UCS4.init length f)
    in
    TextTypeStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let open Text in
    let open Camomile in
    let text = UCS4Encoding.decode CharEncoding.utf8 value in
    let result = TextTypeStruct.create (UCS4.length text) in
    let items = result |-> TextTypeStruct.items in
    let i = ref 0 in
    let f c =
      items +@ !i <-@ Unsigned.UInt32.of_int (UChar.code c);
      i := !i + 1
    in
    UCS4.iter f text;
    result


end

   
module DocAnnotation : sig
   
  
  type t = {
      key :
         string;
      value :
         string;
  }


  val wrap : DocAnnotationStruct.t structure -> t

  val unwrap : t -> DocAnnotationStruct.t structure


end = struct
   
  
  type t = {
      key :
         string;
      value :
         string;
  }


  let wrap c_value = {
    key = TextType.wrap ((getf c_value DocAnnotationStruct.key));
    value = TextType.wrap ((getf c_value DocAnnotationStruct.value));
  }

  let unwrap value =
    let c_value = make DocAnnotationStruct.c_type in
    setf c_value
      DocAnnotationStruct.key
      (TextType.unwrap (value.key));
    setf c_value
      DocAnnotationStruct.value
      (TextType.unwrap (value.value));
    c_value


end

   
module ParamActual : sig
   
  
  type t = {
      param :
         DefiningNameType.t option;
      actual :
         ExprType.t option;
  }


  val wrap : AnalysisContext.t -> ParamActualStruct.t structure -> t

  val unwrap : t -> ParamActualStruct.t structure


end = struct
   
  
  type t = {
      param :
         DefiningNameType.t option;
      actual :
         ExprType.t option;
  }


  let wrap context c_value = {
    param = if is_null (getf (getf c_value ParamActualStruct.param) EntityStruct.node) then None else Some (DefiningNameType.wrap context ((getf c_value ParamActualStruct.param)));
    actual = if is_null (getf (getf c_value ParamActualStruct.actual) EntityStruct.node) then None else Some (ExprType.wrap context ((getf c_value ParamActualStruct.actual)));
  }

  let unwrap value =
    let c_value = make ParamActualStruct.c_type in
    setf c_value
      ParamActualStruct.param
      (match value.param with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    setf c_value
      ParamActualStruct.actual
      (match value.actual with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    c_value


end

   
module RefResult : sig
   
  
  type t = {
      ref :
         BaseIdType.t option;
      kind :
         RefResultKind.t;
  }


  val wrap : AnalysisContext.t -> RefResultStruct.t structure -> t

  val unwrap : t -> RefResultStruct.t structure


end = struct
   
  
  type t = {
      ref :
         BaseIdType.t option;
      kind :
         RefResultKind.t;
  }


  let wrap context c_value = {
    ref = if is_null (getf (getf c_value RefResultStruct.ref) EntityStruct.node) then None else Some (BaseIdType.wrap context ((getf c_value RefResultStruct.ref)));
    kind = (getf c_value RefResultStruct.kind);
  }

  let unwrap value =
    let c_value = make RefResultStruct.c_type in
    setf c_value
      RefResultStruct.ref
      (match value.ref with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    setf c_value
      RefResultStruct.kind
      (value.kind);
    c_value


end

   
module RefdDecl : sig
   
  
  type t = {
      decl :
         BasicDeclType.t option;
      kind :
         RefResultKind.t;
  }


  val wrap : AnalysisContext.t -> RefdDeclStruct.t structure -> t

  val unwrap : t -> RefdDeclStruct.t structure


end = struct
   
  
  type t = {
      decl :
         BasicDeclType.t option;
      kind :
         RefResultKind.t;
  }


  let wrap context c_value = {
    decl = if is_null (getf (getf c_value RefdDeclStruct.decl) EntityStruct.node) then None else Some (BasicDeclType.wrap context ((getf c_value RefdDeclStruct.decl)));
    kind = (getf c_value RefdDeclStruct.kind);
  }

  let unwrap value =
    let c_value = make RefdDeclStruct.c_type in
    setf c_value
      RefdDeclStruct.decl
      (match value.decl with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    setf c_value
      RefdDeclStruct.kind
      (value.kind);
    c_value


end

   
module RefdDef : sig
   
  
  type t = {
      def_name :
         DefiningNameType.t option;
      kind :
         RefResultKind.t;
  }


  val wrap : AnalysisContext.t -> RefdDefStruct.t structure -> t

  val unwrap : t -> RefdDefStruct.t structure


end = struct
   
  
  type t = {
      def_name :
         DefiningNameType.t option;
      kind :
         RefResultKind.t;
  }


  let wrap context c_value = {
    def_name = if is_null (getf (getf c_value RefdDefStruct.def_name) EntityStruct.node) then None else Some (DefiningNameType.wrap context ((getf c_value RefdDefStruct.def_name)));
    kind = (getf c_value RefdDefStruct.kind);
  }

  let unwrap value =
    let c_value = make RefdDefStruct.c_type in
    setf c_value
      RefdDefStruct.def_name
      (match value.def_name with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    setf c_value
      RefdDefStruct.kind
      (value.kind);
    c_value


end

   
module BaseFormalParamDeclArray : sig
   
  type t = BaseFormalParamDeclType.t list

  val wrap : AnalysisContext.t -> AdaNodeArrayStruct.t structure ptr -> t

  val unwrap : t -> AdaNodeArrayStruct.t structure ptr

end = struct
   
  type t = BaseFormalParamDeclType.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value AdaNodeArrayStruct.n in
    let items = c_value @. AdaNodeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate EntityStruct.c_type (!@ (items +@ i))
      in
      BaseFormalParamDeclType.wrap context (!@ fresh)
    in
    let result = List.init length f in
    AdaNodeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = AdaNodeArrayStruct.create (List.length value) in
    let items = result |-> AdaNodeArrayStruct.items in
    let f i v =
      items +@ i <-@
        AdaNodeType.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module AdaNodeArray : sig
   
  type t = AdaNodeType.t list

  val wrap : AnalysisContext.t -> AdaNodeArrayStruct.t structure ptr -> t

  val unwrap : t -> AdaNodeArrayStruct.t structure ptr

end = struct
   
  type t = AdaNodeType.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value AdaNodeArrayStruct.n in
    let items = c_value @. AdaNodeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate EntityStruct.c_type (!@ (items +@ i))
      in
      AdaNodeType.wrap context (!@ fresh)
    in
    let result = List.init length f in
    AdaNodeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = AdaNodeArrayStruct.create (List.length value) in
    let items = result |-> AdaNodeArrayStruct.items in
    let f i v =
      items +@ i <-@
        AdaNodeType.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module DiscriminantValuesArray : sig
   
  type t = DiscriminantValues.t list

  val wrap : AnalysisContext.t -> DiscriminantValuesArrayStruct.t structure ptr -> t

  val unwrap : t -> DiscriminantValuesArrayStruct.t structure ptr

end = struct
   
  type t = DiscriminantValues.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value DiscriminantValuesArrayStruct.n in
    let items = c_value @. DiscriminantValuesArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate DiscriminantValuesStruct.c_type (!@ (items +@ i))
      in
      DiscriminantValues.wrap context (!@ fresh)
    in
    let result = List.init length f in
    DiscriminantValuesArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = DiscriminantValuesArrayStruct.create (List.length value) in
    let items = result |-> DiscriminantValuesArrayStruct.items in
    let f i v =
      items +@ i <-@
        DiscriminantValues.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module Shape : sig
   
  
  type t = {
      components :
         BaseFormalParamDeclType.t list;
      discriminants_values :
         DiscriminantValues.t list;
  }


  val wrap : AnalysisContext.t -> ShapeStruct.t structure -> t

  val unwrap : t -> ShapeStruct.t structure


end = struct
   
  
  type t = {
      components :
         BaseFormalParamDeclType.t list;
      discriminants_values :
         DiscriminantValues.t list;
  }


  let wrap context c_value = {
    components = BaseFormalParamDeclArray.wrap context ((getf c_value ShapeStruct.components));
    discriminants_values = DiscriminantValuesArray.wrap context ((getf c_value ShapeStruct.discriminants_values));
  }

  let unwrap value =
    let c_value = make ShapeStruct.c_type in
    setf c_value
      ShapeStruct.components
      (BaseFormalParamDeclArray.unwrap (value.components));
    setf c_value
      ShapeStruct.discriminants_values
      (DiscriminantValuesArray.unwrap (value.discriminants_values));
    c_value


end

   
module Substitution : sig
   
  
  type t = {
      from_decl :
         BasicDeclType.t option;
      to_value :
         BigInteger.t;
      value_type :
         BaseTypeDeclType.t option;
  }


  val wrap : AnalysisContext.t -> SubstitutionStruct.t structure -> t

  val unwrap : t -> SubstitutionStruct.t structure


end = struct
   
  
  type t = {
      from_decl :
         BasicDeclType.t option;
      to_value :
         BigInteger.t;
      value_type :
         BaseTypeDeclType.t option;
  }


  let wrap context c_value = {
    from_decl = if is_null (getf (getf c_value SubstitutionStruct.from_decl) EntityStruct.node) then None else Some (BasicDeclType.wrap context ((getf c_value SubstitutionStruct.from_decl)));
    to_value = BigInteger.wrap ((getf c_value SubstitutionStruct.to_value));
    value_type = if is_null (getf (getf c_value SubstitutionStruct.value_type) EntityStruct.node) then None else Some (BaseTypeDeclType.wrap context ((getf c_value SubstitutionStruct.value_type)));
  }

  let unwrap value =
    let c_value = make SubstitutionStruct.c_type in
    setf c_value
      SubstitutionStruct.from_decl
      (match value.from_decl with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    setf c_value
      SubstitutionStruct.to_value
      (BigInteger.unwrap (value.to_value));
    setf c_value
      SubstitutionStruct.value_type
      (match value.value_type with Some n -> AdaNodeType.unwrap (n) | None -> make EntityStruct.c_type);
    c_value


end

   
module DocAnnotationArray : sig
   
  type t = DocAnnotation.t list

  val wrap : DocAnnotationArrayStruct.t structure ptr -> t

  val unwrap : t -> DocAnnotationArrayStruct.t structure ptr

end = struct
   
  type t = DocAnnotation.t list

  let wrap c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value DocAnnotationArrayStruct.n in
    let items = c_value @. DocAnnotationArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate DocAnnotationStruct.c_type (!@ (items +@ i))
      in
      DocAnnotation.wrap (!@ fresh)
    in
    let result = List.init length f in
    DocAnnotationArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = DocAnnotationArrayStruct.create (List.length value) in
    let items = result |-> DocAnnotationArrayStruct.items in
    let f i v =
      items +@ i <-@
        DocAnnotation.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module BaseTypeDeclArray : sig
   
  type t = BaseTypeDeclType.t list

  val wrap : AnalysisContext.t -> AdaNodeArrayStruct.t structure ptr -> t

  val unwrap : t -> AdaNodeArrayStruct.t structure ptr

end = struct
   
  type t = BaseTypeDeclType.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value AdaNodeArrayStruct.n in
    let items = c_value @. AdaNodeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate EntityStruct.c_type (!@ (items +@ i))
      in
      BaseTypeDeclType.wrap context (!@ fresh)
    in
    let result = List.init length f in
    AdaNodeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = AdaNodeArrayStruct.create (List.length value) in
    let items = result |-> AdaNodeArrayStruct.items in
    let f i v =
      items +@ i <-@
        AdaNodeType.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module BasicDeclArray : sig
   
  type t = BasicDeclType.t list

  val wrap : AnalysisContext.t -> AdaNodeArrayStruct.t structure ptr -> t

  val unwrap : t -> AdaNodeArrayStruct.t structure ptr

end = struct
   
  type t = BasicDeclType.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value AdaNodeArrayStruct.n in
    let items = c_value @. AdaNodeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate EntityStruct.c_type (!@ (items +@ i))
      in
      BasicDeclType.wrap context (!@ fresh)
    in
    let result = List.init length f in
    AdaNodeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = AdaNodeArrayStruct.create (List.length value) in
    let items = result |-> AdaNodeArrayStruct.items in
    let f i v =
      items +@ i <-@
        AdaNodeType.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module CompilationUnitArray : sig
   
  type t = CompilationUnitType.t list

  val wrap : AnalysisContext.t -> AdaNodeArrayStruct.t structure ptr -> t

  val unwrap : t -> AdaNodeArrayStruct.t structure ptr

end = struct
   
  type t = CompilationUnitType.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value AdaNodeArrayStruct.n in
    let items = c_value @. AdaNodeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate EntityStruct.c_type (!@ (items +@ i))
      in
      CompilationUnitType.wrap context (!@ fresh)
    in
    let result = List.init length f in
    AdaNodeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = AdaNodeArrayStruct.create (List.length value) in
    let items = result |-> AdaNodeArrayStruct.items in
    let f i v =
      items +@ i <-@
        AdaNodeType.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module DefiningNameArray : sig
   
  type t = DefiningNameType.t list

  val wrap : AnalysisContext.t -> AdaNodeArrayStruct.t structure ptr -> t

  val unwrap : t -> AdaNodeArrayStruct.t structure ptr

end = struct
   
  type t = DefiningNameType.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value AdaNodeArrayStruct.n in
    let items = c_value @. AdaNodeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate EntityStruct.c_type (!@ (items +@ i))
      in
      DefiningNameType.wrap context (!@ fresh)
    in
    let result = List.init length f in
    AdaNodeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = AdaNodeArrayStruct.create (List.length value) in
    let items = result |-> AdaNodeArrayStruct.items in
    let f i v =
      items +@ i <-@
        AdaNodeType.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module ExprArray : sig
   
  type t = ExprType.t list

  val wrap : AnalysisContext.t -> AdaNodeArrayStruct.t structure ptr -> t

  val unwrap : t -> AdaNodeArrayStruct.t structure ptr

end = struct
   
  type t = ExprType.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value AdaNodeArrayStruct.n in
    let items = c_value @. AdaNodeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate EntityStruct.c_type (!@ (items +@ i))
      in
      ExprType.wrap context (!@ fresh)
    in
    let result = List.init length f in
    AdaNodeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = AdaNodeArrayStruct.create (List.length value) in
    let items = result |-> AdaNodeArrayStruct.items in
    let f i v =
      items +@ i <-@
        AdaNodeType.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module GenericInstantiationArray : sig
   
  type t = GenericInstantiationType.t list

  val wrap : AnalysisContext.t -> AdaNodeArrayStruct.t structure ptr -> t

  val unwrap : t -> AdaNodeArrayStruct.t structure ptr

end = struct
   
  type t = GenericInstantiationType.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value AdaNodeArrayStruct.n in
    let items = c_value @. AdaNodeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate EntityStruct.c_type (!@ (items +@ i))
      in
      GenericInstantiationType.wrap context (!@ fresh)
    in
    let result = List.init length f in
    AdaNodeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = AdaNodeArrayStruct.create (List.length value) in
    let items = result |-> AdaNodeArrayStruct.items in
    let f i v =
      items +@ i <-@
        AdaNodeType.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module ParamSpecArray : sig
   
  type t = ParamSpecType.t list

  val wrap : AnalysisContext.t -> AdaNodeArrayStruct.t structure ptr -> t

  val unwrap : t -> AdaNodeArrayStruct.t structure ptr

end = struct
   
  type t = ParamSpecType.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value AdaNodeArrayStruct.n in
    let items = c_value @. AdaNodeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate EntityStruct.c_type (!@ (items +@ i))
      in
      ParamSpecType.wrap context (!@ fresh)
    in
    let result = List.init length f in
    AdaNodeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = AdaNodeArrayStruct.create (List.length value) in
    let items = result |-> AdaNodeArrayStruct.items in
    let f i v =
      items +@ i <-@
        AdaNodeType.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module TypeDeclArray : sig
   
  type t = TypeDeclType.t list

  val wrap : AnalysisContext.t -> AdaNodeArrayStruct.t structure ptr -> t

  val unwrap : t -> AdaNodeArrayStruct.t structure ptr

end = struct
   
  type t = TypeDeclType.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value AdaNodeArrayStruct.n in
    let items = c_value @. AdaNodeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate EntityStruct.c_type (!@ (items +@ i))
      in
      TypeDeclType.wrap context (!@ fresh)
    in
    let result = List.init length f in
    AdaNodeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = AdaNodeArrayStruct.create (List.length value) in
    let items = result |-> AdaNodeArrayStruct.items in
    let f i v =
      items +@ i <-@
        AdaNodeType.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module ParamActualArray : sig
   
  type t = ParamActual.t list

  val wrap : AnalysisContext.t -> ParamActualArrayStruct.t structure ptr -> t

  val unwrap : t -> ParamActualArrayStruct.t structure ptr

end = struct
   
  type t = ParamActual.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value ParamActualArrayStruct.n in
    let items = c_value @. ParamActualArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate ParamActualStruct.c_type (!@ (items +@ i))
      in
      ParamActual.wrap context (!@ fresh)
    in
    let result = List.init length f in
    ParamActualArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = ParamActualArrayStruct.create (List.length value) in
    let items = result |-> ParamActualArrayStruct.items in
    let f i v =
      items +@ i <-@
        ParamActual.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module RefResultArray : sig
   
  type t = RefResult.t list

  val wrap : AnalysisContext.t -> RefResultArrayStruct.t structure ptr -> t

  val unwrap : t -> RefResultArrayStruct.t structure ptr

end = struct
   
  type t = RefResult.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value RefResultArrayStruct.n in
    let items = c_value @. RefResultArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate RefResultStruct.c_type (!@ (items +@ i))
      in
      RefResult.wrap context (!@ fresh)
    in
    let result = List.init length f in
    RefResultArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = RefResultArrayStruct.create (List.length value) in
    let items = result |-> RefResultArrayStruct.items in
    let f i v =
      items +@ i <-@
        RefResult.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module ShapeArray : sig
   
  type t = Shape.t list

  val wrap : AnalysisContext.t -> ShapeArrayStruct.t structure ptr -> t

  val unwrap : t -> ShapeArrayStruct.t structure ptr

end = struct
   
  type t = Shape.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value ShapeArrayStruct.n in
    let items = c_value @. ShapeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate ShapeStruct.c_type (!@ (items +@ i))
      in
      Shape.wrap context (!@ fresh)
    in
    let result = List.init length f in
    ShapeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = ShapeArrayStruct.create (List.length value) in
    let items = result |-> ShapeArrayStruct.items in
    let f i v =
      items +@ i <-@
        Shape.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module SubstitutionArray : sig
   
  type t = Substitution.t list

  val wrap : AnalysisContext.t -> SubstitutionArrayStruct.t structure ptr -> t

  val unwrap : t -> SubstitutionArrayStruct.t structure ptr

end = struct
   
  type t = Substitution.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value SubstitutionArrayStruct.n in
    let items = c_value @. SubstitutionArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate SubstitutionStruct.c_type (!@ (items +@ i))
      in
      Substitution.wrap context (!@ fresh)
    in
    let result = List.init length f in
    SubstitutionArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = SubstitutionArrayStruct.create (List.length value) in
    let items = result |-> SubstitutionArrayStruct.items in
    let f i v =
      items +@ i <-@
        Substitution.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module AnalysisUnitArray : sig
   
  type t = AnalysisUnit.t list

  val wrap : AnalysisContext.t -> AnalysisUnitArrayStruct.t structure ptr -> t

  val unwrap : t -> AnalysisUnitArrayStruct.t structure ptr

end = struct
   
  type t = AnalysisUnit.t list

  let wrap context c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value AnalysisUnitArrayStruct.n in
    let items = c_value @. AnalysisUnitArrayStruct.items in
    let f i =
      AnalysisUnit.wrap context (!@ (items +@ i))
    in
    let result = List.init length f in
    AnalysisUnitArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap value =
    let result = AnalysisUnitArrayStruct.create (List.length value) in
    let items = result |-> AnalysisUnitArrayStruct.items in
    let f i v =
      items +@ i <-@
        AnalysisUnit.unwrap (v)
    in
    List.iteri f value;
    result


end

   
module UnboundedTextTypeArray : sig
   
  type t = string list

  val wrap : UnboundedTextTypeArrayStruct.t structure ptr -> t

  val unwrap : AnalysisContext.t -> t -> UnboundedTextTypeArrayStruct.t structure ptr

end = struct
   
  type t = string list

  let wrap c_value_ptr =
    let c_value = !@ c_value_ptr in
    let length = getf c_value UnboundedTextTypeArrayStruct.n in
    let items = c_value @. UnboundedTextTypeArrayStruct.items in
    let f i =
      (* we want to allocate a fresh value for a record, otherwize, the c value
       * will still point to the memory at array location *)
      let fresh =
        allocate Symbol.c_type (!@ (items +@ i))
      in
      Symbol.wrap (!@ fresh)
    in
    let result = List.init length f in
    UnboundedTextTypeArrayStruct.dec_ref c_value_ptr;
    result

  let unwrap context value =
    let result = UnboundedTextTypeArrayStruct.create (List.length value) in
    let items = result |-> UnboundedTextTypeArrayStruct.items in
    let c_context = context.AnalysisContext.c_value in
    let f i v =
      items +@ i <-@
        Symbol.unwrap c_context (v)
    in
    List.iteri f value;
    result


end


let context node =
  (* Given any node, extract the context field *)
  match (node :> AdaNodeType.t) with
  | `AbortAbsent fields -> fields.context
  | `AbortPresent fields -> fields.context
  | `AbstractAbsent fields -> fields.context
  | `AbstractPresent fields -> fields.context
  | `AdaNodeList fields -> fields.context
  | `AbstractStateDeclList fields -> fields.context
  | `AlternativesList fields -> fields.context
  | `ConstraintList fields -> fields.context
  | `DeclList fields -> fields.context
  | `StmtList fields -> fields.context
  | `AspectAssocList fields -> fields.context
  | `BaseAssocList fields -> fields.context
  | `AssocList fields -> fields.context
  | `BasicDeclList fields -> fields.context
  | `CaseExprAlternativeList fields -> fields.context
  | `CaseStmtAlternativeList fields -> fields.context
  | `CompilationUnitList fields -> fields.context
  | `ContractCaseAssocList fields -> fields.context
  | `DefiningNameList fields -> fields.context
  | `DiscriminantSpecList fields -> fields.context
  | `ElsifExprPartList fields -> fields.context
  | `ElsifStmtPartList fields -> fields.context
  | `EnumLiteralDeclList fields -> fields.context
  | `ExprAlternativesList fields -> fields.context
  | `DiscriminantChoiceList fields -> fields.context
  | `NameList fields -> fields.context
  | `ParentList fields -> fields.context
  | `ParamSpecList fields -> fields.context
  | `PragmaNodeList fields -> fields.context
  | `SelectWhenPartList fields -> fields.context
  | `UnconstrainedArrayIndexList fields -> fields.context
  | `VariantList fields -> fields.context
  | `AliasedAbsent fields -> fields.context
  | `AliasedPresent fields -> fields.context
  | `AllAbsent fields -> fields.context
  | `AllPresent fields -> fields.context
  | `ConstrainedArrayIndices fields -> fields.context
  | `UnconstrainedArrayIndices fields -> fields.context
  | `AspectAssoc fields -> fields.context
  | `AtClause fields -> fields.context
  | `AttributeDefClause fields -> fields.context
  | `EnumRepClause fields -> fields.context
  | `RecordRepClause fields -> fields.context
  | `AspectSpec fields -> fields.context
  | `ContractCaseAssoc fields -> fields.context
  | `PragmaArgumentAssoc fields -> fields.context
  | `EntrySpec fields -> fields.context
  | `EnumSubpSpec fields -> fields.context
  | `SubpSpec fields -> fields.context
  | `ComponentList fields -> fields.context
  | `KnownDiscriminantPart fields -> fields.context
  | `UnknownDiscriminantPart fields -> fields.context
  | `EntryCompletionFormalParams fields -> fields.context
  | `GenericFormalPart fields -> fields.context
  | `NullRecordDef fields -> fields.context
  | `RecordDef fields -> fields.context
  | `AggregateAssoc fields -> fields.context
  | `MultiDimArrayAssoc fields -> fields.context
  | `DiscriminantAssoc fields -> fields.context
  | `IteratedAssoc fields -> fields.context
  | `ParamAssoc fields -> fields.context
  | `AbstractStateDecl fields -> fields.context
  | `AnonymousExprDecl fields -> fields.context
  | `ComponentDecl fields -> fields.context
  | `DiscriminantSpec fields -> fields.context
  | `GenericFormalObjDecl fields -> fields.context
  | `GenericFormalPackage fields -> fields.context
  | `GenericFormalSubpDecl fields -> fields.context
  | `GenericFormalTypeDecl fields -> fields.context
  | `ParamSpec fields -> fields.context
  | `GenericPackageInternal fields -> fields.context
  | `PackageDecl fields -> fields.context
  | `DiscreteBaseSubtypeDecl fields -> fields.context
  | `SubtypeDecl fields -> fields.context
  | `ClasswideTypeDecl fields -> fields.context
  | `IncompleteTypeDecl fields -> fields.context
  | `IncompleteTaggedTypeDecl fields -> fields.context
  | `ProtectedTypeDecl fields -> fields.context
  | `TaskTypeDecl fields -> fields.context
  | `SingleTaskTypeDecl fields -> fields.context
  | `TypeDecl fields -> fields.context
  | `AnonymousTypeDecl fields -> fields.context
  | `SynthAnonymousTypeDecl fields -> fields.context
  | `AbstractSubpDecl fields -> fields.context
  | `AbstractFormalSubpDecl fields -> fields.context
  | `ConcreteFormalSubpDecl fields -> fields.context
  | `SubpDecl fields -> fields.context
  | `EntryDecl fields -> fields.context
  | `EnumLiteralDecl fields -> fields.context
  | `GenericSubpInternal fields -> fields.context
  | `ExprFunction fields -> fields.context
  | `NullSubpDecl fields -> fields.context
  | `SubpBody fields -> fields.context
  | `SubpRenamingDecl fields -> fields.context
  | `PackageBodyStub fields -> fields.context
  | `ProtectedBodyStub fields -> fields.context
  | `SubpBodyStub fields -> fields.context
  | `TaskBodyStub fields -> fields.context
  | `EntryBody fields -> fields.context
  | `PackageBody fields -> fields.context
  | `ProtectedBody fields -> fields.context
  | `TaskBody fields -> fields.context
  | `EntryIndexSpec fields -> fields.context
  | `ErrorDecl fields -> fields.context
  | `ExceptionDecl fields -> fields.context
  | `ExceptionHandler fields -> fields.context
  | `ForLoopVarDecl fields -> fields.context
  | `GenericPackageDecl fields -> fields.context
  | `GenericSubpDecl fields -> fields.context
  | `GenericPackageInstantiation fields -> fields.context
  | `GenericSubpInstantiation fields -> fields.context
  | `GenericPackageRenamingDecl fields -> fields.context
  | `GenericSubpRenamingDecl fields -> fields.context
  | `LabelDecl fields -> fields.context
  | `NamedStmtDecl fields -> fields.context
  | `NumberDecl fields -> fields.context
  | `ObjectDecl fields -> fields.context
  | `ExtendedReturnStmtObjectDecl fields -> fields.context
  | `PackageRenamingDecl fields -> fields.context
  | `SingleProtectedDecl fields -> fields.context
  | `SingleTaskDecl fields -> fields.context
  | `CaseStmtAlternative fields -> fields.context
  | `CompilationUnit fields -> fields.context
  | `ComponentClause fields -> fields.context
  | `ComponentDef fields -> fields.context
  | `ConstantAbsent fields -> fields.context
  | `ConstantPresent fields -> fields.context
  | `DeltaConstraint fields -> fields.context
  | `DigitsConstraint fields -> fields.context
  | `DiscriminantConstraint fields -> fields.context
  | `IndexConstraint fields -> fields.context
  | `RangeConstraint fields -> fields.context
  | `DeclarativePart fields -> fields.context
  | `PrivatePart fields -> fields.context
  | `PublicPart fields -> fields.context
  | `ElsifExprPart fields -> fields.context
  | `ElsifStmtPart fields -> fields.context
  | `AbstractStateDeclExpr fields -> fields.context
  | `Allocator fields -> fields.context
  | `Aggregate fields -> fields.context
  | `BracketAggregate fields -> fields.context
  | `DeltaAggregate fields -> fields.context
  | `BracketDeltaAggregate fields -> fields.context
  | `NullRecordAggregate fields -> fields.context
  | `BinOp fields -> fields.context
  | `RelationOp fields -> fields.context
  | `BoxExpr fields -> fields.context
  | `CaseExprAlternative fields -> fields.context
  | `CaseExpr fields -> fields.context
  | `IfExpr fields -> fields.context
  | `ContractCases fields -> fields.context
  | `DeclExpr fields -> fields.context
  | `MembershipExpr fields -> fields.context
  | `AttributeRef fields -> fields.context
  | `UpdateAttributeRef fields -> fields.context
  | `CallExpr fields -> fields.context
  | `DefiningName fields -> fields.context
  | `DiscreteSubtypeName fields -> fields.context
  | `DottedName fields -> fields.context
  | `EndName fields -> fields.context
  | `ExplicitDeref fields -> fields.context
  | `QualExpr fields -> fields.context
  | `CharLiteral fields -> fields.context
  | `Identifier fields -> fields.context
  | `OpAbs fields -> fields.context
  | `OpAnd fields -> fields.context
  | `OpAndThen fields -> fields.context
  | `OpConcat fields -> fields.context
  | `OpDiv fields -> fields.context
  | `OpDoubleDot fields -> fields.context
  | `OpEq fields -> fields.context
  | `OpGt fields -> fields.context
  | `OpGte fields -> fields.context
  | `OpIn fields -> fields.context
  | `OpLt fields -> fields.context
  | `OpLte fields -> fields.context
  | `OpMinus fields -> fields.context
  | `OpMod fields -> fields.context
  | `OpMult fields -> fields.context
  | `OpNeq fields -> fields.context
  | `OpNot fields -> fields.context
  | `OpNotIn fields -> fields.context
  | `OpOr fields -> fields.context
  | `OpOrElse fields -> fields.context
  | `OpPlus fields -> fields.context
  | `OpPow fields -> fields.context
  | `OpRem fields -> fields.context
  | `OpXor fields -> fields.context
  | `StringLiteral fields -> fields.context
  | `NullLiteral fields -> fields.context
  | `IntLiteral fields -> fields.context
  | `RealLiteral fields -> fields.context
  | `TargetName fields -> fields.context
  | `ParenExpr fields -> fields.context
  | `QuantifiedExpr fields -> fields.context
  | `RaiseExpr fields -> fields.context
  | `UnOp fields -> fields.context
  | `HandledStmts fields -> fields.context
  | `InterfaceKindLimited fields -> fields.context
  | `InterfaceKindProtected fields -> fields.context
  | `InterfaceKindSynchronized fields -> fields.context
  | `InterfaceKindTask fields -> fields.context
  | `IterTypeIn fields -> fields.context
  | `IterTypeOf fields -> fields.context
  | `LibraryItem fields -> fields.context
  | `LimitedAbsent fields -> fields.context
  | `LimitedPresent fields -> fields.context
  | `ForLoopSpec fields -> fields.context
  | `WhileLoopSpec fields -> fields.context
  | `ModeDefault fields -> fields.context
  | `ModeIn fields -> fields.context
  | `ModeInOut fields -> fields.context
  | `ModeOut fields -> fields.context
  | `MultiAbstractStateDecl fields -> fields.context
  | `NotNullAbsent fields -> fields.context
  | `NotNullPresent fields -> fields.context
  | `NullComponentDecl fields -> fields.context
  | `OthersDesignator fields -> fields.context
  | `OverridingNotOverriding fields -> fields.context
  | `OverridingOverriding fields -> fields.context
  | `OverridingUnspecified fields -> fields.context
  | `Params fields -> fields.context
  | `ParenAbstractStateDecl fields -> fields.context
  | `PragmaNode fields -> fields.context
  | `PrivateAbsent fields -> fields.context
  | `PrivatePresent fields -> fields.context
  | `ProtectedDef fields -> fields.context
  | `ProtectedAbsent fields -> fields.context
  | `ProtectedPresent fields -> fields.context
  | `QuantifierAll fields -> fields.context
  | `QuantifierSome fields -> fields.context
  | `RangeSpec fields -> fields.context
  | `RenamingClause fields -> fields.context
  | `SyntheticRenamingClause fields -> fields.context
  | `ReverseAbsent fields -> fields.context
  | `ReversePresent fields -> fields.context
  | `SelectWhenPart fields -> fields.context
  | `AcceptStmt fields -> fields.context
  | `AcceptStmtWithStmts fields -> fields.context
  | `ForLoopStmt fields -> fields.context
  | `LoopStmt fields -> fields.context
  | `WhileLoopStmt fields -> fields.context
  | `BeginBlock fields -> fields.context
  | `DeclBlock fields -> fields.context
  | `CaseStmt fields -> fields.context
  | `ExtendedReturnStmt fields -> fields.context
  | `IfStmt fields -> fields.context
  | `NamedStmt fields -> fields.context
  | `SelectStmt fields -> fields.context
  | `ErrorStmt fields -> fields.context
  | `AbortStmt fields -> fields.context
  | `AssignStmt fields -> fields.context
  | `CallStmt fields -> fields.context
  | `DelayStmt fields -> fields.context
  | `ExitStmt fields -> fields.context
  | `GotoStmt fields -> fields.context
  | `Label fields -> fields.context
  | `NullStmt fields -> fields.context
  | `RaiseStmt fields -> fields.context
  | `RequeueStmt fields -> fields.context
  | `ReturnStmt fields -> fields.context
  | `TerminateAlternative fields -> fields.context
  | `SubpKindFunction fields -> fields.context
  | `SubpKindProcedure fields -> fields.context
  | `Subunit fields -> fields.context
  | `SynchronizedAbsent fields -> fields.context
  | `SynchronizedPresent fields -> fields.context
  | `TaggedAbsent fields -> fields.context
  | `TaggedPresent fields -> fields.context
  | `TaskDef fields -> fields.context
  | `AccessToSubpDef fields -> fields.context
  | `AnonymousTypeAccessDef fields -> fields.context
  | `TypeAccessDef fields -> fields.context
  | `ArrayTypeDef fields -> fields.context
  | `DerivedTypeDef fields -> fields.context
  | `EnumTypeDef fields -> fields.context
  | `FormalDiscreteTypeDef fields -> fields.context
  | `InterfaceTypeDef fields -> fields.context
  | `ModIntTypeDef fields -> fields.context
  | `PrivateTypeDef fields -> fields.context
  | `DecimalFixedPointDef fields -> fields.context
  | `FloatingPointDef fields -> fields.context
  | `OrdinaryFixedPointDef fields -> fields.context
  | `RecordTypeDef fields -> fields.context
  | `SignedIntTypeDef fields -> fields.context
  | `AnonymousType fields -> fields.context
  | `EnumLitSynthTypeExpr fields -> fields.context
  | `SubtypeIndication fields -> fields.context
  | `ConstrainedSubtypeIndication fields -> fields.context
  | `DiscreteSubtypeIndication fields -> fields.context
  | `UnconstrainedArrayIndex fields -> fields.context
  | `UntilAbsent fields -> fields.context
  | `UntilPresent fields -> fields.context
  | `UsePackageClause fields -> fields.context
  | `UseTypeClause fields -> fields.context
  | `Variant fields -> fields.context
  | `VariantPart fields -> fields.context
  | `WithClause fields -> fields.context
  | `WithPrivateAbsent fields -> fields.context
  | `WithPrivatePresent fields -> fields.context

type _ node =
  | AdaNode :
      AdaNodeType.t node
  | AbortNode :
      AbortNodeType.t node
  | AbortAbsent :
      AbortAbsentType.t node
  | AbortPresent :
      AbortPresentType.t node
  | AbstractNode :
      AbstractNodeType.t node
  | AbstractAbsent :
      AbstractAbsentType.t node
  | AbstractPresent :
      AbstractPresentType.t node
  | AdaList :
      AdaListType.t node
  | AdaNodeList :
      AdaNodeListType.t node
  | AbstractStateDeclList :
      AbstractStateDeclListType.t node
  | AlternativesList :
      AlternativesListType.t node
  | ConstraintList :
      ConstraintListType.t node
  | DeclList :
      DeclListType.t node
  | StmtList :
      StmtListType.t node
  | AspectAssocList :
      AspectAssocListType.t node
  | BaseAssocList :
      BaseAssocListType.t node
  | BasicAssocList :
      BasicAssocListType.t node
  | AssocList :
      AssocListType.t node
  | BasicDeclList :
      BasicDeclListType.t node
  | CaseExprAlternativeList :
      CaseExprAlternativeListType.t node
  | CaseStmtAlternativeList :
      CaseStmtAlternativeListType.t node
  | CompilationUnitList :
      CompilationUnitListType.t node
  | ContractCaseAssocList :
      ContractCaseAssocListType.t node
  | DefiningNameList :
      DefiningNameListType.t node
  | DiscriminantSpecList :
      DiscriminantSpecListType.t node
  | ElsifExprPartList :
      ElsifExprPartListType.t node
  | ElsifStmtPartList :
      ElsifStmtPartListType.t node
  | EnumLiteralDeclList :
      EnumLiteralDeclListType.t node
  | ExprList :
      ExprListType.t node
  | ExprAlternativesList :
      ExprAlternativesListType.t node
  | IdentifierList :
      IdentifierListType.t node
  | DiscriminantChoiceList :
      DiscriminantChoiceListType.t node
  | NameList :
      NameListType.t node
  | ParentList :
      ParentListType.t node
  | ParamSpecList :
      ParamSpecListType.t node
  | PragmaNodeList :
      PragmaNodeListType.t node
  | SelectWhenPartList :
      SelectWhenPartListType.t node
  | UnconstrainedArrayIndexList :
      UnconstrainedArrayIndexListType.t node
  | VariantList :
      VariantListType.t node
  | AliasedNode :
      AliasedNodeType.t node
  | AliasedAbsent :
      AliasedAbsentType.t node
  | AliasedPresent :
      AliasedPresentType.t node
  | AllNode :
      AllNodeType.t node
  | AllAbsent :
      AllAbsentType.t node
  | AllPresent :
      AllPresentType.t node
  | ArrayIndices :
      ArrayIndicesType.t node
  | ConstrainedArrayIndices :
      ConstrainedArrayIndicesType.t node
  | UnconstrainedArrayIndices :
      UnconstrainedArrayIndicesType.t node
  | AspectAssoc :
      AspectAssocType.t node
  | AspectClause :
      AspectClauseType.t node
  | AtClause :
      AtClauseType.t node
  | AttributeDefClause :
      AttributeDefClauseType.t node
  | EnumRepClause :
      EnumRepClauseType.t node
  | RecordRepClause :
      RecordRepClauseType.t node
  | AspectSpec :
      AspectSpecType.t node
  | BaseAssoc :
      BaseAssocType.t node
  | ContractCaseAssoc :
      ContractCaseAssocType.t node
  | PragmaArgumentAssoc :
      PragmaArgumentAssocType.t node
  | BaseFormalParamHolder :
      BaseFormalParamHolderType.t node
  | BaseSubpSpec :
      BaseSubpSpecType.t node
  | EntrySpec :
      EntrySpecType.t node
  | EnumSubpSpec :
      EnumSubpSpecType.t node
  | SubpSpec :
      SubpSpecType.t node
  | ComponentList :
      ComponentListType.t node
  | DiscriminantPart :
      DiscriminantPartType.t node
  | KnownDiscriminantPart :
      KnownDiscriminantPartType.t node
  | UnknownDiscriminantPart :
      UnknownDiscriminantPartType.t node
  | EntryCompletionFormalParams :
      EntryCompletionFormalParamsType.t node
  | GenericFormalPart :
      GenericFormalPartType.t node
  | BaseRecordDef :
      BaseRecordDefType.t node
  | NullRecordDef :
      NullRecordDefType.t node
  | RecordDef :
      RecordDefType.t node
  | BasicAssoc :
      BasicAssocType.t node
  | AggregateAssoc :
      AggregateAssocType.t node
  | MultiDimArrayAssoc :
      MultiDimArrayAssocType.t node
  | DiscriminantAssoc :
      DiscriminantAssocType.t node
  | IteratedAssoc :
      IteratedAssocType.t node
  | ParamAssoc :
      ParamAssocType.t node
  | BasicDecl :
      BasicDeclType.t node
  | AbstractStateDecl :
      AbstractStateDeclType.t node
  | AnonymousExprDecl :
      AnonymousExprDeclType.t node
  | BaseFormalParamDecl :
      BaseFormalParamDeclType.t node
  | ComponentDecl :
      ComponentDeclType.t node
  | DiscriminantSpec :
      DiscriminantSpecType.t node
  | GenericFormal :
      GenericFormalType.t node
  | GenericFormalObjDecl :
      GenericFormalObjDeclType.t node
  | GenericFormalPackage :
      GenericFormalPackageType.t node
  | GenericFormalSubpDecl :
      GenericFormalSubpDeclType.t node
  | GenericFormalTypeDecl :
      GenericFormalTypeDeclType.t node
  | ParamSpec :
      ParamSpecType.t node
  | BasePackageDecl :
      BasePackageDeclType.t node
  | GenericPackageInternal :
      GenericPackageInternalType.t node
  | PackageDecl :
      PackageDeclType.t node
  | BaseTypeDecl :
      BaseTypeDeclType.t node
  | BaseSubtypeDecl :
      BaseSubtypeDeclType.t node
  | DiscreteBaseSubtypeDecl :
      DiscreteBaseSubtypeDeclType.t node
  | SubtypeDecl :
      SubtypeDeclType.t node
  | ClasswideTypeDecl :
      ClasswideTypeDeclType.t node
  | IncompleteTypeDecl :
      IncompleteTypeDeclType.t node
  | IncompleteTaggedTypeDecl :
      IncompleteTaggedTypeDeclType.t node
  | ProtectedTypeDecl :
      ProtectedTypeDeclType.t node
  | TaskTypeDecl :
      TaskTypeDeclType.t node
  | SingleTaskTypeDecl :
      SingleTaskTypeDeclType.t node
  | TypeDecl :
      TypeDeclType.t node
  | AnonymousTypeDecl :
      AnonymousTypeDeclType.t node
  | SynthAnonymousTypeDecl :
      SynthAnonymousTypeDeclType.t node
  | BasicSubpDecl :
      BasicSubpDeclType.t node
  | ClassicSubpDecl :
      ClassicSubpDeclType.t node
  | AbstractSubpDecl :
      AbstractSubpDeclType.t node
  | FormalSubpDecl :
      FormalSubpDeclType.t node
  | AbstractFormalSubpDecl :
      AbstractFormalSubpDeclType.t node
  | ConcreteFormalSubpDecl :
      ConcreteFormalSubpDeclType.t node
  | SubpDecl :
      SubpDeclType.t node
  | EntryDecl :
      EntryDeclType.t node
  | EnumLiteralDecl :
      EnumLiteralDeclType.t node
  | GenericSubpInternal :
      GenericSubpInternalType.t node
  | BodyNode :
      BodyNodeType.t node
  | BaseSubpBody :
      BaseSubpBodyType.t node
  | ExprFunction :
      ExprFunctionType.t node
  | NullSubpDecl :
      NullSubpDeclType.t node
  | SubpBody :
      SubpBodyType.t node
  | SubpRenamingDecl :
      SubpRenamingDeclType.t node
  | BodyStub :
      BodyStubType.t node
  | PackageBodyStub :
      PackageBodyStubType.t node
  | ProtectedBodyStub :
      ProtectedBodyStubType.t node
  | SubpBodyStub :
      SubpBodyStubType.t node
  | TaskBodyStub :
      TaskBodyStubType.t node
  | EntryBody :
      EntryBodyType.t node
  | PackageBody :
      PackageBodyType.t node
  | ProtectedBody :
      ProtectedBodyType.t node
  | TaskBody :
      TaskBodyType.t node
  | EntryIndexSpec :
      EntryIndexSpecType.t node
  | ErrorDecl :
      ErrorDeclType.t node
  | ExceptionDecl :
      ExceptionDeclType.t node
  | ExceptionHandler :
      ExceptionHandlerType.t node
  | ForLoopVarDecl :
      ForLoopVarDeclType.t node
  | GenericDecl :
      GenericDeclType.t node
  | GenericPackageDecl :
      GenericPackageDeclType.t node
  | GenericSubpDecl :
      GenericSubpDeclType.t node
  | GenericInstantiation :
      GenericInstantiationType.t node
  | GenericPackageInstantiation :
      GenericPackageInstantiationType.t node
  | GenericSubpInstantiation :
      GenericSubpInstantiationType.t node
  | GenericRenamingDecl :
      GenericRenamingDeclType.t node
  | GenericPackageRenamingDecl :
      GenericPackageRenamingDeclType.t node
  | GenericSubpRenamingDecl :
      GenericSubpRenamingDeclType.t node
  | LabelDecl :
      LabelDeclType.t node
  | NamedStmtDecl :
      NamedStmtDeclType.t node
  | NumberDecl :
      NumberDeclType.t node
  | ObjectDecl :
      ObjectDeclType.t node
  | ExtendedReturnStmtObjectDecl :
      ExtendedReturnStmtObjectDeclType.t node
  | PackageRenamingDecl :
      PackageRenamingDeclType.t node
  | SingleProtectedDecl :
      SingleProtectedDeclType.t node
  | SingleTaskDecl :
      SingleTaskDeclType.t node
  | CaseStmtAlternative :
      CaseStmtAlternativeType.t node
  | CompilationUnit :
      CompilationUnitType.t node
  | ComponentClause :
      ComponentClauseType.t node
  | ComponentDef :
      ComponentDefType.t node
  | ConstantNode :
      ConstantNodeType.t node
  | ConstantAbsent :
      ConstantAbsentType.t node
  | ConstantPresent :
      ConstantPresentType.t node
  | Constraint :
      ConstraintType.t node
  | DeltaConstraint :
      DeltaConstraintType.t node
  | DigitsConstraint :
      DigitsConstraintType.t node
  | DiscriminantConstraint :
      DiscriminantConstraintType.t node
  | IndexConstraint :
      IndexConstraintType.t node
  | RangeConstraint :
      RangeConstraintType.t node
  | DeclarativePart :
      DeclarativePartType.t node
  | PrivatePart :
      PrivatePartType.t node
  | PublicPart :
      PublicPartType.t node
  | ElsifExprPart :
      ElsifExprPartType.t node
  | ElsifStmtPart :
      ElsifStmtPartType.t node
  | Expr :
      ExprType.t node
  | AbstractStateDeclExpr :
      AbstractStateDeclExprType.t node
  | Allocator :
      AllocatorType.t node
  | BaseAggregate :
      BaseAggregateType.t node
  | Aggregate :
      AggregateType.t node
  | BracketAggregate :
      BracketAggregateType.t node
  | DeltaAggregate :
      DeltaAggregateType.t node
  | BracketDeltaAggregate :
      BracketDeltaAggregateType.t node
  | NullRecordAggregate :
      NullRecordAggregateType.t node
  | BinOp :
      BinOpType.t node
  | RelationOp :
      RelationOpType.t node
  | BoxExpr :
      BoxExprType.t node
  | CaseExprAlternative :
      CaseExprAlternativeType.t node
  | CondExpr :
      CondExprType.t node
  | CaseExpr :
      CaseExprType.t node
  | IfExpr :
      IfExprType.t node
  | ContractCases :
      ContractCasesType.t node
  | DeclExpr :
      DeclExprType.t node
  | MembershipExpr :
      MembershipExprType.t node
  | Name :
      NameType.t node
  | AttributeRef :
      AttributeRefType.t node
  | UpdateAttributeRef :
      UpdateAttributeRefType.t node
  | CallExpr :
      CallExprType.t node
  | DefiningName :
      DefiningNameType.t node
  | DiscreteSubtypeName :
      DiscreteSubtypeNameType.t node
  | DottedName :
      DottedNameType.t node
  | EndName :
      EndNameType.t node
  | ExplicitDeref :
      ExplicitDerefType.t node
  | QualExpr :
      QualExprType.t node
  | SingleTokNode :
      SingleTokNodeType.t node
  | BaseId :
      BaseIdType.t node
  | CharLiteral :
      CharLiteralType.t node
  | Identifier :
      IdentifierType.t node
  | Op :
      OpType.t node
  | OpAbs :
      OpAbsType.t node
  | OpAnd :
      OpAndType.t node
  | OpAndThen :
      OpAndThenType.t node
  | OpConcat :
      OpConcatType.t node
  | OpDiv :
      OpDivType.t node
  | OpDoubleDot :
      OpDoubleDotType.t node
  | OpEq :
      OpEqType.t node
  | OpGt :
      OpGtType.t node
  | OpGte :
      OpGteType.t node
  | OpIn :
      OpInType.t node
  | OpLt :
      OpLtType.t node
  | OpLte :
      OpLteType.t node
  | OpMinus :
      OpMinusType.t node
  | OpMod :
      OpModType.t node
  | OpMult :
      OpMultType.t node
  | OpNeq :
      OpNeqType.t node
  | OpNot :
      OpNotType.t node
  | OpNotIn :
      OpNotInType.t node
  | OpOr :
      OpOrType.t node
  | OpOrElse :
      OpOrElseType.t node
  | OpPlus :
      OpPlusType.t node
  | OpPow :
      OpPowType.t node
  | OpRem :
      OpRemType.t node
  | OpXor :
      OpXorType.t node
  | StringLiteral :
      StringLiteralType.t node
  | NullLiteral :
      NullLiteralType.t node
  | NumLiteral :
      NumLiteralType.t node
  | IntLiteral :
      IntLiteralType.t node
  | RealLiteral :
      RealLiteralType.t node
  | TargetName :
      TargetNameType.t node
  | ParenExpr :
      ParenExprType.t node
  | QuantifiedExpr :
      QuantifiedExprType.t node
  | RaiseExpr :
      RaiseExprType.t node
  | UnOp :
      UnOpType.t node
  | HandledStmts :
      HandledStmtsType.t node
  | InterfaceKind :
      InterfaceKindType.t node
  | InterfaceKindLimited :
      InterfaceKindLimitedType.t node
  | InterfaceKindProtected :
      InterfaceKindProtectedType.t node
  | InterfaceKindSynchronized :
      InterfaceKindSynchronizedType.t node
  | InterfaceKindTask :
      InterfaceKindTaskType.t node
  | IterType :
      IterTypeType.t node
  | IterTypeIn :
      IterTypeInType.t node
  | IterTypeOf :
      IterTypeOfType.t node
  | LibraryItem :
      LibraryItemType.t node
  | LimitedNode :
      LimitedNodeType.t node
  | LimitedAbsent :
      LimitedAbsentType.t node
  | LimitedPresent :
      LimitedPresentType.t node
  | LoopSpec :
      LoopSpecType.t node
  | ForLoopSpec :
      ForLoopSpecType.t node
  | WhileLoopSpec :
      WhileLoopSpecType.t node
  | Mode :
      ModeType.t node
  | ModeDefault :
      ModeDefaultType.t node
  | ModeIn :
      ModeInType.t node
  | ModeInOut :
      ModeInOutType.t node
  | ModeOut :
      ModeOutType.t node
  | MultiAbstractStateDecl :
      MultiAbstractStateDeclType.t node
  | NotNull :
      NotNullType.t node
  | NotNullAbsent :
      NotNullAbsentType.t node
  | NotNullPresent :
      NotNullPresentType.t node
  | NullComponentDecl :
      NullComponentDeclType.t node
  | OthersDesignator :
      OthersDesignatorType.t node
  | OverridingNode :
      OverridingNodeType.t node
  | OverridingNotOverriding :
      OverridingNotOverridingType.t node
  | OverridingOverriding :
      OverridingOverridingType.t node
  | OverridingUnspecified :
      OverridingUnspecifiedType.t node
  | Params :
      ParamsType.t node
  | ParenAbstractStateDecl :
      ParenAbstractStateDeclType.t node
  | PragmaNode :
      PragmaNodeType.t node
  | PrivateNode :
      PrivateNodeType.t node
  | PrivateAbsent :
      PrivateAbsentType.t node
  | PrivatePresent :
      PrivatePresentType.t node
  | ProtectedDef :
      ProtectedDefType.t node
  | ProtectedNode :
      ProtectedNodeType.t node
  | ProtectedAbsent :
      ProtectedAbsentType.t node
  | ProtectedPresent :
      ProtectedPresentType.t node
  | Quantifier :
      QuantifierType.t node
  | QuantifierAll :
      QuantifierAllType.t node
  | QuantifierSome :
      QuantifierSomeType.t node
  | RangeSpec :
      RangeSpecType.t node
  | RenamingClause :
      RenamingClauseType.t node
  | SyntheticRenamingClause :
      SyntheticRenamingClauseType.t node
  | ReverseNode :
      ReverseNodeType.t node
  | ReverseAbsent :
      ReverseAbsentType.t node
  | ReversePresent :
      ReversePresentType.t node
  | SelectWhenPart :
      SelectWhenPartType.t node
  | Stmt :
      StmtType.t node
  | CompositeStmt :
      CompositeStmtType.t node
  | AcceptStmt :
      AcceptStmtType.t node
  | AcceptStmtWithStmts :
      AcceptStmtWithStmtsType.t node
  | BaseLoopStmt :
      BaseLoopStmtType.t node
  | ForLoopStmt :
      ForLoopStmtType.t node
  | LoopStmt :
      LoopStmtType.t node
  | WhileLoopStmt :
      WhileLoopStmtType.t node
  | BlockStmt :
      BlockStmtType.t node
  | BeginBlock :
      BeginBlockType.t node
  | DeclBlock :
      DeclBlockType.t node
  | CaseStmt :
      CaseStmtType.t node
  | ExtendedReturnStmt :
      ExtendedReturnStmtType.t node
  | IfStmt :
      IfStmtType.t node
  | NamedStmt :
      NamedStmtType.t node
  | SelectStmt :
      SelectStmtType.t node
  | ErrorStmt :
      ErrorStmtType.t node
  | SimpleStmt :
      SimpleStmtType.t node
  | AbortStmt :
      AbortStmtType.t node
  | AssignStmt :
      AssignStmtType.t node
  | CallStmt :
      CallStmtType.t node
  | DelayStmt :
      DelayStmtType.t node
  | ExitStmt :
      ExitStmtType.t node
  | GotoStmt :
      GotoStmtType.t node
  | Label :
      LabelType.t node
  | NullStmt :
      NullStmtType.t node
  | RaiseStmt :
      RaiseStmtType.t node
  | RequeueStmt :
      RequeueStmtType.t node
  | ReturnStmt :
      ReturnStmtType.t node
  | TerminateAlternative :
      TerminateAlternativeType.t node
  | SubpKind :
      SubpKindType.t node
  | SubpKindFunction :
      SubpKindFunctionType.t node
  | SubpKindProcedure :
      SubpKindProcedureType.t node
  | Subunit :
      SubunitType.t node
  | SynchronizedNode :
      SynchronizedNodeType.t node
  | SynchronizedAbsent :
      SynchronizedAbsentType.t node
  | SynchronizedPresent :
      SynchronizedPresentType.t node
  | TaggedNode :
      TaggedNodeType.t node
  | TaggedAbsent :
      TaggedAbsentType.t node
  | TaggedPresent :
      TaggedPresentType.t node
  | TaskDef :
      TaskDefType.t node
  | TypeDef :
      TypeDefType.t node
  | AccessDef :
      AccessDefType.t node
  | AccessToSubpDef :
      AccessToSubpDefType.t node
  | BaseTypeAccessDef :
      BaseTypeAccessDefType.t node
  | AnonymousTypeAccessDef :
      AnonymousTypeAccessDefType.t node
  | TypeAccessDef :
      TypeAccessDefType.t node
  | ArrayTypeDef :
      ArrayTypeDefType.t node
  | DerivedTypeDef :
      DerivedTypeDefType.t node
  | EnumTypeDef :
      EnumTypeDefType.t node
  | FormalDiscreteTypeDef :
      FormalDiscreteTypeDefType.t node
  | InterfaceTypeDef :
      InterfaceTypeDefType.t node
  | ModIntTypeDef :
      ModIntTypeDefType.t node
  | PrivateTypeDef :
      PrivateTypeDefType.t node
  | RealTypeDef :
      RealTypeDefType.t node
  | DecimalFixedPointDef :
      DecimalFixedPointDefType.t node
  | FloatingPointDef :
      FloatingPointDefType.t node
  | OrdinaryFixedPointDef :
      OrdinaryFixedPointDefType.t node
  | RecordTypeDef :
      RecordTypeDefType.t node
  | SignedIntTypeDef :
      SignedIntTypeDefType.t node
  | TypeExpr :
      TypeExprType.t node
  | AnonymousType :
      AnonymousTypeType.t node
  | EnumLitSynthTypeExpr :
      EnumLitSynthTypeExprType.t node
  | SubtypeIndication :
      SubtypeIndicationType.t node
  | ConstrainedSubtypeIndication :
      ConstrainedSubtypeIndicationType.t node
  | DiscreteSubtypeIndication :
      DiscreteSubtypeIndicationType.t node
  | UnconstrainedArrayIndex :
      UnconstrainedArrayIndexType.t node
  | UntilNode :
      UntilNodeType.t node
  | UntilAbsent :
      UntilAbsentType.t node
  | UntilPresent :
      UntilPresentType.t node
  | UseClause :
      UseClauseType.t node
  | UsePackageClause :
      UsePackageClauseType.t node
  | UseTypeClause :
      UseTypeClauseType.t node
  | Variant :
      VariantType.t node
  | VariantPart :
      VariantPartType.t node
  | WithClause :
      WithClauseType.t node
  | WithPrivate :
      WithPrivateType.t node
  | WithPrivateAbsent :
      WithPrivateAbsentType.t node
  | WithPrivatePresent :
      WithPrivatePresentType.t node

module WithPrivatePresent = struct
  type t =
    [
      | `WithPrivatePresent of
          WithPrivatePresentType.fields
    ]

  type fields = WithPrivatePresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module WithPrivateAbsent = struct
  type t =
    [
      | `WithPrivateAbsent of
          WithPrivateAbsentType.fields
    ]

  type fields = WithPrivateAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module WithPrivate = struct
  type t =
    [
      | WithPrivateAbsent.t
      | WithPrivatePresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.with_private_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module WithClause = struct
  type t =
    [
      | `WithClause of
          WithClauseType.fields
    ]

  type fields = WithClauseType.fields =
    
  {
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_has_private: PrivateNodeType.t
    Lazy.t;
         
    f_packages: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_limited node =
    match (node :> WithClauseType.t) with
    | `WithClause fields ->
        Lazy.force fields.f_has_limited
  let f_has_private node =
    match (node :> WithClauseType.t) with
    | `WithClause fields ->
        Lazy.force fields.f_has_private
  let f_packages node =
    match (node :> WithClauseType.t) with
    | `WithClause fields ->
        Lazy.force fields.f_packages



end

module VariantPart = struct
  type t =
    [
      | `VariantPart of
          VariantPartType.fields
    ]

  type fields = VariantPartType.fields =
    
  {
         
    f_discr_name: IdentifierType.t
    Lazy.t;
         
    f_variant: VariantListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_discr_name node =
    match (node :> VariantPartType.t) with
    | `VariantPart fields ->
        Lazy.force fields.f_discr_name
  let f_variant node =
    match (node :> VariantPartType.t) with
    | `VariantPart fields ->
        Lazy.force fields.f_variant



end

module Variant = struct
  type t =
    [
      | `Variant of
          VariantType.fields
    ]

  type fields = VariantType.fields =
    
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_choices node =
    match (node :> VariantType.t) with
    | `Variant fields ->
        Lazy.force fields.f_choices
  let f_components node =
    match (node :> VariantType.t) with
    | `Variant fields ->
        Lazy.force fields.f_components



end

module UseTypeClause = struct
  type t =
    [
      | `UseTypeClause of
          UseTypeClauseType.fields
    ]

  type fields = UseTypeClauseType.fields =
    
  {
         
    f_has_all: AllNodeType.t
    Lazy.t;
         
    f_types: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_all node =
    match (node :> UseTypeClauseType.t) with
    | `UseTypeClause fields ->
        Lazy.force fields.f_has_all
  let f_types node =
    match (node :> UseTypeClauseType.t) with
    | `UseTypeClause fields ->
        Lazy.force fields.f_types



end

module UsePackageClause = struct
  type t =
    [
      | `UsePackageClause of
          UsePackageClauseType.fields
    ]

  type fields = UsePackageClauseType.fields =
    
  {
         
    f_packages: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_packages node =
    match (node :> UsePackageClauseType.t) with
    | `UsePackageClause fields ->
        Lazy.force fields.f_packages



end

module UseClause = struct
  type t =
    [
      | UsePackageClause.t
      | UseTypeClause.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module UntilPresent = struct
  type t =
    [
      | `UntilPresent of
          UntilPresentType.fields
    ]

  type fields = UntilPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module UntilAbsent = struct
  type t =
    [
      | `UntilAbsent of
          UntilAbsentType.fields
    ]

  type fields = UntilAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module UntilNode = struct
  type t =
    [
      | UntilAbsent.t
      | UntilPresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.until_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module UnconstrainedArrayIndex = struct
  type t =
    [
      | `UnconstrainedArrayIndex of
          UnconstrainedArrayIndexType.fields
    ]

  type fields = UnconstrainedArrayIndexType.fields =
    
  {
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_subtype_indication node =
    match (node :> UnconstrainedArrayIndexType.t) with
    | `UnconstrainedArrayIndex fields ->
        Lazy.force fields.f_subtype_indication



end

module DiscreteSubtypeIndication = struct
  type t =
    [
      | `DiscreteSubtypeIndication of
          DiscreteSubtypeIndicationType.fields
    ]

  type fields = DiscreteSubtypeIndicationType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_not_null node =
    match (node :> DiscreteSubtypeIndicationType.t) with
    | `DiscreteSubtypeIndication fields ->
        Lazy.force fields.f_has_not_null
  let f_name node =
    match (node :> DiscreteSubtypeIndicationType.t) with
    | `DiscreteSubtypeIndication fields ->
        Lazy.force fields.f_name
  let f_constraint node =
    match (node :> DiscreteSubtypeIndicationType.t) with
    | `DiscreteSubtypeIndication fields ->
        Lazy.force fields.f_constraint



end

module ConstrainedSubtypeIndication = struct
  type t =
    [
      | `ConstrainedSubtypeIndication of
          ConstrainedSubtypeIndicationType.fields
    ]

  type fields = ConstrainedSubtypeIndicationType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_not_null node =
    match (node :> ConstrainedSubtypeIndicationType.t) with
    | `ConstrainedSubtypeIndication fields ->
        Lazy.force fields.f_has_not_null
  let f_name node =
    match (node :> ConstrainedSubtypeIndicationType.t) with
    | `ConstrainedSubtypeIndication fields ->
        Lazy.force fields.f_name
  let f_constraint node =
    match (node :> ConstrainedSubtypeIndicationType.t) with
    | `ConstrainedSubtypeIndication fields ->
        Lazy.force fields.f_constraint



end

module SubtypeIndication = struct
  type t =
    [
      | `SubtypeIndication of
          SubtypeIndicationType.fields
      | `ConstrainedSubtypeIndication of
          ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication of
          DiscreteSubtypeIndicationType.fields
    ]

  type fields = SubtypeIndicationType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_is_static_subtype
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.subtype_indication_p_is_static_subtype
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      !@ result_ptr


  let f_has_not_null node =
    match (node :> SubtypeIndicationType.t) with
    | `SubtypeIndication fields ->
        Lazy.force fields.f_has_not_null
    | `ConstrainedSubtypeIndication fields ->
        Lazy.force fields.f_has_not_null
    | `DiscreteSubtypeIndication fields ->
        Lazy.force fields.f_has_not_null
  let f_name node =
    match (node :> SubtypeIndicationType.t) with
    | `SubtypeIndication fields ->
        Lazy.force fields.f_name
    | `ConstrainedSubtypeIndication fields ->
        Lazy.force fields.f_name
    | `DiscreteSubtypeIndication fields ->
        Lazy.force fields.f_name
  let f_constraint node =
    match (node :> SubtypeIndicationType.t) with
    | `SubtypeIndication fields ->
        Lazy.force fields.f_constraint
    | `ConstrainedSubtypeIndication fields ->
        Lazy.force fields.f_constraint
    | `DiscreteSubtypeIndication fields ->
        Lazy.force fields.f_constraint



end

module EnumLitSynthTypeExpr = struct
  type t =
    [
      | `EnumLitSynthTypeExpr of
          EnumLitSynthTypeExprType.fields
    ]

  type fields = EnumLitSynthTypeExprType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AnonymousType = struct
  type t =
    [
      | `AnonymousType of
          AnonymousTypeType.fields
    ]

  type fields = AnonymousTypeType.fields =
    
  {
         
    f_type_decl: AnonymousTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_type_decl node =
    match (node :> AnonymousTypeType.t) with
    | `AnonymousType fields ->
        Lazy.force fields.f_type_decl



end

module TypeExpr = struct
  type t =
    [
      | AnonymousType.t
      | EnumLitSynthTypeExpr.t
      | SubtypeIndication.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_type_name
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.type_expr_p_type_name
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (NameType.wrap (context node) (!@ result_ptr))

let p_designated_type_decl
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.type_expr_p_designated_type_decl
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_designated_type_decl_from
    (node)
    (origin_node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin_node =
            
        addr (AdaNodeType.unwrap (origin_node))
      in
      let _ : int =
        CFunctions.type_expr_p_designated_type_decl_from
          (addr (AdaNodeType.unwrap (node)))
          c_origin_node
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))





end

module SignedIntTypeDef = struct
  type t =
    [
      | `SignedIntTypeDef of
          SignedIntTypeDefType.fields
    ]

  type fields = SignedIntTypeDefType.fields =
    
  {
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_range node =
    match (node :> SignedIntTypeDefType.t) with
    | `SignedIntTypeDef fields ->
        Lazy.force fields.f_range



end

module RecordTypeDef = struct
  type t =
    [
      | `RecordTypeDef of
          RecordTypeDefType.fields
    ]

  type fields = RecordTypeDefType.fields =
    
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_tagged: TaggedNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_record_def: BaseRecordDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_abstract node =
    match (node :> RecordTypeDefType.t) with
    | `RecordTypeDef fields ->
        Lazy.force fields.f_has_abstract
  let f_has_tagged node =
    match (node :> RecordTypeDefType.t) with
    | `RecordTypeDef fields ->
        Lazy.force fields.f_has_tagged
  let f_has_limited node =
    match (node :> RecordTypeDefType.t) with
    | `RecordTypeDef fields ->
        Lazy.force fields.f_has_limited
  let f_record_def node =
    match (node :> RecordTypeDefType.t) with
    | `RecordTypeDef fields ->
        Lazy.force fields.f_record_def



end

module OrdinaryFixedPointDef = struct
  type t =
    [
      | `OrdinaryFixedPointDef of
          OrdinaryFixedPointDefType.fields
    ]

  type fields = OrdinaryFixedPointDefType.fields =
    
  {
         
    f_delta: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_delta node =
    match (node :> OrdinaryFixedPointDefType.t) with
    | `OrdinaryFixedPointDef fields ->
        Lazy.force fields.f_delta
  let f_range node =
    match (node :> OrdinaryFixedPointDefType.t) with
    | `OrdinaryFixedPointDef fields ->
        Lazy.force fields.f_range



end

module FloatingPointDef = struct
  type t =
    [
      | `FloatingPointDef of
          FloatingPointDefType.fields
    ]

  type fields = FloatingPointDefType.fields =
    
  {
         
    f_num_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_num_digits node =
    match (node :> FloatingPointDefType.t) with
    | `FloatingPointDef fields ->
        Lazy.force fields.f_num_digits
  let f_range node =
    match (node :> FloatingPointDefType.t) with
    | `FloatingPointDef fields ->
        Lazy.force fields.f_range



end

module DecimalFixedPointDef = struct
  type t =
    [
      | `DecimalFixedPointDef of
          DecimalFixedPointDefType.fields
    ]

  type fields = DecimalFixedPointDefType.fields =
    
  {
         
    f_delta: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_delta node =
    match (node :> DecimalFixedPointDefType.t) with
    | `DecimalFixedPointDef fields ->
        Lazy.force fields.f_delta
  let f_digits node =
    match (node :> DecimalFixedPointDefType.t) with
    | `DecimalFixedPointDef fields ->
        Lazy.force fields.f_digits
  let f_range node =
    match (node :> DecimalFixedPointDefType.t) with
    | `DecimalFixedPointDef fields ->
        Lazy.force fields.f_range



end

module RealTypeDef = struct
  type t =
    [
      | DecimalFixedPointDef.t
      | FloatingPointDef.t
      | OrdinaryFixedPointDef.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module PrivateTypeDef = struct
  type t =
    [
      | `PrivateTypeDef of
          PrivateTypeDefType.fields
    ]

  type fields = PrivateTypeDefType.fields =
    
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_tagged: TaggedNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_abstract node =
    match (node :> PrivateTypeDefType.t) with
    | `PrivateTypeDef fields ->
        Lazy.force fields.f_has_abstract
  let f_has_tagged node =
    match (node :> PrivateTypeDefType.t) with
    | `PrivateTypeDef fields ->
        Lazy.force fields.f_has_tagged
  let f_has_limited node =
    match (node :> PrivateTypeDefType.t) with
    | `PrivateTypeDef fields ->
        Lazy.force fields.f_has_limited



end

module ModIntTypeDef = struct
  type t =
    [
      | `ModIntTypeDef of
          ModIntTypeDefType.fields
    ]

  type fields = ModIntTypeDefType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_expr node =
    match (node :> ModIntTypeDefType.t) with
    | `ModIntTypeDef fields ->
        Lazy.force fields.f_expr



end

module InterfaceTypeDef = struct
  type t =
    [
      | `InterfaceTypeDef of
          InterfaceTypeDefType.fields
    ]

  type fields = InterfaceTypeDefType.fields =
    
  {
         
    f_interface_kind: InterfaceKindType.t
    option
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_interface_kind node =
    match (node :> InterfaceTypeDefType.t) with
    | `InterfaceTypeDef fields ->
        Lazy.force fields.f_interface_kind
  let f_interfaces node =
    match (node :> InterfaceTypeDefType.t) with
    | `InterfaceTypeDef fields ->
        Lazy.force fields.f_interfaces



end

module FormalDiscreteTypeDef = struct
  type t =
    [
      | `FormalDiscreteTypeDef of
          FormalDiscreteTypeDefType.fields
    ]

  type fields = FormalDiscreteTypeDefType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module EnumTypeDef = struct
  type t =
    [
      | `EnumTypeDef of
          EnumTypeDefType.fields
    ]

  type fields = EnumTypeDefType.fields =
    
  {
         
    f_enum_literals: EnumLiteralDeclListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_enum_literals node =
    match (node :> EnumTypeDefType.t) with
    | `EnumTypeDef fields ->
        Lazy.force fields.f_enum_literals



end

module DerivedTypeDef = struct
  type t =
    [
      | `DerivedTypeDef of
          DerivedTypeDefType.fields
    ]

  type fields = DerivedTypeDefType.fields =
    
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_has_synchronized: SynchronizedNodeType.t
    Lazy.t;
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_record_extension: BaseRecordDefType.t
    option
    Lazy.t;
         
    f_has_with_private: WithPrivateType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_abstract node =
    match (node :> DerivedTypeDefType.t) with
    | `DerivedTypeDef fields ->
        Lazy.force fields.f_has_abstract
  let f_has_limited node =
    match (node :> DerivedTypeDefType.t) with
    | `DerivedTypeDef fields ->
        Lazy.force fields.f_has_limited
  let f_has_synchronized node =
    match (node :> DerivedTypeDefType.t) with
    | `DerivedTypeDef fields ->
        Lazy.force fields.f_has_synchronized
  let f_subtype_indication node =
    match (node :> DerivedTypeDefType.t) with
    | `DerivedTypeDef fields ->
        Lazy.force fields.f_subtype_indication
  let f_interfaces node =
    match (node :> DerivedTypeDefType.t) with
    | `DerivedTypeDef fields ->
        Lazy.force fields.f_interfaces
  let f_record_extension node =
    match (node :> DerivedTypeDefType.t) with
    | `DerivedTypeDef fields ->
        Lazy.force fields.f_record_extension
  let f_has_with_private node =
    match (node :> DerivedTypeDefType.t) with
    | `DerivedTypeDef fields ->
        Lazy.force fields.f_has_with_private



end

module ArrayTypeDef = struct
  type t =
    [
      | `ArrayTypeDef of
          ArrayTypeDefType.fields
    ]

  type fields = ArrayTypeDefType.fields =
    
  {
         
    f_indices: ArrayIndicesType.t
    Lazy.t;
         
    f_component_type: ComponentDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_indices node =
    match (node :> ArrayTypeDefType.t) with
    | `ArrayTypeDef fields ->
        Lazy.force fields.f_indices
  let f_component_type node =
    match (node :> ArrayTypeDefType.t) with
    | `ArrayTypeDef fields ->
        Lazy.force fields.f_component_type



end

module TypeAccessDef = struct
  type t =
    [
      | `TypeAccessDef of
          TypeAccessDefType.fields
    ]

  type fields = TypeAccessDefType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_has_all: AllNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_not_null node =
    match (node :> TypeAccessDefType.t) with
    | `TypeAccessDef fields ->
        Lazy.force fields.f_has_not_null
  let f_has_all node =
    match (node :> TypeAccessDefType.t) with
    | `TypeAccessDef fields ->
        Lazy.force fields.f_has_all
  let f_has_constant node =
    match (node :> TypeAccessDefType.t) with
    | `TypeAccessDef fields ->
        Lazy.force fields.f_has_constant
  let f_subtype_indication node =
    match (node :> TypeAccessDefType.t) with
    | `TypeAccessDef fields ->
        Lazy.force fields.f_subtype_indication



end

module AnonymousTypeAccessDef = struct
  type t =
    [
      | `AnonymousTypeAccessDef of
          AnonymousTypeAccessDefType.fields
    ]

  type fields = AnonymousTypeAccessDefType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_type_decl: BaseTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_not_null node =
    match (node :> AnonymousTypeAccessDefType.t) with
    | `AnonymousTypeAccessDef fields ->
        Lazy.force fields.f_has_not_null
  let f_type_decl node =
    match (node :> AnonymousTypeAccessDefType.t) with
    | `AnonymousTypeAccessDef fields ->
        Lazy.force fields.f_type_decl



end

module BaseTypeAccessDef = struct
  type t =
    [
      | AnonymousTypeAccessDef.t
      | TypeAccessDef.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_not_null node =
    match (node :> BaseTypeAccessDefType.t) with
    | `AnonymousTypeAccessDef fields ->
        Lazy.force fields.f_has_not_null
    | `TypeAccessDef fields ->
        Lazy.force fields.f_has_not_null



end

module AccessToSubpDef = struct
  type t =
    [
      | `AccessToSubpDef of
          AccessToSubpDefType.fields
    ]

  type fields = AccessToSubpDefType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_has_protected: ProtectedNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_not_null node =
    match (node :> AccessToSubpDefType.t) with
    | `AccessToSubpDef fields ->
        Lazy.force fields.f_has_not_null
  let f_has_protected node =
    match (node :> AccessToSubpDefType.t) with
    | `AccessToSubpDef fields ->
        Lazy.force fields.f_has_protected
  let f_subp_spec node =
    match (node :> AccessToSubpDefType.t) with
    | `AccessToSubpDef fields ->
        Lazy.force fields.f_subp_spec



end

module AccessDef = struct
  type t =
    [
      | AccessToSubpDef.t
      | BaseTypeAccessDef.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_not_null node =
    match (node :> AccessDefType.t) with
    | `AccessToSubpDef fields ->
        Lazy.force fields.f_has_not_null
    | `AnonymousTypeAccessDef fields ->
        Lazy.force fields.f_has_not_null
    | `TypeAccessDef fields ->
        Lazy.force fields.f_has_not_null



end

module TypeDef = struct
  type t =
    [
      | AccessDef.t
      | ArrayTypeDef.t
      | DerivedTypeDef.t
      | EnumTypeDef.t
      | FormalDiscreteTypeDef.t
      | InterfaceTypeDef.t
      | ModIntTypeDef.t
      | PrivateTypeDef.t
      | RealTypeDef.t
      | RecordTypeDef.t
      | SignedIntTypeDef.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module TaskDef = struct
  type t =
    [
      | `TaskDef of
          TaskDefType.fields
    ]

  type fields = TaskDefType.fields =
    
  {
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_interfaces node =
    match (node :> TaskDefType.t) with
    | `TaskDef fields ->
        Lazy.force fields.f_interfaces
  let f_public_part node =
    match (node :> TaskDefType.t) with
    | `TaskDef fields ->
        Lazy.force fields.f_public_part
  let f_private_part node =
    match (node :> TaskDefType.t) with
    | `TaskDef fields ->
        Lazy.force fields.f_private_part
  let f_end_name node =
    match (node :> TaskDefType.t) with
    | `TaskDef fields ->
        Lazy.force fields.f_end_name



end

module TaggedPresent = struct
  type t =
    [
      | `TaggedPresent of
          TaggedPresentType.fields
    ]

  type fields = TaggedPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module TaggedAbsent = struct
  type t =
    [
      | `TaggedAbsent of
          TaggedAbsentType.fields
    ]

  type fields = TaggedAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module TaggedNode = struct
  type t =
    [
      | TaggedAbsent.t
      | TaggedPresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.tagged_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module SynchronizedPresent = struct
  type t =
    [
      | `SynchronizedPresent of
          SynchronizedPresentType.fields
    ]

  type fields = SynchronizedPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module SynchronizedAbsent = struct
  type t =
    [
      | `SynchronizedAbsent of
          SynchronizedAbsentType.fields
    ]

  type fields = SynchronizedAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module SynchronizedNode = struct
  type t =
    [
      | SynchronizedAbsent.t
      | SynchronizedPresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.synchronized_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module Subunit = struct
  type t =
    [
      | `Subunit of
          SubunitType.fields
    ]

  type fields = SubunitType.fields =
    
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_body: [
      | `PackageBody
          of PackageBodyType.fields
      | `ProtectedBody
          of ProtectedBodyType.fields
      | `SubpBody
          of SubpBodyType.fields
      | `TaskBody
          of TaskBodyType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_body_root
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.subunit_p_body_root
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))


  let f_name node =
    match (node :> SubunitType.t) with
    | `Subunit fields ->
        Lazy.force fields.f_name
  let f_body node =
    match (node :> SubunitType.t) with
    | `Subunit fields ->
        Lazy.force fields.f_body



end

module SubpKindProcedure = struct
  type t =
    [
      | `SubpKindProcedure of
          SubpKindProcedureType.fields
    ]

  type fields = SubpKindProcedureType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module SubpKindFunction = struct
  type t =
    [
      | `SubpKindFunction of
          SubpKindFunctionType.fields
    ]

  type fields = SubpKindFunctionType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module SubpKind = struct
  type t =
    [
      | SubpKindFunction.t
      | SubpKindProcedure.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module TerminateAlternative = struct
  type t =
    [
      | `TerminateAlternative of
          TerminateAlternativeType.fields
    ]

  type fields = TerminateAlternativeType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ReturnStmt = struct
  type t =
    [
      | `ReturnStmt of
          ReturnStmtType.fields
    ]

  type fields = ReturnStmtType.fields =
    
  {
         
    f_return_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_return_expr node =
    match (node :> ReturnStmtType.t) with
    | `ReturnStmt fields ->
        Lazy.force fields.f_return_expr



end

module RequeueStmt = struct
  type t =
    [
      | `RequeueStmt of
          RequeueStmtType.fields
    ]

  type fields = RequeueStmtType.fields =
    
  {
         
    f_call_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_has_abort: AbortNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_call_name node =
    match (node :> RequeueStmtType.t) with
    | `RequeueStmt fields ->
        Lazy.force fields.f_call_name
  let f_has_abort node =
    match (node :> RequeueStmtType.t) with
    | `RequeueStmt fields ->
        Lazy.force fields.f_has_abort



end

module RaiseStmt = struct
  type t =
    [
      | `RaiseStmt of
          RaiseStmtType.fields
    ]

  type fields = RaiseStmtType.fields =
    
  {
         
    f_exception_name: NameType.t
    option
    Lazy.t;
         
    f_error_message: ExprType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_exception_name node =
    match (node :> RaiseStmtType.t) with
    | `RaiseStmt fields ->
        Lazy.force fields.f_exception_name
  let f_error_message node =
    match (node :> RaiseStmtType.t) with
    | `RaiseStmt fields ->
        Lazy.force fields.f_error_message



end

module NullStmt = struct
  type t =
    [
      | `NullStmt of
          NullStmtType.fields
    ]

  type fields = NullStmtType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module Label = struct
  type t =
    [
      | `Label of
          LabelType.fields
    ]

  type fields = LabelType.fields =
    
  {
         
    f_decl: LabelDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decl node =
    match (node :> LabelType.t) with
    | `Label fields ->
        Lazy.force fields.f_decl



end

module GotoStmt = struct
  type t =
    [
      | `GotoStmt of
          GotoStmtType.fields
    ]

  type fields = GotoStmtType.fields =
    
  {
         
    f_label_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_label_name node =
    match (node :> GotoStmtType.t) with
    | `GotoStmt fields ->
        Lazy.force fields.f_label_name



end

module ExitStmt = struct
  type t =
    [
      | `ExitStmt of
          ExitStmtType.fields
    ]

  type fields = ExitStmtType.fields =
    
  {
         
    f_loop_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    option
    Lazy.t;
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_loop_name node =
    match (node :> ExitStmtType.t) with
    | `ExitStmt fields ->
        Lazy.force fields.f_loop_name
  let f_cond_expr node =
    match (node :> ExitStmtType.t) with
    | `ExitStmt fields ->
        Lazy.force fields.f_cond_expr



end

module DelayStmt = struct
  type t =
    [
      | `DelayStmt of
          DelayStmtType.fields
    ]

  type fields = DelayStmtType.fields =
    
  {
         
    f_has_until: UntilNodeType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_until node =
    match (node :> DelayStmtType.t) with
    | `DelayStmt fields ->
        Lazy.force fields.f_has_until
  let f_expr node =
    match (node :> DelayStmtType.t) with
    | `DelayStmt fields ->
        Lazy.force fields.f_expr



end

module CallStmt = struct
  type t =
    [
      | `CallStmt of
          CallStmtType.fields
    ]

  type fields = CallStmtType.fields =
    
  {
         
    f_call: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_call node =
    match (node :> CallStmtType.t) with
    | `CallStmt fields ->
        Lazy.force fields.f_call



end

module AssignStmt = struct
  type t =
    [
      | `AssignStmt of
          AssignStmtType.fields
    ]

  type fields = AssignStmtType.fields =
    
  {
         
    f_dest: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_dest node =
    match (node :> AssignStmtType.t) with
    | `AssignStmt fields ->
        Lazy.force fields.f_dest
  let f_expr node =
    match (node :> AssignStmtType.t) with
    | `AssignStmt fields ->
        Lazy.force fields.f_expr



end

module AbortStmt = struct
  type t =
    [
      | `AbortStmt of
          AbortStmtType.fields
    ]

  type fields = AbortStmtType.fields =
    
  {
         
    f_names: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_names node =
    match (node :> AbortStmtType.t) with
    | `AbortStmt fields ->
        Lazy.force fields.f_names



end

module SimpleStmt = struct
  type t =
    [
      | AbortStmt.t
      | AssignStmt.t
      | CallStmt.t
      | DelayStmt.t
      | ExitStmt.t
      | GotoStmt.t
      | Label.t
      | NullStmt.t
      | RaiseStmt.t
      | RequeueStmt.t
      | ReturnStmt.t
      | TerminateAlternative.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ErrorStmt = struct
  type t =
    [
      | `ErrorStmt of
          ErrorStmtType.fields
    ]

  type fields = ErrorStmtType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module SelectStmt = struct
  type t =
    [
      | `SelectStmt of
          SelectStmtType.fields
    ]

  type fields = SelectStmtType.fields =
    
  {
         
    f_guards: SelectWhenPartListType.t
    Lazy.t;
         
    f_else_stmts: StmtListType.t
    option
    Lazy.t;
         
    f_abort_stmts: StmtListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_guards node =
    match (node :> SelectStmtType.t) with
    | `SelectStmt fields ->
        Lazy.force fields.f_guards
  let f_else_stmts node =
    match (node :> SelectStmtType.t) with
    | `SelectStmt fields ->
        Lazy.force fields.f_else_stmts
  let f_abort_stmts node =
    match (node :> SelectStmtType.t) with
    | `SelectStmt fields ->
        Lazy.force fields.f_abort_stmts



end

module NamedStmt = struct
  type t =
    [
      | `NamedStmt of
          NamedStmtType.fields
    ]

  type fields = NamedStmtType.fields =
    
  {
         
    f_decl: NamedStmtDeclType.t
    Lazy.t;
         
    f_stmt: [
      | `BeginBlock
          of BeginBlockType.fields
      | `DeclBlock
          of DeclBlockType.fields
      | `ForLoopStmt
          of ForLoopStmtType.fields
      | `LoopStmt
          of LoopStmtType.fields
      | `WhileLoopStmt
          of WhileLoopStmtType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decl node =
    match (node :> NamedStmtType.t) with
    | `NamedStmt fields ->
        Lazy.force fields.f_decl
  let f_stmt node =
    match (node :> NamedStmtType.t) with
    | `NamedStmt fields ->
        Lazy.force fields.f_stmt



end

module IfStmt = struct
  type t =
    [
      | `IfStmt of
          IfStmtType.fields
    ]

  type fields = IfStmtType.fields =
    
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_stmts: StmtListType.t
    Lazy.t;
         
    f_alternatives: ElsifStmtPartListType.t
    Lazy.t;
         
    f_else_stmts: StmtListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_cond_expr node =
    match (node :> IfStmtType.t) with
    | `IfStmt fields ->
        Lazy.force fields.f_cond_expr
  let f_then_stmts node =
    match (node :> IfStmtType.t) with
    | `IfStmt fields ->
        Lazy.force fields.f_then_stmts
  let f_alternatives node =
    match (node :> IfStmtType.t) with
    | `IfStmt fields ->
        Lazy.force fields.f_alternatives
  let f_else_stmts node =
    match (node :> IfStmtType.t) with
    | `IfStmt fields ->
        Lazy.force fields.f_else_stmts



end

module ExtendedReturnStmt = struct
  type t =
    [
      | `ExtendedReturnStmt of
          ExtendedReturnStmtType.fields
    ]

  type fields = ExtendedReturnStmtType.fields =
    
  {
         
    f_decl: ExtendedReturnStmtObjectDeclType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decl node =
    match (node :> ExtendedReturnStmtType.t) with
    | `ExtendedReturnStmt fields ->
        Lazy.force fields.f_decl
  let f_stmts node =
    match (node :> ExtendedReturnStmtType.t) with
    | `ExtendedReturnStmt fields ->
        Lazy.force fields.f_stmts



end

module CaseStmt = struct
  type t =
    [
      | `CaseStmt of
          CaseStmtType.fields
    ]

  type fields = CaseStmtType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_alternatives: CaseStmtAlternativeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_expr node =
    match (node :> CaseStmtType.t) with
    | `CaseStmt fields ->
        Lazy.force fields.f_expr
  let f_alternatives node =
    match (node :> CaseStmtType.t) with
    | `CaseStmt fields ->
        Lazy.force fields.f_alternatives



end

module DeclBlock = struct
  type t =
    [
      | `DeclBlock of
          DeclBlockType.fields
    ]

  type fields = DeclBlockType.fields =
    
  {
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decls node =
    match (node :> DeclBlockType.t) with
    | `DeclBlock fields ->
        Lazy.force fields.f_decls
  let f_stmts node =
    match (node :> DeclBlockType.t) with
    | `DeclBlock fields ->
        Lazy.force fields.f_stmts
  let f_end_name node =
    match (node :> DeclBlockType.t) with
    | `DeclBlock fields ->
        Lazy.force fields.f_end_name



end

module BeginBlock = struct
  type t =
    [
      | `BeginBlock of
          BeginBlockType.fields
    ]

  type fields = BeginBlockType.fields =
    
  {
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_stmts node =
    match (node :> BeginBlockType.t) with
    | `BeginBlock fields ->
        Lazy.force fields.f_stmts
  let f_end_name node =
    match (node :> BeginBlockType.t) with
    | `BeginBlock fields ->
        Lazy.force fields.f_end_name



end

module BlockStmt = struct
  type t =
    [
      | BeginBlock.t
      | DeclBlock.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module WhileLoopStmt = struct
  type t =
    [
      | `WhileLoopStmt of
          WhileLoopStmtType.fields
    ]

  type fields = WhileLoopStmtType.fields =
    
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_spec node =
    match (node :> WhileLoopStmtType.t) with
    | `WhileLoopStmt fields ->
        Lazy.force fields.f_spec
  let f_stmts node =
    match (node :> WhileLoopStmtType.t) with
    | `WhileLoopStmt fields ->
        Lazy.force fields.f_stmts
  let f_end_name node =
    match (node :> WhileLoopStmtType.t) with
    | `WhileLoopStmt fields ->
        Lazy.force fields.f_end_name



end

module LoopStmt = struct
  type t =
    [
      | `LoopStmt of
          LoopStmtType.fields
    ]

  type fields = LoopStmtType.fields =
    
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_spec node =
    match (node :> LoopStmtType.t) with
    | `LoopStmt fields ->
        Lazy.force fields.f_spec
  let f_stmts node =
    match (node :> LoopStmtType.t) with
    | `LoopStmt fields ->
        Lazy.force fields.f_stmts
  let f_end_name node =
    match (node :> LoopStmtType.t) with
    | `LoopStmt fields ->
        Lazy.force fields.f_end_name



end

module ForLoopStmt = struct
  type t =
    [
      | `ForLoopStmt of
          ForLoopStmtType.fields
    ]

  type fields = ForLoopStmtType.fields =
    
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_spec node =
    match (node :> ForLoopStmtType.t) with
    | `ForLoopStmt fields ->
        Lazy.force fields.f_spec
  let f_stmts node =
    match (node :> ForLoopStmtType.t) with
    | `ForLoopStmt fields ->
        Lazy.force fields.f_stmts
  let f_end_name node =
    match (node :> ForLoopStmtType.t) with
    | `ForLoopStmt fields ->
        Lazy.force fields.f_end_name



end

module BaseLoopStmt = struct
  type t =
    [
      | ForLoopStmt.t
      | LoopStmt.t
      | WhileLoopStmt.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_spec node =
    match (node :> BaseLoopStmtType.t) with
    | `ForLoopStmt fields ->
        Lazy.force fields.f_spec
    | `LoopStmt fields ->
        Lazy.force fields.f_spec
    | `WhileLoopStmt fields ->
        Lazy.force fields.f_spec
  let f_stmts node =
    match (node :> BaseLoopStmtType.t) with
    | `ForLoopStmt fields ->
        Lazy.force fields.f_stmts
    | `LoopStmt fields ->
        Lazy.force fields.f_stmts
    | `WhileLoopStmt fields ->
        Lazy.force fields.f_stmts
  let f_end_name node =
    match (node :> BaseLoopStmtType.t) with
    | `ForLoopStmt fields ->
        Lazy.force fields.f_end_name
    | `LoopStmt fields ->
        Lazy.force fields.f_end_name
    | `WhileLoopStmt fields ->
        Lazy.force fields.f_end_name



end

module AcceptStmtWithStmts = struct
  type t =
    [
      | `AcceptStmtWithStmts of
          AcceptStmtWithStmtsType.fields
    ]

  type fields = AcceptStmtWithStmtsType.fields =
    
  {
         
    f_name: IdentifierType.t
    Lazy.t;
         
    f_entry_index_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> AcceptStmtWithStmtsType.t) with
    | `AcceptStmtWithStmts fields ->
        Lazy.force fields.f_name
  let f_entry_index_expr node =
    match (node :> AcceptStmtWithStmtsType.t) with
    | `AcceptStmtWithStmts fields ->
        Lazy.force fields.f_entry_index_expr
  let f_params node =
    match (node :> AcceptStmtWithStmtsType.t) with
    | `AcceptStmtWithStmts fields ->
        Lazy.force fields.f_params
  let f_stmts node =
    match (node :> AcceptStmtWithStmtsType.t) with
    | `AcceptStmtWithStmts fields ->
        Lazy.force fields.f_stmts
  let f_end_name node =
    match (node :> AcceptStmtWithStmtsType.t) with
    | `AcceptStmtWithStmts fields ->
        Lazy.force fields.f_end_name



end

module AcceptStmt = struct
  type t =
    [
      | `AcceptStmt of
          AcceptStmtType.fields
      | `AcceptStmtWithStmts of
          AcceptStmtWithStmtsType.fields
    ]

  type fields = AcceptStmtType.fields =
    
  {
         
    f_name: IdentifierType.t
    Lazy.t;
         
    f_entry_index_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> AcceptStmtType.t) with
    | `AcceptStmt fields ->
        Lazy.force fields.f_name
    | `AcceptStmtWithStmts fields ->
        Lazy.force fields.f_name
  let f_entry_index_expr node =
    match (node :> AcceptStmtType.t) with
    | `AcceptStmt fields ->
        Lazy.force fields.f_entry_index_expr
    | `AcceptStmtWithStmts fields ->
        Lazy.force fields.f_entry_index_expr
  let f_params node =
    match (node :> AcceptStmtType.t) with
    | `AcceptStmt fields ->
        Lazy.force fields.f_params
    | `AcceptStmtWithStmts fields ->
        Lazy.force fields.f_params



end

module CompositeStmt = struct
  type t =
    [
      | AcceptStmt.t
      | BaseLoopStmt.t
      | BlockStmt.t
      | CaseStmt.t
      | ExtendedReturnStmt.t
      | IfStmt.t
      | NamedStmt.t
      | SelectStmt.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module Stmt = struct
  type t =
    [
      | CompositeStmt.t
      | ErrorStmt.t
      | SimpleStmt.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module SelectWhenPart = struct
  type t =
    [
      | `SelectWhenPart of
          SelectWhenPartType.fields
    ]

  type fields = SelectWhenPartType.fields =
    
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_cond_expr node =
    match (node :> SelectWhenPartType.t) with
    | `SelectWhenPart fields ->
        Lazy.force fields.f_cond_expr
  let f_stmts node =
    match (node :> SelectWhenPartType.t) with
    | `SelectWhenPart fields ->
        Lazy.force fields.f_stmts



end

module ReversePresent = struct
  type t =
    [
      | `ReversePresent of
          ReversePresentType.fields
    ]

  type fields = ReversePresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ReverseAbsent = struct
  type t =
    [
      | `ReverseAbsent of
          ReverseAbsentType.fields
    ]

  type fields = ReverseAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ReverseNode = struct
  type t =
    [
      | ReverseAbsent.t
      | ReversePresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.reverse_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module SyntheticRenamingClause = struct
  type t =
    [
      | `SyntheticRenamingClause of
          SyntheticRenamingClauseType.fields
    ]

  type fields = SyntheticRenamingClauseType.fields =
    
  {
         
    f_renamed_object: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_renamed_object node =
    match (node :> SyntheticRenamingClauseType.t) with
    | `SyntheticRenamingClause fields ->
        Lazy.force fields.f_renamed_object



end

module RenamingClause = struct
  type t =
    [
      | `RenamingClause of
          RenamingClauseType.fields
      | `SyntheticRenamingClause of
          SyntheticRenamingClauseType.fields
    ]

  type fields = RenamingClauseType.fields =
    
  {
         
    f_renamed_object: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_renamed_object node =
    match (node :> RenamingClauseType.t) with
    | `RenamingClause fields ->
        Lazy.force fields.f_renamed_object
    | `SyntheticRenamingClause fields ->
        Lazy.force fields.f_renamed_object



end

module RangeSpec = struct
  type t =
    [
      | `RangeSpec of
          RangeSpecType.fields
    ]

  type fields = RangeSpecType.fields =
    
  {
         
    f_range: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_range node =
    match (node :> RangeSpecType.t) with
    | `RangeSpec fields ->
        Lazy.force fields.f_range



end

module QuantifierSome = struct
  type t =
    [
      | `QuantifierSome of
          QuantifierSomeType.fields
    ]

  type fields = QuantifierSomeType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module QuantifierAll = struct
  type t =
    [
      | `QuantifierAll of
          QuantifierAllType.fields
    ]

  type fields = QuantifierAllType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module Quantifier = struct
  type t =
    [
      | QuantifierAll.t
      | QuantifierSome.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ProtectedPresent = struct
  type t =
    [
      | `ProtectedPresent of
          ProtectedPresentType.fields
    ]

  type fields = ProtectedPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ProtectedAbsent = struct
  type t =
    [
      | `ProtectedAbsent of
          ProtectedAbsentType.fields
    ]

  type fields = ProtectedAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ProtectedNode = struct
  type t =
    [
      | ProtectedAbsent.t
      | ProtectedPresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.protected_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module ProtectedDef = struct
  type t =
    [
      | `ProtectedDef of
          ProtectedDefType.fields
    ]

  type fields = ProtectedDefType.fields =
    
  {
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_public_part node =
    match (node :> ProtectedDefType.t) with
    | `ProtectedDef fields ->
        Lazy.force fields.f_public_part
  let f_private_part node =
    match (node :> ProtectedDefType.t) with
    | `ProtectedDef fields ->
        Lazy.force fields.f_private_part
  let f_end_name node =
    match (node :> ProtectedDefType.t) with
    | `ProtectedDef fields ->
        Lazy.force fields.f_end_name



end

module PrivatePresent = struct
  type t =
    [
      | `PrivatePresent of
          PrivatePresentType.fields
    ]

  type fields = PrivatePresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module PrivateAbsent = struct
  type t =
    [
      | `PrivateAbsent of
          PrivateAbsentType.fields
    ]

  type fields = PrivateAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module PrivateNode = struct
  type t =
    [
      | PrivateAbsent.t
      | PrivatePresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.private_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module PragmaNode = struct
  type t =
    [
      | `PragmaNode of
          PragmaNodeType.fields
    ]

  type fields = PragmaNodeType.fields =
    
  {
         
    f_id: IdentifierType.t
    Lazy.t;
         
    f_args: BaseAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_associated_decls
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.pragma_node_p_associated_decls
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      BasicDeclArray.wrap (context node) (!@ result_ptr)


  let f_id node =
    match (node :> PragmaNodeType.t) with
    | `PragmaNode fields ->
        Lazy.force fields.f_id
  let f_args node =
    match (node :> PragmaNodeType.t) with
    | `PragmaNode fields ->
        Lazy.force fields.f_args



end

module ParenAbstractStateDecl = struct
  type t =
    [
      | `ParenAbstractStateDecl of
          ParenAbstractStateDeclType.fields
    ]

  type fields = ParenAbstractStateDeclType.fields =
    
  {
         
    f_decl: [
      | `AbstractStateDecl
          of AbstractStateDeclType.fields
      | `ParenAbstractStateDecl
          of fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decl node =
    match (node :> ParenAbstractStateDeclType.t) with
    | `ParenAbstractStateDecl fields ->
        Lazy.force fields.f_decl



end

module Params = struct
  type t =
    [
      | `Params of
          ParamsType.fields
    ]

  type fields = ParamsType.fields =
    
  {
         
    f_params: ParamSpecListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_params node =
    match (node :> ParamsType.t) with
    | `Params fields ->
        Lazy.force fields.f_params



end

module OverridingUnspecified = struct
  type t =
    [
      | `OverridingUnspecified of
          OverridingUnspecifiedType.fields
    ]

  type fields = OverridingUnspecifiedType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OverridingOverriding = struct
  type t =
    [
      | `OverridingOverriding of
          OverridingOverridingType.fields
    ]

  type fields = OverridingOverridingType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OverridingNotOverriding = struct
  type t =
    [
      | `OverridingNotOverriding of
          OverridingNotOverridingType.fields
    ]

  type fields = OverridingNotOverridingType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OverridingNode = struct
  type t =
    [
      | OverridingNotOverriding.t
      | OverridingOverriding.t
      | OverridingUnspecified.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OthersDesignator = struct
  type t =
    [
      | `OthersDesignator of
          OthersDesignatorType.fields
    ]

  type fields = OthersDesignatorType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module NullComponentDecl = struct
  type t =
    [
      | `NullComponentDecl of
          NullComponentDeclType.fields
    ]

  type fields = NullComponentDeclType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module NotNullPresent = struct
  type t =
    [
      | `NotNullPresent of
          NotNullPresentType.fields
    ]

  type fields = NotNullPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module NotNullAbsent = struct
  type t =
    [
      | `NotNullAbsent of
          NotNullAbsentType.fields
    ]

  type fields = NotNullAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module NotNull = struct
  type t =
    [
      | NotNullAbsent.t
      | NotNullPresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.not_null_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module MultiAbstractStateDecl = struct
  type t =
    [
      | `MultiAbstractStateDecl of
          MultiAbstractStateDeclType.fields
    ]

  type fields = MultiAbstractStateDeclType.fields =
    
  {
         
    f_decls: AbstractStateDeclListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decls node =
    match (node :> MultiAbstractStateDeclType.t) with
    | `MultiAbstractStateDecl fields ->
        Lazy.force fields.f_decls



end

module ModeOut = struct
  type t =
    [
      | `ModeOut of
          ModeOutType.fields
    ]

  type fields = ModeOutType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ModeInOut = struct
  type t =
    [
      | `ModeInOut of
          ModeInOutType.fields
    ]

  type fields = ModeInOutType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ModeIn = struct
  type t =
    [
      | `ModeIn of
          ModeInType.fields
    ]

  type fields = ModeInType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ModeDefault = struct
  type t =
    [
      | `ModeDefault of
          ModeDefaultType.fields
    ]

  type fields = ModeDefaultType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module Mode = struct
  type t =
    [
      | ModeDefault.t
      | ModeIn.t
      | ModeInOut.t
      | ModeOut.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module WhileLoopSpec = struct
  type t =
    [
      | `WhileLoopSpec of
          WhileLoopSpecType.fields
    ]

  type fields = WhileLoopSpecType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_expr node =
    match (node :> WhileLoopSpecType.t) with
    | `WhileLoopSpec fields ->
        Lazy.force fields.f_expr



end

module ForLoopSpec = struct
  type t =
    [
      | `ForLoopSpec of
          ForLoopSpecType.fields
    ]

  type fields = ForLoopSpecType.fields =
    
  {
         
    f_var_decl: ForLoopVarDeclType.t
    Lazy.t;
         
    f_loop_type: IterTypeType.t
    Lazy.t;
         
    f_has_reverse: ReverseNodeType.t
    Lazy.t;
         
    f_iter_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_var_decl node =
    match (node :> ForLoopSpecType.t) with
    | `ForLoopSpec fields ->
        Lazy.force fields.f_var_decl
  let f_loop_type node =
    match (node :> ForLoopSpecType.t) with
    | `ForLoopSpec fields ->
        Lazy.force fields.f_loop_type
  let f_has_reverse node =
    match (node :> ForLoopSpecType.t) with
    | `ForLoopSpec fields ->
        Lazy.force fields.f_has_reverse
  let f_iter_expr node =
    match (node :> ForLoopSpecType.t) with
    | `ForLoopSpec fields ->
        Lazy.force fields.f_iter_expr



end

module LoopSpec = struct
  type t =
    [
      | ForLoopSpec.t
      | WhileLoopSpec.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module LimitedPresent = struct
  type t =
    [
      | `LimitedPresent of
          LimitedPresentType.fields
    ]

  type fields = LimitedPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module LimitedAbsent = struct
  type t =
    [
      | `LimitedAbsent of
          LimitedAbsentType.fields
    ]

  type fields = LimitedAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module LimitedNode = struct
  type t =
    [
      | LimitedAbsent.t
      | LimitedPresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.limited_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module LibraryItem = struct
  type t =
    [
      | `LibraryItem of
          LibraryItemType.fields
    ]

  type fields = LibraryItemType.fields =
    
  {
         
    f_has_private: PrivateNodeType.t
    Lazy.t;
         
    f_item: [
      | `AbstractSubpDecl
          of AbstractSubpDeclType.fields
      | `ErrorDecl
          of ErrorDeclType.fields
      | `ExprFunction
          of ExprFunctionType.fields
      | `GenericPackageDecl
          of GenericPackageDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericPackageRenamingDecl
          of GenericPackageRenamingDeclType.fields
      | `GenericSubpDecl
          of GenericSubpDeclType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `GenericSubpRenamingDecl
          of GenericSubpRenamingDeclType.fields
      | `NullSubpDecl
          of NullSubpDeclType.fields
      | `PackageBody
          of PackageBodyType.fields
      | `PackageDecl
          of PackageDeclType.fields
      | `PackageRenamingDecl
          of PackageRenamingDeclType.fields
      | `SubpBody
          of SubpBodyType.fields
      | `SubpDecl
          of SubpDeclType.fields
      | `SubpRenamingDecl
          of SubpRenamingDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_private node =
    match (node :> LibraryItemType.t) with
    | `LibraryItem fields ->
        Lazy.force fields.f_has_private
  let f_item node =
    match (node :> LibraryItemType.t) with
    | `LibraryItem fields ->
        Lazy.force fields.f_item



end

module IterTypeOf = struct
  type t =
    [
      | `IterTypeOf of
          IterTypeOfType.fields
    ]

  type fields = IterTypeOfType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module IterTypeIn = struct
  type t =
    [
      | `IterTypeIn of
          IterTypeInType.fields
    ]

  type fields = IterTypeInType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module IterType = struct
  type t =
    [
      | IterTypeIn.t
      | IterTypeOf.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module InterfaceKindTask = struct
  type t =
    [
      | `InterfaceKindTask of
          InterfaceKindTaskType.fields
    ]

  type fields = InterfaceKindTaskType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module InterfaceKindSynchronized = struct
  type t =
    [
      | `InterfaceKindSynchronized of
          InterfaceKindSynchronizedType.fields
    ]

  type fields = InterfaceKindSynchronizedType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module InterfaceKindProtected = struct
  type t =
    [
      | `InterfaceKindProtected of
          InterfaceKindProtectedType.fields
    ]

  type fields = InterfaceKindProtectedType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module InterfaceKindLimited = struct
  type t =
    [
      | `InterfaceKindLimited of
          InterfaceKindLimitedType.fields
    ]

  type fields = InterfaceKindLimitedType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module InterfaceKind = struct
  type t =
    [
      | InterfaceKindLimited.t
      | InterfaceKindProtected.t
      | InterfaceKindSynchronized.t
      | InterfaceKindTask.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module HandledStmts = struct
  type t =
    [
      | `HandledStmts of
          HandledStmtsType.fields
    ]

  type fields = HandledStmtsType.fields =
    
  {
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_exceptions: AdaNodeListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_stmts node =
    match (node :> HandledStmtsType.t) with
    | `HandledStmts fields ->
        Lazy.force fields.f_stmts
  let f_exceptions node =
    match (node :> HandledStmtsType.t) with
    | `HandledStmts fields ->
        Lazy.force fields.f_exceptions



end

module UnOp = struct
  type t =
    [
      | `UnOp of
          UnOpType.fields
    ]

  type fields = UnOpType.fields =
    
  {
         
    f_op: [
      | `OpAbs
          of OpAbsType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpNot
          of OpNotType.fields
      | `OpPlus
          of OpPlusType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_op node =
    match (node :> UnOpType.t) with
    | `UnOp fields ->
        Lazy.force fields.f_op
  let f_expr node =
    match (node :> UnOpType.t) with
    | `UnOp fields ->
        Lazy.force fields.f_expr



end

module RaiseExpr = struct
  type t =
    [
      | `RaiseExpr of
          RaiseExprType.fields
    ]

  type fields = RaiseExprType.fields =
    
  {
         
    f_exception_name: NameType.t
    option
    Lazy.t;
         
    f_error_message: ExprType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_exception_name node =
    match (node :> RaiseExprType.t) with
    | `RaiseExpr fields ->
        Lazy.force fields.f_exception_name
  let f_error_message node =
    match (node :> RaiseExprType.t) with
    | `RaiseExpr fields ->
        Lazy.force fields.f_error_message



end

module QuantifiedExpr = struct
  type t =
    [
      | `QuantifiedExpr of
          QuantifiedExprType.fields
    ]

  type fields = QuantifiedExprType.fields =
    
  {
         
    f_quantifier: QuantifierType.t
    Lazy.t;
         
    f_loop_spec: ForLoopSpecType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_quantifier node =
    match (node :> QuantifiedExprType.t) with
    | `QuantifiedExpr fields ->
        Lazy.force fields.f_quantifier
  let f_loop_spec node =
    match (node :> QuantifiedExprType.t) with
    | `QuantifiedExpr fields ->
        Lazy.force fields.f_loop_spec
  let f_expr node =
    match (node :> QuantifiedExprType.t) with
    | `QuantifiedExpr fields ->
        Lazy.force fields.f_expr



end

module ParenExpr = struct
  type t =
    [
      | `ParenExpr of
          ParenExprType.fields
    ]

  type fields = ParenExprType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_expr node =
    match (node :> ParenExprType.t) with
    | `ParenExpr fields ->
        Lazy.force fields.f_expr



end

module TargetName = struct
  type t =
    [
      | `TargetName of
          TargetNameType.fields
    ]

  type fields = TargetNameType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module RealLiteral = struct
  type t =
    [
      | `RealLiteral of
          RealLiteralType.fields
    ]

  type fields = RealLiteralType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module IntLiteral = struct
  type t =
    [
      | `IntLiteral of
          IntLiteralType.fields
    ]

  type fields = IntLiteralType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_denoted_value
    (node)
    =
      let result_ptr =
        allocate_n BigInteger.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.int_literal_p_denoted_value
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      BigInteger.wrap (!@ result_ptr)





end

module NumLiteral = struct
  type t =
    [
      | IntLiteral.t
      | RealLiteral.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module NullLiteral = struct
  type t =
    [
      | `NullLiteral of
          NullLiteralType.fields
    ]

  type fields = NullLiteralType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module StringLiteral = struct
  type t =
    [
      | `StringLiteral of
          StringLiteralType.fields
    ]

  type fields = StringLiteralType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_denoted_value
    (node)
    =
      let result_ptr =
        allocate_n TextTypeStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.string_literal_p_denoted_value
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      TextType.wrap (!@ result_ptr)





end

module OpXor = struct
  type t =
    [
      | `OpXor of
          OpXorType.fields
    ]

  type fields = OpXorType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpRem = struct
  type t =
    [
      | `OpRem of
          OpRemType.fields
    ]

  type fields = OpRemType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpPow = struct
  type t =
    [
      | `OpPow of
          OpPowType.fields
    ]

  type fields = OpPowType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpPlus = struct
  type t =
    [
      | `OpPlus of
          OpPlusType.fields
    ]

  type fields = OpPlusType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpOrElse = struct
  type t =
    [
      | `OpOrElse of
          OpOrElseType.fields
    ]

  type fields = OpOrElseType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpOr = struct
  type t =
    [
      | `OpOr of
          OpOrType.fields
    ]

  type fields = OpOrType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpNotIn = struct
  type t =
    [
      | `OpNotIn of
          OpNotInType.fields
    ]

  type fields = OpNotInType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpNot = struct
  type t =
    [
      | `OpNot of
          OpNotType.fields
    ]

  type fields = OpNotType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpNeq = struct
  type t =
    [
      | `OpNeq of
          OpNeqType.fields
    ]

  type fields = OpNeqType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpMult = struct
  type t =
    [
      | `OpMult of
          OpMultType.fields
    ]

  type fields = OpMultType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpMod = struct
  type t =
    [
      | `OpMod of
          OpModType.fields
    ]

  type fields = OpModType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpMinus = struct
  type t =
    [
      | `OpMinus of
          OpMinusType.fields
    ]

  type fields = OpMinusType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpLte = struct
  type t =
    [
      | `OpLte of
          OpLteType.fields
    ]

  type fields = OpLteType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpLt = struct
  type t =
    [
      | `OpLt of
          OpLtType.fields
    ]

  type fields = OpLtType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpIn = struct
  type t =
    [
      | `OpIn of
          OpInType.fields
    ]

  type fields = OpInType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpGte = struct
  type t =
    [
      | `OpGte of
          OpGteType.fields
    ]

  type fields = OpGteType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpGt = struct
  type t =
    [
      | `OpGt of
          OpGtType.fields
    ]

  type fields = OpGtType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpEq = struct
  type t =
    [
      | `OpEq of
          OpEqType.fields
    ]

  type fields = OpEqType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpDoubleDot = struct
  type t =
    [
      | `OpDoubleDot of
          OpDoubleDotType.fields
    ]

  type fields = OpDoubleDotType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpDiv = struct
  type t =
    [
      | `OpDiv of
          OpDivType.fields
    ]

  type fields = OpDivType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpConcat = struct
  type t =
    [
      | `OpConcat of
          OpConcatType.fields
    ]

  type fields = OpConcatType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpAndThen = struct
  type t =
    [
      | `OpAndThen of
          OpAndThenType.fields
    ]

  type fields = OpAndThenType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpAnd = struct
  type t =
    [
      | `OpAnd of
          OpAndType.fields
    ]

  type fields = OpAndType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module OpAbs = struct
  type t =
    [
      | `OpAbs of
          OpAbsType.fields
    ]

  type fields = OpAbsType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module Op = struct
  type t =
    [
      | OpAbs.t
      | OpAnd.t
      | OpAndThen.t
      | OpConcat.t
      | OpDiv.t
      | OpDoubleDot.t
      | OpEq.t
      | OpGt.t
      | OpGte.t
      | OpIn.t
      | OpLt.t
      | OpLte.t
      | OpMinus.t
      | OpMod.t
      | OpMult.t
      | OpNeq.t
      | OpNot.t
      | OpNotIn.t
      | OpOr.t
      | OpOrElse.t
      | OpPlus.t
      | OpPow.t
      | OpRem.t
      | OpXor.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module Identifier = struct
  type t =
    [
      | `Identifier of
          IdentifierType.fields
    ]

  type fields = IdentifierType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module CharLiteral = struct
  type t =
    [
      | `CharLiteral of
          CharLiteralType.fields
    ]

  type fields = CharLiteralType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_denoted_value
    (node)
    =
      let result_ptr =
        allocate_n Character.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.char_literal_p_denoted_value
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module BaseId = struct
  type t =
    [
      | CharLiteral.t
      | Identifier.t
      | Op.t
      | StringLiteral.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module SingleTokNode = struct
  type t =
    [
      | BaseId.t
      | NullLiteral.t
      | NumLiteral.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_canonical_text
    (node)
    =
      let result_ptr =
        allocate_n Symbol.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.single_tok_node_p_canonical_text
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      Symbol.wrap (!@ result_ptr)





end

module QualExpr = struct
  type t =
    [
      | `QualExpr of
          QualExprType.fields
    ]

  type fields = QualExprType.fields =
    
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `Aggregate
          of AggregateType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_prefix node =
    match (node :> QualExprType.t) with
    | `QualExpr fields ->
        Lazy.force fields.f_prefix
  let f_suffix node =
    match (node :> QualExprType.t) with
    | `QualExpr fields ->
        Lazy.force fields.f_suffix



end

module ExplicitDeref = struct
  type t =
    [
      | `ExplicitDeref of
          ExplicitDerefType.fields
    ]

  type fields = ExplicitDerefType.fields =
    
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_prefix node =
    match (node :> ExplicitDerefType.t) with
    | `ExplicitDeref fields ->
        Lazy.force fields.f_prefix



end

module EndName = struct
  type t =
    [
      | `EndName of
          EndNameType.fields
    ]

  type fields = EndNameType.fields =
    
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_basic_decl
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.end_name_p_basic_decl
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))


  let f_name node =
    match (node :> EndNameType.t) with
    | `EndName fields ->
        Lazy.force fields.f_name



end

module DottedName = struct
  type t =
    [
      | `DottedName of
          DottedNameType.fields
    ]

  type fields = DottedNameType.fields =
    
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `CharLiteral
          of CharLiteralType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_prefix node =
    match (node :> DottedNameType.t) with
    | `DottedName fields ->
        Lazy.force fields.f_prefix
  let f_suffix node =
    match (node :> DottedNameType.t) with
    | `DottedName fields ->
        Lazy.force fields.f_suffix



end

module DiscreteSubtypeName = struct
  type t =
    [
      | `DiscreteSubtypeName of
          DiscreteSubtypeNameType.fields
    ]

  type fields = DiscreteSubtypeNameType.fields =
    
  {
         
    f_subtype: DiscreteSubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_subtype node =
    match (node :> DiscreteSubtypeNameType.t) with
    | `DiscreteSubtypeName fields ->
        Lazy.force fields.f_subtype



end

module DefiningName = struct
  type t =
    [
      | `DefiningName of
          DefiningNameType.fields
    ]

  type fields = DefiningNameType.fields =
    
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_basic_decl
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.defining_name_p_basic_decl
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_find_refs
    ?(imprecise_fallback=false)
    (node)
    (root)
    (origin)
    =
      let result_ptr =
        allocate_n RefResultArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_root =
            
        addr (AdaNodeType.unwrap (root))
      in
      let c_origin =
            
        addr (AdaNodeType.unwrap (origin))
      in
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.defining_name_p_find_refs
          (addr (AdaNodeType.unwrap (node)))
          c_root
          c_origin
          c_imprecise_fallback
          (result_ptr)
      in
         
         
         
      RefResultArray.wrap (context node) (!@ result_ptr)

let p_find_all_references
    ?(follow_renamings=false)
    ?(imprecise_fallback=false)
    (node)
    (units)
    =
      let result_ptr =
        allocate_n RefResultArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_units =
            
        AnalysisUnitArray.unwrap (units)
      in
      let c_follow_renamings =
            
        follow_renamings
      in
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.defining_name_p_find_all_references
          (addr (AdaNodeType.unwrap (node)))
          c_units
          c_follow_renamings
          c_imprecise_fallback
          (result_ptr)
      in
         
               AnalysisUnitArrayStruct.dec_ref c_units ;
         
         
      RefResultArray.wrap (context node) (!@ result_ptr)

let p_find_all_calls
    ?(follow_renamings=false)
    ?(imprecise_fallback=false)
    (node)
    (units)
    =
      let result_ptr =
        allocate_n RefResultArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_units =
            
        AnalysisUnitArray.unwrap (units)
      in
      let c_follow_renamings =
            
        follow_renamings
      in
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.defining_name_p_find_all_calls
          (addr (AdaNodeType.unwrap (node)))
          c_units
          c_follow_renamings
          c_imprecise_fallback
          (result_ptr)
      in
         
               AnalysisUnitArrayStruct.dec_ref c_units ;
         
         
      RefResultArray.wrap (context node) (!@ result_ptr)

let p_next_part
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.defining_name_p_next_part
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (DefiningNameType.wrap (context node) (!@ result_ptr))

let p_previous_part
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.defining_name_p_previous_part
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (DefiningNameType.wrap (context node) (!@ result_ptr))

let p_canonical_part
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.defining_name_p_canonical_part
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (DefiningNameType.wrap (context node) (!@ result_ptr))


  let f_name node =
    match (node :> DefiningNameType.t) with
    | `DefiningName fields ->
        Lazy.force fields.f_name



end

module CallExpr = struct
  type t =
    [
      | `CallExpr of
          CallExprType.fields
    ]

  type fields = CallExprType.fields =
    
  {
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `AssocList
          of AssocListType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_is_array_slice
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.call_expr_p_is_array_slice
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr


  let f_name node =
    match (node :> CallExprType.t) with
    | `CallExpr fields ->
        Lazy.force fields.f_name
  let f_suffix node =
    match (node :> CallExprType.t) with
    | `CallExpr fields ->
        Lazy.force fields.f_suffix



end

module UpdateAttributeRef = struct
  type t =
    [
      | `UpdateAttributeRef of
          UpdateAttributeRefType.fields
    ]

  type fields = UpdateAttributeRefType.fields =
    
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of fields
    ]
    Lazy.t;
         
    f_attribute: IdentifierType.t
    Lazy.t;
         
    f_args: AdaNodeType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_prefix node =
    match (node :> UpdateAttributeRefType.t) with
    | `UpdateAttributeRef fields ->
        Lazy.force fields.f_prefix
  let f_attribute node =
    match (node :> UpdateAttributeRefType.t) with
    | `UpdateAttributeRef fields ->
        Lazy.force fields.f_attribute
  let f_args node =
    match (node :> UpdateAttributeRefType.t) with
    | `UpdateAttributeRef fields ->
        Lazy.force fields.f_args



end

module AttributeRef = struct
  type t =
    [
      | `AttributeRef of
          AttributeRefType.fields
      | `UpdateAttributeRef of
          UpdateAttributeRefType.fields
    ]

  type fields = AttributeRefType.fields =
    
  {
         
    f_prefix: [
      | `AttributeRef
          of fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_attribute: IdentifierType.t
    Lazy.t;
         
    f_args: AdaNodeType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_prefix node =
    match (node :> AttributeRefType.t) with
    | `AttributeRef fields ->
        Lazy.force fields.f_prefix
    | `UpdateAttributeRef fields ->
        Lazy.force fields.f_prefix
  let f_attribute node =
    match (node :> AttributeRefType.t) with
    | `AttributeRef fields ->
        Lazy.force fields.f_attribute
    | `UpdateAttributeRef fields ->
        Lazy.force fields.f_attribute
  let f_args node =
    match (node :> AttributeRefType.t) with
    | `AttributeRef fields ->
        Lazy.force fields.f_args
    | `UpdateAttributeRef fields ->
        Lazy.force fields.f_args



end

module Name = struct
  type t =
    [
      | AttributeRef.t
      | CallExpr.t
      | DefiningName.t
      | DiscreteSubtypeName.t
      | DottedName.t
      | EndName.t
      | ExplicitDeref.t
      | QualExpr.t
      | SingleTokNode.t
      | TargetName.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_enclosing_defining_name
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.name_p_enclosing_defining_name
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (DefiningNameType.wrap (context node) (!@ result_ptr))

let p_is_defining
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.name_p_is_defining
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_name_is
    (node)
    (sym)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_sym =
            
        allocate Symbol.c_type (Symbol.unwrap (context node).c_value (sym))
      in
      let _ : int =
        CFunctions.name_p_name_is
          (addr (AdaNodeType.unwrap (node)))
          c_sym
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_direct_call
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.name_p_is_direct_call
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_is_access_call
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.name_p_is_access_call
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_is_call
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.name_p_is_call
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_is_dot_call
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.name_p_is_dot_call
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      !@ result_ptr

let p_failsafe_referenced_def_name
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n RefdDefStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.name_p_failsafe_referenced_def_name
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      RefdDef.wrap (context node) (!@ result_ptr)

let p_referenced_defining_name
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.name_p_referenced_defining_name
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (DefiningNameType.wrap (context node) (!@ result_ptr))

let p_all_env_elements
    ?(seq=true)
    ?seq_from
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_seq =
            
        seq
      in
      let c_seq_from =
           match seq_from with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.name_p_all_env_elements
          (addr (AdaNodeType.unwrap (node)))
          c_seq
          c_seq_from
          (result_ptr)
      in
         
         
      AdaNodeArray.wrap (context node) (!@ result_ptr)

let p_called_subp_spec
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.name_p_called_subp_spec
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseFormalParamHolderType.wrap (context node) (!@ result_ptr))

let p_referenced_decl
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.name_p_referenced_decl
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_failsafe_referenced_decl
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n RefdDeclStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.name_p_failsafe_referenced_decl
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      RefdDecl.wrap (context node) (!@ result_ptr)

let p_referenced_decl_internal
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n RefdDeclStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.name_p_referenced_decl_internal
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      RefdDecl.wrap (context node) (!@ result_ptr)

let p_name_designated_type
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.name_p_name_designated_type
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_is_static_subtype
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.name_p_is_static_subtype
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      !@ result_ptr

let p_name_matches
    (node)
    (n)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_n =
            
        addr (AdaNodeType.unwrap (n))
      in
      let _ : int =
        CFunctions.name_p_name_matches
          (addr (AdaNodeType.unwrap (node)))
          c_n
          (result_ptr)
      in
         
      !@ result_ptr

let p_relative_name
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.name_p_relative_name
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (SingleTokNodeType.wrap (context node) (!@ result_ptr))

let p_is_operator_name
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.name_p_is_operator_name
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_is_write_reference
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.name_p_is_write_reference
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_static_call
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.name_p_is_static_call
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      !@ result_ptr

let p_as_symbol_array
    (node)
    =
      let result_ptr =
        allocate_n UnboundedTextTypeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.name_p_as_symbol_array
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      UnboundedTextTypeArray.wrap (!@ result_ptr)





end

module MembershipExpr = struct
  type t =
    [
      | `MembershipExpr of
          MembershipExprType.fields
    ]

  type fields = MembershipExprType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpIn
          of OpInType.fields
      | `OpNotIn
          of OpNotInType.fields
    ]
    Lazy.t;
         
    f_membership_exprs: ExprListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_expr node =
    match (node :> MembershipExprType.t) with
    | `MembershipExpr fields ->
        Lazy.force fields.f_expr
  let f_op node =
    match (node :> MembershipExprType.t) with
    | `MembershipExpr fields ->
        Lazy.force fields.f_op
  let f_membership_exprs node =
    match (node :> MembershipExprType.t) with
    | `MembershipExpr fields ->
        Lazy.force fields.f_membership_exprs



end

module DeclExpr = struct
  type t =
    [
      | `DeclExpr of
          DeclExprType.fields
    ]

  type fields = DeclExprType.fields =
    
  {
         
    f_decls: BasicDeclListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decls node =
    match (node :> DeclExprType.t) with
    | `DeclExpr fields ->
        Lazy.force fields.f_decls
  let f_expr node =
    match (node :> DeclExprType.t) with
    | `DeclExpr fields ->
        Lazy.force fields.f_expr



end

module ContractCases = struct
  type t =
    [
      | `ContractCases of
          ContractCasesType.fields
    ]

  type fields = ContractCasesType.fields =
    
  {
         
    f_contract_cases: ContractCaseAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_contract_cases node =
    match (node :> ContractCasesType.t) with
    | `ContractCases fields ->
        Lazy.force fields.f_contract_cases



end

module IfExpr = struct
  type t =
    [
      | `IfExpr of
          IfExprType.fields
    ]

  type fields = IfExprType.fields =
    
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_alternatives: ElsifExprPartListType.t
    Lazy.t;
         
    f_else_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_cond_expr node =
    match (node :> IfExprType.t) with
    | `IfExpr fields ->
        Lazy.force fields.f_cond_expr
  let f_then_expr node =
    match (node :> IfExprType.t) with
    | `IfExpr fields ->
        Lazy.force fields.f_then_expr
  let f_alternatives node =
    match (node :> IfExprType.t) with
    | `IfExpr fields ->
        Lazy.force fields.f_alternatives
  let f_else_expr node =
    match (node :> IfExprType.t) with
    | `IfExpr fields ->
        Lazy.force fields.f_else_expr



end

module CaseExpr = struct
  type t =
    [
      | `CaseExpr of
          CaseExprType.fields
    ]

  type fields = CaseExprType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_cases: CaseExprAlternativeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_expr node =
    match (node :> CaseExprType.t) with
    | `CaseExpr fields ->
        Lazy.force fields.f_expr
  let f_cases node =
    match (node :> CaseExprType.t) with
    | `CaseExpr fields ->
        Lazy.force fields.f_cases



end

module CondExpr = struct
  type t =
    [
      | CaseExpr.t
      | IfExpr.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_dependent_exprs
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.cond_expr_p_dependent_exprs
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      ExprArray.wrap (context node) (!@ result_ptr)





end

module CaseExprAlternative = struct
  type t =
    [
      | `CaseExprAlternative of
          CaseExprAlternativeType.fields
    ]

  type fields = CaseExprAlternativeType.fields =
    
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_choices node =
    match (node :> CaseExprAlternativeType.t) with
    | `CaseExprAlternative fields ->
        Lazy.force fields.f_choices
  let f_expr node =
    match (node :> CaseExprAlternativeType.t) with
    | `CaseExprAlternative fields ->
        Lazy.force fields.f_expr



end

module BoxExpr = struct
  type t =
    [
      | `BoxExpr of
          BoxExprType.fields
    ]

  type fields = BoxExprType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module RelationOp = struct
  type t =
    [
      | `RelationOp of
          RelationOpType.fields
    ]

  type fields = RelationOpType.fields =
    
  {
         
    f_left: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpAnd
          of OpAndType.fields
      | `OpAndThen
          of OpAndThenType.fields
      | `OpConcat
          of OpConcatType.fields
      | `OpDiv
          of OpDivType.fields
      | `OpDoubleDot
          of OpDoubleDotType.fields
      | `OpEq
          of OpEqType.fields
      | `OpGt
          of OpGtType.fields
      | `OpGte
          of OpGteType.fields
      | `OpLt
          of OpLtType.fields
      | `OpLte
          of OpLteType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpMod
          of OpModType.fields
      | `OpMult
          of OpMultType.fields
      | `OpNeq
          of OpNeqType.fields
      | `OpOr
          of OpOrType.fields
      | `OpOrElse
          of OpOrElseType.fields
      | `OpPlus
          of OpPlusType.fields
      | `OpPow
          of OpPowType.fields
      | `OpRem
          of OpRemType.fields
      | `OpXor
          of OpXorType.fields
    ]
    Lazy.t;
         
    f_right: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_left node =
    match (node :> RelationOpType.t) with
    | `RelationOp fields ->
        Lazy.force fields.f_left
  let f_op node =
    match (node :> RelationOpType.t) with
    | `RelationOp fields ->
        Lazy.force fields.f_op
  let f_right node =
    match (node :> RelationOpType.t) with
    | `RelationOp fields ->
        Lazy.force fields.f_right



end

module BinOp = struct
  type t =
    [
      | `BinOp of
          BinOpType.fields
      | `RelationOp of
          RelationOpType.fields
    ]

  type fields = BinOpType.fields =
    
  {
         
    f_left: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpAnd
          of OpAndType.fields
      | `OpAndThen
          of OpAndThenType.fields
      | `OpConcat
          of OpConcatType.fields
      | `OpDiv
          of OpDivType.fields
      | `OpDoubleDot
          of OpDoubleDotType.fields
      | `OpEq
          of OpEqType.fields
      | `OpGt
          of OpGtType.fields
      | `OpGte
          of OpGteType.fields
      | `OpLt
          of OpLtType.fields
      | `OpLte
          of OpLteType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpMod
          of OpModType.fields
      | `OpMult
          of OpMultType.fields
      | `OpNeq
          of OpNeqType.fields
      | `OpOr
          of OpOrType.fields
      | `OpOrElse
          of OpOrElseType.fields
      | `OpPlus
          of OpPlusType.fields
      | `OpPow
          of OpPowType.fields
      | `OpRem
          of OpRemType.fields
      | `OpXor
          of OpXorType.fields
    ]
    Lazy.t;
         
    f_right: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_left node =
    match (node :> BinOpType.t) with
    | `BinOp fields ->
        Lazy.force fields.f_left
    | `RelationOp fields ->
        Lazy.force fields.f_left
  let f_op node =
    match (node :> BinOpType.t) with
    | `BinOp fields ->
        Lazy.force fields.f_op
    | `RelationOp fields ->
        Lazy.force fields.f_op
  let f_right node =
    match (node :> BinOpType.t) with
    | `BinOp fields ->
        Lazy.force fields.f_right
    | `RelationOp fields ->
        Lazy.force fields.f_right



end

module NullRecordAggregate = struct
  type t =
    [
      | `NullRecordAggregate of
          NullRecordAggregateType.fields
    ]

  type fields = NullRecordAggregateType.fields =
    
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ancestor_expr node =
    match (node :> NullRecordAggregateType.t) with
    | `NullRecordAggregate fields ->
        Lazy.force fields.f_ancestor_expr
  let f_assocs node =
    match (node :> NullRecordAggregateType.t) with
    | `NullRecordAggregate fields ->
        Lazy.force fields.f_assocs



end

module BracketDeltaAggregate = struct
  type t =
    [
      | `BracketDeltaAggregate of
          BracketDeltaAggregateType.fields
    ]

  type fields = BracketDeltaAggregateType.fields =
    
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ancestor_expr node =
    match (node :> BracketDeltaAggregateType.t) with
    | `BracketDeltaAggregate fields ->
        Lazy.force fields.f_ancestor_expr
  let f_assocs node =
    match (node :> BracketDeltaAggregateType.t) with
    | `BracketDeltaAggregate fields ->
        Lazy.force fields.f_assocs



end

module DeltaAggregate = struct
  type t =
    [
      | `DeltaAggregate of
          DeltaAggregateType.fields
      | `BracketDeltaAggregate of
          BracketDeltaAggregateType.fields
    ]

  type fields = DeltaAggregateType.fields =
    
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ancestor_expr node =
    match (node :> DeltaAggregateType.t) with
    | `DeltaAggregate fields ->
        Lazy.force fields.f_ancestor_expr
    | `BracketDeltaAggregate fields ->
        Lazy.force fields.f_ancestor_expr
  let f_assocs node =
    match (node :> DeltaAggregateType.t) with
    | `DeltaAggregate fields ->
        Lazy.force fields.f_assocs
    | `BracketDeltaAggregate fields ->
        Lazy.force fields.f_assocs



end

module BracketAggregate = struct
  type t =
    [
      | `BracketAggregate of
          BracketAggregateType.fields
    ]

  type fields = BracketAggregateType.fields =
    
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ancestor_expr node =
    match (node :> BracketAggregateType.t) with
    | `BracketAggregate fields ->
        Lazy.force fields.f_ancestor_expr
  let f_assocs node =
    match (node :> BracketAggregateType.t) with
    | `BracketAggregate fields ->
        Lazy.force fields.f_assocs



end

module Aggregate = struct
  type t =
    [
      | `Aggregate of
          AggregateType.fields
      | `BracketAggregate of
          BracketAggregateType.fields
    ]

  type fields = AggregateType.fields =
    
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ancestor_expr node =
    match (node :> AggregateType.t) with
    | `Aggregate fields ->
        Lazy.force fields.f_ancestor_expr
    | `BracketAggregate fields ->
        Lazy.force fields.f_ancestor_expr
  let f_assocs node =
    match (node :> AggregateType.t) with
    | `Aggregate fields ->
        Lazy.force fields.f_assocs
    | `BracketAggregate fields ->
        Lazy.force fields.f_assocs



end

module BaseAggregate = struct
  type t =
    [
      | Aggregate.t
      | DeltaAggregate.t
      | NullRecordAggregate.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ancestor_expr node =
    match (node :> BaseAggregateType.t) with
    | `Aggregate fields ->
        Lazy.force fields.f_ancestor_expr
    | `BracketAggregate fields ->
        Lazy.force fields.f_ancestor_expr
    | `DeltaAggregate fields ->
        Lazy.force fields.f_ancestor_expr
    | `BracketDeltaAggregate fields ->
        Lazy.force fields.f_ancestor_expr
    | `NullRecordAggregate fields ->
        Lazy.force fields.f_ancestor_expr
  let f_assocs node =
    match (node :> BaseAggregateType.t) with
    | `Aggregate fields ->
        Lazy.force fields.f_assocs
    | `BracketAggregate fields ->
        Lazy.force fields.f_assocs
    | `DeltaAggregate fields ->
        Lazy.force fields.f_assocs
    | `BracketDeltaAggregate fields ->
        Lazy.force fields.f_assocs
    | `NullRecordAggregate fields ->
        Lazy.force fields.f_assocs



end

module Allocator = struct
  type t =
    [
      | `Allocator of
          AllocatorType.fields
    ]

  type fields = AllocatorType.fields =
    
  {
         
    f_subpool: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_type_or_expr: [
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `QualExpr
          of QualExprType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_get_allocated_type
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.allocator_p_get_allocated_type
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))


  let f_subpool node =
    match (node :> AllocatorType.t) with
    | `Allocator fields ->
        Lazy.force fields.f_subpool
  let f_type_or_expr node =
    match (node :> AllocatorType.t) with
    | `Allocator fields ->
        Lazy.force fields.f_type_or_expr



end

module AbstractStateDeclExpr = struct
  type t =
    [
      | `AbstractStateDeclExpr of
          AbstractStateDeclExprType.fields
    ]

  type fields = AbstractStateDeclExprType.fields =
    
  {
         
    f_state_decl: [
      | `AbstractStateDecl
          of AbstractStateDeclType.fields
      | `MultiAbstractStateDecl
          of MultiAbstractStateDeclType.fields
      | `ParenAbstractStateDecl
          of ParenAbstractStateDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_state_decl node =
    match (node :> AbstractStateDeclExprType.t) with
    | `AbstractStateDeclExpr fields ->
        Lazy.force fields.f_state_decl



end

module Expr = struct
  type t =
    [
      | AbstractStateDeclExpr.t
      | Allocator.t
      | BaseAggregate.t
      | BinOp.t
      | BoxExpr.t
      | CaseExprAlternative.t
      | CondExpr.t
      | ContractCases.t
      | DeclExpr.t
      | MembershipExpr.t
      | Name.t
      | ParenExpr.t
      | QuantifiedExpr.t
      | RaiseExpr.t
      | UnOp.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_expression_type
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.expr_p_expression_type
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_is_dynamically_tagged
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.expr_p_is_dynamically_tagged
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_dispatching_call
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.expr_p_is_dispatching_call
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_static_expr
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.expr_p_is_static_expr
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      !@ result_ptr

let p_first_corresponding_decl
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.expr_p_first_corresponding_decl
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_eval_as_int
    (node)
    =
      let result_ptr =
        allocate_n BigInteger.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.expr_p_eval_as_int
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      BigInteger.wrap (!@ result_ptr)

let p_eval_as_int_in_env
    (node)
    (env)
    =
      let result_ptr =
        allocate_n BigInteger.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_env =
            
        SubstitutionArray.unwrap (env)
      in
      let _ : int =
        CFunctions.expr_p_eval_as_int_in_env
          (addr (AdaNodeType.unwrap (node)))
          c_env
          (result_ptr)
      in
         
               SubstitutionArrayStruct.dec_ref c_env ;
      BigInteger.wrap (!@ result_ptr)

let p_eval_as_string
    (node)
    =
      let result_ptr =
        allocate_n TextTypeStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.expr_p_eval_as_string
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      TextType.wrap (!@ result_ptr)

let p_eval_as_string_in_env
    (node)
    (env)
    =
      let result_ptr =
        allocate_n TextTypeStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_env =
            
        SubstitutionArray.unwrap (env)
      in
      let _ : int =
        CFunctions.expr_p_eval_as_string_in_env
          (addr (AdaNodeType.unwrap (node)))
          c_env
          (result_ptr)
      in
         
               SubstitutionArrayStruct.dec_ref c_env ;
      TextType.wrap (!@ result_ptr)

let p_matching_nodes
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.expr_p_matching_nodes
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      AdaNodeArray.wrap (context node) (!@ result_ptr)





end

module ElsifStmtPart = struct
  type t =
    [
      | `ElsifStmtPart of
          ElsifStmtPartType.fields
    ]

  type fields = ElsifStmtPartType.fields =
    
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_cond_expr node =
    match (node :> ElsifStmtPartType.t) with
    | `ElsifStmtPart fields ->
        Lazy.force fields.f_cond_expr
  let f_stmts node =
    match (node :> ElsifStmtPartType.t) with
    | `ElsifStmtPart fields ->
        Lazy.force fields.f_stmts



end

module ElsifExprPart = struct
  type t =
    [
      | `ElsifExprPart of
          ElsifExprPartType.fields
    ]

  type fields = ElsifExprPartType.fields =
    
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_cond_expr node =
    match (node :> ElsifExprPartType.t) with
    | `ElsifExprPart fields ->
        Lazy.force fields.f_cond_expr
  let f_then_expr node =
    match (node :> ElsifExprPartType.t) with
    | `ElsifExprPart fields ->
        Lazy.force fields.f_then_expr



end

module PublicPart = struct
  type t =
    [
      | `PublicPart of
          PublicPartType.fields
    ]

  type fields = PublicPartType.fields =
    
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decls node =
    match (node :> PublicPartType.t) with
    | `PublicPart fields ->
        Lazy.force fields.f_decls



end

module PrivatePart = struct
  type t =
    [
      | `PrivatePart of
          PrivatePartType.fields
    ]

  type fields = PrivatePartType.fields =
    
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decls node =
    match (node :> PrivatePartType.t) with
    | `PrivatePart fields ->
        Lazy.force fields.f_decls



end

module DeclarativePart = struct
  type t =
    [
      | `DeclarativePart of
          DeclarativePartType.fields
      | `PrivatePart of
          PrivatePartType.fields
      | `PublicPart of
          PublicPartType.fields
    ]

  type fields = DeclarativePartType.fields =
    
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decls node =
    match (node :> DeclarativePartType.t) with
    | `DeclarativePart fields ->
        Lazy.force fields.f_decls
    | `PrivatePart fields ->
        Lazy.force fields.f_decls
    | `PublicPart fields ->
        Lazy.force fields.f_decls



end

module RangeConstraint = struct
  type t =
    [
      | `RangeConstraint of
          RangeConstraintType.fields
    ]

  type fields = RangeConstraintType.fields =
    
  {
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_range node =
    match (node :> RangeConstraintType.t) with
    | `RangeConstraint fields ->
        Lazy.force fields.f_range



end

module IndexConstraint = struct
  type t =
    [
      | `IndexConstraint of
          IndexConstraintType.fields
    ]

  type fields = IndexConstraintType.fields =
    
  {
         
    f_constraints: ConstraintListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_constraints node =
    match (node :> IndexConstraintType.t) with
    | `IndexConstraint fields ->
        Lazy.force fields.f_constraints



end

module DiscriminantConstraint = struct
  type t =
    [
      | `DiscriminantConstraint of
          DiscriminantConstraintType.fields
    ]

  type fields = DiscriminantConstraintType.fields =
    
  {
         
    f_constraints: BasicAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_constraints node =
    match (node :> DiscriminantConstraintType.t) with
    | `DiscriminantConstraint fields ->
        Lazy.force fields.f_constraints



end

module DigitsConstraint = struct
  type t =
    [
      | `DigitsConstraint of
          DigitsConstraintType.fields
    ]

  type fields = DigitsConstraintType.fields =
    
  {
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_digits node =
    match (node :> DigitsConstraintType.t) with
    | `DigitsConstraint fields ->
        Lazy.force fields.f_digits
  let f_range node =
    match (node :> DigitsConstraintType.t) with
    | `DigitsConstraint fields ->
        Lazy.force fields.f_range



end

module DeltaConstraint = struct
  type t =
    [
      | `DeltaConstraint of
          DeltaConstraintType.fields
    ]

  type fields = DeltaConstraintType.fields =
    
  {
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_digits node =
    match (node :> DeltaConstraintType.t) with
    | `DeltaConstraint fields ->
        Lazy.force fields.f_digits
  let f_range node =
    match (node :> DeltaConstraintType.t) with
    | `DeltaConstraint fields ->
        Lazy.force fields.f_range



end

module Constraint = struct
  type t =
    [
      | DeltaConstraint.t
      | DigitsConstraint.t
      | DiscriminantConstraint.t
      | IndexConstraint.t
      | RangeConstraint.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ConstantPresent = struct
  type t =
    [
      | `ConstantPresent of
          ConstantPresentType.fields
    ]

  type fields = ConstantPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ConstantAbsent = struct
  type t =
    [
      | `ConstantAbsent of
          ConstantAbsentType.fields
    ]

  type fields = ConstantAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ConstantNode = struct
  type t =
    [
      | ConstantAbsent.t
      | ConstantPresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.constant_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module ComponentDef = struct
  type t =
    [
      | `ComponentDef of
          ComponentDefType.fields
    ]

  type fields = ComponentDefType.fields =
    
  {
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_has_aliased node =
    match (node :> ComponentDefType.t) with
    | `ComponentDef fields ->
        Lazy.force fields.f_has_aliased
  let f_has_constant node =
    match (node :> ComponentDefType.t) with
    | `ComponentDef fields ->
        Lazy.force fields.f_has_constant
  let f_type_expr node =
    match (node :> ComponentDefType.t) with
    | `ComponentDef fields ->
        Lazy.force fields.f_type_expr



end

module ComponentClause = struct
  type t =
    [
      | `ComponentClause of
          ComponentClauseType.fields
    ]

  type fields = ComponentClauseType.fields =
    
  {
         
    f_id: IdentifierType.t
    Lazy.t;
         
    f_position: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_id node =
    match (node :> ComponentClauseType.t) with
    | `ComponentClause fields ->
        Lazy.force fields.f_id
  let f_position node =
    match (node :> ComponentClauseType.t) with
    | `ComponentClause fields ->
        Lazy.force fields.f_position
  let f_range node =
    match (node :> ComponentClauseType.t) with
    | `ComponentClause fields ->
        Lazy.force fields.f_range



end

module CompilationUnit = struct
  type t =
    [
      | `CompilationUnit of
          CompilationUnitType.fields
    ]

  type fields = CompilationUnitType.fields =
    
  {
         
    f_prelude: AdaNodeListType.t
    Lazy.t;
         
    f_body: [
      | `LibraryItem
          of LibraryItemType.fields
      | `Subunit
          of SubunitType.fields
    ]
    Lazy.t;
         
    f_pragmas: PragmaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_syntactic_fully_qualified_name
    (node)
    =
      let result_ptr =
        allocate_n UnboundedTextTypeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.compilation_unit_p_syntactic_fully_qualified_name
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      UnboundedTextTypeArray.wrap (!@ result_ptr)

let p_unit_kind
    (node)
    =
      let result_ptr =
        allocate_n AnalysisUnitKind.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.compilation_unit_p_unit_kind
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_withed_units
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.compilation_unit_p_withed_units
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      CompilationUnitArray.wrap (context node) (!@ result_ptr)

let p_imported_units
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.compilation_unit_p_imported_units
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      CompilationUnitArray.wrap (context node) (!@ result_ptr)

let p_unit_dependencies
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.compilation_unit_p_unit_dependencies
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      CompilationUnitArray.wrap (context node) (!@ result_ptr)

let p_decl
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.compilation_unit_p_decl
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_is_preelaborable
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.compilation_unit_p_is_preelaborable
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      !@ result_ptr


  let f_prelude node =
    match (node :> CompilationUnitType.t) with
    | `CompilationUnit fields ->
        Lazy.force fields.f_prelude
  let f_body node =
    match (node :> CompilationUnitType.t) with
    | `CompilationUnit fields ->
        Lazy.force fields.f_body
  let f_pragmas node =
    match (node :> CompilationUnitType.t) with
    | `CompilationUnit fields ->
        Lazy.force fields.f_pragmas



end

module CaseStmtAlternative = struct
  type t =
    [
      | `CaseStmtAlternative of
          CaseStmtAlternativeType.fields
    ]

  type fields = CaseStmtAlternativeType.fields =
    
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_choices node =
    match (node :> CaseStmtAlternativeType.t) with
    | `CaseStmtAlternative fields ->
        Lazy.force fields.f_choices
  let f_stmts node =
    match (node :> CaseStmtAlternativeType.t) with
    | `CaseStmtAlternative fields ->
        Lazy.force fields.f_stmts



end

module SingleTaskDecl = struct
  type t =
    [
      | `SingleTaskDecl of
          SingleTaskDeclType.fields
    ]

  type fields = SingleTaskDeclType.fields =
    
  {
         
    f_task_type: SingleTaskTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_task_type node =
    match (node :> SingleTaskDeclType.t) with
    | `SingleTaskDecl fields ->
        Lazy.force fields.f_task_type



end

module SingleProtectedDecl = struct
  type t =
    [
      | `SingleProtectedDecl of
          SingleProtectedDeclType.fields
    ]

  type fields = SingleProtectedDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_definition: ProtectedDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> SingleProtectedDeclType.t) with
    | `SingleProtectedDecl fields ->
        Lazy.force fields.f_name
  let f_interfaces node =
    match (node :> SingleProtectedDeclType.t) with
    | `SingleProtectedDecl fields ->
        Lazy.force fields.f_interfaces
  let f_definition node =
    match (node :> SingleProtectedDeclType.t) with
    | `SingleProtectedDecl fields ->
        Lazy.force fields.f_definition



end

module PackageRenamingDecl = struct
  type t =
    [
      | `PackageRenamingDecl of
          PackageRenamingDeclType.fields
    ]

  type fields = PackageRenamingDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_renamed_package
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.package_renaming_decl_p_renamed_package
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_final_renamed_package
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.package_renaming_decl_p_final_renamed_package
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))


  let f_name node =
    match (node :> PackageRenamingDeclType.t) with
    | `PackageRenamingDecl fields ->
        Lazy.force fields.f_name
  let f_renames node =
    match (node :> PackageRenamingDeclType.t) with
    | `PackageRenamingDecl fields ->
        Lazy.force fields.f_renames



end

module ExtendedReturnStmtObjectDecl = struct
  type t =
    [
      | `ExtendedReturnStmtObjectDecl of
          ExtendedReturnStmtObjectDeclType.fields
    ]

  type fields = ExtendedReturnStmtObjectDeclType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_renaming_clause: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ids node =
    match (node :> ExtendedReturnStmtObjectDeclType.t) with
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_ids
  let f_has_aliased node =
    match (node :> ExtendedReturnStmtObjectDeclType.t) with
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_has_aliased
  let f_has_constant node =
    match (node :> ExtendedReturnStmtObjectDeclType.t) with
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_has_constant
  let f_mode node =
    match (node :> ExtendedReturnStmtObjectDeclType.t) with
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_mode
  let f_type_expr node =
    match (node :> ExtendedReturnStmtObjectDeclType.t) with
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_type_expr
  let f_default_expr node =
    match (node :> ExtendedReturnStmtObjectDeclType.t) with
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_default_expr
  let f_renaming_clause node =
    match (node :> ExtendedReturnStmtObjectDeclType.t) with
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_renaming_clause



end

module ObjectDecl = struct
  type t =
    [
      | `ObjectDecl of
          ObjectDeclType.fields
      | `ExtendedReturnStmtObjectDecl of
          ExtendedReturnStmtObjectDeclType.fields
    ]

  type fields = ObjectDeclType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_renaming_clause: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_public_part_decl
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.object_decl_p_public_part_decl
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_private_part_decl
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.object_decl_p_private_part_decl
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))


  let f_ids node =
    match (node :> ObjectDeclType.t) with
    | `ObjectDecl fields ->
        Lazy.force fields.f_ids
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_ids
  let f_has_aliased node =
    match (node :> ObjectDeclType.t) with
    | `ObjectDecl fields ->
        Lazy.force fields.f_has_aliased
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_has_aliased
  let f_has_constant node =
    match (node :> ObjectDeclType.t) with
    | `ObjectDecl fields ->
        Lazy.force fields.f_has_constant
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_has_constant
  let f_mode node =
    match (node :> ObjectDeclType.t) with
    | `ObjectDecl fields ->
        Lazy.force fields.f_mode
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_mode
  let f_type_expr node =
    match (node :> ObjectDeclType.t) with
    | `ObjectDecl fields ->
        Lazy.force fields.f_type_expr
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_type_expr
  let f_default_expr node =
    match (node :> ObjectDeclType.t) with
    | `ObjectDecl fields ->
        Lazy.force fields.f_default_expr
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_default_expr
  let f_renaming_clause node =
    match (node :> ObjectDeclType.t) with
    | `ObjectDecl fields ->
        Lazy.force fields.f_renaming_clause
    | `ExtendedReturnStmtObjectDecl fields ->
        Lazy.force fields.f_renaming_clause



end

module NumberDecl = struct
  type t =
    [
      | `NumberDecl of
          NumberDeclType.fields
    ]

  type fields = NumberDeclType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ids node =
    match (node :> NumberDeclType.t) with
    | `NumberDecl fields ->
        Lazy.force fields.f_ids
  let f_expr node =
    match (node :> NumberDeclType.t) with
    | `NumberDecl fields ->
        Lazy.force fields.f_expr



end

module NamedStmtDecl = struct
  type t =
    [
      | `NamedStmtDecl of
          NamedStmtDeclType.fields
    ]

  type fields = NamedStmtDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> NamedStmtDeclType.t) with
    | `NamedStmtDecl fields ->
        Lazy.force fields.f_name



end

module LabelDecl = struct
  type t =
    [
      | `LabelDecl of
          LabelDeclType.fields
    ]

  type fields = LabelDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> LabelDeclType.t) with
    | `LabelDecl fields ->
        Lazy.force fields.f_name



end

module GenericSubpRenamingDecl = struct
  type t =
    [
      | `GenericSubpRenamingDecl of
          GenericSubpRenamingDeclType.fields
    ]

  type fields = GenericSubpRenamingDeclType.fields =
    
  {
         
    f_kind: SubpKindType.t
    Lazy.t;
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_kind node =
    match (node :> GenericSubpRenamingDeclType.t) with
    | `GenericSubpRenamingDecl fields ->
        Lazy.force fields.f_kind
  let f_name node =
    match (node :> GenericSubpRenamingDeclType.t) with
    | `GenericSubpRenamingDecl fields ->
        Lazy.force fields.f_name
  let f_renames node =
    match (node :> GenericSubpRenamingDeclType.t) with
    | `GenericSubpRenamingDecl fields ->
        Lazy.force fields.f_renames



end

module GenericPackageRenamingDecl = struct
  type t =
    [
      | `GenericPackageRenamingDecl of
          GenericPackageRenamingDeclType.fields
    ]

  type fields = GenericPackageRenamingDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> GenericPackageRenamingDeclType.t) with
    | `GenericPackageRenamingDecl fields ->
        Lazy.force fields.f_name
  let f_renames node =
    match (node :> GenericPackageRenamingDeclType.t) with
    | `GenericPackageRenamingDecl fields ->
        Lazy.force fields.f_renames



end

module GenericRenamingDecl = struct
  type t =
    [
      | GenericPackageRenamingDecl.t
      | GenericSubpRenamingDecl.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module GenericSubpInstantiation = struct
  type t =
    [
      | `GenericSubpInstantiation of
          GenericSubpInstantiationType.fields
    ]

  type fields = GenericSubpInstantiationType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_kind: SubpKindType.t
    Lazy.t;
         
    f_subp_name: DefiningNameType.t
    Lazy.t;
         
    f_generic_subp_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_params: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_designated_subp
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.generic_subp_instantiation_p_designated_subp
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AdaNodeType.wrap (context node) (!@ result_ptr))


  let f_overriding node =
    match (node :> GenericSubpInstantiationType.t) with
    | `GenericSubpInstantiation fields ->
        Lazy.force fields.f_overriding
  let f_kind node =
    match (node :> GenericSubpInstantiationType.t) with
    | `GenericSubpInstantiation fields ->
        Lazy.force fields.f_kind
  let f_subp_name node =
    match (node :> GenericSubpInstantiationType.t) with
    | `GenericSubpInstantiation fields ->
        Lazy.force fields.f_subp_name
  let f_generic_subp_name node =
    match (node :> GenericSubpInstantiationType.t) with
    | `GenericSubpInstantiation fields ->
        Lazy.force fields.f_generic_subp_name
  let f_params node =
    match (node :> GenericSubpInstantiationType.t) with
    | `GenericSubpInstantiation fields ->
        Lazy.force fields.f_params



end

module GenericPackageInstantiation = struct
  type t =
    [
      | `GenericPackageInstantiation of
          GenericPackageInstantiationType.fields
    ]

  type fields = GenericPackageInstantiationType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_generic_pkg_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_params: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> GenericPackageInstantiationType.t) with
    | `GenericPackageInstantiation fields ->
        Lazy.force fields.f_name
  let f_generic_pkg_name node =
    match (node :> GenericPackageInstantiationType.t) with
    | `GenericPackageInstantiation fields ->
        Lazy.force fields.f_generic_pkg_name
  let f_params node =
    match (node :> GenericPackageInstantiationType.t) with
    | `GenericPackageInstantiation fields ->
        Lazy.force fields.f_params



end

module GenericInstantiation = struct
  type t =
    [
      | GenericPackageInstantiation.t
      | GenericSubpInstantiation.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_designated_generic_decl
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.generic_instantiation_p_designated_generic_decl
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))





end

module GenericSubpDecl = struct
  type t =
    [
      | `GenericSubpDecl of
          GenericSubpDeclType.fields
    ]

  type fields = GenericSubpDeclType.fields =
    
  {
         
    f_formal_part: GenericFormalPartType.t
    Lazy.t;
         
    f_subp_decl: GenericSubpInternalType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_body_part
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.generic_subp_decl_p_body_part
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseSubpBodyType.wrap (context node) (!@ result_ptr))


  let f_formal_part node =
    match (node :> GenericSubpDeclType.t) with
    | `GenericSubpDecl fields ->
        Lazy.force fields.f_formal_part
  let f_subp_decl node =
    match (node :> GenericSubpDeclType.t) with
    | `GenericSubpDecl fields ->
        Lazy.force fields.f_subp_decl



end

module GenericPackageDecl = struct
  type t =
    [
      | `GenericPackageDecl of
          GenericPackageDeclType.fields
    ]

  type fields = GenericPackageDeclType.fields =
    
  {
         
    f_formal_part: GenericFormalPartType.t
    Lazy.t;
         
    f_package_decl: GenericPackageInternalType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_body_part
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.generic_package_decl_p_body_part
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (PackageBodyType.wrap (context node) (!@ result_ptr))


  let f_formal_part node =
    match (node :> GenericPackageDeclType.t) with
    | `GenericPackageDecl fields ->
        Lazy.force fields.f_formal_part
  let f_package_decl node =
    match (node :> GenericPackageDeclType.t) with
    | `GenericPackageDecl fields ->
        Lazy.force fields.f_package_decl



end

module GenericDecl = struct
  type t =
    [
      | GenericPackageDecl.t
      | GenericSubpDecl.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_formal_part node =
    match (node :> GenericDeclType.t) with
    | `GenericPackageDecl fields ->
        Lazy.force fields.f_formal_part
    | `GenericSubpDecl fields ->
        Lazy.force fields.f_formal_part



end

module ForLoopVarDecl = struct
  type t =
    [
      | `ForLoopVarDecl of
          ForLoopVarDeclType.fields
    ]

  type fields = ForLoopVarDeclType.fields =
    
  {
         
    f_id: DefiningNameType.t
    Lazy.t;
         
    f_id_type: SubtypeIndicationType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_id node =
    match (node :> ForLoopVarDeclType.t) with
    | `ForLoopVarDecl fields ->
        Lazy.force fields.f_id
  let f_id_type node =
    match (node :> ForLoopVarDeclType.t) with
    | `ForLoopVarDecl fields ->
        Lazy.force fields.f_id_type



end

module ExceptionHandler = struct
  type t =
    [
      | `ExceptionHandler of
          ExceptionHandlerType.fields
    ]

  type fields = ExceptionHandlerType.fields =
    
  {
         
    f_exception_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_handled_exceptions: AlternativesListType.t
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_exception_name node =
    match (node :> ExceptionHandlerType.t) with
    | `ExceptionHandler fields ->
        Lazy.force fields.f_exception_name
  let f_handled_exceptions node =
    match (node :> ExceptionHandlerType.t) with
    | `ExceptionHandler fields ->
        Lazy.force fields.f_handled_exceptions
  let f_stmts node =
    match (node :> ExceptionHandlerType.t) with
    | `ExceptionHandler fields ->
        Lazy.force fields.f_stmts



end

module ExceptionDecl = struct
  type t =
    [
      | `ExceptionDecl of
          ExceptionDeclType.fields
    ]

  type fields = ExceptionDeclType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ids node =
    match (node :> ExceptionDeclType.t) with
    | `ExceptionDecl fields ->
        Lazy.force fields.f_ids
  let f_renames node =
    match (node :> ExceptionDeclType.t) with
    | `ExceptionDecl fields ->
        Lazy.force fields.f_renames



end

module ErrorDecl = struct
  type t =
    [
      | `ErrorDecl of
          ErrorDeclType.fields
    ]

  type fields = ErrorDeclType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module EntryIndexSpec = struct
  type t =
    [
      | `EntryIndexSpec of
          EntryIndexSpecType.fields
    ]

  type fields = EntryIndexSpecType.fields =
    
  {
         
    f_id: DefiningNameType.t
    Lazy.t;
         
    f_subtype: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_id node =
    match (node :> EntryIndexSpecType.t) with
    | `EntryIndexSpec fields ->
        Lazy.force fields.f_id
  let f_subtype node =
    match (node :> EntryIndexSpecType.t) with
    | `EntryIndexSpec fields ->
        Lazy.force fields.f_subtype



end

module TaskBody = struct
  type t =
    [
      | `TaskBody of
          TaskBodyType.fields
    ]

  type fields = TaskBodyType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> TaskBodyType.t) with
    | `TaskBody fields ->
        Lazy.force fields.f_name
  let f_decls node =
    match (node :> TaskBodyType.t) with
    | `TaskBody fields ->
        Lazy.force fields.f_decls
  let f_stmts node =
    match (node :> TaskBodyType.t) with
    | `TaskBody fields ->
        Lazy.force fields.f_stmts
  let f_end_name node =
    match (node :> TaskBodyType.t) with
    | `TaskBody fields ->
        Lazy.force fields.f_end_name



end

module ProtectedBody = struct
  type t =
    [
      | `ProtectedBody of
          ProtectedBodyType.fields
    ]

  type fields = ProtectedBodyType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> ProtectedBodyType.t) with
    | `ProtectedBody fields ->
        Lazy.force fields.f_name
  let f_decls node =
    match (node :> ProtectedBodyType.t) with
    | `ProtectedBody fields ->
        Lazy.force fields.f_decls
  let f_end_name node =
    match (node :> ProtectedBodyType.t) with
    | `ProtectedBody fields ->
        Lazy.force fields.f_end_name



end

module PackageBody = struct
  type t =
    [
      | `PackageBody of
          PackageBodyType.fields
    ]

  type fields = PackageBodyType.fields =
    
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_package_name node =
    match (node :> PackageBodyType.t) with
    | `PackageBody fields ->
        Lazy.force fields.f_package_name
  let f_decls node =
    match (node :> PackageBodyType.t) with
    | `PackageBody fields ->
        Lazy.force fields.f_decls
  let f_stmts node =
    match (node :> PackageBodyType.t) with
    | `PackageBody fields ->
        Lazy.force fields.f_stmts
  let f_end_name node =
    match (node :> PackageBodyType.t) with
    | `PackageBody fields ->
        Lazy.force fields.f_end_name



end

module EntryBody = struct
  type t =
    [
      | `EntryBody of
          EntryBodyType.fields
    ]

  type fields = EntryBodyType.fields =
    
  {
         
    f_entry_name: DefiningNameType.t
    Lazy.t;
         
    f_index_spec: EntryIndexSpecType.t
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
         
    f_barrier: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_entry_name node =
    match (node :> EntryBodyType.t) with
    | `EntryBody fields ->
        Lazy.force fields.f_entry_name
  let f_index_spec node =
    match (node :> EntryBodyType.t) with
    | `EntryBody fields ->
        Lazy.force fields.f_index_spec
  let f_params node =
    match (node :> EntryBodyType.t) with
    | `EntryBody fields ->
        Lazy.force fields.f_params
  let f_barrier node =
    match (node :> EntryBodyType.t) with
    | `EntryBody fields ->
        Lazy.force fields.f_barrier
  let f_decls node =
    match (node :> EntryBodyType.t) with
    | `EntryBody fields ->
        Lazy.force fields.f_decls
  let f_stmts node =
    match (node :> EntryBodyType.t) with
    | `EntryBody fields ->
        Lazy.force fields.f_stmts
  let f_end_name node =
    match (node :> EntryBodyType.t) with
    | `EntryBody fields ->
        Lazy.force fields.f_end_name



end

module TaskBodyStub = struct
  type t =
    [
      | `TaskBodyStub of
          TaskBodyStubType.fields
    ]

  type fields = TaskBodyStubType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> TaskBodyStubType.t) with
    | `TaskBodyStub fields ->
        Lazy.force fields.f_name



end

module SubpBodyStub = struct
  type t =
    [
      | `SubpBodyStub of
          SubpBodyStubType.fields
    ]

  type fields = SubpBodyStubType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_overriding node =
    match (node :> SubpBodyStubType.t) with
    | `SubpBodyStub fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> SubpBodyStubType.t) with
    | `SubpBodyStub fields ->
        Lazy.force fields.f_subp_spec



end

module ProtectedBodyStub = struct
  type t =
    [
      | `ProtectedBodyStub of
          ProtectedBodyStubType.fields
    ]

  type fields = ProtectedBodyStubType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> ProtectedBodyStubType.t) with
    | `ProtectedBodyStub fields ->
        Lazy.force fields.f_name



end

module PackageBodyStub = struct
  type t =
    [
      | `PackageBodyStub of
          PackageBodyStubType.fields
    ]

  type fields = PackageBodyStubType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> PackageBodyStubType.t) with
    | `PackageBodyStub fields ->
        Lazy.force fields.f_name



end

module BodyStub = struct
  type t =
    [
      | PackageBodyStub.t
      | ProtectedBodyStub.t
      | SubpBodyStub.t
      | TaskBodyStub.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_syntactic_fully_qualified_name
    (node)
    =
      let result_ptr =
        allocate_n UnboundedTextTypeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.body_stub_p_syntactic_fully_qualified_name
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      UnboundedTextTypeArray.wrap (!@ result_ptr)





end

module SubpRenamingDecl = struct
  type t =
    [
      | `SubpRenamingDecl of
          SubpRenamingDeclType.fields
    ]

  type fields = SubpRenamingDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_overriding node =
    match (node :> SubpRenamingDeclType.t) with
    | `SubpRenamingDecl fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> SubpRenamingDeclType.t) with
    | `SubpRenamingDecl fields ->
        Lazy.force fields.f_subp_spec
  let f_renames node =
    match (node :> SubpRenamingDeclType.t) with
    | `SubpRenamingDecl fields ->
        Lazy.force fields.f_renames



end

module SubpBody = struct
  type t =
    [
      | `SubpBody of
          SubpBodyType.fields
    ]

  type fields = SubpBodyType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_overriding node =
    match (node :> SubpBodyType.t) with
    | `SubpBody fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> SubpBodyType.t) with
    | `SubpBody fields ->
        Lazy.force fields.f_subp_spec
  let f_decls node =
    match (node :> SubpBodyType.t) with
    | `SubpBody fields ->
        Lazy.force fields.f_decls
  let f_stmts node =
    match (node :> SubpBodyType.t) with
    | `SubpBody fields ->
        Lazy.force fields.f_stmts
  let f_end_name node =
    match (node :> SubpBodyType.t) with
    | `SubpBody fields ->
        Lazy.force fields.f_end_name



end

module NullSubpDecl = struct
  type t =
    [
      | `NullSubpDecl of
          NullSubpDeclType.fields
    ]

  type fields = NullSubpDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_overriding node =
    match (node :> NullSubpDeclType.t) with
    | `NullSubpDecl fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> NullSubpDeclType.t) with
    | `NullSubpDecl fields ->
        Lazy.force fields.f_subp_spec



end

module ExprFunction = struct
  type t =
    [
      | `ExprFunction of
          ExprFunctionType.fields
    ]

  type fields = ExprFunctionType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_overriding node =
    match (node :> ExprFunctionType.t) with
    | `ExprFunction fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> ExprFunctionType.t) with
    | `ExprFunction fields ->
        Lazy.force fields.f_subp_spec
  let f_expr node =
    match (node :> ExprFunctionType.t) with
    | `ExprFunction fields ->
        Lazy.force fields.f_expr



end

module BaseSubpBody = struct
  type t =
    [
      | ExprFunction.t
      | NullSubpDecl.t
      | SubpBody.t
      | SubpRenamingDecl.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_overriding node =
    match (node :> BaseSubpBodyType.t) with
    | `ExprFunction fields ->
        Lazy.force fields.f_overriding
    | `NullSubpDecl fields ->
        Lazy.force fields.f_overriding
    | `SubpBody fields ->
        Lazy.force fields.f_overriding
    | `SubpRenamingDecl fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> BaseSubpBodyType.t) with
    | `ExprFunction fields ->
        Lazy.force fields.f_subp_spec
    | `NullSubpDecl fields ->
        Lazy.force fields.f_subp_spec
    | `SubpBody fields ->
        Lazy.force fields.f_subp_spec
    | `SubpRenamingDecl fields ->
        Lazy.force fields.f_subp_spec



end

module BodyNode = struct
  type t =
    [
      | BaseSubpBody.t
      | BodyStub.t
      | EntryBody.t
      | PackageBody.t
      | ProtectedBody.t
      | TaskBody.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_previous_part
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.body_node_p_previous_part
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_decl_part
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.body_node_p_decl_part
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_subunit_root
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.body_node_p_subunit_root
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))





end

module GenericSubpInternal = struct
  type t =
    [
      | `GenericSubpInternal of
          GenericSubpInternalType.fields
    ]

  type fields = GenericSubpInternalType.fields =
    
  {
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_subp_spec node =
    match (node :> GenericSubpInternalType.t) with
    | `GenericSubpInternal fields ->
        Lazy.force fields.f_subp_spec



end

module EnumLiteralDecl = struct
  type t =
    [
      | `EnumLiteralDecl of
          EnumLiteralDeclType.fields
    ]

  type fields = EnumLiteralDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_enum_type
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.enum_literal_decl_p_enum_type
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (TypeDeclType.wrap (context node) (!@ result_ptr))


  let f_name node =
    match (node :> EnumLiteralDeclType.t) with
    | `EnumLiteralDecl fields ->
        Lazy.force fields.f_name



end

module EntryDecl = struct
  type t =
    [
      | `EntryDecl of
          EntryDeclType.fields
    ]

  type fields = EntryDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_spec: EntrySpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_body_part
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.entry_decl_p_body_part
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (EntryBodyType.wrap (context node) (!@ result_ptr))


  let f_overriding node =
    match (node :> EntryDeclType.t) with
    | `EntryDecl fields ->
        Lazy.force fields.f_overriding
  let f_spec node =
    match (node :> EntryDeclType.t) with
    | `EntryDecl fields ->
        Lazy.force fields.f_spec



end

module SubpDecl = struct
  type t =
    [
      | `SubpDecl of
          SubpDeclType.fields
    ]

  type fields = SubpDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_overriding node =
    match (node :> SubpDeclType.t) with
    | `SubpDecl fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> SubpDeclType.t) with
    | `SubpDecl fields ->
        Lazy.force fields.f_subp_spec



end

module ConcreteFormalSubpDecl = struct
  type t =
    [
      | `ConcreteFormalSubpDecl of
          ConcreteFormalSubpDeclType.fields
    ]

  type fields = ConcreteFormalSubpDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_default_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_overriding node =
    match (node :> ConcreteFormalSubpDeclType.t) with
    | `ConcreteFormalSubpDecl fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> ConcreteFormalSubpDeclType.t) with
    | `ConcreteFormalSubpDecl fields ->
        Lazy.force fields.f_subp_spec
  let f_default_expr node =
    match (node :> ConcreteFormalSubpDeclType.t) with
    | `ConcreteFormalSubpDecl fields ->
        Lazy.force fields.f_default_expr



end

module AbstractFormalSubpDecl = struct
  type t =
    [
      | `AbstractFormalSubpDecl of
          AbstractFormalSubpDeclType.fields
    ]

  type fields = AbstractFormalSubpDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_default_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_overriding node =
    match (node :> AbstractFormalSubpDeclType.t) with
    | `AbstractFormalSubpDecl fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> AbstractFormalSubpDeclType.t) with
    | `AbstractFormalSubpDecl fields ->
        Lazy.force fields.f_subp_spec
  let f_default_expr node =
    match (node :> AbstractFormalSubpDeclType.t) with
    | `AbstractFormalSubpDecl fields ->
        Lazy.force fields.f_default_expr



end

module FormalSubpDecl = struct
  type t =
    [
      | AbstractFormalSubpDecl.t
      | ConcreteFormalSubpDecl.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_overriding node =
    match (node :> FormalSubpDeclType.t) with
    | `AbstractFormalSubpDecl fields ->
        Lazy.force fields.f_overriding
    | `ConcreteFormalSubpDecl fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> FormalSubpDeclType.t) with
    | `AbstractFormalSubpDecl fields ->
        Lazy.force fields.f_subp_spec
    | `ConcreteFormalSubpDecl fields ->
        Lazy.force fields.f_subp_spec
  let f_default_expr node =
    match (node :> FormalSubpDeclType.t) with
    | `AbstractFormalSubpDecl fields ->
        Lazy.force fields.f_default_expr
    | `ConcreteFormalSubpDecl fields ->
        Lazy.force fields.f_default_expr



end

module AbstractSubpDecl = struct
  type t =
    [
      | `AbstractSubpDecl of
          AbstractSubpDeclType.fields
    ]

  type fields = AbstractSubpDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_overriding node =
    match (node :> AbstractSubpDeclType.t) with
    | `AbstractSubpDecl fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> AbstractSubpDeclType.t) with
    | `AbstractSubpDecl fields ->
        Lazy.force fields.f_subp_spec



end

module ClassicSubpDecl = struct
  type t =
    [
      | AbstractSubpDecl.t
      | FormalSubpDecl.t
      | SubpDecl.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_body_part
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.classic_subp_decl_p_body_part
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseSubpBodyType.wrap (context node) (!@ result_ptr))


  let f_overriding node =
    match (node :> ClassicSubpDeclType.t) with
    | `AbstractSubpDecl fields ->
        Lazy.force fields.f_overriding
    | `AbstractFormalSubpDecl fields ->
        Lazy.force fields.f_overriding
    | `ConcreteFormalSubpDecl fields ->
        Lazy.force fields.f_overriding
    | `SubpDecl fields ->
        Lazy.force fields.f_overriding
  let f_subp_spec node =
    match (node :> ClassicSubpDeclType.t) with
    | `AbstractSubpDecl fields ->
        Lazy.force fields.f_subp_spec
    | `AbstractFormalSubpDecl fields ->
        Lazy.force fields.f_subp_spec
    | `ConcreteFormalSubpDecl fields ->
        Lazy.force fields.f_subp_spec
    | `SubpDecl fields ->
        Lazy.force fields.f_subp_spec



end

module BasicSubpDecl = struct
  type t =
    [
      | ClassicSubpDecl.t
      | EntryDecl.t
      | EnumLiteralDecl.t
      | GenericSubpInternal.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_subp_decl_spec
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_subp_decl_p_subp_decl_spec
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseSubpSpecType.wrap (context node) (!@ result_ptr))





end

module SynthAnonymousTypeDecl = struct
  type t =
    [
      | `SynthAnonymousTypeDecl of
          SynthAnonymousTypeDeclType.fields
    ]

  type fields = SynthAnonymousTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> SynthAnonymousTypeDeclType.t) with
    | `SynthAnonymousTypeDecl fields ->
        Lazy.force fields.f_name
  let f_discriminants node =
    match (node :> SynthAnonymousTypeDeclType.t) with
    | `SynthAnonymousTypeDecl fields ->
        Lazy.force fields.f_discriminants
  let f_type_def node =
    match (node :> SynthAnonymousTypeDeclType.t) with
    | `SynthAnonymousTypeDecl fields ->
        Lazy.force fields.f_type_def



end

module AnonymousTypeDecl = struct
  type t =
    [
      | `AnonymousTypeDecl of
          AnonymousTypeDeclType.fields
      | `SynthAnonymousTypeDecl of
          SynthAnonymousTypeDeclType.fields
    ]

  type fields = AnonymousTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> AnonymousTypeDeclType.t) with
    | `AnonymousTypeDecl fields ->
        Lazy.force fields.f_name
    | `SynthAnonymousTypeDecl fields ->
        Lazy.force fields.f_name
  let f_discriminants node =
    match (node :> AnonymousTypeDeclType.t) with
    | `AnonymousTypeDecl fields ->
        Lazy.force fields.f_discriminants
    | `SynthAnonymousTypeDecl fields ->
        Lazy.force fields.f_discriminants
  let f_type_def node =
    match (node :> AnonymousTypeDeclType.t) with
    | `AnonymousTypeDecl fields ->
        Lazy.force fields.f_type_def
    | `SynthAnonymousTypeDecl fields ->
        Lazy.force fields.f_type_def



end

module TypeDecl = struct
  type t =
    [
      | `TypeDecl of
          TypeDeclType.fields
      | `AnonymousTypeDecl of
          AnonymousTypeDeclType.fields
      | `SynthAnonymousTypeDecl of
          SynthAnonymousTypeDeclType.fields
    ]

  type fields = TypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_get_primitives
    ?(only_inherited=false)
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_only_inherited =
            
        only_inherited
      in
      let _ : int =
        CFunctions.type_decl_p_get_primitives
          (addr (AdaNodeType.unwrap (node)))
          c_only_inherited
          (result_ptr)
      in
         
      BasicDeclArray.wrap (context node) (!@ result_ptr)


  let f_name node =
    match (node :> TypeDeclType.t) with
    | `TypeDecl fields ->
        Lazy.force fields.f_name
    | `AnonymousTypeDecl fields ->
        Lazy.force fields.f_name
    | `SynthAnonymousTypeDecl fields ->
        Lazy.force fields.f_name
  let f_discriminants node =
    match (node :> TypeDeclType.t) with
    | `TypeDecl fields ->
        Lazy.force fields.f_discriminants
    | `AnonymousTypeDecl fields ->
        Lazy.force fields.f_discriminants
    | `SynthAnonymousTypeDecl fields ->
        Lazy.force fields.f_discriminants
  let f_type_def node =
    match (node :> TypeDeclType.t) with
    | `TypeDecl fields ->
        Lazy.force fields.f_type_def
    | `AnonymousTypeDecl fields ->
        Lazy.force fields.f_type_def
    | `SynthAnonymousTypeDecl fields ->
        Lazy.force fields.f_type_def



end

module SingleTaskTypeDecl = struct
  type t =
    [
      | `SingleTaskTypeDecl of
          SingleTaskTypeDeclType.fields
    ]

  type fields = SingleTaskTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_definition: TaskDefType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> SingleTaskTypeDeclType.t) with
    | `SingleTaskTypeDecl fields ->
        Lazy.force fields.f_name
  let f_discriminants node =
    match (node :> SingleTaskTypeDeclType.t) with
    | `SingleTaskTypeDecl fields ->
        Lazy.force fields.f_discriminants
  let f_definition node =
    match (node :> SingleTaskTypeDeclType.t) with
    | `SingleTaskTypeDecl fields ->
        Lazy.force fields.f_definition



end

module TaskTypeDecl = struct
  type t =
    [
      | `TaskTypeDecl of
          TaskTypeDeclType.fields
      | `SingleTaskTypeDecl of
          SingleTaskTypeDeclType.fields
    ]

  type fields = TaskTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_definition: TaskDefType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> TaskTypeDeclType.t) with
    | `TaskTypeDecl fields ->
        Lazy.force fields.f_name
    | `SingleTaskTypeDecl fields ->
        Lazy.force fields.f_name
  let f_discriminants node =
    match (node :> TaskTypeDeclType.t) with
    | `TaskTypeDecl fields ->
        Lazy.force fields.f_discriminants
    | `SingleTaskTypeDecl fields ->
        Lazy.force fields.f_discriminants
  let f_definition node =
    match (node :> TaskTypeDeclType.t) with
    | `TaskTypeDecl fields ->
        Lazy.force fields.f_definition
    | `SingleTaskTypeDecl fields ->
        Lazy.force fields.f_definition



end

module ProtectedTypeDecl = struct
  type t =
    [
      | `ProtectedTypeDecl of
          ProtectedTypeDeclType.fields
    ]

  type fields = ProtectedTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_definition: ProtectedDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> ProtectedTypeDeclType.t) with
    | `ProtectedTypeDecl fields ->
        Lazy.force fields.f_name
  let f_discriminants node =
    match (node :> ProtectedTypeDeclType.t) with
    | `ProtectedTypeDecl fields ->
        Lazy.force fields.f_discriminants
  let f_interfaces node =
    match (node :> ProtectedTypeDeclType.t) with
    | `ProtectedTypeDecl fields ->
        Lazy.force fields.f_interfaces
  let f_definition node =
    match (node :> ProtectedTypeDeclType.t) with
    | `ProtectedTypeDecl fields ->
        Lazy.force fields.f_definition



end

module IncompleteTaggedTypeDecl = struct
  type t =
    [
      | `IncompleteTaggedTypeDecl of
          IncompleteTaggedTypeDeclType.fields
    ]

  type fields = IncompleteTaggedTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> IncompleteTaggedTypeDeclType.t) with
    | `IncompleteTaggedTypeDecl fields ->
        Lazy.force fields.f_name
  let f_discriminants node =
    match (node :> IncompleteTaggedTypeDeclType.t) with
    | `IncompleteTaggedTypeDecl fields ->
        Lazy.force fields.f_discriminants
  let f_has_abstract node =
    match (node :> IncompleteTaggedTypeDeclType.t) with
    | `IncompleteTaggedTypeDecl fields ->
        Lazy.force fields.f_has_abstract



end

module IncompleteTypeDecl = struct
  type t =
    [
      | `IncompleteTypeDecl of
          IncompleteTypeDeclType.fields
      | `IncompleteTaggedTypeDecl of
          IncompleteTaggedTypeDeclType.fields
    ]

  type fields = IncompleteTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> IncompleteTypeDeclType.t) with
    | `IncompleteTypeDecl fields ->
        Lazy.force fields.f_name
    | `IncompleteTaggedTypeDecl fields ->
        Lazy.force fields.f_name
  let f_discriminants node =
    match (node :> IncompleteTypeDeclType.t) with
    | `IncompleteTypeDecl fields ->
        Lazy.force fields.f_discriminants
    | `IncompleteTaggedTypeDecl fields ->
        Lazy.force fields.f_discriminants



end

module ClasswideTypeDecl = struct
  type t =
    [
      | `ClasswideTypeDecl of
          ClasswideTypeDeclType.fields
    ]

  type fields = ClasswideTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> ClasswideTypeDeclType.t) with
    | `ClasswideTypeDecl fields ->
        Lazy.force fields.f_name



end

module SubtypeDecl = struct
  type t =
    [
      | `SubtypeDecl of
          SubtypeDeclType.fields
    ]

  type fields = SubtypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_subtype: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> SubtypeDeclType.t) with
    | `SubtypeDecl fields ->
        Lazy.force fields.f_name
  let f_subtype node =
    match (node :> SubtypeDeclType.t) with
    | `SubtypeDecl fields ->
        Lazy.force fields.f_subtype



end

module DiscreteBaseSubtypeDecl = struct
  type t =
    [
      | `DiscreteBaseSubtypeDecl of
          DiscreteBaseSubtypeDeclType.fields
    ]

  type fields = DiscreteBaseSubtypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> DiscreteBaseSubtypeDeclType.t) with
    | `DiscreteBaseSubtypeDecl fields ->
        Lazy.force fields.f_name



end

module BaseSubtypeDecl = struct
  type t =
    [
      | DiscreteBaseSubtypeDecl.t
      | SubtypeDecl.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_get_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_subtype_decl_p_get_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))


  let f_name node =
    match (node :> BaseSubtypeDeclType.t) with
    | `DiscreteBaseSubtypeDecl fields ->
        Lazy.force fields.f_name
    | `SubtypeDecl fields ->
        Lazy.force fields.f_name



end

module BaseTypeDecl = struct
  type t =
    [
      | BaseSubtypeDecl.t
      | ClasswideTypeDecl.t
      | IncompleteTypeDecl.t
      | ProtectedTypeDecl.t
      | TaskTypeDecl.t
      | TypeDecl.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_base_subtype
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_base_subtype
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_private_completion
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_type_decl_p_private_completion
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_is_inherited_primitive
    (node)
    (p)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_p =
            
        addr (AdaNodeType.unwrap (p))
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_inherited_primitive
          (addr (AdaNodeType.unwrap (node)))
          c_p
          (result_ptr)
      in
         
      !@ result_ptr

let p_get_record_representation_clause
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.base_type_decl_p_get_record_representation_clause
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (RecordRepClauseType.wrap (context node) (!@ result_ptr))

let p_get_enum_representation_clause
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.base_type_decl_p_get_enum_representation_clause
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (EnumRepClauseType.wrap (context node) (!@ result_ptr))

let p_is_record_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_record_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_array_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_array_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_find_derived_types
    ?(imprecise_fallback=false)
    (node)
    (root)
    (origin)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_root =
            
        addr (AdaNodeType.unwrap (root))
      in
      let c_origin =
            
        addr (AdaNodeType.unwrap (origin))
      in
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.base_type_decl_p_find_derived_types
          (addr (AdaNodeType.unwrap (node)))
          c_root
          c_origin
          c_imprecise_fallback
          (result_ptr)
      in
         
         
         
      TypeDeclArray.wrap (context node) (!@ result_ptr)

let p_is_real_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_real_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_float_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_float_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_fixed_point
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_fixed_point
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_enum_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_enum_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_access_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_access_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_char_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_char_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_discrete_range
    (node)
    =
      let result_ptr =
        allocate_n DiscreteRangeStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_type_decl_p_discrete_range
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      DiscreteRange.wrap (context node) (!@ result_ptr)

let p_is_discrete_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_discrete_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_int_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_int_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_accessed_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_accessed_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_is_tagged_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_tagged_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_base_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_base_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_base_types
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_base_types
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      BaseTypeDeclArray.wrap (context node) (!@ result_ptr)

let p_find_all_derived_types
    ?(imprecise_fallback=false)
    (node)
    (units)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_units =
            
        AnalysisUnitArray.unwrap (units)
      in
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.base_type_decl_p_find_all_derived_types
          (addr (AdaNodeType.unwrap (node)))
          c_units
          c_imprecise_fallback
          (result_ptr)
      in
         
               AnalysisUnitArrayStruct.dec_ref c_units ;
         
      TypeDeclArray.wrap (context node) (!@ result_ptr)

let p_comp_type
    ?(is_subscript=false)
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_is_subscript =
            
        is_subscript
      in
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_comp_type
          (addr (AdaNodeType.unwrap (node)))
          c_is_subscript
          c_origin
          (result_ptr)
      in
         
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_index_type
    ?origin
    (node)
    (dim)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_dim =
            
        dim
      in
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_index_type
          (addr (AdaNodeType.unwrap (node)))
          c_dim
          c_origin
          (result_ptr)
      in
         
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_is_derived_type
    ?origin
    (node)
    (other_type)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_other_type =
            
        addr (AdaNodeType.unwrap (other_type))
      in
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_derived_type
          (addr (AdaNodeType.unwrap (node)))
          c_other_type
          c_origin
          (result_ptr)
      in
         
         
      !@ result_ptr

let p_is_interface_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_interface_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_matching_type
    ?origin
    (node)
    (expected_type)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_expected_type =
            
        addr (AdaNodeType.unwrap (expected_type))
      in
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_matching_type
          (addr (AdaNodeType.unwrap (node)))
          c_expected_type
          c_origin
          (result_ptr)
      in
         
         
      !@ result_ptr

let p_canonical_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_canonical_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_previous_part
    ?(go_to_incomplete=true)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_go_to_incomplete =
            
        go_to_incomplete
      in
      let _ : int =
        CFunctions.base_type_decl_p_previous_part
          (addr (AdaNodeType.unwrap (node)))
          c_go_to_incomplete
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_next_part
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_type_decl_p_next_part
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_full_view
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_type_decl_p_full_view
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_is_definite_subtype
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_is_definite_subtype
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_private
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_type_decl_p_is_private
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_discriminants_list
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_type_decl_p_discriminants_list
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      BaseFormalParamDeclArray.wrap (context node) (!@ result_ptr)

let p_root_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_root_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_shapes
    ?(include_discriminants=true)
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n ShapeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_include_discriminants =
            
        include_discriminants
      in
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_type_decl_p_shapes
          (addr (AdaNodeType.unwrap (node)))
          c_include_discriminants
          c_origin
          (result_ptr)
      in
         
         
      ShapeArray.wrap (context node) (!@ result_ptr)


  let f_name node =
    match (node :> BaseTypeDeclType.t) with
    | `DiscreteBaseSubtypeDecl fields ->
        Lazy.force fields.f_name
    | `SubtypeDecl fields ->
        Lazy.force fields.f_name
    | `ClasswideTypeDecl fields ->
        Lazy.force fields.f_name
    | `IncompleteTypeDecl fields ->
        Lazy.force fields.f_name
    | `IncompleteTaggedTypeDecl fields ->
        Lazy.force fields.f_name
    | `ProtectedTypeDecl fields ->
        Lazy.force fields.f_name
    | `TaskTypeDecl fields ->
        Lazy.force fields.f_name
    | `SingleTaskTypeDecl fields ->
        Lazy.force fields.f_name
    | `TypeDecl fields ->
        Lazy.force fields.f_name
    | `AnonymousTypeDecl fields ->
        Lazy.force fields.f_name
    | `SynthAnonymousTypeDecl fields ->
        Lazy.force fields.f_name



end

module PackageDecl = struct
  type t =
    [
      | `PackageDecl of
          PackageDeclType.fields
    ]

  type fields = PackageDeclType.fields =
    
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_package_name node =
    match (node :> PackageDeclType.t) with
    | `PackageDecl fields ->
        Lazy.force fields.f_package_name
  let f_public_part node =
    match (node :> PackageDeclType.t) with
    | `PackageDecl fields ->
        Lazy.force fields.f_public_part
  let f_private_part node =
    match (node :> PackageDeclType.t) with
    | `PackageDecl fields ->
        Lazy.force fields.f_private_part
  let f_end_name node =
    match (node :> PackageDeclType.t) with
    | `PackageDecl fields ->
        Lazy.force fields.f_end_name



end

module GenericPackageInternal = struct
  type t =
    [
      | `GenericPackageInternal of
          GenericPackageInternalType.fields
    ]

  type fields = GenericPackageInternalType.fields =
    
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_package_name node =
    match (node :> GenericPackageInternalType.t) with
    | `GenericPackageInternal fields ->
        Lazy.force fields.f_package_name
  let f_public_part node =
    match (node :> GenericPackageInternalType.t) with
    | `GenericPackageInternal fields ->
        Lazy.force fields.f_public_part
  let f_private_part node =
    match (node :> GenericPackageInternalType.t) with
    | `GenericPackageInternal fields ->
        Lazy.force fields.f_private_part
  let f_end_name node =
    match (node :> GenericPackageInternalType.t) with
    | `GenericPackageInternal fields ->
        Lazy.force fields.f_end_name



end

module BasePackageDecl = struct
  type t =
    [
      | GenericPackageInternal.t
      | PackageDecl.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_body_part
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_package_decl_p_body_part
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (PackageBodyType.wrap (context node) (!@ result_ptr))


  let f_package_name node =
    match (node :> BasePackageDeclType.t) with
    | `GenericPackageInternal fields ->
        Lazy.force fields.f_package_name
    | `PackageDecl fields ->
        Lazy.force fields.f_package_name
  let f_public_part node =
    match (node :> BasePackageDeclType.t) with
    | `GenericPackageInternal fields ->
        Lazy.force fields.f_public_part
    | `PackageDecl fields ->
        Lazy.force fields.f_public_part
  let f_private_part node =
    match (node :> BasePackageDeclType.t) with
    | `GenericPackageInternal fields ->
        Lazy.force fields.f_private_part
    | `PackageDecl fields ->
        Lazy.force fields.f_private_part
  let f_end_name node =
    match (node :> BasePackageDeclType.t) with
    | `GenericPackageInternal fields ->
        Lazy.force fields.f_end_name
    | `PackageDecl fields ->
        Lazy.force fields.f_end_name



end

module ParamSpec = struct
  type t =
    [
      | `ParamSpec of
          ParamSpecType.fields
    ]

  type fields = ParamSpecType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ids node =
    match (node :> ParamSpecType.t) with
    | `ParamSpec fields ->
        Lazy.force fields.f_ids
  let f_has_aliased node =
    match (node :> ParamSpecType.t) with
    | `ParamSpec fields ->
        Lazy.force fields.f_has_aliased
  let f_mode node =
    match (node :> ParamSpecType.t) with
    | `ParamSpec fields ->
        Lazy.force fields.f_mode
  let f_type_expr node =
    match (node :> ParamSpecType.t) with
    | `ParamSpec fields ->
        Lazy.force fields.f_type_expr
  let f_default_expr node =
    match (node :> ParamSpecType.t) with
    | `ParamSpec fields ->
        Lazy.force fields.f_default_expr



end

module GenericFormalTypeDecl = struct
  type t =
    [
      | `GenericFormalTypeDecl of
          GenericFormalTypeDeclType.fields
    ]

  type fields = GenericFormalTypeDeclType.fields =
    
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decl node =
    match (node :> GenericFormalTypeDeclType.t) with
    | `GenericFormalTypeDecl fields ->
        Lazy.force fields.f_decl



end

module GenericFormalSubpDecl = struct
  type t =
    [
      | `GenericFormalSubpDecl of
          GenericFormalSubpDeclType.fields
    ]

  type fields = GenericFormalSubpDeclType.fields =
    
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decl node =
    match (node :> GenericFormalSubpDeclType.t) with
    | `GenericFormalSubpDecl fields ->
        Lazy.force fields.f_decl



end

module GenericFormalPackage = struct
  type t =
    [
      | `GenericFormalPackage of
          GenericFormalPackageType.fields
    ]

  type fields = GenericFormalPackageType.fields =
    
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decl node =
    match (node :> GenericFormalPackageType.t) with
    | `GenericFormalPackage fields ->
        Lazy.force fields.f_decl



end

module GenericFormalObjDecl = struct
  type t =
    [
      | `GenericFormalObjDecl of
          GenericFormalObjDeclType.fields
    ]

  type fields = GenericFormalObjDeclType.fields =
    
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decl node =
    match (node :> GenericFormalObjDeclType.t) with
    | `GenericFormalObjDecl fields ->
        Lazy.force fields.f_decl



end

module GenericFormal = struct
  type t =
    [
      | GenericFormalObjDecl.t
      | GenericFormalPackage.t
      | GenericFormalSubpDecl.t
      | GenericFormalTypeDecl.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decl node =
    match (node :> GenericFormalType.t) with
    | `GenericFormalObjDecl fields ->
        Lazy.force fields.f_decl
    | `GenericFormalPackage fields ->
        Lazy.force fields.f_decl
    | `GenericFormalSubpDecl fields ->
        Lazy.force fields.f_decl
    | `GenericFormalTypeDecl fields ->
        Lazy.force fields.f_decl



end

module DiscriminantSpec = struct
  type t =
    [
      | `DiscriminantSpec of
          DiscriminantSpecType.fields
    ]

  type fields = DiscriminantSpecType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ids node =
    match (node :> DiscriminantSpecType.t) with
    | `DiscriminantSpec fields ->
        Lazy.force fields.f_ids
  let f_type_expr node =
    match (node :> DiscriminantSpecType.t) with
    | `DiscriminantSpec fields ->
        Lazy.force fields.f_type_expr
  let f_default_expr node =
    match (node :> DiscriminantSpecType.t) with
    | `DiscriminantSpec fields ->
        Lazy.force fields.f_default_expr



end

module ComponentDecl = struct
  type t =
    [
      | `ComponentDecl of
          ComponentDeclType.fields
    ]

  type fields = ComponentDeclType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_component_def: ComponentDefType.t
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ids node =
    match (node :> ComponentDeclType.t) with
    | `ComponentDecl fields ->
        Lazy.force fields.f_ids
  let f_component_def node =
    match (node :> ComponentDeclType.t) with
    | `ComponentDecl fields ->
        Lazy.force fields.f_component_def
  let f_default_expr node =
    match (node :> ComponentDeclType.t) with
    | `ComponentDecl fields ->
        Lazy.force fields.f_default_expr



end

module BaseFormalParamDecl = struct
  type t =
    [
      | ComponentDecl.t
      | DiscriminantSpec.t
      | GenericFormal.t
      | ParamSpec.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_formal_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_formal_param_decl_p_formal_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))





end

module AnonymousExprDecl = struct
  type t =
    [
      | `AnonymousExprDecl of
          AnonymousExprDeclType.fields
    ]

  type fields = AnonymousExprDeclType.fields =
    
  {
         
    f_expr: ExprType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_expr node =
    match (node :> AnonymousExprDeclType.t) with
    | `AnonymousExprDecl fields ->
        Lazy.force fields.f_expr



end

module AbstractStateDecl = struct
  type t =
    [
      | `AbstractStateDecl of
          AbstractStateDeclType.fields
    ]

  type fields = AbstractStateDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> AbstractStateDeclType.t) with
    | `AbstractStateDecl fields ->
        Lazy.force fields.f_name



end

module BasicDecl = struct
  type t =
    [
      | AbstractStateDecl.t
      | AnonymousExprDecl.t
      | BaseFormalParamDecl.t
      | BasePackageDecl.t
      | BaseTypeDecl.t
      | BasicSubpDecl.t
      | BodyNode.t
      | EntryIndexSpec.t
      | ErrorDecl.t
      | ExceptionDecl.t
      | ExceptionHandler.t
      | ForLoopVarDecl.t
      | GenericDecl.t
      | GenericInstantiation.t
      | GenericRenamingDecl.t
      | LabelDecl.t
      | NamedStmtDecl.t
      | NumberDecl.t
      | ObjectDecl.t
      | PackageRenamingDecl.t
      | SingleProtectedDecl.t
      | SingleTaskDecl.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_is_formal
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_is_formal
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_doc_annotations
    (node)
    =
      let result_ptr =
        allocate_n DocAnnotationArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_doc_annotations
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      DocAnnotationArray.wrap (!@ result_ptr)

let p_doc
    (node)
    =
      let result_ptr =
        allocate_n TextTypeStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_doc
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      TextType.wrap (!@ result_ptr)

let p_previous_part_for_decl
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_previous_part_for_decl
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_canonical_part
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_canonical_part
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_all_parts
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_all_parts
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      BasicDeclArray.wrap (context node) (!@ result_ptr)

let p_is_static_decl
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_is_static_decl
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      !@ result_ptr

let p_is_imported
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_is_imported
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_get_aspect_assoc
    (node)
    (name)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_name =
            
        allocate Symbol.c_type (Symbol.unwrap (context node).c_value (name))
      in
      let _ : int =
        CFunctions.basic_decl_p_get_aspect_assoc
          (addr (AdaNodeType.unwrap (node)))
          c_name
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AspectAssocType.wrap (context node) (!@ result_ptr))

let p_get_aspect_spec_expr
    (node)
    (name)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_name =
            
        allocate Symbol.c_type (Symbol.unwrap (context node).c_value (name))
      in
      let _ : int =
        CFunctions.basic_decl_p_get_aspect_spec_expr
          (addr (AdaNodeType.unwrap (node)))
          c_name
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (ExprType.wrap (context node) (!@ result_ptr))

let p_get_aspect
    ?(imprecise_fallback=false)
    (node)
    (name)
    =
      let result_ptr =
        allocate_n AspectStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_name =
            
        allocate Symbol.c_type (Symbol.unwrap (context node).c_value (name))
      in
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_get_aspect
          (addr (AdaNodeType.unwrap (node)))
          c_name
          c_imprecise_fallback
          (result_ptr)
      in
         
         
      Aspect.wrap (context node) (!@ result_ptr)

let p_has_aspect
    ?(imprecise_fallback=false)
    (node)
    (name)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_name =
            
        allocate Symbol.c_type (Symbol.unwrap (context node).c_value (name))
      in
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_has_aspect
          (addr (AdaNodeType.unwrap (node)))
          c_name
          c_imprecise_fallback
          (result_ptr)
      in
         
         
      !@ result_ptr

let p_get_pragma
    (node)
    (name)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_name =
            
        allocate Symbol.c_type (Symbol.unwrap (context node).c_value (name))
      in
      let _ : int =
        CFunctions.basic_decl_p_get_pragma
          (addr (AdaNodeType.unwrap (node)))
          c_name
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (PragmaNodeType.wrap (context node) (!@ result_ptr))

let p_get_representation_clause
    ?(imprecise_fallback=false)
    (node)
    (name)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_name =
            
        allocate Symbol.c_type (Symbol.unwrap (context node).c_value (name))
      in
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_get_representation_clause
          (addr (AdaNodeType.unwrap (node)))
          c_name
          c_imprecise_fallback
          (result_ptr)
      in
         
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AttributeDefClauseType.wrap (context node) (!@ result_ptr))

let p_is_compilation_unit_root
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_is_compilation_unit_root
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_is_visible
    (node)
    (from_node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_from_node =
            
        addr (AdaNodeType.unwrap (from_node))
      in
      let _ : int =
        CFunctions.basic_decl_p_is_visible
          (addr (AdaNodeType.unwrap (node)))
          c_from_node
          (result_ptr)
      in
         
      !@ result_ptr

let p_base_subp_declarations
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_base_subp_declarations
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      BasicDeclArray.wrap (context node) (!@ result_ptr)

let p_root_subp_declarations
    ?origin
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_root_subp_declarations
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          c_imprecise_fallback
          (result_ptr)
      in
         
         
      BasicDeclArray.wrap (context node) (!@ result_ptr)

let p_find_all_overrides
    ?(imprecise_fallback=false)
    (node)
    (units)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_units =
            
        AnalysisUnitArray.unwrap (units)
      in
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_find_all_overrides
          (addr (AdaNodeType.unwrap (node)))
          c_units
          c_imprecise_fallback
          (result_ptr)
      in
         
               AnalysisUnitArrayStruct.dec_ref c_units ;
         
      BasicDeclArray.wrap (context node) (!@ result_ptr)

let p_defining_names
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_defining_names
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      DefiningNameArray.wrap (context node) (!@ result_ptr)

let p_defining_name
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_defining_name
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (DefiningNameType.wrap (context node) (!@ result_ptr))

let p_type_expression
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_type_expression
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (TypeExprType.wrap (context node) (!@ result_ptr))

let p_subp_spec_or_null
    ?(follow_generic=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_follow_generic =
            
        follow_generic
      in
      let _ : int =
        CFunctions.basic_decl_p_subp_spec_or_null
          (addr (AdaNodeType.unwrap (node)))
          c_follow_generic
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseSubpSpecType.wrap (context node) (!@ result_ptr))

let p_is_subprogram
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_is_subprogram
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_relative_name
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_relative_name
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (SingleTokNodeType.wrap (context node) (!@ result_ptr))

let p_relative_name_text
    (node)
    =
      let result_ptr =
        allocate_n Symbol.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_relative_name_text
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      Symbol.wrap (!@ result_ptr)

let p_next_part_for_decl
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_next_part_for_decl
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_body_part_for_decl
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_decl_p_body_part_for_decl
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BodyNodeType.wrap (context node) (!@ result_ptr))

let p_fully_qualified_name_array
    (node)
    =
      let result_ptr =
        allocate_n UnboundedTextTypeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_fully_qualified_name_array
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      UnboundedTextTypeArray.wrap (!@ result_ptr)

let p_fully_qualified_name
    (node)
    =
      let result_ptr =
        allocate_n TextTypeStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_fully_qualified_name
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      TextType.wrap (!@ result_ptr)

let p_canonical_fully_qualified_name
    (node)
    =
      let result_ptr =
        allocate_n TextTypeStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_canonical_fully_qualified_name
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      TextType.wrap (!@ result_ptr)

let p_unique_identifying_name
    (node)
    =
      let result_ptr =
        allocate_n TextTypeStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.basic_decl_p_unique_identifying_name
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      TextType.wrap (!@ result_ptr)





end

module ParamAssoc = struct
  type t =
    [
      | `ParamAssoc of
          ParamAssocType.fields
    ]

  type fields = ParamAssocType.fields =
    
  {
         
    f_designator: [
      | `Identifier
          of IdentifierType.fields
      | `OthersDesignator
          of OthersDesignatorType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_designator node =
    match (node :> ParamAssocType.t) with
    | `ParamAssoc fields ->
        Lazy.force fields.f_designator
  let f_r_expr node =
    match (node :> ParamAssocType.t) with
    | `ParamAssoc fields ->
        Lazy.force fields.f_r_expr



end

module IteratedAssoc = struct
  type t =
    [
      | `IteratedAssoc of
          IteratedAssocType.fields
    ]

  type fields = IteratedAssocType.fields =
    
  {
         
    f_spec: ForLoopSpecType.t
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_spec node =
    match (node :> IteratedAssocType.t) with
    | `IteratedAssoc fields ->
        Lazy.force fields.f_spec
  let f_r_expr node =
    match (node :> IteratedAssocType.t) with
    | `IteratedAssoc fields ->
        Lazy.force fields.f_r_expr



end

module DiscriminantAssoc = struct
  type t =
    [
      | `DiscriminantAssoc of
          DiscriminantAssocType.fields
    ]

  type fields = DiscriminantAssocType.fields =
    
  {
         
    f_ids: IdentifierListType.t
    option
    Lazy.t;
         
    f_discr_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_ids node =
    match (node :> DiscriminantAssocType.t) with
    | `DiscriminantAssoc fields ->
        Lazy.force fields.f_ids
  let f_discr_expr node =
    match (node :> DiscriminantAssocType.t) with
    | `DiscriminantAssoc fields ->
        Lazy.force fields.f_discr_expr



end

module MultiDimArrayAssoc = struct
  type t =
    [
      | `MultiDimArrayAssoc of
          MultiDimArrayAssocType.fields
    ]

  type fields = MultiDimArrayAssocType.fields =
    
  {
         
    f_designators: AlternativesListType.t
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_designators node =
    match (node :> MultiDimArrayAssocType.t) with
    | `MultiDimArrayAssoc fields ->
        Lazy.force fields.f_designators
  let f_r_expr node =
    match (node :> MultiDimArrayAssocType.t) with
    | `MultiDimArrayAssoc fields ->
        Lazy.force fields.f_r_expr



end

module AggregateAssoc = struct
  type t =
    [
      | `AggregateAssoc of
          AggregateAssocType.fields
      | `MultiDimArrayAssoc of
          MultiDimArrayAssocType.fields
    ]

  type fields = AggregateAssocType.fields =
    
  {
         
    f_designators: AlternativesListType.t
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_designators node =
    match (node :> AggregateAssocType.t) with
    | `AggregateAssoc fields ->
        Lazy.force fields.f_designators
    | `MultiDimArrayAssoc fields ->
        Lazy.force fields.f_designators
  let f_r_expr node =
    match (node :> AggregateAssocType.t) with
    | `AggregateAssoc fields ->
        Lazy.force fields.f_r_expr
    | `MultiDimArrayAssoc fields ->
        Lazy.force fields.f_r_expr



end

module BasicAssoc = struct
  type t =
    [
      | AggregateAssoc.t
      | DiscriminantAssoc.t
      | IteratedAssoc.t
      | ParamAssoc.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_get_params
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.basic_assoc_p_get_params
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      DefiningNameArray.wrap (context node) (!@ result_ptr)





end

module RecordDef = struct
  type t =
    [
      | `RecordDef of
          RecordDefType.fields
    ]

  type fields = RecordDefType.fields =
    
  {
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_components node =
    match (node :> RecordDefType.t) with
    | `RecordDef fields ->
        Lazy.force fields.f_components



end

module NullRecordDef = struct
  type t =
    [
      | `NullRecordDef of
          NullRecordDefType.fields
    ]

  type fields = NullRecordDefType.fields =
    
  {
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_components node =
    match (node :> NullRecordDefType.t) with
    | `NullRecordDef fields ->
        Lazy.force fields.f_components



end

module BaseRecordDef = struct
  type t =
    [
      | NullRecordDef.t
      | RecordDef.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_components node =
    match (node :> BaseRecordDefType.t) with
    | `NullRecordDef fields ->
        Lazy.force fields.f_components
    | `RecordDef fields ->
        Lazy.force fields.f_components



end

module GenericFormalPart = struct
  type t =
    [
      | `GenericFormalPart of
          GenericFormalPartType.fields
    ]

  type fields = GenericFormalPartType.fields =
    
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_decls node =
    match (node :> GenericFormalPartType.t) with
    | `GenericFormalPart fields ->
        Lazy.force fields.f_decls



end

module EntryCompletionFormalParams = struct
  type t =
    [
      | `EntryCompletionFormalParams of
          EntryCompletionFormalParamsType.fields
    ]

  type fields = EntryCompletionFormalParamsType.fields =
    
  {
         
    f_params: ParamsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_params node =
    match (node :> EntryCompletionFormalParamsType.t) with
    | `EntryCompletionFormalParams fields ->
        Lazy.force fields.f_params



end

module UnknownDiscriminantPart = struct
  type t =
    [
      | `UnknownDiscriminantPart of
          UnknownDiscriminantPartType.fields
    ]

  type fields = UnknownDiscriminantPartType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module KnownDiscriminantPart = struct
  type t =
    [
      | `KnownDiscriminantPart of
          KnownDiscriminantPartType.fields
    ]

  type fields = KnownDiscriminantPartType.fields =
    
  {
         
    f_discr_specs: DiscriminantSpecListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_discr_specs node =
    match (node :> KnownDiscriminantPartType.t) with
    | `KnownDiscriminantPart fields ->
        Lazy.force fields.f_discr_specs



end

module DiscriminantPart = struct
  type t =
    [
      | KnownDiscriminantPart.t
      | UnknownDiscriminantPart.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module ComponentList = struct
  type t =
    [
      | `ComponentList of
          ComponentListType.fields
    ]

  type fields = ComponentListType.fields =
    
  {
         
    f_components: AdaNodeListType.t
    option
    Lazy.t;
         
    f_variant_part: VariantPartType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_components node =
    match (node :> ComponentListType.t) with
    | `ComponentList fields ->
        Lazy.force fields.f_components
  let f_variant_part node =
    match (node :> ComponentListType.t) with
    | `ComponentList fields ->
        Lazy.force fields.f_variant_part



end

module SubpSpec = struct
  type t =
    [
      | `SubpSpec of
          SubpSpecType.fields
    ]

  type fields = SubpSpecType.fields =
    
  {
         
    f_subp_kind: SubpKindType.t
    Lazy.t;
         
    f_subp_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_subp_params: ParamsType.t
    option
    Lazy.t;
         
    f_subp_returns: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_subp_kind node =
    match (node :> SubpSpecType.t) with
    | `SubpSpec fields ->
        Lazy.force fields.f_subp_kind
  let f_subp_name node =
    match (node :> SubpSpecType.t) with
    | `SubpSpec fields ->
        Lazy.force fields.f_subp_name
  let f_subp_params node =
    match (node :> SubpSpecType.t) with
    | `SubpSpec fields ->
        Lazy.force fields.f_subp_params
  let f_subp_returns node =
    match (node :> SubpSpecType.t) with
    | `SubpSpec fields ->
        Lazy.force fields.f_subp_returns



end

module EnumSubpSpec = struct
  type t =
    [
      | `EnumSubpSpec of
          EnumSubpSpecType.fields
    ]

  type fields = EnumSubpSpecType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module EntrySpec = struct
  type t =
    [
      | `EntrySpec of
          EntrySpecType.fields
    ]

  type fields = EntrySpecType.fields =
    
  {
         
    f_entry_name: DefiningNameType.t
    Lazy.t;
         
    f_family_type: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_entry_params: ParamsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_entry_name node =
    match (node :> EntrySpecType.t) with
    | `EntrySpec fields ->
        Lazy.force fields.f_entry_name
  let f_family_type node =
    match (node :> EntrySpecType.t) with
    | `EntrySpec fields ->
        Lazy.force fields.f_family_type
  let f_entry_params node =
    match (node :> EntrySpecType.t) with
    | `EntrySpec fields ->
        Lazy.force fields.f_entry_params



end

module BaseSubpSpec = struct
  type t =
    [
      | EntrySpec.t
      | EnumSubpSpec.t
      | SubpSpec.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_returns
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_subp_spec_p_returns
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (TypeExprType.wrap (context node) (!@ result_ptr))

let p_params
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_subp_spec_p_params
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      ParamSpecArray.wrap (context node) (!@ result_ptr)

let p_primitive_subp_types
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.base_subp_spec_p_primitive_subp_types
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      BaseTypeDeclArray.wrap (context node) (!@ result_ptr)

let p_primitive_subp_first_type
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.base_subp_spec_p_primitive_subp_first_type
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_primitive_subp_tagged_type
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.base_subp_spec_p_primitive_subp_tagged_type
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))

let p_return_type
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_subp_spec_p_return_type
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BaseTypeDeclType.wrap (context node) (!@ result_ptr))





end

module BaseFormalParamHolder = struct
  type t =
    [
      | BaseSubpSpec.t
      | ComponentList.t
      | DiscriminantPart.t
      | EntryCompletionFormalParams.t
      | GenericFormalPart.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_abstract_formal_params
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_formal_param_holder_p_abstract_formal_params
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      BaseFormalParamDeclArray.wrap (context node) (!@ result_ptr)

let p_nb_min_params
    (node)
    =
      let result_ptr =
        allocate_n int ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_formal_param_holder_p_nb_min_params
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_nb_max_params
    (node)
    =
      let result_ptr =
        allocate_n int ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_formal_param_holder_p_nb_max_params
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_param_types
    ?origin
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_origin =
           match origin with
           | Some node ->
               addr (AdaNodeType.unwrap (node))
           | None -> allocate_n EntityStruct.c_type ~count:1
      in
      let _ : int =
        CFunctions.base_formal_param_holder_p_param_types
          (addr (AdaNodeType.unwrap (node)))
          c_origin
          (result_ptr)
      in
         
      BaseTypeDeclArray.wrap (context node) (!@ result_ptr)





end

module PragmaArgumentAssoc = struct
  type t =
    [
      | `PragmaArgumentAssoc of
          PragmaArgumentAssocType.fields
    ]

  type fields = PragmaArgumentAssocType.fields =
    
  {
         
    f_id: IdentifierType.t
    option
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_id node =
    match (node :> PragmaArgumentAssocType.t) with
    | `PragmaArgumentAssoc fields ->
        Lazy.force fields.f_id
  let f_expr node =
    match (node :> PragmaArgumentAssocType.t) with
    | `PragmaArgumentAssoc fields ->
        Lazy.force fields.f_expr



end

module ContractCaseAssoc = struct
  type t =
    [
      | `ContractCaseAssoc of
          ContractCaseAssocType.fields
    ]

  type fields = ContractCaseAssocType.fields =
    
  {
         
    f_guard: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `OthersDesignator
          of OthersDesignatorType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_consequence: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_guard node =
    match (node :> ContractCaseAssocType.t) with
    | `ContractCaseAssoc fields ->
        Lazy.force fields.f_guard
  let f_consequence node =
    match (node :> ContractCaseAssocType.t) with
    | `ContractCaseAssoc fields ->
        Lazy.force fields.f_consequence



end

module BaseAssoc = struct
  type t =
    [
      | ContractCaseAssoc.t
      | PragmaArgumentAssoc.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_assoc_expr
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.base_assoc_p_assoc_expr
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (ExprType.wrap (context node) (!@ result_ptr))





end

module AspectSpec = struct
  type t =
    [
      | `AspectSpec of
          AspectSpecType.fields
    ]

  type fields = AspectSpecType.fields =
    
  {
         
    f_aspect_assocs: AspectAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_aspect_assocs node =
    match (node :> AspectSpecType.t) with
    | `AspectSpec fields ->
        Lazy.force fields.f_aspect_assocs



end

module RecordRepClause = struct
  type t =
    [
      | `RecordRepClause of
          RecordRepClauseType.fields
    ]

  type fields = RecordRepClauseType.fields =
    
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_at_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_components: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> RecordRepClauseType.t) with
    | `RecordRepClause fields ->
        Lazy.force fields.f_name
  let f_at_expr node =
    match (node :> RecordRepClauseType.t) with
    | `RecordRepClause fields ->
        Lazy.force fields.f_at_expr
  let f_components node =
    match (node :> RecordRepClauseType.t) with
    | `RecordRepClause fields ->
        Lazy.force fields.f_components



end

module EnumRepClause = struct
  type t =
    [
      | `EnumRepClause of
          EnumRepClauseType.fields
    ]

  type fields = EnumRepClauseType.fields =
    
  {
         
    f_type_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_aggregate: BaseAggregateType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_type_name node =
    match (node :> EnumRepClauseType.t) with
    | `EnumRepClause fields ->
        Lazy.force fields.f_type_name
  let f_aggregate node =
    match (node :> EnumRepClauseType.t) with
    | `EnumRepClause fields ->
        Lazy.force fields.f_aggregate



end

module AttributeDefClause = struct
  type t =
    [
      | `AttributeDefClause of
          AttributeDefClauseType.fields
    ]

  type fields = AttributeDefClauseType.fields =
    
  {
         
    f_attribute_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_attribute_expr node =
    match (node :> AttributeDefClauseType.t) with
    | `AttributeDefClause fields ->
        Lazy.force fields.f_attribute_expr
  let f_expr node =
    match (node :> AttributeDefClauseType.t) with
    | `AttributeDefClause fields ->
        Lazy.force fields.f_expr



end

module AtClause = struct
  type t =
    [
      | `AtClause of
          AtClauseType.fields
    ]

  type fields = AtClauseType.fields =
    
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_name node =
    match (node :> AtClauseType.t) with
    | `AtClause fields ->
        Lazy.force fields.f_name
  let f_expr node =
    match (node :> AtClauseType.t) with
    | `AtClause fields ->
        Lazy.force fields.f_expr



end

module AspectClause = struct
  type t =
    [
      | AtClause.t
      | AttributeDefClause.t
      | EnumRepClause.t
      | RecordRepClause.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AspectAssoc = struct
  type t =
    [
      | `AspectAssoc of
          AspectAssocType.fields
    ]

  type fields = AspectAssocType.fields =
    
  {
         
    f_id: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `AbstractStateDeclExpr
          of AbstractStateDeclExprType.fields
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ContractCases
          of ContractCasesType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_id node =
    match (node :> AspectAssocType.t) with
    | `AspectAssoc fields ->
        Lazy.force fields.f_id
  let f_expr node =
    match (node :> AspectAssocType.t) with
    | `AspectAssoc fields ->
        Lazy.force fields.f_expr



end

module UnconstrainedArrayIndices = struct
  type t =
    [
      | `UnconstrainedArrayIndices of
          UnconstrainedArrayIndicesType.fields
    ]

  type fields = UnconstrainedArrayIndicesType.fields =
    
  {
         
    f_types: UnconstrainedArrayIndexListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_types node =
    match (node :> UnconstrainedArrayIndicesType.t) with
    | `UnconstrainedArrayIndices fields ->
        Lazy.force fields.f_types



end

module ConstrainedArrayIndices = struct
  type t =
    [
      | `ConstrainedArrayIndices of
          ConstrainedArrayIndicesType.fields
    ]

  type fields = ConstrainedArrayIndicesType.fields =
    
  {
         
    f_list: ConstraintListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))


  let f_list node =
    match (node :> ConstrainedArrayIndicesType.t) with
    | `ConstrainedArrayIndices fields ->
        Lazy.force fields.f_list



end

module ArrayIndices = struct
  type t =
    [
      | ConstrainedArrayIndices.t
      | UnconstrainedArrayIndices.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AllPresent = struct
  type t =
    [
      | `AllPresent of
          AllPresentType.fields
    ]

  type fields = AllPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AllAbsent = struct
  type t =
    [
      | `AllAbsent of
          AllAbsentType.fields
    ]

  type fields = AllAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AllNode = struct
  type t =
    [
      | AllAbsent.t
      | AllPresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.all_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module AliasedPresent = struct
  type t =
    [
      | `AliasedPresent of
          AliasedPresentType.fields
    ]

  type fields = AliasedPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AliasedAbsent = struct
  type t =
    [
      | `AliasedAbsent of
          AliasedAbsentType.fields
    ]

  type fields = AliasedAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AliasedNode = struct
  type t =
    [
      | AliasedAbsent.t
      | AliasedPresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.aliased_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module VariantList = struct
  type t =
    [
      | `VariantList of
          VariantListType.fields
    ]

  type fields = VariantListType.fields =
    
  {
    list : VariantType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> VariantListType.t) with
    | `VariantList fields ->
        Lazy.force fields.list



end

module UnconstrainedArrayIndexList = struct
  type t =
    [
      | `UnconstrainedArrayIndexList of
          UnconstrainedArrayIndexListType.fields
    ]

  type fields = UnconstrainedArrayIndexListType.fields =
    
  {
    list : UnconstrainedArrayIndexType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> UnconstrainedArrayIndexListType.t) with
    | `UnconstrainedArrayIndexList fields ->
        Lazy.force fields.list



end

module SelectWhenPartList = struct
  type t =
    [
      | `SelectWhenPartList of
          SelectWhenPartListType.fields
    ]

  type fields = SelectWhenPartListType.fields =
    
  {
    list : SelectWhenPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> SelectWhenPartListType.t) with
    | `SelectWhenPartList fields ->
        Lazy.force fields.list



end

module PragmaNodeList = struct
  type t =
    [
      | `PragmaNodeList of
          PragmaNodeListType.fields
    ]

  type fields = PragmaNodeListType.fields =
    
  {
    list : PragmaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> PragmaNodeListType.t) with
    | `PragmaNodeList fields ->
        Lazy.force fields.list



end

module ParamSpecList = struct
  type t =
    [
      | `ParamSpecList of
          ParamSpecListType.fields
    ]

  type fields = ParamSpecListType.fields =
    
  {
    list : ParamSpecType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> ParamSpecListType.t) with
    | `ParamSpecList fields ->
        Lazy.force fields.list



end

module ParentList = struct
  type t =
    [
      | `ParentList of
          ParentListType.fields
    ]

  type fields = ParentListType.fields =
    
  {
    list : NameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> ParentListType.t) with
    | `ParentList fields ->
        Lazy.force fields.list



end

module NameList = struct
  type t =
    [
      | `NameList of
          NameListType.fields
      | `ParentList of
          ParentListType.fields
    ]

  type fields = NameListType.fields =
    
  {
    list : NameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> NameListType.t) with
    | `NameList fields ->
        Lazy.force fields.list
    | `ParentList fields ->
        Lazy.force fields.list



end

module DiscriminantChoiceList = struct
  type t =
    [
      | `DiscriminantChoiceList of
          DiscriminantChoiceListType.fields
    ]

  type fields = DiscriminantChoiceListType.fields =
    
  {
    list : IdentifierType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> DiscriminantChoiceListType.t) with
    | `DiscriminantChoiceList fields ->
        Lazy.force fields.list



end

module IdentifierList = struct
  type t =
    [
      | DiscriminantChoiceList.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> IdentifierListType.t) with
    | `DiscriminantChoiceList fields ->
        Lazy.force fields.list



end

module ExprAlternativesList = struct
  type t =
    [
      | `ExprAlternativesList of
          ExprAlternativesListType.fields
    ]

  type fields = ExprAlternativesListType.fields =
    
  {
    list : ExprType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> ExprAlternativesListType.t) with
    | `ExprAlternativesList fields ->
        Lazy.force fields.list



end

module ExprList = struct
  type t =
    [
      | ExprAlternativesList.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> ExprListType.t) with
    | `ExprAlternativesList fields ->
        Lazy.force fields.list



end

module EnumLiteralDeclList = struct
  type t =
    [
      | `EnumLiteralDeclList of
          EnumLiteralDeclListType.fields
    ]

  type fields = EnumLiteralDeclListType.fields =
    
  {
    list : EnumLiteralDeclType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> EnumLiteralDeclListType.t) with
    | `EnumLiteralDeclList fields ->
        Lazy.force fields.list



end

module ElsifStmtPartList = struct
  type t =
    [
      | `ElsifStmtPartList of
          ElsifStmtPartListType.fields
    ]

  type fields = ElsifStmtPartListType.fields =
    
  {
    list : ElsifStmtPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> ElsifStmtPartListType.t) with
    | `ElsifStmtPartList fields ->
        Lazy.force fields.list



end

module ElsifExprPartList = struct
  type t =
    [
      | `ElsifExprPartList of
          ElsifExprPartListType.fields
    ]

  type fields = ElsifExprPartListType.fields =
    
  {
    list : ElsifExprPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> ElsifExprPartListType.t) with
    | `ElsifExprPartList fields ->
        Lazy.force fields.list



end

module DiscriminantSpecList = struct
  type t =
    [
      | `DiscriminantSpecList of
          DiscriminantSpecListType.fields
    ]

  type fields = DiscriminantSpecListType.fields =
    
  {
    list : DiscriminantSpecType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> DiscriminantSpecListType.t) with
    | `DiscriminantSpecList fields ->
        Lazy.force fields.list



end

module DefiningNameList = struct
  type t =
    [
      | `DefiningNameList of
          DefiningNameListType.fields
    ]

  type fields = DefiningNameListType.fields =
    
  {
    list : DefiningNameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> DefiningNameListType.t) with
    | `DefiningNameList fields ->
        Lazy.force fields.list



end

module ContractCaseAssocList = struct
  type t =
    [
      | `ContractCaseAssocList of
          ContractCaseAssocListType.fields
    ]

  type fields = ContractCaseAssocListType.fields =
    
  {
    list : ContractCaseAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> ContractCaseAssocListType.t) with
    | `ContractCaseAssocList fields ->
        Lazy.force fields.list



end

module CompilationUnitList = struct
  type t =
    [
      | `CompilationUnitList of
          CompilationUnitListType.fields
    ]

  type fields = CompilationUnitListType.fields =
    
  {
    list : CompilationUnitType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> CompilationUnitListType.t) with
    | `CompilationUnitList fields ->
        Lazy.force fields.list



end

module CaseStmtAlternativeList = struct
  type t =
    [
      | `CaseStmtAlternativeList of
          CaseStmtAlternativeListType.fields
    ]

  type fields = CaseStmtAlternativeListType.fields =
    
  {
    list : CaseStmtAlternativeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> CaseStmtAlternativeListType.t) with
    | `CaseStmtAlternativeList fields ->
        Lazy.force fields.list



end

module CaseExprAlternativeList = struct
  type t =
    [
      | `CaseExprAlternativeList of
          CaseExprAlternativeListType.fields
    ]

  type fields = CaseExprAlternativeListType.fields =
    
  {
    list : CaseExprAlternativeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> CaseExprAlternativeListType.t) with
    | `CaseExprAlternativeList fields ->
        Lazy.force fields.list



end

module BasicDeclList = struct
  type t =
    [
      | `BasicDeclList of
          BasicDeclListType.fields
    ]

  type fields = BasicDeclListType.fields =
    
  {
    list : BasicDeclType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> BasicDeclListType.t) with
    | `BasicDeclList fields ->
        Lazy.force fields.list



end

module AssocList = struct
  type t =
    [
      | `AssocList of
          AssocListType.fields
    ]

  type fields = AssocListType.fields =
    
  {
    list : BasicAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_zip_with_params
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n ParamActualArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.assoc_list_p_zip_with_params
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      ParamActualArray.wrap (context node) (!@ result_ptr)



  let f_list node =
    match (node :> AssocListType.t) with
    | `AssocList fields ->
        Lazy.force fields.list



end

module BasicAssocList = struct
  type t =
    [
      | AssocList.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> BasicAssocListType.t) with
    | `AssocList fields ->
        Lazy.force fields.list



end

module BaseAssocList = struct
  type t =
    [
      | `BaseAssocList of
          BaseAssocListType.fields
    ]

  type fields = BaseAssocListType.fields =
    
  {
    list : BaseAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> BaseAssocListType.t) with
    | `BaseAssocList fields ->
        Lazy.force fields.list



end

module AspectAssocList = struct
  type t =
    [
      | `AspectAssocList of
          AspectAssocListType.fields
    ]

  type fields = AspectAssocListType.fields =
    
  {
    list : AspectAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> AspectAssocListType.t) with
    | `AspectAssocList fields ->
        Lazy.force fields.list



end

module StmtList = struct
  type t =
    [
      | `StmtList of
          StmtListType.fields
    ]

  type fields = StmtListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> StmtListType.t) with
    | `StmtList fields ->
        Lazy.force fields.list



end

module DeclList = struct
  type t =
    [
      | `DeclList of
          DeclListType.fields
    ]

  type fields = DeclListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> DeclListType.t) with
    | `DeclList fields ->
        Lazy.force fields.list



end

module ConstraintList = struct
  type t =
    [
      | `ConstraintList of
          ConstraintListType.fields
    ]

  type fields = ConstraintListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> ConstraintListType.t) with
    | `ConstraintList fields ->
        Lazy.force fields.list



end

module AlternativesList = struct
  type t =
    [
      | `AlternativesList of
          AlternativesListType.fields
    ]

  type fields = AlternativesListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> AlternativesListType.t) with
    | `AlternativesList fields ->
        Lazy.force fields.list



end

module AbstractStateDeclList = struct
  type t =
    [
      | `AbstractStateDeclList of
          AbstractStateDeclListType.fields
    ]

  type fields = AbstractStateDeclListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> AbstractStateDeclListType.t) with
    | `AbstractStateDeclList fields ->
        Lazy.force fields.list



end

module AdaNodeList = struct
  type t =
    [
      | `AdaNodeList of
          AdaNodeListType.fields
      | `AbstractStateDeclList of
          AbstractStateDeclListType.fields
      | `AlternativesList of
          AlternativesListType.fields
      | `ConstraintList of
          ConstraintListType.fields
      | `DeclList of
          DeclListType.fields
      | `StmtList of
          StmtListType.fields
    ]

  type fields = AdaNodeListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))



  let f_list node =
    match (node :> AdaNodeListType.t) with
    | `AdaNodeList fields ->
        Lazy.force fields.list
    | `AbstractStateDeclList fields ->
        Lazy.force fields.list
    | `AlternativesList fields ->
        Lazy.force fields.list
    | `ConstraintList fields ->
        Lazy.force fields.list
    | `DeclList fields ->
        Lazy.force fields.list
    | `StmtList fields ->
        Lazy.force fields.list



end

module AdaList = struct
  type t =
    [
      | AdaNodeList.t
      | AspectAssocList.t
      | BaseAssocList.t
      | BasicAssocList.t
      | BasicDeclList.t
      | CaseExprAlternativeList.t
      | CaseStmtAlternativeList.t
      | CompilationUnitList.t
      | ContractCaseAssocList.t
      | DefiningNameList.t
      | DiscriminantSpecList.t
      | ElsifExprPartList.t
      | ElsifStmtPartList.t
      | EnumLiteralDeclList.t
      | ExprList.t
      | IdentifierList.t
      | NameList.t
      | ParamSpecList.t
      | PragmaNodeList.t
      | SelectWhenPartList.t
      | UnconstrainedArrayIndexList.t
      | VariantList.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AbstractPresent = struct
  type t =
    [
      | `AbstractPresent of
          AbstractPresentType.fields
    ]

  type fields = AbstractPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AbstractAbsent = struct
  type t =
    [
      | `AbstractAbsent of
          AbstractAbsentType.fields
    ]

  type fields = AbstractAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AbstractNode = struct
  type t =
    [
      | AbstractAbsent.t
      | AbstractPresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.abstract_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module AbortPresent = struct
  type t =
    [
      | `AbortPresent of
          AbortPresentType.fields
    ]

  type fields = AbortPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AbortAbsent = struct
  type t =
    [
      | `AbortAbsent of
          AbortAbsentType.fields
    ]

  type fields = AbortAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))





end

module AbortNode = struct
  type t =
    [
      | AbortAbsent.t
      | AbortPresent.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_as_bool
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.abort_node_p_as_bool
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr





end

module AdaNode = struct
  type t =
    [
      | AbortNode.t
      | AbstractNode.t
      | AdaList.t
      | AliasedNode.t
      | AllNode.t
      | ArrayIndices.t
      | AspectAssoc.t
      | AspectClause.t
      | AspectSpec.t
      | BaseAssoc.t
      | BaseFormalParamHolder.t
      | BaseRecordDef.t
      | BasicAssoc.t
      | BasicDecl.t
      | CaseStmtAlternative.t
      | CompilationUnit.t
      | ComponentClause.t
      | ComponentDef.t
      | ConstantNode.t
      | Constraint.t
      | DeclarativePart.t
      | ElsifExprPart.t
      | ElsifStmtPart.t
      | Expr.t
      | HandledStmts.t
      | InterfaceKind.t
      | IterType.t
      | LibraryItem.t
      | LimitedNode.t
      | LoopSpec.t
      | Mode.t
      | MultiAbstractStateDecl.t
      | NotNull.t
      | NullComponentDecl.t
      | OthersDesignator.t
      | OverridingNode.t
      | Params.t
      | ParenAbstractStateDecl.t
      | PragmaNode.t
      | PrivateNode.t
      | ProtectedDef.t
      | ProtectedNode.t
      | Quantifier.t
      | RangeSpec.t
      | RenamingClause.t
      | ReverseNode.t
      | SelectWhenPart.t
      | Stmt.t
      | SubpKind.t
      | Subunit.t
      | SynchronizedNode.t
      | TaggedNode.t
      | TaskDef.t
      | TypeDef.t
      | TypeExpr.t
      | UnconstrainedArrayIndex.t
      | UntilNode.t
      | UseClause.t
      | Variant.t
      | VariantPart.t
      | WithClause.t
      | WithPrivate.t
    ]


  let equal node1 node2 =
    Entity.equal
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let compare node1 node2 =
    Entity.compare
      (AdaNodeType.unwrap ((node1 :> AdaNodeType.t)))
      (AdaNodeType.unwrap ((node2 :> AdaNodeType.t)))

  let hash node =
    Entity.hash
      (AdaNodeType.unwrap ((node :> AdaNodeType.t)))

let p_declarative_scope
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_declarative_scope
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (DeclarativePartType.wrap (context node) (!@ result_ptr))

let p_complete
    (node)
    =
      raise (Failure "iterators not implemented yet")

let p_valid_keywords
    (node)
    =
      let result_ptr =
        allocate_n UnboundedTextTypeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_valid_keywords
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      UnboundedTextTypeArray.wrap (!@ result_ptr)

let p_generic_instantiations
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_generic_instantiations
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      GenericInstantiationArray.wrap (context node) (!@ result_ptr)

let p_semantic_parent
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_semantic_parent
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AdaNodeType.wrap (context node) (!@ result_ptr))

let p_parent_basic_decl
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_parent_basic_decl
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_filter_is_imported_by
    (node)
    (units)
    (transitive)
    =
      let result_ptr =
        allocate_n AnalysisUnitArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_units =
            
        AnalysisUnitArray.unwrap (units)
      in
      let c_transitive =
            
        transitive
      in
      let _ : int =
        CFunctions.ada_node_p_filter_is_imported_by
          (addr (AdaNodeType.unwrap (node)))
          c_units
          c_transitive
          (result_ptr)
      in
         
               AnalysisUnitArrayStruct.dec_ref c_units ;
         
      AnalysisUnitArray.wrap (context node) (!@ result_ptr)

let p_xref_entry_point
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_xref_entry_point
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_resolve_names
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_resolve_names
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let p_standard_unit
    (node)
    =
      let result_ptr =
        allocate_n AnalysisUnitStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_standard_unit
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      AnalysisUnit.wrap (context node) (!@ result_ptr)

let p_std_entity
    (node)
    (sym)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_sym =
            
        allocate Symbol.c_type (Symbol.unwrap (context node).c_value (sym))
      in
      let _ : int =
        CFunctions.ada_node_p_std_entity
          (addr (AdaNodeType.unwrap (node)))
          c_sym
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AdaNodeType.wrap (context node) (!@ result_ptr))

let p_bool_type
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_bool_type
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AdaNodeType.wrap (context node) (!@ result_ptr))

let p_int_type
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_int_type
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AdaNodeType.wrap (context node) (!@ result_ptr))

let p_universal_int_type
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_universal_int_type
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AdaNodeType.wrap (context node) (!@ result_ptr))

let p_universal_real_type
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_p_universal_real_type
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AdaNodeType.wrap (context node) (!@ result_ptr))

let p_top_level_decl
    (node)
    (unit)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_unit =
            
        AnalysisUnit.unwrap (unit)
      in
      let _ : int =
        CFunctions.ada_node_p_top_level_decl
          (addr (AdaNodeType.unwrap (node)))
          c_unit
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (BasicDeclType.wrap (context node) (!@ result_ptr))

let p_choice_match
    (node)
    (value)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_value =
            
        allocate BigInteger.c_type (BigInteger.unwrap (value))
      in
      let _ : int =
        CFunctions.ada_node_p_choice_match
          (addr (AdaNodeType.unwrap (node)))
          c_value
          (result_ptr)
      in
         
               BigInteger.decref (!@ c_value) ;
      !@ result_ptr

let p_gnat_xref
    ?(imprecise_fallback=false)
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_imprecise_fallback =
            
        imprecise_fallback
      in
      let _ : int =
        CFunctions.ada_node_p_gnat_xref
          (addr (AdaNodeType.unwrap (node)))
          c_imprecise_fallback
          (result_ptr)
      in
         
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (DefiningNameType.wrap (context node) (!@ result_ptr))

let parent
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_parent
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AdaNodeType.wrap (context node) (!@ result_ptr))

let parents
    ?(with_self=true)
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let c_with_self =
            
        with_self
      in
      let _ : int =
        CFunctions.ada_node_parents
          (addr (AdaNodeType.unwrap (node)))
          c_with_self
          (result_ptr)
      in
         
      AdaNodeArray.wrap (context node) (!@ result_ptr)

let children
    (node)
    =
      let result_ptr =
        allocate_n AdaNodeArrayStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_children
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      AdaNodeArray.wrap (context node) (!@ result_ptr)

let token_start
    (node)
    =
      let result_ptr =
        allocate_n Token.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_token_start
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let token_end
    (node)
    =
      let result_ptr =
        allocate_n Token.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_token_end
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let child_index
    (node)
    =
      let result_ptr =
        allocate_n int ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_child_index
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let previous_sibling
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_previous_sibling
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AdaNodeType.wrap (context node) (!@ result_ptr))

let next_sibling
    (node)
    =
      let result_ptr =
        allocate_n EntityStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_next_sibling
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AdaNodeType.wrap (context node) (!@ result_ptr))

let unit
    (node)
    =
      let result_ptr =
        allocate_n AnalysisUnitStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_unit
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      AnalysisUnit.wrap (context node) (!@ result_ptr)

let is_ghost
    (node)
    =
      let result_ptr =
        allocate_n bool ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_is_ghost
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      !@ result_ptr

let full_sloc_image
    (node)
    =
      let result_ptr =
        allocate_n TextTypeStruct.c_type ~count:1
      in
      (* The result of this call should already be checked by the raisable
         mechanism *)
      let _ : int =
        CFunctions.ada_node_full_sloc_image
          (addr (AdaNodeType.unwrap (node)))
          (result_ptr)
      in
      TextType.wrap (!@ result_ptr)





  let kind_name = function
    | #WithPrivatePresentType.t ->
        "WithPrivatePresent"
    | #WithPrivateAbsentType.t ->
        "WithPrivateAbsent"
    | #WithClauseType.t ->
        "WithClause"
    | #VariantPartType.t ->
        "VariantPart"
    | #VariantType.t ->
        "Variant"
    | #UseTypeClauseType.t ->
        "UseTypeClause"
    | #UsePackageClauseType.t ->
        "UsePackageClause"
    | #UntilPresentType.t ->
        "UntilPresent"
    | #UntilAbsentType.t ->
        "UntilAbsent"
    | #UnconstrainedArrayIndexType.t ->
        "UnconstrainedArrayIndex"
    | #DiscreteSubtypeIndicationType.t ->
        "DiscreteSubtypeIndication"
    | #ConstrainedSubtypeIndicationType.t ->
        "ConstrainedSubtypeIndication"
    | #SubtypeIndicationType.t ->
        "SubtypeIndication"
    | #EnumLitSynthTypeExprType.t ->
        "EnumLitSynthTypeExpr"
    | #AnonymousTypeType.t ->
        "AnonymousType"
    | #SignedIntTypeDefType.t ->
        "SignedIntTypeDef"
    | #RecordTypeDefType.t ->
        "RecordTypeDef"
    | #OrdinaryFixedPointDefType.t ->
        "OrdinaryFixedPointDef"
    | #FloatingPointDefType.t ->
        "FloatingPointDef"
    | #DecimalFixedPointDefType.t ->
        "DecimalFixedPointDef"
    | #PrivateTypeDefType.t ->
        "PrivateTypeDef"
    | #ModIntTypeDefType.t ->
        "ModIntTypeDef"
    | #InterfaceTypeDefType.t ->
        "InterfaceTypeDef"
    | #FormalDiscreteTypeDefType.t ->
        "FormalDiscreteTypeDef"
    | #EnumTypeDefType.t ->
        "EnumTypeDef"
    | #DerivedTypeDefType.t ->
        "DerivedTypeDef"
    | #ArrayTypeDefType.t ->
        "ArrayTypeDef"
    | #TypeAccessDefType.t ->
        "TypeAccessDef"
    | #AnonymousTypeAccessDefType.t ->
        "AnonymousTypeAccessDef"
    | #AccessToSubpDefType.t ->
        "AccessToSubpDef"
    | #TaskDefType.t ->
        "TaskDef"
    | #TaggedPresentType.t ->
        "TaggedPresent"
    | #TaggedAbsentType.t ->
        "TaggedAbsent"
    | #SynchronizedPresentType.t ->
        "SynchronizedPresent"
    | #SynchronizedAbsentType.t ->
        "SynchronizedAbsent"
    | #SubunitType.t ->
        "Subunit"
    | #SubpKindProcedureType.t ->
        "SubpKindProcedure"
    | #SubpKindFunctionType.t ->
        "SubpKindFunction"
    | #TerminateAlternativeType.t ->
        "TerminateAlternative"
    | #ReturnStmtType.t ->
        "ReturnStmt"
    | #RequeueStmtType.t ->
        "RequeueStmt"
    | #RaiseStmtType.t ->
        "RaiseStmt"
    | #NullStmtType.t ->
        "NullStmt"
    | #LabelType.t ->
        "Label"
    | #GotoStmtType.t ->
        "GotoStmt"
    | #ExitStmtType.t ->
        "ExitStmt"
    | #DelayStmtType.t ->
        "DelayStmt"
    | #CallStmtType.t ->
        "CallStmt"
    | #AssignStmtType.t ->
        "AssignStmt"
    | #AbortStmtType.t ->
        "AbortStmt"
    | #ErrorStmtType.t ->
        "ErrorStmt"
    | #SelectStmtType.t ->
        "SelectStmt"
    | #NamedStmtType.t ->
        "NamedStmt"
    | #IfStmtType.t ->
        "IfStmt"
    | #ExtendedReturnStmtType.t ->
        "ExtendedReturnStmt"
    | #CaseStmtType.t ->
        "CaseStmt"
    | #DeclBlockType.t ->
        "DeclBlock"
    | #BeginBlockType.t ->
        "BeginBlock"
    | #WhileLoopStmtType.t ->
        "WhileLoopStmt"
    | #LoopStmtType.t ->
        "LoopStmt"
    | #ForLoopStmtType.t ->
        "ForLoopStmt"
    | #AcceptStmtWithStmtsType.t ->
        "AcceptStmtWithStmts"
    | #AcceptStmtType.t ->
        "AcceptStmt"
    | #SelectWhenPartType.t ->
        "SelectWhenPart"
    | #ReversePresentType.t ->
        "ReversePresent"
    | #ReverseAbsentType.t ->
        "ReverseAbsent"
    | #SyntheticRenamingClauseType.t ->
        "SyntheticRenamingClause"
    | #RenamingClauseType.t ->
        "RenamingClause"
    | #RangeSpecType.t ->
        "RangeSpec"
    | #QuantifierSomeType.t ->
        "QuantifierSome"
    | #QuantifierAllType.t ->
        "QuantifierAll"
    | #ProtectedPresentType.t ->
        "ProtectedPresent"
    | #ProtectedAbsentType.t ->
        "ProtectedAbsent"
    | #ProtectedDefType.t ->
        "ProtectedDef"
    | #PrivatePresentType.t ->
        "PrivatePresent"
    | #PrivateAbsentType.t ->
        "PrivateAbsent"
    | #PragmaNodeType.t ->
        "PragmaNode"
    | #ParenAbstractStateDeclType.t ->
        "ParenAbstractStateDecl"
    | #ParamsType.t ->
        "Params"
    | #OverridingUnspecifiedType.t ->
        "OverridingUnspecified"
    | #OverridingOverridingType.t ->
        "OverridingOverriding"
    | #OverridingNotOverridingType.t ->
        "OverridingNotOverriding"
    | #OthersDesignatorType.t ->
        "OthersDesignator"
    | #NullComponentDeclType.t ->
        "NullComponentDecl"
    | #NotNullPresentType.t ->
        "NotNullPresent"
    | #NotNullAbsentType.t ->
        "NotNullAbsent"
    | #MultiAbstractStateDeclType.t ->
        "MultiAbstractStateDecl"
    | #ModeOutType.t ->
        "ModeOut"
    | #ModeInOutType.t ->
        "ModeInOut"
    | #ModeInType.t ->
        "ModeIn"
    | #ModeDefaultType.t ->
        "ModeDefault"
    | #WhileLoopSpecType.t ->
        "WhileLoopSpec"
    | #ForLoopSpecType.t ->
        "ForLoopSpec"
    | #LimitedPresentType.t ->
        "LimitedPresent"
    | #LimitedAbsentType.t ->
        "LimitedAbsent"
    | #LibraryItemType.t ->
        "LibraryItem"
    | #IterTypeOfType.t ->
        "IterTypeOf"
    | #IterTypeInType.t ->
        "IterTypeIn"
    | #InterfaceKindTaskType.t ->
        "InterfaceKindTask"
    | #InterfaceKindSynchronizedType.t ->
        "InterfaceKindSynchronized"
    | #InterfaceKindProtectedType.t ->
        "InterfaceKindProtected"
    | #InterfaceKindLimitedType.t ->
        "InterfaceKindLimited"
    | #HandledStmtsType.t ->
        "HandledStmts"
    | #UnOpType.t ->
        "UnOp"
    | #RaiseExprType.t ->
        "RaiseExpr"
    | #QuantifiedExprType.t ->
        "QuantifiedExpr"
    | #ParenExprType.t ->
        "ParenExpr"
    | #TargetNameType.t ->
        "TargetName"
    | #RealLiteralType.t ->
        "RealLiteral"
    | #IntLiteralType.t ->
        "IntLiteral"
    | #NullLiteralType.t ->
        "NullLiteral"
    | #StringLiteralType.t ->
        "StringLiteral"
    | #OpXorType.t ->
        "OpXor"
    | #OpRemType.t ->
        "OpRem"
    | #OpPowType.t ->
        "OpPow"
    | #OpPlusType.t ->
        "OpPlus"
    | #OpOrElseType.t ->
        "OpOrElse"
    | #OpOrType.t ->
        "OpOr"
    | #OpNotInType.t ->
        "OpNotIn"
    | #OpNotType.t ->
        "OpNot"
    | #OpNeqType.t ->
        "OpNeq"
    | #OpMultType.t ->
        "OpMult"
    | #OpModType.t ->
        "OpMod"
    | #OpMinusType.t ->
        "OpMinus"
    | #OpLteType.t ->
        "OpLte"
    | #OpLtType.t ->
        "OpLt"
    | #OpInType.t ->
        "OpIn"
    | #OpGteType.t ->
        "OpGte"
    | #OpGtType.t ->
        "OpGt"
    | #OpEqType.t ->
        "OpEq"
    | #OpDoubleDotType.t ->
        "OpDoubleDot"
    | #OpDivType.t ->
        "OpDiv"
    | #OpConcatType.t ->
        "OpConcat"
    | #OpAndThenType.t ->
        "OpAndThen"
    | #OpAndType.t ->
        "OpAnd"
    | #OpAbsType.t ->
        "OpAbs"
    | #IdentifierType.t ->
        "Identifier"
    | #CharLiteralType.t ->
        "CharLiteral"
    | #QualExprType.t ->
        "QualExpr"
    | #ExplicitDerefType.t ->
        "ExplicitDeref"
    | #EndNameType.t ->
        "EndName"
    | #DottedNameType.t ->
        "DottedName"
    | #DiscreteSubtypeNameType.t ->
        "DiscreteSubtypeName"
    | #DefiningNameType.t ->
        "DefiningName"
    | #CallExprType.t ->
        "CallExpr"
    | #UpdateAttributeRefType.t ->
        "UpdateAttributeRef"
    | #AttributeRefType.t ->
        "AttributeRef"
    | #MembershipExprType.t ->
        "MembershipExpr"
    | #DeclExprType.t ->
        "DeclExpr"
    | #ContractCasesType.t ->
        "ContractCases"
    | #IfExprType.t ->
        "IfExpr"
    | #CaseExprType.t ->
        "CaseExpr"
    | #CaseExprAlternativeType.t ->
        "CaseExprAlternative"
    | #BoxExprType.t ->
        "BoxExpr"
    | #RelationOpType.t ->
        "RelationOp"
    | #BinOpType.t ->
        "BinOp"
    | #NullRecordAggregateType.t ->
        "NullRecordAggregate"
    | #BracketDeltaAggregateType.t ->
        "BracketDeltaAggregate"
    | #DeltaAggregateType.t ->
        "DeltaAggregate"
    | #BracketAggregateType.t ->
        "BracketAggregate"
    | #AggregateType.t ->
        "Aggregate"
    | #AllocatorType.t ->
        "Allocator"
    | #AbstractStateDeclExprType.t ->
        "AbstractStateDeclExpr"
    | #ElsifStmtPartType.t ->
        "ElsifStmtPart"
    | #ElsifExprPartType.t ->
        "ElsifExprPart"
    | #PublicPartType.t ->
        "PublicPart"
    | #PrivatePartType.t ->
        "PrivatePart"
    | #DeclarativePartType.t ->
        "DeclarativePart"
    | #RangeConstraintType.t ->
        "RangeConstraint"
    | #IndexConstraintType.t ->
        "IndexConstraint"
    | #DiscriminantConstraintType.t ->
        "DiscriminantConstraint"
    | #DigitsConstraintType.t ->
        "DigitsConstraint"
    | #DeltaConstraintType.t ->
        "DeltaConstraint"
    | #ConstantPresentType.t ->
        "ConstantPresent"
    | #ConstantAbsentType.t ->
        "ConstantAbsent"
    | #ComponentDefType.t ->
        "ComponentDef"
    | #ComponentClauseType.t ->
        "ComponentClause"
    | #CompilationUnitType.t ->
        "CompilationUnit"
    | #CaseStmtAlternativeType.t ->
        "CaseStmtAlternative"
    | #SingleTaskDeclType.t ->
        "SingleTaskDecl"
    | #SingleProtectedDeclType.t ->
        "SingleProtectedDecl"
    | #PackageRenamingDeclType.t ->
        "PackageRenamingDecl"
    | #ExtendedReturnStmtObjectDeclType.t ->
        "ExtendedReturnStmtObjectDecl"
    | #ObjectDeclType.t ->
        "ObjectDecl"
    | #NumberDeclType.t ->
        "NumberDecl"
    | #NamedStmtDeclType.t ->
        "NamedStmtDecl"
    | #LabelDeclType.t ->
        "LabelDecl"
    | #GenericSubpRenamingDeclType.t ->
        "GenericSubpRenamingDecl"
    | #GenericPackageRenamingDeclType.t ->
        "GenericPackageRenamingDecl"
    | #GenericSubpInstantiationType.t ->
        "GenericSubpInstantiation"
    | #GenericPackageInstantiationType.t ->
        "GenericPackageInstantiation"
    | #GenericSubpDeclType.t ->
        "GenericSubpDecl"
    | #GenericPackageDeclType.t ->
        "GenericPackageDecl"
    | #ForLoopVarDeclType.t ->
        "ForLoopVarDecl"
    | #ExceptionHandlerType.t ->
        "ExceptionHandler"
    | #ExceptionDeclType.t ->
        "ExceptionDecl"
    | #ErrorDeclType.t ->
        "ErrorDecl"
    | #EntryIndexSpecType.t ->
        "EntryIndexSpec"
    | #TaskBodyType.t ->
        "TaskBody"
    | #ProtectedBodyType.t ->
        "ProtectedBody"
    | #PackageBodyType.t ->
        "PackageBody"
    | #EntryBodyType.t ->
        "EntryBody"
    | #TaskBodyStubType.t ->
        "TaskBodyStub"
    | #SubpBodyStubType.t ->
        "SubpBodyStub"
    | #ProtectedBodyStubType.t ->
        "ProtectedBodyStub"
    | #PackageBodyStubType.t ->
        "PackageBodyStub"
    | #SubpRenamingDeclType.t ->
        "SubpRenamingDecl"
    | #SubpBodyType.t ->
        "SubpBody"
    | #NullSubpDeclType.t ->
        "NullSubpDecl"
    | #ExprFunctionType.t ->
        "ExprFunction"
    | #GenericSubpInternalType.t ->
        "GenericSubpInternal"
    | #EnumLiteralDeclType.t ->
        "EnumLiteralDecl"
    | #EntryDeclType.t ->
        "EntryDecl"
    | #SubpDeclType.t ->
        "SubpDecl"
    | #ConcreteFormalSubpDeclType.t ->
        "ConcreteFormalSubpDecl"
    | #AbstractFormalSubpDeclType.t ->
        "AbstractFormalSubpDecl"
    | #AbstractSubpDeclType.t ->
        "AbstractSubpDecl"
    | #SynthAnonymousTypeDeclType.t ->
        "SynthAnonymousTypeDecl"
    | #AnonymousTypeDeclType.t ->
        "AnonymousTypeDecl"
    | #TypeDeclType.t ->
        "TypeDecl"
    | #SingleTaskTypeDeclType.t ->
        "SingleTaskTypeDecl"
    | #TaskTypeDeclType.t ->
        "TaskTypeDecl"
    | #ProtectedTypeDeclType.t ->
        "ProtectedTypeDecl"
    | #IncompleteTaggedTypeDeclType.t ->
        "IncompleteTaggedTypeDecl"
    | #IncompleteTypeDeclType.t ->
        "IncompleteTypeDecl"
    | #ClasswideTypeDeclType.t ->
        "ClasswideTypeDecl"
    | #SubtypeDeclType.t ->
        "SubtypeDecl"
    | #DiscreteBaseSubtypeDeclType.t ->
        "DiscreteBaseSubtypeDecl"
    | #PackageDeclType.t ->
        "PackageDecl"
    | #GenericPackageInternalType.t ->
        "GenericPackageInternal"
    | #ParamSpecType.t ->
        "ParamSpec"
    | #GenericFormalTypeDeclType.t ->
        "GenericFormalTypeDecl"
    | #GenericFormalSubpDeclType.t ->
        "GenericFormalSubpDecl"
    | #GenericFormalPackageType.t ->
        "GenericFormalPackage"
    | #GenericFormalObjDeclType.t ->
        "GenericFormalObjDecl"
    | #DiscriminantSpecType.t ->
        "DiscriminantSpec"
    | #ComponentDeclType.t ->
        "ComponentDecl"
    | #AnonymousExprDeclType.t ->
        "AnonymousExprDecl"
    | #AbstractStateDeclType.t ->
        "AbstractStateDecl"
    | #ParamAssocType.t ->
        "ParamAssoc"
    | #IteratedAssocType.t ->
        "IteratedAssoc"
    | #DiscriminantAssocType.t ->
        "DiscriminantAssoc"
    | #MultiDimArrayAssocType.t ->
        "MultiDimArrayAssoc"
    | #AggregateAssocType.t ->
        "AggregateAssoc"
    | #RecordDefType.t ->
        "RecordDef"
    | #NullRecordDefType.t ->
        "NullRecordDef"
    | #GenericFormalPartType.t ->
        "GenericFormalPart"
    | #EntryCompletionFormalParamsType.t ->
        "EntryCompletionFormalParams"
    | #UnknownDiscriminantPartType.t ->
        "UnknownDiscriminantPart"
    | #KnownDiscriminantPartType.t ->
        "KnownDiscriminantPart"
    | #ComponentListType.t ->
        "ComponentList"
    | #SubpSpecType.t ->
        "SubpSpec"
    | #EnumSubpSpecType.t ->
        "EnumSubpSpec"
    | #EntrySpecType.t ->
        "EntrySpec"
    | #PragmaArgumentAssocType.t ->
        "PragmaArgumentAssoc"
    | #ContractCaseAssocType.t ->
        "ContractCaseAssoc"
    | #AspectSpecType.t ->
        "AspectSpec"
    | #RecordRepClauseType.t ->
        "RecordRepClause"
    | #EnumRepClauseType.t ->
        "EnumRepClause"
    | #AttributeDefClauseType.t ->
        "AttributeDefClause"
    | #AtClauseType.t ->
        "AtClause"
    | #AspectAssocType.t ->
        "AspectAssoc"
    | #UnconstrainedArrayIndicesType.t ->
        "UnconstrainedArrayIndices"
    | #ConstrainedArrayIndicesType.t ->
        "ConstrainedArrayIndices"
    | #AllPresentType.t ->
        "AllPresent"
    | #AllAbsentType.t ->
        "AllAbsent"
    | #AliasedPresentType.t ->
        "AliasedPresent"
    | #AliasedAbsentType.t ->
        "AliasedAbsent"
    | #VariantListType.t ->
        "VariantList"
    | #UnconstrainedArrayIndexListType.t ->
        "UnconstrainedArrayIndexList"
    | #SelectWhenPartListType.t ->
        "SelectWhenPartList"
    | #PragmaNodeListType.t ->
        "PragmaNodeList"
    | #ParamSpecListType.t ->
        "ParamSpecList"
    | #ParentListType.t ->
        "ParentList"
    | #NameListType.t ->
        "NameList"
    | #DiscriminantChoiceListType.t ->
        "DiscriminantChoiceList"
    | #ExprAlternativesListType.t ->
        "ExprAlternativesList"
    | #EnumLiteralDeclListType.t ->
        "EnumLiteralDeclList"
    | #ElsifStmtPartListType.t ->
        "ElsifStmtPartList"
    | #ElsifExprPartListType.t ->
        "ElsifExprPartList"
    | #DiscriminantSpecListType.t ->
        "DiscriminantSpecList"
    | #DefiningNameListType.t ->
        "DefiningNameList"
    | #ContractCaseAssocListType.t ->
        "ContractCaseAssocList"
    | #CompilationUnitListType.t ->
        "CompilationUnitList"
    | #CaseStmtAlternativeListType.t ->
        "CaseStmtAlternativeList"
    | #CaseExprAlternativeListType.t ->
        "CaseExprAlternativeList"
    | #BasicDeclListType.t ->
        "BasicDeclList"
    | #AssocListType.t ->
        "AssocList"
    | #BaseAssocListType.t ->
        "BaseAssocList"
    | #AspectAssocListType.t ->
        "AspectAssocList"
    | #StmtListType.t ->
        "StmtList"
    | #DeclListType.t ->
        "DeclList"
    | #ConstraintListType.t ->
        "ConstraintList"
    | #AlternativesListType.t ->
        "AlternativesList"
    | #AbstractStateDeclListType.t ->
        "AbstractStateDeclList"
    | #AdaNodeListType.t ->
        "AdaNodeList"
    | #AbstractPresentType.t ->
        "AbstractPresent"
    | #AbstractAbsentType.t ->
        "AbstractAbsent"
    | #AbortPresentType.t ->
        "AbortPresent"
    | #AbortAbsentType.t ->
        "AbortAbsent"

  let text node =
    Token.text_range (token_start node) (token_end node)

  let image node =
    let c_result_ptr = allocate_n Text.c_type ~count:1 in
    CFunctions.image
      (addr (AdaNodeType.unwrap (node)))
      c_result_ptr;
    !@ c_result_ptr

  let entity_image node =
    let c_result_ptr = allocate_n Text.c_type ~count:1 in
    let node_c_value = AdaNodeType.unwrap (node) in
    CFunctions.entity_image (addr node_c_value) c_result_ptr;
    !@ c_result_ptr

  let is_token_node node =
    let node_c_value = AdaNodeType.unwrap (node) in
    CFunctions.node_is_token_node (addr node_c_value)

  let sloc_range node =
    let c_result_ptr = allocate_n SlocRange.c_type ~count:1 in
    CFunctions.node_sloc_range
      (addr (AdaNodeType.unwrap (node)))
      c_result_ptr;
    !@ c_result_ptr

  
  let fold_tokens f init node =
    let tok_start = token_start node in
    let tok_end = token_end node in
    let rec aux acc tok_curr =
      let new_acc = f acc tok_curr in
      if Token.equal tok_curr tok_end then
        new_acc
      else
        aux new_acc (Token.next tok_curr)
    in
    aux init tok_start

  let iter_tokens f node =
    let tok_start = token_start node in
    let tok_end = token_end node in
    let rec aux tok_curr =
      f tok_curr;
      if not (Token.equal tok_curr tok_end) then
        aux (Token.next tok_curr)
    in
    aux tok_start

  let map_tokens f node =
    let tok_start = token_start node in
    let tok_end = token_end node in
    let rec aux tok_curr =
      let value = f tok_curr in
      if Token.equal tok_curr tok_end then
        [value]
      else
        value :: aux (Token.next tok_curr)
    in
    aux tok_start

  let tokens node =
    map_tokens (fun x -> x) node


  let lookup node sloc =
    let node_c_value = AdaNodeType.unwrap (node) in
    let sloc_ptr = allocate Sloc.c_type sloc in
    let result_ptr = allocate_n EntityStruct.c_type ~count:1 in
    CFunctions.lookup_in_node
      (addr node_c_value) sloc_ptr result_ptr;
    if is_null (getf !@ result_ptr EntityStruct.node) then None else Some (AdaNodeType.wrap (context node) (!@ result_ptr))

  let children_opt node =
    let node_c_value = AdaNodeType.unwrap (node) in
    let context = context node in
    let c_value_ptr = allocate_n AdaNodeArrayStruct.c_type ~count:1 in
    let _ : int =
      CFunctions.ada_node_children
        (addr node_c_value)
        (c_value_ptr)
    in
    let c_value = !@(!@(c_value_ptr)) in
    let length = getf c_value AdaNodeArrayStruct.n in
    let items = c_value @. AdaNodeArrayStruct.items in
    let f i =
      let fresh = allocate EntityStruct.c_type !@(items +@ i) in
      if is_null (getf !@ fresh EntityStruct.node) then None else Some (AdaNodeType.wrap context (!@ fresh))
    in
    let result = List.init length f in
    AdaNodeArrayStruct.dec_ref (!@ c_value_ptr);
    result

  let iter_fields f node =
    children_opt (node :> AdaNodeType.t)
    |> List.iter (function None -> () | Some node -> f node)

  let fold_fields f acc node =
    children_opt (node :> AdaNodeType.t)
    |> List.fold_left (fun x -> function None -> x | Some node -> f x node) acc

  let exists_fields p node =
    children_opt (node :> AdaNodeType.t)
    |> List.exists (function | None -> false | Some node -> p node)

  let for_all_fields p node =
    children_opt (node :> AdaNodeType.t)
    |> List.for_all (function | None -> true | Some node -> p node)

  let fold f acc node =
    (* Use an auxiliary function here to have a better type for the function *)
    let rec aux acc node = fold_fields aux (f acc node) node in
    aux acc (node :> AdaNodeType.t)

  let iter f node =
    (* Use an auxiliary function here to have a better type for the function *)
    let rec aux node = f node; iter_fields aux node in
    aux (node :> AdaNodeType.t)

  let filter p node =
    fold (fun acc node -> if p node then node :: acc else acc) [] node
    |> List.rev

  let exists p node =
    (* Use an auxiliary function here to have a better type for the function *)
    let rec aux node =
      p node || exists_fields aux node in aux (node :> AdaNodeType.t)

  let for_all p node =
    (* Use an auxiliary function here to have a better type for the function *)
    let rec aux node = p node && for_all_fields aux node in
    aux (node :> AdaNodeType.t)

  let lookup_with_kind :
    type a. a node -> [< AdaNodeType.t] -> Sloc.t -> a option =
    fun node_type node sloc ->
      let lookup_res = lookup node sloc in
      let rec aux : a node -> [< AdaNodeType.t] -> a option =
        fun node_type node ->
        match node_type, node with
        | AdaNode
          , (#AdaNodeType.t as node) ->
          Some node
        | AbortNode
          , (#AbortNodeType.t as node) ->
          Some node
        | AbortAbsent
          , (#AbortAbsentType.t as node) ->
          Some node
        | AbortPresent
          , (#AbortPresentType.t as node) ->
          Some node
        | AbstractNode
          , (#AbstractNodeType.t as node) ->
          Some node
        | AbstractAbsent
          , (#AbstractAbsentType.t as node) ->
          Some node
        | AbstractPresent
          , (#AbstractPresentType.t as node) ->
          Some node
        | AdaList
          , (#AdaListType.t as node) ->
          Some node
        | AdaNodeList
          , (#AdaNodeListType.t as node) ->
          Some node
        | AbstractStateDeclList
          , (#AbstractStateDeclListType.t as node) ->
          Some node
        | AlternativesList
          , (#AlternativesListType.t as node) ->
          Some node
        | ConstraintList
          , (#ConstraintListType.t as node) ->
          Some node
        | DeclList
          , (#DeclListType.t as node) ->
          Some node
        | StmtList
          , (#StmtListType.t as node) ->
          Some node
        | AspectAssocList
          , (#AspectAssocListType.t as node) ->
          Some node
        | BaseAssocList
          , (#BaseAssocListType.t as node) ->
          Some node
        | BasicAssocList
          , (#BasicAssocListType.t as node) ->
          Some node
        | AssocList
          , (#AssocListType.t as node) ->
          Some node
        | BasicDeclList
          , (#BasicDeclListType.t as node) ->
          Some node
        | CaseExprAlternativeList
          , (#CaseExprAlternativeListType.t as node) ->
          Some node
        | CaseStmtAlternativeList
          , (#CaseStmtAlternativeListType.t as node) ->
          Some node
        | CompilationUnitList
          , (#CompilationUnitListType.t as node) ->
          Some node
        | ContractCaseAssocList
          , (#ContractCaseAssocListType.t as node) ->
          Some node
        | DefiningNameList
          , (#DefiningNameListType.t as node) ->
          Some node
        | DiscriminantSpecList
          , (#DiscriminantSpecListType.t as node) ->
          Some node
        | ElsifExprPartList
          , (#ElsifExprPartListType.t as node) ->
          Some node
        | ElsifStmtPartList
          , (#ElsifStmtPartListType.t as node) ->
          Some node
        | EnumLiteralDeclList
          , (#EnumLiteralDeclListType.t as node) ->
          Some node
        | ExprList
          , (#ExprListType.t as node) ->
          Some node
        | ExprAlternativesList
          , (#ExprAlternativesListType.t as node) ->
          Some node
        | IdentifierList
          , (#IdentifierListType.t as node) ->
          Some node
        | DiscriminantChoiceList
          , (#DiscriminantChoiceListType.t as node) ->
          Some node
        | NameList
          , (#NameListType.t as node) ->
          Some node
        | ParentList
          , (#ParentListType.t as node) ->
          Some node
        | ParamSpecList
          , (#ParamSpecListType.t as node) ->
          Some node
        | PragmaNodeList
          , (#PragmaNodeListType.t as node) ->
          Some node
        | SelectWhenPartList
          , (#SelectWhenPartListType.t as node) ->
          Some node
        | UnconstrainedArrayIndexList
          , (#UnconstrainedArrayIndexListType.t as node) ->
          Some node
        | VariantList
          , (#VariantListType.t as node) ->
          Some node
        | AliasedNode
          , (#AliasedNodeType.t as node) ->
          Some node
        | AliasedAbsent
          , (#AliasedAbsentType.t as node) ->
          Some node
        | AliasedPresent
          , (#AliasedPresentType.t as node) ->
          Some node
        | AllNode
          , (#AllNodeType.t as node) ->
          Some node
        | AllAbsent
          , (#AllAbsentType.t as node) ->
          Some node
        | AllPresent
          , (#AllPresentType.t as node) ->
          Some node
        | ArrayIndices
          , (#ArrayIndicesType.t as node) ->
          Some node
        | ConstrainedArrayIndices
          , (#ConstrainedArrayIndicesType.t as node) ->
          Some node
        | UnconstrainedArrayIndices
          , (#UnconstrainedArrayIndicesType.t as node) ->
          Some node
        | AspectAssoc
          , (#AspectAssocType.t as node) ->
          Some node
        | AspectClause
          , (#AspectClauseType.t as node) ->
          Some node
        | AtClause
          , (#AtClauseType.t as node) ->
          Some node
        | AttributeDefClause
          , (#AttributeDefClauseType.t as node) ->
          Some node
        | EnumRepClause
          , (#EnumRepClauseType.t as node) ->
          Some node
        | RecordRepClause
          , (#RecordRepClauseType.t as node) ->
          Some node
        | AspectSpec
          , (#AspectSpecType.t as node) ->
          Some node
        | BaseAssoc
          , (#BaseAssocType.t as node) ->
          Some node
        | ContractCaseAssoc
          , (#ContractCaseAssocType.t as node) ->
          Some node
        | PragmaArgumentAssoc
          , (#PragmaArgumentAssocType.t as node) ->
          Some node
        | BaseFormalParamHolder
          , (#BaseFormalParamHolderType.t as node) ->
          Some node
        | BaseSubpSpec
          , (#BaseSubpSpecType.t as node) ->
          Some node
        | EntrySpec
          , (#EntrySpecType.t as node) ->
          Some node
        | EnumSubpSpec
          , (#EnumSubpSpecType.t as node) ->
          Some node
        | SubpSpec
          , (#SubpSpecType.t as node) ->
          Some node
        | ComponentList
          , (#ComponentListType.t as node) ->
          Some node
        | DiscriminantPart
          , (#DiscriminantPartType.t as node) ->
          Some node
        | KnownDiscriminantPart
          , (#KnownDiscriminantPartType.t as node) ->
          Some node
        | UnknownDiscriminantPart
          , (#UnknownDiscriminantPartType.t as node) ->
          Some node
        | EntryCompletionFormalParams
          , (#EntryCompletionFormalParamsType.t as node) ->
          Some node
        | GenericFormalPart
          , (#GenericFormalPartType.t as node) ->
          Some node
        | BaseRecordDef
          , (#BaseRecordDefType.t as node) ->
          Some node
        | NullRecordDef
          , (#NullRecordDefType.t as node) ->
          Some node
        | RecordDef
          , (#RecordDefType.t as node) ->
          Some node
        | BasicAssoc
          , (#BasicAssocType.t as node) ->
          Some node
        | AggregateAssoc
          , (#AggregateAssocType.t as node) ->
          Some node
        | MultiDimArrayAssoc
          , (#MultiDimArrayAssocType.t as node) ->
          Some node
        | DiscriminantAssoc
          , (#DiscriminantAssocType.t as node) ->
          Some node
        | IteratedAssoc
          , (#IteratedAssocType.t as node) ->
          Some node
        | ParamAssoc
          , (#ParamAssocType.t as node) ->
          Some node
        | BasicDecl
          , (#BasicDeclType.t as node) ->
          Some node
        | AbstractStateDecl
          , (#AbstractStateDeclType.t as node) ->
          Some node
        | AnonymousExprDecl
          , (#AnonymousExprDeclType.t as node) ->
          Some node
        | BaseFormalParamDecl
          , (#BaseFormalParamDeclType.t as node) ->
          Some node
        | ComponentDecl
          , (#ComponentDeclType.t as node) ->
          Some node
        | DiscriminantSpec
          , (#DiscriminantSpecType.t as node) ->
          Some node
        | GenericFormal
          , (#GenericFormalType.t as node) ->
          Some node
        | GenericFormalObjDecl
          , (#GenericFormalObjDeclType.t as node) ->
          Some node
        | GenericFormalPackage
          , (#GenericFormalPackageType.t as node) ->
          Some node
        | GenericFormalSubpDecl
          , (#GenericFormalSubpDeclType.t as node) ->
          Some node
        | GenericFormalTypeDecl
          , (#GenericFormalTypeDeclType.t as node) ->
          Some node
        | ParamSpec
          , (#ParamSpecType.t as node) ->
          Some node
        | BasePackageDecl
          , (#BasePackageDeclType.t as node) ->
          Some node
        | GenericPackageInternal
          , (#GenericPackageInternalType.t as node) ->
          Some node
        | PackageDecl
          , (#PackageDeclType.t as node) ->
          Some node
        | BaseTypeDecl
          , (#BaseTypeDeclType.t as node) ->
          Some node
        | BaseSubtypeDecl
          , (#BaseSubtypeDeclType.t as node) ->
          Some node
        | DiscreteBaseSubtypeDecl
          , (#DiscreteBaseSubtypeDeclType.t as node) ->
          Some node
        | SubtypeDecl
          , (#SubtypeDeclType.t as node) ->
          Some node
        | ClasswideTypeDecl
          , (#ClasswideTypeDeclType.t as node) ->
          Some node
        | IncompleteTypeDecl
          , (#IncompleteTypeDeclType.t as node) ->
          Some node
        | IncompleteTaggedTypeDecl
          , (#IncompleteTaggedTypeDeclType.t as node) ->
          Some node
        | ProtectedTypeDecl
          , (#ProtectedTypeDeclType.t as node) ->
          Some node
        | TaskTypeDecl
          , (#TaskTypeDeclType.t as node) ->
          Some node
        | SingleTaskTypeDecl
          , (#SingleTaskTypeDeclType.t as node) ->
          Some node
        | TypeDecl
          , (#TypeDeclType.t as node) ->
          Some node
        | AnonymousTypeDecl
          , (#AnonymousTypeDeclType.t as node) ->
          Some node
        | SynthAnonymousTypeDecl
          , (#SynthAnonymousTypeDeclType.t as node) ->
          Some node
        | BasicSubpDecl
          , (#BasicSubpDeclType.t as node) ->
          Some node
        | ClassicSubpDecl
          , (#ClassicSubpDeclType.t as node) ->
          Some node
        | AbstractSubpDecl
          , (#AbstractSubpDeclType.t as node) ->
          Some node
        | FormalSubpDecl
          , (#FormalSubpDeclType.t as node) ->
          Some node
        | AbstractFormalSubpDecl
          , (#AbstractFormalSubpDeclType.t as node) ->
          Some node
        | ConcreteFormalSubpDecl
          , (#ConcreteFormalSubpDeclType.t as node) ->
          Some node
        | SubpDecl
          , (#SubpDeclType.t as node) ->
          Some node
        | EntryDecl
          , (#EntryDeclType.t as node) ->
          Some node
        | EnumLiteralDecl
          , (#EnumLiteralDeclType.t as node) ->
          Some node
        | GenericSubpInternal
          , (#GenericSubpInternalType.t as node) ->
          Some node
        | BodyNode
          , (#BodyNodeType.t as node) ->
          Some node
        | BaseSubpBody
          , (#BaseSubpBodyType.t as node) ->
          Some node
        | ExprFunction
          , (#ExprFunctionType.t as node) ->
          Some node
        | NullSubpDecl
          , (#NullSubpDeclType.t as node) ->
          Some node
        | SubpBody
          , (#SubpBodyType.t as node) ->
          Some node
        | SubpRenamingDecl
          , (#SubpRenamingDeclType.t as node) ->
          Some node
        | BodyStub
          , (#BodyStubType.t as node) ->
          Some node
        | PackageBodyStub
          , (#PackageBodyStubType.t as node) ->
          Some node
        | ProtectedBodyStub
          , (#ProtectedBodyStubType.t as node) ->
          Some node
        | SubpBodyStub
          , (#SubpBodyStubType.t as node) ->
          Some node
        | TaskBodyStub
          , (#TaskBodyStubType.t as node) ->
          Some node
        | EntryBody
          , (#EntryBodyType.t as node) ->
          Some node
        | PackageBody
          , (#PackageBodyType.t as node) ->
          Some node
        | ProtectedBody
          , (#ProtectedBodyType.t as node) ->
          Some node
        | TaskBody
          , (#TaskBodyType.t as node) ->
          Some node
        | EntryIndexSpec
          , (#EntryIndexSpecType.t as node) ->
          Some node
        | ErrorDecl
          , (#ErrorDeclType.t as node) ->
          Some node
        | ExceptionDecl
          , (#ExceptionDeclType.t as node) ->
          Some node
        | ExceptionHandler
          , (#ExceptionHandlerType.t as node) ->
          Some node
        | ForLoopVarDecl
          , (#ForLoopVarDeclType.t as node) ->
          Some node
        | GenericDecl
          , (#GenericDeclType.t as node) ->
          Some node
        | GenericPackageDecl
          , (#GenericPackageDeclType.t as node) ->
          Some node
        | GenericSubpDecl
          , (#GenericSubpDeclType.t as node) ->
          Some node
        | GenericInstantiation
          , (#GenericInstantiationType.t as node) ->
          Some node
        | GenericPackageInstantiation
          , (#GenericPackageInstantiationType.t as node) ->
          Some node
        | GenericSubpInstantiation
          , (#GenericSubpInstantiationType.t as node) ->
          Some node
        | GenericRenamingDecl
          , (#GenericRenamingDeclType.t as node) ->
          Some node
        | GenericPackageRenamingDecl
          , (#GenericPackageRenamingDeclType.t as node) ->
          Some node
        | GenericSubpRenamingDecl
          , (#GenericSubpRenamingDeclType.t as node) ->
          Some node
        | LabelDecl
          , (#LabelDeclType.t as node) ->
          Some node
        | NamedStmtDecl
          , (#NamedStmtDeclType.t as node) ->
          Some node
        | NumberDecl
          , (#NumberDeclType.t as node) ->
          Some node
        | ObjectDecl
          , (#ObjectDeclType.t as node) ->
          Some node
        | ExtendedReturnStmtObjectDecl
          , (#ExtendedReturnStmtObjectDeclType.t as node) ->
          Some node
        | PackageRenamingDecl
          , (#PackageRenamingDeclType.t as node) ->
          Some node
        | SingleProtectedDecl
          , (#SingleProtectedDeclType.t as node) ->
          Some node
        | SingleTaskDecl
          , (#SingleTaskDeclType.t as node) ->
          Some node
        | CaseStmtAlternative
          , (#CaseStmtAlternativeType.t as node) ->
          Some node
        | CompilationUnit
          , (#CompilationUnitType.t as node) ->
          Some node
        | ComponentClause
          , (#ComponentClauseType.t as node) ->
          Some node
        | ComponentDef
          , (#ComponentDefType.t as node) ->
          Some node
        | ConstantNode
          , (#ConstantNodeType.t as node) ->
          Some node
        | ConstantAbsent
          , (#ConstantAbsentType.t as node) ->
          Some node
        | ConstantPresent
          , (#ConstantPresentType.t as node) ->
          Some node
        | Constraint
          , (#ConstraintType.t as node) ->
          Some node
        | DeltaConstraint
          , (#DeltaConstraintType.t as node) ->
          Some node
        | DigitsConstraint
          , (#DigitsConstraintType.t as node) ->
          Some node
        | DiscriminantConstraint
          , (#DiscriminantConstraintType.t as node) ->
          Some node
        | IndexConstraint
          , (#IndexConstraintType.t as node) ->
          Some node
        | RangeConstraint
          , (#RangeConstraintType.t as node) ->
          Some node
        | DeclarativePart
          , (#DeclarativePartType.t as node) ->
          Some node
        | PrivatePart
          , (#PrivatePartType.t as node) ->
          Some node
        | PublicPart
          , (#PublicPartType.t as node) ->
          Some node
        | ElsifExprPart
          , (#ElsifExprPartType.t as node) ->
          Some node
        | ElsifStmtPart
          , (#ElsifStmtPartType.t as node) ->
          Some node
        | Expr
          , (#ExprType.t as node) ->
          Some node
        | AbstractStateDeclExpr
          , (#AbstractStateDeclExprType.t as node) ->
          Some node
        | Allocator
          , (#AllocatorType.t as node) ->
          Some node
        | BaseAggregate
          , (#BaseAggregateType.t as node) ->
          Some node
        | Aggregate
          , (#AggregateType.t as node) ->
          Some node
        | BracketAggregate
          , (#BracketAggregateType.t as node) ->
          Some node
        | DeltaAggregate
          , (#DeltaAggregateType.t as node) ->
          Some node
        | BracketDeltaAggregate
          , (#BracketDeltaAggregateType.t as node) ->
          Some node
        | NullRecordAggregate
          , (#NullRecordAggregateType.t as node) ->
          Some node
        | BinOp
          , (#BinOpType.t as node) ->
          Some node
        | RelationOp
          , (#RelationOpType.t as node) ->
          Some node
        | BoxExpr
          , (#BoxExprType.t as node) ->
          Some node
        | CaseExprAlternative
          , (#CaseExprAlternativeType.t as node) ->
          Some node
        | CondExpr
          , (#CondExprType.t as node) ->
          Some node
        | CaseExpr
          , (#CaseExprType.t as node) ->
          Some node
        | IfExpr
          , (#IfExprType.t as node) ->
          Some node
        | ContractCases
          , (#ContractCasesType.t as node) ->
          Some node
        | DeclExpr
          , (#DeclExprType.t as node) ->
          Some node
        | MembershipExpr
          , (#MembershipExprType.t as node) ->
          Some node
        | Name
          , (#NameType.t as node) ->
          Some node
        | AttributeRef
          , (#AttributeRefType.t as node) ->
          Some node
        | UpdateAttributeRef
          , (#UpdateAttributeRefType.t as node) ->
          Some node
        | CallExpr
          , (#CallExprType.t as node) ->
          Some node
        | DefiningName
          , (#DefiningNameType.t as node) ->
          Some node
        | DiscreteSubtypeName
          , (#DiscreteSubtypeNameType.t as node) ->
          Some node
        | DottedName
          , (#DottedNameType.t as node) ->
          Some node
        | EndName
          , (#EndNameType.t as node) ->
          Some node
        | ExplicitDeref
          , (#ExplicitDerefType.t as node) ->
          Some node
        | QualExpr
          , (#QualExprType.t as node) ->
          Some node
        | SingleTokNode
          , (#SingleTokNodeType.t as node) ->
          Some node
        | BaseId
          , (#BaseIdType.t as node) ->
          Some node
        | CharLiteral
          , (#CharLiteralType.t as node) ->
          Some node
        | Identifier
          , (#IdentifierType.t as node) ->
          Some node
        | Op
          , (#OpType.t as node) ->
          Some node
        | OpAbs
          , (#OpAbsType.t as node) ->
          Some node
        | OpAnd
          , (#OpAndType.t as node) ->
          Some node
        | OpAndThen
          , (#OpAndThenType.t as node) ->
          Some node
        | OpConcat
          , (#OpConcatType.t as node) ->
          Some node
        | OpDiv
          , (#OpDivType.t as node) ->
          Some node
        | OpDoubleDot
          , (#OpDoubleDotType.t as node) ->
          Some node
        | OpEq
          , (#OpEqType.t as node) ->
          Some node
        | OpGt
          , (#OpGtType.t as node) ->
          Some node
        | OpGte
          , (#OpGteType.t as node) ->
          Some node
        | OpIn
          , (#OpInType.t as node) ->
          Some node
        | OpLt
          , (#OpLtType.t as node) ->
          Some node
        | OpLte
          , (#OpLteType.t as node) ->
          Some node
        | OpMinus
          , (#OpMinusType.t as node) ->
          Some node
        | OpMod
          , (#OpModType.t as node) ->
          Some node
        | OpMult
          , (#OpMultType.t as node) ->
          Some node
        | OpNeq
          , (#OpNeqType.t as node) ->
          Some node
        | OpNot
          , (#OpNotType.t as node) ->
          Some node
        | OpNotIn
          , (#OpNotInType.t as node) ->
          Some node
        | OpOr
          , (#OpOrType.t as node) ->
          Some node
        | OpOrElse
          , (#OpOrElseType.t as node) ->
          Some node
        | OpPlus
          , (#OpPlusType.t as node) ->
          Some node
        | OpPow
          , (#OpPowType.t as node) ->
          Some node
        | OpRem
          , (#OpRemType.t as node) ->
          Some node
        | OpXor
          , (#OpXorType.t as node) ->
          Some node
        | StringLiteral
          , (#StringLiteralType.t as node) ->
          Some node
        | NullLiteral
          , (#NullLiteralType.t as node) ->
          Some node
        | NumLiteral
          , (#NumLiteralType.t as node) ->
          Some node
        | IntLiteral
          , (#IntLiteralType.t as node) ->
          Some node
        | RealLiteral
          , (#RealLiteralType.t as node) ->
          Some node
        | TargetName
          , (#TargetNameType.t as node) ->
          Some node
        | ParenExpr
          , (#ParenExprType.t as node) ->
          Some node
        | QuantifiedExpr
          , (#QuantifiedExprType.t as node) ->
          Some node
        | RaiseExpr
          , (#RaiseExprType.t as node) ->
          Some node
        | UnOp
          , (#UnOpType.t as node) ->
          Some node
        | HandledStmts
          , (#HandledStmtsType.t as node) ->
          Some node
        | InterfaceKind
          , (#InterfaceKindType.t as node) ->
          Some node
        | InterfaceKindLimited
          , (#InterfaceKindLimitedType.t as node) ->
          Some node
        | InterfaceKindProtected
          , (#InterfaceKindProtectedType.t as node) ->
          Some node
        | InterfaceKindSynchronized
          , (#InterfaceKindSynchronizedType.t as node) ->
          Some node
        | InterfaceKindTask
          , (#InterfaceKindTaskType.t as node) ->
          Some node
        | IterType
          , (#IterTypeType.t as node) ->
          Some node
        | IterTypeIn
          , (#IterTypeInType.t as node) ->
          Some node
        | IterTypeOf
          , (#IterTypeOfType.t as node) ->
          Some node
        | LibraryItem
          , (#LibraryItemType.t as node) ->
          Some node
        | LimitedNode
          , (#LimitedNodeType.t as node) ->
          Some node
        | LimitedAbsent
          , (#LimitedAbsentType.t as node) ->
          Some node
        | LimitedPresent
          , (#LimitedPresentType.t as node) ->
          Some node
        | LoopSpec
          , (#LoopSpecType.t as node) ->
          Some node
        | ForLoopSpec
          , (#ForLoopSpecType.t as node) ->
          Some node
        | WhileLoopSpec
          , (#WhileLoopSpecType.t as node) ->
          Some node
        | Mode
          , (#ModeType.t as node) ->
          Some node
        | ModeDefault
          , (#ModeDefaultType.t as node) ->
          Some node
        | ModeIn
          , (#ModeInType.t as node) ->
          Some node
        | ModeInOut
          , (#ModeInOutType.t as node) ->
          Some node
        | ModeOut
          , (#ModeOutType.t as node) ->
          Some node
        | MultiAbstractStateDecl
          , (#MultiAbstractStateDeclType.t as node) ->
          Some node
        | NotNull
          , (#NotNullType.t as node) ->
          Some node
        | NotNullAbsent
          , (#NotNullAbsentType.t as node) ->
          Some node
        | NotNullPresent
          , (#NotNullPresentType.t as node) ->
          Some node
        | NullComponentDecl
          , (#NullComponentDeclType.t as node) ->
          Some node
        | OthersDesignator
          , (#OthersDesignatorType.t as node) ->
          Some node
        | OverridingNode
          , (#OverridingNodeType.t as node) ->
          Some node
        | OverridingNotOverriding
          , (#OverridingNotOverridingType.t as node) ->
          Some node
        | OverridingOverriding
          , (#OverridingOverridingType.t as node) ->
          Some node
        | OverridingUnspecified
          , (#OverridingUnspecifiedType.t as node) ->
          Some node
        | Params
          , (#ParamsType.t as node) ->
          Some node
        | ParenAbstractStateDecl
          , (#ParenAbstractStateDeclType.t as node) ->
          Some node
        | PragmaNode
          , (#PragmaNodeType.t as node) ->
          Some node
        | PrivateNode
          , (#PrivateNodeType.t as node) ->
          Some node
        | PrivateAbsent
          , (#PrivateAbsentType.t as node) ->
          Some node
        | PrivatePresent
          , (#PrivatePresentType.t as node) ->
          Some node
        | ProtectedDef
          , (#ProtectedDefType.t as node) ->
          Some node
        | ProtectedNode
          , (#ProtectedNodeType.t as node) ->
          Some node
        | ProtectedAbsent
          , (#ProtectedAbsentType.t as node) ->
          Some node
        | ProtectedPresent
          , (#ProtectedPresentType.t as node) ->
          Some node
        | Quantifier
          , (#QuantifierType.t as node) ->
          Some node
        | QuantifierAll
          , (#QuantifierAllType.t as node) ->
          Some node
        | QuantifierSome
          , (#QuantifierSomeType.t as node) ->
          Some node
        | RangeSpec
          , (#RangeSpecType.t as node) ->
          Some node
        | RenamingClause
          , (#RenamingClauseType.t as node) ->
          Some node
        | SyntheticRenamingClause
          , (#SyntheticRenamingClauseType.t as node) ->
          Some node
        | ReverseNode
          , (#ReverseNodeType.t as node) ->
          Some node
        | ReverseAbsent
          , (#ReverseAbsentType.t as node) ->
          Some node
        | ReversePresent
          , (#ReversePresentType.t as node) ->
          Some node
        | SelectWhenPart
          , (#SelectWhenPartType.t as node) ->
          Some node
        | Stmt
          , (#StmtType.t as node) ->
          Some node
        | CompositeStmt
          , (#CompositeStmtType.t as node) ->
          Some node
        | AcceptStmt
          , (#AcceptStmtType.t as node) ->
          Some node
        | AcceptStmtWithStmts
          , (#AcceptStmtWithStmtsType.t as node) ->
          Some node
        | BaseLoopStmt
          , (#BaseLoopStmtType.t as node) ->
          Some node
        | ForLoopStmt
          , (#ForLoopStmtType.t as node) ->
          Some node
        | LoopStmt
          , (#LoopStmtType.t as node) ->
          Some node
        | WhileLoopStmt
          , (#WhileLoopStmtType.t as node) ->
          Some node
        | BlockStmt
          , (#BlockStmtType.t as node) ->
          Some node
        | BeginBlock
          , (#BeginBlockType.t as node) ->
          Some node
        | DeclBlock
          , (#DeclBlockType.t as node) ->
          Some node
        | CaseStmt
          , (#CaseStmtType.t as node) ->
          Some node
        | ExtendedReturnStmt
          , (#ExtendedReturnStmtType.t as node) ->
          Some node
        | IfStmt
          , (#IfStmtType.t as node) ->
          Some node
        | NamedStmt
          , (#NamedStmtType.t as node) ->
          Some node
        | SelectStmt
          , (#SelectStmtType.t as node) ->
          Some node
        | ErrorStmt
          , (#ErrorStmtType.t as node) ->
          Some node
        | SimpleStmt
          , (#SimpleStmtType.t as node) ->
          Some node
        | AbortStmt
          , (#AbortStmtType.t as node) ->
          Some node
        | AssignStmt
          , (#AssignStmtType.t as node) ->
          Some node
        | CallStmt
          , (#CallStmtType.t as node) ->
          Some node
        | DelayStmt
          , (#DelayStmtType.t as node) ->
          Some node
        | ExitStmt
          , (#ExitStmtType.t as node) ->
          Some node
        | GotoStmt
          , (#GotoStmtType.t as node) ->
          Some node
        | Label
          , (#LabelType.t as node) ->
          Some node
        | NullStmt
          , (#NullStmtType.t as node) ->
          Some node
        | RaiseStmt
          , (#RaiseStmtType.t as node) ->
          Some node
        | RequeueStmt
          , (#RequeueStmtType.t as node) ->
          Some node
        | ReturnStmt
          , (#ReturnStmtType.t as node) ->
          Some node
        | TerminateAlternative
          , (#TerminateAlternativeType.t as node) ->
          Some node
        | SubpKind
          , (#SubpKindType.t as node) ->
          Some node
        | SubpKindFunction
          , (#SubpKindFunctionType.t as node) ->
          Some node
        | SubpKindProcedure
          , (#SubpKindProcedureType.t as node) ->
          Some node
        | Subunit
          , (#SubunitType.t as node) ->
          Some node
        | SynchronizedNode
          , (#SynchronizedNodeType.t as node) ->
          Some node
        | SynchronizedAbsent
          , (#SynchronizedAbsentType.t as node) ->
          Some node
        | SynchronizedPresent
          , (#SynchronizedPresentType.t as node) ->
          Some node
        | TaggedNode
          , (#TaggedNodeType.t as node) ->
          Some node
        | TaggedAbsent
          , (#TaggedAbsentType.t as node) ->
          Some node
        | TaggedPresent
          , (#TaggedPresentType.t as node) ->
          Some node
        | TaskDef
          , (#TaskDefType.t as node) ->
          Some node
        | TypeDef
          , (#TypeDefType.t as node) ->
          Some node
        | AccessDef
          , (#AccessDefType.t as node) ->
          Some node
        | AccessToSubpDef
          , (#AccessToSubpDefType.t as node) ->
          Some node
        | BaseTypeAccessDef
          , (#BaseTypeAccessDefType.t as node) ->
          Some node
        | AnonymousTypeAccessDef
          , (#AnonymousTypeAccessDefType.t as node) ->
          Some node
        | TypeAccessDef
          , (#TypeAccessDefType.t as node) ->
          Some node
        | ArrayTypeDef
          , (#ArrayTypeDefType.t as node) ->
          Some node
        | DerivedTypeDef
          , (#DerivedTypeDefType.t as node) ->
          Some node
        | EnumTypeDef
          , (#EnumTypeDefType.t as node) ->
          Some node
        | FormalDiscreteTypeDef
          , (#FormalDiscreteTypeDefType.t as node) ->
          Some node
        | InterfaceTypeDef
          , (#InterfaceTypeDefType.t as node) ->
          Some node
        | ModIntTypeDef
          , (#ModIntTypeDefType.t as node) ->
          Some node
        | PrivateTypeDef
          , (#PrivateTypeDefType.t as node) ->
          Some node
        | RealTypeDef
          , (#RealTypeDefType.t as node) ->
          Some node
        | DecimalFixedPointDef
          , (#DecimalFixedPointDefType.t as node) ->
          Some node
        | FloatingPointDef
          , (#FloatingPointDefType.t as node) ->
          Some node
        | OrdinaryFixedPointDef
          , (#OrdinaryFixedPointDefType.t as node) ->
          Some node
        | RecordTypeDef
          , (#RecordTypeDefType.t as node) ->
          Some node
        | SignedIntTypeDef
          , (#SignedIntTypeDefType.t as node) ->
          Some node
        | TypeExpr
          , (#TypeExprType.t as node) ->
          Some node
        | AnonymousType
          , (#AnonymousTypeType.t as node) ->
          Some node
        | EnumLitSynthTypeExpr
          , (#EnumLitSynthTypeExprType.t as node) ->
          Some node
        | SubtypeIndication
          , (#SubtypeIndicationType.t as node) ->
          Some node
        | ConstrainedSubtypeIndication
          , (#ConstrainedSubtypeIndicationType.t as node) ->
          Some node
        | DiscreteSubtypeIndication
          , (#DiscreteSubtypeIndicationType.t as node) ->
          Some node
        | UnconstrainedArrayIndex
          , (#UnconstrainedArrayIndexType.t as node) ->
          Some node
        | UntilNode
          , (#UntilNodeType.t as node) ->
          Some node
        | UntilAbsent
          , (#UntilAbsentType.t as node) ->
          Some node
        | UntilPresent
          , (#UntilPresentType.t as node) ->
          Some node
        | UseClause
          , (#UseClauseType.t as node) ->
          Some node
        | UsePackageClause
          , (#UsePackageClauseType.t as node) ->
          Some node
        | UseTypeClause
          , (#UseTypeClauseType.t as node) ->
          Some node
        | Variant
          , (#VariantType.t as node) ->
          Some node
        | VariantPart
          , (#VariantPartType.t as node) ->
          Some node
        | WithClause
          , (#WithClauseType.t as node) ->
          Some node
        | WithPrivate
          , (#WithPrivateType.t as node) ->
          Some node
        | WithPrivateAbsent
          , (#WithPrivateAbsentType.t as node) ->
          Some node
        | WithPrivatePresent
          , (#WithPrivatePresentType.t as node) ->
          Some node
        | _ -> (match parent node with
                | Some parent_node -> aux node_type parent_node
                | _ -> None) in
    match lookup_res with
      | Some node -> aux node_type node
      | _ -> None

  let as_a : type a. a node -> [< AdaNodeType.t ] -> a option =
   fun node_type node ->
    match node_type, (node :> AdaNodeType.t) with
    | AdaNode
      , (#AdaNodeType.t as node) ->
        Some node
    | AbortNode
      , (#AbortNodeType.t as node) ->
        Some node
    | AbortAbsent
      , (#AbortAbsentType.t as node) ->
        Some node
    | AbortPresent
      , (#AbortPresentType.t as node) ->
        Some node
    | AbstractNode
      , (#AbstractNodeType.t as node) ->
        Some node
    | AbstractAbsent
      , (#AbstractAbsentType.t as node) ->
        Some node
    | AbstractPresent
      , (#AbstractPresentType.t as node) ->
        Some node
    | AdaList
      , (#AdaListType.t as node) ->
        Some node
    | AdaNodeList
      , (#AdaNodeListType.t as node) ->
        Some node
    | AbstractStateDeclList
      , (#AbstractStateDeclListType.t as node) ->
        Some node
    | AlternativesList
      , (#AlternativesListType.t as node) ->
        Some node
    | ConstraintList
      , (#ConstraintListType.t as node) ->
        Some node
    | DeclList
      , (#DeclListType.t as node) ->
        Some node
    | StmtList
      , (#StmtListType.t as node) ->
        Some node
    | AspectAssocList
      , (#AspectAssocListType.t as node) ->
        Some node
    | BaseAssocList
      , (#BaseAssocListType.t as node) ->
        Some node
    | BasicAssocList
      , (#BasicAssocListType.t as node) ->
        Some node
    | AssocList
      , (#AssocListType.t as node) ->
        Some node
    | BasicDeclList
      , (#BasicDeclListType.t as node) ->
        Some node
    | CaseExprAlternativeList
      , (#CaseExprAlternativeListType.t as node) ->
        Some node
    | CaseStmtAlternativeList
      , (#CaseStmtAlternativeListType.t as node) ->
        Some node
    | CompilationUnitList
      , (#CompilationUnitListType.t as node) ->
        Some node
    | ContractCaseAssocList
      , (#ContractCaseAssocListType.t as node) ->
        Some node
    | DefiningNameList
      , (#DefiningNameListType.t as node) ->
        Some node
    | DiscriminantSpecList
      , (#DiscriminantSpecListType.t as node) ->
        Some node
    | ElsifExprPartList
      , (#ElsifExprPartListType.t as node) ->
        Some node
    | ElsifStmtPartList
      , (#ElsifStmtPartListType.t as node) ->
        Some node
    | EnumLiteralDeclList
      , (#EnumLiteralDeclListType.t as node) ->
        Some node
    | ExprList
      , (#ExprListType.t as node) ->
        Some node
    | ExprAlternativesList
      , (#ExprAlternativesListType.t as node) ->
        Some node
    | IdentifierList
      , (#IdentifierListType.t as node) ->
        Some node
    | DiscriminantChoiceList
      , (#DiscriminantChoiceListType.t as node) ->
        Some node
    | NameList
      , (#NameListType.t as node) ->
        Some node
    | ParentList
      , (#ParentListType.t as node) ->
        Some node
    | ParamSpecList
      , (#ParamSpecListType.t as node) ->
        Some node
    | PragmaNodeList
      , (#PragmaNodeListType.t as node) ->
        Some node
    | SelectWhenPartList
      , (#SelectWhenPartListType.t as node) ->
        Some node
    | UnconstrainedArrayIndexList
      , (#UnconstrainedArrayIndexListType.t as node) ->
        Some node
    | VariantList
      , (#VariantListType.t as node) ->
        Some node
    | AliasedNode
      , (#AliasedNodeType.t as node) ->
        Some node
    | AliasedAbsent
      , (#AliasedAbsentType.t as node) ->
        Some node
    | AliasedPresent
      , (#AliasedPresentType.t as node) ->
        Some node
    | AllNode
      , (#AllNodeType.t as node) ->
        Some node
    | AllAbsent
      , (#AllAbsentType.t as node) ->
        Some node
    | AllPresent
      , (#AllPresentType.t as node) ->
        Some node
    | ArrayIndices
      , (#ArrayIndicesType.t as node) ->
        Some node
    | ConstrainedArrayIndices
      , (#ConstrainedArrayIndicesType.t as node) ->
        Some node
    | UnconstrainedArrayIndices
      , (#UnconstrainedArrayIndicesType.t as node) ->
        Some node
    | AspectAssoc
      , (#AspectAssocType.t as node) ->
        Some node
    | AspectClause
      , (#AspectClauseType.t as node) ->
        Some node
    | AtClause
      , (#AtClauseType.t as node) ->
        Some node
    | AttributeDefClause
      , (#AttributeDefClauseType.t as node) ->
        Some node
    | EnumRepClause
      , (#EnumRepClauseType.t as node) ->
        Some node
    | RecordRepClause
      , (#RecordRepClauseType.t as node) ->
        Some node
    | AspectSpec
      , (#AspectSpecType.t as node) ->
        Some node
    | BaseAssoc
      , (#BaseAssocType.t as node) ->
        Some node
    | ContractCaseAssoc
      , (#ContractCaseAssocType.t as node) ->
        Some node
    | PragmaArgumentAssoc
      , (#PragmaArgumentAssocType.t as node) ->
        Some node
    | BaseFormalParamHolder
      , (#BaseFormalParamHolderType.t as node) ->
        Some node
    | BaseSubpSpec
      , (#BaseSubpSpecType.t as node) ->
        Some node
    | EntrySpec
      , (#EntrySpecType.t as node) ->
        Some node
    | EnumSubpSpec
      , (#EnumSubpSpecType.t as node) ->
        Some node
    | SubpSpec
      , (#SubpSpecType.t as node) ->
        Some node
    | ComponentList
      , (#ComponentListType.t as node) ->
        Some node
    | DiscriminantPart
      , (#DiscriminantPartType.t as node) ->
        Some node
    | KnownDiscriminantPart
      , (#KnownDiscriminantPartType.t as node) ->
        Some node
    | UnknownDiscriminantPart
      , (#UnknownDiscriminantPartType.t as node) ->
        Some node
    | EntryCompletionFormalParams
      , (#EntryCompletionFormalParamsType.t as node) ->
        Some node
    | GenericFormalPart
      , (#GenericFormalPartType.t as node) ->
        Some node
    | BaseRecordDef
      , (#BaseRecordDefType.t as node) ->
        Some node
    | NullRecordDef
      , (#NullRecordDefType.t as node) ->
        Some node
    | RecordDef
      , (#RecordDefType.t as node) ->
        Some node
    | BasicAssoc
      , (#BasicAssocType.t as node) ->
        Some node
    | AggregateAssoc
      , (#AggregateAssocType.t as node) ->
        Some node
    | MultiDimArrayAssoc
      , (#MultiDimArrayAssocType.t as node) ->
        Some node
    | DiscriminantAssoc
      , (#DiscriminantAssocType.t as node) ->
        Some node
    | IteratedAssoc
      , (#IteratedAssocType.t as node) ->
        Some node
    | ParamAssoc
      , (#ParamAssocType.t as node) ->
        Some node
    | BasicDecl
      , (#BasicDeclType.t as node) ->
        Some node
    | AbstractStateDecl
      , (#AbstractStateDeclType.t as node) ->
        Some node
    | AnonymousExprDecl
      , (#AnonymousExprDeclType.t as node) ->
        Some node
    | BaseFormalParamDecl
      , (#BaseFormalParamDeclType.t as node) ->
        Some node
    | ComponentDecl
      , (#ComponentDeclType.t as node) ->
        Some node
    | DiscriminantSpec
      , (#DiscriminantSpecType.t as node) ->
        Some node
    | GenericFormal
      , (#GenericFormalType.t as node) ->
        Some node
    | GenericFormalObjDecl
      , (#GenericFormalObjDeclType.t as node) ->
        Some node
    | GenericFormalPackage
      , (#GenericFormalPackageType.t as node) ->
        Some node
    | GenericFormalSubpDecl
      , (#GenericFormalSubpDeclType.t as node) ->
        Some node
    | GenericFormalTypeDecl
      , (#GenericFormalTypeDeclType.t as node) ->
        Some node
    | ParamSpec
      , (#ParamSpecType.t as node) ->
        Some node
    | BasePackageDecl
      , (#BasePackageDeclType.t as node) ->
        Some node
    | GenericPackageInternal
      , (#GenericPackageInternalType.t as node) ->
        Some node
    | PackageDecl
      , (#PackageDeclType.t as node) ->
        Some node
    | BaseTypeDecl
      , (#BaseTypeDeclType.t as node) ->
        Some node
    | BaseSubtypeDecl
      , (#BaseSubtypeDeclType.t as node) ->
        Some node
    | DiscreteBaseSubtypeDecl
      , (#DiscreteBaseSubtypeDeclType.t as node) ->
        Some node
    | SubtypeDecl
      , (#SubtypeDeclType.t as node) ->
        Some node
    | ClasswideTypeDecl
      , (#ClasswideTypeDeclType.t as node) ->
        Some node
    | IncompleteTypeDecl
      , (#IncompleteTypeDeclType.t as node) ->
        Some node
    | IncompleteTaggedTypeDecl
      , (#IncompleteTaggedTypeDeclType.t as node) ->
        Some node
    | ProtectedTypeDecl
      , (#ProtectedTypeDeclType.t as node) ->
        Some node
    | TaskTypeDecl
      , (#TaskTypeDeclType.t as node) ->
        Some node
    | SingleTaskTypeDecl
      , (#SingleTaskTypeDeclType.t as node) ->
        Some node
    | TypeDecl
      , (#TypeDeclType.t as node) ->
        Some node
    | AnonymousTypeDecl
      , (#AnonymousTypeDeclType.t as node) ->
        Some node
    | SynthAnonymousTypeDecl
      , (#SynthAnonymousTypeDeclType.t as node) ->
        Some node
    | BasicSubpDecl
      , (#BasicSubpDeclType.t as node) ->
        Some node
    | ClassicSubpDecl
      , (#ClassicSubpDeclType.t as node) ->
        Some node
    | AbstractSubpDecl
      , (#AbstractSubpDeclType.t as node) ->
        Some node
    | FormalSubpDecl
      , (#FormalSubpDeclType.t as node) ->
        Some node
    | AbstractFormalSubpDecl
      , (#AbstractFormalSubpDeclType.t as node) ->
        Some node
    | ConcreteFormalSubpDecl
      , (#ConcreteFormalSubpDeclType.t as node) ->
        Some node
    | SubpDecl
      , (#SubpDeclType.t as node) ->
        Some node
    | EntryDecl
      , (#EntryDeclType.t as node) ->
        Some node
    | EnumLiteralDecl
      , (#EnumLiteralDeclType.t as node) ->
        Some node
    | GenericSubpInternal
      , (#GenericSubpInternalType.t as node) ->
        Some node
    | BodyNode
      , (#BodyNodeType.t as node) ->
        Some node
    | BaseSubpBody
      , (#BaseSubpBodyType.t as node) ->
        Some node
    | ExprFunction
      , (#ExprFunctionType.t as node) ->
        Some node
    | NullSubpDecl
      , (#NullSubpDeclType.t as node) ->
        Some node
    | SubpBody
      , (#SubpBodyType.t as node) ->
        Some node
    | SubpRenamingDecl
      , (#SubpRenamingDeclType.t as node) ->
        Some node
    | BodyStub
      , (#BodyStubType.t as node) ->
        Some node
    | PackageBodyStub
      , (#PackageBodyStubType.t as node) ->
        Some node
    | ProtectedBodyStub
      , (#ProtectedBodyStubType.t as node) ->
        Some node
    | SubpBodyStub
      , (#SubpBodyStubType.t as node) ->
        Some node
    | TaskBodyStub
      , (#TaskBodyStubType.t as node) ->
        Some node
    | EntryBody
      , (#EntryBodyType.t as node) ->
        Some node
    | PackageBody
      , (#PackageBodyType.t as node) ->
        Some node
    | ProtectedBody
      , (#ProtectedBodyType.t as node) ->
        Some node
    | TaskBody
      , (#TaskBodyType.t as node) ->
        Some node
    | EntryIndexSpec
      , (#EntryIndexSpecType.t as node) ->
        Some node
    | ErrorDecl
      , (#ErrorDeclType.t as node) ->
        Some node
    | ExceptionDecl
      , (#ExceptionDeclType.t as node) ->
        Some node
    | ExceptionHandler
      , (#ExceptionHandlerType.t as node) ->
        Some node
    | ForLoopVarDecl
      , (#ForLoopVarDeclType.t as node) ->
        Some node
    | GenericDecl
      , (#GenericDeclType.t as node) ->
        Some node
    | GenericPackageDecl
      , (#GenericPackageDeclType.t as node) ->
        Some node
    | GenericSubpDecl
      , (#GenericSubpDeclType.t as node) ->
        Some node
    | GenericInstantiation
      , (#GenericInstantiationType.t as node) ->
        Some node
    | GenericPackageInstantiation
      , (#GenericPackageInstantiationType.t as node) ->
        Some node
    | GenericSubpInstantiation
      , (#GenericSubpInstantiationType.t as node) ->
        Some node
    | GenericRenamingDecl
      , (#GenericRenamingDeclType.t as node) ->
        Some node
    | GenericPackageRenamingDecl
      , (#GenericPackageRenamingDeclType.t as node) ->
        Some node
    | GenericSubpRenamingDecl
      , (#GenericSubpRenamingDeclType.t as node) ->
        Some node
    | LabelDecl
      , (#LabelDeclType.t as node) ->
        Some node
    | NamedStmtDecl
      , (#NamedStmtDeclType.t as node) ->
        Some node
    | NumberDecl
      , (#NumberDeclType.t as node) ->
        Some node
    | ObjectDecl
      , (#ObjectDeclType.t as node) ->
        Some node
    | ExtendedReturnStmtObjectDecl
      , (#ExtendedReturnStmtObjectDeclType.t as node) ->
        Some node
    | PackageRenamingDecl
      , (#PackageRenamingDeclType.t as node) ->
        Some node
    | SingleProtectedDecl
      , (#SingleProtectedDeclType.t as node) ->
        Some node
    | SingleTaskDecl
      , (#SingleTaskDeclType.t as node) ->
        Some node
    | CaseStmtAlternative
      , (#CaseStmtAlternativeType.t as node) ->
        Some node
    | CompilationUnit
      , (#CompilationUnitType.t as node) ->
        Some node
    | ComponentClause
      , (#ComponentClauseType.t as node) ->
        Some node
    | ComponentDef
      , (#ComponentDefType.t as node) ->
        Some node
    | ConstantNode
      , (#ConstantNodeType.t as node) ->
        Some node
    | ConstantAbsent
      , (#ConstantAbsentType.t as node) ->
        Some node
    | ConstantPresent
      , (#ConstantPresentType.t as node) ->
        Some node
    | Constraint
      , (#ConstraintType.t as node) ->
        Some node
    | DeltaConstraint
      , (#DeltaConstraintType.t as node) ->
        Some node
    | DigitsConstraint
      , (#DigitsConstraintType.t as node) ->
        Some node
    | DiscriminantConstraint
      , (#DiscriminantConstraintType.t as node) ->
        Some node
    | IndexConstraint
      , (#IndexConstraintType.t as node) ->
        Some node
    | RangeConstraint
      , (#RangeConstraintType.t as node) ->
        Some node
    | DeclarativePart
      , (#DeclarativePartType.t as node) ->
        Some node
    | PrivatePart
      , (#PrivatePartType.t as node) ->
        Some node
    | PublicPart
      , (#PublicPartType.t as node) ->
        Some node
    | ElsifExprPart
      , (#ElsifExprPartType.t as node) ->
        Some node
    | ElsifStmtPart
      , (#ElsifStmtPartType.t as node) ->
        Some node
    | Expr
      , (#ExprType.t as node) ->
        Some node
    | AbstractStateDeclExpr
      , (#AbstractStateDeclExprType.t as node) ->
        Some node
    | Allocator
      , (#AllocatorType.t as node) ->
        Some node
    | BaseAggregate
      , (#BaseAggregateType.t as node) ->
        Some node
    | Aggregate
      , (#AggregateType.t as node) ->
        Some node
    | BracketAggregate
      , (#BracketAggregateType.t as node) ->
        Some node
    | DeltaAggregate
      , (#DeltaAggregateType.t as node) ->
        Some node
    | BracketDeltaAggregate
      , (#BracketDeltaAggregateType.t as node) ->
        Some node
    | NullRecordAggregate
      , (#NullRecordAggregateType.t as node) ->
        Some node
    | BinOp
      , (#BinOpType.t as node) ->
        Some node
    | RelationOp
      , (#RelationOpType.t as node) ->
        Some node
    | BoxExpr
      , (#BoxExprType.t as node) ->
        Some node
    | CaseExprAlternative
      , (#CaseExprAlternativeType.t as node) ->
        Some node
    | CondExpr
      , (#CondExprType.t as node) ->
        Some node
    | CaseExpr
      , (#CaseExprType.t as node) ->
        Some node
    | IfExpr
      , (#IfExprType.t as node) ->
        Some node
    | ContractCases
      , (#ContractCasesType.t as node) ->
        Some node
    | DeclExpr
      , (#DeclExprType.t as node) ->
        Some node
    | MembershipExpr
      , (#MembershipExprType.t as node) ->
        Some node
    | Name
      , (#NameType.t as node) ->
        Some node
    | AttributeRef
      , (#AttributeRefType.t as node) ->
        Some node
    | UpdateAttributeRef
      , (#UpdateAttributeRefType.t as node) ->
        Some node
    | CallExpr
      , (#CallExprType.t as node) ->
        Some node
    | DefiningName
      , (#DefiningNameType.t as node) ->
        Some node
    | DiscreteSubtypeName
      , (#DiscreteSubtypeNameType.t as node) ->
        Some node
    | DottedName
      , (#DottedNameType.t as node) ->
        Some node
    | EndName
      , (#EndNameType.t as node) ->
        Some node
    | ExplicitDeref
      , (#ExplicitDerefType.t as node) ->
        Some node
    | QualExpr
      , (#QualExprType.t as node) ->
        Some node
    | SingleTokNode
      , (#SingleTokNodeType.t as node) ->
        Some node
    | BaseId
      , (#BaseIdType.t as node) ->
        Some node
    | CharLiteral
      , (#CharLiteralType.t as node) ->
        Some node
    | Identifier
      , (#IdentifierType.t as node) ->
        Some node
    | Op
      , (#OpType.t as node) ->
        Some node
    | OpAbs
      , (#OpAbsType.t as node) ->
        Some node
    | OpAnd
      , (#OpAndType.t as node) ->
        Some node
    | OpAndThen
      , (#OpAndThenType.t as node) ->
        Some node
    | OpConcat
      , (#OpConcatType.t as node) ->
        Some node
    | OpDiv
      , (#OpDivType.t as node) ->
        Some node
    | OpDoubleDot
      , (#OpDoubleDotType.t as node) ->
        Some node
    | OpEq
      , (#OpEqType.t as node) ->
        Some node
    | OpGt
      , (#OpGtType.t as node) ->
        Some node
    | OpGte
      , (#OpGteType.t as node) ->
        Some node
    | OpIn
      , (#OpInType.t as node) ->
        Some node
    | OpLt
      , (#OpLtType.t as node) ->
        Some node
    | OpLte
      , (#OpLteType.t as node) ->
        Some node
    | OpMinus
      , (#OpMinusType.t as node) ->
        Some node
    | OpMod
      , (#OpModType.t as node) ->
        Some node
    | OpMult
      , (#OpMultType.t as node) ->
        Some node
    | OpNeq
      , (#OpNeqType.t as node) ->
        Some node
    | OpNot
      , (#OpNotType.t as node) ->
        Some node
    | OpNotIn
      , (#OpNotInType.t as node) ->
        Some node
    | OpOr
      , (#OpOrType.t as node) ->
        Some node
    | OpOrElse
      , (#OpOrElseType.t as node) ->
        Some node
    | OpPlus
      , (#OpPlusType.t as node) ->
        Some node
    | OpPow
      , (#OpPowType.t as node) ->
        Some node
    | OpRem
      , (#OpRemType.t as node) ->
        Some node
    | OpXor
      , (#OpXorType.t as node) ->
        Some node
    | StringLiteral
      , (#StringLiteralType.t as node) ->
        Some node
    | NullLiteral
      , (#NullLiteralType.t as node) ->
        Some node
    | NumLiteral
      , (#NumLiteralType.t as node) ->
        Some node
    | IntLiteral
      , (#IntLiteralType.t as node) ->
        Some node
    | RealLiteral
      , (#RealLiteralType.t as node) ->
        Some node
    | TargetName
      , (#TargetNameType.t as node) ->
        Some node
    | ParenExpr
      , (#ParenExprType.t as node) ->
        Some node
    | QuantifiedExpr
      , (#QuantifiedExprType.t as node) ->
        Some node
    | RaiseExpr
      , (#RaiseExprType.t as node) ->
        Some node
    | UnOp
      , (#UnOpType.t as node) ->
        Some node
    | HandledStmts
      , (#HandledStmtsType.t as node) ->
        Some node
    | InterfaceKind
      , (#InterfaceKindType.t as node) ->
        Some node
    | InterfaceKindLimited
      , (#InterfaceKindLimitedType.t as node) ->
        Some node
    | InterfaceKindProtected
      , (#InterfaceKindProtectedType.t as node) ->
        Some node
    | InterfaceKindSynchronized
      , (#InterfaceKindSynchronizedType.t as node) ->
        Some node
    | InterfaceKindTask
      , (#InterfaceKindTaskType.t as node) ->
        Some node
    | IterType
      , (#IterTypeType.t as node) ->
        Some node
    | IterTypeIn
      , (#IterTypeInType.t as node) ->
        Some node
    | IterTypeOf
      , (#IterTypeOfType.t as node) ->
        Some node
    | LibraryItem
      , (#LibraryItemType.t as node) ->
        Some node
    | LimitedNode
      , (#LimitedNodeType.t as node) ->
        Some node
    | LimitedAbsent
      , (#LimitedAbsentType.t as node) ->
        Some node
    | LimitedPresent
      , (#LimitedPresentType.t as node) ->
        Some node
    | LoopSpec
      , (#LoopSpecType.t as node) ->
        Some node
    | ForLoopSpec
      , (#ForLoopSpecType.t as node) ->
        Some node
    | WhileLoopSpec
      , (#WhileLoopSpecType.t as node) ->
        Some node
    | Mode
      , (#ModeType.t as node) ->
        Some node
    | ModeDefault
      , (#ModeDefaultType.t as node) ->
        Some node
    | ModeIn
      , (#ModeInType.t as node) ->
        Some node
    | ModeInOut
      , (#ModeInOutType.t as node) ->
        Some node
    | ModeOut
      , (#ModeOutType.t as node) ->
        Some node
    | MultiAbstractStateDecl
      , (#MultiAbstractStateDeclType.t as node) ->
        Some node
    | NotNull
      , (#NotNullType.t as node) ->
        Some node
    | NotNullAbsent
      , (#NotNullAbsentType.t as node) ->
        Some node
    | NotNullPresent
      , (#NotNullPresentType.t as node) ->
        Some node
    | NullComponentDecl
      , (#NullComponentDeclType.t as node) ->
        Some node
    | OthersDesignator
      , (#OthersDesignatorType.t as node) ->
        Some node
    | OverridingNode
      , (#OverridingNodeType.t as node) ->
        Some node
    | OverridingNotOverriding
      , (#OverridingNotOverridingType.t as node) ->
        Some node
    | OverridingOverriding
      , (#OverridingOverridingType.t as node) ->
        Some node
    | OverridingUnspecified
      , (#OverridingUnspecifiedType.t as node) ->
        Some node
    | Params
      , (#ParamsType.t as node) ->
        Some node
    | ParenAbstractStateDecl
      , (#ParenAbstractStateDeclType.t as node) ->
        Some node
    | PragmaNode
      , (#PragmaNodeType.t as node) ->
        Some node
    | PrivateNode
      , (#PrivateNodeType.t as node) ->
        Some node
    | PrivateAbsent
      , (#PrivateAbsentType.t as node) ->
        Some node
    | PrivatePresent
      , (#PrivatePresentType.t as node) ->
        Some node
    | ProtectedDef
      , (#ProtectedDefType.t as node) ->
        Some node
    | ProtectedNode
      , (#ProtectedNodeType.t as node) ->
        Some node
    | ProtectedAbsent
      , (#ProtectedAbsentType.t as node) ->
        Some node
    | ProtectedPresent
      , (#ProtectedPresentType.t as node) ->
        Some node
    | Quantifier
      , (#QuantifierType.t as node) ->
        Some node
    | QuantifierAll
      , (#QuantifierAllType.t as node) ->
        Some node
    | QuantifierSome
      , (#QuantifierSomeType.t as node) ->
        Some node
    | RangeSpec
      , (#RangeSpecType.t as node) ->
        Some node
    | RenamingClause
      , (#RenamingClauseType.t as node) ->
        Some node
    | SyntheticRenamingClause
      , (#SyntheticRenamingClauseType.t as node) ->
        Some node
    | ReverseNode
      , (#ReverseNodeType.t as node) ->
        Some node
    | ReverseAbsent
      , (#ReverseAbsentType.t as node) ->
        Some node
    | ReversePresent
      , (#ReversePresentType.t as node) ->
        Some node
    | SelectWhenPart
      , (#SelectWhenPartType.t as node) ->
        Some node
    | Stmt
      , (#StmtType.t as node) ->
        Some node
    | CompositeStmt
      , (#CompositeStmtType.t as node) ->
        Some node
    | AcceptStmt
      , (#AcceptStmtType.t as node) ->
        Some node
    | AcceptStmtWithStmts
      , (#AcceptStmtWithStmtsType.t as node) ->
        Some node
    | BaseLoopStmt
      , (#BaseLoopStmtType.t as node) ->
        Some node
    | ForLoopStmt
      , (#ForLoopStmtType.t as node) ->
        Some node
    | LoopStmt
      , (#LoopStmtType.t as node) ->
        Some node
    | WhileLoopStmt
      , (#WhileLoopStmtType.t as node) ->
        Some node
    | BlockStmt
      , (#BlockStmtType.t as node) ->
        Some node
    | BeginBlock
      , (#BeginBlockType.t as node) ->
        Some node
    | DeclBlock
      , (#DeclBlockType.t as node) ->
        Some node
    | CaseStmt
      , (#CaseStmtType.t as node) ->
        Some node
    | ExtendedReturnStmt
      , (#ExtendedReturnStmtType.t as node) ->
        Some node
    | IfStmt
      , (#IfStmtType.t as node) ->
        Some node
    | NamedStmt
      , (#NamedStmtType.t as node) ->
        Some node
    | SelectStmt
      , (#SelectStmtType.t as node) ->
        Some node
    | ErrorStmt
      , (#ErrorStmtType.t as node) ->
        Some node
    | SimpleStmt
      , (#SimpleStmtType.t as node) ->
        Some node
    | AbortStmt
      , (#AbortStmtType.t as node) ->
        Some node
    | AssignStmt
      , (#AssignStmtType.t as node) ->
        Some node
    | CallStmt
      , (#CallStmtType.t as node) ->
        Some node
    | DelayStmt
      , (#DelayStmtType.t as node) ->
        Some node
    | ExitStmt
      , (#ExitStmtType.t as node) ->
        Some node
    | GotoStmt
      , (#GotoStmtType.t as node) ->
        Some node
    | Label
      , (#LabelType.t as node) ->
        Some node
    | NullStmt
      , (#NullStmtType.t as node) ->
        Some node
    | RaiseStmt
      , (#RaiseStmtType.t as node) ->
        Some node
    | RequeueStmt
      , (#RequeueStmtType.t as node) ->
        Some node
    | ReturnStmt
      , (#ReturnStmtType.t as node) ->
        Some node
    | TerminateAlternative
      , (#TerminateAlternativeType.t as node) ->
        Some node
    | SubpKind
      , (#SubpKindType.t as node) ->
        Some node
    | SubpKindFunction
      , (#SubpKindFunctionType.t as node) ->
        Some node
    | SubpKindProcedure
      , (#SubpKindProcedureType.t as node) ->
        Some node
    | Subunit
      , (#SubunitType.t as node) ->
        Some node
    | SynchronizedNode
      , (#SynchronizedNodeType.t as node) ->
        Some node
    | SynchronizedAbsent
      , (#SynchronizedAbsentType.t as node) ->
        Some node
    | SynchronizedPresent
      , (#SynchronizedPresentType.t as node) ->
        Some node
    | TaggedNode
      , (#TaggedNodeType.t as node) ->
        Some node
    | TaggedAbsent
      , (#TaggedAbsentType.t as node) ->
        Some node
    | TaggedPresent
      , (#TaggedPresentType.t as node) ->
        Some node
    | TaskDef
      , (#TaskDefType.t as node) ->
        Some node
    | TypeDef
      , (#TypeDefType.t as node) ->
        Some node
    | AccessDef
      , (#AccessDefType.t as node) ->
        Some node
    | AccessToSubpDef
      , (#AccessToSubpDefType.t as node) ->
        Some node
    | BaseTypeAccessDef
      , (#BaseTypeAccessDefType.t as node) ->
        Some node
    | AnonymousTypeAccessDef
      , (#AnonymousTypeAccessDefType.t as node) ->
        Some node
    | TypeAccessDef
      , (#TypeAccessDefType.t as node) ->
        Some node
    | ArrayTypeDef
      , (#ArrayTypeDefType.t as node) ->
        Some node
    | DerivedTypeDef
      , (#DerivedTypeDefType.t as node) ->
        Some node
    | EnumTypeDef
      , (#EnumTypeDefType.t as node) ->
        Some node
    | FormalDiscreteTypeDef
      , (#FormalDiscreteTypeDefType.t as node) ->
        Some node
    | InterfaceTypeDef
      , (#InterfaceTypeDefType.t as node) ->
        Some node
    | ModIntTypeDef
      , (#ModIntTypeDefType.t as node) ->
        Some node
    | PrivateTypeDef
      , (#PrivateTypeDefType.t as node) ->
        Some node
    | RealTypeDef
      , (#RealTypeDefType.t as node) ->
        Some node
    | DecimalFixedPointDef
      , (#DecimalFixedPointDefType.t as node) ->
        Some node
    | FloatingPointDef
      , (#FloatingPointDefType.t as node) ->
        Some node
    | OrdinaryFixedPointDef
      , (#OrdinaryFixedPointDefType.t as node) ->
        Some node
    | RecordTypeDef
      , (#RecordTypeDefType.t as node) ->
        Some node
    | SignedIntTypeDef
      , (#SignedIntTypeDefType.t as node) ->
        Some node
    | TypeExpr
      , (#TypeExprType.t as node) ->
        Some node
    | AnonymousType
      , (#AnonymousTypeType.t as node) ->
        Some node
    | EnumLitSynthTypeExpr
      , (#EnumLitSynthTypeExprType.t as node) ->
        Some node
    | SubtypeIndication
      , (#SubtypeIndicationType.t as node) ->
        Some node
    | ConstrainedSubtypeIndication
      , (#ConstrainedSubtypeIndicationType.t as node) ->
        Some node
    | DiscreteSubtypeIndication
      , (#DiscreteSubtypeIndicationType.t as node) ->
        Some node
    | UnconstrainedArrayIndex
      , (#UnconstrainedArrayIndexType.t as node) ->
        Some node
    | UntilNode
      , (#UntilNodeType.t as node) ->
        Some node
    | UntilAbsent
      , (#UntilAbsentType.t as node) ->
        Some node
    | UntilPresent
      , (#UntilPresentType.t as node) ->
        Some node
    | UseClause
      , (#UseClauseType.t as node) ->
        Some node
    | UsePackageClause
      , (#UsePackageClauseType.t as node) ->
        Some node
    | UseTypeClause
      , (#UseTypeClauseType.t as node) ->
        Some node
    | Variant
      , (#VariantType.t as node) ->
        Some node
    | VariantPart
      , (#VariantPartType.t as node) ->
        Some node
    | WithClause
      , (#WithClauseType.t as node) ->
        Some node
    | WithPrivate
      , (#WithPrivateType.t as node) ->
        Some node
    | WithPrivateAbsent
      , (#WithPrivateAbsentType.t as node) ->
        Some node
    | WithPrivatePresent
      , (#WithPrivatePresentType.t as node) ->
        Some node
    | _ ->
        None

  let find : type a. a node ->  [< AdaNodeType.t ] -> a =
    fun node_type node ->
      let exception Found of a in
      let aux node =
        match node_type, node with
        | AdaNode
          , (#AdaNodeType.t as node) ->
            raise (Found node)
        | AbortNode
          , (#AbortNodeType.t as node) ->
            raise (Found node)
        | AbortAbsent
          , (#AbortAbsentType.t as node) ->
            raise (Found node)
        | AbortPresent
          , (#AbortPresentType.t as node) ->
            raise (Found node)
        | AbstractNode
          , (#AbstractNodeType.t as node) ->
            raise (Found node)
        | AbstractAbsent
          , (#AbstractAbsentType.t as node) ->
            raise (Found node)
        | AbstractPresent
          , (#AbstractPresentType.t as node) ->
            raise (Found node)
        | AdaList
          , (#AdaListType.t as node) ->
            raise (Found node)
        | AdaNodeList
          , (#AdaNodeListType.t as node) ->
            raise (Found node)
        | AbstractStateDeclList
          , (#AbstractStateDeclListType.t as node) ->
            raise (Found node)
        | AlternativesList
          , (#AlternativesListType.t as node) ->
            raise (Found node)
        | ConstraintList
          , (#ConstraintListType.t as node) ->
            raise (Found node)
        | DeclList
          , (#DeclListType.t as node) ->
            raise (Found node)
        | StmtList
          , (#StmtListType.t as node) ->
            raise (Found node)
        | AspectAssocList
          , (#AspectAssocListType.t as node) ->
            raise (Found node)
        | BaseAssocList
          , (#BaseAssocListType.t as node) ->
            raise (Found node)
        | BasicAssocList
          , (#BasicAssocListType.t as node) ->
            raise (Found node)
        | AssocList
          , (#AssocListType.t as node) ->
            raise (Found node)
        | BasicDeclList
          , (#BasicDeclListType.t as node) ->
            raise (Found node)
        | CaseExprAlternativeList
          , (#CaseExprAlternativeListType.t as node) ->
            raise (Found node)
        | CaseStmtAlternativeList
          , (#CaseStmtAlternativeListType.t as node) ->
            raise (Found node)
        | CompilationUnitList
          , (#CompilationUnitListType.t as node) ->
            raise (Found node)
        | ContractCaseAssocList
          , (#ContractCaseAssocListType.t as node) ->
            raise (Found node)
        | DefiningNameList
          , (#DefiningNameListType.t as node) ->
            raise (Found node)
        | DiscriminantSpecList
          , (#DiscriminantSpecListType.t as node) ->
            raise (Found node)
        | ElsifExprPartList
          , (#ElsifExprPartListType.t as node) ->
            raise (Found node)
        | ElsifStmtPartList
          , (#ElsifStmtPartListType.t as node) ->
            raise (Found node)
        | EnumLiteralDeclList
          , (#EnumLiteralDeclListType.t as node) ->
            raise (Found node)
        | ExprList
          , (#ExprListType.t as node) ->
            raise (Found node)
        | ExprAlternativesList
          , (#ExprAlternativesListType.t as node) ->
            raise (Found node)
        | IdentifierList
          , (#IdentifierListType.t as node) ->
            raise (Found node)
        | DiscriminantChoiceList
          , (#DiscriminantChoiceListType.t as node) ->
            raise (Found node)
        | NameList
          , (#NameListType.t as node) ->
            raise (Found node)
        | ParentList
          , (#ParentListType.t as node) ->
            raise (Found node)
        | ParamSpecList
          , (#ParamSpecListType.t as node) ->
            raise (Found node)
        | PragmaNodeList
          , (#PragmaNodeListType.t as node) ->
            raise (Found node)
        | SelectWhenPartList
          , (#SelectWhenPartListType.t as node) ->
            raise (Found node)
        | UnconstrainedArrayIndexList
          , (#UnconstrainedArrayIndexListType.t as node) ->
            raise (Found node)
        | VariantList
          , (#VariantListType.t as node) ->
            raise (Found node)
        | AliasedNode
          , (#AliasedNodeType.t as node) ->
            raise (Found node)
        | AliasedAbsent
          , (#AliasedAbsentType.t as node) ->
            raise (Found node)
        | AliasedPresent
          , (#AliasedPresentType.t as node) ->
            raise (Found node)
        | AllNode
          , (#AllNodeType.t as node) ->
            raise (Found node)
        | AllAbsent
          , (#AllAbsentType.t as node) ->
            raise (Found node)
        | AllPresent
          , (#AllPresentType.t as node) ->
            raise (Found node)
        | ArrayIndices
          , (#ArrayIndicesType.t as node) ->
            raise (Found node)
        | ConstrainedArrayIndices
          , (#ConstrainedArrayIndicesType.t as node) ->
            raise (Found node)
        | UnconstrainedArrayIndices
          , (#UnconstrainedArrayIndicesType.t as node) ->
            raise (Found node)
        | AspectAssoc
          , (#AspectAssocType.t as node) ->
            raise (Found node)
        | AspectClause
          , (#AspectClauseType.t as node) ->
            raise (Found node)
        | AtClause
          , (#AtClauseType.t as node) ->
            raise (Found node)
        | AttributeDefClause
          , (#AttributeDefClauseType.t as node) ->
            raise (Found node)
        | EnumRepClause
          , (#EnumRepClauseType.t as node) ->
            raise (Found node)
        | RecordRepClause
          , (#RecordRepClauseType.t as node) ->
            raise (Found node)
        | AspectSpec
          , (#AspectSpecType.t as node) ->
            raise (Found node)
        | BaseAssoc
          , (#BaseAssocType.t as node) ->
            raise (Found node)
        | ContractCaseAssoc
          , (#ContractCaseAssocType.t as node) ->
            raise (Found node)
        | PragmaArgumentAssoc
          , (#PragmaArgumentAssocType.t as node) ->
            raise (Found node)
        | BaseFormalParamHolder
          , (#BaseFormalParamHolderType.t as node) ->
            raise (Found node)
        | BaseSubpSpec
          , (#BaseSubpSpecType.t as node) ->
            raise (Found node)
        | EntrySpec
          , (#EntrySpecType.t as node) ->
            raise (Found node)
        | EnumSubpSpec
          , (#EnumSubpSpecType.t as node) ->
            raise (Found node)
        | SubpSpec
          , (#SubpSpecType.t as node) ->
            raise (Found node)
        | ComponentList
          , (#ComponentListType.t as node) ->
            raise (Found node)
        | DiscriminantPart
          , (#DiscriminantPartType.t as node) ->
            raise (Found node)
        | KnownDiscriminantPart
          , (#KnownDiscriminantPartType.t as node) ->
            raise (Found node)
        | UnknownDiscriminantPart
          , (#UnknownDiscriminantPartType.t as node) ->
            raise (Found node)
        | EntryCompletionFormalParams
          , (#EntryCompletionFormalParamsType.t as node) ->
            raise (Found node)
        | GenericFormalPart
          , (#GenericFormalPartType.t as node) ->
            raise (Found node)
        | BaseRecordDef
          , (#BaseRecordDefType.t as node) ->
            raise (Found node)
        | NullRecordDef
          , (#NullRecordDefType.t as node) ->
            raise (Found node)
        | RecordDef
          , (#RecordDefType.t as node) ->
            raise (Found node)
        | BasicAssoc
          , (#BasicAssocType.t as node) ->
            raise (Found node)
        | AggregateAssoc
          , (#AggregateAssocType.t as node) ->
            raise (Found node)
        | MultiDimArrayAssoc
          , (#MultiDimArrayAssocType.t as node) ->
            raise (Found node)
        | DiscriminantAssoc
          , (#DiscriminantAssocType.t as node) ->
            raise (Found node)
        | IteratedAssoc
          , (#IteratedAssocType.t as node) ->
            raise (Found node)
        | ParamAssoc
          , (#ParamAssocType.t as node) ->
            raise (Found node)
        | BasicDecl
          , (#BasicDeclType.t as node) ->
            raise (Found node)
        | AbstractStateDecl
          , (#AbstractStateDeclType.t as node) ->
            raise (Found node)
        | AnonymousExprDecl
          , (#AnonymousExprDeclType.t as node) ->
            raise (Found node)
        | BaseFormalParamDecl
          , (#BaseFormalParamDeclType.t as node) ->
            raise (Found node)
        | ComponentDecl
          , (#ComponentDeclType.t as node) ->
            raise (Found node)
        | DiscriminantSpec
          , (#DiscriminantSpecType.t as node) ->
            raise (Found node)
        | GenericFormal
          , (#GenericFormalType.t as node) ->
            raise (Found node)
        | GenericFormalObjDecl
          , (#GenericFormalObjDeclType.t as node) ->
            raise (Found node)
        | GenericFormalPackage
          , (#GenericFormalPackageType.t as node) ->
            raise (Found node)
        | GenericFormalSubpDecl
          , (#GenericFormalSubpDeclType.t as node) ->
            raise (Found node)
        | GenericFormalTypeDecl
          , (#GenericFormalTypeDeclType.t as node) ->
            raise (Found node)
        | ParamSpec
          , (#ParamSpecType.t as node) ->
            raise (Found node)
        | BasePackageDecl
          , (#BasePackageDeclType.t as node) ->
            raise (Found node)
        | GenericPackageInternal
          , (#GenericPackageInternalType.t as node) ->
            raise (Found node)
        | PackageDecl
          , (#PackageDeclType.t as node) ->
            raise (Found node)
        | BaseTypeDecl
          , (#BaseTypeDeclType.t as node) ->
            raise (Found node)
        | BaseSubtypeDecl
          , (#BaseSubtypeDeclType.t as node) ->
            raise (Found node)
        | DiscreteBaseSubtypeDecl
          , (#DiscreteBaseSubtypeDeclType.t as node) ->
            raise (Found node)
        | SubtypeDecl
          , (#SubtypeDeclType.t as node) ->
            raise (Found node)
        | ClasswideTypeDecl
          , (#ClasswideTypeDeclType.t as node) ->
            raise (Found node)
        | IncompleteTypeDecl
          , (#IncompleteTypeDeclType.t as node) ->
            raise (Found node)
        | IncompleteTaggedTypeDecl
          , (#IncompleteTaggedTypeDeclType.t as node) ->
            raise (Found node)
        | ProtectedTypeDecl
          , (#ProtectedTypeDeclType.t as node) ->
            raise (Found node)
        | TaskTypeDecl
          , (#TaskTypeDeclType.t as node) ->
            raise (Found node)
        | SingleTaskTypeDecl
          , (#SingleTaskTypeDeclType.t as node) ->
            raise (Found node)
        | TypeDecl
          , (#TypeDeclType.t as node) ->
            raise (Found node)
        | AnonymousTypeDecl
          , (#AnonymousTypeDeclType.t as node) ->
            raise (Found node)
        | SynthAnonymousTypeDecl
          , (#SynthAnonymousTypeDeclType.t as node) ->
            raise (Found node)
        | BasicSubpDecl
          , (#BasicSubpDeclType.t as node) ->
            raise (Found node)
        | ClassicSubpDecl
          , (#ClassicSubpDeclType.t as node) ->
            raise (Found node)
        | AbstractSubpDecl
          , (#AbstractSubpDeclType.t as node) ->
            raise (Found node)
        | FormalSubpDecl
          , (#FormalSubpDeclType.t as node) ->
            raise (Found node)
        | AbstractFormalSubpDecl
          , (#AbstractFormalSubpDeclType.t as node) ->
            raise (Found node)
        | ConcreteFormalSubpDecl
          , (#ConcreteFormalSubpDeclType.t as node) ->
            raise (Found node)
        | SubpDecl
          , (#SubpDeclType.t as node) ->
            raise (Found node)
        | EntryDecl
          , (#EntryDeclType.t as node) ->
            raise (Found node)
        | EnumLiteralDecl
          , (#EnumLiteralDeclType.t as node) ->
            raise (Found node)
        | GenericSubpInternal
          , (#GenericSubpInternalType.t as node) ->
            raise (Found node)
        | BodyNode
          , (#BodyNodeType.t as node) ->
            raise (Found node)
        | BaseSubpBody
          , (#BaseSubpBodyType.t as node) ->
            raise (Found node)
        | ExprFunction
          , (#ExprFunctionType.t as node) ->
            raise (Found node)
        | NullSubpDecl
          , (#NullSubpDeclType.t as node) ->
            raise (Found node)
        | SubpBody
          , (#SubpBodyType.t as node) ->
            raise (Found node)
        | SubpRenamingDecl
          , (#SubpRenamingDeclType.t as node) ->
            raise (Found node)
        | BodyStub
          , (#BodyStubType.t as node) ->
            raise (Found node)
        | PackageBodyStub
          , (#PackageBodyStubType.t as node) ->
            raise (Found node)
        | ProtectedBodyStub
          , (#ProtectedBodyStubType.t as node) ->
            raise (Found node)
        | SubpBodyStub
          , (#SubpBodyStubType.t as node) ->
            raise (Found node)
        | TaskBodyStub
          , (#TaskBodyStubType.t as node) ->
            raise (Found node)
        | EntryBody
          , (#EntryBodyType.t as node) ->
            raise (Found node)
        | PackageBody
          , (#PackageBodyType.t as node) ->
            raise (Found node)
        | ProtectedBody
          , (#ProtectedBodyType.t as node) ->
            raise (Found node)
        | TaskBody
          , (#TaskBodyType.t as node) ->
            raise (Found node)
        | EntryIndexSpec
          , (#EntryIndexSpecType.t as node) ->
            raise (Found node)
        | ErrorDecl
          , (#ErrorDeclType.t as node) ->
            raise (Found node)
        | ExceptionDecl
          , (#ExceptionDeclType.t as node) ->
            raise (Found node)
        | ExceptionHandler
          , (#ExceptionHandlerType.t as node) ->
            raise (Found node)
        | ForLoopVarDecl
          , (#ForLoopVarDeclType.t as node) ->
            raise (Found node)
        | GenericDecl
          , (#GenericDeclType.t as node) ->
            raise (Found node)
        | GenericPackageDecl
          , (#GenericPackageDeclType.t as node) ->
            raise (Found node)
        | GenericSubpDecl
          , (#GenericSubpDeclType.t as node) ->
            raise (Found node)
        | GenericInstantiation
          , (#GenericInstantiationType.t as node) ->
            raise (Found node)
        | GenericPackageInstantiation
          , (#GenericPackageInstantiationType.t as node) ->
            raise (Found node)
        | GenericSubpInstantiation
          , (#GenericSubpInstantiationType.t as node) ->
            raise (Found node)
        | GenericRenamingDecl
          , (#GenericRenamingDeclType.t as node) ->
            raise (Found node)
        | GenericPackageRenamingDecl
          , (#GenericPackageRenamingDeclType.t as node) ->
            raise (Found node)
        | GenericSubpRenamingDecl
          , (#GenericSubpRenamingDeclType.t as node) ->
            raise (Found node)
        | LabelDecl
          , (#LabelDeclType.t as node) ->
            raise (Found node)
        | NamedStmtDecl
          , (#NamedStmtDeclType.t as node) ->
            raise (Found node)
        | NumberDecl
          , (#NumberDeclType.t as node) ->
            raise (Found node)
        | ObjectDecl
          , (#ObjectDeclType.t as node) ->
            raise (Found node)
        | ExtendedReturnStmtObjectDecl
          , (#ExtendedReturnStmtObjectDeclType.t as node) ->
            raise (Found node)
        | PackageRenamingDecl
          , (#PackageRenamingDeclType.t as node) ->
            raise (Found node)
        | SingleProtectedDecl
          , (#SingleProtectedDeclType.t as node) ->
            raise (Found node)
        | SingleTaskDecl
          , (#SingleTaskDeclType.t as node) ->
            raise (Found node)
        | CaseStmtAlternative
          , (#CaseStmtAlternativeType.t as node) ->
            raise (Found node)
        | CompilationUnit
          , (#CompilationUnitType.t as node) ->
            raise (Found node)
        | ComponentClause
          , (#ComponentClauseType.t as node) ->
            raise (Found node)
        | ComponentDef
          , (#ComponentDefType.t as node) ->
            raise (Found node)
        | ConstantNode
          , (#ConstantNodeType.t as node) ->
            raise (Found node)
        | ConstantAbsent
          , (#ConstantAbsentType.t as node) ->
            raise (Found node)
        | ConstantPresent
          , (#ConstantPresentType.t as node) ->
            raise (Found node)
        | Constraint
          , (#ConstraintType.t as node) ->
            raise (Found node)
        | DeltaConstraint
          , (#DeltaConstraintType.t as node) ->
            raise (Found node)
        | DigitsConstraint
          , (#DigitsConstraintType.t as node) ->
            raise (Found node)
        | DiscriminantConstraint
          , (#DiscriminantConstraintType.t as node) ->
            raise (Found node)
        | IndexConstraint
          , (#IndexConstraintType.t as node) ->
            raise (Found node)
        | RangeConstraint
          , (#RangeConstraintType.t as node) ->
            raise (Found node)
        | DeclarativePart
          , (#DeclarativePartType.t as node) ->
            raise (Found node)
        | PrivatePart
          , (#PrivatePartType.t as node) ->
            raise (Found node)
        | PublicPart
          , (#PublicPartType.t as node) ->
            raise (Found node)
        | ElsifExprPart
          , (#ElsifExprPartType.t as node) ->
            raise (Found node)
        | ElsifStmtPart
          , (#ElsifStmtPartType.t as node) ->
            raise (Found node)
        | Expr
          , (#ExprType.t as node) ->
            raise (Found node)
        | AbstractStateDeclExpr
          , (#AbstractStateDeclExprType.t as node) ->
            raise (Found node)
        | Allocator
          , (#AllocatorType.t as node) ->
            raise (Found node)
        | BaseAggregate
          , (#BaseAggregateType.t as node) ->
            raise (Found node)
        | Aggregate
          , (#AggregateType.t as node) ->
            raise (Found node)
        | BracketAggregate
          , (#BracketAggregateType.t as node) ->
            raise (Found node)
        | DeltaAggregate
          , (#DeltaAggregateType.t as node) ->
            raise (Found node)
        | BracketDeltaAggregate
          , (#BracketDeltaAggregateType.t as node) ->
            raise (Found node)
        | NullRecordAggregate
          , (#NullRecordAggregateType.t as node) ->
            raise (Found node)
        | BinOp
          , (#BinOpType.t as node) ->
            raise (Found node)
        | RelationOp
          , (#RelationOpType.t as node) ->
            raise (Found node)
        | BoxExpr
          , (#BoxExprType.t as node) ->
            raise (Found node)
        | CaseExprAlternative
          , (#CaseExprAlternativeType.t as node) ->
            raise (Found node)
        | CondExpr
          , (#CondExprType.t as node) ->
            raise (Found node)
        | CaseExpr
          , (#CaseExprType.t as node) ->
            raise (Found node)
        | IfExpr
          , (#IfExprType.t as node) ->
            raise (Found node)
        | ContractCases
          , (#ContractCasesType.t as node) ->
            raise (Found node)
        | DeclExpr
          , (#DeclExprType.t as node) ->
            raise (Found node)
        | MembershipExpr
          , (#MembershipExprType.t as node) ->
            raise (Found node)
        | Name
          , (#NameType.t as node) ->
            raise (Found node)
        | AttributeRef
          , (#AttributeRefType.t as node) ->
            raise (Found node)
        | UpdateAttributeRef
          , (#UpdateAttributeRefType.t as node) ->
            raise (Found node)
        | CallExpr
          , (#CallExprType.t as node) ->
            raise (Found node)
        | DefiningName
          , (#DefiningNameType.t as node) ->
            raise (Found node)
        | DiscreteSubtypeName
          , (#DiscreteSubtypeNameType.t as node) ->
            raise (Found node)
        | DottedName
          , (#DottedNameType.t as node) ->
            raise (Found node)
        | EndName
          , (#EndNameType.t as node) ->
            raise (Found node)
        | ExplicitDeref
          , (#ExplicitDerefType.t as node) ->
            raise (Found node)
        | QualExpr
          , (#QualExprType.t as node) ->
            raise (Found node)
        | SingleTokNode
          , (#SingleTokNodeType.t as node) ->
            raise (Found node)
        | BaseId
          , (#BaseIdType.t as node) ->
            raise (Found node)
        | CharLiteral
          , (#CharLiteralType.t as node) ->
            raise (Found node)
        | Identifier
          , (#IdentifierType.t as node) ->
            raise (Found node)
        | Op
          , (#OpType.t as node) ->
            raise (Found node)
        | OpAbs
          , (#OpAbsType.t as node) ->
            raise (Found node)
        | OpAnd
          , (#OpAndType.t as node) ->
            raise (Found node)
        | OpAndThen
          , (#OpAndThenType.t as node) ->
            raise (Found node)
        | OpConcat
          , (#OpConcatType.t as node) ->
            raise (Found node)
        | OpDiv
          , (#OpDivType.t as node) ->
            raise (Found node)
        | OpDoubleDot
          , (#OpDoubleDotType.t as node) ->
            raise (Found node)
        | OpEq
          , (#OpEqType.t as node) ->
            raise (Found node)
        | OpGt
          , (#OpGtType.t as node) ->
            raise (Found node)
        | OpGte
          , (#OpGteType.t as node) ->
            raise (Found node)
        | OpIn
          , (#OpInType.t as node) ->
            raise (Found node)
        | OpLt
          , (#OpLtType.t as node) ->
            raise (Found node)
        | OpLte
          , (#OpLteType.t as node) ->
            raise (Found node)
        | OpMinus
          , (#OpMinusType.t as node) ->
            raise (Found node)
        | OpMod
          , (#OpModType.t as node) ->
            raise (Found node)
        | OpMult
          , (#OpMultType.t as node) ->
            raise (Found node)
        | OpNeq
          , (#OpNeqType.t as node) ->
            raise (Found node)
        | OpNot
          , (#OpNotType.t as node) ->
            raise (Found node)
        | OpNotIn
          , (#OpNotInType.t as node) ->
            raise (Found node)
        | OpOr
          , (#OpOrType.t as node) ->
            raise (Found node)
        | OpOrElse
          , (#OpOrElseType.t as node) ->
            raise (Found node)
        | OpPlus
          , (#OpPlusType.t as node) ->
            raise (Found node)
        | OpPow
          , (#OpPowType.t as node) ->
            raise (Found node)
        | OpRem
          , (#OpRemType.t as node) ->
            raise (Found node)
        | OpXor
          , (#OpXorType.t as node) ->
            raise (Found node)
        | StringLiteral
          , (#StringLiteralType.t as node) ->
            raise (Found node)
        | NullLiteral
          , (#NullLiteralType.t as node) ->
            raise (Found node)
        | NumLiteral
          , (#NumLiteralType.t as node) ->
            raise (Found node)
        | IntLiteral
          , (#IntLiteralType.t as node) ->
            raise (Found node)
        | RealLiteral
          , (#RealLiteralType.t as node) ->
            raise (Found node)
        | TargetName
          , (#TargetNameType.t as node) ->
            raise (Found node)
        | ParenExpr
          , (#ParenExprType.t as node) ->
            raise (Found node)
        | QuantifiedExpr
          , (#QuantifiedExprType.t as node) ->
            raise (Found node)
        | RaiseExpr
          , (#RaiseExprType.t as node) ->
            raise (Found node)
        | UnOp
          , (#UnOpType.t as node) ->
            raise (Found node)
        | HandledStmts
          , (#HandledStmtsType.t as node) ->
            raise (Found node)
        | InterfaceKind
          , (#InterfaceKindType.t as node) ->
            raise (Found node)
        | InterfaceKindLimited
          , (#InterfaceKindLimitedType.t as node) ->
            raise (Found node)
        | InterfaceKindProtected
          , (#InterfaceKindProtectedType.t as node) ->
            raise (Found node)
        | InterfaceKindSynchronized
          , (#InterfaceKindSynchronizedType.t as node) ->
            raise (Found node)
        | InterfaceKindTask
          , (#InterfaceKindTaskType.t as node) ->
            raise (Found node)
        | IterType
          , (#IterTypeType.t as node) ->
            raise (Found node)
        | IterTypeIn
          , (#IterTypeInType.t as node) ->
            raise (Found node)
        | IterTypeOf
          , (#IterTypeOfType.t as node) ->
            raise (Found node)
        | LibraryItem
          , (#LibraryItemType.t as node) ->
            raise (Found node)
        | LimitedNode
          , (#LimitedNodeType.t as node) ->
            raise (Found node)
        | LimitedAbsent
          , (#LimitedAbsentType.t as node) ->
            raise (Found node)
        | LimitedPresent
          , (#LimitedPresentType.t as node) ->
            raise (Found node)
        | LoopSpec
          , (#LoopSpecType.t as node) ->
            raise (Found node)
        | ForLoopSpec
          , (#ForLoopSpecType.t as node) ->
            raise (Found node)
        | WhileLoopSpec
          , (#WhileLoopSpecType.t as node) ->
            raise (Found node)
        | Mode
          , (#ModeType.t as node) ->
            raise (Found node)
        | ModeDefault
          , (#ModeDefaultType.t as node) ->
            raise (Found node)
        | ModeIn
          , (#ModeInType.t as node) ->
            raise (Found node)
        | ModeInOut
          , (#ModeInOutType.t as node) ->
            raise (Found node)
        | ModeOut
          , (#ModeOutType.t as node) ->
            raise (Found node)
        | MultiAbstractStateDecl
          , (#MultiAbstractStateDeclType.t as node) ->
            raise (Found node)
        | NotNull
          , (#NotNullType.t as node) ->
            raise (Found node)
        | NotNullAbsent
          , (#NotNullAbsentType.t as node) ->
            raise (Found node)
        | NotNullPresent
          , (#NotNullPresentType.t as node) ->
            raise (Found node)
        | NullComponentDecl
          , (#NullComponentDeclType.t as node) ->
            raise (Found node)
        | OthersDesignator
          , (#OthersDesignatorType.t as node) ->
            raise (Found node)
        | OverridingNode
          , (#OverridingNodeType.t as node) ->
            raise (Found node)
        | OverridingNotOverriding
          , (#OverridingNotOverridingType.t as node) ->
            raise (Found node)
        | OverridingOverriding
          , (#OverridingOverridingType.t as node) ->
            raise (Found node)
        | OverridingUnspecified
          , (#OverridingUnspecifiedType.t as node) ->
            raise (Found node)
        | Params
          , (#ParamsType.t as node) ->
            raise (Found node)
        | ParenAbstractStateDecl
          , (#ParenAbstractStateDeclType.t as node) ->
            raise (Found node)
        | PragmaNode
          , (#PragmaNodeType.t as node) ->
            raise (Found node)
        | PrivateNode
          , (#PrivateNodeType.t as node) ->
            raise (Found node)
        | PrivateAbsent
          , (#PrivateAbsentType.t as node) ->
            raise (Found node)
        | PrivatePresent
          , (#PrivatePresentType.t as node) ->
            raise (Found node)
        | ProtectedDef
          , (#ProtectedDefType.t as node) ->
            raise (Found node)
        | ProtectedNode
          , (#ProtectedNodeType.t as node) ->
            raise (Found node)
        | ProtectedAbsent
          , (#ProtectedAbsentType.t as node) ->
            raise (Found node)
        | ProtectedPresent
          , (#ProtectedPresentType.t as node) ->
            raise (Found node)
        | Quantifier
          , (#QuantifierType.t as node) ->
            raise (Found node)
        | QuantifierAll
          , (#QuantifierAllType.t as node) ->
            raise (Found node)
        | QuantifierSome
          , (#QuantifierSomeType.t as node) ->
            raise (Found node)
        | RangeSpec
          , (#RangeSpecType.t as node) ->
            raise (Found node)
        | RenamingClause
          , (#RenamingClauseType.t as node) ->
            raise (Found node)
        | SyntheticRenamingClause
          , (#SyntheticRenamingClauseType.t as node) ->
            raise (Found node)
        | ReverseNode
          , (#ReverseNodeType.t as node) ->
            raise (Found node)
        | ReverseAbsent
          , (#ReverseAbsentType.t as node) ->
            raise (Found node)
        | ReversePresent
          , (#ReversePresentType.t as node) ->
            raise (Found node)
        | SelectWhenPart
          , (#SelectWhenPartType.t as node) ->
            raise (Found node)
        | Stmt
          , (#StmtType.t as node) ->
            raise (Found node)
        | CompositeStmt
          , (#CompositeStmtType.t as node) ->
            raise (Found node)
        | AcceptStmt
          , (#AcceptStmtType.t as node) ->
            raise (Found node)
        | AcceptStmtWithStmts
          , (#AcceptStmtWithStmtsType.t as node) ->
            raise (Found node)
        | BaseLoopStmt
          , (#BaseLoopStmtType.t as node) ->
            raise (Found node)
        | ForLoopStmt
          , (#ForLoopStmtType.t as node) ->
            raise (Found node)
        | LoopStmt
          , (#LoopStmtType.t as node) ->
            raise (Found node)
        | WhileLoopStmt
          , (#WhileLoopStmtType.t as node) ->
            raise (Found node)
        | BlockStmt
          , (#BlockStmtType.t as node) ->
            raise (Found node)
        | BeginBlock
          , (#BeginBlockType.t as node) ->
            raise (Found node)
        | DeclBlock
          , (#DeclBlockType.t as node) ->
            raise (Found node)
        | CaseStmt
          , (#CaseStmtType.t as node) ->
            raise (Found node)
        | ExtendedReturnStmt
          , (#ExtendedReturnStmtType.t as node) ->
            raise (Found node)
        | IfStmt
          , (#IfStmtType.t as node) ->
            raise (Found node)
        | NamedStmt
          , (#NamedStmtType.t as node) ->
            raise (Found node)
        | SelectStmt
          , (#SelectStmtType.t as node) ->
            raise (Found node)
        | ErrorStmt
          , (#ErrorStmtType.t as node) ->
            raise (Found node)
        | SimpleStmt
          , (#SimpleStmtType.t as node) ->
            raise (Found node)
        | AbortStmt
          , (#AbortStmtType.t as node) ->
            raise (Found node)
        | AssignStmt
          , (#AssignStmtType.t as node) ->
            raise (Found node)
        | CallStmt
          , (#CallStmtType.t as node) ->
            raise (Found node)
        | DelayStmt
          , (#DelayStmtType.t as node) ->
            raise (Found node)
        | ExitStmt
          , (#ExitStmtType.t as node) ->
            raise (Found node)
        | GotoStmt
          , (#GotoStmtType.t as node) ->
            raise (Found node)
        | Label
          , (#LabelType.t as node) ->
            raise (Found node)
        | NullStmt
          , (#NullStmtType.t as node) ->
            raise (Found node)
        | RaiseStmt
          , (#RaiseStmtType.t as node) ->
            raise (Found node)
        | RequeueStmt
          , (#RequeueStmtType.t as node) ->
            raise (Found node)
        | ReturnStmt
          , (#ReturnStmtType.t as node) ->
            raise (Found node)
        | TerminateAlternative
          , (#TerminateAlternativeType.t as node) ->
            raise (Found node)
        | SubpKind
          , (#SubpKindType.t as node) ->
            raise (Found node)
        | SubpKindFunction
          , (#SubpKindFunctionType.t as node) ->
            raise (Found node)
        | SubpKindProcedure
          , (#SubpKindProcedureType.t as node) ->
            raise (Found node)
        | Subunit
          , (#SubunitType.t as node) ->
            raise (Found node)
        | SynchronizedNode
          , (#SynchronizedNodeType.t as node) ->
            raise (Found node)
        | SynchronizedAbsent
          , (#SynchronizedAbsentType.t as node) ->
            raise (Found node)
        | SynchronizedPresent
          , (#SynchronizedPresentType.t as node) ->
            raise (Found node)
        | TaggedNode
          , (#TaggedNodeType.t as node) ->
            raise (Found node)
        | TaggedAbsent
          , (#TaggedAbsentType.t as node) ->
            raise (Found node)
        | TaggedPresent
          , (#TaggedPresentType.t as node) ->
            raise (Found node)
        | TaskDef
          , (#TaskDefType.t as node) ->
            raise (Found node)
        | TypeDef
          , (#TypeDefType.t as node) ->
            raise (Found node)
        | AccessDef
          , (#AccessDefType.t as node) ->
            raise (Found node)
        | AccessToSubpDef
          , (#AccessToSubpDefType.t as node) ->
            raise (Found node)
        | BaseTypeAccessDef
          , (#BaseTypeAccessDefType.t as node) ->
            raise (Found node)
        | AnonymousTypeAccessDef
          , (#AnonymousTypeAccessDefType.t as node) ->
            raise (Found node)
        | TypeAccessDef
          , (#TypeAccessDefType.t as node) ->
            raise (Found node)
        | ArrayTypeDef
          , (#ArrayTypeDefType.t as node) ->
            raise (Found node)
        | DerivedTypeDef
          , (#DerivedTypeDefType.t as node) ->
            raise (Found node)
        | EnumTypeDef
          , (#EnumTypeDefType.t as node) ->
            raise (Found node)
        | FormalDiscreteTypeDef
          , (#FormalDiscreteTypeDefType.t as node) ->
            raise (Found node)
        | InterfaceTypeDef
          , (#InterfaceTypeDefType.t as node) ->
            raise (Found node)
        | ModIntTypeDef
          , (#ModIntTypeDefType.t as node) ->
            raise (Found node)
        | PrivateTypeDef
          , (#PrivateTypeDefType.t as node) ->
            raise (Found node)
        | RealTypeDef
          , (#RealTypeDefType.t as node) ->
            raise (Found node)
        | DecimalFixedPointDef
          , (#DecimalFixedPointDefType.t as node) ->
            raise (Found node)
        | FloatingPointDef
          , (#FloatingPointDefType.t as node) ->
            raise (Found node)
        | OrdinaryFixedPointDef
          , (#OrdinaryFixedPointDefType.t as node) ->
            raise (Found node)
        | RecordTypeDef
          , (#RecordTypeDefType.t as node) ->
            raise (Found node)
        | SignedIntTypeDef
          , (#SignedIntTypeDefType.t as node) ->
            raise (Found node)
        | TypeExpr
          , (#TypeExprType.t as node) ->
            raise (Found node)
        | AnonymousType
          , (#AnonymousTypeType.t as node) ->
            raise (Found node)
        | EnumLitSynthTypeExpr
          , (#EnumLitSynthTypeExprType.t as node) ->
            raise (Found node)
        | SubtypeIndication
          , (#SubtypeIndicationType.t as node) ->
            raise (Found node)
        | ConstrainedSubtypeIndication
          , (#ConstrainedSubtypeIndicationType.t as node) ->
            raise (Found node)
        | DiscreteSubtypeIndication
          , (#DiscreteSubtypeIndicationType.t as node) ->
            raise (Found node)
        | UnconstrainedArrayIndex
          , (#UnconstrainedArrayIndexType.t as node) ->
            raise (Found node)
        | UntilNode
          , (#UntilNodeType.t as node) ->
            raise (Found node)
        | UntilAbsent
          , (#UntilAbsentType.t as node) ->
            raise (Found node)
        | UntilPresent
          , (#UntilPresentType.t as node) ->
            raise (Found node)
        | UseClause
          , (#UseClauseType.t as node) ->
            raise (Found node)
        | UsePackageClause
          , (#UsePackageClauseType.t as node) ->
            raise (Found node)
        | UseTypeClause
          , (#UseTypeClauseType.t as node) ->
            raise (Found node)
        | Variant
          , (#VariantType.t as node) ->
            raise (Found node)
        | VariantPart
          , (#VariantPartType.t as node) ->
            raise (Found node)
        | WithClause
          , (#WithClauseType.t as node) ->
            raise (Found node)
        | WithPrivate
          , (#WithPrivateType.t as node) ->
            raise (Found node)
        | WithPrivateAbsent
          , (#WithPrivateAbsentType.t as node) ->
            raise (Found node)
        | WithPrivatePresent
          , (#WithPrivatePresentType.t as node) ->
            raise (Found node)
        | _ ->
          ()
      in
      try
        iter aux node;
        raise Not_found
      with (Found node) -> node



  let findall : type a. a node ->  [< AdaNodeType.t ] -> a list =
    fun node_type node ->
      let aux : a list -> [< AdaNodeType.t ] -> a list =
       fun acc node ->
        match node_type, node with
        | AdaNode
          , (#AdaNodeType.t as node) ->
            node :: acc
        | AbortNode
          , (#AbortNodeType.t as node) ->
            node :: acc
        | AbortAbsent
          , (#AbortAbsentType.t as node) ->
            node :: acc
        | AbortPresent
          , (#AbortPresentType.t as node) ->
            node :: acc
        | AbstractNode
          , (#AbstractNodeType.t as node) ->
            node :: acc
        | AbstractAbsent
          , (#AbstractAbsentType.t as node) ->
            node :: acc
        | AbstractPresent
          , (#AbstractPresentType.t as node) ->
            node :: acc
        | AdaList
          , (#AdaListType.t as node) ->
            node :: acc
        | AdaNodeList
          , (#AdaNodeListType.t as node) ->
            node :: acc
        | AbstractStateDeclList
          , (#AbstractStateDeclListType.t as node) ->
            node :: acc
        | AlternativesList
          , (#AlternativesListType.t as node) ->
            node :: acc
        | ConstraintList
          , (#ConstraintListType.t as node) ->
            node :: acc
        | DeclList
          , (#DeclListType.t as node) ->
            node :: acc
        | StmtList
          , (#StmtListType.t as node) ->
            node :: acc
        | AspectAssocList
          , (#AspectAssocListType.t as node) ->
            node :: acc
        | BaseAssocList
          , (#BaseAssocListType.t as node) ->
            node :: acc
        | BasicAssocList
          , (#BasicAssocListType.t as node) ->
            node :: acc
        | AssocList
          , (#AssocListType.t as node) ->
            node :: acc
        | BasicDeclList
          , (#BasicDeclListType.t as node) ->
            node :: acc
        | CaseExprAlternativeList
          , (#CaseExprAlternativeListType.t as node) ->
            node :: acc
        | CaseStmtAlternativeList
          , (#CaseStmtAlternativeListType.t as node) ->
            node :: acc
        | CompilationUnitList
          , (#CompilationUnitListType.t as node) ->
            node :: acc
        | ContractCaseAssocList
          , (#ContractCaseAssocListType.t as node) ->
            node :: acc
        | DefiningNameList
          , (#DefiningNameListType.t as node) ->
            node :: acc
        | DiscriminantSpecList
          , (#DiscriminantSpecListType.t as node) ->
            node :: acc
        | ElsifExprPartList
          , (#ElsifExprPartListType.t as node) ->
            node :: acc
        | ElsifStmtPartList
          , (#ElsifStmtPartListType.t as node) ->
            node :: acc
        | EnumLiteralDeclList
          , (#EnumLiteralDeclListType.t as node) ->
            node :: acc
        | ExprList
          , (#ExprListType.t as node) ->
            node :: acc
        | ExprAlternativesList
          , (#ExprAlternativesListType.t as node) ->
            node :: acc
        | IdentifierList
          , (#IdentifierListType.t as node) ->
            node :: acc
        | DiscriminantChoiceList
          , (#DiscriminantChoiceListType.t as node) ->
            node :: acc
        | NameList
          , (#NameListType.t as node) ->
            node :: acc
        | ParentList
          , (#ParentListType.t as node) ->
            node :: acc
        | ParamSpecList
          , (#ParamSpecListType.t as node) ->
            node :: acc
        | PragmaNodeList
          , (#PragmaNodeListType.t as node) ->
            node :: acc
        | SelectWhenPartList
          , (#SelectWhenPartListType.t as node) ->
            node :: acc
        | UnconstrainedArrayIndexList
          , (#UnconstrainedArrayIndexListType.t as node) ->
            node :: acc
        | VariantList
          , (#VariantListType.t as node) ->
            node :: acc
        | AliasedNode
          , (#AliasedNodeType.t as node) ->
            node :: acc
        | AliasedAbsent
          , (#AliasedAbsentType.t as node) ->
            node :: acc
        | AliasedPresent
          , (#AliasedPresentType.t as node) ->
            node :: acc
        | AllNode
          , (#AllNodeType.t as node) ->
            node :: acc
        | AllAbsent
          , (#AllAbsentType.t as node) ->
            node :: acc
        | AllPresent
          , (#AllPresentType.t as node) ->
            node :: acc
        | ArrayIndices
          , (#ArrayIndicesType.t as node) ->
            node :: acc
        | ConstrainedArrayIndices
          , (#ConstrainedArrayIndicesType.t as node) ->
            node :: acc
        | UnconstrainedArrayIndices
          , (#UnconstrainedArrayIndicesType.t as node) ->
            node :: acc
        | AspectAssoc
          , (#AspectAssocType.t as node) ->
            node :: acc
        | AspectClause
          , (#AspectClauseType.t as node) ->
            node :: acc
        | AtClause
          , (#AtClauseType.t as node) ->
            node :: acc
        | AttributeDefClause
          , (#AttributeDefClauseType.t as node) ->
            node :: acc
        | EnumRepClause
          , (#EnumRepClauseType.t as node) ->
            node :: acc
        | RecordRepClause
          , (#RecordRepClauseType.t as node) ->
            node :: acc
        | AspectSpec
          , (#AspectSpecType.t as node) ->
            node :: acc
        | BaseAssoc
          , (#BaseAssocType.t as node) ->
            node :: acc
        | ContractCaseAssoc
          , (#ContractCaseAssocType.t as node) ->
            node :: acc
        | PragmaArgumentAssoc
          , (#PragmaArgumentAssocType.t as node) ->
            node :: acc
        | BaseFormalParamHolder
          , (#BaseFormalParamHolderType.t as node) ->
            node :: acc
        | BaseSubpSpec
          , (#BaseSubpSpecType.t as node) ->
            node :: acc
        | EntrySpec
          , (#EntrySpecType.t as node) ->
            node :: acc
        | EnumSubpSpec
          , (#EnumSubpSpecType.t as node) ->
            node :: acc
        | SubpSpec
          , (#SubpSpecType.t as node) ->
            node :: acc
        | ComponentList
          , (#ComponentListType.t as node) ->
            node :: acc
        | DiscriminantPart
          , (#DiscriminantPartType.t as node) ->
            node :: acc
        | KnownDiscriminantPart
          , (#KnownDiscriminantPartType.t as node) ->
            node :: acc
        | UnknownDiscriminantPart
          , (#UnknownDiscriminantPartType.t as node) ->
            node :: acc
        | EntryCompletionFormalParams
          , (#EntryCompletionFormalParamsType.t as node) ->
            node :: acc
        | GenericFormalPart
          , (#GenericFormalPartType.t as node) ->
            node :: acc
        | BaseRecordDef
          , (#BaseRecordDefType.t as node) ->
            node :: acc
        | NullRecordDef
          , (#NullRecordDefType.t as node) ->
            node :: acc
        | RecordDef
          , (#RecordDefType.t as node) ->
            node :: acc
        | BasicAssoc
          , (#BasicAssocType.t as node) ->
            node :: acc
        | AggregateAssoc
          , (#AggregateAssocType.t as node) ->
            node :: acc
        | MultiDimArrayAssoc
          , (#MultiDimArrayAssocType.t as node) ->
            node :: acc
        | DiscriminantAssoc
          , (#DiscriminantAssocType.t as node) ->
            node :: acc
        | IteratedAssoc
          , (#IteratedAssocType.t as node) ->
            node :: acc
        | ParamAssoc
          , (#ParamAssocType.t as node) ->
            node :: acc
        | BasicDecl
          , (#BasicDeclType.t as node) ->
            node :: acc
        | AbstractStateDecl
          , (#AbstractStateDeclType.t as node) ->
            node :: acc
        | AnonymousExprDecl
          , (#AnonymousExprDeclType.t as node) ->
            node :: acc
        | BaseFormalParamDecl
          , (#BaseFormalParamDeclType.t as node) ->
            node :: acc
        | ComponentDecl
          , (#ComponentDeclType.t as node) ->
            node :: acc
        | DiscriminantSpec
          , (#DiscriminantSpecType.t as node) ->
            node :: acc
        | GenericFormal
          , (#GenericFormalType.t as node) ->
            node :: acc
        | GenericFormalObjDecl
          , (#GenericFormalObjDeclType.t as node) ->
            node :: acc
        | GenericFormalPackage
          , (#GenericFormalPackageType.t as node) ->
            node :: acc
        | GenericFormalSubpDecl
          , (#GenericFormalSubpDeclType.t as node) ->
            node :: acc
        | GenericFormalTypeDecl
          , (#GenericFormalTypeDeclType.t as node) ->
            node :: acc
        | ParamSpec
          , (#ParamSpecType.t as node) ->
            node :: acc
        | BasePackageDecl
          , (#BasePackageDeclType.t as node) ->
            node :: acc
        | GenericPackageInternal
          , (#GenericPackageInternalType.t as node) ->
            node :: acc
        | PackageDecl
          , (#PackageDeclType.t as node) ->
            node :: acc
        | BaseTypeDecl
          , (#BaseTypeDeclType.t as node) ->
            node :: acc
        | BaseSubtypeDecl
          , (#BaseSubtypeDeclType.t as node) ->
            node :: acc
        | DiscreteBaseSubtypeDecl
          , (#DiscreteBaseSubtypeDeclType.t as node) ->
            node :: acc
        | SubtypeDecl
          , (#SubtypeDeclType.t as node) ->
            node :: acc
        | ClasswideTypeDecl
          , (#ClasswideTypeDeclType.t as node) ->
            node :: acc
        | IncompleteTypeDecl
          , (#IncompleteTypeDeclType.t as node) ->
            node :: acc
        | IncompleteTaggedTypeDecl
          , (#IncompleteTaggedTypeDeclType.t as node) ->
            node :: acc
        | ProtectedTypeDecl
          , (#ProtectedTypeDeclType.t as node) ->
            node :: acc
        | TaskTypeDecl
          , (#TaskTypeDeclType.t as node) ->
            node :: acc
        | SingleTaskTypeDecl
          , (#SingleTaskTypeDeclType.t as node) ->
            node :: acc
        | TypeDecl
          , (#TypeDeclType.t as node) ->
            node :: acc
        | AnonymousTypeDecl
          , (#AnonymousTypeDeclType.t as node) ->
            node :: acc
        | SynthAnonymousTypeDecl
          , (#SynthAnonymousTypeDeclType.t as node) ->
            node :: acc
        | BasicSubpDecl
          , (#BasicSubpDeclType.t as node) ->
            node :: acc
        | ClassicSubpDecl
          , (#ClassicSubpDeclType.t as node) ->
            node :: acc
        | AbstractSubpDecl
          , (#AbstractSubpDeclType.t as node) ->
            node :: acc
        | FormalSubpDecl
          , (#FormalSubpDeclType.t as node) ->
            node :: acc
        | AbstractFormalSubpDecl
          , (#AbstractFormalSubpDeclType.t as node) ->
            node :: acc
        | ConcreteFormalSubpDecl
          , (#ConcreteFormalSubpDeclType.t as node) ->
            node :: acc
        | SubpDecl
          , (#SubpDeclType.t as node) ->
            node :: acc
        | EntryDecl
          , (#EntryDeclType.t as node) ->
            node :: acc
        | EnumLiteralDecl
          , (#EnumLiteralDeclType.t as node) ->
            node :: acc
        | GenericSubpInternal
          , (#GenericSubpInternalType.t as node) ->
            node :: acc
        | BodyNode
          , (#BodyNodeType.t as node) ->
            node :: acc
        | BaseSubpBody
          , (#BaseSubpBodyType.t as node) ->
            node :: acc
        | ExprFunction
          , (#ExprFunctionType.t as node) ->
            node :: acc
        | NullSubpDecl
          , (#NullSubpDeclType.t as node) ->
            node :: acc
        | SubpBody
          , (#SubpBodyType.t as node) ->
            node :: acc
        | SubpRenamingDecl
          , (#SubpRenamingDeclType.t as node) ->
            node :: acc
        | BodyStub
          , (#BodyStubType.t as node) ->
            node :: acc
        | PackageBodyStub
          , (#PackageBodyStubType.t as node) ->
            node :: acc
        | ProtectedBodyStub
          , (#ProtectedBodyStubType.t as node) ->
            node :: acc
        | SubpBodyStub
          , (#SubpBodyStubType.t as node) ->
            node :: acc
        | TaskBodyStub
          , (#TaskBodyStubType.t as node) ->
            node :: acc
        | EntryBody
          , (#EntryBodyType.t as node) ->
            node :: acc
        | PackageBody
          , (#PackageBodyType.t as node) ->
            node :: acc
        | ProtectedBody
          , (#ProtectedBodyType.t as node) ->
            node :: acc
        | TaskBody
          , (#TaskBodyType.t as node) ->
            node :: acc
        | EntryIndexSpec
          , (#EntryIndexSpecType.t as node) ->
            node :: acc
        | ErrorDecl
          , (#ErrorDeclType.t as node) ->
            node :: acc
        | ExceptionDecl
          , (#ExceptionDeclType.t as node) ->
            node :: acc
        | ExceptionHandler
          , (#ExceptionHandlerType.t as node) ->
            node :: acc
        | ForLoopVarDecl
          , (#ForLoopVarDeclType.t as node) ->
            node :: acc
        | GenericDecl
          , (#GenericDeclType.t as node) ->
            node :: acc
        | GenericPackageDecl
          , (#GenericPackageDeclType.t as node) ->
            node :: acc
        | GenericSubpDecl
          , (#GenericSubpDeclType.t as node) ->
            node :: acc
        | GenericInstantiation
          , (#GenericInstantiationType.t as node) ->
            node :: acc
        | GenericPackageInstantiation
          , (#GenericPackageInstantiationType.t as node) ->
            node :: acc
        | GenericSubpInstantiation
          , (#GenericSubpInstantiationType.t as node) ->
            node :: acc
        | GenericRenamingDecl
          , (#GenericRenamingDeclType.t as node) ->
            node :: acc
        | GenericPackageRenamingDecl
          , (#GenericPackageRenamingDeclType.t as node) ->
            node :: acc
        | GenericSubpRenamingDecl
          , (#GenericSubpRenamingDeclType.t as node) ->
            node :: acc
        | LabelDecl
          , (#LabelDeclType.t as node) ->
            node :: acc
        | NamedStmtDecl
          , (#NamedStmtDeclType.t as node) ->
            node :: acc
        | NumberDecl
          , (#NumberDeclType.t as node) ->
            node :: acc
        | ObjectDecl
          , (#ObjectDeclType.t as node) ->
            node :: acc
        | ExtendedReturnStmtObjectDecl
          , (#ExtendedReturnStmtObjectDeclType.t as node) ->
            node :: acc
        | PackageRenamingDecl
          , (#PackageRenamingDeclType.t as node) ->
            node :: acc
        | SingleProtectedDecl
          , (#SingleProtectedDeclType.t as node) ->
            node :: acc
        | SingleTaskDecl
          , (#SingleTaskDeclType.t as node) ->
            node :: acc
        | CaseStmtAlternative
          , (#CaseStmtAlternativeType.t as node) ->
            node :: acc
        | CompilationUnit
          , (#CompilationUnitType.t as node) ->
            node :: acc
        | ComponentClause
          , (#ComponentClauseType.t as node) ->
            node :: acc
        | ComponentDef
          , (#ComponentDefType.t as node) ->
            node :: acc
        | ConstantNode
          , (#ConstantNodeType.t as node) ->
            node :: acc
        | ConstantAbsent
          , (#ConstantAbsentType.t as node) ->
            node :: acc
        | ConstantPresent
          , (#ConstantPresentType.t as node) ->
            node :: acc
        | Constraint
          , (#ConstraintType.t as node) ->
            node :: acc
        | DeltaConstraint
          , (#DeltaConstraintType.t as node) ->
            node :: acc
        | DigitsConstraint
          , (#DigitsConstraintType.t as node) ->
            node :: acc
        | DiscriminantConstraint
          , (#DiscriminantConstraintType.t as node) ->
            node :: acc
        | IndexConstraint
          , (#IndexConstraintType.t as node) ->
            node :: acc
        | RangeConstraint
          , (#RangeConstraintType.t as node) ->
            node :: acc
        | DeclarativePart
          , (#DeclarativePartType.t as node) ->
            node :: acc
        | PrivatePart
          , (#PrivatePartType.t as node) ->
            node :: acc
        | PublicPart
          , (#PublicPartType.t as node) ->
            node :: acc
        | ElsifExprPart
          , (#ElsifExprPartType.t as node) ->
            node :: acc
        | ElsifStmtPart
          , (#ElsifStmtPartType.t as node) ->
            node :: acc
        | Expr
          , (#ExprType.t as node) ->
            node :: acc
        | AbstractStateDeclExpr
          , (#AbstractStateDeclExprType.t as node) ->
            node :: acc
        | Allocator
          , (#AllocatorType.t as node) ->
            node :: acc
        | BaseAggregate
          , (#BaseAggregateType.t as node) ->
            node :: acc
        | Aggregate
          , (#AggregateType.t as node) ->
            node :: acc
        | BracketAggregate
          , (#BracketAggregateType.t as node) ->
            node :: acc
        | DeltaAggregate
          , (#DeltaAggregateType.t as node) ->
            node :: acc
        | BracketDeltaAggregate
          , (#BracketDeltaAggregateType.t as node) ->
            node :: acc
        | NullRecordAggregate
          , (#NullRecordAggregateType.t as node) ->
            node :: acc
        | BinOp
          , (#BinOpType.t as node) ->
            node :: acc
        | RelationOp
          , (#RelationOpType.t as node) ->
            node :: acc
        | BoxExpr
          , (#BoxExprType.t as node) ->
            node :: acc
        | CaseExprAlternative
          , (#CaseExprAlternativeType.t as node) ->
            node :: acc
        | CondExpr
          , (#CondExprType.t as node) ->
            node :: acc
        | CaseExpr
          , (#CaseExprType.t as node) ->
            node :: acc
        | IfExpr
          , (#IfExprType.t as node) ->
            node :: acc
        | ContractCases
          , (#ContractCasesType.t as node) ->
            node :: acc
        | DeclExpr
          , (#DeclExprType.t as node) ->
            node :: acc
        | MembershipExpr
          , (#MembershipExprType.t as node) ->
            node :: acc
        | Name
          , (#NameType.t as node) ->
            node :: acc
        | AttributeRef
          , (#AttributeRefType.t as node) ->
            node :: acc
        | UpdateAttributeRef
          , (#UpdateAttributeRefType.t as node) ->
            node :: acc
        | CallExpr
          , (#CallExprType.t as node) ->
            node :: acc
        | DefiningName
          , (#DefiningNameType.t as node) ->
            node :: acc
        | DiscreteSubtypeName
          , (#DiscreteSubtypeNameType.t as node) ->
            node :: acc
        | DottedName
          , (#DottedNameType.t as node) ->
            node :: acc
        | EndName
          , (#EndNameType.t as node) ->
            node :: acc
        | ExplicitDeref
          , (#ExplicitDerefType.t as node) ->
            node :: acc
        | QualExpr
          , (#QualExprType.t as node) ->
            node :: acc
        | SingleTokNode
          , (#SingleTokNodeType.t as node) ->
            node :: acc
        | BaseId
          , (#BaseIdType.t as node) ->
            node :: acc
        | CharLiteral
          , (#CharLiteralType.t as node) ->
            node :: acc
        | Identifier
          , (#IdentifierType.t as node) ->
            node :: acc
        | Op
          , (#OpType.t as node) ->
            node :: acc
        | OpAbs
          , (#OpAbsType.t as node) ->
            node :: acc
        | OpAnd
          , (#OpAndType.t as node) ->
            node :: acc
        | OpAndThen
          , (#OpAndThenType.t as node) ->
            node :: acc
        | OpConcat
          , (#OpConcatType.t as node) ->
            node :: acc
        | OpDiv
          , (#OpDivType.t as node) ->
            node :: acc
        | OpDoubleDot
          , (#OpDoubleDotType.t as node) ->
            node :: acc
        | OpEq
          , (#OpEqType.t as node) ->
            node :: acc
        | OpGt
          , (#OpGtType.t as node) ->
            node :: acc
        | OpGte
          , (#OpGteType.t as node) ->
            node :: acc
        | OpIn
          , (#OpInType.t as node) ->
            node :: acc
        | OpLt
          , (#OpLtType.t as node) ->
            node :: acc
        | OpLte
          , (#OpLteType.t as node) ->
            node :: acc
        | OpMinus
          , (#OpMinusType.t as node) ->
            node :: acc
        | OpMod
          , (#OpModType.t as node) ->
            node :: acc
        | OpMult
          , (#OpMultType.t as node) ->
            node :: acc
        | OpNeq
          , (#OpNeqType.t as node) ->
            node :: acc
        | OpNot
          , (#OpNotType.t as node) ->
            node :: acc
        | OpNotIn
          , (#OpNotInType.t as node) ->
            node :: acc
        | OpOr
          , (#OpOrType.t as node) ->
            node :: acc
        | OpOrElse
          , (#OpOrElseType.t as node) ->
            node :: acc
        | OpPlus
          , (#OpPlusType.t as node) ->
            node :: acc
        | OpPow
          , (#OpPowType.t as node) ->
            node :: acc
        | OpRem
          , (#OpRemType.t as node) ->
            node :: acc
        | OpXor
          , (#OpXorType.t as node) ->
            node :: acc
        | StringLiteral
          , (#StringLiteralType.t as node) ->
            node :: acc
        | NullLiteral
          , (#NullLiteralType.t as node) ->
            node :: acc
        | NumLiteral
          , (#NumLiteralType.t as node) ->
            node :: acc
        | IntLiteral
          , (#IntLiteralType.t as node) ->
            node :: acc
        | RealLiteral
          , (#RealLiteralType.t as node) ->
            node :: acc
        | TargetName
          , (#TargetNameType.t as node) ->
            node :: acc
        | ParenExpr
          , (#ParenExprType.t as node) ->
            node :: acc
        | QuantifiedExpr
          , (#QuantifiedExprType.t as node) ->
            node :: acc
        | RaiseExpr
          , (#RaiseExprType.t as node) ->
            node :: acc
        | UnOp
          , (#UnOpType.t as node) ->
            node :: acc
        | HandledStmts
          , (#HandledStmtsType.t as node) ->
            node :: acc
        | InterfaceKind
          , (#InterfaceKindType.t as node) ->
            node :: acc
        | InterfaceKindLimited
          , (#InterfaceKindLimitedType.t as node) ->
            node :: acc
        | InterfaceKindProtected
          , (#InterfaceKindProtectedType.t as node) ->
            node :: acc
        | InterfaceKindSynchronized
          , (#InterfaceKindSynchronizedType.t as node) ->
            node :: acc
        | InterfaceKindTask
          , (#InterfaceKindTaskType.t as node) ->
            node :: acc
        | IterType
          , (#IterTypeType.t as node) ->
            node :: acc
        | IterTypeIn
          , (#IterTypeInType.t as node) ->
            node :: acc
        | IterTypeOf
          , (#IterTypeOfType.t as node) ->
            node :: acc
        | LibraryItem
          , (#LibraryItemType.t as node) ->
            node :: acc
        | LimitedNode
          , (#LimitedNodeType.t as node) ->
            node :: acc
        | LimitedAbsent
          , (#LimitedAbsentType.t as node) ->
            node :: acc
        | LimitedPresent
          , (#LimitedPresentType.t as node) ->
            node :: acc
        | LoopSpec
          , (#LoopSpecType.t as node) ->
            node :: acc
        | ForLoopSpec
          , (#ForLoopSpecType.t as node) ->
            node :: acc
        | WhileLoopSpec
          , (#WhileLoopSpecType.t as node) ->
            node :: acc
        | Mode
          , (#ModeType.t as node) ->
            node :: acc
        | ModeDefault
          , (#ModeDefaultType.t as node) ->
            node :: acc
        | ModeIn
          , (#ModeInType.t as node) ->
            node :: acc
        | ModeInOut
          , (#ModeInOutType.t as node) ->
            node :: acc
        | ModeOut
          , (#ModeOutType.t as node) ->
            node :: acc
        | MultiAbstractStateDecl
          , (#MultiAbstractStateDeclType.t as node) ->
            node :: acc
        | NotNull
          , (#NotNullType.t as node) ->
            node :: acc
        | NotNullAbsent
          , (#NotNullAbsentType.t as node) ->
            node :: acc
        | NotNullPresent
          , (#NotNullPresentType.t as node) ->
            node :: acc
        | NullComponentDecl
          , (#NullComponentDeclType.t as node) ->
            node :: acc
        | OthersDesignator
          , (#OthersDesignatorType.t as node) ->
            node :: acc
        | OverridingNode
          , (#OverridingNodeType.t as node) ->
            node :: acc
        | OverridingNotOverriding
          , (#OverridingNotOverridingType.t as node) ->
            node :: acc
        | OverridingOverriding
          , (#OverridingOverridingType.t as node) ->
            node :: acc
        | OverridingUnspecified
          , (#OverridingUnspecifiedType.t as node) ->
            node :: acc
        | Params
          , (#ParamsType.t as node) ->
            node :: acc
        | ParenAbstractStateDecl
          , (#ParenAbstractStateDeclType.t as node) ->
            node :: acc
        | PragmaNode
          , (#PragmaNodeType.t as node) ->
            node :: acc
        | PrivateNode
          , (#PrivateNodeType.t as node) ->
            node :: acc
        | PrivateAbsent
          , (#PrivateAbsentType.t as node) ->
            node :: acc
        | PrivatePresent
          , (#PrivatePresentType.t as node) ->
            node :: acc
        | ProtectedDef
          , (#ProtectedDefType.t as node) ->
            node :: acc
        | ProtectedNode
          , (#ProtectedNodeType.t as node) ->
            node :: acc
        | ProtectedAbsent
          , (#ProtectedAbsentType.t as node) ->
            node :: acc
        | ProtectedPresent
          , (#ProtectedPresentType.t as node) ->
            node :: acc
        | Quantifier
          , (#QuantifierType.t as node) ->
            node :: acc
        | QuantifierAll
          , (#QuantifierAllType.t as node) ->
            node :: acc
        | QuantifierSome
          , (#QuantifierSomeType.t as node) ->
            node :: acc
        | RangeSpec
          , (#RangeSpecType.t as node) ->
            node :: acc
        | RenamingClause
          , (#RenamingClauseType.t as node) ->
            node :: acc
        | SyntheticRenamingClause
          , (#SyntheticRenamingClauseType.t as node) ->
            node :: acc
        | ReverseNode
          , (#ReverseNodeType.t as node) ->
            node :: acc
        | ReverseAbsent
          , (#ReverseAbsentType.t as node) ->
            node :: acc
        | ReversePresent
          , (#ReversePresentType.t as node) ->
            node :: acc
        | SelectWhenPart
          , (#SelectWhenPartType.t as node) ->
            node :: acc
        | Stmt
          , (#StmtType.t as node) ->
            node :: acc
        | CompositeStmt
          , (#CompositeStmtType.t as node) ->
            node :: acc
        | AcceptStmt
          , (#AcceptStmtType.t as node) ->
            node :: acc
        | AcceptStmtWithStmts
          , (#AcceptStmtWithStmtsType.t as node) ->
            node :: acc
        | BaseLoopStmt
          , (#BaseLoopStmtType.t as node) ->
            node :: acc
        | ForLoopStmt
          , (#ForLoopStmtType.t as node) ->
            node :: acc
        | LoopStmt
          , (#LoopStmtType.t as node) ->
            node :: acc
        | WhileLoopStmt
          , (#WhileLoopStmtType.t as node) ->
            node :: acc
        | BlockStmt
          , (#BlockStmtType.t as node) ->
            node :: acc
        | BeginBlock
          , (#BeginBlockType.t as node) ->
            node :: acc
        | DeclBlock
          , (#DeclBlockType.t as node) ->
            node :: acc
        | CaseStmt
          , (#CaseStmtType.t as node) ->
            node :: acc
        | ExtendedReturnStmt
          , (#ExtendedReturnStmtType.t as node) ->
            node :: acc
        | IfStmt
          , (#IfStmtType.t as node) ->
            node :: acc
        | NamedStmt
          , (#NamedStmtType.t as node) ->
            node :: acc
        | SelectStmt
          , (#SelectStmtType.t as node) ->
            node :: acc
        | ErrorStmt
          , (#ErrorStmtType.t as node) ->
            node :: acc
        | SimpleStmt
          , (#SimpleStmtType.t as node) ->
            node :: acc
        | AbortStmt
          , (#AbortStmtType.t as node) ->
            node :: acc
        | AssignStmt
          , (#AssignStmtType.t as node) ->
            node :: acc
        | CallStmt
          , (#CallStmtType.t as node) ->
            node :: acc
        | DelayStmt
          , (#DelayStmtType.t as node) ->
            node :: acc
        | ExitStmt
          , (#ExitStmtType.t as node) ->
            node :: acc
        | GotoStmt
          , (#GotoStmtType.t as node) ->
            node :: acc
        | Label
          , (#LabelType.t as node) ->
            node :: acc
        | NullStmt
          , (#NullStmtType.t as node) ->
            node :: acc
        | RaiseStmt
          , (#RaiseStmtType.t as node) ->
            node :: acc
        | RequeueStmt
          , (#RequeueStmtType.t as node) ->
            node :: acc
        | ReturnStmt
          , (#ReturnStmtType.t as node) ->
            node :: acc
        | TerminateAlternative
          , (#TerminateAlternativeType.t as node) ->
            node :: acc
        | SubpKind
          , (#SubpKindType.t as node) ->
            node :: acc
        | SubpKindFunction
          , (#SubpKindFunctionType.t as node) ->
            node :: acc
        | SubpKindProcedure
          , (#SubpKindProcedureType.t as node) ->
            node :: acc
        | Subunit
          , (#SubunitType.t as node) ->
            node :: acc
        | SynchronizedNode
          , (#SynchronizedNodeType.t as node) ->
            node :: acc
        | SynchronizedAbsent
          , (#SynchronizedAbsentType.t as node) ->
            node :: acc
        | SynchronizedPresent
          , (#SynchronizedPresentType.t as node) ->
            node :: acc
        | TaggedNode
          , (#TaggedNodeType.t as node) ->
            node :: acc
        | TaggedAbsent
          , (#TaggedAbsentType.t as node) ->
            node :: acc
        | TaggedPresent
          , (#TaggedPresentType.t as node) ->
            node :: acc
        | TaskDef
          , (#TaskDefType.t as node) ->
            node :: acc
        | TypeDef
          , (#TypeDefType.t as node) ->
            node :: acc
        | AccessDef
          , (#AccessDefType.t as node) ->
            node :: acc
        | AccessToSubpDef
          , (#AccessToSubpDefType.t as node) ->
            node :: acc
        | BaseTypeAccessDef
          , (#BaseTypeAccessDefType.t as node) ->
            node :: acc
        | AnonymousTypeAccessDef
          , (#AnonymousTypeAccessDefType.t as node) ->
            node :: acc
        | TypeAccessDef
          , (#TypeAccessDefType.t as node) ->
            node :: acc
        | ArrayTypeDef
          , (#ArrayTypeDefType.t as node) ->
            node :: acc
        | DerivedTypeDef
          , (#DerivedTypeDefType.t as node) ->
            node :: acc
        | EnumTypeDef
          , (#EnumTypeDefType.t as node) ->
            node :: acc
        | FormalDiscreteTypeDef
          , (#FormalDiscreteTypeDefType.t as node) ->
            node :: acc
        | InterfaceTypeDef
          , (#InterfaceTypeDefType.t as node) ->
            node :: acc
        | ModIntTypeDef
          , (#ModIntTypeDefType.t as node) ->
            node :: acc
        | PrivateTypeDef
          , (#PrivateTypeDefType.t as node) ->
            node :: acc
        | RealTypeDef
          , (#RealTypeDefType.t as node) ->
            node :: acc
        | DecimalFixedPointDef
          , (#DecimalFixedPointDefType.t as node) ->
            node :: acc
        | FloatingPointDef
          , (#FloatingPointDefType.t as node) ->
            node :: acc
        | OrdinaryFixedPointDef
          , (#OrdinaryFixedPointDefType.t as node) ->
            node :: acc
        | RecordTypeDef
          , (#RecordTypeDefType.t as node) ->
            node :: acc
        | SignedIntTypeDef
          , (#SignedIntTypeDefType.t as node) ->
            node :: acc
        | TypeExpr
          , (#TypeExprType.t as node) ->
            node :: acc
        | AnonymousType
          , (#AnonymousTypeType.t as node) ->
            node :: acc
        | EnumLitSynthTypeExpr
          , (#EnumLitSynthTypeExprType.t as node) ->
            node :: acc
        | SubtypeIndication
          , (#SubtypeIndicationType.t as node) ->
            node :: acc
        | ConstrainedSubtypeIndication
          , (#ConstrainedSubtypeIndicationType.t as node) ->
            node :: acc
        | DiscreteSubtypeIndication
          , (#DiscreteSubtypeIndicationType.t as node) ->
            node :: acc
        | UnconstrainedArrayIndex
          , (#UnconstrainedArrayIndexType.t as node) ->
            node :: acc
        | UntilNode
          , (#UntilNodeType.t as node) ->
            node :: acc
        | UntilAbsent
          , (#UntilAbsentType.t as node) ->
            node :: acc
        | UntilPresent
          , (#UntilPresentType.t as node) ->
            node :: acc
        | UseClause
          , (#UseClauseType.t as node) ->
            node :: acc
        | UsePackageClause
          , (#UsePackageClauseType.t as node) ->
            node :: acc
        | UseTypeClause
          , (#UseTypeClauseType.t as node) ->
            node :: acc
        | Variant
          , (#VariantType.t as node) ->
            node :: acc
        | VariantPart
          , (#VariantPartType.t as node) ->
            node :: acc
        | WithClause
          , (#WithClauseType.t as node) ->
            node :: acc
        | WithPrivate
          , (#WithPrivateType.t as node) ->
            node :: acc
        | WithPrivateAbsent
          , (#WithPrivateAbsentType.t as node) ->
            node :: acc
        | WithPrivatePresent
          , (#WithPrivatePresentType.t as node) ->
            node :: acc
        | _ ->
          acc
      in
      List.rev (fold aux [] node)

  let fields_with_names node =
    let aux i x =
      (Format.sprintf "item_%d" i), x
    in
    match (node :> AdaNodeType.t) with
    | `WithPrivatePresent value ->
        [
        ]
    | `WithPrivateAbsent value ->
        [
        ]
    | `WithClause value ->
        [
        (try
           ("has_limited"
           , Some (Lazy.force value.f_has_limited
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_limited", None) );
        (try
           ("has_private"
           , Some (Lazy.force value.f_has_private
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_private", None) );
        (try
           ("packages"
           , Some (Lazy.force value.f_packages
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("packages", None) );
        ]
    | `VariantPart value ->
        [
        (try
           ("discr_name"
           , Some (Lazy.force value.f_discr_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("discr_name", None) );
        (try
           ("variant"
           , Some (Lazy.force value.f_variant
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("variant", None) );
        ]
    | `Variant value ->
        [
        (try
           ("choices"
           , Some (Lazy.force value.f_choices
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("choices", None) );
        (try
           ("components"
           , Some (Lazy.force value.f_components
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("components", None) );
        ]
    | `UseTypeClause value ->
        [
        (try
           ("has_all"
           , Some (Lazy.force value.f_has_all
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_all", None) );
        (try
           ("types"
           , Some (Lazy.force value.f_types
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("types", None) );
        ]
    | `UsePackageClause value ->
        [
        (try
           ("packages"
           , Some (Lazy.force value.f_packages
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("packages", None) );
        ]
    | `UntilPresent value ->
        [
        ]
    | `UntilAbsent value ->
        [
        ]
    | `UnconstrainedArrayIndex value ->
        [
        (try
           ("subtype_indication"
           , Some (Lazy.force value.f_subtype_indication
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subtype_indication", None) );
        ]
    | `DiscreteSubtypeIndication value ->
        [
        (try
           ("has_not_null"
           , Some (Lazy.force value.f_has_not_null
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_not_null", None) );
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ("constraint"
        , (Lazy.force value.f_constraint
           :> AdaNodeType.t option));
        ]
    | `ConstrainedSubtypeIndication value ->
        [
        (try
           ("has_not_null"
           , Some (Lazy.force value.f_has_not_null
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_not_null", None) );
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ("constraint"
        , (Lazy.force value.f_constraint
           :> AdaNodeType.t option));
        ]
    | `SubtypeIndication value ->
        [
        (try
           ("has_not_null"
           , Some (Lazy.force value.f_has_not_null
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_not_null", None) );
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ("constraint"
        , (Lazy.force value.f_constraint
           :> AdaNodeType.t option));
        ]
    | `EnumLitSynthTypeExpr value ->
        [
        ]
    | `AnonymousType value ->
        [
        (try
           ("type_decl"
           , Some (Lazy.force value.f_type_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_decl", None) );
        ]
    | `SignedIntTypeDef value ->
        [
        (try
           ("range"
           , Some (Lazy.force value.f_range
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("range", None) );
        ]
    | `RecordTypeDef value ->
        [
        (try
           ("has_abstract"
           , Some (Lazy.force value.f_has_abstract
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_abstract", None) );
        (try
           ("has_tagged"
           , Some (Lazy.force value.f_has_tagged
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_tagged", None) );
        (try
           ("has_limited"
           , Some (Lazy.force value.f_has_limited
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_limited", None) );
        (try
           ("record_def"
           , Some (Lazy.force value.f_record_def
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("record_def", None) );
        ]
    | `OrdinaryFixedPointDef value ->
        [
        (try
           ("delta"
           , Some (Lazy.force value.f_delta
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("delta", None) );
        ("range"
        , (Lazy.force value.f_range
           :> AdaNodeType.t option));
        ]
    | `FloatingPointDef value ->
        [
        (try
           ("num_digits"
           , Some (Lazy.force value.f_num_digits
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("num_digits", None) );
        ("range"
        , (Lazy.force value.f_range
           :> AdaNodeType.t option));
        ]
    | `DecimalFixedPointDef value ->
        [
        (try
           ("delta"
           , Some (Lazy.force value.f_delta
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("delta", None) );
        (try
           ("digits"
           , Some (Lazy.force value.f_digits
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("digits", None) );
        ("range"
        , (Lazy.force value.f_range
           :> AdaNodeType.t option));
        ]
    | `PrivateTypeDef value ->
        [
        (try
           ("has_abstract"
           , Some (Lazy.force value.f_has_abstract
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_abstract", None) );
        (try
           ("has_tagged"
           , Some (Lazy.force value.f_has_tagged
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_tagged", None) );
        (try
           ("has_limited"
           , Some (Lazy.force value.f_has_limited
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_limited", None) );
        ]
    | `ModIntTypeDef value ->
        [
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `InterfaceTypeDef value ->
        [
        ("interface_kind"
        , (Lazy.force value.f_interface_kind
           :> AdaNodeType.t option));
        ("interfaces"
        , (Lazy.force value.f_interfaces
           :> AdaNodeType.t option));
        ]
    | `FormalDiscreteTypeDef value ->
        [
        ]
    | `EnumTypeDef value ->
        [
        (try
           ("enum_literals"
           , Some (Lazy.force value.f_enum_literals
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("enum_literals", None) );
        ]
    | `DerivedTypeDef value ->
        [
        (try
           ("has_abstract"
           , Some (Lazy.force value.f_has_abstract
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_abstract", None) );
        (try
           ("has_limited"
           , Some (Lazy.force value.f_has_limited
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_limited", None) );
        (try
           ("has_synchronized"
           , Some (Lazy.force value.f_has_synchronized
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_synchronized", None) );
        (try
           ("subtype_indication"
           , Some (Lazy.force value.f_subtype_indication
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subtype_indication", None) );
        ("interfaces"
        , (Lazy.force value.f_interfaces
           :> AdaNodeType.t option));
        ("record_extension"
        , (Lazy.force value.f_record_extension
           :> AdaNodeType.t option));
        (try
           ("has_with_private"
           , Some (Lazy.force value.f_has_with_private
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_with_private", None) );
        ]
    | `ArrayTypeDef value ->
        [
        (try
           ("indices"
           , Some (Lazy.force value.f_indices
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("indices", None) );
        (try
           ("component_type"
           , Some (Lazy.force value.f_component_type
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("component_type", None) );
        ]
    | `TypeAccessDef value ->
        [
        (try
           ("has_not_null"
           , Some (Lazy.force value.f_has_not_null
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_not_null", None) );
        (try
           ("has_all"
           , Some (Lazy.force value.f_has_all
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_all", None) );
        (try
           ("has_constant"
           , Some (Lazy.force value.f_has_constant
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_constant", None) );
        (try
           ("subtype_indication"
           , Some (Lazy.force value.f_subtype_indication
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subtype_indication", None) );
        ]
    | `AnonymousTypeAccessDef value ->
        [
        (try
           ("has_not_null"
           , Some (Lazy.force value.f_has_not_null
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_not_null", None) );
        (try
           ("type_decl"
           , Some (Lazy.force value.f_type_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_decl", None) );
        ]
    | `AccessToSubpDef value ->
        [
        (try
           ("has_not_null"
           , Some (Lazy.force value.f_has_not_null
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_not_null", None) );
        (try
           ("has_protected"
           , Some (Lazy.force value.f_has_protected
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_protected", None) );
        (try
           ("subp_spec"
           , Some (Lazy.force value.f_subp_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_spec", None) );
        ]
    | `TaskDef value ->
        [
        ("interfaces"
        , (Lazy.force value.f_interfaces
           :> AdaNodeType.t option));
        (try
           ("public_part"
           , Some (Lazy.force value.f_public_part
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("public_part", None) );
        ("private_part"
        , (Lazy.force value.f_private_part
           :> AdaNodeType.t option));
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `TaggedPresent value ->
        [
        ]
    | `TaggedAbsent value ->
        [
        ]
    | `SynchronizedPresent value ->
        [
        ]
    | `SynchronizedAbsent value ->
        [
        ]
    | `Subunit value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        (try
           ("body"
           , Some (Lazy.force value.f_body
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("body", None) );
        ]
    | `SubpKindProcedure value ->
        [
        ]
    | `SubpKindFunction value ->
        [
        ]
    | `TerminateAlternative value ->
        [
        ]
    | `ReturnStmt value ->
        [
        ("return_expr"
        , (Lazy.force value.f_return_expr
           :> AdaNodeType.t option));
        ]
    | `RequeueStmt value ->
        [
        (try
           ("call_name"
           , Some (Lazy.force value.f_call_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("call_name", None) );
        (try
           ("has_abort"
           , Some (Lazy.force value.f_has_abort
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_abort", None) );
        ]
    | `RaiseStmt value ->
        [
        ("exception_name"
        , (Lazy.force value.f_exception_name
           :> AdaNodeType.t option));
        ("error_message"
        , (Lazy.force value.f_error_message
           :> AdaNodeType.t option));
        ]
    | `NullStmt value ->
        [
        ]
    | `Label value ->
        [
        (try
           ("decl"
           , Some (Lazy.force value.f_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decl", None) );
        ]
    | `GotoStmt value ->
        [
        (try
           ("label_name"
           , Some (Lazy.force value.f_label_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("label_name", None) );
        ]
    | `ExitStmt value ->
        [
        ("loop_name"
        , (Lazy.force value.f_loop_name
           :> AdaNodeType.t option));
        ("cond_expr"
        , (Lazy.force value.f_cond_expr
           :> AdaNodeType.t option));
        ]
    | `DelayStmt value ->
        [
        (try
           ("has_until"
           , Some (Lazy.force value.f_has_until
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_until", None) );
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `CallStmt value ->
        [
        (try
           ("call"
           , Some (Lazy.force value.f_call
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("call", None) );
        ]
    | `AssignStmt value ->
        [
        (try
           ("dest"
           , Some (Lazy.force value.f_dest
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("dest", None) );
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `AbortStmt value ->
        [
        (try
           ("names"
           , Some (Lazy.force value.f_names
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("names", None) );
        ]
    | `ErrorStmt value ->
        [
        ]
    | `SelectStmt value ->
        [
        (try
           ("guards"
           , Some (Lazy.force value.f_guards
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("guards", None) );
        ("else_stmts"
        , (Lazy.force value.f_else_stmts
           :> AdaNodeType.t option));
        ("abort_stmts"
        , (Lazy.force value.f_abort_stmts
           :> AdaNodeType.t option));
        ]
    | `NamedStmt value ->
        [
        (try
           ("decl"
           , Some (Lazy.force value.f_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decl", None) );
        (try
           ("stmt"
           , Some (Lazy.force value.f_stmt
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmt", None) );
        ]
    | `IfStmt value ->
        [
        (try
           ("cond_expr"
           , Some (Lazy.force value.f_cond_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("cond_expr", None) );
        (try
           ("then_stmts"
           , Some (Lazy.force value.f_then_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("then_stmts", None) );
        (try
           ("alternatives"
           , Some (Lazy.force value.f_alternatives
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("alternatives", None) );
        ("else_stmts"
        , (Lazy.force value.f_else_stmts
           :> AdaNodeType.t option));
        ]
    | `ExtendedReturnStmt value ->
        [
        (try
           ("decl"
           , Some (Lazy.force value.f_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decl", None) );
        ("stmts"
        , (Lazy.force value.f_stmts
           :> AdaNodeType.t option));
        ]
    | `CaseStmt value ->
        [
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        (try
           ("alternatives"
           , Some (Lazy.force value.f_alternatives
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("alternatives", None) );
        ]
    | `DeclBlock value ->
        [
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `BeginBlock value ->
        [
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `WhileLoopStmt value ->
        [
        ("spec"
        , (Lazy.force value.f_spec
           :> AdaNodeType.t option));
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `LoopStmt value ->
        [
        ("spec"
        , (Lazy.force value.f_spec
           :> AdaNodeType.t option));
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `ForLoopStmt value ->
        [
        ("spec"
        , (Lazy.force value.f_spec
           :> AdaNodeType.t option));
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `AcceptStmtWithStmts value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ("entry_index_expr"
        , (Lazy.force value.f_entry_index_expr
           :> AdaNodeType.t option));
        (try
           ("params"
           , Some (Lazy.force value.f_params
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("params", None) );
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `AcceptStmt value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ("entry_index_expr"
        , (Lazy.force value.f_entry_index_expr
           :> AdaNodeType.t option));
        (try
           ("params"
           , Some (Lazy.force value.f_params
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("params", None) );
        ]
    | `SelectWhenPart value ->
        [
        ("cond_expr"
        , (Lazy.force value.f_cond_expr
           :> AdaNodeType.t option));
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ]
    | `ReversePresent value ->
        [
        ]
    | `ReverseAbsent value ->
        [
        ]
    | `SyntheticRenamingClause value ->
        [
        (try
           ("renamed_object"
           , Some (Lazy.force value.f_renamed_object
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("renamed_object", None) );
        ]
    | `RenamingClause value ->
        [
        (try
           ("renamed_object"
           , Some (Lazy.force value.f_renamed_object
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("renamed_object", None) );
        ]
    | `RangeSpec value ->
        [
        (try
           ("range"
           , Some (Lazy.force value.f_range
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("range", None) );
        ]
    | `QuantifierSome value ->
        [
        ]
    | `QuantifierAll value ->
        [
        ]
    | `ProtectedPresent value ->
        [
        ]
    | `ProtectedAbsent value ->
        [
        ]
    | `ProtectedDef value ->
        [
        (try
           ("public_part"
           , Some (Lazy.force value.f_public_part
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("public_part", None) );
        ("private_part"
        , (Lazy.force value.f_private_part
           :> AdaNodeType.t option));
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `PrivatePresent value ->
        [
        ]
    | `PrivateAbsent value ->
        [
        ]
    | `PragmaNode value ->
        [
        (try
           ("id"
           , Some (Lazy.force value.f_id
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("id", None) );
        ("args"
        , (Lazy.force value.f_args
           :> AdaNodeType.t option));
        ]
    | `ParenAbstractStateDecl value ->
        [
        (try
           ("decl"
           , Some (Lazy.force value.f_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decl", None) );
        ]
    | `Params value ->
        [
        (try
           ("params"
           , Some (Lazy.force value.f_params
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("params", None) );
        ]
    | `OverridingUnspecified value ->
        [
        ]
    | `OverridingOverriding value ->
        [
        ]
    | `OverridingNotOverriding value ->
        [
        ]
    | `OthersDesignator value ->
        [
        ]
    | `NullComponentDecl value ->
        [
        ]
    | `NotNullPresent value ->
        [
        ]
    | `NotNullAbsent value ->
        [
        ]
    | `MultiAbstractStateDecl value ->
        [
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        ]
    | `ModeOut value ->
        [
        ]
    | `ModeInOut value ->
        [
        ]
    | `ModeIn value ->
        [
        ]
    | `ModeDefault value ->
        [
        ]
    | `WhileLoopSpec value ->
        [
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `ForLoopSpec value ->
        [
        (try
           ("var_decl"
           , Some (Lazy.force value.f_var_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("var_decl", None) );
        (try
           ("loop_type"
           , Some (Lazy.force value.f_loop_type
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("loop_type", None) );
        (try
           ("has_reverse"
           , Some (Lazy.force value.f_has_reverse
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_reverse", None) );
        (try
           ("iter_expr"
           , Some (Lazy.force value.f_iter_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("iter_expr", None) );
        ]
    | `LimitedPresent value ->
        [
        ]
    | `LimitedAbsent value ->
        [
        ]
    | `LibraryItem value ->
        [
        (try
           ("has_private"
           , Some (Lazy.force value.f_has_private
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_private", None) );
        (try
           ("item"
           , Some (Lazy.force value.f_item
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("item", None) );
        ]
    | `IterTypeOf value ->
        [
        ]
    | `IterTypeIn value ->
        [
        ]
    | `InterfaceKindTask value ->
        [
        ]
    | `InterfaceKindSynchronized value ->
        [
        ]
    | `InterfaceKindProtected value ->
        [
        ]
    | `InterfaceKindLimited value ->
        [
        ]
    | `HandledStmts value ->
        [
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ("exceptions"
        , (Lazy.force value.f_exceptions
           :> AdaNodeType.t option));
        ]
    | `UnOp value ->
        [
        (try
           ("op"
           , Some (Lazy.force value.f_op
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("op", None) );
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `RaiseExpr value ->
        [
        ("exception_name"
        , (Lazy.force value.f_exception_name
           :> AdaNodeType.t option));
        ("error_message"
        , (Lazy.force value.f_error_message
           :> AdaNodeType.t option));
        ]
    | `QuantifiedExpr value ->
        [
        (try
           ("quantifier"
           , Some (Lazy.force value.f_quantifier
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("quantifier", None) );
        (try
           ("loop_spec"
           , Some (Lazy.force value.f_loop_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("loop_spec", None) );
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `ParenExpr value ->
        [
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `TargetName value ->
        [
        ]
    | `RealLiteral value ->
        [
        ]
    | `IntLiteral value ->
        [
        ]
    | `NullLiteral value ->
        [
        ]
    | `StringLiteral value ->
        [
        ]
    | `OpXor value ->
        [
        ]
    | `OpRem value ->
        [
        ]
    | `OpPow value ->
        [
        ]
    | `OpPlus value ->
        [
        ]
    | `OpOrElse value ->
        [
        ]
    | `OpOr value ->
        [
        ]
    | `OpNotIn value ->
        [
        ]
    | `OpNot value ->
        [
        ]
    | `OpNeq value ->
        [
        ]
    | `OpMult value ->
        [
        ]
    | `OpMod value ->
        [
        ]
    | `OpMinus value ->
        [
        ]
    | `OpLte value ->
        [
        ]
    | `OpLt value ->
        [
        ]
    | `OpIn value ->
        [
        ]
    | `OpGte value ->
        [
        ]
    | `OpGt value ->
        [
        ]
    | `OpEq value ->
        [
        ]
    | `OpDoubleDot value ->
        [
        ]
    | `OpDiv value ->
        [
        ]
    | `OpConcat value ->
        [
        ]
    | `OpAndThen value ->
        [
        ]
    | `OpAnd value ->
        [
        ]
    | `OpAbs value ->
        [
        ]
    | `Identifier value ->
        [
        ]
    | `CharLiteral value ->
        [
        ]
    | `QualExpr value ->
        [
        (try
           ("prefix"
           , Some (Lazy.force value.f_prefix
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("prefix", None) );
        (try
           ("suffix"
           , Some (Lazy.force value.f_suffix
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("suffix", None) );
        ]
    | `ExplicitDeref value ->
        [
        (try
           ("prefix"
           , Some (Lazy.force value.f_prefix
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("prefix", None) );
        ]
    | `EndName value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ]
    | `DottedName value ->
        [
        (try
           ("prefix"
           , Some (Lazy.force value.f_prefix
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("prefix", None) );
        (try
           ("suffix"
           , Some (Lazy.force value.f_suffix
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("suffix", None) );
        ]
    | `DiscreteSubtypeName value ->
        [
        (try
           ("subtype"
           , Some (Lazy.force value.f_subtype
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subtype", None) );
        ]
    | `DefiningName value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ]
    | `CallExpr value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        (try
           ("suffix"
           , Some (Lazy.force value.f_suffix
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("suffix", None) );
        ]
    | `UpdateAttributeRef value ->
        [
        (try
           ("prefix"
           , Some (Lazy.force value.f_prefix
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("prefix", None) );
        (try
           ("attribute"
           , Some (Lazy.force value.f_attribute
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("attribute", None) );
        ("args"
        , (Lazy.force value.f_args
           :> AdaNodeType.t option));
        ]
    | `AttributeRef value ->
        [
        (try
           ("prefix"
           , Some (Lazy.force value.f_prefix
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("prefix", None) );
        (try
           ("attribute"
           , Some (Lazy.force value.f_attribute
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("attribute", None) );
        ("args"
        , (Lazy.force value.f_args
           :> AdaNodeType.t option));
        ]
    | `MembershipExpr value ->
        [
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        (try
           ("op"
           , Some (Lazy.force value.f_op
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("op", None) );
        (try
           ("membership_exprs"
           , Some (Lazy.force value.f_membership_exprs
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("membership_exprs", None) );
        ]
    | `DeclExpr value ->
        [
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `ContractCases value ->
        [
        (try
           ("contract_cases"
           , Some (Lazy.force value.f_contract_cases
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("contract_cases", None) );
        ]
    | `IfExpr value ->
        [
        (try
           ("cond_expr"
           , Some (Lazy.force value.f_cond_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("cond_expr", None) );
        (try
           ("then_expr"
           , Some (Lazy.force value.f_then_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("then_expr", None) );
        (try
           ("alternatives"
           , Some (Lazy.force value.f_alternatives
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("alternatives", None) );
        ("else_expr"
        , (Lazy.force value.f_else_expr
           :> AdaNodeType.t option));
        ]
    | `CaseExpr value ->
        [
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        (try
           ("cases"
           , Some (Lazy.force value.f_cases
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("cases", None) );
        ]
    | `CaseExprAlternative value ->
        [
        (try
           ("choices"
           , Some (Lazy.force value.f_choices
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("choices", None) );
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `BoxExpr value ->
        [
        ]
    | `RelationOp value ->
        [
        (try
           ("left"
           , Some (Lazy.force value.f_left
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("left", None) );
        (try
           ("op"
           , Some (Lazy.force value.f_op
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("op", None) );
        (try
           ("right"
           , Some (Lazy.force value.f_right
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("right", None) );
        ]
    | `BinOp value ->
        [
        (try
           ("left"
           , Some (Lazy.force value.f_left
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("left", None) );
        (try
           ("op"
           , Some (Lazy.force value.f_op
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("op", None) );
        (try
           ("right"
           , Some (Lazy.force value.f_right
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("right", None) );
        ]
    | `NullRecordAggregate value ->
        [
        ("ancestor_expr"
        , (Lazy.force value.f_ancestor_expr
           :> AdaNodeType.t option));
        ("assocs"
        , (Lazy.force value.f_assocs
           :> AdaNodeType.t option));
        ]
    | `BracketDeltaAggregate value ->
        [
        ("ancestor_expr"
        , (Lazy.force value.f_ancestor_expr
           :> AdaNodeType.t option));
        ("assocs"
        , (Lazy.force value.f_assocs
           :> AdaNodeType.t option));
        ]
    | `DeltaAggregate value ->
        [
        ("ancestor_expr"
        , (Lazy.force value.f_ancestor_expr
           :> AdaNodeType.t option));
        ("assocs"
        , (Lazy.force value.f_assocs
           :> AdaNodeType.t option));
        ]
    | `BracketAggregate value ->
        [
        ("ancestor_expr"
        , (Lazy.force value.f_ancestor_expr
           :> AdaNodeType.t option));
        ("assocs"
        , (Lazy.force value.f_assocs
           :> AdaNodeType.t option));
        ]
    | `Aggregate value ->
        [
        ("ancestor_expr"
        , (Lazy.force value.f_ancestor_expr
           :> AdaNodeType.t option));
        ("assocs"
        , (Lazy.force value.f_assocs
           :> AdaNodeType.t option));
        ]
    | `Allocator value ->
        [
        ("subpool"
        , (Lazy.force value.f_subpool
           :> AdaNodeType.t option));
        (try
           ("type_or_expr"
           , Some (Lazy.force value.f_type_or_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_or_expr", None) );
        ]
    | `AbstractStateDeclExpr value ->
        [
        (try
           ("state_decl"
           , Some (Lazy.force value.f_state_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("state_decl", None) );
        ]
    | `ElsifStmtPart value ->
        [
        (try
           ("cond_expr"
           , Some (Lazy.force value.f_cond_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("cond_expr", None) );
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ]
    | `ElsifExprPart value ->
        [
        (try
           ("cond_expr"
           , Some (Lazy.force value.f_cond_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("cond_expr", None) );
        (try
           ("then_expr"
           , Some (Lazy.force value.f_then_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("then_expr", None) );
        ]
    | `PublicPart value ->
        [
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        ]
    | `PrivatePart value ->
        [
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        ]
    | `DeclarativePart value ->
        [
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        ]
    | `RangeConstraint value ->
        [
        (try
           ("range"
           , Some (Lazy.force value.f_range
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("range", None) );
        ]
    | `IndexConstraint value ->
        [
        (try
           ("constraints"
           , Some (Lazy.force value.f_constraints
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("constraints", None) );
        ]
    | `DiscriminantConstraint value ->
        [
        (try
           ("constraints"
           , Some (Lazy.force value.f_constraints
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("constraints", None) );
        ]
    | `DigitsConstraint value ->
        [
        (try
           ("digits"
           , Some (Lazy.force value.f_digits
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("digits", None) );
        ("range"
        , (Lazy.force value.f_range
           :> AdaNodeType.t option));
        ]
    | `DeltaConstraint value ->
        [
        (try
           ("digits"
           , Some (Lazy.force value.f_digits
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("digits", None) );
        ("range"
        , (Lazy.force value.f_range
           :> AdaNodeType.t option));
        ]
    | `ConstantPresent value ->
        [
        ]
    | `ConstantAbsent value ->
        [
        ]
    | `ComponentDef value ->
        [
        (try
           ("has_aliased"
           , Some (Lazy.force value.f_has_aliased
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_aliased", None) );
        (try
           ("has_constant"
           , Some (Lazy.force value.f_has_constant
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_constant", None) );
        (try
           ("type_expr"
           , Some (Lazy.force value.f_type_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_expr", None) );
        ]
    | `ComponentClause value ->
        [
        (try
           ("id"
           , Some (Lazy.force value.f_id
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("id", None) );
        (try
           ("position"
           , Some (Lazy.force value.f_position
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("position", None) );
        (try
           ("range"
           , Some (Lazy.force value.f_range
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("range", None) );
        ]
    | `CompilationUnit value ->
        [
        (try
           ("prelude"
           , Some (Lazy.force value.f_prelude
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("prelude", None) );
        (try
           ("body"
           , Some (Lazy.force value.f_body
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("body", None) );
        (try
           ("pragmas"
           , Some (Lazy.force value.f_pragmas
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("pragmas", None) );
        ]
    | `CaseStmtAlternative value ->
        [
        (try
           ("choices"
           , Some (Lazy.force value.f_choices
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("choices", None) );
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ]
    | `SingleTaskDecl value ->
        [
        (try
           ("task_type"
           , Some (Lazy.force value.f_task_type
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("task_type", None) );
        ]
    | `SingleProtectedDecl value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ("interfaces"
        , (Lazy.force value.f_interfaces
           :> AdaNodeType.t option));
        (try
           ("definition"
           , Some (Lazy.force value.f_definition
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("definition", None) );
        ]
    | `PackageRenamingDecl value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        (try
           ("renames"
           , Some (Lazy.force value.f_renames
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("renames", None) );
        ]
    | `ExtendedReturnStmtObjectDecl value ->
        [
        (try
           ("ids"
           , Some (Lazy.force value.f_ids
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("ids", None) );
        (try
           ("has_aliased"
           , Some (Lazy.force value.f_has_aliased
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_aliased", None) );
        (try
           ("has_constant"
           , Some (Lazy.force value.f_has_constant
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_constant", None) );
        ("mode"
        , (Lazy.force value.f_mode
           :> AdaNodeType.t option));
        (try
           ("type_expr"
           , Some (Lazy.force value.f_type_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_expr", None) );
        ("default_expr"
        , (Lazy.force value.f_default_expr
           :> AdaNodeType.t option));
        ("renaming_clause"
        , (Lazy.force value.f_renaming_clause
           :> AdaNodeType.t option));
        ]
    | `ObjectDecl value ->
        [
        (try
           ("ids"
           , Some (Lazy.force value.f_ids
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("ids", None) );
        (try
           ("has_aliased"
           , Some (Lazy.force value.f_has_aliased
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_aliased", None) );
        (try
           ("has_constant"
           , Some (Lazy.force value.f_has_constant
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_constant", None) );
        ("mode"
        , (Lazy.force value.f_mode
           :> AdaNodeType.t option));
        (try
           ("type_expr"
           , Some (Lazy.force value.f_type_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_expr", None) );
        ("default_expr"
        , (Lazy.force value.f_default_expr
           :> AdaNodeType.t option));
        ("renaming_clause"
        , (Lazy.force value.f_renaming_clause
           :> AdaNodeType.t option));
        ]
    | `NumberDecl value ->
        [
        (try
           ("ids"
           , Some (Lazy.force value.f_ids
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("ids", None) );
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `NamedStmtDecl value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ]
    | `LabelDecl value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ]
    | `GenericSubpRenamingDecl value ->
        [
        (try
           ("kind"
           , Some (Lazy.force value.f_kind
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("kind", None) );
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        (try
           ("renames"
           , Some (Lazy.force value.f_renames
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("renames", None) );
        ]
    | `GenericPackageRenamingDecl value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        (try
           ("renames"
           , Some (Lazy.force value.f_renames
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("renames", None) );
        ]
    | `GenericSubpInstantiation value ->
        [
        (try
           ("overriding"
           , Some (Lazy.force value.f_overriding
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("overriding", None) );
        (try
           ("kind"
           , Some (Lazy.force value.f_kind
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("kind", None) );
        (try
           ("subp_name"
           , Some (Lazy.force value.f_subp_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_name", None) );
        (try
           ("generic_subp_name"
           , Some (Lazy.force value.f_generic_subp_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("generic_subp_name", None) );
        ("params"
        , (Lazy.force value.f_params
           :> AdaNodeType.t option));
        ]
    | `GenericPackageInstantiation value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        (try
           ("generic_pkg_name"
           , Some (Lazy.force value.f_generic_pkg_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("generic_pkg_name", None) );
        ("params"
        , (Lazy.force value.f_params
           :> AdaNodeType.t option));
        ]
    | `GenericSubpDecl value ->
        [
        (try
           ("formal_part"
           , Some (Lazy.force value.f_formal_part
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("formal_part", None) );
        (try
           ("subp_decl"
           , Some (Lazy.force value.f_subp_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_decl", None) );
        ]
    | `GenericPackageDecl value ->
        [
        (try
           ("formal_part"
           , Some (Lazy.force value.f_formal_part
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("formal_part", None) );
        (try
           ("package_decl"
           , Some (Lazy.force value.f_package_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("package_decl", None) );
        ]
    | `ForLoopVarDecl value ->
        [
        (try
           ("id"
           , Some (Lazy.force value.f_id
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("id", None) );
        ("id_type"
        , (Lazy.force value.f_id_type
           :> AdaNodeType.t option));
        ]
    | `ExceptionHandler value ->
        [
        ("exception_name"
        , (Lazy.force value.f_exception_name
           :> AdaNodeType.t option));
        (try
           ("handled_exceptions"
           , Some (Lazy.force value.f_handled_exceptions
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("handled_exceptions", None) );
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ]
    | `ExceptionDecl value ->
        [
        (try
           ("ids"
           , Some (Lazy.force value.f_ids
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("ids", None) );
        ("renames"
        , (Lazy.force value.f_renames
           :> AdaNodeType.t option));
        ]
    | `ErrorDecl value ->
        [
        ]
    | `EntryIndexSpec value ->
        [
        (try
           ("id"
           , Some (Lazy.force value.f_id
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("id", None) );
        (try
           ("subtype"
           , Some (Lazy.force value.f_subtype
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subtype", None) );
        ]
    | `TaskBody value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `ProtectedBody value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `PackageBody value ->
        [
        (try
           ("package_name"
           , Some (Lazy.force value.f_package_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("package_name", None) );
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        ("stmts"
        , (Lazy.force value.f_stmts
           :> AdaNodeType.t option));
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `EntryBody value ->
        [
        (try
           ("entry_name"
           , Some (Lazy.force value.f_entry_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("entry_name", None) );
        ("index_spec"
        , (Lazy.force value.f_index_spec
           :> AdaNodeType.t option));
        (try
           ("params"
           , Some (Lazy.force value.f_params
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("params", None) );
        (try
           ("barrier"
           , Some (Lazy.force value.f_barrier
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("barrier", None) );
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `TaskBodyStub value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ]
    | `SubpBodyStub value ->
        [
        (try
           ("overriding"
           , Some (Lazy.force value.f_overriding
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("overriding", None) );
        (try
           ("subp_spec"
           , Some (Lazy.force value.f_subp_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_spec", None) );
        ]
    | `ProtectedBodyStub value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ]
    | `PackageBodyStub value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ]
    | `SubpRenamingDecl value ->
        [
        (try
           ("overriding"
           , Some (Lazy.force value.f_overriding
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("overriding", None) );
        (try
           ("subp_spec"
           , Some (Lazy.force value.f_subp_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_spec", None) );
        (try
           ("renames"
           , Some (Lazy.force value.f_renames
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("renames", None) );
        ]
    | `SubpBody value ->
        [
        (try
           ("overriding"
           , Some (Lazy.force value.f_overriding
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("overriding", None) );
        (try
           ("subp_spec"
           , Some (Lazy.force value.f_subp_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_spec", None) );
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        (try
           ("stmts"
           , Some (Lazy.force value.f_stmts
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("stmts", None) );
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `NullSubpDecl value ->
        [
        (try
           ("overriding"
           , Some (Lazy.force value.f_overriding
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("overriding", None) );
        (try
           ("subp_spec"
           , Some (Lazy.force value.f_subp_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_spec", None) );
        ]
    | `ExprFunction value ->
        [
        (try
           ("overriding"
           , Some (Lazy.force value.f_overriding
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("overriding", None) );
        (try
           ("subp_spec"
           , Some (Lazy.force value.f_subp_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_spec", None) );
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `GenericSubpInternal value ->
        [
        (try
           ("subp_spec"
           , Some (Lazy.force value.f_subp_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_spec", None) );
        ]
    | `EnumLiteralDecl value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ]
    | `EntryDecl value ->
        [
        (try
           ("overriding"
           , Some (Lazy.force value.f_overriding
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("overriding", None) );
        (try
           ("spec"
           , Some (Lazy.force value.f_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("spec", None) );
        ]
    | `SubpDecl value ->
        [
        (try
           ("overriding"
           , Some (Lazy.force value.f_overriding
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("overriding", None) );
        (try
           ("subp_spec"
           , Some (Lazy.force value.f_subp_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_spec", None) );
        ]
    | `ConcreteFormalSubpDecl value ->
        [
        (try
           ("overriding"
           , Some (Lazy.force value.f_overriding
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("overriding", None) );
        (try
           ("subp_spec"
           , Some (Lazy.force value.f_subp_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_spec", None) );
        ("default_expr"
        , (Lazy.force value.f_default_expr
           :> AdaNodeType.t option));
        ]
    | `AbstractFormalSubpDecl value ->
        [
        (try
           ("overriding"
           , Some (Lazy.force value.f_overriding
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("overriding", None) );
        (try
           ("subp_spec"
           , Some (Lazy.force value.f_subp_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_spec", None) );
        ("default_expr"
        , (Lazy.force value.f_default_expr
           :> AdaNodeType.t option));
        ]
    | `AbstractSubpDecl value ->
        [
        (try
           ("overriding"
           , Some (Lazy.force value.f_overriding
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("overriding", None) );
        (try
           ("subp_spec"
           , Some (Lazy.force value.f_subp_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_spec", None) );
        ]
    | `SynthAnonymousTypeDecl value ->
        [
        ("name"
        , (Lazy.force value.f_name
           :> AdaNodeType.t option));
        ("discriminants"
        , (Lazy.force value.f_discriminants
           :> AdaNodeType.t option));
        (try
           ("type_def"
           , Some (Lazy.force value.f_type_def
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_def", None) );
        ]
    | `AnonymousTypeDecl value ->
        [
        ("name"
        , (Lazy.force value.f_name
           :> AdaNodeType.t option));
        ("discriminants"
        , (Lazy.force value.f_discriminants
           :> AdaNodeType.t option));
        (try
           ("type_def"
           , Some (Lazy.force value.f_type_def
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_def", None) );
        ]
    | `TypeDecl value ->
        [
        ("name"
        , (Lazy.force value.f_name
           :> AdaNodeType.t option));
        ("discriminants"
        , (Lazy.force value.f_discriminants
           :> AdaNodeType.t option));
        (try
           ("type_def"
           , Some (Lazy.force value.f_type_def
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_def", None) );
        ]
    | `SingleTaskTypeDecl value ->
        [
        ("name"
        , (Lazy.force value.f_name
           :> AdaNodeType.t option));
        ("discriminants"
        , (Lazy.force value.f_discriminants
           :> AdaNodeType.t option));
        ("definition"
        , (Lazy.force value.f_definition
           :> AdaNodeType.t option));
        ]
    | `TaskTypeDecl value ->
        [
        ("name"
        , (Lazy.force value.f_name
           :> AdaNodeType.t option));
        ("discriminants"
        , (Lazy.force value.f_discriminants
           :> AdaNodeType.t option));
        ("definition"
        , (Lazy.force value.f_definition
           :> AdaNodeType.t option));
        ]
    | `ProtectedTypeDecl value ->
        [
        ("name"
        , (Lazy.force value.f_name
           :> AdaNodeType.t option));
        ("discriminants"
        , (Lazy.force value.f_discriminants
           :> AdaNodeType.t option));
        ("interfaces"
        , (Lazy.force value.f_interfaces
           :> AdaNodeType.t option));
        (try
           ("definition"
           , Some (Lazy.force value.f_definition
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("definition", None) );
        ]
    | `IncompleteTaggedTypeDecl value ->
        [
        ("name"
        , (Lazy.force value.f_name
           :> AdaNodeType.t option));
        ("discriminants"
        , (Lazy.force value.f_discriminants
           :> AdaNodeType.t option));
        (try
           ("has_abstract"
           , Some (Lazy.force value.f_has_abstract
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_abstract", None) );
        ]
    | `IncompleteTypeDecl value ->
        [
        ("name"
        , (Lazy.force value.f_name
           :> AdaNodeType.t option));
        ("discriminants"
        , (Lazy.force value.f_discriminants
           :> AdaNodeType.t option));
        ]
    | `ClasswideTypeDecl value ->
        [
        ("name"
        , (Lazy.force value.f_name
           :> AdaNodeType.t option));
        ]
    | `SubtypeDecl value ->
        [
        ("name"
        , (Lazy.force value.f_name
           :> AdaNodeType.t option));
        (try
           ("subtype"
           , Some (Lazy.force value.f_subtype
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subtype", None) );
        ]
    | `DiscreteBaseSubtypeDecl value ->
        [
        ("name"
        , (Lazy.force value.f_name
           :> AdaNodeType.t option));
        ]
    | `PackageDecl value ->
        [
        (try
           ("package_name"
           , Some (Lazy.force value.f_package_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("package_name", None) );
        (try
           ("public_part"
           , Some (Lazy.force value.f_public_part
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("public_part", None) );
        ("private_part"
        , (Lazy.force value.f_private_part
           :> AdaNodeType.t option));
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `GenericPackageInternal value ->
        [
        (try
           ("package_name"
           , Some (Lazy.force value.f_package_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("package_name", None) );
        (try
           ("public_part"
           , Some (Lazy.force value.f_public_part
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("public_part", None) );
        ("private_part"
        , (Lazy.force value.f_private_part
           :> AdaNodeType.t option));
        ("end_name"
        , (Lazy.force value.f_end_name
           :> AdaNodeType.t option));
        ]
    | `ParamSpec value ->
        [
        (try
           ("ids"
           , Some (Lazy.force value.f_ids
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("ids", None) );
        (try
           ("has_aliased"
           , Some (Lazy.force value.f_has_aliased
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("has_aliased", None) );
        ("mode"
        , (Lazy.force value.f_mode
           :> AdaNodeType.t option));
        (try
           ("type_expr"
           , Some (Lazy.force value.f_type_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_expr", None) );
        ("default_expr"
        , (Lazy.force value.f_default_expr
           :> AdaNodeType.t option));
        ]
    | `GenericFormalTypeDecl value ->
        [
        (try
           ("decl"
           , Some (Lazy.force value.f_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decl", None) );
        ]
    | `GenericFormalSubpDecl value ->
        [
        (try
           ("decl"
           , Some (Lazy.force value.f_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decl", None) );
        ]
    | `GenericFormalPackage value ->
        [
        (try
           ("decl"
           , Some (Lazy.force value.f_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decl", None) );
        ]
    | `GenericFormalObjDecl value ->
        [
        (try
           ("decl"
           , Some (Lazy.force value.f_decl
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decl", None) );
        ]
    | `DiscriminantSpec value ->
        [
        (try
           ("ids"
           , Some (Lazy.force value.f_ids
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("ids", None) );
        (try
           ("type_expr"
           , Some (Lazy.force value.f_type_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_expr", None) );
        ("default_expr"
        , (Lazy.force value.f_default_expr
           :> AdaNodeType.t option));
        ]
    | `ComponentDecl value ->
        [
        (try
           ("ids"
           , Some (Lazy.force value.f_ids
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("ids", None) );
        (try
           ("component_def"
           , Some (Lazy.force value.f_component_def
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("component_def", None) );
        ("default_expr"
        , (Lazy.force value.f_default_expr
           :> AdaNodeType.t option));
        ]
    | `AnonymousExprDecl value ->
        [
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `AbstractStateDecl value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ]
    | `ParamAssoc value ->
        [
        ("designator"
        , (Lazy.force value.f_designator
           :> AdaNodeType.t option));
        (try
           ("r_expr"
           , Some (Lazy.force value.f_r_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("r_expr", None) );
        ]
    | `IteratedAssoc value ->
        [
        (try
           ("spec"
           , Some (Lazy.force value.f_spec
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("spec", None) );
        (try
           ("r_expr"
           , Some (Lazy.force value.f_r_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("r_expr", None) );
        ]
    | `DiscriminantAssoc value ->
        [
        ("ids"
        , (Lazy.force value.f_ids
           :> AdaNodeType.t option));
        (try
           ("discr_expr"
           , Some (Lazy.force value.f_discr_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("discr_expr", None) );
        ]
    | `MultiDimArrayAssoc value ->
        [
        ("designators"
        , (Lazy.force value.f_designators
           :> AdaNodeType.t option));
        (try
           ("r_expr"
           , Some (Lazy.force value.f_r_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("r_expr", None) );
        ]
    | `AggregateAssoc value ->
        [
        ("designators"
        , (Lazy.force value.f_designators
           :> AdaNodeType.t option));
        (try
           ("r_expr"
           , Some (Lazy.force value.f_r_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("r_expr", None) );
        ]
    | `RecordDef value ->
        [
        (try
           ("components"
           , Some (Lazy.force value.f_components
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("components", None) );
        ]
    | `NullRecordDef value ->
        [
        (try
           ("components"
           , Some (Lazy.force value.f_components
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("components", None) );
        ]
    | `GenericFormalPart value ->
        [
        (try
           ("decls"
           , Some (Lazy.force value.f_decls
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("decls", None) );
        ]
    | `EntryCompletionFormalParams value ->
        [
        ("params"
        , (Lazy.force value.f_params
           :> AdaNodeType.t option));
        ]
    | `UnknownDiscriminantPart value ->
        [
        ]
    | `KnownDiscriminantPart value ->
        [
        (try
           ("discr_specs"
           , Some (Lazy.force value.f_discr_specs
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("discr_specs", None) );
        ]
    | `ComponentList value ->
        [
        ("components"
        , (Lazy.force value.f_components
           :> AdaNodeType.t option));
        ("variant_part"
        , (Lazy.force value.f_variant_part
           :> AdaNodeType.t option));
        ]
    | `SubpSpec value ->
        [
        (try
           ("subp_kind"
           , Some (Lazy.force value.f_subp_kind
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("subp_kind", None) );
        ("subp_name"
        , (Lazy.force value.f_subp_name
           :> AdaNodeType.t option));
        ("subp_params"
        , (Lazy.force value.f_subp_params
           :> AdaNodeType.t option));
        ("subp_returns"
        , (Lazy.force value.f_subp_returns
           :> AdaNodeType.t option));
        ]
    | `EnumSubpSpec value ->
        [
        ]
    | `EntrySpec value ->
        [
        (try
           ("entry_name"
           , Some (Lazy.force value.f_entry_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("entry_name", None) );
        ("family_type"
        , (Lazy.force value.f_family_type
           :> AdaNodeType.t option));
        ("entry_params"
        , (Lazy.force value.f_entry_params
           :> AdaNodeType.t option));
        ]
    | `PragmaArgumentAssoc value ->
        [
        ("id"
        , (Lazy.force value.f_id
           :> AdaNodeType.t option));
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `ContractCaseAssoc value ->
        [
        (try
           ("guard"
           , Some (Lazy.force value.f_guard
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("guard", None) );
        (try
           ("consequence"
           , Some (Lazy.force value.f_consequence
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("consequence", None) );
        ]
    | `AspectSpec value ->
        [
        (try
           ("aspect_assocs"
           , Some (Lazy.force value.f_aspect_assocs
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("aspect_assocs", None) );
        ]
    | `RecordRepClause value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        ("at_expr"
        , (Lazy.force value.f_at_expr
           :> AdaNodeType.t option));
        (try
           ("components"
           , Some (Lazy.force value.f_components
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("components", None) );
        ]
    | `EnumRepClause value ->
        [
        (try
           ("type_name"
           , Some (Lazy.force value.f_type_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("type_name", None) );
        (try
           ("aggregate"
           , Some (Lazy.force value.f_aggregate
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("aggregate", None) );
        ]
    | `AttributeDefClause value ->
        [
        (try
           ("attribute_expr"
           , Some (Lazy.force value.f_attribute_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("attribute_expr", None) );
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `AtClause value ->
        [
        (try
           ("name"
           , Some (Lazy.force value.f_name
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("name", None) );
        (try
           ("expr"
           , Some (Lazy.force value.f_expr
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("expr", None) );
        ]
    | `AspectAssoc value ->
        [
        (try
           ("id"
           , Some (Lazy.force value.f_id
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("id", None) );
        ("expr"
        , (Lazy.force value.f_expr
           :> AdaNodeType.t option));
        ]
    | `UnconstrainedArrayIndices value ->
        [
        (try
           ("types"
           , Some (Lazy.force value.f_types
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("types", None) );
        ]
    | `ConstrainedArrayIndices value ->
        [
        (try
           ("list"
           , Some (Lazy.force value.f_list
                    :> AdaNodeType.t))
        with SyntaxError ->
          ("list", None) );
        ]
    | `AllPresent value ->
        [
        ]
    | `AllAbsent value ->
        [
        ]
    | `AliasedPresent value ->
        [
        ]
    | `AliasedAbsent value ->
        [
        ]
    | `VariantList value ->
        List.mapi aux (children_opt node)
    | `UnconstrainedArrayIndexList value ->
        List.mapi aux (children_opt node)
    | `SelectWhenPartList value ->
        List.mapi aux (children_opt node)
    | `PragmaNodeList value ->
        List.mapi aux (children_opt node)
    | `ParamSpecList value ->
        List.mapi aux (children_opt node)
    | `ParentList value ->
        List.mapi aux (children_opt node)
    | `NameList value ->
        List.mapi aux (children_opt node)
    | `DiscriminantChoiceList value ->
        List.mapi aux (children_opt node)
    | `ExprAlternativesList value ->
        List.mapi aux (children_opt node)
    | `EnumLiteralDeclList value ->
        List.mapi aux (children_opt node)
    | `ElsifStmtPartList value ->
        List.mapi aux (children_opt node)
    | `ElsifExprPartList value ->
        List.mapi aux (children_opt node)
    | `DiscriminantSpecList value ->
        List.mapi aux (children_opt node)
    | `DefiningNameList value ->
        List.mapi aux (children_opt node)
    | `ContractCaseAssocList value ->
        List.mapi aux (children_opt node)
    | `CompilationUnitList value ->
        List.mapi aux (children_opt node)
    | `CaseStmtAlternativeList value ->
        List.mapi aux (children_opt node)
    | `CaseExprAlternativeList value ->
        List.mapi aux (children_opt node)
    | `BasicDeclList value ->
        List.mapi aux (children_opt node)
    | `AssocList value ->
        List.mapi aux (children_opt node)
    | `BaseAssocList value ->
        List.mapi aux (children_opt node)
    | `AspectAssocList value ->
        List.mapi aux (children_opt node)
    | `StmtList value ->
        List.mapi aux (children_opt node)
    | `DeclList value ->
        List.mapi aux (children_opt node)
    | `ConstraintList value ->
        List.mapi aux (children_opt node)
    | `AlternativesList value ->
        List.mapi aux (children_opt node)
    | `AbstractStateDeclList value ->
        List.mapi aux (children_opt node)
    | `AdaNodeList value ->
        List.mapi aux (children_opt node)
    | `AbstractPresent value ->
        [
        ]
    | `AbstractAbsent value ->
        [
        ]
    | `AbortPresent value ->
        [
        ]
    | `AbortAbsent value ->
        [
        ]

  let rec pp_tree fmt node =
    let rec pp_node_field fmt (name, node) =
      match node with
      | Some node ->
          Format.fprintf fmt "@[<v 2>%s:@ %a@]" name pp_node node
      | None ->
          Format.fprintf fmt "@[<v 2>%s: None@]" name
    and pp_node_fields fmt node =
      let name_field_list = fields_with_names node in
      match name_field_list with
      | [] ->
          ()
      | l ->
          Format.fprintf fmt "@ @[<v>%a@]"
            (Format.pp_print_list pp_node_field) l
    and pp_node fmt node =
      let repr = entity_image node in
      let len = String.length repr in
      let erepr = String.sub repr 1 (len - 2) in
      Format.fprintf fmt "@[<v 2>%s%s%a@]"
        erepr
        (if is_token_node node then (": " ^ (text node)) else "")
        pp_node_fields node
    in
    let default = Format.pp_get_formatter_out_functions fmt () in
    let out_indent n =
      let the_end = n in
      let rec make n =
        if n = the_end then ""
        else (if n mod 4 = 2 then "|" else " ") ^ make (n + 1)
      in
      default.out_string (make 0) 0 n
    in
    Format.pp_set_formatter_out_functions fmt {default with out_indent} ;
    Format.fprintf fmt "%a%!" pp_node (node :> AdaNodeType.t);
    Format.pp_set_formatter_out_functions fmt default


end

