





(**
 * OCaml binding of the Libadalang API.
 *)

(**
 * Raised when introspection functions (``Libadalang.Introspection``) are
 * provided mismatching types/values.
 *)
exception BadTypeError of string

(**
 * Raised when introspection functions (``Libadalang.Introspection``) are
 * passed an out of bounds index.
 *)
exception OutOfBoundsError of string

(**
 * Raised by lexing functions (``Libadalang.Lexer``) when the input contains an
 * invalid byte sequence.
 *)
exception InvalidInput of string

(**
 * Exception raise when an invalid symbol is passed to a subprogram.
 *)
exception InvalidSymbolError of string

(**
 * Raised when an invalid unit name is provided.
 *)
exception InvalidUnitNameError of string

(**
 * Exception raised in language bindings when the underlying C API reports an
 * unexpected error that occurred in the library.

 * This kind of exception is raised for internal errors: they should never
 * happen in normal situations and if they are raised at some point, it means
 * the library state is potentially corrupted.

 * Nevertheless, the library does its best not to crash the program,
 * materializing internal errors using this kind of exception.
 *)
exception NativeException of string

(**
 * Exception raised when an API is called while its preconditions are not
 * satisfied.
 *)
exception PreconditionFailure of string

(**
 * Exception that is raised when an error occurs while evaluating any AST node
 * method whose name starts with ``p_``. This is the only exceptions that such
 * functions can raise.
 *)
exception PropertyError of string

(**
 * Exception raised when the provided arguments for a template don't match what
 * the template expects.
 *)
exception TemplateArgsError of string

(**
 * Exception raised when a template has an invalid syntax, such as badly
 * formatted placeholders.
 *)
exception TemplateFormatError of string

(**
 * Exception raised when the instantiation of a template cannot be parsed.
 *)
exception TemplateInstantiationError of string

(**
 * Exception raised while trying to access data that was deallocated. This
 * happens when one tries to use a node whose unit has been reparsed, for
 * instance.
 *)
exception StaleReferenceError of string

(**
 * Raised by lexing functions (``Libadalang.Lexer``) when the input charset is
 * not supported.
 *)
exception UnknownCharset of string



      (**
 * Raised when an error occurs while loading a project file.
 *)
exception InvalidProjectError of string



(**
 * Raised if a field of a node is null due to a syntax error
 *)
exception SyntaxError

module AnalysisUnitKind : sig
  (**
   * Specify a kind of analysis unit. Specification units provide an interface
   * to the outer world while body units provide an implementation for the
   * corresponding interface.
   *)

  type t =
  | UnitSpecification
  | UnitBody

  val name : unit -> string
end

module LookupKind : sig
  

  type t =
  | Recursive
  | Flat
  | Minimal

  val name : unit -> string
end

module FindAllMode : sig
  

  type t =
  | References
  | Derivedtypes

  val name : unit -> string
end

module RefResultKind : sig
  (**
   * Kind for the result of a cross reference operation.
  
   * - ``NoRef`` is for no reference, it is the null value for this enum.
  
   * - ``Precise`` is when the reference result is precise.
  
   * - ``Imprecise`` is when there was an error computing the precise result,
   * and a result was gotten in an imprecise fashion.
  
   * - ``Error`` is for unrecoverable errors (either there is no imprecise path
   * for the request you made, or the imprecise path errored out too.
   *)

  type t =
  | Noref
  | Precise
  | Imprecise
  | Error

  val name : unit -> string
end

module GrammarRule : sig
  (**
   * Gramar rule to use for parsing.
   *)

  type t =
  | ParentListRule
  | ProtectedTypeDeclRule
  | ProtectedOpRule
  | ProtectedElRule
  | ProtectedDefRule
  | ProtectedDeclRule
  | TaskItemRule
  | TaskDefRule
  | TaskTypeDeclRule
  | SubtypeDeclRule
  | InterfaceTypeDefRule
  | UnconstrainedIndexRule
  | ArrayTypeDefRule
  | DiscreteSubtypeDefinitionRule
  | ConstraintListRule
  | SignedIntTypeDefRule
  | ModIntTypeDefRule
  | DerivedTypeDefRule
  | DiscriminantAssocRule
  | DiscriminantConstraintRule
  | IndexConstraintRule
  | DigitsConstraintRule
  | DeltaConstraintRule
  | RangeConstraintRule
  | ConstraintRule
  | DiscriminantSpecRule
  | DiscrSpecListRule
  | DiscriminantPartRule
  | EnumLiteralDeclRule
  | FormalDiscreteTypeDefRule
  | RecordDefRule
  | RangeSpecRule
  | RealTypeDefRule
  | SexprOrBoxRule
  | OrdinaryFixedPointDefRule
  | DecimalFixedPointDefRule
  | FloatingPointDefRule
  | RecordTypeDefRule
  | AccessDefRule
  | EnumTypeDefRule
  | TypeDefRule
  | VariantRule
  | AnonymousTypeDeclRule
  | TypeDeclRule
  | VariantPartRule
  | ComponentDefRule
  | ComponentItemRule
  | ComponentDeclRule
  | ComponentListRule
  | GenericDeclRule
  | GenericFormalPartRule
  | GenericFormalDeclRule
  | FormalSubpDeclRule
  | RenamingClauseRule
  | GenericRenamingDeclRule
  | GenericInstantiationRule
  | ExceptionDeclRule
  | BasicDeclsRule
  | PackageRenamingDeclRule
  | PackageDeclRule
  | BasicDeclRule
  | ObjectDeclRule
  | SubObjectDeclRule
  | ExtRetStmtObjectDeclRule
  | DefiningIdListRule
  | NumberDeclRule
  | ContractCaseAssocRule
  | ContractCasesExprRule
  | AbstractStateDeclRule
  | MultiAbstractStateDeclRule
  | AspectAssocRule
  | AspectSpecRule
  | SingleTaskDeclRule
  | OverridingIndicatorRule
  | EntryDeclRule
  | ComponentClauseRule
  | AspectClauseRule
  | ParamSpecRule
  | ParamSpecsRule
  | SubpSpecRule
  | ExprFnRule
  | NullSubpDeclRule
  | AbstractSubpDeclRule
  | SubpRenamingDeclRule
  | SimpleSubpDeclRule
  | SubpDeclRule
  | WithClauseRule
  | ContextItemRule
  | UseClauseRule
  | UsePackageClauseRule
  | UseTypeClauseRule
  | SubtypeIndicationRule
  | DiscreteSubtypeIndicationRule
  | ConstrainedSubtypeIndicationRule
  | TypeExprRule
  | AnonymousTypeRule
  | ModeRule
  | PragmaArgumentRule
  | PragmaRule
  | SubunitRule
  | LibraryUnitBodyRule
  | LibraryUnitRenamingDeclRule
  | LibraryItemRule
  | CompilationUnitRule
  | CompilationRule
  | DeclPartRule
  | EntryBodyRule
  | ProtectedBodyRule
  | ProtectedBodyStubRule
  | TaskBodyRule
  | TaskBodyStubRule
  | PackageBodyStubRule
  | PackageBodyRule
  | TerminateAlternativeRule
  | SelectStmtRule
  | AcceptStmtRule
  | CaseAltRule
  | CaseStmtRule
  | ExtReturnStmtRule
  | IblockStmtRule
  | BlockStmtRule
  | IloopStmtRule
  | LoopStmtRule
  | CompoundStmtRule
  | IfStmtRule
  | RaiseStmtRule
  | DelayStmtRule
  | AbortStmtRule
  | BodyRule
  | BodyStubRule
  | SubpBodyStubRule
  | RecovDeclPartRule
  | SubpBodyRule
  | HandledStmtsRule
  | ExceptionHandlerRule
  | StmtsRule
  | LabelRule
  | StmtRule
  | CallStmtRule
  | SimpleStmtRule
  | NullStmtRule
  | AssignmentStmtRule
  | GotoStmtRule
  | ExitStmtRule
  | ReturnStmtRule
  | RequeueStmtRule
  | IdentifierRule
  | CharLiteralRule
  | StringLiteralRule
  | DefiningIdRule
  | DecLiteralRule
  | IntLiteralRule
  | NumLiteralRule
  | NullLiteralRule
  | AllocatorRule
  | ForLoopParamSpecRule
  | QuantifiedExprRule
  | CaseExprRule
  | CaseExprAltRule
  | RaiseExprRule
  | IfExprRule
  | ConditionalExprRule
  | BoxExprRule
  | OthersDesignatorRule
  | AggregateAssocRule
  | RegularAggregateRule
  | BracketAggregateRule
  | AggregateRule
  | DirectNameRule
  | ParamAssocRule
  | CallSuffixRule
  | QualifiedNameRule
  | QualNameInternalRule
  | NameRule
  | DefiningNameRule
  | DirectNameOrTargetNameRule
  | TargetNameRule
  | UpdateAttrAggregateRule
  | UpdateAttrContentRule
  | MultidimArrayAssocRule
  | SubtypeNameRule
  | StaticNameRule
  | PrimaryRule
  | ParenExprRule
  | DeclareExprRule
  | FactorRule
  | TermRule
  | UnopTermRule
  | SimpleExprRule
  | BooleanOpRule
  | DiscreteRangeRule
  | ChoiceRule
  | ChoiceListRule
  | RelOpRule
  | MembershipChoiceRule
  | MembershipChoiceListRule
  | RelationRule
  | ExprRule

  val name : unit -> string
end

val default_grammar_rule : GrammarRule.t

module Camomile : CamomileLibrary.Type

module Sloc : sig
  (**
   * Location in a source file. Line and column numbers are one-based.
   *)

  type t = {
    line : int;
    column : int;
  }
end

module SlocRange : sig
  (**
   * Location of a span of text in a source file.
   *)

  type t = {
    loc_start : Sloc.t;
    loc_end : Sloc.t;
  }

  val pp : Format.formatter -> t -> unit
end

module Diagnostic : sig
  (**
   * Diagnostic for an analysis unit: cannot open the source file, parsing
   * error, ...
   *)

  type t = {
    sloc_range : SlocRange.t;
    message : string;
  }
end

module TokenData : sig
  type t
end

module Token : sig
  (**
   * Reference to a token in an analysis unit.
   *)

  type t = {
    token_data : TokenData.t;
    token_index : int;
    trivia_index : int;
    kind : int;
    text : string;
    sloc_range : SlocRange.t;
  }

  val kind_name : t -> string
  (**
   * Return a human-readable name for a token kind.
   *)

  val text_range : t -> t -> string
  (**
   * Compute the source buffer slice corresponding to the text that spans
   * between the ``First`` and ``Last`` tokens (both included). This yields an
   * empty slice if ``Last`` actually appears before ``First``.
  
   * if ``First`` and ``Last`` don't belong to the same analysis unit.
   *)

  val is_trivia : t -> bool
  (**
   * Return whether this token is a trivia. If it's not, it's a regular token.
   *)

  val index : t -> int
  (**
   * Zero-based index for this token/trivia. Tokens and trivias get their own
   * index space.
   *)

  val next : t -> t
  (**
   * Return a reference to the first token scanned in this unit.
   *)

  val previous : t -> t
  (**
   * Return a reference to the last token scanned in this unit.
   *)

  val compare : t -> t -> int

  val equal : t -> t -> bool

  val hash : t -> int

  val is_equivalent : t -> t -> bool
  (**
   * Return whether ``L`` and ``R`` are structurally equivalent tokens. This
   * means that their position in the stream won't be taken into account, only
   * the kind and text of the token.
   *)

  val pp : Format.formatter -> t -> unit
end

module BigInteger : sig
  type t = Z.t
end

module BareNode : sig
  type t
end

module Rebindings : sig
  type t
end

module UnitProvider : sig
  (**
   * Interface to fetch analysis units from a name and a unit kind.
  
   * The unit provider mechanism provides an abstraction which assumes that to
   * any couple (unit name, unit kind) we can associate at most one source file.
   * This means that several couples can be associated to the same source file,
   * but on the other hand, only one one source file can be associated to a
   * couple.
  
   * This is used to make the semantic analysis able to switch from one analysis
   * units to another.
   *)

  type t

  
        val for_project :
    ?project:string
    -> ?scenario_vars:(string * string) list
    -> ?target:string
    -> ?runtime:string
    -> string
    -> t
  (**
   * Load the project file at ``Project_File`` and return a unit provider that
   * uses it.
  
   * If ``Project`` is passed, use it to provide units, otherwise, try use the
   * whole project tree.
  
   * As unit providers must guarantee that there exists at most one source file
   * for each couple (unit name, unit kind), aggregate projects that contains
   * several conflicting units are not supported: trying to load one will yield
   * an error (see below).
  
   * If not ``None``, ``Scenario_Vars`` must point to an array of
   * ``ada_project_scenario_variable`` couples to provide scenario variables for
   * this project. The last element of this array must end with a ``{ None, None
   * }`` couple.
  
   * If not ``None``, ``target`` and ``runtime`` must point to valid NULL-
   * terminated strings.
  
   * If the requested project is invalid (error while opening the file, error
   * while analysing its syntax, ...), or if it is an unsupported aggregate
   * project, this returns ``None``.
   *)

  val auto : string list -> t
  (**
   * Return a unit provider that knows which compilation units are to be found
   * in the given list of source files.
  
   * This knowledge is built trying to parse all given input files as Ada source
   * files and listing the compilation units found there. Files that cannot be
   * parsed properly are discarded. If two compilation units are found for the
   * same unit, the first that is found in the given input files is taken and
   * the other ones are discarded.
  
   * Source files are decoded using the given charset. If it is ``None``, the
   * default charset (ISO-8859-1) is used.
  
   * .. admonition:: todo
  
   * Find a way to report discarded source files/compilation units.
   *)


end

module rec Entity : sig
  type t

  val info : t -> EntityInfo.t
end

and AdaNodeType : sig

  
   
  (**
    * AbortNodeType.t
    * AbstractNodeType.t
    * AdaListType.t
    * AliasedNodeType.t
    * AllNodeType.t
    * ArrayIndicesType.t
    * AspectAssocType.t
    * AspectClauseType.t
    * AspectSpecType.t
    * BaseAssocType.t
    * BaseFormalParamHolderType.t
    * BaseRecordDefType.t
    * BasicAssocType.t
    * BasicDeclType.t
    * CaseStmtAlternativeType.t
    * CompilationUnitType.t
    * ComponentClauseType.t
    * ComponentDefType.t
    * ConstantNodeType.t
    * ConstraintType.t
    * DeclarativePartType.t
    * ElsifExprPartType.t
    * ElsifStmtPartType.t
    * ExprType.t
    * HandledStmtsType.t
    * InterfaceKindType.t
    * IterTypeType.t
    * LibraryItemType.t
    * LimitedNodeType.t
    * LoopSpecType.t
    * ModeType.t
    * MultiAbstractStateDeclType.t
    * NotNullType.t
    * NullComponentDeclType.t
    * OthersDesignatorType.t
    * OverridingNodeType.t
    * ParamsType.t
    * ParenAbstractStateDeclType.t
    * PragmaNodeType.t
    * PrivateNodeType.t
    * ProtectedDefType.t
    * ProtectedNodeType.t
    * QuantifierType.t
    * RangeSpecType.t
    * RenamingClauseType.t
    * ReverseNodeType.t
    * SelectWhenPartType.t
    * StmtType.t
    * SubpKindType.t
    * SubunitType.t
    * SynchronizedNodeType.t
    * TaggedNodeType.t
    * TaskDefType.t
    * TypeDefType.t
    * TypeExprType.t
    * UnconstrainedArrayIndexType.t
    * UntilNodeType.t
    * UseClauseType.t
    * VariantType.t
    * VariantPartType.t
    * WithClauseType.t
    * WithPrivateType.t
    *)
  type t = [
    | `AbortAbsent
        of AbortAbsentType.fields
    | `AbortPresent
        of AbortPresentType.fields
    | `AbstractAbsent
        of AbstractAbsentType.fields
    | `AbstractPresent
        of AbstractPresentType.fields
    | `AdaNodeList
        of AdaNodeListType.fields
    | `AbstractStateDeclList
        of AbstractStateDeclListType.fields
    | `AlternativesList
        of AlternativesListType.fields
    | `ConstraintList
        of ConstraintListType.fields
    | `DeclList
        of DeclListType.fields
    | `StmtList
        of StmtListType.fields
    | `AspectAssocList
        of AspectAssocListType.fields
    | `BaseAssocList
        of BaseAssocListType.fields
    | `AssocList
        of AssocListType.fields
    | `BasicDeclList
        of BasicDeclListType.fields
    | `CaseExprAlternativeList
        of CaseExprAlternativeListType.fields
    | `CaseStmtAlternativeList
        of CaseStmtAlternativeListType.fields
    | `CompilationUnitList
        of CompilationUnitListType.fields
    | `ContractCaseAssocList
        of ContractCaseAssocListType.fields
    | `DefiningNameList
        of DefiningNameListType.fields
    | `DiscriminantSpecList
        of DiscriminantSpecListType.fields
    | `ElsifExprPartList
        of ElsifExprPartListType.fields
    | `ElsifStmtPartList
        of ElsifStmtPartListType.fields
    | `EnumLiteralDeclList
        of EnumLiteralDeclListType.fields
    | `ExprAlternativesList
        of ExprAlternativesListType.fields
    | `DiscriminantChoiceList
        of DiscriminantChoiceListType.fields
    | `NameList
        of NameListType.fields
    | `ParentList
        of ParentListType.fields
    | `ParamSpecList
        of ParamSpecListType.fields
    | `PragmaNodeList
        of PragmaNodeListType.fields
    | `SelectWhenPartList
        of SelectWhenPartListType.fields
    | `UnconstrainedArrayIndexList
        of UnconstrainedArrayIndexListType.fields
    | `VariantList
        of VariantListType.fields
    | `AliasedAbsent
        of AliasedAbsentType.fields
    | `AliasedPresent
        of AliasedPresentType.fields
    | `AllAbsent
        of AllAbsentType.fields
    | `AllPresent
        of AllPresentType.fields
    | `ConstrainedArrayIndices
        of ConstrainedArrayIndicesType.fields
    | `UnconstrainedArrayIndices
        of UnconstrainedArrayIndicesType.fields
    | `AspectAssoc
        of AspectAssocType.fields
    | `AtClause
        of AtClauseType.fields
    | `AttributeDefClause
        of AttributeDefClauseType.fields
    | `EnumRepClause
        of EnumRepClauseType.fields
    | `RecordRepClause
        of RecordRepClauseType.fields
    | `AspectSpec
        of AspectSpecType.fields
    | `ContractCaseAssoc
        of ContractCaseAssocType.fields
    | `PragmaArgumentAssoc
        of PragmaArgumentAssocType.fields
    | `EntrySpec
        of EntrySpecType.fields
    | `EnumSubpSpec
        of EnumSubpSpecType.fields
    | `SubpSpec
        of SubpSpecType.fields
    | `ComponentList
        of ComponentListType.fields
    | `KnownDiscriminantPart
        of KnownDiscriminantPartType.fields
    | `UnknownDiscriminantPart
        of UnknownDiscriminantPartType.fields
    | `EntryCompletionFormalParams
        of EntryCompletionFormalParamsType.fields
    | `GenericFormalPart
        of GenericFormalPartType.fields
    | `NullRecordDef
        of NullRecordDefType.fields
    | `RecordDef
        of RecordDefType.fields
    | `AggregateAssoc
        of AggregateAssocType.fields
    | `MultiDimArrayAssoc
        of MultiDimArrayAssocType.fields
    | `DiscriminantAssoc
        of DiscriminantAssocType.fields
    | `IteratedAssoc
        of IteratedAssocType.fields
    | `ParamAssoc
        of ParamAssocType.fields
    | `AbstractStateDecl
        of AbstractStateDeclType.fields
    | `AnonymousExprDecl
        of AnonymousExprDeclType.fields
    | `ComponentDecl
        of ComponentDeclType.fields
    | `DiscriminantSpec
        of DiscriminantSpecType.fields
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
    | `ParamSpec
        of ParamSpecType.fields
    | `GenericPackageInternal
        of GenericPackageInternalType.fields
    | `PackageDecl
        of PackageDeclType.fields
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
    | `ClasswideTypeDecl
        of ClasswideTypeDeclType.fields
    | `IncompleteTypeDecl
        of IncompleteTypeDeclType.fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
    | `ProtectedTypeDecl
        of ProtectedTypeDeclType.fields
    | `TaskTypeDecl
        of TaskTypeDeclType.fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
    | `TypeDecl
        of TypeDeclType.fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
    | `EntryDecl
        of EntryDeclType.fields
    | `EnumLiteralDecl
        of EnumLiteralDeclType.fields
    | `GenericSubpInternal
        of GenericSubpInternalType.fields
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
    | `EntryBody
        of EntryBodyType.fields
    | `PackageBody
        of PackageBodyType.fields
    | `ProtectedBody
        of ProtectedBodyType.fields
    | `TaskBody
        of TaskBodyType.fields
    | `EntryIndexSpec
        of EntryIndexSpecType.fields
    | `ErrorDecl
        of ErrorDeclType.fields
    | `ExceptionDecl
        of ExceptionDeclType.fields
    | `ExceptionHandler
        of ExceptionHandlerType.fields
    | `ForLoopVarDecl
        of ForLoopVarDeclType.fields
    | `GenericPackageDecl
        of GenericPackageDeclType.fields
    | `GenericSubpDecl
        of GenericSubpDeclType.fields
    | `GenericPackageInstantiation
        of GenericPackageInstantiationType.fields
    | `GenericSubpInstantiation
        of GenericSubpInstantiationType.fields
    | `GenericPackageRenamingDecl
        of GenericPackageRenamingDeclType.fields
    | `GenericSubpRenamingDecl
        of GenericSubpRenamingDeclType.fields
    | `LabelDecl
        of LabelDeclType.fields
    | `NamedStmtDecl
        of NamedStmtDeclType.fields
    | `NumberDecl
        of NumberDeclType.fields
    | `ObjectDecl
        of ObjectDeclType.fields
    | `ExtendedReturnStmtObjectDecl
        of ExtendedReturnStmtObjectDeclType.fields
    | `PackageRenamingDecl
        of PackageRenamingDeclType.fields
    | `SingleProtectedDecl
        of SingleProtectedDeclType.fields
    | `SingleTaskDecl
        of SingleTaskDeclType.fields
    | `CaseStmtAlternative
        of CaseStmtAlternativeType.fields
    | `CompilationUnit
        of CompilationUnitType.fields
    | `ComponentClause
        of ComponentClauseType.fields
    | `ComponentDef
        of ComponentDefType.fields
    | `ConstantAbsent
        of ConstantAbsentType.fields
    | `ConstantPresent
        of ConstantPresentType.fields
    | `DeltaConstraint
        of DeltaConstraintType.fields
    | `DigitsConstraint
        of DigitsConstraintType.fields
    | `DiscriminantConstraint
        of DiscriminantConstraintType.fields
    | `IndexConstraint
        of IndexConstraintType.fields
    | `RangeConstraint
        of RangeConstraintType.fields
    | `DeclarativePart
        of DeclarativePartType.fields
    | `PrivatePart
        of PrivatePartType.fields
    | `PublicPart
        of PublicPartType.fields
    | `ElsifExprPart
        of ElsifExprPartType.fields
    | `ElsifStmtPart
        of ElsifStmtPartType.fields
    | `AbstractStateDeclExpr
        of AbstractStateDeclExprType.fields
    | `Allocator
        of AllocatorType.fields
    | `Aggregate
        of AggregateType.fields
    | `BracketAggregate
        of BracketAggregateType.fields
    | `DeltaAggregate
        of DeltaAggregateType.fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
    | `NullRecordAggregate
        of NullRecordAggregateType.fields
    | `BinOp
        of BinOpType.fields
    | `RelationOp
        of RelationOpType.fields
    | `BoxExpr
        of BoxExprType.fields
    | `CaseExprAlternative
        of CaseExprAlternativeType.fields
    | `CaseExpr
        of CaseExprType.fields
    | `IfExpr
        of IfExprType.fields
    | `ContractCases
        of ContractCasesType.fields
    | `DeclExpr
        of DeclExprType.fields
    | `MembershipExpr
        of MembershipExprType.fields
    | `AttributeRef
        of AttributeRefType.fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
    | `CallExpr
        of CallExprType.fields
    | `DefiningName
        of DefiningNameType.fields
    | `DiscreteSubtypeName
        of DiscreteSubtypeNameType.fields
    | `DottedName
        of DottedNameType.fields
    | `EndName
        of EndNameType.fields
    | `ExplicitDeref
        of ExplicitDerefType.fields
    | `QualExpr
        of QualExprType.fields
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
    | `TargetName
        of TargetNameType.fields
    | `ParenExpr
        of ParenExprType.fields
    | `QuantifiedExpr
        of QuantifiedExprType.fields
    | `RaiseExpr
        of RaiseExprType.fields
    | `UnOp
        of UnOpType.fields
    | `HandledStmts
        of HandledStmtsType.fields
    | `InterfaceKindLimited
        of InterfaceKindLimitedType.fields
    | `InterfaceKindProtected
        of InterfaceKindProtectedType.fields
    | `InterfaceKindSynchronized
        of InterfaceKindSynchronizedType.fields
    | `InterfaceKindTask
        of InterfaceKindTaskType.fields
    | `IterTypeIn
        of IterTypeInType.fields
    | `IterTypeOf
        of IterTypeOfType.fields
    | `LibraryItem
        of LibraryItemType.fields
    | `LimitedAbsent
        of LimitedAbsentType.fields
    | `LimitedPresent
        of LimitedPresentType.fields
    | `ForLoopSpec
        of ForLoopSpecType.fields
    | `WhileLoopSpec
        of WhileLoopSpecType.fields
    | `ModeDefault
        of ModeDefaultType.fields
    | `ModeIn
        of ModeInType.fields
    | `ModeInOut
        of ModeInOutType.fields
    | `ModeOut
        of ModeOutType.fields
    | `MultiAbstractStateDecl
        of MultiAbstractStateDeclType.fields
    | `NotNullAbsent
        of NotNullAbsentType.fields
    | `NotNullPresent
        of NotNullPresentType.fields
    | `NullComponentDecl
        of NullComponentDeclType.fields
    | `OthersDesignator
        of OthersDesignatorType.fields
    | `OverridingNotOverriding
        of OverridingNotOverridingType.fields
    | `OverridingOverriding
        of OverridingOverridingType.fields
    | `OverridingUnspecified
        of OverridingUnspecifiedType.fields
    | `Params
        of ParamsType.fields
    | `ParenAbstractStateDecl
        of ParenAbstractStateDeclType.fields
    | `PragmaNode
        of PragmaNodeType.fields
    | `PrivateAbsent
        of PrivateAbsentType.fields
    | `PrivatePresent
        of PrivatePresentType.fields
    | `ProtectedDef
        of ProtectedDefType.fields
    | `ProtectedAbsent
        of ProtectedAbsentType.fields
    | `ProtectedPresent
        of ProtectedPresentType.fields
    | `QuantifierAll
        of QuantifierAllType.fields
    | `QuantifierSome
        of QuantifierSomeType.fields
    | `RangeSpec
        of RangeSpecType.fields
    | `RenamingClause
        of RenamingClauseType.fields
    | `SyntheticRenamingClause
        of SyntheticRenamingClauseType.fields
    | `ReverseAbsent
        of ReverseAbsentType.fields
    | `ReversePresent
        of ReversePresentType.fields
    | `SelectWhenPart
        of SelectWhenPartType.fields
    | `AcceptStmt
        of AcceptStmtType.fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
    | `CaseStmt
        of CaseStmtType.fields
    | `ExtendedReturnStmt
        of ExtendedReturnStmtType.fields
    | `IfStmt
        of IfStmtType.fields
    | `NamedStmt
        of NamedStmtType.fields
    | `SelectStmt
        of SelectStmtType.fields
    | `ErrorStmt
        of ErrorStmtType.fields
    | `AbortStmt
        of AbortStmtType.fields
    | `AssignStmt
        of AssignStmtType.fields
    | `CallStmt
        of CallStmtType.fields
    | `DelayStmt
        of DelayStmtType.fields
    | `ExitStmt
        of ExitStmtType.fields
    | `GotoStmt
        of GotoStmtType.fields
    | `Label
        of LabelType.fields
    | `NullStmt
        of NullStmtType.fields
    | `RaiseStmt
        of RaiseStmtType.fields
    | `RequeueStmt
        of RequeueStmtType.fields
    | `ReturnStmt
        of ReturnStmtType.fields
    | `TerminateAlternative
        of TerminateAlternativeType.fields
    | `SubpKindFunction
        of SubpKindFunctionType.fields
    | `SubpKindProcedure
        of SubpKindProcedureType.fields
    | `Subunit
        of SubunitType.fields
    | `SynchronizedAbsent
        of SynchronizedAbsentType.fields
    | `SynchronizedPresent
        of SynchronizedPresentType.fields
    | `TaggedAbsent
        of TaggedAbsentType.fields
    | `TaggedPresent
        of TaggedPresentType.fields
    | `TaskDef
        of TaskDefType.fields
    | `AccessToSubpDef
        of AccessToSubpDefType.fields
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
    | `ArrayTypeDef
        of ArrayTypeDefType.fields
    | `DerivedTypeDef
        of DerivedTypeDefType.fields
    | `EnumTypeDef
        of EnumTypeDefType.fields
    | `FormalDiscreteTypeDef
        of FormalDiscreteTypeDefType.fields
    | `InterfaceTypeDef
        of InterfaceTypeDefType.fields
    | `ModIntTypeDef
        of ModIntTypeDefType.fields
    | `PrivateTypeDef
        of PrivateTypeDefType.fields
    | `DecimalFixedPointDef
        of DecimalFixedPointDefType.fields
    | `FloatingPointDef
        of FloatingPointDefType.fields
    | `OrdinaryFixedPointDef
        of OrdinaryFixedPointDefType.fields
    | `RecordTypeDef
        of RecordTypeDefType.fields
    | `SignedIntTypeDef
        of SignedIntTypeDefType.fields
    | `AnonymousType
        of AnonymousTypeType.fields
    | `EnumLitSynthTypeExpr
        of EnumLitSynthTypeExprType.fields
    | `SubtypeIndication
        of SubtypeIndicationType.fields
    | `ConstrainedSubtypeIndication
        of ConstrainedSubtypeIndicationType.fields
    | `DiscreteSubtypeIndication
        of DiscreteSubtypeIndicationType.fields
    | `UnconstrainedArrayIndex
        of UnconstrainedArrayIndexType.fields
    | `UntilAbsent
        of UntilAbsentType.fields
    | `UntilPresent
        of UntilPresentType.fields
    | `UsePackageClause
        of UsePackageClauseType.fields
    | `UseTypeClause
        of UseTypeClauseType.fields
    | `Variant
        of VariantType.fields
    | `VariantPart
        of VariantPartType.fields
    | `WithClause
        of WithClauseType.fields
    | `WithPrivateAbsent
        of WithPrivateAbsentType.fields
    | `WithPrivatePresent
        of WithPrivatePresentType.fields
  ]


end
and AbortNodeType : sig

  
   
  type t = [
    | `AbortAbsent
        of AbortAbsentType.fields
    | `AbortPresent
        of AbortPresentType.fields
  ]


end
and AbortAbsentType : sig

  
   
  type t = [
    | `AbortAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AbortPresentType : sig

  
   
  type t = [
    | `AbortPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AbstractNodeType : sig

  
   
  type t = [
    | `AbstractAbsent
        of AbstractAbsentType.fields
    | `AbstractPresent
        of AbstractPresentType.fields
  ]


end
and AbstractAbsentType : sig

  
   
  type t = [
    | `AbstractAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AbstractPresentType : sig

  
   
  type t = [
    | `AbstractPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AdaListType : sig

  
   
  (**
    * AdaNodeListType.t
    * AspectAssocListType.t
    * BaseAssocListType.t
    * BasicAssocListType.t
    * BasicDeclListType.t
    * CaseExprAlternativeListType.t
    * CaseStmtAlternativeListType.t
    * CompilationUnitListType.t
    * ContractCaseAssocListType.t
    * DefiningNameListType.t
    * DiscriminantSpecListType.t
    * ElsifExprPartListType.t
    * ElsifStmtPartListType.t
    * EnumLiteralDeclListType.t
    * ExprListType.t
    * IdentifierListType.t
    * NameListType.t
    * ParamSpecListType.t
    * PragmaNodeListType.t
    * SelectWhenPartListType.t
    * UnconstrainedArrayIndexListType.t
    * VariantListType.t
    *)
  type t = [
    | `AdaNodeList
        of AdaNodeListType.fields
    | `AbstractStateDeclList
        of AbstractStateDeclListType.fields
    | `AlternativesList
        of AlternativesListType.fields
    | `ConstraintList
        of ConstraintListType.fields
    | `DeclList
        of DeclListType.fields
    | `StmtList
        of StmtListType.fields
    | `AspectAssocList
        of AspectAssocListType.fields
    | `BaseAssocList
        of BaseAssocListType.fields
    | `AssocList
        of AssocListType.fields
    | `BasicDeclList
        of BasicDeclListType.fields
    | `CaseExprAlternativeList
        of CaseExprAlternativeListType.fields
    | `CaseStmtAlternativeList
        of CaseStmtAlternativeListType.fields
    | `CompilationUnitList
        of CompilationUnitListType.fields
    | `ContractCaseAssocList
        of ContractCaseAssocListType.fields
    | `DefiningNameList
        of DefiningNameListType.fields
    | `DiscriminantSpecList
        of DiscriminantSpecListType.fields
    | `ElsifExprPartList
        of ElsifExprPartListType.fields
    | `ElsifStmtPartList
        of ElsifStmtPartListType.fields
    | `EnumLiteralDeclList
        of EnumLiteralDeclListType.fields
    | `ExprAlternativesList
        of ExprAlternativesListType.fields
    | `DiscriminantChoiceList
        of DiscriminantChoiceListType.fields
    | `NameList
        of NameListType.fields
    | `ParentList
        of ParentListType.fields
    | `ParamSpecList
        of ParamSpecListType.fields
    | `PragmaNodeList
        of PragmaNodeListType.fields
    | `SelectWhenPartList
        of SelectWhenPartListType.fields
    | `UnconstrainedArrayIndexList
        of UnconstrainedArrayIndexListType.fields
    | `VariantList
        of VariantListType.fields
  ]


end
and AdaNodeListType : sig

  
   
  type t = [
    | `AdaNodeList
        of fields
    | `AbstractStateDeclList
        of AbstractStateDeclListType.fields
    | `AlternativesList
        of AlternativesListType.fields
    | `ConstraintList
        of ConstraintListType.fields
    | `DeclList
        of DeclListType.fields
    | `StmtList
        of StmtListType.fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AbstractStateDeclListType : sig

  
   
  type t = [
    | `AbstractStateDeclList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AlternativesListType : sig

  
   
  type t = [
    | `AlternativesList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ConstraintListType : sig

  
   
  type t = [
    | `ConstraintList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DeclListType : sig

  
   
  type t = [
    | `DeclList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and StmtListType : sig

  
   
  type t = [
    | `StmtList
        of fields
  ]
  and fields = 
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AspectAssocListType : sig

  
   
  type t = [
    | `AspectAssocList
        of fields
  ]
  and fields = 
  {
    list : AspectAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BaseAssocListType : sig

  
   
  type t = [
    | `BaseAssocList
        of fields
  ]
  and fields = 
  {
    list : BaseAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BasicAssocListType : sig

  
   
  type t = [
    | `AssocList
        of AssocListType.fields
  ]


end
and AssocListType : sig

  
   
  type t = [
    | `AssocList
        of fields
  ]
  and fields = 
  {
    list : BasicAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BasicDeclListType : sig

  
   
  type t = [
    | `BasicDeclList
        of fields
  ]
  and fields = 
  {
    list : BasicDeclType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and CaseExprAlternativeListType : sig

  
   
  type t = [
    | `CaseExprAlternativeList
        of fields
  ]
  and fields = 
  {
    list : CaseExprAlternativeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and CaseStmtAlternativeListType : sig

  
   
  type t = [
    | `CaseStmtAlternativeList
        of fields
  ]
  and fields = 
  {
    list : CaseStmtAlternativeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and CompilationUnitListType : sig

  
   
  type t = [
    | `CompilationUnitList
        of fields
  ]
  and fields = 
  {
    list : CompilationUnitType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ContractCaseAssocListType : sig

  
   
  type t = [
    | `ContractCaseAssocList
        of fields
  ]
  and fields = 
  {
    list : ContractCaseAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DefiningNameListType : sig

  
   
  type t = [
    | `DefiningNameList
        of fields
  ]
  and fields = 
  {
    list : DefiningNameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DiscriminantSpecListType : sig

  
   
  type t = [
    | `DiscriminantSpecList
        of fields
  ]
  and fields = 
  {
    list : DiscriminantSpecType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ElsifExprPartListType : sig

  
   
  type t = [
    | `ElsifExprPartList
        of fields
  ]
  and fields = 
  {
    list : ElsifExprPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ElsifStmtPartListType : sig

  
   
  type t = [
    | `ElsifStmtPartList
        of fields
  ]
  and fields = 
  {
    list : ElsifStmtPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and EnumLiteralDeclListType : sig

  
   
  type t = [
    | `EnumLiteralDeclList
        of fields
  ]
  and fields = 
  {
    list : EnumLiteralDeclType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ExprListType : sig

  
   
  type t = [
    | `ExprAlternativesList
        of ExprAlternativesListType.fields
  ]


end
and ExprAlternativesListType : sig

  
   
  type t = [
    | `ExprAlternativesList
        of fields
  ]
  and fields = 
  {
    list : ExprType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and IdentifierListType : sig

  
   
  type t = [
    | `DiscriminantChoiceList
        of DiscriminantChoiceListType.fields
  ]


end
and DiscriminantChoiceListType : sig

  
   
  type t = [
    | `DiscriminantChoiceList
        of fields
  ]
  and fields = 
  {
    list : IdentifierType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NameListType : sig

  
   
  type t = [
    | `NameList
        of fields
    | `ParentList
        of ParentListType.fields
  ]
  and fields = 
  {
    list : NameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ParentListType : sig

  
   
  type t = [
    | `ParentList
        of fields
  ]
  and fields = 
  {
    list : NameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ParamSpecListType : sig

  
   
  type t = [
    | `ParamSpecList
        of fields
  ]
  and fields = 
  {
    list : ParamSpecType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and PragmaNodeListType : sig

  
   
  type t = [
    | `PragmaNodeList
        of fields
  ]
  and fields = 
  {
    list : PragmaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SelectWhenPartListType : sig

  
   
  type t = [
    | `SelectWhenPartList
        of fields
  ]
  and fields = 
  {
    list : SelectWhenPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and UnconstrainedArrayIndexListType : sig

  
   
  type t = [
    | `UnconstrainedArrayIndexList
        of fields
  ]
  and fields = 
  {
    list : UnconstrainedArrayIndexType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and VariantListType : sig

  
   
  type t = [
    | `VariantList
        of fields
  ]
  and fields = 
  {
    list : VariantType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AliasedNodeType : sig

  
   
  type t = [
    | `AliasedAbsent
        of AliasedAbsentType.fields
    | `AliasedPresent
        of AliasedPresentType.fields
  ]


end
and AliasedAbsentType : sig

  
   
  type t = [
    | `AliasedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AliasedPresentType : sig

  
   
  type t = [
    | `AliasedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AllNodeType : sig

  
   
  type t = [
    | `AllAbsent
        of AllAbsentType.fields
    | `AllPresent
        of AllPresentType.fields
  ]


end
and AllAbsentType : sig

  
   
  type t = [
    | `AllAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AllPresentType : sig

  
   
  type t = [
    | `AllPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ArrayIndicesType : sig

  
   
  type t = [
    | `ConstrainedArrayIndices
        of ConstrainedArrayIndicesType.fields
    | `UnconstrainedArrayIndices
        of UnconstrainedArrayIndicesType.fields
  ]


end
and ConstrainedArrayIndicesType : sig

  
   
  type t = [
    | `ConstrainedArrayIndices
        of fields
  ]
  and fields = 
  {
         
    f_list: ConstraintListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and UnconstrainedArrayIndicesType : sig

  
   
  type t = [
    | `UnconstrainedArrayIndices
        of fields
  ]
  and fields = 
  {
         
    f_types: UnconstrainedArrayIndexListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AspectAssocType : sig

  
   
  type t = [
    | `AspectAssoc
        of fields
  ]
  and fields = 
  {
         
    f_id: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `AbstractStateDeclExpr
          of AbstractStateDeclExprType.fields
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ContractCases
          of ContractCasesType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AspectClauseType : sig

  
   
  type t = [
    | `AtClause
        of AtClauseType.fields
    | `AttributeDefClause
        of AttributeDefClauseType.fields
    | `EnumRepClause
        of EnumRepClauseType.fields
    | `RecordRepClause
        of RecordRepClauseType.fields
  ]


end
and AtClauseType : sig

  
   
  type t = [
    | `AtClause
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AttributeDefClauseType : sig

  
   
  type t = [
    | `AttributeDefClause
        of fields
  ]
  and fields = 
  {
         
    f_attribute_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and EnumRepClauseType : sig

  
   
  type t = [
    | `EnumRepClause
        of fields
  ]
  and fields = 
  {
         
    f_type_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_aggregate: BaseAggregateType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RecordRepClauseType : sig

  
   
  type t = [
    | `RecordRepClause
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_at_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_components: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AspectSpecType : sig

  
   
  type t = [
    | `AspectSpec
        of fields
  ]
  and fields = 
  {
         
    f_aspect_assocs: AspectAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BaseAssocType : sig

  
   
  type t = [
    | `ContractCaseAssoc
        of ContractCaseAssocType.fields
    | `PragmaArgumentAssoc
        of PragmaArgumentAssocType.fields
  ]


end
and ContractCaseAssocType : sig

  
   
  type t = [
    | `ContractCaseAssoc
        of fields
  ]
  and fields = 
  {
         
    f_guard: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `OthersDesignator
          of OthersDesignatorType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_consequence: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and PragmaArgumentAssocType : sig

  
   
  type t = [
    | `PragmaArgumentAssoc
        of fields
  ]
  and fields = 
  {
         
    f_id: IdentifierType.t
    option
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BaseFormalParamHolderType : sig

  
   
  (**
    * BaseSubpSpecType.t
    * ComponentListType.t
    * DiscriminantPartType.t
    * EntryCompletionFormalParamsType.t
    * GenericFormalPartType.t
    *)
  type t = [
    | `EntrySpec
        of EntrySpecType.fields
    | `EnumSubpSpec
        of EnumSubpSpecType.fields
    | `SubpSpec
        of SubpSpecType.fields
    | `ComponentList
        of ComponentListType.fields
    | `KnownDiscriminantPart
        of KnownDiscriminantPartType.fields
    | `UnknownDiscriminantPart
        of UnknownDiscriminantPartType.fields
    | `EntryCompletionFormalParams
        of EntryCompletionFormalParamsType.fields
    | `GenericFormalPart
        of GenericFormalPartType.fields
  ]


end
and BaseSubpSpecType : sig

  
   
  type t = [
    | `EntrySpec
        of EntrySpecType.fields
    | `EnumSubpSpec
        of EnumSubpSpecType.fields
    | `SubpSpec
        of SubpSpecType.fields
  ]


end
and EntrySpecType : sig

  
   
  type t = [
    | `EntrySpec
        of fields
  ]
  and fields = 
  {
         
    f_entry_name: DefiningNameType.t
    Lazy.t;
         
    f_family_type: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_entry_params: ParamsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and EnumSubpSpecType : sig

  
   
  type t = [
    | `EnumSubpSpec
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SubpSpecType : sig

  
   
  type t = [
    | `SubpSpec
        of fields
  ]
  and fields = 
  {
         
    f_subp_kind: SubpKindType.t
    Lazy.t;
         
    f_subp_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_subp_params: ParamsType.t
    option
    Lazy.t;
         
    f_subp_returns: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ComponentListType : sig

  
   
  type t = [
    | `ComponentList
        of fields
  ]
  and fields = 
  {
         
    f_components: AdaNodeListType.t
    option
    Lazy.t;
         
    f_variant_part: VariantPartType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DiscriminantPartType : sig

  
   
  type t = [
    | `KnownDiscriminantPart
        of KnownDiscriminantPartType.fields
    | `UnknownDiscriminantPart
        of UnknownDiscriminantPartType.fields
  ]


end
and KnownDiscriminantPartType : sig

  
   
  type t = [
    | `KnownDiscriminantPart
        of fields
  ]
  and fields = 
  {
         
    f_discr_specs: DiscriminantSpecListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and UnknownDiscriminantPartType : sig

  
   
  type t = [
    | `UnknownDiscriminantPart
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and EntryCompletionFormalParamsType : sig

  
   
  type t = [
    | `EntryCompletionFormalParams
        of fields
  ]
  and fields = 
  {
         
    f_params: ParamsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericFormalPartType : sig

  
   
  type t = [
    | `GenericFormalPart
        of fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BaseRecordDefType : sig

  
   
  type t = [
    | `NullRecordDef
        of NullRecordDefType.fields
    | `RecordDef
        of RecordDefType.fields
  ]


end
and NullRecordDefType : sig

  
   
  type t = [
    | `NullRecordDef
        of fields
  ]
  and fields = 
  {
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RecordDefType : sig

  
   
  type t = [
    | `RecordDef
        of fields
  ]
  and fields = 
  {
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BasicAssocType : sig

  
   
  (**
    * AggregateAssocType.t
    * DiscriminantAssocType.t
    * IteratedAssocType.t
    * ParamAssocType.t
    *)
  type t = [
    | `AggregateAssoc
        of AggregateAssocType.fields
    | `MultiDimArrayAssoc
        of MultiDimArrayAssocType.fields
    | `DiscriminantAssoc
        of DiscriminantAssocType.fields
    | `IteratedAssoc
        of IteratedAssocType.fields
    | `ParamAssoc
        of ParamAssocType.fields
  ]


end
and AggregateAssocType : sig

  
   
  type t = [
    | `AggregateAssoc
        of fields
    | `MultiDimArrayAssoc
        of MultiDimArrayAssocType.fields
  ]
  and fields = 
  {
         
    f_designators: AlternativesListType.t
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and MultiDimArrayAssocType : sig

  
   
  type t = [
    | `MultiDimArrayAssoc
        of fields
  ]
  and fields = 
  {
         
    f_designators: AlternativesListType.t
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DiscriminantAssocType : sig

  
   
  type t = [
    | `DiscriminantAssoc
        of fields
  ]
  and fields = 
  {
         
    f_ids: IdentifierListType.t
    option
    Lazy.t;
         
    f_discr_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and IteratedAssocType : sig

  
   
  type t = [
    | `IteratedAssoc
        of fields
  ]
  and fields = 
  {
         
    f_spec: ForLoopSpecType.t
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ParamAssocType : sig

  
   
  type t = [
    | `ParamAssoc
        of fields
  ]
  and fields = 
  {
         
    f_designator: [
      | `Identifier
          of IdentifierType.fields
      | `OthersDesignator
          of OthersDesignatorType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BasicDeclType : sig

  
   
  (**
    * AbstractStateDeclType.t
    * AnonymousExprDeclType.t
    * BaseFormalParamDeclType.t
    * BasePackageDeclType.t
    * BaseTypeDeclType.t
    * BasicSubpDeclType.t
    * BodyNodeType.t
    * EntryIndexSpecType.t
    * ErrorDeclType.t
    * ExceptionDeclType.t
    * ExceptionHandlerType.t
    * ForLoopVarDeclType.t
    * GenericDeclType.t
    * GenericInstantiationType.t
    * GenericRenamingDeclType.t
    * LabelDeclType.t
    * NamedStmtDeclType.t
    * NumberDeclType.t
    * ObjectDeclType.t
    * PackageRenamingDeclType.t
    * SingleProtectedDeclType.t
    * SingleTaskDeclType.t
    *)
  type t = [
    | `AbstractStateDecl
        of AbstractStateDeclType.fields
    | `AnonymousExprDecl
        of AnonymousExprDeclType.fields
    | `ComponentDecl
        of ComponentDeclType.fields
    | `DiscriminantSpec
        of DiscriminantSpecType.fields
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
    | `ParamSpec
        of ParamSpecType.fields
    | `GenericPackageInternal
        of GenericPackageInternalType.fields
    | `PackageDecl
        of PackageDeclType.fields
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
    | `ClasswideTypeDecl
        of ClasswideTypeDeclType.fields
    | `IncompleteTypeDecl
        of IncompleteTypeDeclType.fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
    | `ProtectedTypeDecl
        of ProtectedTypeDeclType.fields
    | `TaskTypeDecl
        of TaskTypeDeclType.fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
    | `TypeDecl
        of TypeDeclType.fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
    | `EntryDecl
        of EntryDeclType.fields
    | `EnumLiteralDecl
        of EnumLiteralDeclType.fields
    | `GenericSubpInternal
        of GenericSubpInternalType.fields
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
    | `EntryBody
        of EntryBodyType.fields
    | `PackageBody
        of PackageBodyType.fields
    | `ProtectedBody
        of ProtectedBodyType.fields
    | `TaskBody
        of TaskBodyType.fields
    | `EntryIndexSpec
        of EntryIndexSpecType.fields
    | `ErrorDecl
        of ErrorDeclType.fields
    | `ExceptionDecl
        of ExceptionDeclType.fields
    | `ExceptionHandler
        of ExceptionHandlerType.fields
    | `ForLoopVarDecl
        of ForLoopVarDeclType.fields
    | `GenericPackageDecl
        of GenericPackageDeclType.fields
    | `GenericSubpDecl
        of GenericSubpDeclType.fields
    | `GenericPackageInstantiation
        of GenericPackageInstantiationType.fields
    | `GenericSubpInstantiation
        of GenericSubpInstantiationType.fields
    | `GenericPackageRenamingDecl
        of GenericPackageRenamingDeclType.fields
    | `GenericSubpRenamingDecl
        of GenericSubpRenamingDeclType.fields
    | `LabelDecl
        of LabelDeclType.fields
    | `NamedStmtDecl
        of NamedStmtDeclType.fields
    | `NumberDecl
        of NumberDeclType.fields
    | `ObjectDecl
        of ObjectDeclType.fields
    | `ExtendedReturnStmtObjectDecl
        of ExtendedReturnStmtObjectDeclType.fields
    | `PackageRenamingDecl
        of PackageRenamingDeclType.fields
    | `SingleProtectedDecl
        of SingleProtectedDeclType.fields
    | `SingleTaskDecl
        of SingleTaskDeclType.fields
  ]


end
and AbstractStateDeclType : sig

  
   
  type t = [
    | `AbstractStateDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AnonymousExprDeclType : sig

  
   
  type t = [
    | `AnonymousExprDecl
        of fields
  ]
  and fields = 
  {
         
    f_expr: ExprType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BaseFormalParamDeclType : sig

  
   
  (**
    * ComponentDeclType.t
    * DiscriminantSpecType.t
    * GenericFormalType.t
    * ParamSpecType.t
    *)
  type t = [
    | `ComponentDecl
        of ComponentDeclType.fields
    | `DiscriminantSpec
        of DiscriminantSpecType.fields
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
    | `ParamSpec
        of ParamSpecType.fields
  ]


end
and ComponentDeclType : sig

  
   
  type t = [
    | `ComponentDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_component_def: ComponentDefType.t
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DiscriminantSpecType : sig

  
   
  type t = [
    | `DiscriminantSpec
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericFormalType : sig

  
   
  type t = [
    | `GenericFormalObjDecl
        of GenericFormalObjDeclType.fields
    | `GenericFormalPackage
        of GenericFormalPackageType.fields
    | `GenericFormalSubpDecl
        of GenericFormalSubpDeclType.fields
    | `GenericFormalTypeDecl
        of GenericFormalTypeDeclType.fields
  ]


end
and GenericFormalObjDeclType : sig

  
   
  type t = [
    | `GenericFormalObjDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericFormalPackageType : sig

  
   
  type t = [
    | `GenericFormalPackage
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericFormalSubpDeclType : sig

  
   
  type t = [
    | `GenericFormalSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericFormalTypeDeclType : sig

  
   
  type t = [
    | `GenericFormalTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ParamSpecType : sig

  
   
  type t = [
    | `ParamSpec
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BasePackageDeclType : sig

  
   
  type t = [
    | `GenericPackageInternal
        of GenericPackageInternalType.fields
    | `PackageDecl
        of PackageDeclType.fields
  ]


end
and GenericPackageInternalType : sig

  
   
  type t = [
    | `GenericPackageInternal
        of fields
  ]
  and fields = 
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and PackageDeclType : sig

  
   
  type t = [
    | `PackageDecl
        of fields
  ]
  and fields = 
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BaseTypeDeclType : sig

  
   
  (**
    * BaseSubtypeDeclType.t
    * ClasswideTypeDeclType.t
    * IncompleteTypeDeclType.t
    * ProtectedTypeDeclType.t
    * TaskTypeDeclType.t
    * TypeDeclType.t
    *)
  type t = [
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
    | `ClasswideTypeDecl
        of ClasswideTypeDeclType.fields
    | `IncompleteTypeDecl
        of IncompleteTypeDeclType.fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
    | `ProtectedTypeDecl
        of ProtectedTypeDeclType.fields
    | `TaskTypeDecl
        of TaskTypeDeclType.fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
    | `TypeDecl
        of TypeDeclType.fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
  ]


end
and BaseSubtypeDeclType : sig

  
   
  type t = [
    | `DiscreteBaseSubtypeDecl
        of DiscreteBaseSubtypeDeclType.fields
    | `SubtypeDecl
        of SubtypeDeclType.fields
  ]


end
and DiscreteBaseSubtypeDeclType : sig

  
   
  type t = [
    | `DiscreteBaseSubtypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SubtypeDeclType : sig

  
   
  type t = [
    | `SubtypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_subtype: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ClasswideTypeDeclType : sig

  
   
  type t = [
    | `ClasswideTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and IncompleteTypeDeclType : sig

  
   
  type t = [
    | `IncompleteTypeDecl
        of fields
    | `IncompleteTaggedTypeDecl
        of IncompleteTaggedTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and IncompleteTaggedTypeDeclType : sig

  
   
  type t = [
    | `IncompleteTaggedTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ProtectedTypeDeclType : sig

  
   
  type t = [
    | `ProtectedTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_definition: ProtectedDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TaskTypeDeclType : sig

  
   
  type t = [
    | `TaskTypeDecl
        of fields
    | `SingleTaskTypeDecl
        of SingleTaskTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_definition: TaskDefType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SingleTaskTypeDeclType : sig

  
   
  type t = [
    | `SingleTaskTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_definition: TaskDefType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TypeDeclType : sig

  
   
  type t = [
    | `TypeDecl
        of fields
    | `AnonymousTypeDecl
        of AnonymousTypeDeclType.fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AnonymousTypeDeclType : sig

  
   
  type t = [
    | `AnonymousTypeDecl
        of fields
    | `SynthAnonymousTypeDecl
        of SynthAnonymousTypeDeclType.fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SynthAnonymousTypeDeclType : sig

  
   
  type t = [
    | `SynthAnonymousTypeDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BasicSubpDeclType : sig

  
   
  (**
    * ClassicSubpDeclType.t
    * EntryDeclType.t
    * EnumLiteralDeclType.t
    * GenericSubpInternalType.t
    *)
  type t = [
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
    | `EntryDecl
        of EntryDeclType.fields
    | `EnumLiteralDecl
        of EnumLiteralDeclType.fields
    | `GenericSubpInternal
        of GenericSubpInternalType.fields
  ]


end
and ClassicSubpDeclType : sig

  
   
  (**
    * AbstractSubpDeclType.t
    * FormalSubpDeclType.t
    * SubpDeclType.t
    *)
  type t = [
    | `AbstractSubpDecl
        of AbstractSubpDeclType.fields
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
    | `SubpDecl
        of SubpDeclType.fields
  ]


end
and AbstractSubpDeclType : sig

  
   
  type t = [
    | `AbstractSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and FormalSubpDeclType : sig

  
   
  type t = [
    | `AbstractFormalSubpDecl
        of AbstractFormalSubpDeclType.fields
    | `ConcreteFormalSubpDecl
        of ConcreteFormalSubpDeclType.fields
  ]


end
and AbstractFormalSubpDeclType : sig

  
   
  type t = [
    | `AbstractFormalSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_default_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ConcreteFormalSubpDeclType : sig

  
   
  type t = [
    | `ConcreteFormalSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_default_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SubpDeclType : sig

  
   
  type t = [
    | `SubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and EntryDeclType : sig

  
   
  type t = [
    | `EntryDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_spec: EntrySpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and EnumLiteralDeclType : sig

  
   
  type t = [
    | `EnumLiteralDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericSubpInternalType : sig

  
   
  type t = [
    | `GenericSubpInternal
        of fields
  ]
  and fields = 
  {
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BodyNodeType : sig

  
   
  (**
    * BaseSubpBodyType.t
    * BodyStubType.t
    * EntryBodyType.t
    * PackageBodyType.t
    * ProtectedBodyType.t
    * TaskBodyType.t
    *)
  type t = [
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
    | `EntryBody
        of EntryBodyType.fields
    | `PackageBody
        of PackageBodyType.fields
    | `ProtectedBody
        of ProtectedBodyType.fields
    | `TaskBody
        of TaskBodyType.fields
  ]


end
and BaseSubpBodyType : sig

  
   
  type t = [
    | `ExprFunction
        of ExprFunctionType.fields
    | `NullSubpDecl
        of NullSubpDeclType.fields
    | `SubpBody
        of SubpBodyType.fields
    | `SubpRenamingDecl
        of SubpRenamingDeclType.fields
  ]


end
and ExprFunctionType : sig

  
   
  type t = [
    | `ExprFunction
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NullSubpDeclType : sig

  
   
  type t = [
    | `NullSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SubpBodyType : sig

  
   
  type t = [
    | `SubpBody
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SubpRenamingDeclType : sig

  
   
  type t = [
    | `SubpRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BodyStubType : sig

  
   
  type t = [
    | `PackageBodyStub
        of PackageBodyStubType.fields
    | `ProtectedBodyStub
        of ProtectedBodyStubType.fields
    | `SubpBodyStub
        of SubpBodyStubType.fields
    | `TaskBodyStub
        of TaskBodyStubType.fields
  ]


end
and PackageBodyStubType : sig

  
   
  type t = [
    | `PackageBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ProtectedBodyStubType : sig

  
   
  type t = [
    | `ProtectedBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SubpBodyStubType : sig

  
   
  type t = [
    | `SubpBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TaskBodyStubType : sig

  
   
  type t = [
    | `TaskBodyStub
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and EntryBodyType : sig

  
   
  type t = [
    | `EntryBody
        of fields
  ]
  and fields = 
  {
         
    f_entry_name: DefiningNameType.t
    Lazy.t;
         
    f_index_spec: EntryIndexSpecType.t
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
         
    f_barrier: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and PackageBodyType : sig

  
   
  type t = [
    | `PackageBody
        of fields
  ]
  and fields = 
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ProtectedBodyType : sig

  
   
  type t = [
    | `ProtectedBody
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TaskBodyType : sig

  
   
  type t = [
    | `TaskBody
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and EntryIndexSpecType : sig

  
   
  type t = [
    | `EntryIndexSpec
        of fields
  ]
  and fields = 
  {
         
    f_id: DefiningNameType.t
    Lazy.t;
         
    f_subtype: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ErrorDeclType : sig

  
   
  type t = [
    | `ErrorDecl
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ExceptionDeclType : sig

  
   
  type t = [
    | `ExceptionDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ExceptionHandlerType : sig

  
   
  type t = [
    | `ExceptionHandler
        of fields
  ]
  and fields = 
  {
         
    f_exception_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_handled_exceptions: AlternativesListType.t
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ForLoopVarDeclType : sig

  
   
  type t = [
    | `ForLoopVarDecl
        of fields
  ]
  and fields = 
  {
         
    f_id: DefiningNameType.t
    Lazy.t;
         
    f_id_type: SubtypeIndicationType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericDeclType : sig

  
   
  type t = [
    | `GenericPackageDecl
        of GenericPackageDeclType.fields
    | `GenericSubpDecl
        of GenericSubpDeclType.fields
  ]


end
and GenericPackageDeclType : sig

  
   
  type t = [
    | `GenericPackageDecl
        of fields
  ]
  and fields = 
  {
         
    f_formal_part: GenericFormalPartType.t
    Lazy.t;
         
    f_package_decl: GenericPackageInternalType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericSubpDeclType : sig

  
   
  type t = [
    | `GenericSubpDecl
        of fields
  ]
  and fields = 
  {
         
    f_formal_part: GenericFormalPartType.t
    Lazy.t;
         
    f_subp_decl: GenericSubpInternalType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericInstantiationType : sig

  
   
  type t = [
    | `GenericPackageInstantiation
        of GenericPackageInstantiationType.fields
    | `GenericSubpInstantiation
        of GenericSubpInstantiationType.fields
  ]


end
and GenericPackageInstantiationType : sig

  
   
  type t = [
    | `GenericPackageInstantiation
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_generic_pkg_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_params: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericSubpInstantiationType : sig

  
   
  type t = [
    | `GenericSubpInstantiation
        of fields
  ]
  and fields = 
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_kind: SubpKindType.t
    Lazy.t;
         
    f_subp_name: DefiningNameType.t
    Lazy.t;
         
    f_generic_subp_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_params: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericRenamingDeclType : sig

  
   
  type t = [
    | `GenericPackageRenamingDecl
        of GenericPackageRenamingDeclType.fields
    | `GenericSubpRenamingDecl
        of GenericSubpRenamingDeclType.fields
  ]


end
and GenericPackageRenamingDeclType : sig

  
   
  type t = [
    | `GenericPackageRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GenericSubpRenamingDeclType : sig

  
   
  type t = [
    | `GenericSubpRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_kind: SubpKindType.t
    Lazy.t;
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and LabelDeclType : sig

  
   
  type t = [
    | `LabelDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NamedStmtDeclType : sig

  
   
  type t = [
    | `NamedStmtDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NumberDeclType : sig

  
   
  type t = [
    | `NumberDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ObjectDeclType : sig

  
   
  type t = [
    | `ObjectDecl
        of fields
    | `ExtendedReturnStmtObjectDecl
        of ExtendedReturnStmtObjectDeclType.fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_renaming_clause: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ExtendedReturnStmtObjectDeclType : sig

  
   
  type t = [
    | `ExtendedReturnStmtObjectDecl
        of fields
  ]
  and fields = 
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_renaming_clause: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and PackageRenamingDeclType : sig

  
   
  type t = [
    | `PackageRenamingDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SingleProtectedDeclType : sig

  
   
  type t = [
    | `SingleProtectedDecl
        of fields
  ]
  and fields = 
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_definition: ProtectedDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SingleTaskDeclType : sig

  
   
  type t = [
    | `SingleTaskDecl
        of fields
  ]
  and fields = 
  {
         
    f_task_type: SingleTaskTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and CaseStmtAlternativeType : sig

  
   
  type t = [
    | `CaseStmtAlternative
        of fields
  ]
  and fields = 
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and CompilationUnitType : sig

  
   
  type t = [
    | `CompilationUnit
        of fields
  ]
  and fields = 
  {
         
    f_prelude: AdaNodeListType.t
    Lazy.t;
         
    f_body: [
      | `LibraryItem
          of LibraryItemType.fields
      | `Subunit
          of SubunitType.fields
    ]
    Lazy.t;
         
    f_pragmas: PragmaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ComponentClauseType : sig

  
   
  type t = [
    | `ComponentClause
        of fields
  ]
  and fields = 
  {
         
    f_id: IdentifierType.t
    Lazy.t;
         
    f_position: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ComponentDefType : sig

  
   
  type t = [
    | `ComponentDef
        of fields
  ]
  and fields = 
  {
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ConstantNodeType : sig

  
   
  type t = [
    | `ConstantAbsent
        of ConstantAbsentType.fields
    | `ConstantPresent
        of ConstantPresentType.fields
  ]


end
and ConstantAbsentType : sig

  
   
  type t = [
    | `ConstantAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ConstantPresentType : sig

  
   
  type t = [
    | `ConstantPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ConstraintType : sig

  
   
  type t = [
    | `DeltaConstraint
        of DeltaConstraintType.fields
    | `DigitsConstraint
        of DigitsConstraintType.fields
    | `DiscriminantConstraint
        of DiscriminantConstraintType.fields
    | `IndexConstraint
        of IndexConstraintType.fields
    | `RangeConstraint
        of RangeConstraintType.fields
  ]


end
and DeltaConstraintType : sig

  
   
  type t = [
    | `DeltaConstraint
        of fields
  ]
  and fields = 
  {
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DigitsConstraintType : sig

  
   
  type t = [
    | `DigitsConstraint
        of fields
  ]
  and fields = 
  {
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DiscriminantConstraintType : sig

  
   
  type t = [
    | `DiscriminantConstraint
        of fields
  ]
  and fields = 
  {
         
    f_constraints: BasicAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and IndexConstraintType : sig

  
   
  type t = [
    | `IndexConstraint
        of fields
  ]
  and fields = 
  {
         
    f_constraints: ConstraintListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RangeConstraintType : sig

  
   
  type t = [
    | `RangeConstraint
        of fields
  ]
  and fields = 
  {
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DeclarativePartType : sig

  
   
  type t = [
    | `DeclarativePart
        of fields
    | `PrivatePart
        of PrivatePartType.fields
    | `PublicPart
        of PublicPartType.fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and PrivatePartType : sig

  
   
  type t = [
    | `PrivatePart
        of fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and PublicPartType : sig

  
   
  type t = [
    | `PublicPart
        of fields
  ]
  and fields = 
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ElsifExprPartType : sig

  
   
  type t = [
    | `ElsifExprPart
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ElsifStmtPartType : sig

  
   
  type t = [
    | `ElsifStmtPart
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ExprType : sig

  
   
  (**
    * AbstractStateDeclExprType.t
    * AllocatorType.t
    * BaseAggregateType.t
    * BinOpType.t
    * BoxExprType.t
    * CaseExprAlternativeType.t
    * CondExprType.t
    * ContractCasesType.t
    * DeclExprType.t
    * MembershipExprType.t
    * NameType.t
    * ParenExprType.t
    * QuantifiedExprType.t
    * RaiseExprType.t
    * UnOpType.t
    *)
  type t = [
    | `AbstractStateDeclExpr
        of AbstractStateDeclExprType.fields
    | `Allocator
        of AllocatorType.fields
    | `Aggregate
        of AggregateType.fields
    | `BracketAggregate
        of BracketAggregateType.fields
    | `DeltaAggregate
        of DeltaAggregateType.fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
    | `NullRecordAggregate
        of NullRecordAggregateType.fields
    | `BinOp
        of BinOpType.fields
    | `RelationOp
        of RelationOpType.fields
    | `BoxExpr
        of BoxExprType.fields
    | `CaseExprAlternative
        of CaseExprAlternativeType.fields
    | `CaseExpr
        of CaseExprType.fields
    | `IfExpr
        of IfExprType.fields
    | `ContractCases
        of ContractCasesType.fields
    | `DeclExpr
        of DeclExprType.fields
    | `MembershipExpr
        of MembershipExprType.fields
    | `AttributeRef
        of AttributeRefType.fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
    | `CallExpr
        of CallExprType.fields
    | `DefiningName
        of DefiningNameType.fields
    | `DiscreteSubtypeName
        of DiscreteSubtypeNameType.fields
    | `DottedName
        of DottedNameType.fields
    | `EndName
        of EndNameType.fields
    | `ExplicitDeref
        of ExplicitDerefType.fields
    | `QualExpr
        of QualExprType.fields
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
    | `TargetName
        of TargetNameType.fields
    | `ParenExpr
        of ParenExprType.fields
    | `QuantifiedExpr
        of QuantifiedExprType.fields
    | `RaiseExpr
        of RaiseExprType.fields
    | `UnOp
        of UnOpType.fields
  ]


end
and AbstractStateDeclExprType : sig

  
   
  type t = [
    | `AbstractStateDeclExpr
        of fields
  ]
  and fields = 
  {
         
    f_state_decl: [
      | `AbstractStateDecl
          of AbstractStateDeclType.fields
      | `MultiAbstractStateDecl
          of MultiAbstractStateDeclType.fields
      | `ParenAbstractStateDecl
          of ParenAbstractStateDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AllocatorType : sig

  
   
  type t = [
    | `Allocator
        of fields
  ]
  and fields = 
  {
         
    f_subpool: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_type_or_expr: [
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `QualExpr
          of QualExprType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BaseAggregateType : sig

  
   
  (**
    * AggregateType.t
    * DeltaAggregateType.t
    * NullRecordAggregateType.t
    *)
  type t = [
    | `Aggregate
        of AggregateType.fields
    | `BracketAggregate
        of BracketAggregateType.fields
    | `DeltaAggregate
        of DeltaAggregateType.fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
    | `NullRecordAggregate
        of NullRecordAggregateType.fields
  ]


end
and AggregateType : sig

  
   
  type t = [
    | `Aggregate
        of fields
    | `BracketAggregate
        of BracketAggregateType.fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BracketAggregateType : sig

  
   
  type t = [
    | `BracketAggregate
        of fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DeltaAggregateType : sig

  
   
  type t = [
    | `DeltaAggregate
        of fields
    | `BracketDeltaAggregate
        of BracketDeltaAggregateType.fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BracketDeltaAggregateType : sig

  
   
  type t = [
    | `BracketDeltaAggregate
        of fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NullRecordAggregateType : sig

  
   
  type t = [
    | `NullRecordAggregate
        of fields
  ]
  and fields = 
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BinOpType : sig

  
   
  type t = [
    | `BinOp
        of fields
    | `RelationOp
        of RelationOpType.fields
  ]
  and fields = 
  {
         
    f_left: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpAnd
          of OpAndType.fields
      | `OpAndThen
          of OpAndThenType.fields
      | `OpConcat
          of OpConcatType.fields
      | `OpDiv
          of OpDivType.fields
      | `OpDoubleDot
          of OpDoubleDotType.fields
      | `OpEq
          of OpEqType.fields
      | `OpGt
          of OpGtType.fields
      | `OpGte
          of OpGteType.fields
      | `OpLt
          of OpLtType.fields
      | `OpLte
          of OpLteType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpMod
          of OpModType.fields
      | `OpMult
          of OpMultType.fields
      | `OpNeq
          of OpNeqType.fields
      | `OpOr
          of OpOrType.fields
      | `OpOrElse
          of OpOrElseType.fields
      | `OpPlus
          of OpPlusType.fields
      | `OpPow
          of OpPowType.fields
      | `OpRem
          of OpRemType.fields
      | `OpXor
          of OpXorType.fields
    ]
    Lazy.t;
         
    f_right: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RelationOpType : sig

  
   
  type t = [
    | `RelationOp
        of fields
  ]
  and fields = 
  {
         
    f_left: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpAnd
          of OpAndType.fields
      | `OpAndThen
          of OpAndThenType.fields
      | `OpConcat
          of OpConcatType.fields
      | `OpDiv
          of OpDivType.fields
      | `OpDoubleDot
          of OpDoubleDotType.fields
      | `OpEq
          of OpEqType.fields
      | `OpGt
          of OpGtType.fields
      | `OpGte
          of OpGteType.fields
      | `OpLt
          of OpLtType.fields
      | `OpLte
          of OpLteType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpMod
          of OpModType.fields
      | `OpMult
          of OpMultType.fields
      | `OpNeq
          of OpNeqType.fields
      | `OpOr
          of OpOrType.fields
      | `OpOrElse
          of OpOrElseType.fields
      | `OpPlus
          of OpPlusType.fields
      | `OpPow
          of OpPowType.fields
      | `OpRem
          of OpRemType.fields
      | `OpXor
          of OpXorType.fields
    ]
    Lazy.t;
         
    f_right: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BoxExprType : sig

  
   
  type t = [
    | `BoxExpr
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and CaseExprAlternativeType : sig

  
   
  type t = [
    | `CaseExprAlternative
        of fields
  ]
  and fields = 
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and CondExprType : sig

  
   
  type t = [
    | `CaseExpr
        of CaseExprType.fields
    | `IfExpr
        of IfExprType.fields
  ]


end
and CaseExprType : sig

  
   
  type t = [
    | `CaseExpr
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_cases: CaseExprAlternativeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and IfExprType : sig

  
   
  type t = [
    | `IfExpr
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_alternatives: ElsifExprPartListType.t
    Lazy.t;
         
    f_else_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ContractCasesType : sig

  
   
  type t = [
    | `ContractCases
        of fields
  ]
  and fields = 
  {
         
    f_contract_cases: ContractCaseAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DeclExprType : sig

  
   
  type t = [
    | `DeclExpr
        of fields
  ]
  and fields = 
  {
         
    f_decls: BasicDeclListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and MembershipExprType : sig

  
   
  type t = [
    | `MembershipExpr
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpIn
          of OpInType.fields
      | `OpNotIn
          of OpNotInType.fields
    ]
    Lazy.t;
         
    f_membership_exprs: ExprListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NameType : sig

  
   
  (**
    * AttributeRefType.t
    * CallExprType.t
    * DefiningNameType.t
    * DiscreteSubtypeNameType.t
    * DottedNameType.t
    * EndNameType.t
    * ExplicitDerefType.t
    * QualExprType.t
    * SingleTokNodeType.t
    * TargetNameType.t
    *)
  type t = [
    | `AttributeRef
        of AttributeRefType.fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
    | `CallExpr
        of CallExprType.fields
    | `DefiningName
        of DefiningNameType.fields
    | `DiscreteSubtypeName
        of DiscreteSubtypeNameType.fields
    | `DottedName
        of DottedNameType.fields
    | `EndName
        of EndNameType.fields
    | `ExplicitDeref
        of ExplicitDerefType.fields
    | `QualExpr
        of QualExprType.fields
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
    | `TargetName
        of TargetNameType.fields
  ]


end
and AttributeRefType : sig

  
   
  type t = [
    | `AttributeRef
        of fields
    | `UpdateAttributeRef
        of UpdateAttributeRefType.fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_attribute: IdentifierType.t
    Lazy.t;
         
    f_args: AdaNodeType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and UpdateAttributeRefType : sig

  
   
  type t = [
    | `UpdateAttributeRef
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of fields
    ]
    Lazy.t;
         
    f_attribute: IdentifierType.t
    Lazy.t;
         
    f_args: AdaNodeType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and CallExprType : sig

  
   
  type t = [
    | `CallExpr
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `AssocList
          of AssocListType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DefiningNameType : sig

  
   
  type t = [
    | `DefiningName
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DiscreteSubtypeNameType : sig

  
   
  type t = [
    | `DiscreteSubtypeName
        of fields
  ]
  and fields = 
  {
         
    f_subtype: DiscreteSubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DottedNameType : sig

  
   
  type t = [
    | `DottedName
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `CharLiteral
          of CharLiteralType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and EndNameType : sig

  
   
  type t = [
    | `EndName
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ExplicitDerefType : sig

  
   
  type t = [
    | `ExplicitDeref
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and QualExprType : sig

  
   
  type t = [
    | `QualExpr
        of fields
  ]
  and fields = 
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `Aggregate
          of AggregateType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SingleTokNodeType : sig

  
   
  (**
    * BaseIdType.t
    * NullLiteralType.t
    * NumLiteralType.t
    *)
  type t = [
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
    | `NullLiteral
        of NullLiteralType.fields
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
  ]


end
and BaseIdType : sig

  
   
  (**
    * CharLiteralType.t
    * IdentifierType.t
    * OpType.t
    * StringLiteralType.t
    *)
  type t = [
    | `CharLiteral
        of CharLiteralType.fields
    | `Identifier
        of IdentifierType.fields
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
    | `StringLiteral
        of StringLiteralType.fields
  ]


end
and CharLiteralType : sig

  
   
  type t = [
    | `CharLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and IdentifierType : sig

  
   
  type t = [
    | `Identifier
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpType : sig

  
   
  type t = [
    | `OpAbs
        of OpAbsType.fields
    | `OpAnd
        of OpAndType.fields
    | `OpAndThen
        of OpAndThenType.fields
    | `OpConcat
        of OpConcatType.fields
    | `OpDiv
        of OpDivType.fields
    | `OpDoubleDot
        of OpDoubleDotType.fields
    | `OpEq
        of OpEqType.fields
    | `OpGt
        of OpGtType.fields
    | `OpGte
        of OpGteType.fields
    | `OpIn
        of OpInType.fields
    | `OpLt
        of OpLtType.fields
    | `OpLte
        of OpLteType.fields
    | `OpMinus
        of OpMinusType.fields
    | `OpMod
        of OpModType.fields
    | `OpMult
        of OpMultType.fields
    | `OpNeq
        of OpNeqType.fields
    | `OpNot
        of OpNotType.fields
    | `OpNotIn
        of OpNotInType.fields
    | `OpOr
        of OpOrType.fields
    | `OpOrElse
        of OpOrElseType.fields
    | `OpPlus
        of OpPlusType.fields
    | `OpPow
        of OpPowType.fields
    | `OpRem
        of OpRemType.fields
    | `OpXor
        of OpXorType.fields
  ]


end
and OpAbsType : sig

  
   
  type t = [
    | `OpAbs
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpAndType : sig

  
   
  type t = [
    | `OpAnd
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpAndThenType : sig

  
   
  type t = [
    | `OpAndThen
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpConcatType : sig

  
   
  type t = [
    | `OpConcat
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpDivType : sig

  
   
  type t = [
    | `OpDiv
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpDoubleDotType : sig

  
   
  type t = [
    | `OpDoubleDot
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpEqType : sig

  
   
  type t = [
    | `OpEq
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpGtType : sig

  
   
  type t = [
    | `OpGt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpGteType : sig

  
   
  type t = [
    | `OpGte
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpInType : sig

  
   
  type t = [
    | `OpIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpLtType : sig

  
   
  type t = [
    | `OpLt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpLteType : sig

  
   
  type t = [
    | `OpLte
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpMinusType : sig

  
   
  type t = [
    | `OpMinus
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpModType : sig

  
   
  type t = [
    | `OpMod
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpMultType : sig

  
   
  type t = [
    | `OpMult
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpNeqType : sig

  
   
  type t = [
    | `OpNeq
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpNotType : sig

  
   
  type t = [
    | `OpNot
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpNotInType : sig

  
   
  type t = [
    | `OpNotIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpOrType : sig

  
   
  type t = [
    | `OpOr
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpOrElseType : sig

  
   
  type t = [
    | `OpOrElse
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpPlusType : sig

  
   
  type t = [
    | `OpPlus
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpPowType : sig

  
   
  type t = [
    | `OpPow
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpRemType : sig

  
   
  type t = [
    | `OpRem
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OpXorType : sig

  
   
  type t = [
    | `OpXor
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and StringLiteralType : sig

  
   
  type t = [
    | `StringLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NullLiteralType : sig

  
   
  type t = [
    | `NullLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NumLiteralType : sig

  
   
  type t = [
    | `IntLiteral
        of IntLiteralType.fields
    | `RealLiteral
        of RealLiteralType.fields
  ]


end
and IntLiteralType : sig

  
   
  type t = [
    | `IntLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RealLiteralType : sig

  
   
  type t = [
    | `RealLiteral
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TargetNameType : sig

  
   
  type t = [
    | `TargetName
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ParenExprType : sig

  
   
  type t = [
    | `ParenExpr
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and QuantifiedExprType : sig

  
   
  type t = [
    | `QuantifiedExpr
        of fields
  ]
  and fields = 
  {
         
    f_quantifier: QuantifierType.t
    Lazy.t;
         
    f_loop_spec: ForLoopSpecType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RaiseExprType : sig

  
   
  type t = [
    | `RaiseExpr
        of fields
  ]
  and fields = 
  {
         
    f_exception_name: NameType.t
    option
    Lazy.t;
         
    f_error_message: ExprType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and UnOpType : sig

  
   
  type t = [
    | `UnOp
        of fields
  ]
  and fields = 
  {
         
    f_op: [
      | `OpAbs
          of OpAbsType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpNot
          of OpNotType.fields
      | `OpPlus
          of OpPlusType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and HandledStmtsType : sig

  
   
  type t = [
    | `HandledStmts
        of fields
  ]
  and fields = 
  {
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_exceptions: AdaNodeListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and InterfaceKindType : sig

  
   
  type t = [
    | `InterfaceKindLimited
        of InterfaceKindLimitedType.fields
    | `InterfaceKindProtected
        of InterfaceKindProtectedType.fields
    | `InterfaceKindSynchronized
        of InterfaceKindSynchronizedType.fields
    | `InterfaceKindTask
        of InterfaceKindTaskType.fields
  ]


end
and InterfaceKindLimitedType : sig

  
   
  type t = [
    | `InterfaceKindLimited
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and InterfaceKindProtectedType : sig

  
   
  type t = [
    | `InterfaceKindProtected
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and InterfaceKindSynchronizedType : sig

  
   
  type t = [
    | `InterfaceKindSynchronized
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and InterfaceKindTaskType : sig

  
   
  type t = [
    | `InterfaceKindTask
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and IterTypeType : sig

  
   
  type t = [
    | `IterTypeIn
        of IterTypeInType.fields
    | `IterTypeOf
        of IterTypeOfType.fields
  ]


end
and IterTypeInType : sig

  
   
  type t = [
    | `IterTypeIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and IterTypeOfType : sig

  
   
  type t = [
    | `IterTypeOf
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and LibraryItemType : sig

  
   
  type t = [
    | `LibraryItem
        of fields
  ]
  and fields = 
  {
         
    f_has_private: PrivateNodeType.t
    Lazy.t;
         
    f_item: [
      | `AbstractSubpDecl
          of AbstractSubpDeclType.fields
      | `ErrorDecl
          of ErrorDeclType.fields
      | `ExprFunction
          of ExprFunctionType.fields
      | `GenericPackageDecl
          of GenericPackageDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericPackageRenamingDecl
          of GenericPackageRenamingDeclType.fields
      | `GenericSubpDecl
          of GenericSubpDeclType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `GenericSubpRenamingDecl
          of GenericSubpRenamingDeclType.fields
      | `NullSubpDecl
          of NullSubpDeclType.fields
      | `PackageBody
          of PackageBodyType.fields
      | `PackageDecl
          of PackageDeclType.fields
      | `PackageRenamingDecl
          of PackageRenamingDeclType.fields
      | `SubpBody
          of SubpBodyType.fields
      | `SubpDecl
          of SubpDeclType.fields
      | `SubpRenamingDecl
          of SubpRenamingDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and LimitedNodeType : sig

  
   
  type t = [
    | `LimitedAbsent
        of LimitedAbsentType.fields
    | `LimitedPresent
        of LimitedPresentType.fields
  ]


end
and LimitedAbsentType : sig

  
   
  type t = [
    | `LimitedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and LimitedPresentType : sig

  
   
  type t = [
    | `LimitedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and LoopSpecType : sig

  
   
  type t = [
    | `ForLoopSpec
        of ForLoopSpecType.fields
    | `WhileLoopSpec
        of WhileLoopSpecType.fields
  ]


end
and ForLoopSpecType : sig

  
   
  type t = [
    | `ForLoopSpec
        of fields
  ]
  and fields = 
  {
         
    f_var_decl: ForLoopVarDeclType.t
    Lazy.t;
         
    f_loop_type: IterTypeType.t
    Lazy.t;
         
    f_has_reverse: ReverseNodeType.t
    Lazy.t;
         
    f_iter_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and WhileLoopSpecType : sig

  
   
  type t = [
    | `WhileLoopSpec
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ModeType : sig

  
   
  type t = [
    | `ModeDefault
        of ModeDefaultType.fields
    | `ModeIn
        of ModeInType.fields
    | `ModeInOut
        of ModeInOutType.fields
    | `ModeOut
        of ModeOutType.fields
  ]


end
and ModeDefaultType : sig

  
   
  type t = [
    | `ModeDefault
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ModeInType : sig

  
   
  type t = [
    | `ModeIn
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ModeInOutType : sig

  
   
  type t = [
    | `ModeInOut
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ModeOutType : sig

  
   
  type t = [
    | `ModeOut
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and MultiAbstractStateDeclType : sig

  
   
  type t = [
    | `MultiAbstractStateDecl
        of fields
  ]
  and fields = 
  {
         
    f_decls: AbstractStateDeclListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NotNullType : sig

  
   
  type t = [
    | `NotNullAbsent
        of NotNullAbsentType.fields
    | `NotNullPresent
        of NotNullPresentType.fields
  ]


end
and NotNullAbsentType : sig

  
   
  type t = [
    | `NotNullAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NotNullPresentType : sig

  
   
  type t = [
    | `NotNullPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NullComponentDeclType : sig

  
   
  type t = [
    | `NullComponentDecl
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OthersDesignatorType : sig

  
   
  type t = [
    | `OthersDesignator
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OverridingNodeType : sig

  
   
  type t = [
    | `OverridingNotOverriding
        of OverridingNotOverridingType.fields
    | `OverridingOverriding
        of OverridingOverridingType.fields
    | `OverridingUnspecified
        of OverridingUnspecifiedType.fields
  ]


end
and OverridingNotOverridingType : sig

  
   
  type t = [
    | `OverridingNotOverriding
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OverridingOverridingType : sig

  
   
  type t = [
    | `OverridingOverriding
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OverridingUnspecifiedType : sig

  
   
  type t = [
    | `OverridingUnspecified
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ParamsType : sig

  
   
  type t = [
    | `Params
        of fields
  ]
  and fields = 
  {
         
    f_params: ParamSpecListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ParenAbstractStateDeclType : sig

  
   
  type t = [
    | `ParenAbstractStateDecl
        of fields
  ]
  and fields = 
  {
         
    f_decl: [
      | `AbstractStateDecl
          of AbstractStateDeclType.fields
      | `ParenAbstractStateDecl
          of fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and PragmaNodeType : sig

  
   
  type t = [
    | `PragmaNode
        of fields
  ]
  and fields = 
  {
         
    f_id: IdentifierType.t
    Lazy.t;
         
    f_args: BaseAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and PrivateNodeType : sig

  
   
  type t = [
    | `PrivateAbsent
        of PrivateAbsentType.fields
    | `PrivatePresent
        of PrivatePresentType.fields
  ]


end
and PrivateAbsentType : sig

  
   
  type t = [
    | `PrivateAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and PrivatePresentType : sig

  
   
  type t = [
    | `PrivatePresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ProtectedDefType : sig

  
   
  type t = [
    | `ProtectedDef
        of fields
  ]
  and fields = 
  {
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ProtectedNodeType : sig

  
   
  type t = [
    | `ProtectedAbsent
        of ProtectedAbsentType.fields
    | `ProtectedPresent
        of ProtectedPresentType.fields
  ]


end
and ProtectedAbsentType : sig

  
   
  type t = [
    | `ProtectedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ProtectedPresentType : sig

  
   
  type t = [
    | `ProtectedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and QuantifierType : sig

  
   
  type t = [
    | `QuantifierAll
        of QuantifierAllType.fields
    | `QuantifierSome
        of QuantifierSomeType.fields
  ]


end
and QuantifierAllType : sig

  
   
  type t = [
    | `QuantifierAll
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and QuantifierSomeType : sig

  
   
  type t = [
    | `QuantifierSome
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RangeSpecType : sig

  
   
  type t = [
    | `RangeSpec
        of fields
  ]
  and fields = 
  {
         
    f_range: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RenamingClauseType : sig

  
   
  type t = [
    | `RenamingClause
        of fields
    | `SyntheticRenamingClause
        of SyntheticRenamingClauseType.fields
  ]
  and fields = 
  {
         
    f_renamed_object: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SyntheticRenamingClauseType : sig

  
   
  type t = [
    | `SyntheticRenamingClause
        of fields
  ]
  and fields = 
  {
         
    f_renamed_object: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ReverseNodeType : sig

  
   
  type t = [
    | `ReverseAbsent
        of ReverseAbsentType.fields
    | `ReversePresent
        of ReversePresentType.fields
  ]


end
and ReverseAbsentType : sig

  
   
  type t = [
    | `ReverseAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ReversePresentType : sig

  
   
  type t = [
    | `ReversePresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SelectWhenPartType : sig

  
   
  type t = [
    | `SelectWhenPart
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and StmtType : sig

  
   
  (**
    * CompositeStmtType.t
    * ErrorStmtType.t
    * SimpleStmtType.t
    *)
  type t = [
    | `AcceptStmt
        of AcceptStmtType.fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
    | `CaseStmt
        of CaseStmtType.fields
    | `ExtendedReturnStmt
        of ExtendedReturnStmtType.fields
    | `IfStmt
        of IfStmtType.fields
    | `NamedStmt
        of NamedStmtType.fields
    | `SelectStmt
        of SelectStmtType.fields
    | `ErrorStmt
        of ErrorStmtType.fields
    | `AbortStmt
        of AbortStmtType.fields
    | `AssignStmt
        of AssignStmtType.fields
    | `CallStmt
        of CallStmtType.fields
    | `DelayStmt
        of DelayStmtType.fields
    | `ExitStmt
        of ExitStmtType.fields
    | `GotoStmt
        of GotoStmtType.fields
    | `Label
        of LabelType.fields
    | `NullStmt
        of NullStmtType.fields
    | `RaiseStmt
        of RaiseStmtType.fields
    | `RequeueStmt
        of RequeueStmtType.fields
    | `ReturnStmt
        of ReturnStmtType.fields
    | `TerminateAlternative
        of TerminateAlternativeType.fields
  ]


end
and CompositeStmtType : sig

  
   
  (**
    * AcceptStmtType.t
    * BaseLoopStmtType.t
    * BlockStmtType.t
    * CaseStmtType.t
    * ExtendedReturnStmtType.t
    * IfStmtType.t
    * NamedStmtType.t
    * SelectStmtType.t
    *)
  type t = [
    | `AcceptStmt
        of AcceptStmtType.fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
    | `CaseStmt
        of CaseStmtType.fields
    | `ExtendedReturnStmt
        of ExtendedReturnStmtType.fields
    | `IfStmt
        of IfStmtType.fields
    | `NamedStmt
        of NamedStmtType.fields
    | `SelectStmt
        of SelectStmtType.fields
  ]


end
and AcceptStmtType : sig

  
   
  type t = [
    | `AcceptStmt
        of fields
    | `AcceptStmtWithStmts
        of AcceptStmtWithStmtsType.fields
  ]
  and fields = 
  {
         
    f_name: IdentifierType.t
    Lazy.t;
         
    f_entry_index_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AcceptStmtWithStmtsType : sig

  
   
  type t = [
    | `AcceptStmtWithStmts
        of fields
  ]
  and fields = 
  {
         
    f_name: IdentifierType.t
    Lazy.t;
         
    f_entry_index_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BaseLoopStmtType : sig

  
   
  type t = [
    | `ForLoopStmt
        of ForLoopStmtType.fields
    | `LoopStmt
        of LoopStmtType.fields
    | `WhileLoopStmt
        of WhileLoopStmtType.fields
  ]


end
and ForLoopStmtType : sig

  
   
  type t = [
    | `ForLoopStmt
        of fields
  ]
  and fields = 
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and LoopStmtType : sig

  
   
  type t = [
    | `LoopStmt
        of fields
  ]
  and fields = 
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and WhileLoopStmtType : sig

  
   
  type t = [
    | `WhileLoopStmt
        of fields
  ]
  and fields = 
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BlockStmtType : sig

  
   
  type t = [
    | `BeginBlock
        of BeginBlockType.fields
    | `DeclBlock
        of DeclBlockType.fields
  ]


end
and BeginBlockType : sig

  
   
  type t = [
    | `BeginBlock
        of fields
  ]
  and fields = 
  {
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DeclBlockType : sig

  
   
  type t = [
    | `DeclBlock
        of fields
  ]
  and fields = 
  {
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and CaseStmtType : sig

  
   
  type t = [
    | `CaseStmt
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_alternatives: CaseStmtAlternativeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ExtendedReturnStmtType : sig

  
   
  type t = [
    | `ExtendedReturnStmt
        of fields
  ]
  and fields = 
  {
         
    f_decl: ExtendedReturnStmtObjectDeclType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and IfStmtType : sig

  
   
  type t = [
    | `IfStmt
        of fields
  ]
  and fields = 
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_stmts: StmtListType.t
    Lazy.t;
         
    f_alternatives: ElsifStmtPartListType.t
    Lazy.t;
         
    f_else_stmts: StmtListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NamedStmtType : sig

  
   
  type t = [
    | `NamedStmt
        of fields
  ]
  and fields = 
  {
         
    f_decl: NamedStmtDeclType.t
    Lazy.t;
         
    f_stmt: [
      | `BeginBlock
          of BeginBlockType.fields
      | `DeclBlock
          of DeclBlockType.fields
      | `ForLoopStmt
          of ForLoopStmtType.fields
      | `LoopStmt
          of LoopStmtType.fields
      | `WhileLoopStmt
          of WhileLoopStmtType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SelectStmtType : sig

  
   
  type t = [
    | `SelectStmt
        of fields
  ]
  and fields = 
  {
         
    f_guards: SelectWhenPartListType.t
    Lazy.t;
         
    f_else_stmts: StmtListType.t
    option
    Lazy.t;
         
    f_abort_stmts: StmtListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ErrorStmtType : sig

  
   
  type t = [
    | `ErrorStmt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SimpleStmtType : sig

  
   
  type t = [
    | `AbortStmt
        of AbortStmtType.fields
    | `AssignStmt
        of AssignStmtType.fields
    | `CallStmt
        of CallStmtType.fields
    | `DelayStmt
        of DelayStmtType.fields
    | `ExitStmt
        of ExitStmtType.fields
    | `GotoStmt
        of GotoStmtType.fields
    | `Label
        of LabelType.fields
    | `NullStmt
        of NullStmtType.fields
    | `RaiseStmt
        of RaiseStmtType.fields
    | `RequeueStmt
        of RequeueStmtType.fields
    | `ReturnStmt
        of ReturnStmtType.fields
    | `TerminateAlternative
        of TerminateAlternativeType.fields
  ]


end
and AbortStmtType : sig

  
   
  type t = [
    | `AbortStmt
        of fields
  ]
  and fields = 
  {
         
    f_names: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and AssignStmtType : sig

  
   
  type t = [
    | `AssignStmt
        of fields
  ]
  and fields = 
  {
         
    f_dest: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and CallStmtType : sig

  
   
  type t = [
    | `CallStmt
        of fields
  ]
  and fields = 
  {
         
    f_call: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DelayStmtType : sig

  
   
  type t = [
    | `DelayStmt
        of fields
  ]
  and fields = 
  {
         
    f_has_until: UntilNodeType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ExitStmtType : sig

  
   
  type t = [
    | `ExitStmt
        of fields
  ]
  and fields = 
  {
         
    f_loop_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    option
    Lazy.t;
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and GotoStmtType : sig

  
   
  type t = [
    | `GotoStmt
        of fields
  ]
  and fields = 
  {
         
    f_label_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and LabelType : sig

  
   
  type t = [
    | `Label
        of fields
  ]
  and fields = 
  {
         
    f_decl: LabelDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and NullStmtType : sig

  
   
  type t = [
    | `NullStmt
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RaiseStmtType : sig

  
   
  type t = [
    | `RaiseStmt
        of fields
  ]
  and fields = 
  {
         
    f_exception_name: NameType.t
    option
    Lazy.t;
         
    f_error_message: ExprType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RequeueStmtType : sig

  
   
  type t = [
    | `RequeueStmt
        of fields
  ]
  and fields = 
  {
         
    f_call_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_has_abort: AbortNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ReturnStmtType : sig

  
   
  type t = [
    | `ReturnStmt
        of fields
  ]
  and fields = 
  {
         
    f_return_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TerminateAlternativeType : sig

  
   
  type t = [
    | `TerminateAlternative
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SubpKindType : sig

  
   
  type t = [
    | `SubpKindFunction
        of SubpKindFunctionType.fields
    | `SubpKindProcedure
        of SubpKindProcedureType.fields
  ]


end
and SubpKindFunctionType : sig

  
   
  type t = [
    | `SubpKindFunction
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SubpKindProcedureType : sig

  
   
  type t = [
    | `SubpKindProcedure
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SubunitType : sig

  
   
  type t = [
    | `Subunit
        of fields
  ]
  and fields = 
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_body: [
      | `PackageBody
          of PackageBodyType.fields
      | `ProtectedBody
          of ProtectedBodyType.fields
      | `SubpBody
          of SubpBodyType.fields
      | `TaskBody
          of TaskBodyType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SynchronizedNodeType : sig

  
   
  type t = [
    | `SynchronizedAbsent
        of SynchronizedAbsentType.fields
    | `SynchronizedPresent
        of SynchronizedPresentType.fields
  ]


end
and SynchronizedAbsentType : sig

  
   
  type t = [
    | `SynchronizedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SynchronizedPresentType : sig

  
   
  type t = [
    | `SynchronizedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TaggedNodeType : sig

  
   
  type t = [
    | `TaggedAbsent
        of TaggedAbsentType.fields
    | `TaggedPresent
        of TaggedPresentType.fields
  ]


end
and TaggedAbsentType : sig

  
   
  type t = [
    | `TaggedAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TaggedPresentType : sig

  
   
  type t = [
    | `TaggedPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TaskDefType : sig

  
   
  type t = [
    | `TaskDef
        of fields
  ]
  and fields = 
  {
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TypeDefType : sig

  
   
  (**
    * AccessDefType.t
    * ArrayTypeDefType.t
    * DerivedTypeDefType.t
    * EnumTypeDefType.t
    * FormalDiscreteTypeDefType.t
    * InterfaceTypeDefType.t
    * ModIntTypeDefType.t
    * PrivateTypeDefType.t
    * RealTypeDefType.t
    * RecordTypeDefType.t
    * SignedIntTypeDefType.t
    *)
  type t = [
    | `AccessToSubpDef
        of AccessToSubpDefType.fields
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
    | `ArrayTypeDef
        of ArrayTypeDefType.fields
    | `DerivedTypeDef
        of DerivedTypeDefType.fields
    | `EnumTypeDef
        of EnumTypeDefType.fields
    | `FormalDiscreteTypeDef
        of FormalDiscreteTypeDefType.fields
    | `InterfaceTypeDef
        of InterfaceTypeDefType.fields
    | `ModIntTypeDef
        of ModIntTypeDefType.fields
    | `PrivateTypeDef
        of PrivateTypeDefType.fields
    | `DecimalFixedPointDef
        of DecimalFixedPointDefType.fields
    | `FloatingPointDef
        of FloatingPointDefType.fields
    | `OrdinaryFixedPointDef
        of OrdinaryFixedPointDefType.fields
    | `RecordTypeDef
        of RecordTypeDefType.fields
    | `SignedIntTypeDef
        of SignedIntTypeDefType.fields
  ]


end
and AccessDefType : sig

  
   
  (**
    * AccessToSubpDefType.t
    * BaseTypeAccessDefType.t
    *)
  type t = [
    | `AccessToSubpDef
        of AccessToSubpDefType.fields
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
  ]


end
and AccessToSubpDefType : sig

  
   
  type t = [
    | `AccessToSubpDef
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_has_protected: ProtectedNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and BaseTypeAccessDefType : sig

  
   
  type t = [
    | `AnonymousTypeAccessDef
        of AnonymousTypeAccessDefType.fields
    | `TypeAccessDef
        of TypeAccessDefType.fields
  ]


end
and AnonymousTypeAccessDefType : sig

  
   
  type t = [
    | `AnonymousTypeAccessDef
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_type_decl: BaseTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TypeAccessDefType : sig

  
   
  type t = [
    | `TypeAccessDef
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_has_all: AllNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ArrayTypeDefType : sig

  
   
  type t = [
    | `ArrayTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_indices: ArrayIndicesType.t
    Lazy.t;
         
    f_component_type: ComponentDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DerivedTypeDefType : sig

  
   
  type t = [
    | `DerivedTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_has_synchronized: SynchronizedNodeType.t
    Lazy.t;
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_record_extension: BaseRecordDefType.t
    option
    Lazy.t;
         
    f_has_with_private: WithPrivateType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and EnumTypeDefType : sig

  
   
  type t = [
    | `EnumTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_enum_literals: EnumLiteralDeclListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and FormalDiscreteTypeDefType : sig

  
   
  type t = [
    | `FormalDiscreteTypeDef
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and InterfaceTypeDefType : sig

  
   
  type t = [
    | `InterfaceTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_interface_kind: InterfaceKindType.t
    option
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ModIntTypeDefType : sig

  
   
  type t = [
    | `ModIntTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and PrivateTypeDefType : sig

  
   
  type t = [
    | `PrivateTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_tagged: TaggedNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RealTypeDefType : sig

  
   
  type t = [
    | `DecimalFixedPointDef
        of DecimalFixedPointDefType.fields
    | `FloatingPointDef
        of FloatingPointDefType.fields
    | `OrdinaryFixedPointDef
        of OrdinaryFixedPointDefType.fields
  ]


end
and DecimalFixedPointDefType : sig

  
   
  type t = [
    | `DecimalFixedPointDef
        of fields
  ]
  and fields = 
  {
         
    f_delta: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and FloatingPointDefType : sig

  
   
  type t = [
    | `FloatingPointDef
        of fields
  ]
  and fields = 
  {
         
    f_num_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and OrdinaryFixedPointDefType : sig

  
   
  type t = [
    | `OrdinaryFixedPointDef
        of fields
  ]
  and fields = 
  {
         
    f_delta: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and RecordTypeDefType : sig

  
   
  type t = [
    | `RecordTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_tagged: TaggedNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_record_def: BaseRecordDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SignedIntTypeDefType : sig

  
   
  type t = [
    | `SignedIntTypeDef
        of fields
  ]
  and fields = 
  {
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and TypeExprType : sig

  
   
  (**
    * AnonymousTypeType.t
    * EnumLitSynthTypeExprType.t
    * SubtypeIndicationType.t
    *)
  type t = [
    | `AnonymousType
        of AnonymousTypeType.fields
    | `EnumLitSynthTypeExpr
        of EnumLitSynthTypeExprType.fields
    | `SubtypeIndication
        of SubtypeIndicationType.fields
    | `ConstrainedSubtypeIndication
        of ConstrainedSubtypeIndicationType.fields
    | `DiscreteSubtypeIndication
        of DiscreteSubtypeIndicationType.fields
  ]


end
and AnonymousTypeType : sig

  
   
  type t = [
    | `AnonymousType
        of fields
  ]
  and fields = 
  {
         
    f_type_decl: AnonymousTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and EnumLitSynthTypeExprType : sig

  
   
  type t = [
    | `EnumLitSynthTypeExpr
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and SubtypeIndicationType : sig

  
   
  type t = [
    | `SubtypeIndication
        of fields
    | `ConstrainedSubtypeIndication
        of ConstrainedSubtypeIndicationType.fields
    | `DiscreteSubtypeIndication
        of DiscreteSubtypeIndicationType.fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and ConstrainedSubtypeIndicationType : sig

  
   
  type t = [
    | `ConstrainedSubtypeIndication
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and DiscreteSubtypeIndicationType : sig

  
   
  type t = [
    | `DiscreteSubtypeIndication
        of fields
  ]
  and fields = 
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and UnconstrainedArrayIndexType : sig

  
   
  type t = [
    | `UnconstrainedArrayIndex
        of fields
  ]
  and fields = 
  {
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and UntilNodeType : sig

  
   
  type t = [
    | `UntilAbsent
        of UntilAbsentType.fields
    | `UntilPresent
        of UntilPresentType.fields
  ]


end
and UntilAbsentType : sig

  
   
  type t = [
    | `UntilAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and UntilPresentType : sig

  
   
  type t = [
    | `UntilPresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and UseClauseType : sig

  
   
  type t = [
    | `UsePackageClause
        of UsePackageClauseType.fields
    | `UseTypeClause
        of UseTypeClauseType.fields
  ]


end
and UsePackageClauseType : sig

  
   
  type t = [
    | `UsePackageClause
        of fields
  ]
  and fields = 
  {
         
    f_packages: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and UseTypeClauseType : sig

  
   
  type t = [
    | `UseTypeClause
        of fields
  ]
  and fields = 
  {
         
    f_has_all: AllNodeType.t
    Lazy.t;
         
    f_types: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and VariantType : sig

  
   
  type t = [
    | `Variant
        of fields
  ]
  and fields = 
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and VariantPartType : sig

  
   
  type t = [
    | `VariantPart
        of fields
  ]
  and fields = 
  {
         
    f_discr_name: IdentifierType.t
    Lazy.t;
         
    f_variant: VariantListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and WithClauseType : sig

  
   
  type t = [
    | `WithClause
        of fields
  ]
  and fields = 
  {
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_has_private: PrivateNodeType.t
    Lazy.t;
         
    f_packages: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and WithPrivateType : sig

  
   
  type t = [
    | `WithPrivateAbsent
        of WithPrivateAbsentType.fields
    | `WithPrivatePresent
        of WithPrivatePresentType.fields
  ]


end
and WithPrivateAbsentType : sig

  
   
  type t = [
    | `WithPrivateAbsent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end
and WithPrivatePresentType : sig

  
   
  type t = [
    | `WithPrivatePresent
        of fields
  ]
  and fields = 
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }



end

and AnalysisUnit : sig
  (**
   * This type represents the analysis of a single file.
  
   * This type has strong-reference semantics and is ref-counted. Furthermore, a
   * reference to a unit contains an implicit reference to the context that owns
   * it. This means that keeping a reference to a unit will keep the context and
   * all the unit it contains allocated.
   *)

  type t

  val root : t -> AdaNodeType.t option
  (**
   * Return the root node for this unit, or ``None`` if there is none.
   *)

  val diagnostics : t -> Diagnostic.t list
  (**
   * Diagnostics for this unit.
   *)

  val filename : t -> string
  (**
   * Return the filename this unit is associated to.
   *)

  val reparse : ?charset:string -> ?buffer:string -> t -> unit
  (**
   * Reparse an analysis unit from a buffer, if provided, or from the original
   * file otherwise. If ``Charset`` is empty or ``None``, use the last charset
   * successfuly used for this unit, otherwise use it to decode the content of
   * the source file.
  
   * If any failure occurs, such as decoding, lexing or parsing failure,
   * diagnostic are emitted to explain what happened.
   *)

  val first_token : t -> Token.t
  (**
   * Return a reference to the first token scanned in this unit.
   *)

  val last_token : t -> Token.t
  (**
   * Return a reference to the last token scanned in this unit.
   *)

  val token_count : t -> int
  (**
   * Return the number of tokens in this unit.
   *)

  val trivia_count : t -> int
  (**
   * Return the number of trivias in this unit. This is 0 for units that were
   * parsed with trivia analysis disabled.
   *)

  
  val fold_tokens : ('a -> Token.t -> 'a) -> 'a -> t -> 'a
  (**
   * Fold all the token this node contains by calling f on each token.
   *)

  val iter_tokens : (Token.t -> unit) -> t-> unit
  (**
   * Iterate over all token this node contains by calling f on each token.
   *)

  val map_tokens : (Token.t -> 'a) -> t -> 'a list
  (**
   * Map each token calling the given function
   *)

  val tokens : t -> Token.t list
  (**
   * Return a list of tokens for the given node
   *)

end

and AnalysisContext : sig
  (**
   * This type represents a context for all source analysis. This is the first
   * type you need to create to use Libadalang. It will contain the results of
   * all analysis, and is the main holder for all the data.
  
   * You can create several analysis contexts if you need to, which enables you,
   * for example to:
  
   * * analyze several different projects at the same time;
  
   * * analyze different parts of the same projects in parallel.
  
   * In the current design, contexts always keep all of their analysis units
   * allocated. If you need to get this memory released, the only option at your
   * disposal is to destroy your analysis context instance.
   *)

  type t

  val create :
    ?charset:string
    -> ?with_trivia:bool
    -> ?tab_stop:int
    -> ?unit_provider:UnitProvider.t
    -> unit
    -> t
  (**
   * Create a new analysis context.
  
   * ``Charset`` will be used as a default charset to decode input sources in
   * analysis units. Please see ``GNATCOLL.Iconv`` for several supported
   * charsets. Be careful: passing an unsupported charset is not guaranteed to
   * raise an error here. If no charset is provided, ``"iso-8859-1"`` is the
   * default.
  
   * .. todo:: Passing an unsupported charset here is not guaranteed to raise an
   *    error right here, but this would be really helpful for users.
  
   * When ``With_Trivia`` is true, the parsed analysis units will contain
   * trivias.
  
   * If provided, ``File_Reader`` will be used to fetch the contents of source
   * files instead of the default, which is to just read it from the filesystem
   * and decode it using the regular charset rules. Note that if provided, all
   * parsing APIs that provide a buffer are forbidden, and any use of the
   * rewriting API with the returned context is rejected.
  
   * If provided, ``Unit_Provider`` will be used to query the file name that
   * corresponds to a unit reference during semantic analysis. If it is
   * ``None``, the default one is used instead.
  
   * ``Tab_Stop`` is a positive number to describe the effect of tabulation
   * characters on the column number in source files.
   *)

  val get_from_file :
    ?charset:string
    -> ?reparse:bool
    -> ?grammar_rule:GrammarRule.t
    -> t
    -> string
    -> AnalysisUnit.t
  (**
   * Create a new analysis unit for ``Filename`` or return the existing one if
   * any. If ``Reparse`` is true and the analysis unit already exists, reparse
   * it from ``Filename``.
  
   * ``Rule`` controls which grammar rule is used to parse the unit.
  
   * Use ``Charset`` in order to decode the source. If ``Charset`` is empty then
   * use the context's default charset.
  
   * If any failure occurs, such as file opening, decoding, lexing or parsing
   * failure, return an analysis unit anyway: errors are described as
   * diagnostics of the returned analysis unit.
   *)

  val get_from_buffer :
    ?charset:string
    -> ?grammar_rule:GrammarRule.t
    -> t
    -> string
    -> string
    -> AnalysisUnit.t
  (**
   * Create a new analysis unit for ``Filename`` or return the existing one if
   * any. Whether the analysis unit already exists or not, (re)parse it from the
   * source code in ``Buffer``.
  
   * ``Rule`` controls which grammar rule is used to parse the unit.
  
   * Use ``Charset`` in order to decode the source. If ``Charset`` is empty then
   * use the context's default charset.
  
   * If any failure occurs, such as file opening, decoding, lexing or parsing
   * failure, return an analysis unit anyway: errors are described as
   * diagnostics of the returned analysis unit.
   *)
end

   
and Metadata : sig
  

  
  type t = {
      dottable_subp :
         bool;
      access_entity :
         bool;
      primitive :
         BareNode.t;
      primitive_real_type :
         BareNode.t;
  }

end

   
and EntityInfo : sig
  

  
  type t = {
      md :
         Metadata.t;
      rebindings :
         Rebindings.t;
      from_rebound :
         bool;
  }

end

   
and Aspect : sig
  (**
   * Composite field representing the aspect of an entity (RM 13).
   *)

  
  type t = {
      exists :
         bool;
      node :
         AdaNodeType.t option;
      value :
         ExprType.t option;
  }

end

   
and CompletionItem : sig
  

  
  type t = {
      decl :
         BasicDeclType.t option;
      is_dot_call :
         bool;
      is_visible :
         bool;
  }

end

   
and DiscreteRange : sig
  (**
   * Represent the range of a discrete type or subtype. The bounds are not
   * evaluated, you need to call ``eval_as_int`` on them, if they're static, to
   * get their value.
   *)

  
  type t = {
      low_bound :
         ExprType.t option;
      high_bound :
         ExprType.t option;
  }

end

   
and DiscriminantValues : sig
  (**
   * Represent a set of values (as a list of choices) on a discriminant.
   *)

  
  type t = {
      discriminant :
         IdentifierType.t option;
      values :
         AlternativesListType.t option;
  }

end

   
and DocAnnotation : sig
  (**
   * Documentation annotation.
   *)

  
  type t = {
      key :
         string;
      value :
         string;
  }

end

   
and ParamActual : sig
  (**
   * Data structure used by zip_with_params property. Associates an expression
   * (the actual) to a formal param declaration (the parameter).
   *)

  
  type t = {
      param :
         DefiningNameType.t option;
      actual :
         ExprType.t option;
  }

end

   
and RefResult : sig
  (**
   * Result for a cross reference query returning a reference.
   *)

  
  type t = {
      ref :
         BaseIdType.t option;
      kind :
         RefResultKind.t;
  }

end

   
and RefdDecl : sig
  (**
   * Result for a cross reference query returning a referenced decl.
   *)

  
  type t = {
      decl :
         BasicDeclType.t option;
      kind :
         RefResultKind.t;
  }

end

   
and RefdDef : sig
  (**
   * Result for a cross reference query returning a referenced defining name.
   *)

  
  type t = {
      def_name :
         DefiningNameType.t option;
      kind :
         RefResultKind.t;
  }

end

   
and Shape : sig
  (**
   * Represent one of the shapes that a variant record can have, as a list of
   * the available components.
   *)

  
  type t = {
      components :
         BaseFormalParamDeclType.t list;
      discriminants_values :
         DiscriminantValues.t list;
  }

end

   
and Substitution : sig
  (**
   * Represent a substitution of a BasicDecl by a given value. This can then be
   * used as part of an environment in the eval_as_*_in_env property. See the
   * declaration of those properties for more details.
   *)

  
  type t = {
      from_decl :
         BasicDeclType.t option;
      to_value :
         BigInteger.t;
      value_type :
         BaseTypeDeclType.t option;
  }

end


type _ node =
  | AdaNode :
      AdaNodeType.t node
  | AbortNode :
      AbortNodeType.t node
  | AbortAbsent :
      AbortAbsentType.t node
  | AbortPresent :
      AbortPresentType.t node
  | AbstractNode :
      AbstractNodeType.t node
  | AbstractAbsent :
      AbstractAbsentType.t node
  | AbstractPresent :
      AbstractPresentType.t node
  | AdaList :
      AdaListType.t node
  | AdaNodeList :
      AdaNodeListType.t node
  | AbstractStateDeclList :
      AbstractStateDeclListType.t node
  | AlternativesList :
      AlternativesListType.t node
  | ConstraintList :
      ConstraintListType.t node
  | DeclList :
      DeclListType.t node
  | StmtList :
      StmtListType.t node
  | AspectAssocList :
      AspectAssocListType.t node
  | BaseAssocList :
      BaseAssocListType.t node
  | BasicAssocList :
      BasicAssocListType.t node
  | AssocList :
      AssocListType.t node
  | BasicDeclList :
      BasicDeclListType.t node
  | CaseExprAlternativeList :
      CaseExprAlternativeListType.t node
  | CaseStmtAlternativeList :
      CaseStmtAlternativeListType.t node
  | CompilationUnitList :
      CompilationUnitListType.t node
  | ContractCaseAssocList :
      ContractCaseAssocListType.t node
  | DefiningNameList :
      DefiningNameListType.t node
  | DiscriminantSpecList :
      DiscriminantSpecListType.t node
  | ElsifExprPartList :
      ElsifExprPartListType.t node
  | ElsifStmtPartList :
      ElsifStmtPartListType.t node
  | EnumLiteralDeclList :
      EnumLiteralDeclListType.t node
  | ExprList :
      ExprListType.t node
  | ExprAlternativesList :
      ExprAlternativesListType.t node
  | IdentifierList :
      IdentifierListType.t node
  | DiscriminantChoiceList :
      DiscriminantChoiceListType.t node
  | NameList :
      NameListType.t node
  | ParentList :
      ParentListType.t node
  | ParamSpecList :
      ParamSpecListType.t node
  | PragmaNodeList :
      PragmaNodeListType.t node
  | SelectWhenPartList :
      SelectWhenPartListType.t node
  | UnconstrainedArrayIndexList :
      UnconstrainedArrayIndexListType.t node
  | VariantList :
      VariantListType.t node
  | AliasedNode :
      AliasedNodeType.t node
  | AliasedAbsent :
      AliasedAbsentType.t node
  | AliasedPresent :
      AliasedPresentType.t node
  | AllNode :
      AllNodeType.t node
  | AllAbsent :
      AllAbsentType.t node
  | AllPresent :
      AllPresentType.t node
  | ArrayIndices :
      ArrayIndicesType.t node
  | ConstrainedArrayIndices :
      ConstrainedArrayIndicesType.t node
  | UnconstrainedArrayIndices :
      UnconstrainedArrayIndicesType.t node
  | AspectAssoc :
      AspectAssocType.t node
  | AspectClause :
      AspectClauseType.t node
  | AtClause :
      AtClauseType.t node
  | AttributeDefClause :
      AttributeDefClauseType.t node
  | EnumRepClause :
      EnumRepClauseType.t node
  | RecordRepClause :
      RecordRepClauseType.t node
  | AspectSpec :
      AspectSpecType.t node
  | BaseAssoc :
      BaseAssocType.t node
  | ContractCaseAssoc :
      ContractCaseAssocType.t node
  | PragmaArgumentAssoc :
      PragmaArgumentAssocType.t node
  | BaseFormalParamHolder :
      BaseFormalParamHolderType.t node
  | BaseSubpSpec :
      BaseSubpSpecType.t node
  | EntrySpec :
      EntrySpecType.t node
  | EnumSubpSpec :
      EnumSubpSpecType.t node
  | SubpSpec :
      SubpSpecType.t node
  | ComponentList :
      ComponentListType.t node
  | DiscriminantPart :
      DiscriminantPartType.t node
  | KnownDiscriminantPart :
      KnownDiscriminantPartType.t node
  | UnknownDiscriminantPart :
      UnknownDiscriminantPartType.t node
  | EntryCompletionFormalParams :
      EntryCompletionFormalParamsType.t node
  | GenericFormalPart :
      GenericFormalPartType.t node
  | BaseRecordDef :
      BaseRecordDefType.t node
  | NullRecordDef :
      NullRecordDefType.t node
  | RecordDef :
      RecordDefType.t node
  | BasicAssoc :
      BasicAssocType.t node
  | AggregateAssoc :
      AggregateAssocType.t node
  | MultiDimArrayAssoc :
      MultiDimArrayAssocType.t node
  | DiscriminantAssoc :
      DiscriminantAssocType.t node
  | IteratedAssoc :
      IteratedAssocType.t node
  | ParamAssoc :
      ParamAssocType.t node
  | BasicDecl :
      BasicDeclType.t node
  | AbstractStateDecl :
      AbstractStateDeclType.t node
  | AnonymousExprDecl :
      AnonymousExprDeclType.t node
  | BaseFormalParamDecl :
      BaseFormalParamDeclType.t node
  | ComponentDecl :
      ComponentDeclType.t node
  | DiscriminantSpec :
      DiscriminantSpecType.t node
  | GenericFormal :
      GenericFormalType.t node
  | GenericFormalObjDecl :
      GenericFormalObjDeclType.t node
  | GenericFormalPackage :
      GenericFormalPackageType.t node
  | GenericFormalSubpDecl :
      GenericFormalSubpDeclType.t node
  | GenericFormalTypeDecl :
      GenericFormalTypeDeclType.t node
  | ParamSpec :
      ParamSpecType.t node
  | BasePackageDecl :
      BasePackageDeclType.t node
  | GenericPackageInternal :
      GenericPackageInternalType.t node
  | PackageDecl :
      PackageDeclType.t node
  | BaseTypeDecl :
      BaseTypeDeclType.t node
  | BaseSubtypeDecl :
      BaseSubtypeDeclType.t node
  | DiscreteBaseSubtypeDecl :
      DiscreteBaseSubtypeDeclType.t node
  | SubtypeDecl :
      SubtypeDeclType.t node
  | ClasswideTypeDecl :
      ClasswideTypeDeclType.t node
  | IncompleteTypeDecl :
      IncompleteTypeDeclType.t node
  | IncompleteTaggedTypeDecl :
      IncompleteTaggedTypeDeclType.t node
  | ProtectedTypeDecl :
      ProtectedTypeDeclType.t node
  | TaskTypeDecl :
      TaskTypeDeclType.t node
  | SingleTaskTypeDecl :
      SingleTaskTypeDeclType.t node
  | TypeDecl :
      TypeDeclType.t node
  | AnonymousTypeDecl :
      AnonymousTypeDeclType.t node
  | SynthAnonymousTypeDecl :
      SynthAnonymousTypeDeclType.t node
  | BasicSubpDecl :
      BasicSubpDeclType.t node
  | ClassicSubpDecl :
      ClassicSubpDeclType.t node
  | AbstractSubpDecl :
      AbstractSubpDeclType.t node
  | FormalSubpDecl :
      FormalSubpDeclType.t node
  | AbstractFormalSubpDecl :
      AbstractFormalSubpDeclType.t node
  | ConcreteFormalSubpDecl :
      ConcreteFormalSubpDeclType.t node
  | SubpDecl :
      SubpDeclType.t node
  | EntryDecl :
      EntryDeclType.t node
  | EnumLiteralDecl :
      EnumLiteralDeclType.t node
  | GenericSubpInternal :
      GenericSubpInternalType.t node
  | BodyNode :
      BodyNodeType.t node
  | BaseSubpBody :
      BaseSubpBodyType.t node
  | ExprFunction :
      ExprFunctionType.t node
  | NullSubpDecl :
      NullSubpDeclType.t node
  | SubpBody :
      SubpBodyType.t node
  | SubpRenamingDecl :
      SubpRenamingDeclType.t node
  | BodyStub :
      BodyStubType.t node
  | PackageBodyStub :
      PackageBodyStubType.t node
  | ProtectedBodyStub :
      ProtectedBodyStubType.t node
  | SubpBodyStub :
      SubpBodyStubType.t node
  | TaskBodyStub :
      TaskBodyStubType.t node
  | EntryBody :
      EntryBodyType.t node
  | PackageBody :
      PackageBodyType.t node
  | ProtectedBody :
      ProtectedBodyType.t node
  | TaskBody :
      TaskBodyType.t node
  | EntryIndexSpec :
      EntryIndexSpecType.t node
  | ErrorDecl :
      ErrorDeclType.t node
  | ExceptionDecl :
      ExceptionDeclType.t node
  | ExceptionHandler :
      ExceptionHandlerType.t node
  | ForLoopVarDecl :
      ForLoopVarDeclType.t node
  | GenericDecl :
      GenericDeclType.t node
  | GenericPackageDecl :
      GenericPackageDeclType.t node
  | GenericSubpDecl :
      GenericSubpDeclType.t node
  | GenericInstantiation :
      GenericInstantiationType.t node
  | GenericPackageInstantiation :
      GenericPackageInstantiationType.t node
  | GenericSubpInstantiation :
      GenericSubpInstantiationType.t node
  | GenericRenamingDecl :
      GenericRenamingDeclType.t node
  | GenericPackageRenamingDecl :
      GenericPackageRenamingDeclType.t node
  | GenericSubpRenamingDecl :
      GenericSubpRenamingDeclType.t node
  | LabelDecl :
      LabelDeclType.t node
  | NamedStmtDecl :
      NamedStmtDeclType.t node
  | NumberDecl :
      NumberDeclType.t node
  | ObjectDecl :
      ObjectDeclType.t node
  | ExtendedReturnStmtObjectDecl :
      ExtendedReturnStmtObjectDeclType.t node
  | PackageRenamingDecl :
      PackageRenamingDeclType.t node
  | SingleProtectedDecl :
      SingleProtectedDeclType.t node
  | SingleTaskDecl :
      SingleTaskDeclType.t node
  | CaseStmtAlternative :
      CaseStmtAlternativeType.t node
  | CompilationUnit :
      CompilationUnitType.t node
  | ComponentClause :
      ComponentClauseType.t node
  | ComponentDef :
      ComponentDefType.t node
  | ConstantNode :
      ConstantNodeType.t node
  | ConstantAbsent :
      ConstantAbsentType.t node
  | ConstantPresent :
      ConstantPresentType.t node
  | Constraint :
      ConstraintType.t node
  | DeltaConstraint :
      DeltaConstraintType.t node
  | DigitsConstraint :
      DigitsConstraintType.t node
  | DiscriminantConstraint :
      DiscriminantConstraintType.t node
  | IndexConstraint :
      IndexConstraintType.t node
  | RangeConstraint :
      RangeConstraintType.t node
  | DeclarativePart :
      DeclarativePartType.t node
  | PrivatePart :
      PrivatePartType.t node
  | PublicPart :
      PublicPartType.t node
  | ElsifExprPart :
      ElsifExprPartType.t node
  | ElsifStmtPart :
      ElsifStmtPartType.t node
  | Expr :
      ExprType.t node
  | AbstractStateDeclExpr :
      AbstractStateDeclExprType.t node
  | Allocator :
      AllocatorType.t node
  | BaseAggregate :
      BaseAggregateType.t node
  | Aggregate :
      AggregateType.t node
  | BracketAggregate :
      BracketAggregateType.t node
  | DeltaAggregate :
      DeltaAggregateType.t node
  | BracketDeltaAggregate :
      BracketDeltaAggregateType.t node
  | NullRecordAggregate :
      NullRecordAggregateType.t node
  | BinOp :
      BinOpType.t node
  | RelationOp :
      RelationOpType.t node
  | BoxExpr :
      BoxExprType.t node
  | CaseExprAlternative :
      CaseExprAlternativeType.t node
  | CondExpr :
      CondExprType.t node
  | CaseExpr :
      CaseExprType.t node
  | IfExpr :
      IfExprType.t node
  | ContractCases :
      ContractCasesType.t node
  | DeclExpr :
      DeclExprType.t node
  | MembershipExpr :
      MembershipExprType.t node
  | Name :
      NameType.t node
  | AttributeRef :
      AttributeRefType.t node
  | UpdateAttributeRef :
      UpdateAttributeRefType.t node
  | CallExpr :
      CallExprType.t node
  | DefiningName :
      DefiningNameType.t node
  | DiscreteSubtypeName :
      DiscreteSubtypeNameType.t node
  | DottedName :
      DottedNameType.t node
  | EndName :
      EndNameType.t node
  | ExplicitDeref :
      ExplicitDerefType.t node
  | QualExpr :
      QualExprType.t node
  | SingleTokNode :
      SingleTokNodeType.t node
  | BaseId :
      BaseIdType.t node
  | CharLiteral :
      CharLiteralType.t node
  | Identifier :
      IdentifierType.t node
  | Op :
      OpType.t node
  | OpAbs :
      OpAbsType.t node
  | OpAnd :
      OpAndType.t node
  | OpAndThen :
      OpAndThenType.t node
  | OpConcat :
      OpConcatType.t node
  | OpDiv :
      OpDivType.t node
  | OpDoubleDot :
      OpDoubleDotType.t node
  | OpEq :
      OpEqType.t node
  | OpGt :
      OpGtType.t node
  | OpGte :
      OpGteType.t node
  | OpIn :
      OpInType.t node
  | OpLt :
      OpLtType.t node
  | OpLte :
      OpLteType.t node
  | OpMinus :
      OpMinusType.t node
  | OpMod :
      OpModType.t node
  | OpMult :
      OpMultType.t node
  | OpNeq :
      OpNeqType.t node
  | OpNot :
      OpNotType.t node
  | OpNotIn :
      OpNotInType.t node
  | OpOr :
      OpOrType.t node
  | OpOrElse :
      OpOrElseType.t node
  | OpPlus :
      OpPlusType.t node
  | OpPow :
      OpPowType.t node
  | OpRem :
      OpRemType.t node
  | OpXor :
      OpXorType.t node
  | StringLiteral :
      StringLiteralType.t node
  | NullLiteral :
      NullLiteralType.t node
  | NumLiteral :
      NumLiteralType.t node
  | IntLiteral :
      IntLiteralType.t node
  | RealLiteral :
      RealLiteralType.t node
  | TargetName :
      TargetNameType.t node
  | ParenExpr :
      ParenExprType.t node
  | QuantifiedExpr :
      QuantifiedExprType.t node
  | RaiseExpr :
      RaiseExprType.t node
  | UnOp :
      UnOpType.t node
  | HandledStmts :
      HandledStmtsType.t node
  | InterfaceKind :
      InterfaceKindType.t node
  | InterfaceKindLimited :
      InterfaceKindLimitedType.t node
  | InterfaceKindProtected :
      InterfaceKindProtectedType.t node
  | InterfaceKindSynchronized :
      InterfaceKindSynchronizedType.t node
  | InterfaceKindTask :
      InterfaceKindTaskType.t node
  | IterType :
      IterTypeType.t node
  | IterTypeIn :
      IterTypeInType.t node
  | IterTypeOf :
      IterTypeOfType.t node
  | LibraryItem :
      LibraryItemType.t node
  | LimitedNode :
      LimitedNodeType.t node
  | LimitedAbsent :
      LimitedAbsentType.t node
  | LimitedPresent :
      LimitedPresentType.t node
  | LoopSpec :
      LoopSpecType.t node
  | ForLoopSpec :
      ForLoopSpecType.t node
  | WhileLoopSpec :
      WhileLoopSpecType.t node
  | Mode :
      ModeType.t node
  | ModeDefault :
      ModeDefaultType.t node
  | ModeIn :
      ModeInType.t node
  | ModeInOut :
      ModeInOutType.t node
  | ModeOut :
      ModeOutType.t node
  | MultiAbstractStateDecl :
      MultiAbstractStateDeclType.t node
  | NotNull :
      NotNullType.t node
  | NotNullAbsent :
      NotNullAbsentType.t node
  | NotNullPresent :
      NotNullPresentType.t node
  | NullComponentDecl :
      NullComponentDeclType.t node
  | OthersDesignator :
      OthersDesignatorType.t node
  | OverridingNode :
      OverridingNodeType.t node
  | OverridingNotOverriding :
      OverridingNotOverridingType.t node
  | OverridingOverriding :
      OverridingOverridingType.t node
  | OverridingUnspecified :
      OverridingUnspecifiedType.t node
  | Params :
      ParamsType.t node
  | ParenAbstractStateDecl :
      ParenAbstractStateDeclType.t node
  | PragmaNode :
      PragmaNodeType.t node
  | PrivateNode :
      PrivateNodeType.t node
  | PrivateAbsent :
      PrivateAbsentType.t node
  | PrivatePresent :
      PrivatePresentType.t node
  | ProtectedDef :
      ProtectedDefType.t node
  | ProtectedNode :
      ProtectedNodeType.t node
  | ProtectedAbsent :
      ProtectedAbsentType.t node
  | ProtectedPresent :
      ProtectedPresentType.t node
  | Quantifier :
      QuantifierType.t node
  | QuantifierAll :
      QuantifierAllType.t node
  | QuantifierSome :
      QuantifierSomeType.t node
  | RangeSpec :
      RangeSpecType.t node
  | RenamingClause :
      RenamingClauseType.t node
  | SyntheticRenamingClause :
      SyntheticRenamingClauseType.t node
  | ReverseNode :
      ReverseNodeType.t node
  | ReverseAbsent :
      ReverseAbsentType.t node
  | ReversePresent :
      ReversePresentType.t node
  | SelectWhenPart :
      SelectWhenPartType.t node
  | Stmt :
      StmtType.t node
  | CompositeStmt :
      CompositeStmtType.t node
  | AcceptStmt :
      AcceptStmtType.t node
  | AcceptStmtWithStmts :
      AcceptStmtWithStmtsType.t node
  | BaseLoopStmt :
      BaseLoopStmtType.t node
  | ForLoopStmt :
      ForLoopStmtType.t node
  | LoopStmt :
      LoopStmtType.t node
  | WhileLoopStmt :
      WhileLoopStmtType.t node
  | BlockStmt :
      BlockStmtType.t node
  | BeginBlock :
      BeginBlockType.t node
  | DeclBlock :
      DeclBlockType.t node
  | CaseStmt :
      CaseStmtType.t node
  | ExtendedReturnStmt :
      ExtendedReturnStmtType.t node
  | IfStmt :
      IfStmtType.t node
  | NamedStmt :
      NamedStmtType.t node
  | SelectStmt :
      SelectStmtType.t node
  | ErrorStmt :
      ErrorStmtType.t node
  | SimpleStmt :
      SimpleStmtType.t node
  | AbortStmt :
      AbortStmtType.t node
  | AssignStmt :
      AssignStmtType.t node
  | CallStmt :
      CallStmtType.t node
  | DelayStmt :
      DelayStmtType.t node
  | ExitStmt :
      ExitStmtType.t node
  | GotoStmt :
      GotoStmtType.t node
  | Label :
      LabelType.t node
  | NullStmt :
      NullStmtType.t node
  | RaiseStmt :
      RaiseStmtType.t node
  | RequeueStmt :
      RequeueStmtType.t node
  | ReturnStmt :
      ReturnStmtType.t node
  | TerminateAlternative :
      TerminateAlternativeType.t node
  | SubpKind :
      SubpKindType.t node
  | SubpKindFunction :
      SubpKindFunctionType.t node
  | SubpKindProcedure :
      SubpKindProcedureType.t node
  | Subunit :
      SubunitType.t node
  | SynchronizedNode :
      SynchronizedNodeType.t node
  | SynchronizedAbsent :
      SynchronizedAbsentType.t node
  | SynchronizedPresent :
      SynchronizedPresentType.t node
  | TaggedNode :
      TaggedNodeType.t node
  | TaggedAbsent :
      TaggedAbsentType.t node
  | TaggedPresent :
      TaggedPresentType.t node
  | TaskDef :
      TaskDefType.t node
  | TypeDef :
      TypeDefType.t node
  | AccessDef :
      AccessDefType.t node
  | AccessToSubpDef :
      AccessToSubpDefType.t node
  | BaseTypeAccessDef :
      BaseTypeAccessDefType.t node
  | AnonymousTypeAccessDef :
      AnonymousTypeAccessDefType.t node
  | TypeAccessDef :
      TypeAccessDefType.t node
  | ArrayTypeDef :
      ArrayTypeDefType.t node
  | DerivedTypeDef :
      DerivedTypeDefType.t node
  | EnumTypeDef :
      EnumTypeDefType.t node
  | FormalDiscreteTypeDef :
      FormalDiscreteTypeDefType.t node
  | InterfaceTypeDef :
      InterfaceTypeDefType.t node
  | ModIntTypeDef :
      ModIntTypeDefType.t node
  | PrivateTypeDef :
      PrivateTypeDefType.t node
  | RealTypeDef :
      RealTypeDefType.t node
  | DecimalFixedPointDef :
      DecimalFixedPointDefType.t node
  | FloatingPointDef :
      FloatingPointDefType.t node
  | OrdinaryFixedPointDef :
      OrdinaryFixedPointDefType.t node
  | RecordTypeDef :
      RecordTypeDefType.t node
  | SignedIntTypeDef :
      SignedIntTypeDefType.t node
  | TypeExpr :
      TypeExprType.t node
  | AnonymousType :
      AnonymousTypeType.t node
  | EnumLitSynthTypeExpr :
      EnumLitSynthTypeExprType.t node
  | SubtypeIndication :
      SubtypeIndicationType.t node
  | ConstrainedSubtypeIndication :
      ConstrainedSubtypeIndicationType.t node
  | DiscreteSubtypeIndication :
      DiscreteSubtypeIndicationType.t node
  | UnconstrainedArrayIndex :
      UnconstrainedArrayIndexType.t node
  | UntilNode :
      UntilNodeType.t node
  | UntilAbsent :
      UntilAbsentType.t node
  | UntilPresent :
      UntilPresentType.t node
  | UseClause :
      UseClauseType.t node
  | UsePackageClause :
      UsePackageClauseType.t node
  | UseTypeClause :
      UseTypeClauseType.t node
  | Variant :
      VariantType.t node
  | VariantPart :
      VariantPartType.t node
  | WithClause :
      WithClauseType.t node
  | WithPrivate :
      WithPrivateType.t node
  | WithPrivateAbsent :
      WithPrivateAbsentType.t node
  | WithPrivatePresent :
      WithPrivatePresentType.t node

module WithPrivatePresent : sig
  

  type t =
    [
      | `WithPrivatePresent of
          WithPrivatePresentType.fields
    ]

  type fields = WithPrivatePresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module WithPrivateAbsent : sig
  

  type t =
    [
      | `WithPrivateAbsent of
          WithPrivateAbsentType.fields
    ]

  type fields = WithPrivateAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module WithPrivate : sig
  (**
   * Qualifier for the ``private`` keyword in ``with private`` record clauses.
   *)

  type t =
    [
      | WithPrivateAbsent.t
      | WithPrivatePresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< WithPrivateType.t ]
    -> bool
  (**
   * Return whether this is an instance of WithPrivatePresent
   *)



end

module WithClause : sig
  (**
   * With clause.
   *)

  type t =
    [
      | `WithClause of
          WithClauseType.fields
    ]

  type fields = WithClauseType.fields =
    
  {
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_has_private: PrivateNodeType.t
    Lazy.t;
         
    f_packages: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_limited :
    [< WithClauseType.t]
    -> LimitedNodeType.t

      
  val f_has_private :
    [< WithClauseType.t]
    -> PrivateNodeType.t

      
  val f_packages :
    [< WithClauseType.t]
    -> NameListType.t


end

module VariantPart : sig
  (**
   * Variant part in a discriminated type record declaration.
  
   * This corresponds to the whole ``case ... is ... end case;`` block.
   *)

  type t =
    [
      | `VariantPart of
          VariantPartType.fields
    ]

  type fields = VariantPartType.fields =
    
  {
         
    f_discr_name: IdentifierType.t
    Lazy.t;
         
    f_variant: VariantListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_discr_name :
    [< VariantPartType.t]
    -> IdentifierType.t

      
  val f_variant :
    [< VariantPartType.t]
    -> VariantListType.t


end

module Variant : sig
  (**
   * Single variant in a discriminated type record declaration.
  
   * This corresponds to a ``when ... => ...`` section in a variant part.
   *)

  type t =
    [
      | `Variant of
          VariantType.fields
    ]

  type fields = VariantType.fields =
    
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_choices :
    [< VariantType.t]
    -> AlternativesListType.t

      
  val f_components :
    [< VariantType.t]
    -> ComponentListType.t


end

module UseTypeClause : sig
  (**
   * Use clause for types.
   *)

  type t =
    [
      | `UseTypeClause of
          UseTypeClauseType.fields
    ]

  type fields = UseTypeClauseType.fields =
    
  {
         
    f_has_all: AllNodeType.t
    Lazy.t;
         
    f_types: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_all :
    [< UseTypeClauseType.t]
    -> AllNodeType.t

      
  val f_types :
    [< UseTypeClauseType.t]
    -> NameListType.t


end

module UsePackageClause : sig
  (**
   * Use clause for packages.
   *)

  type t =
    [
      | `UsePackageClause of
          UsePackageClauseType.fields
    ]

  type fields = UsePackageClauseType.fields =
    
  {
         
    f_packages: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_packages :
    [< UsePackageClauseType.t]
    -> NameListType.t


end

module UseClause : sig
  (**
   * Base class for use clauses.
   *)

  type t =
    [
      | UsePackageClause.t
      | UseTypeClause.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module UntilPresent : sig
  

  type t =
    [
      | `UntilPresent of
          UntilPresentType.fields
    ]

  type fields = UntilPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module UntilAbsent : sig
  

  type t =
    [
      | `UntilAbsent of
          UntilAbsentType.fields
    ]

  type fields = UntilAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module UntilNode : sig
  (**
   * Qualifier for the ``until`` keyword.
   *)

  type t =
    [
      | UntilAbsent.t
      | UntilPresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< UntilNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of UntilPresent
   *)



end

module UnconstrainedArrayIndex : sig
  (**
   * List of unconstrained array indexes.
   *)

  type t =
    [
      | `UnconstrainedArrayIndex of
          UnconstrainedArrayIndexType.fields
    ]

  type fields = UnconstrainedArrayIndexType.fields =
    
  {
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_subtype_indication :
    [< UnconstrainedArrayIndexType.t]
    -> SubtypeIndicationType.t


end

module DiscreteSubtypeIndication : sig
  (**
   * Reference to a type with a general constraint.
   *)

  type t =
    [
      | `DiscreteSubtypeIndication of
          DiscreteSubtypeIndicationType.fields
    ]

  type fields = DiscreteSubtypeIndicationType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_not_null :
    [< DiscreteSubtypeIndicationType.t]
    -> NotNullType.t

      
  val f_name :
    [< DiscreteSubtypeIndicationType.t]
    -> [AttributeRefType.t | CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]

      
  val f_constraint :
    [< DiscreteSubtypeIndicationType.t]
    -> ConstraintType.t option


end

module ConstrainedSubtypeIndication : sig
  (**
   * Reference to a type with a range constraint.
   *)

  type t =
    [
      | `ConstrainedSubtypeIndication of
          ConstrainedSubtypeIndicationType.fields
    ]

  type fields = ConstrainedSubtypeIndicationType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_not_null :
    [< ConstrainedSubtypeIndicationType.t]
    -> NotNullType.t

      
  val f_name :
    [< ConstrainedSubtypeIndicationType.t]
    -> [AttributeRefType.t | CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]

      
  val f_constraint :
    [< ConstrainedSubtypeIndicationType.t]
    -> ConstraintType.t option


end

module SubtypeIndication : sig
  (**
   * Reference to a type by name.
   *)

  type t =
    [
      | `SubtypeIndication of
          SubtypeIndicationType.fields
      | `ConstrainedSubtypeIndication of
          ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication of
          DiscreteSubtypeIndicationType.fields
    ]

  type fields = SubtypeIndicationType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_constraint: ConstraintType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_is_static_subtype :
    ?imprecise_fallback:
    bool
    -> [< SubtypeIndicationType.t ]
    -> bool
  (**
   * Returns whether Self denotes a static subtype or not.
   *)


      
  val f_has_not_null :
    [< SubtypeIndicationType.t]
    -> NotNullType.t

      
  val f_name :
    [< SubtypeIndicationType.t]
    -> [AttributeRefType.t | CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]

      
  val f_constraint :
    [< SubtypeIndicationType.t]
    -> ConstraintType.t option


end

module EnumLitSynthTypeExpr : sig
  (**
   * Synthetic node. Represents the type expression for an enum literal.
   *)

  type t =
    [
      | `EnumLitSynthTypeExpr of
          EnumLitSynthTypeExprType.fields
    ]

  type fields = EnumLitSynthTypeExprType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AnonymousType : sig
  (**
   * Container for inline anonymous array and access types declarations.
   *)

  type t =
    [
      | `AnonymousType of
          AnonymousTypeType.fields
    ]

  type fields = AnonymousTypeType.fields =
    
  {
         
    f_type_decl: AnonymousTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_type_decl :
    [< AnonymousTypeType.t]
    -> AnonymousTypeDeclType.t


end

module TypeExpr : sig
  (**
   * A type expression is an abstract node that embodies the concept of a
   * reference to a type.
  
   * Since Ada has both subtype_indications and anonymous (inline) type
   * declarations, a type expression contains one or the other.
   *)

  type t =
    [
      | AnonymousType.t
      | EnumLitSynthTypeExpr.t
      | SubtypeIndication.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_type_name :
    [< TypeExprType.t ]
    -> NameType.t option
  (**
   * Return the name node for this type expression, if applicable, else null
   *)

      
  val p_designated_type_decl :
    [< TypeExprType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Returns the type declaration designated by this type expression.
   *)

      
  val p_designated_type_decl_from :
    [< TypeExprType.t ]
    -> AdaNodeType.t
    -> BaseTypeDeclType.t option
  (**
   * Return the type declaration designated by this type expression as viewed
   * from the node given by origin_node.
   *)



end

module SignedIntTypeDef : sig
  (**
   * Type definition for a signed integer type.
   *)

  type t =
    [
      | `SignedIntTypeDef of
          SignedIntTypeDefType.fields
    ]

  type fields = SignedIntTypeDefType.fields =
    
  {
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_range :
    [< SignedIntTypeDefType.t]
    -> RangeSpecType.t


end

module RecordTypeDef : sig
  (**
   * Type definition for a record.
   *)

  type t =
    [
      | `RecordTypeDef of
          RecordTypeDefType.fields
    ]

  type fields = RecordTypeDefType.fields =
    
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_tagged: TaggedNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_record_def: BaseRecordDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_abstract :
    [< RecordTypeDefType.t]
    -> AbstractNodeType.t

      
  val f_has_tagged :
    [< RecordTypeDefType.t]
    -> TaggedNodeType.t

      
  val f_has_limited :
    [< RecordTypeDefType.t]
    -> LimitedNodeType.t

      
  val f_record_def :
    [< RecordTypeDefType.t]
    -> BaseRecordDefType.t


end

module OrdinaryFixedPointDef : sig
  (**
   * Type definition for ordinary fixed-point numbers.
   *)

  type t =
    [
      | `OrdinaryFixedPointDef of
          OrdinaryFixedPointDefType.fields
    ]

  type fields = OrdinaryFixedPointDefType.fields =
    
  {
         
    f_delta: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_delta :
    [< OrdinaryFixedPointDefType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_range :
    [< OrdinaryFixedPointDefType.t]
    -> RangeSpecType.t option


end

module FloatingPointDef : sig
  (**
   * Type definition for floating-point numbers.
   *)

  type t =
    [
      | `FloatingPointDef of
          FloatingPointDefType.fields
    ]

  type fields = FloatingPointDefType.fields =
    
  {
         
    f_num_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_num_digits :
    [< FloatingPointDefType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_range :
    [< FloatingPointDefType.t]
    -> RangeSpecType.t option


end

module DecimalFixedPointDef : sig
  (**
   * Type definition for decimal fixed-point numbers.
   *)

  type t =
    [
      | `DecimalFixedPointDef of
          DecimalFixedPointDefType.fields
    ]

  type fields = DecimalFixedPointDefType.fields =
    
  {
         
    f_delta: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_delta :
    [< DecimalFixedPointDefType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_digits :
    [< DecimalFixedPointDefType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_range :
    [< DecimalFixedPointDefType.t]
    -> RangeSpecType.t option


end

module RealTypeDef : sig
  (**
   * Type definition for real numbers.
   *)

  type t =
    [
      | DecimalFixedPointDef.t
      | FloatingPointDef.t
      | OrdinaryFixedPointDef.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module PrivateTypeDef : sig
  (**
   * Type definition for a private type.
   *)

  type t =
    [
      | `PrivateTypeDef of
          PrivateTypeDefType.fields
    ]

  type fields = PrivateTypeDefType.fields =
    
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_tagged: TaggedNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_abstract :
    [< PrivateTypeDefType.t]
    -> AbstractNodeType.t

      
  val f_has_tagged :
    [< PrivateTypeDefType.t]
    -> TaggedNodeType.t

      
  val f_has_limited :
    [< PrivateTypeDefType.t]
    -> LimitedNodeType.t


end

module ModIntTypeDef : sig
  (**
   * Type definition for a modular integer type.
   *)

  type t =
    [
      | `ModIntTypeDef of
          ModIntTypeDefType.fields
    ]

  type fields = ModIntTypeDefType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_expr :
    [< ModIntTypeDefType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module InterfaceTypeDef : sig
  (**
   * Type definition for an interface.
   *)

  type t =
    [
      | `InterfaceTypeDef of
          InterfaceTypeDefType.fields
    ]

  type fields = InterfaceTypeDefType.fields =
    
  {
         
    f_interface_kind: InterfaceKindType.t
    option
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_interface_kind :
    [< InterfaceTypeDefType.t]
    -> InterfaceKindType.t option

      
  val f_interfaces :
    [< InterfaceTypeDefType.t]
    -> ParentListType.t option


end

module FormalDiscreteTypeDef : sig
  (**
   * Type definition for discrete types in generic formals.
   *)

  type t =
    [
      | `FormalDiscreteTypeDef of
          FormalDiscreteTypeDefType.fields
    ]

  type fields = FormalDiscreteTypeDefType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module EnumTypeDef : sig
  (**
   * Type definition for enumerations.
   *)

  type t =
    [
      | `EnumTypeDef of
          EnumTypeDefType.fields
    ]

  type fields = EnumTypeDefType.fields =
    
  {
         
    f_enum_literals: EnumLiteralDeclListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_enum_literals :
    [< EnumTypeDefType.t]
    -> EnumLiteralDeclListType.t


end

module DerivedTypeDef : sig
  (**
   * Type definition for a derived type.
   *)

  type t =
    [
      | `DerivedTypeDef of
          DerivedTypeDefType.fields
    ]

  type fields = DerivedTypeDefType.fields =
    
  {
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
         
    f_has_limited: LimitedNodeType.t
    Lazy.t;
         
    f_has_synchronized: SynchronizedNodeType.t
    Lazy.t;
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_record_extension: BaseRecordDefType.t
    option
    Lazy.t;
         
    f_has_with_private: WithPrivateType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_abstract :
    [< DerivedTypeDefType.t]
    -> AbstractNodeType.t

      
  val f_has_limited :
    [< DerivedTypeDefType.t]
    -> LimitedNodeType.t

      
  val f_has_synchronized :
    [< DerivedTypeDefType.t]
    -> SynchronizedNodeType.t

      
  val f_subtype_indication :
    [< DerivedTypeDefType.t]
    -> SubtypeIndicationType.t

      
  val f_interfaces :
    [< DerivedTypeDefType.t]
    -> ParentListType.t option

      
  val f_record_extension :
    [< DerivedTypeDefType.t]
    -> BaseRecordDefType.t option

      
  val f_has_with_private :
    [< DerivedTypeDefType.t]
    -> WithPrivateType.t


end

module ArrayTypeDef : sig
  (**
   * Type definition for an array.
   *)

  type t =
    [
      | `ArrayTypeDef of
          ArrayTypeDefType.fields
    ]

  type fields = ArrayTypeDefType.fields =
    
  {
         
    f_indices: ArrayIndicesType.t
    Lazy.t;
         
    f_component_type: ComponentDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_indices :
    [< ArrayTypeDefType.t]
    -> ArrayIndicesType.t

      
  val f_component_type :
    [< ArrayTypeDefType.t]
    -> ComponentDefType.t


end

module TypeAccessDef : sig
  (**
   * Syntactic type definition for accesses.
   *)

  type t =
    [
      | `TypeAccessDef of
          TypeAccessDefType.fields
    ]

  type fields = TypeAccessDefType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_has_all: AllNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_subtype_indication: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_not_null :
    [< TypeAccessDefType.t]
    -> NotNullType.t

      
  val f_has_all :
    [< TypeAccessDefType.t]
    -> AllNodeType.t

      
  val f_has_constant :
    [< TypeAccessDefType.t]
    -> ConstantNodeType.t

      
  val f_subtype_indication :
    [< TypeAccessDefType.t]
    -> SubtypeIndicationType.t


end

module AnonymousTypeAccessDef : sig
  (**
   * Synthetic type access, that will directly reference a type decl. It is used
   * to generate synthetic anonymous access types.
   *)

  type t =
    [
      | `AnonymousTypeAccessDef of
          AnonymousTypeAccessDefType.fields
    ]

  type fields = AnonymousTypeAccessDefType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_type_decl: BaseTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_not_null :
    [< AnonymousTypeAccessDefType.t]
    -> NotNullType.t

      
  val f_type_decl :
    [< AnonymousTypeAccessDefType.t]
    -> BaseTypeDeclType.t


end

module BaseTypeAccessDef : sig
  (**
   * Base class for access type definitions.
   *)

  type t =
    [
      | AnonymousTypeAccessDef.t
      | TypeAccessDef.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_not_null :
    [< BaseTypeAccessDefType.t]
    -> NotNullType.t


end

module AccessToSubpDef : sig
  (**
   * Type definition for accesses to subprograms.
   *)

  type t =
    [
      | `AccessToSubpDef of
          AccessToSubpDefType.fields
    ]

  type fields = AccessToSubpDefType.fields =
    
  {
         
    f_has_not_null: NotNullType.t
    Lazy.t;
         
    f_has_protected: ProtectedNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_not_null :
    [< AccessToSubpDefType.t]
    -> NotNullType.t

      
  val f_has_protected :
    [< AccessToSubpDefType.t]
    -> ProtectedNodeType.t

      
  val f_subp_spec :
    [< AccessToSubpDefType.t]
    -> SubpSpecType.t


end

module AccessDef : sig
  (**
   * Base class for access type definitions.
   *)

  type t =
    [
      | AccessToSubpDef.t
      | BaseTypeAccessDef.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_not_null :
    [< AccessDefType.t]
    -> NotNullType.t


end

module TypeDef : sig
  (**
   * Base class for type definitions.
   *)

  type t =
    [
      | AccessDef.t
      | ArrayTypeDef.t
      | DerivedTypeDef.t
      | EnumTypeDef.t
      | FormalDiscreteTypeDef.t
      | InterfaceTypeDef.t
      | ModIntTypeDef.t
      | PrivateTypeDef.t
      | RealTypeDef.t
      | RecordTypeDef.t
      | SignedIntTypeDef.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module TaskDef : sig
  (**
   * Type definition for a task type.
   *)

  type t =
    [
      | `TaskDef of
          TaskDefType.fields
    ]

  type fields = TaskDefType.fields =
    
  {
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_interfaces :
    [< TaskDefType.t]
    -> ParentListType.t option

      
  val f_public_part :
    [< TaskDefType.t]
    -> PublicPartType.t

      
  val f_private_part :
    [< TaskDefType.t]
    -> PrivatePartType.t option

      
  val f_end_name :
    [< TaskDefType.t]
    -> EndNameType.t option


end

module TaggedPresent : sig
  

  type t =
    [
      | `TaggedPresent of
          TaggedPresentType.fields
    ]

  type fields = TaggedPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module TaggedAbsent : sig
  

  type t =
    [
      | `TaggedAbsent of
          TaggedAbsentType.fields
    ]

  type fields = TaggedAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module TaggedNode : sig
  (**
   * Qualifier for the ``tagged`` keyword.
   *)

  type t =
    [
      | TaggedAbsent.t
      | TaggedPresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< TaggedNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of TaggedPresent
   *)



end

module SynchronizedPresent : sig
  

  type t =
    [
      | `SynchronizedPresent of
          SynchronizedPresentType.fields
    ]

  type fields = SynchronizedPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module SynchronizedAbsent : sig
  

  type t =
    [
      | `SynchronizedAbsent of
          SynchronizedAbsentType.fields
    ]

  type fields = SynchronizedAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module SynchronizedNode : sig
  (**
   * Qualifier for the ``synchronized`` keyword.
   *)

  type t =
    [
      | SynchronizedAbsent.t
      | SynchronizedPresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< SynchronizedNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of SynchronizedPresent
   *)



end

module Subunit : sig
  (**
   * Subunit (``separate``).
   *)

  type t =
    [
      | `Subunit of
          SubunitType.fields
    ]

  type fields = SubunitType.fields =
    
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_body: [
      | `PackageBody
          of PackageBodyType.fields
      | `ProtectedBody
          of ProtectedBodyType.fields
      | `SubpBody
          of SubpBodyType.fields
      | `TaskBody
          of TaskBodyType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_body_root :
    [< SubunitType.t ]
    -> BasicDeclType.t option
  (**
   * Return the body in which this subunit is rooted.
   *)


      
  val f_name :
    [< SubunitType.t]
    -> [CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]

      
  val f_body :
    [< SubunitType.t]
    -> [PackageBodyType.t | ProtectedBodyType.t | SubpBodyType.t | TaskBodyType.t]


end

module SubpKindProcedure : sig
  

  type t =
    [
      | `SubpKindProcedure of
          SubpKindProcedureType.fields
    ]

  type fields = SubpKindProcedureType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module SubpKindFunction : sig
  

  type t =
    [
      | `SubpKindFunction of
          SubpKindFunctionType.fields
    ]

  type fields = SubpKindFunctionType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module SubpKind : sig
  (**
   * Qualifier for a subprogram kind.
   *)

  type t =
    [
      | SubpKindFunction.t
      | SubpKindProcedure.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module TerminateAlternative : sig
  (**
   * ``terminate`` alternative in a ``select`` statement.
   *)

  type t =
    [
      | `TerminateAlternative of
          TerminateAlternativeType.fields
    ]

  type fields = TerminateAlternativeType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ReturnStmt : sig
  (**
   * ``return`` statement.
   *)

  type t =
    [
      | `ReturnStmt of
          ReturnStmtType.fields
    ]

  type fields = ReturnStmtType.fields =
    
  {
         
    f_return_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_return_expr :
    [< ReturnStmtType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option


end

module RequeueStmt : sig
  (**
   * ``requeue`` statement.
   *)

  type t =
    [
      | `RequeueStmt of
          RequeueStmtType.fields
    ]

  type fields = RequeueStmtType.fields =
    
  {
         
    f_call_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_has_abort: AbortNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_call_name :
    [< RequeueStmtType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]

      
  val f_has_abort :
    [< RequeueStmtType.t]
    -> AbortNodeType.t


end

module RaiseStmt : sig
  (**
   * ``raise`` statement.
   *)

  type t =
    [
      | `RaiseStmt of
          RaiseStmtType.fields
    ]

  type fields = RaiseStmtType.fields =
    
  {
         
    f_exception_name: NameType.t
    option
    Lazy.t;
         
    f_error_message: ExprType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_exception_name :
    [< RaiseStmtType.t]
    -> NameType.t option

      
  val f_error_message :
    [< RaiseStmtType.t]
    -> ExprType.t option


end

module NullStmt : sig
  (**
   * ``null;`` statement.
   *)

  type t =
    [
      | `NullStmt of
          NullStmtType.fields
    ]

  type fields = NullStmtType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module Label : sig
  (**
   * Statement to declare a code label.
   *)

  type t =
    [
      | `Label of
          LabelType.fields
    ]

  type fields = LabelType.fields =
    
  {
         
    f_decl: LabelDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decl :
    [< LabelType.t]
    -> LabelDeclType.t


end

module GotoStmt : sig
  (**
   * ``goto`` statement.
   *)

  type t =
    [
      | `GotoStmt of
          GotoStmtType.fields
    ]

  type fields = GotoStmtType.fields =
    
  {
         
    f_label_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_label_name :
    [< GotoStmtType.t]
    -> [CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]


end

module ExitStmt : sig
  (**
   * ``exit`` statement.
   *)

  type t =
    [
      | `ExitStmt of
          ExitStmtType.fields
    ]

  type fields = ExitStmtType.fields =
    
  {
         
    f_loop_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    option
    Lazy.t;
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_loop_name :
    [< ExitStmtType.t]
    -> [CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t] option

      
  val f_cond_expr :
    [< ExitStmtType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option


end

module DelayStmt : sig
  (**
   * ``delay`` statement.
   *)

  type t =
    [
      | `DelayStmt of
          DelayStmtType.fields
    ]

  type fields = DelayStmtType.fields =
    
  {
         
    f_has_until: UntilNodeType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_until :
    [< DelayStmtType.t]
    -> UntilNodeType.t

      
  val f_expr :
    [< DelayStmtType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module CallStmt : sig
  (**
   * Statement for entry or procedure calls.
   *)

  type t =
    [
      | `CallStmt of
          CallStmtType.fields
    ]

  type fields = CallStmtType.fields =
    
  {
         
    f_call: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_call :
    [< CallStmtType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]


end

module AssignStmt : sig
  (**
   * Statement for assignments.
   *)

  type t =
    [
      | `AssignStmt of
          AssignStmtType.fields
    ]

  type fields = AssignStmtType.fields =
    
  {
         
    f_dest: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_dest :
    [< AssignStmtType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]

      
  val f_expr :
    [< AssignStmtType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module AbortStmt : sig
  (**
   * ``abort`` statement.
   *)

  type t =
    [
      | `AbortStmt of
          AbortStmtType.fields
    ]

  type fields = AbortStmtType.fields =
    
  {
         
    f_names: NameListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_names :
    [< AbortStmtType.t]
    -> NameListType.t


end

module SimpleStmt : sig
  (**
   * Base class for simple statements.
   *)

  type t =
    [
      | AbortStmt.t
      | AssignStmt.t
      | CallStmt.t
      | DelayStmt.t
      | ExitStmt.t
      | GotoStmt.t
      | Label.t
      | NullStmt.t
      | RaiseStmt.t
      | RequeueStmt.t
      | ReturnStmt.t
      | TerminateAlternative.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ErrorStmt : sig
  (**
   * Placeholder node for syntax errors in lists of statements.
   *)

  type t =
    [
      | `ErrorStmt of
          ErrorStmtType.fields
    ]

  type fields = ErrorStmtType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module SelectStmt : sig
  (**
   * ``select`` statements block.
   *)

  type t =
    [
      | `SelectStmt of
          SelectStmtType.fields
    ]

  type fields = SelectStmtType.fields =
    
  {
         
    f_guards: SelectWhenPartListType.t
    Lazy.t;
         
    f_else_stmts: StmtListType.t
    option
    Lazy.t;
         
    f_abort_stmts: StmtListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_guards :
    [< SelectStmtType.t]
    -> SelectWhenPartListType.t

      
  val f_else_stmts :
    [< SelectStmtType.t]
    -> StmtListType.t option

      
  val f_abort_stmts :
    [< SelectStmtType.t]
    -> StmtListType.t option


end

module NamedStmt : sig
  (**
   * Wrapper class, used for composite statements that can be named (declare
   * blocks, loops). This allows to both have a BasicDecl for the named entity
   * declared, and a CompositeStmt for the statement hierarchy.
   *)

  type t =
    [
      | `NamedStmt of
          NamedStmtType.fields
    ]

  type fields = NamedStmtType.fields =
    
  {
         
    f_decl: NamedStmtDeclType.t
    Lazy.t;
         
    f_stmt: [
      | `BeginBlock
          of BeginBlockType.fields
      | `DeclBlock
          of DeclBlockType.fields
      | `ForLoopStmt
          of ForLoopStmtType.fields
      | `LoopStmt
          of LoopStmtType.fields
      | `WhileLoopStmt
          of WhileLoopStmtType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decl :
    [< NamedStmtType.t]
    -> NamedStmtDeclType.t

      
  val f_stmt :
    [< NamedStmtType.t]
    -> [BaseLoopStmtType.t | BlockStmtType.t]


end

module IfStmt : sig
  (**
   * ``if`` statement block.
   *)

  type t =
    [
      | `IfStmt of
          IfStmtType.fields
    ]

  type fields = IfStmtType.fields =
    
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_stmts: StmtListType.t
    Lazy.t;
         
    f_alternatives: ElsifStmtPartListType.t
    Lazy.t;
         
    f_else_stmts: StmtListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_cond_expr :
    [< IfStmtType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_then_stmts :
    [< IfStmtType.t]
    -> StmtListType.t

      
  val f_alternatives :
    [< IfStmtType.t]
    -> ElsifStmtPartListType.t

      
  val f_else_stmts :
    [< IfStmtType.t]
    -> StmtListType.t option


end

module ExtendedReturnStmt : sig
  (**
   * Extended ``return`` statement.
   *)

  type t =
    [
      | `ExtendedReturnStmt of
          ExtendedReturnStmtType.fields
    ]

  type fields = ExtendedReturnStmtType.fields =
    
  {
         
    f_decl: ExtendedReturnStmtObjectDeclType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decl :
    [< ExtendedReturnStmtType.t]
    -> ExtendedReturnStmtObjectDeclType.t

      
  val f_stmts :
    [< ExtendedReturnStmtType.t]
    -> HandledStmtsType.t option


end

module CaseStmt : sig
  (**
   * ``case`` statement.
   *)

  type t =
    [
      | `CaseStmt of
          CaseStmtType.fields
    ]

  type fields = CaseStmtType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_alternatives: CaseStmtAlternativeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_expr :
    [< CaseStmtType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_alternatives :
    [< CaseStmtType.t]
    -> CaseStmtAlternativeListType.t


end

module DeclBlock : sig
  (**
   * Statement block with a declarative part.
   *)

  type t =
    [
      | `DeclBlock of
          DeclBlockType.fields
    ]

  type fields = DeclBlockType.fields =
    
  {
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decls :
    [< DeclBlockType.t]
    -> DeclarativePartType.t

      
  val f_stmts :
    [< DeclBlockType.t]
    -> HandledStmtsType.t

      
  val f_end_name :
    [< DeclBlockType.t]
    -> EndNameType.t option


end

module BeginBlock : sig
  (**
   * Statement block with no declarative part.
   *)

  type t =
    [
      | `BeginBlock of
          BeginBlockType.fields
    ]

  type fields = BeginBlockType.fields =
    
  {
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_stmts :
    [< BeginBlockType.t]
    -> HandledStmtsType.t

      
  val f_end_name :
    [< BeginBlockType.t]
    -> EndNameType.t option


end

module BlockStmt : sig
  (**
   * Base class for statement blocks.
   *)

  type t =
    [
      | BeginBlock.t
      | DeclBlock.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module WhileLoopStmt : sig
  (**
   * Statement for ``while`` loops (``while ... loop ... end loop;``).
   *)

  type t =
    [
      | `WhileLoopStmt of
          WhileLoopStmtType.fields
    ]

  type fields = WhileLoopStmtType.fields =
    
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_spec :
    [< WhileLoopStmtType.t]
    -> LoopSpecType.t option

      
  val f_stmts :
    [< WhileLoopStmtType.t]
    -> StmtListType.t

      
  val f_end_name :
    [< WhileLoopStmtType.t]
    -> EndNameType.t option


end

module LoopStmt : sig
  (**
   * Statement for simple loops (``loop ... end loop;``).
   *)

  type t =
    [
      | `LoopStmt of
          LoopStmtType.fields
    ]

  type fields = LoopStmtType.fields =
    
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_spec :
    [< LoopStmtType.t]
    -> LoopSpecType.t option

      
  val f_stmts :
    [< LoopStmtType.t]
    -> StmtListType.t

      
  val f_end_name :
    [< LoopStmtType.t]
    -> EndNameType.t option


end

module ForLoopStmt : sig
  (**
   * Statement for ``for`` loops (``for ... loop ... end loop;``).
   *)

  type t =
    [
      | `ForLoopStmt of
          ForLoopStmtType.fields
    ]

  type fields = ForLoopStmtType.fields =
    
  {
         
    f_spec: LoopSpecType.t
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_spec :
    [< ForLoopStmtType.t]
    -> LoopSpecType.t option

      
  val f_stmts :
    [< ForLoopStmtType.t]
    -> StmtListType.t

      
  val f_end_name :
    [< ForLoopStmtType.t]
    -> EndNameType.t option


end

module BaseLoopStmt : sig
  (**
   * Base class for loop statements.
   *)

  type t =
    [
      | ForLoopStmt.t
      | LoopStmt.t
      | WhileLoopStmt.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_spec :
    [< BaseLoopStmtType.t]
    -> LoopSpecType.t option

      
  val f_stmts :
    [< BaseLoopStmtType.t]
    -> StmtListType.t

      
  val f_end_name :
    [< BaseLoopStmtType.t]
    -> EndNameType.t option


end

module AcceptStmtWithStmts : sig
  (**
   * Extended ``accept`` statement.
   *)

  type t =
    [
      | `AcceptStmtWithStmts of
          AcceptStmtWithStmtsType.fields
    ]

  type fields = AcceptStmtWithStmtsType.fields =
    
  {
         
    f_name: IdentifierType.t
    Lazy.t;
         
    f_entry_index_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< AcceptStmtWithStmtsType.t]
    -> IdentifierType.t

      
  val f_entry_index_expr :
    [< AcceptStmtWithStmtsType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option

      
  val f_params :
    [< AcceptStmtWithStmtsType.t]
    -> EntryCompletionFormalParamsType.t

      
  val f_stmts :
    [< AcceptStmtWithStmtsType.t]
    -> HandledStmtsType.t

      
  val f_end_name :
    [< AcceptStmtWithStmtsType.t]
    -> EndNameType.t option


end

module AcceptStmt : sig
  (**
   * ``accept`` statement.
   *)

  type t =
    [
      | `AcceptStmt of
          AcceptStmtType.fields
      | `AcceptStmtWithStmts of
          AcceptStmtWithStmtsType.fields
    ]

  type fields = AcceptStmtType.fields =
    
  {
         
    f_name: IdentifierType.t
    Lazy.t;
         
    f_entry_index_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< AcceptStmtType.t]
    -> IdentifierType.t

      
  val f_entry_index_expr :
    [< AcceptStmtType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option

      
  val f_params :
    [< AcceptStmtType.t]
    -> EntryCompletionFormalParamsType.t


end

module CompositeStmt : sig
  (**
   * Base class for composite statements.
   *)

  type t =
    [
      | AcceptStmt.t
      | BaseLoopStmt.t
      | BlockStmt.t
      | CaseStmt.t
      | ExtendedReturnStmt.t
      | IfStmt.t
      | NamedStmt.t
      | SelectStmt.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module Stmt : sig
  (**
   * Bass class for statements.
   *)

  type t =
    [
      | CompositeStmt.t
      | ErrorStmt.t
      | SimpleStmt.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module SelectWhenPart : sig
  (**
   * Alternative part in a ``select`` statements block.
   *)

  type t =
    [
      | `SelectWhenPart of
          SelectWhenPartType.fields
    ]

  type fields = SelectWhenPartType.fields =
    
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_cond_expr :
    [< SelectWhenPartType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option

      
  val f_stmts :
    [< SelectWhenPartType.t]
    -> StmtListType.t


end

module ReversePresent : sig
  

  type t =
    [
      | `ReversePresent of
          ReversePresentType.fields
    ]

  type fields = ReversePresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ReverseAbsent : sig
  

  type t =
    [
      | `ReverseAbsent of
          ReverseAbsentType.fields
    ]

  type fields = ReverseAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ReverseNode : sig
  (**
   * Qualifier for the ``reverse`` keyword.
   *)

  type t =
    [
      | ReverseAbsent.t
      | ReversePresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< ReverseNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of ReversePresent
   *)



end

module SyntheticRenamingClause : sig
  (**
   * Synthetic renaming clause. Used to synthesize object decls with renamings.
   * (See to_anonymous_object_decl).
   *)

  type t =
    [
      | `SyntheticRenamingClause of
          SyntheticRenamingClauseType.fields
    ]

  type fields = SyntheticRenamingClauseType.fields =
    
  {
         
    f_renamed_object: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_renamed_object :
    [< SyntheticRenamingClauseType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]


end

module RenamingClause : sig
  (**
   * Renaming clause, used everywhere renamings are valid.
   *)

  type t =
    [
      | `RenamingClause of
          RenamingClauseType.fields
      | `SyntheticRenamingClause of
          SyntheticRenamingClauseType.fields
    ]

  type fields = RenamingClauseType.fields =
    
  {
         
    f_renamed_object: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_renamed_object :
    [< RenamingClauseType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]


end

module RangeSpec : sig
  (**
   * Range specification.
   *)

  type t =
    [
      | `RangeSpec of
          RangeSpecType.fields
    ]

  type fields = RangeSpecType.fields =
    
  {
         
    f_range: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_range :
    [< RangeSpecType.t]
    -> [AttributeRefType.t | BinOpType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]


end

module QuantifierSome : sig
  

  type t =
    [
      | `QuantifierSome of
          QuantifierSomeType.fields
    ]

  type fields = QuantifierSomeType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module QuantifierAll : sig
  

  type t =
    [
      | `QuantifierAll of
          QuantifierAllType.fields
    ]

  type fields = QuantifierAllType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module Quantifier : sig
  (**
   * Type for quantified expressions.
   *)

  type t =
    [
      | QuantifierAll.t
      | QuantifierSome.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ProtectedPresent : sig
  

  type t =
    [
      | `ProtectedPresent of
          ProtectedPresentType.fields
    ]

  type fields = ProtectedPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ProtectedAbsent : sig
  

  type t =
    [
      | `ProtectedAbsent of
          ProtectedAbsentType.fields
    ]

  type fields = ProtectedAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ProtectedNode : sig
  (**
   * Qualifier for the ``protected`` keyword.
   *)

  type t =
    [
      | ProtectedAbsent.t
      | ProtectedPresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< ProtectedNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of ProtectedPresent
   *)



end

module ProtectedDef : sig
  (**
   * Type definition for a protected object.
   *)

  type t =
    [
      | `ProtectedDef of
          ProtectedDefType.fields
    ]

  type fields = ProtectedDefType.fields =
    
  {
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_public_part :
    [< ProtectedDefType.t]
    -> PublicPartType.t

      
  val f_private_part :
    [< ProtectedDefType.t]
    -> PrivatePartType.t option

      
  val f_end_name :
    [< ProtectedDefType.t]
    -> EndNameType.t option


end

module PrivatePresent : sig
  

  type t =
    [
      | `PrivatePresent of
          PrivatePresentType.fields
    ]

  type fields = PrivatePresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module PrivateAbsent : sig
  

  type t =
    [
      | `PrivateAbsent of
          PrivateAbsentType.fields
    ]

  type fields = PrivateAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module PrivateNode : sig
  (**
   * Qualifier for the ``private`` keyword.
   *)

  type t =
    [
      | PrivateAbsent.t
      | PrivatePresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< PrivateNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of PrivatePresent
   *)



end

module PragmaNode : sig
  (**
   * Class for pragmas (RM 2.8). Pragmas are compiler directives, that can be
   * language or compiler defined.
   *)

  type t =
    [
      | `PragmaNode of
          PragmaNodeType.fields
    ]

  type fields = PragmaNodeType.fields =
    
  {
         
    f_id: IdentifierType.t
    Lazy.t;
         
    f_args: BaseAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_associated_decls :
    [< PragmaNodeType.t ]
    -> BasicDeclType.t list
  (**
   * Return an array of ``BasicDecl`` instances associated with this pragma, or
   * an empty array if non applicable.
   *)


      
  val f_id :
    [< PragmaNodeType.t]
    -> IdentifierType.t

      
  val f_args :
    [< PragmaNodeType.t]
    -> BaseAssocListType.t option


end

module ParenAbstractStateDecl : sig
  (**
   * Holds an AbstractStateDecl between parentheses. Needed to support the
   * syntax:
  
   * .. code:: ada
  
   * package Pkg with Abstract_State => (A, (B with Some_Aspect))
   *)

  type t =
    [
      | `ParenAbstractStateDecl of
          ParenAbstractStateDeclType.fields
    ]

  type fields = ParenAbstractStateDeclType.fields =
    
  {
         
    f_decl: [
      | `AbstractStateDecl
          of AbstractStateDeclType.fields
      | `ParenAbstractStateDecl
          of fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decl :
    [< ParenAbstractStateDeclType.t]
    -> [AbstractStateDeclType.t | ParenAbstractStateDeclType.t]


end

module Params : sig
  (**
   * List of parameter specifications.
   *)

  type t =
    [
      | `Params of
          ParamsType.fields
    ]

  type fields = ParamsType.fields =
    
  {
         
    f_params: ParamSpecListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_params :
    [< ParamsType.t]
    -> ParamSpecListType.t


end

module OverridingUnspecified : sig
  

  type t =
    [
      | `OverridingUnspecified of
          OverridingUnspecifiedType.fields
    ]

  type fields = OverridingUnspecifiedType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OverridingOverriding : sig
  

  type t =
    [
      | `OverridingOverriding of
          OverridingOverridingType.fields
    ]

  type fields = OverridingOverridingType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OverridingNotOverriding : sig
  

  type t =
    [
      | `OverridingNotOverriding of
          OverridingNotOverridingType.fields
    ]

  type fields = OverridingNotOverridingType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OverridingNode : sig
  (**
   * Syntactic indicators for subprogram overriding modes.
   *)

  type t =
    [
      | OverridingNotOverriding.t
      | OverridingOverriding.t
      | OverridingUnspecified.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OthersDesignator : sig
  (**
   * ``other`` designator.
   *)

  type t =
    [
      | `OthersDesignator of
          OthersDesignatorType.fields
    ]

  type fields = OthersDesignatorType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module NullComponentDecl : sig
  (**
   * Placeholder for the ``null`` in lists of components.
   *)

  type t =
    [
      | `NullComponentDecl of
          NullComponentDeclType.fields
    ]

  type fields = NullComponentDeclType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module NotNullPresent : sig
  

  type t =
    [
      | `NotNullPresent of
          NotNullPresentType.fields
    ]

  type fields = NotNullPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module NotNullAbsent : sig
  

  type t =
    [
      | `NotNullAbsent of
          NotNullAbsentType.fields
    ]

  type fields = NotNullAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module NotNull : sig
  (**
   * Qualifier for the ``not null`` keywords.
   *)

  type t =
    [
      | NotNullAbsent.t
      | NotNullPresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< NotNullType.t ]
    -> bool
  (**
   * Return whether this is an instance of NotNullPresent
   *)



end

module MultiAbstractStateDecl : sig
  (**
   * Node that holds several AbstractStateDecl nodes, which is necessary when
   * the Abstract_State aspect is associated with an aggregate in order to
   * declare a list of abstract states.
   *)

  type t =
    [
      | `MultiAbstractStateDecl of
          MultiAbstractStateDeclType.fields
    ]

  type fields = MultiAbstractStateDeclType.fields =
    
  {
         
    f_decls: AbstractStateDeclListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decls :
    [< MultiAbstractStateDeclType.t]
    -> AbstractStateDeclListType.t


end

module ModeOut : sig
  

  type t =
    [
      | `ModeOut of
          ModeOutType.fields
    ]

  type fields = ModeOutType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ModeInOut : sig
  

  type t =
    [
      | `ModeInOut of
          ModeInOutType.fields
    ]

  type fields = ModeInOutType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ModeIn : sig
  

  type t =
    [
      | `ModeIn of
          ModeInType.fields
    ]

  type fields = ModeInType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ModeDefault : sig
  

  type t =
    [
      | `ModeDefault of
          ModeDefaultType.fields
    ]

  type fields = ModeDefaultType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module Mode : sig
  (**
   * Syntactic indicators for passing modes in formals.
   *)

  type t =
    [
      | ModeDefault.t
      | ModeIn.t
      | ModeInOut.t
      | ModeOut.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module WhileLoopSpec : sig
  (**
   * Specification for a ``while`` loop.
   *)

  type t =
    [
      | `WhileLoopSpec of
          WhileLoopSpecType.fields
    ]

  type fields = WhileLoopSpecType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_expr :
    [< WhileLoopSpecType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module ForLoopSpec : sig
  (**
   * Specification for a ``for`` loop.
   *)

  type t =
    [
      | `ForLoopSpec of
          ForLoopSpecType.fields
    ]

  type fields = ForLoopSpecType.fields =
    
  {
         
    f_var_decl: ForLoopVarDeclType.t
    Lazy.t;
         
    f_loop_type: IterTypeType.t
    Lazy.t;
         
    f_has_reverse: ReverseNodeType.t
    Lazy.t;
         
    f_iter_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_var_decl :
    [< ForLoopSpecType.t]
    -> ForLoopVarDeclType.t

      
  val f_loop_type :
    [< ForLoopSpecType.t]
    -> IterTypeType.t

      
  val f_has_reverse :
    [< ForLoopSpecType.t]
    -> ReverseNodeType.t

      
  val f_iter_expr :
    [< ForLoopSpecType.t]
    -> [AttributeRefType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | DiscreteSubtypeIndicationType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]


end

module LoopSpec : sig
  (**
   * Base class for loop specifications.
   *)

  type t =
    [
      | ForLoopSpec.t
      | WhileLoopSpec.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module LimitedPresent : sig
  

  type t =
    [
      | `LimitedPresent of
          LimitedPresentType.fields
    ]

  type fields = LimitedPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module LimitedAbsent : sig
  

  type t =
    [
      | `LimitedAbsent of
          LimitedAbsentType.fields
    ]

  type fields = LimitedAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module LimitedNode : sig
  (**
   * Qualifier for the ``limited`` keyword.
   *)

  type t =
    [
      | LimitedAbsent.t
      | LimitedPresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< LimitedNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of LimitedPresent
   *)



end

module LibraryItem : sig
  (**
   * Library item in a compilation unit.
   *)

  type t =
    [
      | `LibraryItem of
          LibraryItemType.fields
    ]

  type fields = LibraryItemType.fields =
    
  {
         
    f_has_private: PrivateNodeType.t
    Lazy.t;
         
    f_item: [
      | `AbstractSubpDecl
          of AbstractSubpDeclType.fields
      | `ErrorDecl
          of ErrorDeclType.fields
      | `ExprFunction
          of ExprFunctionType.fields
      | `GenericPackageDecl
          of GenericPackageDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericPackageRenamingDecl
          of GenericPackageRenamingDeclType.fields
      | `GenericSubpDecl
          of GenericSubpDeclType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `GenericSubpRenamingDecl
          of GenericSubpRenamingDeclType.fields
      | `NullSubpDecl
          of NullSubpDeclType.fields
      | `PackageBody
          of PackageBodyType.fields
      | `PackageDecl
          of PackageDeclType.fields
      | `PackageRenamingDecl
          of PackageRenamingDeclType.fields
      | `SubpBody
          of SubpBodyType.fields
      | `SubpDecl
          of SubpDeclType.fields
      | `SubpRenamingDecl
          of SubpRenamingDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_private :
    [< LibraryItemType.t]
    -> PrivateNodeType.t

      
  val f_item :
    [< LibraryItemType.t]
    -> [AbstractSubpDeclType.t | BaseSubpBodyType.t | ErrorDeclType.t | GenericDeclType.t | GenericInstantiationType.t | GenericRenamingDeclType.t | PackageBodyType.t | PackageDeclType.t | PackageRenamingDeclType.t | SubpDeclType.t]


end

module IterTypeOf : sig
  

  type t =
    [
      | `IterTypeOf of
          IterTypeOfType.fields
    ]

  type fields = IterTypeOfType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module IterTypeIn : sig
  

  type t =
    [
      | `IterTypeIn of
          IterTypeInType.fields
    ]

  type fields = IterTypeInType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module IterType : sig
  (**
   * Iteration type for ``for`` loops.
   *)

  type t =
    [
      | IterTypeIn.t
      | IterTypeOf.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module InterfaceKindTask : sig
  

  type t =
    [
      | `InterfaceKindTask of
          InterfaceKindTaskType.fields
    ]

  type fields = InterfaceKindTaskType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module InterfaceKindSynchronized : sig
  

  type t =
    [
      | `InterfaceKindSynchronized of
          InterfaceKindSynchronizedType.fields
    ]

  type fields = InterfaceKindSynchronizedType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module InterfaceKindProtected : sig
  

  type t =
    [
      | `InterfaceKindProtected of
          InterfaceKindProtectedType.fields
    ]

  type fields = InterfaceKindProtectedType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module InterfaceKindLimited : sig
  

  type t =
    [
      | `InterfaceKindLimited of
          InterfaceKindLimitedType.fields
    ]

  type fields = InterfaceKindLimitedType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module InterfaceKind : sig
  (**
   * Kind of interface type.
   *)

  type t =
    [
      | InterfaceKindLimited.t
      | InterfaceKindProtected.t
      | InterfaceKindSynchronized.t
      | InterfaceKindTask.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module HandledStmts : sig
  (**
   * List of statements, with optional exception handlers.
   *)

  type t =
    [
      | `HandledStmts of
          HandledStmtsType.fields
    ]

  type fields = HandledStmtsType.fields =
    
  {
         
    f_stmts: StmtListType.t
    Lazy.t;
         
    f_exceptions: AdaNodeListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_stmts :
    [< HandledStmtsType.t]
    -> StmtListType.t

      
  val f_exceptions :
    [< HandledStmtsType.t]
    -> AdaNodeListType.t option


end

module UnOp : sig
  (**
   * Unary expression.
   *)

  type t =
    [
      | `UnOp of
          UnOpType.fields
    ]

  type fields = UnOpType.fields =
    
  {
         
    f_op: [
      | `OpAbs
          of OpAbsType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpNot
          of OpNotType.fields
      | `OpPlus
          of OpPlusType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_op :
    [< UnOpType.t]
    -> [OpAbsType.t | OpMinusType.t | OpNotType.t | OpPlusType.t]

      
  val f_expr :
    [< UnOpType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module RaiseExpr : sig
  (**
   * Expression to raise an exception.
   *)

  type t =
    [
      | `RaiseExpr of
          RaiseExprType.fields
    ]

  type fields = RaiseExprType.fields =
    
  {
         
    f_exception_name: NameType.t
    option
    Lazy.t;
         
    f_error_message: ExprType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_exception_name :
    [< RaiseExprType.t]
    -> NameType.t option

      
  val f_error_message :
    [< RaiseExprType.t]
    -> ExprType.t option


end

module QuantifiedExpr : sig
  (**
   * Quantified expression.
   *)

  type t =
    [
      | `QuantifiedExpr of
          QuantifiedExprType.fields
    ]

  type fields = QuantifiedExprType.fields =
    
  {
         
    f_quantifier: QuantifierType.t
    Lazy.t;
         
    f_loop_spec: ForLoopSpecType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_quantifier :
    [< QuantifiedExprType.t]
    -> QuantifierType.t

      
  val f_loop_spec :
    [< QuantifiedExprType.t]
    -> ForLoopSpecType.t

      
  val f_expr :
    [< QuantifiedExprType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module ParenExpr : sig
  (**
   * Parenthesized expression.
   *)

  type t =
    [
      | `ParenExpr of
          ParenExprType.fields
    ]

  type fields = ParenExprType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_expr :
    [< ParenExprType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module TargetName : sig
  (**
   * Name for Ada 2020 ``@``.
   *)

  type t =
    [
      | `TargetName of
          TargetNameType.fields
    ]

  type fields = TargetNameType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module RealLiteral : sig
  (**
   * Literal for a real number.
   *)

  type t =
    [
      | `RealLiteral of
          RealLiteralType.fields
    ]

  type fields = RealLiteralType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module IntLiteral : sig
  (**
   * Literal for an integer.
   *)

  type t =
    [
      | `IntLiteral of
          IntLiteralType.fields
    ]

  type fields = IntLiteralType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_denoted_value :
    [< IntLiteralType.t ]
    -> BigInteger.t
  (**
   * Return the value that this literal denotes.
   *)



end

module NumLiteral : sig
  (**
   * Base class for number literals.
   *)

  type t =
    [
      | IntLiteral.t
      | RealLiteral.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module NullLiteral : sig
  (**
   * The ``null`` literal.
   *)

  type t =
    [
      | `NullLiteral of
          NullLiteralType.fields
    ]

  type fields = NullLiteralType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module StringLiteral : sig
  (**
   * String literal.
   *)

  type t =
    [
      | `StringLiteral of
          StringLiteralType.fields
    ]

  type fields = StringLiteralType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_denoted_value :
    [< StringLiteralType.t ]
    -> string
  (**
   * Return the value that this literal denotes.
   *)



end

module OpXor : sig
  

  type t =
    [
      | `OpXor of
          OpXorType.fields
    ]

  type fields = OpXorType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpRem : sig
  

  type t =
    [
      | `OpRem of
          OpRemType.fields
    ]

  type fields = OpRemType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpPow : sig
  

  type t =
    [
      | `OpPow of
          OpPowType.fields
    ]

  type fields = OpPowType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpPlus : sig
  

  type t =
    [
      | `OpPlus of
          OpPlusType.fields
    ]

  type fields = OpPlusType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpOrElse : sig
  

  type t =
    [
      | `OpOrElse of
          OpOrElseType.fields
    ]

  type fields = OpOrElseType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpOr : sig
  

  type t =
    [
      | `OpOr of
          OpOrType.fields
    ]

  type fields = OpOrType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpNotIn : sig
  

  type t =
    [
      | `OpNotIn of
          OpNotInType.fields
    ]

  type fields = OpNotInType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpNot : sig
  

  type t =
    [
      | `OpNot of
          OpNotType.fields
    ]

  type fields = OpNotType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpNeq : sig
  

  type t =
    [
      | `OpNeq of
          OpNeqType.fields
    ]

  type fields = OpNeqType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpMult : sig
  

  type t =
    [
      | `OpMult of
          OpMultType.fields
    ]

  type fields = OpMultType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpMod : sig
  

  type t =
    [
      | `OpMod of
          OpModType.fields
    ]

  type fields = OpModType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpMinus : sig
  

  type t =
    [
      | `OpMinus of
          OpMinusType.fields
    ]

  type fields = OpMinusType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpLte : sig
  

  type t =
    [
      | `OpLte of
          OpLteType.fields
    ]

  type fields = OpLteType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpLt : sig
  

  type t =
    [
      | `OpLt of
          OpLtType.fields
    ]

  type fields = OpLtType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpIn : sig
  

  type t =
    [
      | `OpIn of
          OpInType.fields
    ]

  type fields = OpInType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpGte : sig
  

  type t =
    [
      | `OpGte of
          OpGteType.fields
    ]

  type fields = OpGteType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpGt : sig
  

  type t =
    [
      | `OpGt of
          OpGtType.fields
    ]

  type fields = OpGtType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpEq : sig
  

  type t =
    [
      | `OpEq of
          OpEqType.fields
    ]

  type fields = OpEqType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpDoubleDot : sig
  

  type t =
    [
      | `OpDoubleDot of
          OpDoubleDotType.fields
    ]

  type fields = OpDoubleDotType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpDiv : sig
  

  type t =
    [
      | `OpDiv of
          OpDivType.fields
    ]

  type fields = OpDivType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpConcat : sig
  

  type t =
    [
      | `OpConcat of
          OpConcatType.fields
    ]

  type fields = OpConcatType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpAndThen : sig
  

  type t =
    [
      | `OpAndThen of
          OpAndThenType.fields
    ]

  type fields = OpAndThenType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpAnd : sig
  

  type t =
    [
      | `OpAnd of
          OpAndType.fields
    ]

  type fields = OpAndType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module OpAbs : sig
  

  type t =
    [
      | `OpAbs of
          OpAbsType.fields
    ]

  type fields = OpAbsType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module Op : sig
  (**
   * Operation in a binary expression.
  
   * Note that the ARM does not consider "double_dot" ("..") as a binary
   * operator, but we process it this way here anyway to keep things simple.
   *)

  type t =
    [
      | OpAbs.t
      | OpAnd.t
      | OpAndThen.t
      | OpConcat.t
      | OpDiv.t
      | OpDoubleDot.t
      | OpEq.t
      | OpGt.t
      | OpGte.t
      | OpIn.t
      | OpLt.t
      | OpLte.t
      | OpMinus.t
      | OpMod.t
      | OpMult.t
      | OpNeq.t
      | OpNot.t
      | OpNotIn.t
      | OpOr.t
      | OpOrElse.t
      | OpPlus.t
      | OpPow.t
      | OpRem.t
      | OpXor.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module Identifier : sig
  (**
   * Regular identifier.
   *)

  type t =
    [
      | `Identifier of
          IdentifierType.fields
    ]

  type fields = IdentifierType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module CharLiteral : sig
  (**
   * Character literal.
   *)

  type t =
    [
      | `CharLiteral of
          CharLiteralType.fields
    ]

  type fields = CharLiteralType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_denoted_value :
    [< CharLiteralType.t ]
    -> string
  (**
   * Return the value that this literal denotes.
   *)



end

module BaseId : sig
  (**
   * Base class for identifiers.
   *)

  type t =
    [
      | CharLiteral.t
      | Identifier.t
      | Op.t
      | StringLiteral.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module SingleTokNode : sig
  (**
   * Base class for nodes that are made up of a single token.
   *)

  type t =
    [
      | BaseId.t
      | NullLiteral.t
      | NumLiteral.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_canonical_text :
    [< SingleTokNodeType.t ]
    -> string
  (**
   * Return a canonicalized version of this node's text.
   *)



end

module QualExpr : sig
  (**
   * Qualified expression (``...'(...)``).
   *)

  type t =
    [
      | `QualExpr of
          QualExprType.fields
    ]

  type fields = QualExprType.fields =
    
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `Aggregate
          of AggregateType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_prefix :
    [< QualExprType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]

      
  val f_suffix :
    [< QualExprType.t]
    -> [BaseAggregateType.t | ParenExprType.t]


end

module ExplicitDeref : sig
  (**
   * Explicit dereference expression (``.all``).
   *)

  type t =
    [
      | `ExplicitDeref of
          ExplicitDerefType.fields
    ]

  type fields = ExplicitDerefType.fields =
    
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_prefix :
    [< ExplicitDerefType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]


end

module EndName : sig
  (**
   * Entity name in ``end ...;`` syntactic constructs.
   *)

  type t =
    [
      | `EndName of
          EndNameType.fields
    ]

  type fields = EndNameType.fields =
    
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_basic_decl :
    [< EndNameType.t ]
    -> BasicDeclType.t option
  (**
   * Returns this EndName's basic declaration
   *)


      
  val f_name :
    [< EndNameType.t]
    -> [CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]


end

module DottedName : sig
  (**
   * Name to select a suffix in a prefix.
   *)

  type t =
    [
      | `DottedName of
          DottedNameType.fields
    ]

  type fields = DottedNameType.fields =
    
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `CharLiteral
          of CharLiteralType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_prefix :
    [< DottedNameType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]

      
  val f_suffix :
    [< DottedNameType.t]
    -> [CharLiteralType.t | IdentifierType.t | StringLiteralType.t]


end

module DiscreteSubtypeName : sig
  (**
   * Subtype name for membership test expressions.
   *)

  type t =
    [
      | `DiscreteSubtypeName of
          DiscreteSubtypeNameType.fields
    ]

  type fields = DiscreteSubtypeNameType.fields =
    
  {
         
    f_subtype: DiscreteSubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_subtype :
    [< DiscreteSubtypeNameType.t]
    -> DiscreteSubtypeIndicationType.t


end

module DefiningName : sig
  (**
   * Name that defines an entity.
   *)

  type t =
    [
      | `DefiningName of
          DefiningNameType.fields
    ]

  type fields = DefiningNameType.fields =
    
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_basic_decl :
    [< DefiningNameType.t ]
    -> BasicDeclType.t option
  (**
   * Returns this DefiningName's basic declaration
   *)

      
  val p_find_refs :
    ?imprecise_fallback:
    bool
    -> [< DefiningNameType.t ]
    -> AdaNodeType.t
    -> AdaNodeType.t
    -> RefResult.t list
  (**
   * Find all references to this defining name in the given ``root`` and its
   * children.
   *)

      
  val p_find_all_references :
    ?follow_renamings:
    bool
    -> ?imprecise_fallback:
    bool
    -> [< DefiningNameType.t ]
    -> AnalysisUnit.t list
    -> RefResult.t list
  (**
   * Searches all references to this defining name in the given list of units.
  
   * If ``follow_renamings`` is True, also this also includes references that
   * ultimately refer to this defining name, by unwinding renaming clauses.
   *)

      
  val p_find_all_calls :
    ?follow_renamings:
    bool
    -> ?imprecise_fallback:
    bool
    -> [< DefiningNameType.t ]
    -> AnalysisUnit.t list
    -> RefResult.t list
  (**
   * Return the list of all possible calls to the subprogram which Self is the
   * defining name of.
  
   * This will return the name corresponding to the call, excluding the
   * parameters if there are any. For instance, it will return `A` for the `A
   * (B)` call.
  
   * .. note:: This does not yet support calls done inside generics.
   *)

      
  val p_next_part :
    ?imprecise_fallback:
    bool
    -> [< DefiningNameType.t ]
    -> DefiningNameType.t option
  (**
   * Like ``BasicDecl.next_part_for_decl`` on a defining name
   *)

      
  val p_previous_part :
    ?imprecise_fallback:
    bool
    -> [< DefiningNameType.t ]
    -> DefiningNameType.t option
  (**
   * Like ``BasicDecl.previous_part_for_decl`` on a defining name
   *)

      
  val p_canonical_part :
    ?imprecise_fallback:
    bool
    -> [< DefiningNameType.t ]
    -> DefiningNameType.t option
  (**
   * Like ``BasicDecl.canonical_part`` on a defining name
   *)


      
  val f_name :
    [< DefiningNameType.t]
    -> [CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]


end

module CallExpr : sig
  (**
   * Represent a syntactic call expression.
  
   * At the semantic level, this can be either a subprogram call, an array
   * subcomponent access expression, an array slice or a type conversion.
   *)

  type t =
    [
      | `CallExpr of
          CallExprType.fields
    ]

  type fields = CallExprType.fields =
    
  {
         
    f_name: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_suffix: [
      | `AssocList
          of AssocListType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_is_array_slice :
    [< CallExprType.t ]
    -> bool
  (**
   * Return whether this CallExpr is actually an access to a slice of the array
   * denoted by the prefix of this CallExpr.
   *)


      
  val f_name :
    [< CallExprType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]

      
  val f_suffix :
    [< CallExprType.t]
    -> [AttributeRefType.t | BasicAssocListType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | DiscreteSubtypeIndicationType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]


end

module UpdateAttributeRef : sig
  (**
   * Reference to the ``Update`` attribute.
   *)

  type t =
    [
      | `UpdateAttributeRef of
          UpdateAttributeRefType.fields
    ]

  type fields = UpdateAttributeRefType.fields =
    
  {
         
    f_prefix: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of fields
    ]
    Lazy.t;
         
    f_attribute: IdentifierType.t
    Lazy.t;
         
    f_args: AdaNodeType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_prefix :
    [< UpdateAttributeRefType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]

      
  val f_attribute :
    [< UpdateAttributeRefType.t]
    -> IdentifierType.t

      
  val f_args :
    [< UpdateAttributeRefType.t]
    -> AdaNodeType.t option


end

module AttributeRef : sig
  (**
   * Expression to reference an attribute.
   *)

  type t =
    [
      | `AttributeRef of
          AttributeRefType.fields
      | `UpdateAttributeRef of
          UpdateAttributeRefType.fields
    ]

  type fields = AttributeRefType.fields =
    
  {
         
    f_prefix: [
      | `AttributeRef
          of fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_attribute: IdentifierType.t
    Lazy.t;
         
    f_args: AdaNodeType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_prefix :
    [< AttributeRefType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]

      
  val f_attribute :
    [< AttributeRefType.t]
    -> IdentifierType.t

      
  val f_args :
    [< AttributeRefType.t]
    -> AdaNodeType.t option


end

module Name : sig
  (**
   * Base class for names.
   *)

  type t =
    [
      | AttributeRef.t
      | CallExpr.t
      | DefiningName.t
      | DiscreteSubtypeName.t
      | DottedName.t
      | EndName.t
      | ExplicitDeref.t
      | QualExpr.t
      | SingleTokNode.t
      | TargetName.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_enclosing_defining_name :
    [< NameType.t ]
    -> DefiningNameType.t option
  (**
   * If this name is part of a defining name, return the enclosing defining name
   * node.
   *)

      
  val p_is_defining :
    [< NameType.t ]
    -> bool
  (**
   * Return True if this name is part of a defining name.
   *)

      
  val p_name_is :
    [< NameType.t ]
    -> string
    -> bool
  (**
   * Helper. Check that this name matches ``sym``.
   *)

      
  val p_is_direct_call :
    [< NameType.t ]
    -> bool
  (**
   * Return True iff this name represents a call to a subprogram which is
   * referred by its defining name. (i.e. not through a subprogram access).
   *)

      
  val p_is_access_call :
    [< NameType.t ]
    -> bool
  (**
   * Return True iff this name represents a call to subprogram through an access
   * type.
   *)

      
  val p_is_call :
    [< NameType.t ]
    -> bool
  (**
   * Returns True if this Name corresponds to a call.
   *)

      
  val p_is_dot_call :
    ?imprecise_fallback:
    bool
    -> [< NameType.t ]
    -> bool
  (**
   * Returns True if this Name corresponds to a dot notation call.
   *)

      
  val p_failsafe_referenced_def_name :
    ?imprecise_fallback:
    bool
    -> [< NameType.t ]
    -> RefdDef.t
  (**
   * Failsafe version of ``referenced_defining_name``. Returns a ``RefdDef``,
   * which can be precise, imprecise, or error.
   *)

      
  val p_referenced_defining_name :
    ?imprecise_fallback:
    bool
    -> [< NameType.t ]
    -> DefiningNameType.t option
  (**
   * Like ``referenced_decl``, but will return the defining identifier for the
   * decl, rather than the basic declaration node itself.
   *)

      
  val p_all_env_elements :
    ?seq:
    bool
    -> ?seq_from:
    AdaNodeType.t
    -> [< NameType.t ]
    -> AdaNodeType.t list
  (**
   * Return all elements in self's scope that are lexically named like Self.
   *)

      
  val p_called_subp_spec :
    [< NameType.t ]
    -> BaseFormalParamHolderType.t option
  (**
   * Return the subprogram specification of the subprogram or subprogram access
   * that is being called by this exact Name, if relevant.
   *)

      
  val p_referenced_decl :
    ?imprecise_fallback:
    bool
    -> [< NameType.t ]
    -> BasicDeclType.t option
  (**
   * Return the declaration this node references after name resolution. If
   * imprecise_fallback is True, errors raised during resolution of the xref
   * equation are catched and a fallback mechanism is triggered, which tries to
   * find the referenced declaration in an ad-hoc way.
   *)

      
  val p_failsafe_referenced_decl :
    ?imprecise_fallback:
    bool
    -> [< NameType.t ]
    -> RefdDecl.t
  (**
   * Failsafe version of ``referenced_decl``. Returns a ``RefdDecl``, which can
   * be precise, imprecise, or error.
   *)

      
  val p_referenced_decl_internal :
    ?imprecise_fallback:
    bool
    -> [< NameType.t ]
    -> RefdDecl.t
  (**
   * Return the declaration this node references. Try not to run name res if
   * already resolved. INTERNAL USE ONLY.
   *)

      
  val p_name_designated_type :
    [< NameType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Like SubtypeIndication.designated_type, but on names, since because of
   * Ada's ambiguous grammar, some subtype indications will be parsed as names.
   *)

      
  val p_is_static_subtype :
    ?imprecise_fallback:
    bool
    -> [< NameType.t ]
    -> bool
  (**
   * Returns whether Self denotes a static subtype or not.
   *)

      
  val p_name_matches :
    [< NameType.t ]
    -> NameType.t
    -> bool
  (**
   * Return whether two names match each other.
  
   * This compares the symbol for Identifier and StringLiteral nodes. We
   * consider that there is no match for all other node kinds.
   *)

      
  val p_relative_name :
    [< NameType.t ]
    -> SingleTokNodeType.t option
  (**
   * Returns the relative name of this instance. For example, for a prefix
   * A.B.C, this will return C.
   *)

      
  val p_is_operator_name :
    [< NameType.t ]
    -> bool
  (**
   * Return whether the name that Self designates is an operator.
   *)

      
  val p_is_write_reference :
    ?imprecise_fallback:
    bool
    -> [< NameType.t ]
    -> bool
  (**
   * Whether this name is a write reference.
  
   * For example, `X` is a write reference in the following cases:
  
   * 1. `X := 2;` 2. `X (2) := 2;` 3. `P(F => X)` where F is declared `out` or
   * `in out`. 4. `X'Access`. 5. `X.C := 2`, `R.X := 2` 6. `X.P` where the
   * formal for X is declared `out` or `in out`.
  
   * .. note:: This is an experimental feature. There might be some discrepancy
   *    with the GNAT concept of "write reference".
   *)

      
  val p_is_static_call :
    ?imprecise_fallback:
    bool
    -> [< NameType.t ]
    -> bool
  (**
   * Returns True if this Name corresponds to a static non-dispatching call. In
   * other words, this will return True if and only if the target of the call is
   * known statically.
  
   * .. note:: This is an experimental feature. There might be some discrepancy
   *    with the GNAT concept of "static call".
   *)

      
  val p_as_symbol_array :
    [< NameType.t ]
    -> string list
  (**
   * Turn this name into an array of symbols.
  
   * For instance, a node with name ``A.B.C`` is turned into ``['A', 'B',
   * 'C']``.
   *)



end

module MembershipExpr : sig
  (**
   * Represent a membership test (in/not in operators).
  
   * Note that we don't consider them as binary operators since multiple
   * expressions on the right hand side are allowed.
   *)

  type t =
    [
      | `MembershipExpr of
          MembershipExprType.fields
    ]

  type fields = MembershipExprType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpIn
          of OpInType.fields
      | `OpNotIn
          of OpNotInType.fields
    ]
    Lazy.t;
         
    f_membership_exprs: ExprListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_expr :
    [< MembershipExprType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_op :
    [< MembershipExprType.t]
    -> [OpInType.t | OpNotInType.t]

      
  val f_membership_exprs :
    [< MembershipExprType.t]
    -> ExprListType.t


end

module DeclExpr : sig
  (**
   * Declare expression (Ada 2020).
   *)

  type t =
    [
      | `DeclExpr of
          DeclExprType.fields
    ]

  type fields = DeclExprType.fields =
    
  {
         
    f_decls: BasicDeclListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decls :
    [< DeclExprType.t]
    -> BasicDeclListType.t

      
  val f_expr :
    [< DeclExprType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module ContractCases : sig
  (**
   * List of associations for the ``Contract_Case`` aspect.
   *)

  type t =
    [
      | `ContractCases of
          ContractCasesType.fields
    ]

  type fields = ContractCasesType.fields =
    
  {
         
    f_contract_cases: ContractCaseAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_contract_cases :
    [< ContractCasesType.t]
    -> ContractCaseAssocListType.t


end

module IfExpr : sig
  (**
   * ``if`` expression (RM 4.5.7).
   *)

  type t =
    [
      | `IfExpr of
          IfExprType.fields
    ]

  type fields = IfExprType.fields =
    
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_alternatives: ElsifExprPartListType.t
    Lazy.t;
         
    f_else_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_cond_expr :
    [< IfExprType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_then_expr :
    [< IfExprType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_alternatives :
    [< IfExprType.t]
    -> ElsifExprPartListType.t

      
  val f_else_expr :
    [< IfExprType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option


end

module CaseExpr : sig
  (**
   * ``case`` expression (RM 4.5.7).
   *)

  type t =
    [
      | `CaseExpr of
          CaseExprType.fields
    ]

  type fields = CaseExprType.fields =
    
  {
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_cases: CaseExprAlternativeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_expr :
    [< CaseExprType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_cases :
    [< CaseExprType.t]
    -> CaseExprAlternativeListType.t


end

module CondExpr : sig
  (**
   * Base class for a conditional expressions (RM 4.5.7).
   *)

  type t =
    [
      | CaseExpr.t
      | IfExpr.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_dependent_exprs :
    [< CondExprType.t ]
    -> ExprType.t list
  (**
   * Return the dependent expressions for this conditional expression.
   *)



end

module CaseExprAlternative : sig
  (**
   * Alternative in a ``case`` expression (``when ... => ...``).
   *)

  type t =
    [
      | `CaseExprAlternative of
          CaseExprAlternativeType.fields
    ]

  type fields = CaseExprAlternativeType.fields =
    
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_choices :
    [< CaseExprAlternativeType.t]
    -> AlternativesListType.t

      
  val f_expr :
    [< CaseExprAlternativeType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module BoxExpr : sig
  (**
   * Box expression (``<>``).
   *)

  type t =
    [
      | `BoxExpr of
          BoxExprType.fields
    ]

  type fields = BoxExprType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module RelationOp : sig
  (**
   * Binary operation that compares two value, producing a boolean.
   *)

  type t =
    [
      | `RelationOp of
          RelationOpType.fields
    ]

  type fields = RelationOpType.fields =
    
  {
         
    f_left: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpAnd
          of OpAndType.fields
      | `OpAndThen
          of OpAndThenType.fields
      | `OpConcat
          of OpConcatType.fields
      | `OpDiv
          of OpDivType.fields
      | `OpDoubleDot
          of OpDoubleDotType.fields
      | `OpEq
          of OpEqType.fields
      | `OpGt
          of OpGtType.fields
      | `OpGte
          of OpGteType.fields
      | `OpLt
          of OpLtType.fields
      | `OpLte
          of OpLteType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpMod
          of OpModType.fields
      | `OpMult
          of OpMultType.fields
      | `OpNeq
          of OpNeqType.fields
      | `OpOr
          of OpOrType.fields
      | `OpOrElse
          of OpOrElseType.fields
      | `OpPlus
          of OpPlusType.fields
      | `OpPow
          of OpPowType.fields
      | `OpRem
          of OpRemType.fields
      | `OpXor
          of OpXorType.fields
    ]
    Lazy.t;
         
    f_right: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_left :
    [< RelationOpType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_op :
    [< RelationOpType.t]
    -> [OpAndType.t | OpAndThenType.t | OpConcatType.t | OpDivType.t | OpDoubleDotType.t | OpEqType.t | OpGtType.t | OpGteType.t | OpLtType.t | OpLteType.t | OpMinusType.t | OpModType.t | OpMultType.t | OpNeqType.t | OpOrType.t | OpOrElseType.t | OpPlusType.t | OpPowType.t | OpRemType.t | OpXorType.t]

      
  val f_right :
    [< RelationOpType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module BinOp : sig
  (**
   * Binary expression.
   *)

  type t =
    [
      | `BinOp of
          BinOpType.fields
      | `RelationOp of
          RelationOpType.fields
    ]

  type fields = BinOpType.fields =
    
  {
         
    f_left: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_op: [
      | `OpAnd
          of OpAndType.fields
      | `OpAndThen
          of OpAndThenType.fields
      | `OpConcat
          of OpConcatType.fields
      | `OpDiv
          of OpDivType.fields
      | `OpDoubleDot
          of OpDoubleDotType.fields
      | `OpEq
          of OpEqType.fields
      | `OpGt
          of OpGtType.fields
      | `OpGte
          of OpGteType.fields
      | `OpLt
          of OpLtType.fields
      | `OpLte
          of OpLteType.fields
      | `OpMinus
          of OpMinusType.fields
      | `OpMod
          of OpModType.fields
      | `OpMult
          of OpMultType.fields
      | `OpNeq
          of OpNeqType.fields
      | `OpOr
          of OpOrType.fields
      | `OpOrElse
          of OpOrElseType.fields
      | `OpPlus
          of OpPlusType.fields
      | `OpPow
          of OpPowType.fields
      | `OpRem
          of OpRemType.fields
      | `OpXor
          of OpXorType.fields
    ]
    Lazy.t;
         
    f_right: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_left :
    [< BinOpType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_op :
    [< BinOpType.t]
    -> [OpAndType.t | OpAndThenType.t | OpConcatType.t | OpDivType.t | OpDoubleDotType.t | OpEqType.t | OpGtType.t | OpGteType.t | OpLtType.t | OpLteType.t | OpMinusType.t | OpModType.t | OpMultType.t | OpNeqType.t | OpOrType.t | OpOrElseType.t | OpPlusType.t | OpPowType.t | OpRemType.t | OpXorType.t]

      
  val f_right :
    [< BinOpType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module NullRecordAggregate : sig
  (**
   * Aggregate for ``null record``.
   *)

  type t =
    [
      | `NullRecordAggregate of
          NullRecordAggregateType.fields
    ]

  type fields = NullRecordAggregateType.fields =
    
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ancestor_expr :
    [< NullRecordAggregateType.t]
    -> ExprType.t option

      
  val f_assocs :
    [< NullRecordAggregateType.t]
    -> BasicAssocListType.t option


end

module BracketDeltaAggregate : sig
  (**
   * Bracket delta aggregate (Ada 2020).
   *)

  type t =
    [
      | `BracketDeltaAggregate of
          BracketDeltaAggregateType.fields
    ]

  type fields = BracketDeltaAggregateType.fields =
    
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ancestor_expr :
    [< BracketDeltaAggregateType.t]
    -> ExprType.t option

      
  val f_assocs :
    [< BracketDeltaAggregateType.t]
    -> BasicAssocListType.t option


end

module DeltaAggregate : sig
  

  type t =
    [
      | `DeltaAggregate of
          DeltaAggregateType.fields
      | `BracketDeltaAggregate of
          BracketDeltaAggregateType.fields
    ]

  type fields = DeltaAggregateType.fields =
    
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ancestor_expr :
    [< DeltaAggregateType.t]
    -> ExprType.t option

      
  val f_assocs :
    [< DeltaAggregateType.t]
    -> BasicAssocListType.t option


end

module BracketAggregate : sig
  (**
   * Bracket array or container aggregate (Ada 2020).
   *)

  type t =
    [
      | `BracketAggregate of
          BracketAggregateType.fields
    ]

  type fields = BracketAggregateType.fields =
    
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ancestor_expr :
    [< BracketAggregateType.t]
    -> ExprType.t option

      
  val f_assocs :
    [< BracketAggregateType.t]
    -> BasicAssocListType.t option


end

module Aggregate : sig
  (**
   * Aggregate that is not a ``null record`` aggregate.
   *)

  type t =
    [
      | `Aggregate of
          AggregateType.fields
      | `BracketAggregate of
          BracketAggregateType.fields
    ]

  type fields = AggregateType.fields =
    
  {
         
    f_ancestor_expr: ExprType.t
    option
    Lazy.t;
         
    f_assocs: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ancestor_expr :
    [< AggregateType.t]
    -> ExprType.t option

      
  val f_assocs :
    [< AggregateType.t]
    -> BasicAssocListType.t option


end

module BaseAggregate : sig
  (**
   * Base class for aggregates.
   *)

  type t =
    [
      | Aggregate.t
      | DeltaAggregate.t
      | NullRecordAggregate.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ancestor_expr :
    [< BaseAggregateType.t]
    -> ExprType.t option

      
  val f_assocs :
    [< BaseAggregateType.t]
    -> BasicAssocListType.t option


end

module Allocator : sig
  (**
   * Allocator expression (``new ...``).
   *)

  type t =
    [
      | `Allocator of
          AllocatorType.fields
    ]

  type fields = AllocatorType.fields =
    
  {
         
    f_subpool: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_type_or_expr: [
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `QualExpr
          of QualExprType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_get_allocated_type :
    [< AllocatorType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Return the allocated type for this allocator.
   *)


      
  val f_subpool :
    [< AllocatorType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t] option

      
  val f_type_or_expr :
    [< AllocatorType.t]
    -> [QualExprType.t | SubtypeIndicationType.t]


end

module AbstractStateDeclExpr : sig
  (**
   * Directly corresponds to the right-hand side of the Abstract_State aspect.
   * Only exists because the RHS of an AspectAssoc must be an expression: the
   * actual logic is in AbstractStateDecl.
   *)

  type t =
    [
      | `AbstractStateDeclExpr of
          AbstractStateDeclExprType.fields
    ]

  type fields = AbstractStateDeclExprType.fields =
    
  {
         
    f_state_decl: [
      | `AbstractStateDecl
          of AbstractStateDeclType.fields
      | `MultiAbstractStateDecl
          of MultiAbstractStateDeclType.fields
      | `ParenAbstractStateDecl
          of ParenAbstractStateDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_state_decl :
    [< AbstractStateDeclExprType.t]
    -> [AbstractStateDeclType.t | MultiAbstractStateDeclType.t | ParenAbstractStateDeclType.t]


end

module Expr : sig
  (**
   * Base class for expressions.
   *)

  type t =
    [
      | AbstractStateDeclExpr.t
      | Allocator.t
      | BaseAggregate.t
      | BinOp.t
      | BoxExpr.t
      | CaseExprAlternative.t
      | CondExpr.t
      | ContractCases.t
      | DeclExpr.t
      | MembershipExpr.t
      | Name.t
      | ParenExpr.t
      | QuantifiedExpr.t
      | RaiseExpr.t
      | UnOp.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_expression_type :
    [< ExprType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Return the declaration corresponding to the type of this expression after
   * name resolution.
   *)

      
  val p_is_dynamically_tagged :
    ?imprecise_fallback:
    bool
    -> [< ExprType.t ]
    -> bool
  (**
   * Returns whether this expression is dynamically tagged (See RM 3.9.2).
   *)

      
  val p_is_dispatching_call :
    ?imprecise_fallback:
    bool
    -> [< ExprType.t ]
    -> bool
  (**
   * Returns True if this Name corresponds to a dispatching call, including:
  
   * - Calls done through subprogram access types.
  
   * - Calls to dispatching subprograms, in the object-oriented sense.
  
   * .. note:: This is an experimental feature. There might be some discrepancy
   *    with the GNAT concept of "dispatching call".
  
   * .. note:: This should only be called on a Name and UnOp or a BinOp.
   *)

      
  val p_is_static_expr :
    ?imprecise_fallback:
    bool
    -> [< ExprType.t ]
    -> bool
  (**
   * Return whether this expression is static according to the ARM definition of
   * static. See RM 4.9.
   *)

      
  val p_first_corresponding_decl :
    [< ExprType.t ]
    -> BasicDeclType.t option
  (**
   * Return the first decl that is lexically named like self in self's scope.
   *)

      
  val p_eval_as_int :
    [< ExprType.t ]
    -> BigInteger.t
  (**
   * Statically evaluates self, and returns the value of the evaluation as an
   * integer.
  
   * .. note:: In order for a call to this not to raise, the expression needs to
   *    be a static expression, as specified in the ARM section 4.9. You can
   *    verify whether an expression is static with the ``is_static_expr``
   *    property.
  
   * .. ATTENTION:: This is an experimental feature, so even if it is exposed to
   *    allow experiments, it is totally unsupported and the API and behavior
   *    are very likely to change in the future.
   *)

      
  val p_eval_as_int_in_env :
    [< ExprType.t ]
    -> Substitution.t list
    -> BigInteger.t
  (**
   * Statically evaluates self, and returns the value of the evaluation as an
   * integer. The given environment is used to substitute references to
   * declarations by actual values.
  
   * .. note:: In order for a call to this not to raise, the expression needs to
   *    be a static expression, as specified in the ARM section 4.9. You can
   *    verify whether an expression is static with the ``is_static_expr``
   *    property.
  
   * .. ATTENTION:: This is an experimental feature, so even if it is exposed to
   *    allow experiments, it is totally unsupported and the API and behavior
   *    are very likely to change in the future.
   *)

      
  val p_eval_as_string :
    [< ExprType.t ]
    -> string
  (**
   * Statically evaluates self, and returns the value of the evaluation as a
   * string.
  
   * .. note:: In order for a call to this not to raise, the expression needs to
   *    be a static expression, as specified in the ARM section 4.9. You can
   *    verify whether an expression is static with the ``is_static_expr``
   *    property.
  
   * .. ATTENTION:: This is an experimental feature, so even if it is exposed to
   *    allow experiments, it is totally unsupported and the API and behavior
   *    are very likely to change in the future.
   *)

      
  val p_eval_as_string_in_env :
    [< ExprType.t ]
    -> Substitution.t list
    -> string
  (**
   * Statically evaluates self, and returns the value of the evaluation as a
   * string. The given environment is used to substitute references to
   * declarations by actual values.
  
   * .. note:: In order for a call to this not to raise, the expression needs to
   *    be a static expression, as specified in the ARM section 4.9. You can
   *    verify whether an expression is static with the ``is_static_expr``
   *    property.
  
   * .. ATTENTION:: This is an experimental feature, so even if it is exposed to
   *    allow experiments, it is totally unsupported and the API and behavior
   *    are very likely to change in the future.
   *)

      
  val p_matching_nodes :
    [< ExprType.t ]
    -> AdaNodeType.t list
  (**
   * Return the list of AST nodes that can be a match for this expression before
   * overloading analysis.
   *)



end

module ElsifStmtPart : sig
  (**
   * ``elsif`` part in an ``if`` statement block.
   *)

  type t =
    [
      | `ElsifStmtPart of
          ElsifStmtPartType.fields
    ]

  type fields = ElsifStmtPartType.fields =
    
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_cond_expr :
    [< ElsifStmtPartType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_stmts :
    [< ElsifStmtPartType.t]
    -> StmtListType.t


end

module ElsifExprPart : sig
  (**
   * ``elsif`` block, part of an ``if`` expression.
   *)

  type t =
    [
      | `ElsifExprPart of
          ElsifExprPartType.fields
    ]

  type fields = ElsifExprPartType.fields =
    
  {
         
    f_cond_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_then_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_cond_expr :
    [< ElsifExprPartType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_then_expr :
    [< ElsifExprPartType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module PublicPart : sig
  (**
   * List of declarations in a public part.
   *)

  type t =
    [
      | `PublicPart of
          PublicPartType.fields
    ]

  type fields = PublicPartType.fields =
    
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decls :
    [< PublicPartType.t]
    -> AdaNodeListType.t


end

module PrivatePart : sig
  (**
   * List of declarations in a private part.
   *)

  type t =
    [
      | `PrivatePart of
          PrivatePartType.fields
    ]

  type fields = PrivatePartType.fields =
    
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decls :
    [< PrivatePartType.t]
    -> AdaNodeListType.t


end

module DeclarativePart : sig
  (**
   * List of declarations.
   *)

  type t =
    [
      | `DeclarativePart of
          DeclarativePartType.fields
      | `PrivatePart of
          PrivatePartType.fields
      | `PublicPart of
          PublicPartType.fields
    ]

  type fields = DeclarativePartType.fields =
    
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decls :
    [< DeclarativePartType.t]
    -> AdaNodeListType.t


end

module RangeConstraint : sig
  (**
   * Range-based type constraint.
   *)

  type t =
    [
      | `RangeConstraint of
          RangeConstraintType.fields
    ]

  type fields = RangeConstraintType.fields =
    
  {
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_range :
    [< RangeConstraintType.t]
    -> RangeSpecType.t


end

module IndexConstraint : sig
  (**
   * List of type constraints.
   *)

  type t =
    [
      | `IndexConstraint of
          IndexConstraintType.fields
    ]

  type fields = IndexConstraintType.fields =
    
  {
         
    f_constraints: ConstraintListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_constraints :
    [< IndexConstraintType.t]
    -> ConstraintListType.t


end

module DiscriminantConstraint : sig
  (**
   * List of constraints that relate to type discriminants.
   *)

  type t =
    [
      | `DiscriminantConstraint of
          DiscriminantConstraintType.fields
    ]

  type fields = DiscriminantConstraintType.fields =
    
  {
         
    f_constraints: BasicAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_constraints :
    [< DiscriminantConstraintType.t]
    -> BasicAssocListType.t


end

module DigitsConstraint : sig
  (**
   * Digits and range type constraint.
   *)

  type t =
    [
      | `DigitsConstraint of
          DigitsConstraintType.fields
    ]

  type fields = DigitsConstraintType.fields =
    
  {
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_digits :
    [< DigitsConstraintType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_range :
    [< DigitsConstraintType.t]
    -> RangeSpecType.t option


end

module DeltaConstraint : sig
  (**
   * Delta and range type constraint.
   *)

  type t =
    [
      | `DeltaConstraint of
          DeltaConstraintType.fields
    ]

  type fields = DeltaConstraintType.fields =
    
  {
         
    f_digits: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_digits :
    [< DeltaConstraintType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_range :
    [< DeltaConstraintType.t]
    -> RangeSpecType.t option


end

module Constraint : sig
  (**
   * Base class for type constraints.
   *)

  type t =
    [
      | DeltaConstraint.t
      | DigitsConstraint.t
      | DiscriminantConstraint.t
      | IndexConstraint.t
      | RangeConstraint.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ConstantPresent : sig
  

  type t =
    [
      | `ConstantPresent of
          ConstantPresentType.fields
    ]

  type fields = ConstantPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ConstantAbsent : sig
  

  type t =
    [
      | `ConstantAbsent of
          ConstantAbsentType.fields
    ]

  type fields = ConstantAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ConstantNode : sig
  (**
   * Qualifier for the ``constant`` keyword.
   *)

  type t =
    [
      | ConstantAbsent.t
      | ConstantPresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< ConstantNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of ConstantPresent
   *)



end

module ComponentDef : sig
  (**
   * Definition for a component.
   *)

  type t =
    [
      | `ComponentDef of
          ComponentDefType.fields
    ]

  type fields = ComponentDefType.fields =
    
  {
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_has_aliased :
    [< ComponentDefType.t]
    -> AliasedNodeType.t

      
  val f_has_constant :
    [< ComponentDefType.t]
    -> ConstantNodeType.t

      
  val f_type_expr :
    [< ComponentDefType.t]
    -> [AnonymousTypeType.t | SubtypeIndicationType.t]


end

module ComponentClause : sig
  (**
   * Representation clause for a single component.
   *)

  type t =
    [
      | `ComponentClause of
          ComponentClauseType.fields
    ]

  type fields = ComponentClauseType.fields =
    
  {
         
    f_id: IdentifierType.t
    Lazy.t;
         
    f_position: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_range: RangeSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_id :
    [< ComponentClauseType.t]
    -> IdentifierType.t

      
  val f_position :
    [< ComponentClauseType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_range :
    [< ComponentClauseType.t]
    -> RangeSpecType.t


end

module CompilationUnit : sig
  (**
   * Root node for all Ada analysis units.
   *)

  type t =
    [
      | `CompilationUnit of
          CompilationUnitType.fields
    ]

  type fields = CompilationUnitType.fields =
    
  {
         
    f_prelude: AdaNodeListType.t
    Lazy.t;
         
    f_body: [
      | `LibraryItem
          of LibraryItemType.fields
      | `Subunit
          of SubunitType.fields
    ]
    Lazy.t;
         
    f_pragmas: PragmaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_syntactic_fully_qualified_name :
    [< CompilationUnitType.t ]
    -> string list
  (**
   * Return the syntactic fully qualified name of this compilation unit.
   *)

      
  val p_unit_kind :
    [< CompilationUnitType.t ]
    -> AnalysisUnitKind.t
  (**
   * Return the kind corresponding to this analysis unit.
   *)

      
  val p_withed_units :
    [< CompilationUnitType.t ]
    -> CompilationUnitType.t list
  (**
   * Look for all "with" clauses at the top of this compilation unit and return
   * all the compilation units designated by them.
   *)

      
  val p_imported_units :
    [< CompilationUnitType.t ]
    -> CompilationUnitType.t list
  (**
   * Return all the compilation units that are directly imported by this one.
   * This includes "with"ed units as well as the direct parent unit.
   *)

      
  val p_unit_dependencies :
    [< CompilationUnitType.t ]
    -> CompilationUnitType.t list
  (**
   * Return the list of all the compilation units that are (direct and indirect)
   * dependencies of this one.
   *)

      
  val p_decl :
    [< CompilationUnitType.t ]
    -> BasicDeclType.t option
  (**
   * Get the root basic decl defined in this compilation unit.
   *)

      
  val p_is_preelaborable :
    ?imprecise_fallback:
    bool
    -> [< CompilationUnitType.t ]
    -> bool
  (**
   * Whether this compilation unit is preelaborable or not.
   *)


      
  val f_prelude :
    [< CompilationUnitType.t]
    -> AdaNodeListType.t

      
  val f_body :
    [< CompilationUnitType.t]
    -> [LibraryItemType.t | SubunitType.t]

      
  val f_pragmas :
    [< CompilationUnitType.t]
    -> PragmaNodeListType.t


end

module CaseStmtAlternative : sig
  (**
   * Alternative in a ``case`` statement (``when ... => ...``).
   *)

  type t =
    [
      | `CaseStmtAlternative of
          CaseStmtAlternativeType.fields
    ]

  type fields = CaseStmtAlternativeType.fields =
    
  {
         
    f_choices: AlternativesListType.t
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_choices :
    [< CaseStmtAlternativeType.t]
    -> AlternativesListType.t

      
  val f_stmts :
    [< CaseStmtAlternativeType.t]
    -> StmtListType.t


end

module SingleTaskDecl : sig
  (**
   * Declaration for a single task.
   *)

  type t =
    [
      | `SingleTaskDecl of
          SingleTaskDeclType.fields
    ]

  type fields = SingleTaskDeclType.fields =
    
  {
         
    f_task_type: SingleTaskTypeDeclType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_task_type :
    [< SingleTaskDeclType.t]
    -> SingleTaskTypeDeclType.t


end

module SingleProtectedDecl : sig
  (**
   * Declaration for a single protected object.
   *)

  type t =
    [
      | `SingleProtectedDecl of
          SingleProtectedDeclType.fields
    ]

  type fields = SingleProtectedDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_definition: ProtectedDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< SingleProtectedDeclType.t]
    -> DefiningNameType.t

      
  val f_interfaces :
    [< SingleProtectedDeclType.t]
    -> ParentListType.t option

      
  val f_definition :
    [< SingleProtectedDeclType.t]
    -> ProtectedDefType.t


end

module PackageRenamingDecl : sig
  (**
   * Declaration for a package renaming.
   *)

  type t =
    [
      | `PackageRenamingDecl of
          PackageRenamingDeclType.fields
    ]

  type fields = PackageRenamingDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_renamed_package :
    [< PackageRenamingDeclType.t ]
    -> BasicDeclType.t option
  (**
   * Return the declaration of the package that is renamed by Self.
   *)

      
  val p_final_renamed_package :
    [< PackageRenamingDeclType.t ]
    -> BasicDeclType.t option
  (**
   * Return the declaration of the package that is ultimately renamed by Self,
   * skipping through all intermediate package renamings.
   *)


      
  val f_name :
    [< PackageRenamingDeclType.t]
    -> DefiningNameType.t

      
  val f_renames :
    [< PackageRenamingDeclType.t]
    -> RenamingClauseType.t


end

module ExtendedReturnStmtObjectDecl : sig
  (**
   * Object declaration that is part of an extended return statement.
   *)

  type t =
    [
      | `ExtendedReturnStmtObjectDecl of
          ExtendedReturnStmtObjectDeclType.fields
    ]

  type fields = ExtendedReturnStmtObjectDeclType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_renaming_clause: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ids :
    [< ExtendedReturnStmtObjectDeclType.t]
    -> DefiningNameListType.t

      
  val f_has_aliased :
    [< ExtendedReturnStmtObjectDeclType.t]
    -> AliasedNodeType.t

      
  val f_has_constant :
    [< ExtendedReturnStmtObjectDeclType.t]
    -> ConstantNodeType.t

      
  val f_mode :
    [< ExtendedReturnStmtObjectDeclType.t]
    -> ModeType.t option

      
  val f_type_expr :
    [< ExtendedReturnStmtObjectDeclType.t]
    -> [AnonymousTypeType.t | SubtypeIndicationType.t]

      
  val f_default_expr :
    [< ExtendedReturnStmtObjectDeclType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option

      
  val f_renaming_clause :
    [< ExtendedReturnStmtObjectDeclType.t]
    -> RenamingClauseType.t option


end

module ObjectDecl : sig
  (**
   * Base class for Ada object declarations (RM 3.3.1). Ada object declarations
   * are variables/constants declarations that can be declared in any
   * declarative scope.
   *)

  type t =
    [
      | `ObjectDecl of
          ObjectDeclType.fields
      | `ExtendedReturnStmtObjectDecl of
          ExtendedReturnStmtObjectDeclType.fields
    ]

  type fields = ObjectDeclType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_has_constant: ConstantNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_renaming_clause: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_public_part_decl :
    [< ObjectDeclType.t ]
    -> BasicDeclType.t option
  (**
   * If this object decl is the constant completion of an object decl in the
   * public part, return the object decl from the public part.
   *)

      
  val p_private_part_decl :
    [< ObjectDeclType.t ]
    -> BasicDeclType.t option
  (**
   * If this object decl is the incomplete declaration of a constant in a public
   * part, return its completion in the private part.
   *)


      
  val f_ids :
    [< ObjectDeclType.t]
    -> DefiningNameListType.t

      
  val f_has_aliased :
    [< ObjectDeclType.t]
    -> AliasedNodeType.t

      
  val f_has_constant :
    [< ObjectDeclType.t]
    -> ConstantNodeType.t

      
  val f_mode :
    [< ObjectDeclType.t]
    -> ModeType.t option

      
  val f_type_expr :
    [< ObjectDeclType.t]
    -> [AnonymousTypeType.t | SubtypeIndicationType.t]

      
  val f_default_expr :
    [< ObjectDeclType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option

      
  val f_renaming_clause :
    [< ObjectDeclType.t]
    -> RenamingClauseType.t option


end

module NumberDecl : sig
  (**
   * Declaration for a static constant number.
   *)

  type t =
    [
      | `NumberDecl of
          NumberDeclType.fields
    ]

  type fields = NumberDeclType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ids :
    [< NumberDeclType.t]
    -> DefiningNameListType.t

      
  val f_expr :
    [< NumberDeclType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module NamedStmtDecl : sig
  (**
   * BasicDecl that is always the declaration inside a named statement.
   *)

  type t =
    [
      | `NamedStmtDecl of
          NamedStmtDeclType.fields
    ]

  type fields = NamedStmtDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< NamedStmtDeclType.t]
    -> DefiningNameType.t


end

module LabelDecl : sig
  (**
   * Declaration for a code label.
   *)

  type t =
    [
      | `LabelDecl of
          LabelDeclType.fields
    ]

  type fields = LabelDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< LabelDeclType.t]
    -> DefiningNameType.t


end

module GenericSubpRenamingDecl : sig
  (**
   * Declaration for a generic subprogram renaming.
   *)

  type t =
    [
      | `GenericSubpRenamingDecl of
          GenericSubpRenamingDeclType.fields
    ]

  type fields = GenericSubpRenamingDeclType.fields =
    
  {
         
    f_kind: SubpKindType.t
    Lazy.t;
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_kind :
    [< GenericSubpRenamingDeclType.t]
    -> SubpKindType.t

      
  val f_name :
    [< GenericSubpRenamingDeclType.t]
    -> DefiningNameType.t

      
  val f_renames :
    [< GenericSubpRenamingDeclType.t]
    -> [CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]


end

module GenericPackageRenamingDecl : sig
  (**
   * Declaration for a generic package renaming.
   *)

  type t =
    [
      | `GenericPackageRenamingDecl of
          GenericPackageRenamingDeclType.fields
    ]

  type fields = GenericPackageRenamingDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_renames: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< GenericPackageRenamingDeclType.t]
    -> DefiningNameType.t

      
  val f_renames :
    [< GenericPackageRenamingDeclType.t]
    -> [CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]


end

module GenericRenamingDecl : sig
  (**
   * Base node for all generic renaming declarations.
   *)

  type t =
    [
      | GenericPackageRenamingDecl.t
      | GenericSubpRenamingDecl.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module GenericSubpInstantiation : sig
  (**
   * Instantiations of a generic subprogram.
   *)

  type t =
    [
      | `GenericSubpInstantiation of
          GenericSubpInstantiationType.fields
    ]

  type fields = GenericSubpInstantiationType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_kind: SubpKindType.t
    Lazy.t;
         
    f_subp_name: DefiningNameType.t
    Lazy.t;
         
    f_generic_subp_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_params: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_designated_subp :
    [< GenericSubpInstantiationType.t ]
    -> AdaNodeType.t option
  (**
   * Return the subprogram decl designated by this instantiation.
   *)


      
  val f_overriding :
    [< GenericSubpInstantiationType.t]
    -> OverridingNodeType.t

      
  val f_kind :
    [< GenericSubpInstantiationType.t]
    -> SubpKindType.t

      
  val f_subp_name :
    [< GenericSubpInstantiationType.t]
    -> DefiningNameType.t

      
  val f_generic_subp_name :
    [< GenericSubpInstantiationType.t]
    -> [CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]

      
  val f_params :
    [< GenericSubpInstantiationType.t]
    -> BasicAssocListType.t option


end

module GenericPackageInstantiation : sig
  (**
   * Instantiations of a generic package.
   *)

  type t =
    [
      | `GenericPackageInstantiation of
          GenericPackageInstantiationType.fields
    ]

  type fields = GenericPackageInstantiationType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_generic_pkg_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_params: BasicAssocListType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< GenericPackageInstantiationType.t]
    -> DefiningNameType.t

      
  val f_generic_pkg_name :
    [< GenericPackageInstantiationType.t]
    -> [CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]

      
  val f_params :
    [< GenericPackageInstantiationType.t]
    -> BasicAssocListType.t option


end

module GenericInstantiation : sig
  (**
   * Instantiations of generics.
   *)

  type t =
    [
      | GenericPackageInstantiation.t
      | GenericSubpInstantiation.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_designated_generic_decl :
    [< GenericInstantiationType.t ]
    -> BasicDeclType.t option
  (**
   * Return the generic decl entity designated by this instantiation, containing
   * the generic context. This is equivalent to the expanded generic unit in
   * GNAT.
   *)



end

module GenericSubpDecl : sig
  (**
   * Generic subprogram declaration.
   *)

  type t =
    [
      | `GenericSubpDecl of
          GenericSubpDeclType.fields
    ]

  type fields = GenericSubpDeclType.fields =
    
  {
         
    f_formal_part: GenericFormalPartType.t
    Lazy.t;
         
    f_subp_decl: GenericSubpInternalType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_body_part :
    ?imprecise_fallback:
    bool
    -> [< GenericSubpDeclType.t ]
    -> BaseSubpBodyType.t option
  (**
   * Return the BaseSubpBody corresponding to this node.
   *)


      
  val f_formal_part :
    [< GenericSubpDeclType.t]
    -> GenericFormalPartType.t

      
  val f_subp_decl :
    [< GenericSubpDeclType.t]
    -> GenericSubpInternalType.t


end

module GenericPackageDecl : sig
  (**
   * Generic package declaration.
   *)

  type t =
    [
      | `GenericPackageDecl of
          GenericPackageDeclType.fields
    ]

  type fields = GenericPackageDeclType.fields =
    
  {
         
    f_formal_part: GenericFormalPartType.t
    Lazy.t;
         
    f_package_decl: GenericPackageInternalType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_body_part :
    [< GenericPackageDeclType.t ]
    -> PackageBodyType.t option
  (**
   * Return the PackageBody corresponding to this node, or null if there is
   * none.
   *)


      
  val f_formal_part :
    [< GenericPackageDeclType.t]
    -> GenericFormalPartType.t

      
  val f_package_decl :
    [< GenericPackageDeclType.t]
    -> GenericPackageInternalType.t


end

module GenericDecl : sig
  (**
   * Base class for generic declarations.
   *)

  type t =
    [
      | GenericPackageDecl.t
      | GenericSubpDecl.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_formal_part :
    [< GenericDeclType.t]
    -> GenericFormalPartType.t


end

module ForLoopVarDecl : sig
  (**
   * Declaration for the controlling variable in a ``for`` loop.
   *)

  type t =
    [
      | `ForLoopVarDecl of
          ForLoopVarDeclType.fields
    ]

  type fields = ForLoopVarDeclType.fields =
    
  {
         
    f_id: DefiningNameType.t
    Lazy.t;
         
    f_id_type: SubtypeIndicationType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_id :
    [< ForLoopVarDeclType.t]
    -> DefiningNameType.t

      
  val f_id_type :
    [< ForLoopVarDeclType.t]
    -> SubtypeIndicationType.t option


end

module ExceptionHandler : sig
  (**
   * Exception handler.
   *)

  type t =
    [
      | `ExceptionHandler of
          ExceptionHandlerType.fields
    ]

  type fields = ExceptionHandlerType.fields =
    
  {
         
    f_exception_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_handled_exceptions: AlternativesListType.t
    Lazy.t;
         
    f_stmts: StmtListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_exception_name :
    [< ExceptionHandlerType.t]
    -> DefiningNameType.t option

      
  val f_handled_exceptions :
    [< ExceptionHandlerType.t]
    -> AlternativesListType.t

      
  val f_stmts :
    [< ExceptionHandlerType.t]
    -> StmtListType.t


end

module ExceptionDecl : sig
  (**
   * Exception declarations.
   *)

  type t =
    [
      | `ExceptionDecl of
          ExceptionDeclType.fields
    ]

  type fields = ExceptionDeclType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ids :
    [< ExceptionDeclType.t]
    -> DefiningNameListType.t

      
  val f_renames :
    [< ExceptionDeclType.t]
    -> RenamingClauseType.t option


end

module ErrorDecl : sig
  (**
   * Placeholder node for syntax errors in lists of declarations.
   *)

  type t =
    [
      | `ErrorDecl of
          ErrorDeclType.fields
    ]

  type fields = ErrorDeclType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module EntryIndexSpec : sig
  (**
   * Index specification for an entry body.
   *)

  type t =
    [
      | `EntryIndexSpec of
          EntryIndexSpecType.fields
    ]

  type fields = EntryIndexSpecType.fields =
    
  {
         
    f_id: DefiningNameType.t
    Lazy.t;
         
    f_subtype: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_id :
    [< EntryIndexSpecType.t]
    -> DefiningNameType.t

      
  val f_subtype :
    [< EntryIndexSpecType.t]
    -> [AttributeRefType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | SubtypeIndicationType.t | TargetNameType.t]


end

module TaskBody : sig
  (**
   * Task body.
   *)

  type t =
    [
      | `TaskBody of
          TaskBodyType.fields
    ]

  type fields = TaskBodyType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< TaskBodyType.t]
    -> DefiningNameType.t

      
  val f_decls :
    [< TaskBodyType.t]
    -> DeclarativePartType.t

      
  val f_stmts :
    [< TaskBodyType.t]
    -> HandledStmtsType.t

      
  val f_end_name :
    [< TaskBodyType.t]
    -> EndNameType.t option


end

module ProtectedBody : sig
  (**
   * Protected object body.
   *)

  type t =
    [
      | `ProtectedBody of
          ProtectedBodyType.fields
    ]

  type fields = ProtectedBodyType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< ProtectedBodyType.t]
    -> DefiningNameType.t

      
  val f_decls :
    [< ProtectedBodyType.t]
    -> DeclarativePartType.t

      
  val f_end_name :
    [< ProtectedBodyType.t]
    -> EndNameType.t option


end

module PackageBody : sig
  (**
   * Package body.
   *)

  type t =
    [
      | `PackageBody of
          PackageBodyType.fields
    ]

  type fields = PackageBodyType.fields =
    
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_package_name :
    [< PackageBodyType.t]
    -> DefiningNameType.t

      
  val f_decls :
    [< PackageBodyType.t]
    -> DeclarativePartType.t

      
  val f_stmts :
    [< PackageBodyType.t]
    -> HandledStmtsType.t option

      
  val f_end_name :
    [< PackageBodyType.t]
    -> EndNameType.t option


end

module EntryBody : sig
  (**
   * Entry body.
   *)

  type t =
    [
      | `EntryBody of
          EntryBodyType.fields
    ]

  type fields = EntryBodyType.fields =
    
  {
         
    f_entry_name: DefiningNameType.t
    Lazy.t;
         
    f_index_spec: EntryIndexSpecType.t
    option
    Lazy.t;
         
    f_params: EntryCompletionFormalParamsType.t
    Lazy.t;
         
    f_barrier: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_entry_name :
    [< EntryBodyType.t]
    -> DefiningNameType.t

      
  val f_index_spec :
    [< EntryBodyType.t]
    -> EntryIndexSpecType.t option

      
  val f_params :
    [< EntryBodyType.t]
    -> EntryCompletionFormalParamsType.t

      
  val f_barrier :
    [< EntryBodyType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_decls :
    [< EntryBodyType.t]
    -> DeclarativePartType.t

      
  val f_stmts :
    [< EntryBodyType.t]
    -> HandledStmtsType.t

      
  val f_end_name :
    [< EntryBodyType.t]
    -> EndNameType.t option


end

module TaskBodyStub : sig
  (**
   * Stub for a task body (``is separate``).
   *)

  type t =
    [
      | `TaskBodyStub of
          TaskBodyStubType.fields
    ]

  type fields = TaskBodyStubType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< TaskBodyStubType.t]
    -> DefiningNameType.t


end

module SubpBodyStub : sig
  (**
   * Stub for a subprogram body (``is separate``).
   *)

  type t =
    [
      | `SubpBodyStub of
          SubpBodyStubType.fields
    ]

  type fields = SubpBodyStubType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_overriding :
    [< SubpBodyStubType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< SubpBodyStubType.t]
    -> SubpSpecType.t


end

module ProtectedBodyStub : sig
  (**
   * Stub for a protected object body (``is separate``).
   *)

  type t =
    [
      | `ProtectedBodyStub of
          ProtectedBodyStubType.fields
    ]

  type fields = ProtectedBodyStubType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< ProtectedBodyStubType.t]
    -> DefiningNameType.t


end

module PackageBodyStub : sig
  (**
   * Stub for a package body (``is separate``).
   *)

  type t =
    [
      | `PackageBodyStub of
          PackageBodyStubType.fields
    ]

  type fields = PackageBodyStubType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< PackageBodyStubType.t]
    -> DefiningNameType.t


end

module BodyStub : sig
  (**
   * Base class for a body stub (RM 10.1.3). A body stub is meant to be
   * completed by .
   *)

  type t =
    [
      | PackageBodyStub.t
      | ProtectedBodyStub.t
      | SubpBodyStub.t
      | TaskBodyStub.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_syntactic_fully_qualified_name :
    [< BodyStubType.t ]
    -> string list
  (**
   * Return the syntactic fully qualified name to refer to this body.
  
   * Note that this can raise a Property_Error when the stub is in an illegal
   * place (too nested, in a declare block, etc.).
   *)



end

module SubpRenamingDecl : sig
  (**
   * Declaration for a subprogram renaming.
   *)

  type t =
    [
      | `SubpRenamingDecl of
          SubpRenamingDeclType.fields
    ]

  type fields = SubpRenamingDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_renames: RenamingClauseType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_overriding :
    [< SubpRenamingDeclType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< SubpRenamingDeclType.t]
    -> SubpSpecType.t

      
  val f_renames :
    [< SubpRenamingDeclType.t]
    -> RenamingClauseType.t


end

module SubpBody : sig
  (**
   * Subprogram body.
   *)

  type t =
    [
      | `SubpBody of
          SubpBodyType.fields
    ]

  type fields = SubpBodyType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_decls: DeclarativePartType.t
    Lazy.t;
         
    f_stmts: HandledStmtsType.t
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_overriding :
    [< SubpBodyType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< SubpBodyType.t]
    -> SubpSpecType.t

      
  val f_decls :
    [< SubpBodyType.t]
    -> DeclarativePartType.t

      
  val f_stmts :
    [< SubpBodyType.t]
    -> HandledStmtsType.t

      
  val f_end_name :
    [< SubpBodyType.t]
    -> EndNameType.t option


end

module NullSubpDecl : sig
  (**
   * Declaration for a null subprogram.
   *)

  type t =
    [
      | `NullSubpDecl of
          NullSubpDeclType.fields
    ]

  type fields = NullSubpDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_overriding :
    [< NullSubpDeclType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< NullSubpDeclType.t]
    -> SubpSpecType.t


end

module ExprFunction : sig
  (**
   * Expression function.
   *)

  type t =
    [
      | `ExprFunction of
          ExprFunctionType.fields
    ]

  type fields = ExprFunctionType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_overriding :
    [< ExprFunctionType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< ExprFunctionType.t]
    -> SubpSpecType.t

      
  val f_expr :
    [< ExprFunctionType.t]
    -> [BaseAggregateType.t | ParenExprType.t]


end

module BaseSubpBody : sig
  (**
   * Base class for subprogram bodies.
   *)

  type t =
    [
      | ExprFunction.t
      | NullSubpDecl.t
      | SubpBody.t
      | SubpRenamingDecl.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_overriding :
    [< BaseSubpBodyType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< BaseSubpBodyType.t]
    -> SubpSpecType.t


end

module BodyNode : sig
  (**
   * Base class for an Ada body (RM 3.11). A body is the completion of a
   * declaration.
   *)

  type t =
    [
      | BaseSubpBody.t
      | BodyStub.t
      | EntryBody.t
      | PackageBody.t
      | ProtectedBody.t
      | TaskBody.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_previous_part :
    ?imprecise_fallback:
    bool
    -> [< BodyNodeType.t ]
    -> BasicDeclType.t option
  (**
   * Return the previous part for this body. Might be a declaration or a body
   * stub.
   *)

      
  val p_decl_part :
    ?imprecise_fallback:
    bool
    -> [< BodyNodeType.t ]
    -> BasicDeclType.t option
  (**
   * Return the decl corresponding to this node if applicable.
   *)

      
  val p_subunit_root :
    [< BodyNodeType.t ]
    -> BasicDeclType.t option
  (**
   * If self is a subunit, return the body in which it is rooted.
   *)



end

module GenericSubpInternal : sig
  (**
   * Internal node for generic subprograms.
   *)

  type t =
    [
      | `GenericSubpInternal of
          GenericSubpInternalType.fields
    ]

  type fields = GenericSubpInternalType.fields =
    
  {
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_subp_spec :
    [< GenericSubpInternalType.t]
    -> SubpSpecType.t


end

module EnumLiteralDecl : sig
  (**
   * Declaration for an enumeration literal.
   *)

  type t =
    [
      | `EnumLiteralDecl of
          EnumLiteralDeclType.fields
    ]

  type fields = EnumLiteralDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_enum_type :
    [< EnumLiteralDeclType.t ]
    -> TypeDeclType.t option
  (**
   * Return the enum type corresponding to this enum literal.
   *)


      
  val f_name :
    [< EnumLiteralDeclType.t]
    -> DefiningNameType.t


end

module EntryDecl : sig
  (**
   * Entry declaration.
   *)

  type t =
    [
      | `EntryDecl of
          EntryDeclType.fields
    ]

  type fields = EntryDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_spec: EntrySpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_body_part :
    ?imprecise_fallback:
    bool
    -> [< EntryDeclType.t ]
    -> EntryBodyType.t option
  (**
   * Return the entry body associated to this entry declaration.
   *)


      
  val f_overriding :
    [< EntryDeclType.t]
    -> OverridingNodeType.t

      
  val f_spec :
    [< EntryDeclType.t]
    -> EntrySpecType.t


end

module SubpDecl : sig
  (**
   * Regular subprogram declaration.
   *)

  type t =
    [
      | `SubpDecl of
          SubpDeclType.fields
    ]

  type fields = SubpDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_overriding :
    [< SubpDeclType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< SubpDeclType.t]
    -> SubpSpecType.t


end

module ConcreteFormalSubpDecl : sig
  (**
   * Formal declaration for a concrete subprogram.
   *)

  type t =
    [
      | `ConcreteFormalSubpDecl of
          ConcreteFormalSubpDeclType.fields
    ]

  type fields = ConcreteFormalSubpDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_default_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_overriding :
    [< ConcreteFormalSubpDeclType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< ConcreteFormalSubpDeclType.t]
    -> SubpSpecType.t

      
  val f_default_expr :
    [< ConcreteFormalSubpDeclType.t]
    -> [AttributeRefType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t] option


end

module AbstractFormalSubpDecl : sig
  (**
   * Formal declaration for an abstract subprogram.
   *)

  type t =
    [
      | `AbstractFormalSubpDecl of
          AbstractFormalSubpDeclType.fields
    ]

  type fields = AbstractFormalSubpDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
         
    f_default_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_overriding :
    [< AbstractFormalSubpDeclType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< AbstractFormalSubpDeclType.t]
    -> SubpSpecType.t

      
  val f_default_expr :
    [< AbstractFormalSubpDeclType.t]
    -> [AttributeRefType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t] option


end

module FormalSubpDecl : sig
  (**
   * Formal subprogram declarations, in generic declarations formal parts.
   *)

  type t =
    [
      | AbstractFormalSubpDecl.t
      | ConcreteFormalSubpDecl.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_overriding :
    [< FormalSubpDeclType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< FormalSubpDeclType.t]
    -> SubpSpecType.t

      
  val f_default_expr :
    [< FormalSubpDeclType.t]
    -> [AttributeRefType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t] option


end

module AbstractSubpDecl : sig
  (**
   * Declaration for an abstract subprogram.
   *)

  type t =
    [
      | `AbstractSubpDecl of
          AbstractSubpDeclType.fields
    ]

  type fields = AbstractSubpDeclType.fields =
    
  {
         
    f_overriding: OverridingNodeType.t
    Lazy.t;
         
    f_subp_spec: SubpSpecType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_overriding :
    [< AbstractSubpDeclType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< AbstractSubpDeclType.t]
    -> SubpSpecType.t


end

module ClassicSubpDecl : sig
  (**
   * This is an intermediate abstract class for subprogram declarations with a
   * common structure: overriding indicator, ``SubpSpec``, aspects, <other
   * fields>.
   *)

  type t =
    [
      | AbstractSubpDecl.t
      | FormalSubpDecl.t
      | SubpDecl.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_body_part :
    ?imprecise_fallback:
    bool
    -> [< ClassicSubpDeclType.t ]
    -> BaseSubpBodyType.t option
  (**
   * Return the BaseSubpBody corresponding to this node.
   *)


      
  val f_overriding :
    [< ClassicSubpDeclType.t]
    -> OverridingNodeType.t

      
  val f_subp_spec :
    [< ClassicSubpDeclType.t]
    -> SubpSpecType.t


end

module BasicSubpDecl : sig
  (**
   * Base class for subprogram declarations.
   *)

  type t =
    [
      | ClassicSubpDecl.t
      | EntryDecl.t
      | EnumLiteralDecl.t
      | GenericSubpInternal.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_subp_decl_spec :
    [< BasicSubpDeclType.t ]
    -> BaseSubpSpecType.t option
  (**
   * Return the specification for this subprogram
   *)



end

module SynthAnonymousTypeDecl : sig
  (**
   * Synthetic anonymous type decl. Used to generate anonymous access types.
   *)

  type t =
    [
      | `SynthAnonymousTypeDecl of
          SynthAnonymousTypeDeclType.fields
    ]

  type fields = SynthAnonymousTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< SynthAnonymousTypeDeclType.t]
    -> DefiningNameType.t option

      
  val f_discriminants :
    [< SynthAnonymousTypeDeclType.t]
    -> DiscriminantPartType.t option

      
  val f_type_def :
    [< SynthAnonymousTypeDeclType.t]
    -> TypeDefType.t


end

module AnonymousTypeDecl : sig
  (**
   * Anonymous type declaration (for anonymous array or access types).
   *)

  type t =
    [
      | `AnonymousTypeDecl of
          AnonymousTypeDeclType.fields
      | `SynthAnonymousTypeDecl of
          SynthAnonymousTypeDeclType.fields
    ]

  type fields = AnonymousTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< AnonymousTypeDeclType.t]
    -> DefiningNameType.t option

      
  val f_discriminants :
    [< AnonymousTypeDeclType.t]
    -> DiscriminantPartType.t option

      
  val f_type_def :
    [< AnonymousTypeDeclType.t]
    -> TypeDefType.t


end

module TypeDecl : sig
  (**
   * Type declarations that embed a type definition node.
   *)

  type t =
    [
      | `TypeDecl of
          TypeDeclType.fields
      | `AnonymousTypeDecl of
          AnonymousTypeDeclType.fields
      | `SynthAnonymousTypeDecl of
          SynthAnonymousTypeDeclType.fields
    ]

  type fields = TypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_type_def: TypeDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_get_primitives :
    ?only_inherited:
    bool
    -> [< TypeDeclType.t ]
    -> BasicDeclType.t list
  (**
   * Return the list of all primitive operations that are available on this
   * type. If `only_inherited` is True, it will only return the primitives that
   * are implicitly inherited by this type, discarding those explicitly defined
   * on this type.
   *)


      
  val f_name :
    [< TypeDeclType.t]
    -> DefiningNameType.t option

      
  val f_discriminants :
    [< TypeDeclType.t]
    -> DiscriminantPartType.t option

      
  val f_type_def :
    [< TypeDeclType.t]
    -> TypeDefType.t


end

module SingleTaskTypeDecl : sig
  (**
   * Type declaration for a single task.
   *)

  type t =
    [
      | `SingleTaskTypeDecl of
          SingleTaskTypeDeclType.fields
    ]

  type fields = SingleTaskTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_definition: TaskDefType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< SingleTaskTypeDeclType.t]
    -> DefiningNameType.t option

      
  val f_discriminants :
    [< SingleTaskTypeDeclType.t]
    -> DiscriminantPartType.t option

      
  val f_definition :
    [< SingleTaskTypeDeclType.t]
    -> TaskDefType.t option


end

module TaskTypeDecl : sig
  (**
   * Declaration for a task type.
   *)

  type t =
    [
      | `TaskTypeDecl of
          TaskTypeDeclType.fields
      | `SingleTaskTypeDecl of
          SingleTaskTypeDeclType.fields
    ]

  type fields = TaskTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_definition: TaskDefType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< TaskTypeDeclType.t]
    -> DefiningNameType.t option

      
  val f_discriminants :
    [< TaskTypeDeclType.t]
    -> DiscriminantPartType.t option

      
  val f_definition :
    [< TaskTypeDeclType.t]
    -> TaskDefType.t option


end

module ProtectedTypeDecl : sig
  (**
   * Declaration for a protected type.
   *)

  type t =
    [
      | `ProtectedTypeDecl of
          ProtectedTypeDeclType.fields
    ]

  type fields = ProtectedTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_interfaces: ParentListType.t
    option
    Lazy.t;
         
    f_definition: ProtectedDefType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< ProtectedTypeDeclType.t]
    -> DefiningNameType.t option

      
  val f_discriminants :
    [< ProtectedTypeDeclType.t]
    -> DiscriminantPartType.t option

      
  val f_interfaces :
    [< ProtectedTypeDeclType.t]
    -> ParentListType.t option

      
  val f_definition :
    [< ProtectedTypeDeclType.t]
    -> ProtectedDefType.t


end

module IncompleteTaggedTypeDecl : sig
  (**
   * Incomplete declaration for a tagged type.
   *)

  type t =
    [
      | `IncompleteTaggedTypeDecl of
          IncompleteTaggedTypeDeclType.fields
    ]

  type fields = IncompleteTaggedTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
         
    f_has_abstract: AbstractNodeType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< IncompleteTaggedTypeDeclType.t]
    -> DefiningNameType.t option

      
  val f_discriminants :
    [< IncompleteTaggedTypeDeclType.t]
    -> DiscriminantPartType.t option

      
  val f_has_abstract :
    [< IncompleteTaggedTypeDeclType.t]
    -> AbstractNodeType.t


end

module IncompleteTypeDecl : sig
  (**
   * Incomplete declaration for a type.
   *)

  type t =
    [
      | `IncompleteTypeDecl of
          IncompleteTypeDeclType.fields
      | `IncompleteTaggedTypeDecl of
          IncompleteTaggedTypeDeclType.fields
    ]

  type fields = IncompleteTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_discriminants: DiscriminantPartType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< IncompleteTypeDeclType.t]
    -> DefiningNameType.t option

      
  val f_discriminants :
    [< IncompleteTypeDeclType.t]
    -> DiscriminantPartType.t option


end

module ClasswideTypeDecl : sig
  (**
   * Synthetic node (not parsed, generated from a property call). Refers to the
   * classwide type for a given tagged type. The aim is that those be mostly
   * equivalent to their non-classwide type, except for some resolution rules.
   *)

  type t =
    [
      | `ClasswideTypeDecl of
          ClasswideTypeDeclType.fields
    ]

  type fields = ClasswideTypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< ClasswideTypeDeclType.t]
    -> DefiningNameType.t option


end

module SubtypeDecl : sig
  (**
   * Subtype declaration.
   *)

  type t =
    [
      | `SubtypeDecl of
          SubtypeDeclType.fields
    ]

  type fields = SubtypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_subtype: SubtypeIndicationType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< SubtypeDeclType.t]
    -> DefiningNameType.t option

      
  val f_subtype :
    [< SubtypeDeclType.t]
    -> SubtypeIndicationType.t


end

module DiscreteBaseSubtypeDecl : sig
  (**
   * Specific ``BaseSubtypeDecl`` synthetic subclass for the base type of scalar
   * types.
   *)

  type t =
    [
      | `DiscreteBaseSubtypeDecl of
          DiscreteBaseSubtypeDeclType.fields
    ]

  type fields = DiscreteBaseSubtypeDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< DiscreteBaseSubtypeDeclType.t]
    -> DefiningNameType.t option


end

module BaseSubtypeDecl : sig
  (**
   * Base class for subtype declarations.
   *)

  type t =
    [
      | DiscreteBaseSubtypeDecl.t
      | SubtypeDecl.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_get_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseSubtypeDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Get the type for this subtype.
   *)


      
  val f_name :
    [< BaseSubtypeDeclType.t]
    -> DefiningNameType.t option


end

module BaseTypeDecl : sig
  (**
   * Base class for type declarations.
   *)

  type t =
    [
      | BaseSubtypeDecl.t
      | ClasswideTypeDecl.t
      | IncompleteTypeDecl.t
      | ProtectedTypeDecl.t
      | TaskTypeDecl.t
      | TypeDecl.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_base_subtype :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * If this type decl is a subtype decl, return the base subtype. If not,
   * return ``Self``.
   *)

      
  val p_private_completion :
    [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Return the private completion for this type, if there is one.
   *)

      
  val p_is_inherited_primitive :
    [< BaseTypeDeclType.t ]
    -> BasicDeclType.t
    -> bool
  (**
   * Assuming that P is a primitive of Self, return whether the given primitive
   * P is inherited from one of Self's parents.
   *)

      
  val p_get_record_representation_clause :
    ?imprecise_fallback:
    bool
    -> [< BaseTypeDeclType.t ]
    -> RecordRepClauseType.t option
  (**
   * Return the record representation clause associated to this type decl, if
   * applicable (i.e. this type decl defines a record type).
   *)

      
  val p_get_enum_representation_clause :
    ?imprecise_fallback:
    bool
    -> [< BaseTypeDeclType.t ]
    -> EnumRepClauseType.t option
  (**
   * Return the enum representation clause associated to this type decl, if
   * applicable (i.e. this type decl defines an enum type).
   *)

      
  val p_is_record_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Return whether this type is a record type.
   *)

      
  val p_is_array_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Return whether this type is an array type.
   *)

      
  val p_find_derived_types :
    ?imprecise_fallback:
    bool
    -> [< BaseTypeDeclType.t ]
    -> AdaNodeType.t
    -> AdaNodeType.t
    -> TypeDeclType.t list
  (**
   * Find types derived from self in the given ``root`` and its children.
   *)

      
  val p_is_real_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Whether type is a real type or not.
   *)

      
  val p_is_float_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Whether type is a float type or not.
   *)

      
  val p_is_fixed_point :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Whether type is a fixed point type or not.
   *)

      
  val p_is_enum_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Whether type is an enum type
   *)

      
  val p_is_access_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Whether Self is an access type or not
   *)

      
  val p_is_char_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Whether type is a character type or not
   *)

      
  val p_discrete_range :
    [< BaseTypeDeclType.t ]
    -> DiscreteRange.t
  (**
   * Return the discrete range for this type decl, if applicable.
   *)

      
  val p_is_discrete_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Whether type is a discrete type or not.
   *)

      
  val p_is_int_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Whether type is an integer type or not.
   *)

      
  val p_accessed_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * If this type is an access type, or a type with an Implicit_Dereference
   * aspect, return the type of a dereference of an instance of this type.
   *)

      
  val p_is_tagged_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Whether type is tagged or not
   *)

      
  val p_base_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Return the base type entity for this derived type declaration
   *)

      
  val p_base_types :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t list
  (**
   * Return the list of base types for Self.
   *)

      
  val p_find_all_derived_types :
    ?imprecise_fallback:
    bool
    -> [< BaseTypeDeclType.t ]
    -> AnalysisUnit.t list
    -> TypeDeclType.t list
  (**
   * Return the list of all types that inherit (directly or inderictly) from
   * Self among the given units.
   *)

      
  val p_comp_type :
    ?is_subscript:
    bool
    -> ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Return the component type of `Self`, if applicable. The component type is
   * the type you'll get if you call a value whose type is `Self`.  So it can
   * either be:
  
   * 1. The component type for an array. 2. The return type for an access to
   * function.
   *)

      
  val p_index_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> int
    -> BaseTypeDeclType.t option
  (**
   * Return the index type for dimension ``dim`` for this type, if applicable.
   *)

      
  val p_is_derived_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t
    -> bool
  (**
   * Whether Self is derived from other_type.
   *)

      
  val p_is_interface_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Return True iff this type declaration is an interface definition.
   *)

      
  val p_matching_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t
    -> bool
  (**
   * Return whether ``self`` matches ``expected_type``.
   *)

      
  val p_canonical_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Return the canonical type declaration for this type declaration. For
   * subtypes, it will return the base type declaration.
   *)

      
  val p_previous_part :
    ?go_to_incomplete:
    bool
    -> [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Returns the previous part for this type decl.
   *)

      
  val p_next_part :
    [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Returns the next part for this type decl.
   *)

      
  val p_full_view :
    [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Return the full completion of this type.
   *)

      
  val p_is_definite_subtype :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Returns whether this is a definite subtype.
  
   * For convenience, this will return ``False`` for incomplete types, even
   * though the correct answer is more akin to "non applicable".
   *)

      
  val p_is_private :
    [< BaseTypeDeclType.t ]
    -> bool
  (**
   * Whether node is a private view of corresponding type.
   *)

      
  val p_discriminants_list :
    [< BaseTypeDeclType.t ]
    -> BaseFormalParamDeclType.t list
  (**
   * Return the list of all discriminants of this type. If this type has no
   * discriminant or only unknown discriminants, an empty list is returned.
   *)

      
  val p_root_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Return the type that is at the root of the derivation hierarchy (ignoring
   * secondary interfaces derivations for tagged types)
   *)

      
  val p_shapes :
    ?include_discriminants:
    bool
    -> ?origin:
    AdaNodeType.t
    -> [< BaseTypeDeclType.t ]
    -> Shape.t list
  (**
   * Must be called on a record (sub-)type declaration. Return all the possible
   * shapes that a value of this record type can take. For example, consider the
   * following record definition:
  
   * .. code::
  
   * type R (A : Integer; B : Integer) is record X : Integer; case A is when 1
   * .. 10 => Y_1 : Integer; case B is when 1 .. 10 => Z_1 : Integer; when
   * others => null; end case; when 11 .. 20 => Y_2 : Integer; case B is when 1
   * .. 10 => Z_2 : Integer; when others => null; end case; when others => null;
   * end case; end record;
  
   * For this instance, this property will return the following results:
  
   * .. code::
  
   * [ [X, Y_1, Z_1], [X, Y_1], [X, Y_2, Z_2], [X, Y_2], [X] ]
  
   * .. ATTENTION:: This property is inaccurate when called on a record
   *    extension which defines components under a certain condition C, and this
   *    same condition is used to define some components in the parent record:
   *    in that case, any feasible shape will in practice contain either both
   *    the components defined under condition C in the child record and the
   *    parent record, or none of them.
  
   * However, due to the simplified algorithm we use here to compute the
   * feasible shapes, we will also return shapes that include the components of
   * the child record but not the parent record, and conversely.
   *)


      
  val f_name :
    [< BaseTypeDeclType.t]
    -> DefiningNameType.t option


end

module PackageDecl : sig
  (**
   * Non-generic package declarations.
   *)

  type t =
    [
      | `PackageDecl of
          PackageDeclType.fields
    ]

  type fields = PackageDeclType.fields =
    
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_package_name :
    [< PackageDeclType.t]
    -> DefiningNameType.t

      
  val f_public_part :
    [< PackageDeclType.t]
    -> PublicPartType.t

      
  val f_private_part :
    [< PackageDeclType.t]
    -> PrivatePartType.t option

      
  val f_end_name :
    [< PackageDeclType.t]
    -> EndNameType.t option


end

module GenericPackageInternal : sig
  (**
   * This class denotes the internal package contained by a GenericPackageDecl.
   *)

  type t =
    [
      | `GenericPackageInternal of
          GenericPackageInternalType.fields
    ]

  type fields = GenericPackageInternalType.fields =
    
  {
         
    f_package_name: DefiningNameType.t
    Lazy.t;
         
    f_public_part: PublicPartType.t
    Lazy.t;
         
    f_private_part: PrivatePartType.t
    option
    Lazy.t;
         
    f_end_name: EndNameType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_package_name :
    [< GenericPackageInternalType.t]
    -> DefiningNameType.t

      
  val f_public_part :
    [< GenericPackageInternalType.t]
    -> PublicPartType.t

      
  val f_private_part :
    [< GenericPackageInternalType.t]
    -> PrivatePartType.t option

      
  val f_end_name :
    [< GenericPackageInternalType.t]
    -> EndNameType.t option


end

module BasePackageDecl : sig
  (**
   * Package declarations. Concrete instances of this class will be created in
   * generic package declarations. Other non-generic package declarations will
   * be instances of PackageDecl.
  
   * The behavior is the same, the only difference is that BasePackageDecl and
   * PackageDecl have different behavior regarding lexical environments. In the
   * case of generic package declarations, we use BasePackageDecl which has no
   * env_spec, and the environment behavior is handled by the GenericPackageDecl
   * instance.
   *)

  type t =
    [
      | GenericPackageInternal.t
      | PackageDecl.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_body_part :
    [< BasePackageDeclType.t ]
    -> PackageBodyType.t option
  (**
   * Return the PackageBody corresponding to this node.
   *)


      
  val f_package_name :
    [< BasePackageDeclType.t]
    -> DefiningNameType.t

      
  val f_public_part :
    [< BasePackageDeclType.t]
    -> PublicPartType.t

      
  val f_private_part :
    [< BasePackageDeclType.t]
    -> PrivatePartType.t option

      
  val f_end_name :
    [< BasePackageDeclType.t]
    -> EndNameType.t option


end

module ParamSpec : sig
  (**
   * Specification for a parameter.
   *)

  type t =
    [
      | `ParamSpec of
          ParamSpecType.fields
    ]

  type fields = ParamSpecType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_has_aliased: AliasedNodeType.t
    Lazy.t;
         
    f_mode: ModeType.t
    option
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ids :
    [< ParamSpecType.t]
    -> DefiningNameListType.t

      
  val f_has_aliased :
    [< ParamSpecType.t]
    -> AliasedNodeType.t

      
  val f_mode :
    [< ParamSpecType.t]
    -> ModeType.t option

      
  val f_type_expr :
    [< ParamSpecType.t]
    -> [AnonymousTypeType.t | SubtypeIndicationType.t]

      
  val f_default_expr :
    [< ParamSpecType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option


end

module GenericFormalTypeDecl : sig
  (**
   * Formal declaration for a type.
   *)

  type t =
    [
      | `GenericFormalTypeDecl of
          GenericFormalTypeDeclType.fields
    ]

  type fields = GenericFormalTypeDeclType.fields =
    
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decl :
    [< GenericFormalTypeDeclType.t]
    -> [FormalSubpDeclType.t | GenericInstantiationType.t | IncompleteTypeDeclType.t | NumberDeclType.t | ObjectDeclType.t | SingleProtectedDeclType.t | SingleTaskDeclType.t | TypeDeclType.t]


end

module GenericFormalSubpDecl : sig
  (**
   * Formal declaration for a subprogram.
   *)

  type t =
    [
      | `GenericFormalSubpDecl of
          GenericFormalSubpDeclType.fields
    ]

  type fields = GenericFormalSubpDeclType.fields =
    
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decl :
    [< GenericFormalSubpDeclType.t]
    -> [FormalSubpDeclType.t | GenericInstantiationType.t | IncompleteTypeDeclType.t | NumberDeclType.t | ObjectDeclType.t | SingleProtectedDeclType.t | SingleTaskDeclType.t | TypeDeclType.t]


end

module GenericFormalPackage : sig
  (**
   * Formal declaration for a package.
   *)

  type t =
    [
      | `GenericFormalPackage of
          GenericFormalPackageType.fields
    ]

  type fields = GenericFormalPackageType.fields =
    
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decl :
    [< GenericFormalPackageType.t]
    -> [FormalSubpDeclType.t | GenericInstantiationType.t | IncompleteTypeDeclType.t | NumberDeclType.t | ObjectDeclType.t | SingleProtectedDeclType.t | SingleTaskDeclType.t | TypeDeclType.t]


end

module GenericFormalObjDecl : sig
  (**
   * Formal declaration for an object.
   *)

  type t =
    [
      | `GenericFormalObjDecl of
          GenericFormalObjDeclType.fields
    ]

  type fields = GenericFormalObjDeclType.fields =
    
  {
         
    f_decl: [
      | `AbstractFormalSubpDecl
          of AbstractFormalSubpDeclType.fields
      | `AnonymousTypeDecl
          of AnonymousTypeDeclType.fields
      | `ConcreteFormalSubpDecl
          of ConcreteFormalSubpDeclType.fields
      | `ExtendedReturnStmtObjectDecl
          of ExtendedReturnStmtObjectDeclType.fields
      | `GenericPackageInstantiation
          of GenericPackageInstantiationType.fields
      | `GenericSubpInstantiation
          of GenericSubpInstantiationType.fields
      | `IncompleteTaggedTypeDecl
          of IncompleteTaggedTypeDeclType.fields
      | `IncompleteTypeDecl
          of IncompleteTypeDeclType.fields
      | `NumberDecl
          of NumberDeclType.fields
      | `ObjectDecl
          of ObjectDeclType.fields
      | `SingleProtectedDecl
          of SingleProtectedDeclType.fields
      | `SingleTaskDecl
          of SingleTaskDeclType.fields
      | `SynthAnonymousTypeDecl
          of SynthAnonymousTypeDeclType.fields
      | `TypeDecl
          of TypeDeclType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decl :
    [< GenericFormalObjDeclType.t]
    -> [FormalSubpDeclType.t | GenericInstantiationType.t | IncompleteTypeDeclType.t | NumberDeclType.t | ObjectDeclType.t | SingleProtectedDeclType.t | SingleTaskDeclType.t | TypeDeclType.t]


end

module GenericFormal : sig
  (**
   * Enclosing declaration for a generic formal. The real declaration is
   * accessible via the ``decl`` field.
   *)

  type t =
    [
      | GenericFormalObjDecl.t
      | GenericFormalPackage.t
      | GenericFormalSubpDecl.t
      | GenericFormalTypeDecl.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decl :
    [< GenericFormalType.t]
    -> [FormalSubpDeclType.t | GenericInstantiationType.t | IncompleteTypeDeclType.t | NumberDeclType.t | ObjectDeclType.t | SingleProtectedDeclType.t | SingleTaskDeclType.t | TypeDeclType.t]


end

module DiscriminantSpec : sig
  (**
   * Known list of discriminants in type declarations.
   *)

  type t =
    [
      | `DiscriminantSpec of
          DiscriminantSpecType.fields
    ]

  type fields = DiscriminantSpecType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_type_expr: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ids :
    [< DiscriminantSpecType.t]
    -> DefiningNameListType.t

      
  val f_type_expr :
    [< DiscriminantSpecType.t]
    -> [AnonymousTypeType.t | SubtypeIndicationType.t]

      
  val f_default_expr :
    [< DiscriminantSpecType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option


end

module ComponentDecl : sig
  (**
   * Declaration for a component.
   *)

  type t =
    [
      | `ComponentDecl of
          ComponentDeclType.fields
    ]

  type fields = ComponentDeclType.fields =
    
  {
         
    f_ids: DefiningNameListType.t
    Lazy.t;
         
    f_component_def: ComponentDefType.t
    Lazy.t;
         
    f_default_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ids :
    [< ComponentDeclType.t]
    -> DefiningNameListType.t

      
  val f_component_def :
    [< ComponentDeclType.t]
    -> ComponentDefType.t

      
  val f_default_expr :
    [< ComponentDeclType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option


end

module BaseFormalParamDecl : sig
  (**
   * Base class for formal parameter declarations. This is used both for records
   * components and for subprogram parameters.
  
   * This is a Libadalang abstraction, that has no ARM existence.
   *)

  type t =
    [
      | ComponentDecl.t
      | DiscriminantSpec.t
      | GenericFormal.t
      | ParamSpec.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_formal_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseFormalParamDeclType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Return the type for this formal.
   *)



end

module AnonymousExprDecl : sig
  (**
   * Represents a anonymous declaration that holds an expression.
  
   * This is used to store the results of queries such as ``referenced_decl``
   * called on references to object formals from inside a instantiated generic
   * in order to return the relevant actual.
  
   * Indeed, ``referenced_decl`` must return a ``BasicDecl``, but actuals of
   * generic instantiations are ``Expr``s. This wrapper node is therefore a way
   * to both satisfy the ``BasicDecl`` inteface, and provide to the user the
   * expression of the actual through the ``expr`` field.
   *)

  type t =
    [
      | `AnonymousExprDecl of
          AnonymousExprDeclType.fields
    ]

  type fields = AnonymousExprDeclType.fields =
    
  {
         
    f_expr: ExprType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_expr :
    [< AnonymousExprDeclType.t]
    -> ExprType.t


end

module AbstractStateDecl : sig
  (**
   * Contained (directly or indirectly) in an AbstractStateDeclExpr, and is used
   * to represent the BasicDecl associated with the abstract state introduced by
   * the Abstract_State aspect. This node is necessary because all of our name
   * resolution routines expect BasicDecls as environments' values.
  
   * The only purpose of this node is to populate the env with the abstract
   * state declared through this node, so it can be referred in SPARK aspects
   * such as Global, Depends, Refined_State, etc.
   *)

  type t =
    [
      | `AbstractStateDecl of
          AbstractStateDeclType.fields
    ]

  type fields = AbstractStateDeclType.fields =
    
  {
         
    f_name: DefiningNameType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< AbstractStateDeclType.t]
    -> DefiningNameType.t


end

module BasicDecl : sig
  (**
   * Root class for an Ada declaration (RM 3.1). A declaration associates a name
   * with a language entity, for example a type or a variable.
   *)

  type t =
    [
      | AbstractStateDecl.t
      | AnonymousExprDecl.t
      | BaseFormalParamDecl.t
      | BasePackageDecl.t
      | BaseTypeDecl.t
      | BasicSubpDecl.t
      | BodyNode.t
      | EntryIndexSpec.t
      | ErrorDecl.t
      | ExceptionDecl.t
      | ExceptionHandler.t
      | ForLoopVarDecl.t
      | GenericDecl.t
      | GenericInstantiation.t
      | GenericRenamingDecl.t
      | LabelDecl.t
      | NamedStmtDecl.t
      | NumberDecl.t
      | ObjectDecl.t
      | PackageRenamingDecl.t
      | SingleProtectedDecl.t
      | SingleTaskDecl.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_is_formal :
    [< BasicDeclType.t ]
    -> bool
  (**
   * Whether this decl is the nested decl of a generic formal declaration.
   *)

      
  val p_doc_annotations :
    [< BasicDeclType.t ]
    -> DocAnnotation.t list
  (**
   * Return the documentation annotations associated with this decl. Annotations
   * are any comment line of the form::
  
   * --% [annotation_name]: [annotation]
  
   * Raises a property error if the doc is incorrectly formatted.
  
   * .. ATTENTION:: This is an experimental feature, so even if it is exposed to
   *    allow experiments, it is totally unsupported and the API and behavior
   *    are very likely to change in the future.
   *)

      
  val p_doc :
    [< BasicDeclType.t ]
    -> string
  (**
   * Return the documentation associated with this decl. Raises a property error
   * if the doc is incorrectly formatted.
  
   * .. ATTENTION:: This is an experimental feature, so even if it is exposed to
   *    allow experiments, it is totally unsupported and the API and behavior
   *    are very likely to change in the future.
   *)

      
  val p_previous_part_for_decl :
    ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> BasicDeclType.t option
  (**
   * Return the previous part for this decl, if applicable.
  
   * .. note:: It is not named previous_part, because BaseTypeDecl has a more
   *    precise version of previous_part that returns a BaseTypeDecl. Probably,
   *    we want to rename the specific versions, and have the root property be
   *    named previous_part. (TODO R925-008)
   *)

      
  val p_canonical_part :
    ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> BasicDeclType.t option
  (**
   * Return the canonical part for this decl. In the case of decls composed of
   * several parts, the canonical part will be the first part.
   *)

      
  val p_all_parts :
    ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> BasicDeclType.t list
  (**
   * Return all parts that define this entity, sorted from first part to last
   * part.
   *)

      
  val p_is_static_decl :
    ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> bool
  (**
   * Return whether this declaration is static.
   *)

      
  val p_is_imported :
    [< BasicDeclType.t ]
    -> bool
  (**
   * Whether this declaration is imported from another language.
   *)

      
  val p_get_aspect_assoc :
    [< BasicDeclType.t ]
    -> string
    -> AspectAssocType.t option
  (**
   * Return the aspect with name ``name`` for this entity.
   *)

      
  val p_get_aspect_spec_expr :
    [< BasicDeclType.t ]
    -> string
    -> ExprType.t option
  (**
   * Return the expression associated to the aspect with name ``name`` for this
   * entity.
   *)

      
  val p_get_aspect :
    ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> string
    -> Aspect.t
  (**
   * Return the aspect with name ``name`` associated to this entity.
  
   * Aspects are properties of entities that can be specified by the Ada
   * program, either via aspect specifications, pragmas, or attributes.
  
   * This will return the syntactic node corresponding to attribute directly.
  
   * Note: for some aspects (e.g. Inline), Libadalang will check if they are
   * defined on any part of the entity.
   *)

      
  val p_has_aspect :
    ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> string
    -> bool
  (**
   * Returns whether the boolean aspect named ``name`` is set on the entity
   * represented by this node.
  
   * "Aspect" is used as in RM terminology (see RM 13).
   *)

      
  val p_get_pragma :
    [< BasicDeclType.t ]
    -> string
    -> PragmaNodeType.t option
  (**
   * Return the pragma with name ``name`` associated to this entity.
   *)

      
  val p_get_representation_clause :
    ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> string
    -> AttributeDefClauseType.t option
  (**
   * Return the representation clause associated to this type decl that defines
   * the given attribute name.
   *)

      
  val p_is_compilation_unit_root :
    [< BasicDeclType.t ]
    -> bool
  (**
   * Whether a BasicDecl is the root decl for its unit.
   *)

      
  val p_is_visible :
    [< BasicDeclType.t ]
    -> AdaNodeType.t
    -> bool
  (**
   * Return whether this declaration is visible from the point of view of the
   * given ``origin`` node.
  
   * .. ATTENTION:: Only package-level (public or private) declarations are
   *    supported for now.
   *)

      
  val p_base_subp_declarations :
    ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> BasicDeclType.t list
  (**
   * If Self declares a primitive subprogram of some tagged type T, return the
   * set of all subprogram declarations that it overrides (including itself).
  
   * NOTE: for the moment this only works for tagged types. Remains to be seen
   * if we need to extend it.
   *)

      
  val p_root_subp_declarations :
    ?origin:
    AdaNodeType.t
    -> ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> BasicDeclType.t list
  (**
   * If Self declares a primitive subprogram of some tagged type T, return the
   * root subprogram declarations that it overrides. There can be several, as in
   * the following scenario:
  
   * - package Root defines the root tagged type T and subprogram Foo.
  
   * - package Itf defines interface I and abstract subprogram Foo.
  
   * - package D defines "type U is new Root.T and Itf.I" and an overriding
   * subprogram Foo.
  
   * Here, root_subp_declarations of Foo defined in package D will return both
   * Foo from package Root and Foo from package Itf.
   *)

      
  val p_find_all_overrides :
    ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> AnalysisUnit.t list
    -> BasicDeclType.t list
  (**
   * If Self is the declaration of a primitive of some type T, return the list
   * of all subprogram that override this subprogram among the given units.
   *)

      
  val p_defining_names :
    [< BasicDeclType.t ]
    -> DefiningNameType.t list
  (**
   * Get all the names of this basic declaration.
   *)

      
  val p_defining_name :
    [< BasicDeclType.t ]
    -> DefiningNameType.t option
  (**
   * Get the name of this declaration. If this declaration has several names, it
   * will return the first one.
   *)

      
  val p_type_expression :
    [< BasicDeclType.t ]
    -> TypeExprType.t option
  (**
   * Return the type expression for this BasicDecl if applicable, a null
   * otherwise.
   *)

      
  val p_subp_spec_or_null :
    ?follow_generic:
    bool
    -> [< BasicDeclType.t ]
    -> BaseSubpSpecType.t option
  (**
   * If Self is a Subp, returns the specification of this subprogram.
  
   * If ``follow_generic`` is True, will also work for instances of
   * ``GenericSubpDecl``.
   *)

      
  val p_is_subprogram :
    [< BasicDeclType.t ]
    -> bool
  (**
   * Return True if self is a subprogram node in the general sense (which is, an
   * entity that can be called). This includes separates and entries.
  
   * .. attention: This is a purely syntactic query and will return True for
   *    everything that is a syntactic entity that can be called like a
   *    subprogram in some contexts, even generic formal subprograms for
   *    example.
   *)

      
  val p_relative_name :
    [< BasicDeclType.t ]
    -> SingleTokNodeType.t option
  (**
   * Return the relative name for Self. If Self's defining name is ``A.B.C``,
   * return C as a node.
   *)

      
  val p_relative_name_text :
    [< BasicDeclType.t ]
    -> string
  (**
   * Return the relative name for Self, as text.
   *)

      
  val p_next_part_for_decl :
    ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> BasicDeclType.t option
  (**
   * Return the next part of this declaration, if applicable.
  
   * .. note:: It is not named next_part, because BaseTypeDecl has a more
   *    precise version of next_part that returns a BaseTypeDecl. Probably, we
   *    want to rename the specific versions, and have the root property be
   *    named next_part. (TODO R925-008)
   *)

      
  val p_body_part_for_decl :
    ?imprecise_fallback:
    bool
    -> [< BasicDeclType.t ]
    -> BodyNodeType.t option
  (**
   * Return the body corresponding to this declaration, if applicable.
  
   * .. note:: It is not named body_part, subclasses have more precise versions
   *    named body_part and returning a more precise result. Probably, we want
   *    to rename the specific versions, and have the root property be named
   *    body_part. (TODO R925-008)
   *)

      
  val p_fully_qualified_name_array :
    [< BasicDeclType.t ]
    -> string list
  (**
   * Return the fully qualified name corresponding to this declaration, as an
   * array of symbols.
   *)

      
  val p_fully_qualified_name :
    [< BasicDeclType.t ]
    -> string
  (**
   * Return the fully qualified name corresponding to this declaration.
   *)

      
  val p_canonical_fully_qualified_name :
    [< BasicDeclType.t ]
    -> string
  (**
   * Return a canonical representation of the fully qualified name corresponding
   * to this declaration.
   *)

      
  val p_unique_identifying_name :
    [< BasicDeclType.t ]
    -> string
  (**
   * Return a unique identifying name for this declaration, provided this
   * declaration is a public declaration. In the case of subprograms, this will
   * include the profile.
  
   * .. attention:: This will only return a unique name for public declarations.
   *    Notably, anything nested in an unnamed declare block won't be handled
   *    correctly.
   *)



end

module ParamAssoc : sig
  (**
   * Assocation (X => Y) used for aggregates and parameter associations.
   *)

  type t =
    [
      | `ParamAssoc of
          ParamAssocType.fields
    ]

  type fields = ParamAssocType.fields =
    
  {
         
    f_designator: [
      | `Identifier
          of IdentifierType.fields
      | `OthersDesignator
          of OthersDesignatorType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_designator :
    [< ParamAssocType.t]
    -> [IdentifierType.t | OthersDesignatorType.t | StringLiteralType.t] option

      
  val f_r_expr :
    [< ParamAssocType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module IteratedAssoc : sig
  (**
   * Iterated association (Ada 2020).
   *)

  type t =
    [
      | `IteratedAssoc of
          IteratedAssocType.fields
    ]

  type fields = IteratedAssocType.fields =
    
  {
         
    f_spec: ForLoopSpecType.t
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_spec :
    [< IteratedAssocType.t]
    -> ForLoopSpecType.t

      
  val f_r_expr :
    [< IteratedAssocType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module DiscriminantAssoc : sig
  (**
   * Association of discriminant names to an expression.
   *)

  type t =
    [
      | `DiscriminantAssoc of
          DiscriminantAssocType.fields
    ]

  type fields = DiscriminantAssocType.fields =
    
  {
         
    f_ids: IdentifierListType.t
    option
    Lazy.t;
         
    f_discr_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_ids :
    [< DiscriminantAssocType.t]
    -> IdentifierListType.t option

      
  val f_discr_expr :
    [< DiscriminantAssocType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module MultiDimArrayAssoc : sig
  (**
   * Association used for multi-dimension array aggregates.
   *)

  type t =
    [
      | `MultiDimArrayAssoc of
          MultiDimArrayAssocType.fields
    ]

  type fields = MultiDimArrayAssocType.fields =
    
  {
         
    f_designators: AlternativesListType.t
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_designators :
    [< MultiDimArrayAssocType.t]
    -> AlternativesListType.t option

      
  val f_r_expr :
    [< MultiDimArrayAssocType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module AggregateAssoc : sig
  (**
   * Assocation (X => Y) used for aggregates and parameter associations.
   *)

  type t =
    [
      | `AggregateAssoc of
          AggregateAssocType.fields
      | `MultiDimArrayAssoc of
          MultiDimArrayAssocType.fields
    ]

  type fields = AggregateAssocType.fields =
    
  {
         
    f_designators: AlternativesListType.t
    option
    Lazy.t;
         
    f_r_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BoxExpr
          of BoxExprType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_designators :
    [< AggregateAssocType.t]
    -> AlternativesListType.t option

      
  val f_r_expr :
    [< AggregateAssocType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | BoxExprType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module BasicAssoc : sig
  (**
   * Association of one or several names to an expression.
   *)

  type t =
    [
      | AggregateAssoc.t
      | DiscriminantAssoc.t
      | IteratedAssoc.t
      | ParamAssoc.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_get_params :
    ?imprecise_fallback:
    bool
    -> [< BasicAssocType.t ]
    -> DefiningNameType.t list
  (**
   * Return the list of parameters that this association refers to.
   *)



end

module RecordDef : sig
  (**
   * Record definition that contains components (``record ... end record``).
   *)

  type t =
    [
      | `RecordDef of
          RecordDefType.fields
    ]

  type fields = RecordDefType.fields =
    
  {
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_components :
    [< RecordDefType.t]
    -> ComponentListType.t


end

module NullRecordDef : sig
  (**
   * Record definition for ``null record``.
   *)

  type t =
    [
      | `NullRecordDef of
          NullRecordDefType.fields
    ]

  type fields = NullRecordDefType.fields =
    
  {
         
    f_components: ComponentListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_components :
    [< NullRecordDefType.t]
    -> ComponentListType.t


end

module BaseRecordDef : sig
  (**
   * Base class for record definitions.
   *)

  type t =
    [
      | NullRecordDef.t
      | RecordDef.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_components :
    [< BaseRecordDefType.t]
    -> ComponentListType.t


end

module GenericFormalPart : sig
  (**
   * List of declaration for generic formals.
   *)

  type t =
    [
      | `GenericFormalPart of
          GenericFormalPartType.fields
    ]

  type fields = GenericFormalPartType.fields =
    
  {
         
    f_decls: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_decls :
    [< GenericFormalPartType.t]
    -> AdaNodeListType.t


end

module EntryCompletionFormalParams : sig
  (**
   * Formal parameters for the completion of an ``EntryDecl`` (either an
   * ``EntryBody`` or an ``AcceptStmt``).
   *)

  type t =
    [
      | `EntryCompletionFormalParams of
          EntryCompletionFormalParamsType.fields
    ]

  type fields = EntryCompletionFormalParamsType.fields =
    
  {
         
    f_params: ParamsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_params :
    [< EntryCompletionFormalParamsType.t]
    -> ParamsType.t option


end

module UnknownDiscriminantPart : sig
  (**
   * Unknown list of discriminants in type declarations.
   *)

  type t =
    [
      | `UnknownDiscriminantPart of
          UnknownDiscriminantPartType.fields
    ]

  type fields = UnknownDiscriminantPartType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module KnownDiscriminantPart : sig
  (**
   * Known list of discriminants in type declarations.
   *)

  type t =
    [
      | `KnownDiscriminantPart of
          KnownDiscriminantPartType.fields
    ]

  type fields = KnownDiscriminantPartType.fields =
    
  {
         
    f_discr_specs: DiscriminantSpecListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_discr_specs :
    [< KnownDiscriminantPartType.t]
    -> DiscriminantSpecListType.t


end

module DiscriminantPart : sig
  (**
   * Specification for discriminants in type declarations.
   *)

  type t =
    [
      | KnownDiscriminantPart.t
      | UnknownDiscriminantPart.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module ComponentList : sig
  (**
   * List of component declarations.
   *)

  type t =
    [
      | `ComponentList of
          ComponentListType.fields
    ]

  type fields = ComponentListType.fields =
    
  {
         
    f_components: AdaNodeListType.t
    option
    Lazy.t;
         
    f_variant_part: VariantPartType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_components :
    [< ComponentListType.t]
    -> AdaNodeListType.t option

      
  val f_variant_part :
    [< ComponentListType.t]
    -> VariantPartType.t option


end

module SubpSpec : sig
  (**
   * Subprogram specification.
   *)

  type t =
    [
      | `SubpSpec of
          SubpSpecType.fields
    ]

  type fields = SubpSpecType.fields =
    
  {
         
    f_subp_kind: SubpKindType.t
    Lazy.t;
         
    f_subp_name: DefiningNameType.t
    option
    Lazy.t;
         
    f_subp_params: ParamsType.t
    option
    Lazy.t;
         
    f_subp_returns: [
      | `AnonymousType
          of AnonymousTypeType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_subp_kind :
    [< SubpSpecType.t]
    -> SubpKindType.t

      
  val f_subp_name :
    [< SubpSpecType.t]
    -> DefiningNameType.t option

      
  val f_subp_params :
    [< SubpSpecType.t]
    -> ParamsType.t option

      
  val f_subp_returns :
    [< SubpSpecType.t]
    -> [AnonymousTypeType.t | SubtypeIndicationType.t] option


end

module EnumSubpSpec : sig
  (**
   * Synthetic node for the abstract subprogram spec of an enum literal.
  
   * NOTE: This has no existence in the ARM. While enum literals are functions
   * semantically, they're not such syntactically.
   *)

  type t =
    [
      | `EnumSubpSpec of
          EnumSubpSpecType.fields
    ]

  type fields = EnumSubpSpecType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module EntrySpec : sig
  (**
   * Entry specification.
   *)

  type t =
    [
      | `EntrySpec of
          EntrySpecType.fields
    ]

  type fields = EntrySpecType.fields =
    
  {
         
    f_entry_name: DefiningNameType.t
    Lazy.t;
         
    f_family_type: [
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ConstrainedSubtypeIndication
          of ConstrainedSubtypeIndicationType.fields
      | `DiscreteSubtypeIndication
          of DiscreteSubtypeIndicationType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `SubtypeIndication
          of SubtypeIndicationType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_entry_params: ParamsType.t
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_entry_name :
    [< EntrySpecType.t]
    -> DefiningNameType.t

      
  val f_family_type :
    [< EntrySpecType.t]
    -> [AttributeRefType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | SubtypeIndicationType.t | TargetNameType.t] option

      
  val f_entry_params :
    [< EntrySpecType.t]
    -> ParamsType.t option


end

module BaseSubpSpec : sig
  (**
   * Base class for subprogram specifications.
   *)

  type t =
    [
      | EntrySpec.t
      | EnumSubpSpec.t
      | SubpSpec.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_returns :
    [< BaseSubpSpecType.t ]
    -> TypeExprType.t option
  (**
   * Syntax property. Return the type expression node corresponding to the
   * return of this subprogram spec.
   *)

      
  val p_params :
    [< BaseSubpSpecType.t ]
    -> ParamSpecType.t list
  (**
   * Returns the array of parameters specification for this subprogram spec.
   *)

      
  val p_primitive_subp_types :
    ?imprecise_fallback:
    bool
    -> [< BaseSubpSpecType.t ]
    -> BaseTypeDeclType.t list
  (**
   * Return the types of which this subprogram is a primitive of.
   *)

      
  val p_primitive_subp_first_type :
    ?imprecise_fallback:
    bool
    -> [< BaseSubpSpecType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Return the first type of which this subprogram is a primitive of.
   *)

      
  val p_primitive_subp_tagged_type :
    ?imprecise_fallback:
    bool
    -> [< BaseSubpSpecType.t ]
    -> BaseTypeDeclType.t option
  (**
   * If this subprogram is a primitive for a tagged type, then return this type.
   *)

      
  val p_return_type :
    ?origin:
    AdaNodeType.t
    -> [< BaseSubpSpecType.t ]
    -> BaseTypeDeclType.t option
  (**
   * Returns the return type of Self, if applicable (e.g. if Self is a
   * subprogram). Else, returns null.
   *)



end

module BaseFormalParamHolder : sig
  (**
   * Base class for lists of formal parameters. This is used both for subprogram
   * specifications and for records, so that we can share the matching and
   * unpacking logic.
   *)

  type t =
    [
      | BaseSubpSpec.t
      | ComponentList.t
      | DiscriminantPart.t
      | EntryCompletionFormalParams.t
      | GenericFormalPart.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_abstract_formal_params :
    [< BaseFormalParamHolderType.t ]
    -> BaseFormalParamDeclType.t list
  (**
   * Return the list of abstract formal parameters for this holder.
   *)

      
  val p_nb_min_params :
    [< BaseFormalParamHolderType.t ]
    -> int
  (**
   * Return the minimum number of parameters this subprogram can be called while
   * still being a legal call.
   *)

      
  val p_nb_max_params :
    [< BaseFormalParamHolderType.t ]
    -> int
  (**
   * Return the maximum number of parameters this subprogram can be called while
   * still being a legal call.
   *)

      
  val p_param_types :
    ?origin:
    AdaNodeType.t
    -> [< BaseFormalParamHolderType.t ]
    -> BaseTypeDeclType.t list
  (**
   * Returns the type of each parameter of Self.
   *)



end

module PragmaArgumentAssoc : sig
  (**
   * Argument assocation in a pragma.
   *)

  type t =
    [
      | `PragmaArgumentAssoc of
          PragmaArgumentAssocType.fields
    ]

  type fields = PragmaArgumentAssocType.fields =
    
  {
         
    f_id: IdentifierType.t
    option
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_id :
    [< PragmaArgumentAssocType.t]
    -> IdentifierType.t option

      
  val f_expr :
    [< PragmaArgumentAssocType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module ContractCaseAssoc : sig
  (**
   * Single association for the ``Contract_Case`` aspect.
   *)

  type t =
    [
      | `ContractCaseAssoc of
          ContractCaseAssocType.fields
    ]

  type fields = ContractCaseAssocType.fields =
    
  {
         
    f_guard: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `OthersDesignator
          of OthersDesignatorType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_consequence: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_guard :
    [< ContractCaseAssocType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | OthersDesignatorType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]

      
  val f_consequence :
    [< ContractCaseAssocType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module BaseAssoc : sig
  (**
   * Abstract class for a key/value association, where the value is an
   * expression.
   *)

  type t =
    [
      | ContractCaseAssoc.t
      | PragmaArgumentAssoc.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_assoc_expr :
    [< BaseAssocType.t ]
    -> ExprType.t option
  (**
   * Returns the expression side of this assoc node.
   *)



end

module AspectSpec : sig
  (**
   * List of aspects in a declaration.
   *)

  type t =
    [
      | `AspectSpec of
          AspectSpecType.fields
    ]

  type fields = AspectSpecType.fields =
    
  {
         
    f_aspect_assocs: AspectAssocListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_aspect_assocs :
    [< AspectSpecType.t]
    -> AspectAssocListType.t


end

module RecordRepClause : sig
  (**
   * Representation clause for a record type.
   *)

  type t =
    [
      | `RecordRepClause of
          RecordRepClauseType.fields
    ]

  type fields = RecordRepClauseType.fields =
    
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_at_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
         
    f_components: AdaNodeListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< RecordRepClauseType.t]
    -> [CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]

      
  val f_at_expr :
    [< RecordRepClauseType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option

      
  val f_components :
    [< RecordRepClauseType.t]
    -> AdaNodeListType.t


end

module EnumRepClause : sig
  (**
   * Representation clause for enumeration types.
   *)

  type t =
    [
      | `EnumRepClause of
          EnumRepClauseType.fields
    ]

  type fields = EnumRepClauseType.fields =
    
  {
         
    f_type_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_aggregate: BaseAggregateType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_type_name :
    [< EnumRepClauseType.t]
    -> [CharLiteralType.t | DottedNameType.t | IdentifierType.t | StringLiteralType.t]

      
  val f_aggregate :
    [< EnumRepClauseType.t]
    -> BaseAggregateType.t


end

module AttributeDefClause : sig
  (**
   * Clause for an attribute definition (``for ...'Attribute use ...;``).
   *)

  type t =
    [
      | `AttributeDefClause of
          AttributeDefClauseType.fields
    ]

  type fields = AttributeDefClauseType.fields =
    
  {
         
    f_attribute_expr: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_attribute_expr :
    [< AttributeDefClauseType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]

      
  val f_expr :
    [< AttributeDefClauseType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module AtClause : sig
  (**
   * Representation clause (``for .. use at ...;``).
   *)

  type t =
    [
      | `AtClause of
          AtClauseType.fields
    ]

  type fields = AtClauseType.fields =
    
  {
         
    f_name: [
      | `CharLiteral
          of CharLiteralType.fields
      | `Identifier
          of IdentifierType.fields
      | `StringLiteral
          of StringLiteralType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_name :
    [< AtClauseType.t]
    -> [CharLiteralType.t | IdentifierType.t | StringLiteralType.t]

      
  val f_expr :
    [< AtClauseType.t]
    -> [AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t]


end

module AspectClause : sig
  (**
   * Base class for aspect clauses.
   *)

  type t =
    [
      | AtClause.t
      | AttributeDefClause.t
      | EnumRepClause.t
      | RecordRepClause.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AspectAssoc : sig
  (**
   * Name/expression association in an aspect.
   *)

  type t =
    [
      | `AspectAssoc of
          AspectAssocType.fields
    ]

  type fields = AspectAssocType.fields =
    
  {
         
    f_id: [
      | `AttributeRef
          of AttributeRefType.fields
      | `CallExpr
          of CallExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `QualExpr
          of QualExprType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    Lazy.t;
         
    f_expr: [
      | `AbstractStateDeclExpr
          of AbstractStateDeclExprType.fields
      | `Aggregate
          of AggregateType.fields
      | `Allocator
          of AllocatorType.fields
      | `AttributeRef
          of AttributeRefType.fields
      | `BinOp
          of BinOpType.fields
      | `BracketAggregate
          of BracketAggregateType.fields
      | `BracketDeltaAggregate
          of BracketDeltaAggregateType.fields
      | `CallExpr
          of CallExprType.fields
      | `CaseExpr
          of CaseExprType.fields
      | `CharLiteral
          of CharLiteralType.fields
      | `ContractCases
          of ContractCasesType.fields
      | `DeclExpr
          of DeclExprType.fields
      | `DeltaAggregate
          of DeltaAggregateType.fields
      | `DottedName
          of DottedNameType.fields
      | `ExplicitDeref
          of ExplicitDerefType.fields
      | `Identifier
          of IdentifierType.fields
      | `IfExpr
          of IfExprType.fields
      | `IntLiteral
          of IntLiteralType.fields
      | `MembershipExpr
          of MembershipExprType.fields
      | `NullLiteral
          of NullLiteralType.fields
      | `NullRecordAggregate
          of NullRecordAggregateType.fields
      | `ParenExpr
          of ParenExprType.fields
      | `QualExpr
          of QualExprType.fields
      | `QuantifiedExpr
          of QuantifiedExprType.fields
      | `RaiseExpr
          of RaiseExprType.fields
      | `RealLiteral
          of RealLiteralType.fields
      | `RelationOp
          of RelationOpType.fields
      | `StringLiteral
          of StringLiteralType.fields
      | `TargetName
          of TargetNameType.fields
      | `UnOp
          of UnOpType.fields
      | `UpdateAttributeRef
          of UpdateAttributeRefType.fields
    ]
    option
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_id :
    [< AspectAssocType.t]
    -> [AttributeRefType.t | CallExprType.t | CharLiteralType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | QualExprType.t | StringLiteralType.t | TargetNameType.t]

      
  val f_expr :
    [< AspectAssocType.t]
    -> [AbstractStateDeclExprType.t | AllocatorType.t | AttributeRefType.t | BaseAggregateType.t | BinOpType.t | CallExprType.t | CharLiteralType.t | CondExprType.t | ContractCasesType.t | DeclExprType.t | DottedNameType.t | ExplicitDerefType.t | IdentifierType.t | MembershipExprType.t | NullLiteralType.t | NumLiteralType.t | ParenExprType.t | QualExprType.t | QuantifiedExprType.t | RaiseExprType.t | StringLiteralType.t | TargetNameType.t | UnOpType.t] option


end

module UnconstrainedArrayIndices : sig
  (**
   * Unconstrained specification for array indexes.
   *)

  type t =
    [
      | `UnconstrainedArrayIndices of
          UnconstrainedArrayIndicesType.fields
    ]

  type fields = UnconstrainedArrayIndicesType.fields =
    
  {
         
    f_types: UnconstrainedArrayIndexListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_types :
    [< UnconstrainedArrayIndicesType.t]
    -> UnconstrainedArrayIndexListType.t


end

module ConstrainedArrayIndices : sig
  (**
   * Constrained specification for array indexes.
   *)

  type t =
    [
      | `ConstrainedArrayIndices of
          ConstrainedArrayIndicesType.fields
    ]

  type fields = ConstrainedArrayIndicesType.fields =
    
  {
         
    f_list: ConstraintListType.t
    Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int



      
  val f_list :
    [< ConstrainedArrayIndicesType.t]
    -> ConstraintListType.t


end

module ArrayIndices : sig
  (**
   * Specification for array indexes.
   *)

  type t =
    [
      | ConstrainedArrayIndices.t
      | UnconstrainedArrayIndices.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AllPresent : sig
  

  type t =
    [
      | `AllPresent of
          AllPresentType.fields
    ]

  type fields = AllPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AllAbsent : sig
  

  type t =
    [
      | `AllAbsent of
          AllAbsentType.fields
    ]

  type fields = AllAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AllNode : sig
  (**
   * Qualifier for the ``all`` keyword.
   *)

  type t =
    [
      | AllAbsent.t
      | AllPresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< AllNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of AllPresent
   *)



end

module AliasedPresent : sig
  

  type t =
    [
      | `AliasedPresent of
          AliasedPresentType.fields
    ]

  type fields = AliasedPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AliasedAbsent : sig
  

  type t =
    [
      | `AliasedAbsent of
          AliasedAbsentType.fields
    ]

  type fields = AliasedAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AliasedNode : sig
  (**
   * Qualifier for the ``aliased`` keyword.
   *)

  type t =
    [
      | AliasedAbsent.t
      | AliasedPresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< AliasedNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of AliasedPresent
   *)



end

module VariantList : sig
  (**
   * List of Variant.
   *)

  type t =
    [
      | `VariantList of
          VariantListType.fields
    ]

  type fields = VariantListType.fields =
    
  {
    list : VariantType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< VariantListType.t]
    -> VariantType.t list

end

module UnconstrainedArrayIndexList : sig
  (**
   * List of UnconstrainedArrayIndex.
   *)

  type t =
    [
      | `UnconstrainedArrayIndexList of
          UnconstrainedArrayIndexListType.fields
    ]

  type fields = UnconstrainedArrayIndexListType.fields =
    
  {
    list : UnconstrainedArrayIndexType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< UnconstrainedArrayIndexListType.t]
    -> UnconstrainedArrayIndexType.t list

end

module SelectWhenPartList : sig
  (**
   * List of SelectWhenPart.
   *)

  type t =
    [
      | `SelectWhenPartList of
          SelectWhenPartListType.fields
    ]

  type fields = SelectWhenPartListType.fields =
    
  {
    list : SelectWhenPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< SelectWhenPartListType.t]
    -> SelectWhenPartType.t list

end

module PragmaNodeList : sig
  (**
   * List of Pragma.
   *)

  type t =
    [
      | `PragmaNodeList of
          PragmaNodeListType.fields
    ]

  type fields = PragmaNodeListType.fields =
    
  {
    list : PragmaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< PragmaNodeListType.t]
    -> PragmaNodeType.t list

end

module ParamSpecList : sig
  (**
   * List of ParamSpec.
   *)

  type t =
    [
      | `ParamSpecList of
          ParamSpecListType.fields
    ]

  type fields = ParamSpecListType.fields =
    
  {
    list : ParamSpecType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< ParamSpecListType.t]
    -> ParamSpecType.t list

end

module ParentList : sig
  (**
   * List of parents in a type declaration.
  
   * This list node can contain one of the following nodes:
  
   * * CharLiteralType.t
  
   * * DottedNameType.t
  
   * * IdentifierType.t
  
   * * StringLiteralType.t
   *)

  type t =
    [
      | `ParentList of
          ParentListType.fields
    ]

  type fields = ParentListType.fields =
    
  {
    list : NameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< ParentListType.t]
    -> NameType.t list

end

module NameList : sig
  (**
   * List of Name.
  
   * This list node can contain one of the following nodes:
  
   * * AttributeRefType.t
  
   * * CallExprType.t
  
   * * CharLiteralType.t
  
   * * DottedNameType.t
  
   * * ExplicitDerefType.t
  
   * * IdentifierType.t
  
   * * QualExprType.t
  
   * * StringLiteralType.t
  
   * * TargetNameType.t
   *)

  type t =
    [
      | `NameList of
          NameListType.fields
      | `ParentList of
          ParentListType.fields
    ]

  type fields = NameListType.fields =
    
  {
    list : NameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< NameListType.t]
    -> NameType.t list

end

module DiscriminantChoiceList : sig
  (**
   * List of discriminant associations.
   *)

  type t =
    [
      | `DiscriminantChoiceList of
          DiscriminantChoiceListType.fields
    ]

  type fields = DiscriminantChoiceListType.fields =
    
  {
    list : IdentifierType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< DiscriminantChoiceListType.t]
    -> IdentifierType.t list

end

module IdentifierList : sig
  (**
   * List of Identifier.
   *)

  type t =
    [
      | DiscriminantChoiceList.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< IdentifierListType.t]
    -> IdentifierType.t list

end

module ExprAlternativesList : sig
  (**
   * List of alternatives in a membership test expression.
  
   * This list node can contain one of the following nodes:
  
   * * AllocatorType.t
  
   * * AttributeRefType.t
  
   * * BaseAggregateType.t
  
   * * BinOpType.t
  
   * * CallExprType.t
  
   * * CharLiteralType.t
  
   * * CondExprType.t
  
   * * DeclExprType.t
  
   * * DiscreteSubtypeNameType.t
  
   * * DottedNameType.t
  
   * * ExplicitDerefType.t
  
   * * IdentifierType.t
  
   * * NullLiteralType.t
  
   * * NumLiteralType.t
  
   * * ParenExprType.t
  
   * * QualExprType.t
  
   * * QuantifiedExprType.t
  
   * * RaiseExprType.t
  
   * * StringLiteralType.t
  
   * * TargetNameType.t
  
   * * UnOpType.t
   *)

  type t =
    [
      | `ExprAlternativesList of
          ExprAlternativesListType.fields
    ]

  type fields = ExprAlternativesListType.fields =
    
  {
    list : ExprType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< ExprAlternativesListType.t]
    -> ExprType.t list

end

module ExprList : sig
  (**
   * List of Expr.
  
   * This list node can contain one of the following nodes:
  
   * * AllocatorType.t
  
   * * AttributeRefType.t
  
   * * BaseAggregateType.t
  
   * * BinOpType.t
  
   * * CallExprType.t
  
   * * CharLiteralType.t
  
   * * CondExprType.t
  
   * * DeclExprType.t
  
   * * DiscreteSubtypeNameType.t
  
   * * DottedNameType.t
  
   * * ExplicitDerefType.t
  
   * * IdentifierType.t
  
   * * NullLiteralType.t
  
   * * NumLiteralType.t
  
   * * ParenExprType.t
  
   * * QualExprType.t
  
   * * QuantifiedExprType.t
  
   * * RaiseExprType.t
  
   * * StringLiteralType.t
  
   * * TargetNameType.t
  
   * * UnOpType.t
   *)

  type t =
    [
      | ExprAlternativesList.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< ExprListType.t]
    -> ExprType.t list

end

module EnumLiteralDeclList : sig
  (**
   * List of EnumLiteralDecl.
   *)

  type t =
    [
      | `EnumLiteralDeclList of
          EnumLiteralDeclListType.fields
    ]

  type fields = EnumLiteralDeclListType.fields =
    
  {
    list : EnumLiteralDeclType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< EnumLiteralDeclListType.t]
    -> EnumLiteralDeclType.t list

end

module ElsifStmtPartList : sig
  (**
   * List of ElsifStmtPart.
   *)

  type t =
    [
      | `ElsifStmtPartList of
          ElsifStmtPartListType.fields
    ]

  type fields = ElsifStmtPartListType.fields =
    
  {
    list : ElsifStmtPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< ElsifStmtPartListType.t]
    -> ElsifStmtPartType.t list

end

module ElsifExprPartList : sig
  (**
   * List of ElsifExprPart.
   *)

  type t =
    [
      | `ElsifExprPartList of
          ElsifExprPartListType.fields
    ]

  type fields = ElsifExprPartListType.fields =
    
  {
    list : ElsifExprPartType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< ElsifExprPartListType.t]
    -> ElsifExprPartType.t list

end

module DiscriminantSpecList : sig
  (**
   * List of DiscriminantSpec.
   *)

  type t =
    [
      | `DiscriminantSpecList of
          DiscriminantSpecListType.fields
    ]

  type fields = DiscriminantSpecListType.fields =
    
  {
    list : DiscriminantSpecType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< DiscriminantSpecListType.t]
    -> DiscriminantSpecType.t list

end

module DefiningNameList : sig
  (**
   * List of DefiningName.
   *)

  type t =
    [
      | `DefiningNameList of
          DefiningNameListType.fields
    ]

  type fields = DefiningNameListType.fields =
    
  {
    list : DefiningNameType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< DefiningNameListType.t]
    -> DefiningNameType.t list

end

module ContractCaseAssocList : sig
  (**
   * List of ContractCaseAssoc.
   *)

  type t =
    [
      | `ContractCaseAssocList of
          ContractCaseAssocListType.fields
    ]

  type fields = ContractCaseAssocListType.fields =
    
  {
    list : ContractCaseAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< ContractCaseAssocListType.t]
    -> ContractCaseAssocType.t list

end

module CompilationUnitList : sig
  (**
   * List of CompilationUnit.
   *)

  type t =
    [
      | `CompilationUnitList of
          CompilationUnitListType.fields
    ]

  type fields = CompilationUnitListType.fields =
    
  {
    list : CompilationUnitType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< CompilationUnitListType.t]
    -> CompilationUnitType.t list

end

module CaseStmtAlternativeList : sig
  (**
   * List of CaseStmtAlternative.
   *)

  type t =
    [
      | `CaseStmtAlternativeList of
          CaseStmtAlternativeListType.fields
    ]

  type fields = CaseStmtAlternativeListType.fields =
    
  {
    list : CaseStmtAlternativeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< CaseStmtAlternativeListType.t]
    -> CaseStmtAlternativeType.t list

end

module CaseExprAlternativeList : sig
  (**
   * List of CaseExprAlternative.
   *)

  type t =
    [
      | `CaseExprAlternativeList of
          CaseExprAlternativeListType.fields
    ]

  type fields = CaseExprAlternativeListType.fields =
    
  {
    list : CaseExprAlternativeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< CaseExprAlternativeListType.t]
    -> CaseExprAlternativeType.t list

end

module BasicDeclList : sig
  (**
   * List of BasicDecl.
  
   * This list node can contain one of the following nodes:
  
   * * NumberDeclType.t
  
   * * ObjectDeclType.t
  
   * * SingleProtectedDeclType.t
  
   * * SingleTaskDeclType.t
   *)

  type t =
    [
      | `BasicDeclList of
          BasicDeclListType.fields
    ]

  type fields = BasicDeclListType.fields =
    
  {
    list : BasicDeclType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< BasicDeclListType.t]
    -> BasicDeclType.t list

end

module AssocList : sig
  (**
   * List of associations.
   *)

  type t =
    [
      | `AssocList of
          AssocListType.fields
    ]

  type fields = AssocListType.fields =
    
  {
    list : BasicAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_zip_with_params :
    ?imprecise_fallback:
    bool
    -> [< AssocListType.t ]
    -> ParamActual.t list
  (**
   * Returns an array of pairs, associating formal parameters to actual
   * expressions. The formals to match are retrieved by resolving the call which
   * this AssocList represents the actuals of.
   *)



  val f_list :
    [< AssocListType.t]
    -> BasicAssocType.t list

end

module BasicAssocList : sig
  (**
   * List of BasicAssoc.
   *)

  type t =
    [
      | AssocList.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< BasicAssocListType.t]
    -> BasicAssocType.t list

end

module BaseAssocList : sig
  (**
   * List of BaseAssoc.
   *)

  type t =
    [
      | `BaseAssocList of
          BaseAssocListType.fields
    ]

  type fields = BaseAssocListType.fields =
    
  {
    list : BaseAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< BaseAssocListType.t]
    -> BaseAssocType.t list

end

module AspectAssocList : sig
  (**
   * List of AspectAssoc.
   *)

  type t =
    [
      | `AspectAssocList of
          AspectAssocListType.fields
    ]

  type fields = AspectAssocListType.fields =
    
  {
    list : AspectAssocType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< AspectAssocListType.t]
    -> AspectAssocType.t list

end

module StmtList : sig
  (**
   * List of statements.
  
   * This list node can contain one of the following nodes:
  
   * * PragmaNodeType.t
  
   * * StmtType.t
   *)

  type t =
    [
      | `StmtList of
          StmtListType.fields
    ]

  type fields = StmtListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< StmtListType.t]
    -> AdaNodeType.t list

end

module DeclList : sig
  (**
   * List of declarations.
  
   * This list node can contain one of the following nodes:
  
   * * AbstractSubpDeclType.t
  
   * * AspectClauseType.t
  
   * * ComponentDeclType.t
  
   * * EntryDeclType.t
  
   * * ExprFunctionType.t
  
   * * NullSubpDeclType.t
  
   * * PragmaNodeType.t
  
   * * SubpDeclType.t
  
   * * SubpRenamingDeclType.t
   *)

  type t =
    [
      | `DeclList of
          DeclListType.fields
    ]

  type fields = DeclListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< DeclListType.t]
    -> AdaNodeType.t list

end

module ConstraintList : sig
  (**
   * List of constraints.
  
   * This list node can contain one of the following nodes:
  
   * * AttributeRefType.t
  
   * * BinOpType.t
  
   * * CallExprType.t
  
   * * CharLiteralType.t
  
   * * DottedNameType.t
  
   * * ExplicitDerefType.t
  
   * * IdentifierType.t
  
   * * QualExprType.t
  
   * * StringLiteralType.t
  
   * * SubtypeIndicationType.t
  
   * * TargetNameType.t
   *)

  type t =
    [
      | `ConstraintList of
          ConstraintListType.fields
    ]

  type fields = ConstraintListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< ConstraintListType.t]
    -> AdaNodeType.t list

end

module AlternativesList : sig
  (**
   * List of alternatives in a ``when ...`` clause.
  
   * This list node can contain one of the following nodes:
  
   * * AllocatorType.t
  
   * * AttributeRefType.t
  
   * * BaseAggregateType.t
  
   * * BinOpType.t
  
   * * CallExprType.t
  
   * * CharLiteralType.t
  
   * * CondExprType.t
  
   * * DeclExprType.t
  
   * * DiscreteSubtypeIndicationType.t
  
   * * DottedNameType.t
  
   * * ExplicitDerefType.t
  
   * * IdentifierType.t
  
   * * MembershipExprType.t
  
   * * NullLiteralType.t
  
   * * NumLiteralType.t
  
   * * OthersDesignatorType.t
  
   * * ParenExprType.t
  
   * * QualExprType.t
  
   * * QuantifiedExprType.t
  
   * * RaiseExprType.t
  
   * * StringLiteralType.t
  
   * * TargetNameType.t
  
   * * UnOpType.t
   *)

  type t =
    [
      | `AlternativesList of
          AlternativesListType.fields
    ]

  type fields = AlternativesListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< AlternativesListType.t]
    -> AdaNodeType.t list

end

module AbstractStateDeclList : sig
  (**
   * List of AbstractStateDecls.
  
   * This list node can contain one of the following nodes:
  
   * * AbstractStateDeclType.t
  
   * * ParenAbstractStateDeclType.t
   *)

  type t =
    [
      | `AbstractStateDeclList of
          AbstractStateDeclListType.fields
    ]

  type fields = AbstractStateDeclListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< AbstractStateDeclListType.t]
    -> AdaNodeType.t list

end

module AdaNodeList : sig
  (**
   * List of AdaNode.
  
   * This list node can contain one of the following nodes:
  
   * * AbstractStateDeclType.t
  
   * * AbstractSubpDeclType.t
  
   * * AllocatorType.t
  
   * * AspectClauseType.t
  
   * * AttributeRefType.t
  
   * * BaseAggregateType.t
  
   * * BinOpType.t
  
   * * BodyNodeType.t
  
   * * CallExprType.t
  
   * * CharLiteralType.t
  
   * * ComponentClauseType.t
  
   * * ComponentDeclType.t
  
   * * CondExprType.t
  
   * * DeclExprType.t
  
   * * DottedNameType.t
  
   * * EntryDeclType.t
  
   * * ErrorDeclType.t
  
   * * ExceptionDeclType.t
  
   * * ExceptionHandlerType.t
  
   * * ExplicitDerefType.t
  
   * * GenericDeclType.t
  
   * * GenericFormalType.t
  
   * * GenericInstantiationType.t
  
   * * GenericRenamingDeclType.t
  
   * * IdentifierType.t
  
   * * IncompleteTypeDeclType.t
  
   * * MembershipExprType.t
  
   * * NullComponentDeclType.t
  
   * * NullLiteralType.t
  
   * * NumLiteralType.t
  
   * * NumberDeclType.t
  
   * * ObjectDeclType.t
  
   * * OthersDesignatorType.t
  
   * * PackageDeclType.t
  
   * * PackageRenamingDeclType.t
  
   * * ParenAbstractStateDeclType.t
  
   * * ParenExprType.t
  
   * * PragmaNodeType.t
  
   * * ProtectedTypeDeclType.t
  
   * * QualExprType.t
  
   * * QuantifiedExprType.t
  
   * * RaiseExprType.t
  
   * * SingleProtectedDeclType.t
  
   * * SingleTaskDeclType.t
  
   * * StmtType.t
  
   * * StringLiteralType.t
  
   * * SubpDeclType.t
  
   * * SubtypeDeclType.t
  
   * * SubtypeIndicationType.t
  
   * * TargetNameType.t
  
   * * TaskTypeDeclType.t
  
   * * TypeDeclType.t
  
   * * UnOpType.t
  
   * * UseClauseType.t
  
   * * WithClauseType.t
   *)

  type t =
    [
      | `AdaNodeList of
          AdaNodeListType.fields
      | `AbstractStateDeclList of
          AbstractStateDeclListType.fields
      | `AlternativesList of
          AlternativesListType.fields
      | `ConstraintList of
          ConstraintListType.fields
      | `DeclList of
          DeclListType.fields
      | `StmtList of
          StmtListType.fields
    ]

  type fields = AdaNodeListType.fields =
    
  {
    list : AdaNodeType.t list Lazy.t;
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




  val f_list :
    [< AdaNodeListType.t]
    -> AdaNodeType.t list

end

module AdaList : sig
  

  type t =
    [
      | AdaNodeList.t
      | AspectAssocList.t
      | BaseAssocList.t
      | BasicAssocList.t
      | BasicDeclList.t
      | CaseExprAlternativeList.t
      | CaseStmtAlternativeList.t
      | CompilationUnitList.t
      | ContractCaseAssocList.t
      | DefiningNameList.t
      | DiscriminantSpecList.t
      | ElsifExprPartList.t
      | ElsifStmtPartList.t
      | EnumLiteralDeclList.t
      | ExprList.t
      | IdentifierList.t
      | NameList.t
      | ParamSpecList.t
      | PragmaNodeList.t
      | SelectWhenPartList.t
      | UnconstrainedArrayIndexList.t
      | VariantList.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AbstractPresent : sig
  

  type t =
    [
      | `AbstractPresent of
          AbstractPresentType.fields
    ]

  type fields = AbstractPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AbstractAbsent : sig
  

  type t =
    [
      | `AbstractAbsent of
          AbstractAbsentType.fields
    ]

  type fields = AbstractAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AbstractNode : sig
  (**
   * Qualifier for the ``abstract`` keyword.
   *)

  type t =
    [
      | AbstractAbsent.t
      | AbstractPresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< AbstractNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of AbstractPresent
   *)



end

module AbortPresent : sig
  

  type t =
    [
      | `AbortPresent of
          AbortPresentType.fields
    ]

  type fields = AbortPresentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AbortAbsent : sig
  

  type t =
    [
      | `AbortAbsent of
          AbortAbsentType.fields
    ]

  type fields = AbortAbsentType.fields =
    
  {
    c_value : Entity.t;
    context : AnalysisContext.t
  }


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int




end

module AbortNode : sig
  (**
   * Qualifier for the ``abort`` keyword.
   *)

  type t =
    [
      | AbortAbsent.t
      | AbortPresent.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int


      
  val p_as_bool :
    [< AbortNodeType.t ]
    -> bool
  (**
   * Return whether this is an instance of AbortPresent
   *)



end

module AdaNode : sig
  (**
   * Root node class for the Ada syntax tree.
   *)

  type t =
    [
      | AbortNode.t
      | AbstractNode.t
      | AdaList.t
      | AliasedNode.t
      | AllNode.t
      | ArrayIndices.t
      | AspectAssoc.t
      | AspectClause.t
      | AspectSpec.t
      | BaseAssoc.t
      | BaseFormalParamHolder.t
      | BaseRecordDef.t
      | BasicAssoc.t
      | BasicDecl.t
      | CaseStmtAlternative.t
      | CompilationUnit.t
      | ComponentClause.t
      | ComponentDef.t
      | ConstantNode.t
      | Constraint.t
      | DeclarativePart.t
      | ElsifExprPart.t
      | ElsifStmtPart.t
      | Expr.t
      | HandledStmts.t
      | InterfaceKind.t
      | IterType.t
      | LibraryItem.t
      | LimitedNode.t
      | LoopSpec.t
      | Mode.t
      | MultiAbstractStateDecl.t
      | NotNull.t
      | NullComponentDecl.t
      | OthersDesignator.t
      | OverridingNode.t
      | Params.t
      | ParenAbstractStateDecl.t
      | PragmaNode.t
      | PrivateNode.t
      | ProtectedDef.t
      | ProtectedNode.t
      | Quantifier.t
      | RangeSpec.t
      | RenamingClause.t
      | ReverseNode.t
      | SelectWhenPart.t
      | Stmt.t
      | SubpKind.t
      | Subunit.t
      | SynchronizedNode.t
      | TaggedNode.t
      | TaskDef.t
      | TypeDef.t
      | TypeExpr.t
      | UnconstrainedArrayIndex.t
      | UntilNode.t
      | UseClause.t
      | Variant.t
      | VariantPart.t
      | WithClause.t
      | WithPrivate.t
    ]


  val equal : [< t] -> [< t] -> bool

  val hash : [< t] -> int

  val compare : [< t] -> [< t] -> int

  val kind_name : [< AdaNodeType.t] -> string
  (**
   * Return the kind of this node.
   *)

  val text : [< AdaNodeType.t ] -> string
  (**
   * Return the source buffer slice corresponding to the text that spans
   * between the first and the last tokens of the given node.
   *)

  val image : [< AdaNodeType.t ] -> string
  (**
   * Return a representation of this node as a string.
   *)

  val sloc_range : [< AdaNodeType.t ] -> SlocRange.t
  (**
   * Return the spanning source location range for this node.
  
   * Note that this returns the sloc of the parent for synthetic nodes.
   *)

  val lookup : [< AdaNodeType.t ] -> Sloc.t -> AdaNodeType.t option
  (**
   * Return the bottom-most node from in ``Node`` and its children which
   * contains ``Sloc``, or ``None`` if there is none.
   *)

  
  val fold_tokens : ('a -> Token.t -> 'a) -> 'a -> [< AdaNodeType.t] -> 'a
  (**
   * Fold all the token this node contains by calling f on each token.
   *)

  val iter_tokens : (Token.t -> unit) -> [< AdaNodeType.t]-> unit
  (**
   * Iterate over all token this node contains by calling f on each token.
   *)

  val map_tokens : (Token.t -> 'a) -> [< AdaNodeType.t] -> 'a list
  (**
   * Map each token calling the given function
   *)

  val tokens : [< AdaNodeType.t] -> Token.t list
  (**
   * Return a list of tokens for the given node
   *)


  val entity_image : [< AdaNodeType.t ] -> string

  val children_opt : [< AdaNodeType.t ] -> AdaNodeType.t option list
  (**
   * Return an optional list of nodes which are the children of the given node.
   * Each child is optional because it can either be because of a syntax error,
   * or an optional field evaluated to null.
   *)

  val fold_fields :
    ('a -> AdaNodeType.t -> 'a) -> 'a -> [< AdaNodeType.t ] -> 'a
  (**
   * Fold all fields of the given node. This skips any child that is None
   * because of a syntax error or because the field is optional
   *)

  val iter_fields :
    (AdaNodeType.t -> unit) -> [< AdaNodeType.t ] -> unit
  (**
   * Iter all fields of the given node. This skips any child that is None
   * because of a syntax error or because the field is optional
   *)

  val exists_fields :
    (AdaNodeType.t -> bool) -> [< AdaNodeType.t ] -> bool
  (**
   * Fold all fields of the given node. Return true if the given predicate is
   * evaluated to true for at least one node.
   *)

  val for_all_fields :
    (AdaNodeType.t -> bool) -> [< AdaNodeType.t ] -> bool
  (**
   * Fold all fields of the given node. Return true if the given predicate is
   * evaluated to true for all nodes.
   *)

  val fold :
    ('a -> AdaNodeType.t -> 'a) -> 'a -> [< AdaNodeType.t ] -> 'a
  (**
   * Fold the entire AST, below the given node, and call the given function on
   * each node in prefix order.
   *)

  val iter :
    (AdaNodeType.t -> unit) -> [< AdaNodeType.t ] -> unit
  (**
   * Iterate over the entire AST, below the given node, and call the given
   * function on each node in prefix order.
   *)

  val filter :
    (AdaNodeType.t -> bool)
    -> [< AdaNodeType.t ]
    -> AdaNodeType.t list
  (**
   * Fold the entire AST, below the given node, and return the list of node
   * evaluated to true by the given function
   *)

  val exists :
    (AdaNodeType.t -> bool) -> [< AdaNodeType.t ] -> bool
  (**
   * Fold the entire AST, below the given node, and return true if the given
   * predicate is evaluated to true by at least one node.
   *)

  val for_all :
    (AdaNodeType.t -> bool) -> [< AdaNodeType.t ] -> bool
  (**
   * Fold the entire AST, below the given node, and return true if the given
   * predicate is evaluated to true for all nodes.
   *)

  val lookup_with_kind :
    'a node -> [< AdaNodeType.t] -> Sloc.t -> 'a option
  (**
   * Given the kind of a node, a source location and a node, return the deepest
   * node containing the source location and of the right kind. Returns None if
   * there is no match.
   *)

  val as_a : 'a node -> [< AdaNodeType.t ] -> 'a option
  (**
   * Given the kind of a node, try to cast the given node to this kind. Return
   * None if the node is not of this type and thus, cannot be cast.
   *)

  val find : 'a node -> [< AdaNodeType.t ] -> 'a
  (**
   * Given the kind of node, return the first node found by walking the given
   * node. The type of the resulting node depends on the desired kind
   *)


  val findall : 'a node -> [< AdaNodeType.t ] -> 'a list
  (**
   * Given the kind of node, return the all nodes of this kind found by walking
   * the given node. The type of the resulting nodes depends on the desired
   * kind
   *)

  val fields_with_names :
    [< AdaNodeType.t ] -> (string * AdaNodeType.t option) list
  (**
   * Given a node, return the list of it's fields, together with the name of
   * the field. This function does not raise SyntaxError, but instead the
   * returned node is None.
   *)

  val pp_tree : Format.formatter -> [< AdaNodeType.t] -> unit
  (**
   * Pretty print the whole tree by completely walking it.
   *)


      
  val p_declarative_scope :
    [< AdaNodeType.t ]
    -> DeclarativePartType.t option
  (**
   * Return the scope of definition of this basic declaration.
   *)

      
  val p_complete :
    [< AdaNodeType.t ]
    -> unit
  (**
   * Return possible completions at this point in the file.
   *)

      
  val p_valid_keywords :
    [< AdaNodeType.t ]
    -> string list
  (**
   * Return the list of keywords that are valid at this point in the file.
  
   * .. note:: This is work in progress. It will return all keywords for now,
   *    without looking at the context.
   *)

      
  val p_generic_instantiations :
    [< AdaNodeType.t ]
    -> GenericInstantiationType.t list
  (**
   * Return the potentially empty list of generic package/subprogram
   * instantiations that led to the creation of this entity. Outer-most
   * instantiations appear last.
   *)

      
  val p_semantic_parent :
    [< AdaNodeType.t ]
    -> AdaNodeType.t option
  (**
   * Return the semantic parent for this node, if applicable, null otherwise.
  
   * .. note:: A node lying outside of a library item's declaration or subunit's
   *    body does not have a parent environment, meaning that this property will
   *    return null.
   *)

      
  val p_parent_basic_decl :
    [< AdaNodeType.t ]
    -> BasicDeclType.t option
  (**
   * Return the parent basic decl for this node, if applicable, null otherwise.
  
   * .. note:: If the parent BasicDecl of the given node is a generic
   *    declaration, this call will return the instantiation from which the node
   *    was retrieved instead, if any.
   *)

      
  val p_filter_is_imported_by :
    [< AdaNodeType.t ]
    -> AnalysisUnit.t list
    -> bool
    -> AnalysisUnit.t list
  (**
   * Filters out among the list of given units those that cannot refer to the
   * unit in which this node lies. If transitive is True, the whole transitive
   * closure of imports will be used to find a reference to the unit of this
   * node.
   *)

      
  val p_xref_entry_point :
    [< AdaNodeType.t ]
    -> bool
  (**
   * Designates entities that are entry point for the xref solving
   * infrastructure. If this returns true, then resolve_names can be called on
   * it.
  
   * .. note:: For convenience, and unlike what is defined in the ARM wrt.
   *    complete contexts for name resolution, ``xref_entry_points`` can be
   *    nested.
   *)

      
  val p_resolve_names :
    [< AdaNodeType.t ]
    -> bool
  (**
   * This will resolve names for this node. If the operation is successful, then
   * type_var and ref_var will be bound on appropriate subnodes of the
   * statement.
   *)

      
  val p_standard_unit :
    [< AdaNodeType.t ]
    -> AnalysisUnit.t
  (**
   * Static method. Return the analysis unit corresponding to the Standard
   * package.
   *)

      
  val p_std_entity :
    [< AdaNodeType.t ]
    -> string
    -> AdaNodeType.t option
  (**
   * Static property. Return an entity from the standard package with name
   * `sym`.
   *)

      
  val p_bool_type :
    [< AdaNodeType.t ]
    -> AdaNodeType.t option
  (**
   * Static method. Return the standard Boolean type.
   *)

      
  val p_int_type :
    [< AdaNodeType.t ]
    -> AdaNodeType.t option
  (**
   * Static method. Return the standard Integer type.
   *)

      
  val p_universal_int_type :
    [< AdaNodeType.t ]
    -> AdaNodeType.t option
  (**
   * Static method. Return the standard Universal Integer type.
   *)

      
  val p_universal_real_type :
    [< AdaNodeType.t ]
    -> AdaNodeType.t option
  (**
   * Static method. Return the standard Universal Real type.
   *)

      
  val p_top_level_decl :
    [< AdaNodeType.t ]
    -> AnalysisUnit.t
    -> BasicDeclType.t option
  (**
   * Static method. Get the top-level decl in ``unit``.  This is the body of a
   * Subunit, or the item of a ``LibraryItem``.
   *)

      
  val p_choice_match :
    [< AdaNodeType.t ]
    -> BigInteger.t
    -> bool
  (**
   * Assuming that self is a choice expression (such as what can appear in an
   * alternative of a case statement or in the RHS of a membership expression,
   * this property returns whether the given value satisfies it.
  
   * .. ATTENTION:: This is an experimental feature, so even if it is exposed to
   *    allow experiments, it is totally unsupported and the API and behavior
   *    are very likely to change in the future.
   *)

      
  val p_gnat_xref :
    ?imprecise_fallback:
    bool
    -> [< AdaNodeType.t ]
    -> DefiningNameType.t option
  (**
   * Return a cross reference from this name to a defining identifier, trying to
   * mimic GNAT's xrefs as much as possible.
   *)

      
  val parent :
    [< AdaNodeType.t ]
    -> AdaNodeType.t option
  (**
   * Return the syntactic parent for this node. Return null for the root node.
   *)

      
  val parents :
    ?with_self:
    bool
    -> [< AdaNodeType.t ]
    -> AdaNodeType.t list
  (**
   * Return an array that contains the lexical parents, this node included iff
   * ``with_self`` is True. Nearer parents are first in the list.
   *)

      
  val children :
    [< AdaNodeType.t ]
    -> AdaNodeType.t list
  (**
   * Return an array that contains the direct lexical children.
   *)

      
  val token_start :
    [< AdaNodeType.t ]
    -> Token.t
  (**
   * Return the first token used to parse this node.
   *)

      
  val token_end :
    [< AdaNodeType.t ]
    -> Token.t
  (**
   * Return the last token used to parse this node.
   *)

      
  val child_index :
    [< AdaNodeType.t ]
    -> int
  (**
   * Return the 0-based index for Node in its parent's children.
   *)

      
  val previous_sibling :
    [< AdaNodeType.t ]
    -> AdaNodeType.t option
  (**
   * Return the node's previous sibling, or null if there is no such sibling.
   *)

      
  val next_sibling :
    [< AdaNodeType.t ]
    -> AdaNodeType.t option
  (**
   * Return the node's next sibling, or null if there is no such sibling.
   *)

      
  val unit :
    [< AdaNodeType.t ]
    -> AnalysisUnit.t
  (**
   * Return the analysis unit owning this node.
   *)

      
  val is_ghost :
    [< AdaNodeType.t ]
    -> bool
  (**
   * Return whether the node is a ghost.
  
   * Unlike regular nodes, ghost nodes cover no token in the input source: they
   * are logically located instead between two tokens. The "token_first" of all
   * ghost nodes is the token right after this logical position, while they have
   * no "token_last".
   *)

      
  val full_sloc_image :
    [< AdaNodeType.t ]
    -> string
  (**
   * Return a string containing the filename + the sloc in GNU conformant
   * format. Useful to create diagnostics from a node.
   *)



end


